/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.acquisition.aquisitionserverside.interfaces.impl;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionServerException;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DiskModelDescription;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DiskModelUtil;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DiskModelWrapper;
import com.netapp.acquisition.aquisitionserverside.interfaces.remote.DiskModelRemote;
import com.netapp.oci.platform.common.ServerConstant;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.table.TableResultParserDefault;
import org.apache.log4j.Logger;
import org.jboss.ejb3.annotation.RemoteBinding;
import org.jboss.ejb3.annotation.RemoteBindings;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import java.io.*;
import java.sql.Connection;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

@SuppressWarnings({"ConstantConditions", "UnusedDeclaration"})
@Stateless
@RemoteBindings({
                        @RemoteBinding(jndiBinding = DiskModelBean.JNDI_NAME),
                        @RemoteBinding(jndiBinding = ServerConstant.HTTPS_JNDI_BINDING_PREFIX + DiskModelBean.JNDI_NAME)})
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"admin", "acquisition"})
public class DiskModelBean implements DiskModelRemote {
    private static final Logger logger = Logger.getLogger("com.netapp.acquisition.aquisitionserverside.interfaces.impl.AcquisitionServerSessionBean");

    private static final int FIVE_MINUTES_IN_MS = 5 * 60 * 1000;

    private static final String JBOSS_DIR = System.getProperty("jboss.server.home.dir", "C:/Program Files/SANscreen/jboss/server/onaro/");
    private static final String DATASOURCE_FOLDER = new File(JBOSS_DIR, "deploy/datasources.war/").getAbsolutePath();
    private static final String UPLOAD_FOLDER = new File(JBOSS_DIR, "deploy/upload.war").getAbsolutePath();
    private static final String REMOTE_DISK_MODEL_NAME = "disk_models." + DiskModelUtil.DISK_MODEL_FORMAT_CURRENT_VERSION + ".txt";
    private static final String REMOTE_DISK_MODEL_PATH = "/diskmodels/" + REMOTE_DISK_MODEL_NAME;
    private static final String REMOTE_DISK_MODEL_CACHE = new File(UPLOAD_FOLDER, REMOTE_DISK_MODEL_NAME).getAbsolutePath();


    private static volatile long leaseStartTime = 0;                     // lease time for the combined disk model
    private static volatile String remoteDiskModelLastModifyTime = null; // a string as returned by the remote server
    private static volatile long customDiskModelLastCheckTime = 0;       // new_disk_models.txt lease start time
    private static volatile long customDiskModelLastModifyTime = 0;

    // The current disk model that is cached in memory and returned to the acq units when they request it.
    // It may be updated in a separate thread - when the acq unit renews its lease
    private static volatile DiskModelWrapper diskModelCache = null;

    /**
     * used for a connection to the database. It is passed to the updates checker class to retrieve proxy settings
     */
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private javax.sql.DataSource dataSource;

    /**
     * gets the disk model
     *
     * @return a DiskModelWrapper that contains disk models and the timestamp of the latest disk model update
     */
    public DiskModelWrapper getDiskModel() {
        return getDiskModelIfNewer(-1);
    }

    /**
     * gets the disk model if it is newer than <code>acqDiskModelTimeStampMillis</code>
     *
     * @param acqDiskModelTimeStampMillis <= 0 to always get the latest or time stamp in millis
     *
     * @return a DiskModelWrapper or null
     */
    public DiskModelWrapper getDiskModelIfNewer(final long acqDiskModelTimeStampMillis) {

        final long lastDiskModelUpdateTimeMillis = diskModelCache != null ? diskModelCache.getUpdateTimeStamp() : -1;
        if (diskModelCache == null || (acqDiskModelTimeStampMillis <= 0) || acqDiskModelTimeStampMillis < lastDiskModelUpdateTimeMillis) {
            if (diskModelCache == null) {
                updateDiskModel();
            }
            return diskModelCache;
        }
        else {
            return null;
        }
    }

    /**
     * updates the single instance of the disk model cache if new configuration files are available.
     *
     * @throws AcquisitionServerException if disk model configurations can not be read or parsed
     */
    public void updateDiskModel() {
        java.sql.Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            updateDiskModel(connection);
        }
        catch (AcquisitionServerException e) {
            logger.error("Error updating disk model: " + e.getMessage(), e);
        }
        finally {
            ConnectionHelper.close(connection);
        }

    }

    /**
     * updates the single instance of the disk model cache if new content is available
     *
     * @throws AcquisitionServerException if disk model configurations can not be read or parsed
     */
    protected void updateDiskModel(final Connection connection) throws AcquisitionServerException {

        boolean checkForUpdatedContents = false;
        if (diskModelCache == null) {
            // if the cache is empty always update
            checkForUpdatedContents = true;
        }

        final long now = System.currentTimeMillis();
        final long leaseTimeMillis = getDiskModelLeaseTimeMillis(connection);

        // force check the local custom disk model file every 5 minutes so the server doesn't have to restart to pick up changes
        boolean customDiskModelIsNew = false;
        SortedMap<String, DiskModelDescription> customDiskModelMap = null;
        final boolean customDiskModelLeaseExpired = (now - customDiskModelLastCheckTime) > FIVE_MINUTES_IN_MS;

        if (checkForUpdatedContents || customDiskModelLeaseExpired) {
            final File customDiskModelFile = new File(DATASOURCE_FOLDER, DiskModelUtil.DISK_DESCR_CUSTOM_STORE);
            if (customDiskModelFile.exists()) {
                // add custom disk models
                customDiskModelIsNew = (customDiskModelFile.lastModified() > customDiskModelLastModifyTime);
                if (customDiskModelIsNew || checkForUpdatedContents) {
                    try {
                        customDiskModelMap = DiskModelUtil.parseDiskModelDescription(getContents(customDiskModelFile));
                    }
                    catch (Exception e) {
                        throw new AcquisitionServerException("Error reading custom disk model configuration: " + DiskModelUtil.DISK_DESCR_CUSTOM_STORE, e);
                    }

                }
                if (customDiskModelIsNew) {
                    checkForUpdatedContents = true;
                }
                customDiskModelLastCheckTime = now;
                customDiskModelLastModifyTime = customDiskModelFile.lastModified();
            }
        }

        if (!checkForUpdatedContents) {
            if (logger.isDebugEnabled()) {
                logger.debug("Disk Model - skipping update because no contents changed");
            }
            return;
        }

        //leases have expired (or the cache is empty) so check if the disk model needs to be updated - always update the current least time
        final SortedMap<String, DiskModelDescription> newDiskModelCache = reloadCacheIfNecessary(customDiskModelMap, customDiskModelIsNew);
        if (logger.isDebugEnabled()) {
            logger.debug("Disk Model - new content size: " + ((newDiskModelCache == null) ? null : newDiskModelCache.size()) + " new lease start time : "  + leaseStartTime);
        }

        leaseStartTime = System.currentTimeMillis();  // there may be no disk model changes, but we checked so update the lease time
        if (newDiskModelCache != null) {
            synchronized (getClass()) {
                diskModelCache = new DiskModelWrapper(newDiskModelCache, leaseStartTime);
            }
        }


    }

    /**
     * gets the configured time to wait before checking for a newer disk model file on the phone home server
     *
     * get the value from the param table
     *
     * @param connection the connection to the database
     *
     * @return the time in ms
     */
    private long getDiskModelLeaseTimeMillis(final Connection connection) {

        long leaseTimeHours;
        try {
            leaseTimeHours = getParamLong(connection, "DISK_MODEL_LEASE_HOURS", DiskModelUtil.DEFAULT_DISK_MODEL_LEASE_TIME_HOURS);
        }
        catch (Exception e) {
            leaseTimeHours = DiskModelUtil.DEFAULT_DISK_MODEL_LEASE_TIME_HOURS;
        }
        final long leaseTimeMillis;
        if (leaseTimeHours < 0) {
            leaseTimeMillis = -1L;
        }
        else if (leaseTimeHours == 0) {
            // we don't want to check gate.onaro.com on every poll
            // this is a magic setting for 5 minutes  - great for testing :)
            leaseTimeMillis = FIVE_MINUTES_IN_MS;
        }
        else {
            leaseTimeMillis = leaseTimeHours * 60 * 60 * 1000; // convert to MilliSeconds
        }

        return leaseTimeMillis;
    }

    /**
     * gets the latest disk models from gate.onaro.com and custom disk models.
     * If the disk model file can not be retrieved from gate.onaro.com, the cache of that file is used. If there is no cache,
     * the default disk model in the .jar is used. Then, any custom disk models are added.
     *
     * @return null if there are no disk model updates. Otherwise a sorted map of disk default and custom disk model descriptions.
     *
     * @throws AcquisitionServerException if disk model configurations can not be read or parsed
     */
    private SortedMap<String, DiskModelDescription> reloadCacheIfNecessary(final SortedMap<String, DiskModelDescription> customDiskModels, final boolean customDiskModelIsNew) throws AcquisitionServerException {
        // if the local cache is up to date an there

        final SortedMap<String, DiskModelDescription> diskModelDescriptions = new TreeMap<String, DiskModelDescription>();
        // if the gate.onaro.com disk model is newer or the custom disk model is newer then
        // parse them and return a new cache

        // check custom disk model
        boolean remoteDiskModelIsNew = false;

        // update if the cache has not been initialized, there is an updated version on gate.onaro.com, or in the local custom disk model
        final boolean updateCash = diskModelCache == null ||
                                   remoteDiskModelIsNew ||
                                   customDiskModelIsNew;

        if (logger.isDebugEnabled()) {
            logger.debug("Disk Model - updateCache = " + updateCash +
                         " (customDiskModelIsNew = " + customDiskModelIsNew +
                         ", customDiskModel = " + ((customDiskModels == null) ? null : customDiskModels.size()) +
                         ", diskModelCache = " + ((diskModelCache == null) ? null : diskModelCache.getDescriptions().size()) + ")");
        }
        if (!updateCash) {
            return null;
        }

        diskModelDescriptions.putAll(getLatestOrDefaultDiskModels());

        // add custom disk models if there are any
        if (customDiskModels != null) {
            diskModelDescriptions.putAll(customDiskModels);
        }

        return diskModelDescriptions;

    }

    private void writeDiskModelCache(final File latestDiskModelCacheFile, final String diskModelContents) throws IOException {
        FileWriter fw = null;
        try {
            final File parentDir = latestDiskModelCacheFile.getParentFile();
            if (!parentDir.exists()) {
                if (!parentDir.mkdirs()) {
                    throw new IOException("failed to create path for disk model cache: " + parentDir);
                }
            }
            fw = new FileWriter(latestDiskModelCacheFile);
            fw.write(diskModelContents);
        }
        finally {
            if (fw != null) {
                fw.close();
            }
        }
    }

    /**
     * gets, in the following order, disk models from gate.onaro.com, the latest cache from gate.onaro.com,
     * or the default disk models in the version.
     *
     * @return A map of disk model descriptions. Never returns null.
     *
     * @throws AcquisitionServerException if disk model configurations can not be read or parsed
     */
    private SortedMap<String, DiskModelDescription> getLatestOrDefaultDiskModels() throws AcquisitionServerException {
        // return the disk models based on the following algorithm
        // 1) use the contents of the downloaded disk model
        // 2) use the cache of the latest downloaded disk model
        // 3) use the default disk model in the jar
        final File diskModelCacheFile = new File(REMOTE_DISK_MODEL_CACHE);

        // if getting/parsing the downloaded disk models fails, use the cached version
        if (diskModelCacheFile.exists() && diskModelCacheFile.canRead()) {
            try {
                final SortedMap<String, DiskModelDescription> diskModelDescriptions =
                        DiskModelUtil.parseDiskModelDescriptionStream(new FileInputStream(diskModelCacheFile), diskModelCacheFile.getCanonicalPath(), false);
                if (diskModelDescriptions.isEmpty()) {
                    throw new AcquisitionServerException("the disk model cache contains no valid disk models");
                }
                return diskModelDescriptions;
            }
            catch (Exception e) {
                logger.warn("failed to parse the cached disk model", e);
            }
        }

        // if getting/parsing the downloaded disk models fails and the cache doesn't exist/failed, the use the default version
        return getDefaultDiskModels();
    }

    /**
     * gets the default disk models in the version
     *
     * @return A map of disk model descriptions
     *
     * @throws AcquisitionServerException if disk model configurations can not be read or parsed
     */
    private SortedMap<String, DiskModelDescription> getDefaultDiskModels() throws AcquisitionServerException {

        // populate the map with either the latest disk model file (if available), or the default disk model
        // then add any custom disk model entries
        // add disk models from the release (diskmodel.jar)
        // Download the SANscreen defined default disk model description
        final SortedMap<String, DiskModelDescription> diskModels = new TreeMap<String, DiskModelDescription>();

        Exception exc = null;
        final File diskModelFile = new File(DATASOURCE_FOLDER, "diskmodel.jar");
        try {
            final JarFile diskModelJarFile = new JarFile(diskModelFile);
            final InputStream defaultDiskDescriptionStream = diskModelJarFile.getInputStream(new ZipEntry(DiskModelUtil.DISK_DESCR_DEFAULT_STORE));
            diskModels.putAll(DiskModelUtil.parseDiskModelDescriptionStream(defaultDiskDescriptionStream, DiskModelUtil.DISK_DESCR_DEFAULT_STORE, false));
        }
        catch (Exception e) {
            exc = e;
        }
        if (diskModels.isEmpty() || exc != null) {
            throw new AcquisitionServerException("Error reading default disk model file: " + diskModelFile.getAbsolutePath(), exc);
        }
        return diskModels;
    }

    /**
     * gets a string from the param table
     *
     * @param connection the connection to the database
     * @param param      the parameter name
     *
     * @return the value of the parameter name or null if the parameter doesn't exist or is empty
     */
    private String getParamString(final java.sql.Connection connection, final String param) {
        try {
            final String value = PreparedStatementWrapper.create(connection, "SELECT value FROM sanscreen.param WHERE param='" + param + "'", TableResultParserDefault.STRING).executeQueryEx().iterator().next();
            if ((value != null) && !value.isEmpty()) {
                return value;
            }
        }
        catch (TableException e) {
            // ignore;
        }
        return null;
    }

    /**
     * gets a long value from the param table for the parameter named <code>param</code>
     *
     * @param connection   the connection to the database
     * @param param        the name of the parameter
     * @param defaultValue the value to use if the param does not exist or could not be retrieved
     *
     * @return
     */
    private long getParamLong(final java.sql.Connection connection, final String param, final long defaultValue) {
        try {
            return PreparedStatementWrapper.executeFunctionQuery(connection, "SELECT value FROM sanscreen.param WHERE param='" + param + "'", defaultValue);
        }
        catch (TableException e) {
            return defaultValue;
        }
    }

    private String getContents(final File file) throws IOException {
        final Writer writer = new StringWriter();
        Reader reader = null;
        try {
            reader = new BufferedReader(new InputStreamReader(new FileInputStream(file)));
            final char[] buffer = new char[1024];
            int n;
            while ((n = reader.read(buffer)) != -1) {
                writer.write(buffer, 0, n);
            }
        }
        finally {
            if (reader != null) {
                try {
                    reader.close();
                }
                catch (IOException e) {
                    // ignore
                }
            }
        }
        return writer.toString();
    }
}