/*
 * $Id: //depot/prod/newport_news/main/modules/crypto-util/src/main/java/com/netapp/crypto/SymmetricCipherDelegate.java#1 $
 *
 * Copyright (c) 2013 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.crypto;

import java.io.IOException;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.util.concurrent.Callable;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.KeyGenerator;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;



/**
 * A delegate used by the {@link com.netapp.crypto.SymmetricCipherFacade} to perform
 * <ul>
 * <li>encryption</li>
 * <li>decryption</li>
 * <li>key generation</li>
 * </ul>
 */
abstract class SymmetricCipherDelegate {

    /**
     * Unique identifier for this delegate.
     * Must be unique within the {@link com.netapp.crypto.SymmetricCipherFacade}.
     */
    private final int id;

    /**
     * Transformation used for the cipher (e.g. {@link Aes128CbcPkcs5PaddingCipherDelegate#TRANSFORMATION}).
     */
    private final String cipherTransformation;
    /**
     * Length in bits of the initialization vector.
     */
    private final int cipherIvBitLength;
    /**
     * Algorithm used for key generation (e.g. {@link Aes128CbcPkcs5PaddingCipherDelegate#KEY_GENERATION_ALGORITHM}).
     */
    private final String keyGeneratorAlgorithm;
    /**
     * Length in bits of the key.
     */
    private final int keyBitLength;

    /**
     * PRNG used to generate IVs and secret keys.
     * <p>
     * {@link ThreadLocal} is used because reusing the same {@link SecureRandom}
     * in multiple threads can cause a lot of contention.
     * <p>
     */
    private final ThreadLocal<SecureRandom> secureRandomLocal;

    /**
     * Cipher used for encryption/decryption.
     * <p>
     * Ciphers are expensive to construct, and are not threadsafe,
     * so use a ThreadLocal to construct a max of one per thread.
     */
    private final ThreadLocal<Cipher> cipherLocal = new ThreadLocal<Cipher>() {
        @Override
        protected Cipher initialValue() {
            try {
                return Cipher.getInstance(cipherTransformation);
            } catch (NoSuchAlgorithmException e) {
                throw new CryptoException(e);
            } catch (NoSuchPaddingException e) {
                throw new CryptoException(e);
            }
        }
    };

    /**
     * Key generator.
     * <p>
     *
     * KeyGenerators are expensive to construct, and are not threadsafe,
     * so use a ThreadLocal to construct a max of one per thread.
     */
    private final ThreadLocal<KeyGenerator> keyGenerator = new ThreadLocal<KeyGenerator>() {
        @Override
        protected KeyGenerator initialValue() {
            try {
                KeyGenerator generator = KeyGenerator.getInstance(keyGeneratorAlgorithm);
                generator.init(keyBitLength, secureRandomLocal.get());
                return generator;
            } catch (NoSuchAlgorithmException e) {
                throw new CryptoException(e);
            }
        }
    };

    protected SymmetricCipherDelegate(
            int id,
            String cipherTransformation,
            int cipherIvBitLength,
            String keyGeneratorAlgorithm,
            int keyBitLength,
            final Callable<SecureRandom> secureRandomFactory) {

        super();
        this.id = id;
        this.cipherTransformation = cipherTransformation;
        this.cipherIvBitLength = cipherIvBitLength;
        this.keyGeneratorAlgorithm = keyGeneratorAlgorithm;
        this.keyBitLength = keyBitLength;
        this.secureRandomLocal = new ThreadLocal<SecureRandom>() {
            @Override
            protected SecureRandom initialValue() {
                try {
                    return secureRandomFactory.call();
                } catch (Exception e) {
                    throw new CryptoException("Unable to create SecureRandom", e);
                }
            }
        };
    }

    public int getId() {
        return this.id;
    }

    protected Cipher getCipher() {
        return this.cipherLocal.get();
    }

    protected int getCipherIvBitLength() {
        return this.cipherIvBitLength;
    }

    protected KeyGenerator getKeyGenerator() {
        return this.keyGenerator.get();
    }

    protected SecureRandom getSecureRandom() {
        return this.secureRandomLocal.get();
    }

    /**
     * Generates a key that can be used for encryption/decryption this delegate.
     */
    public SecretKey generateKey() {
        return getKeyGenerator().generateKey();
    }

    /**
     * Encrypts the given plaintext with the given key.
     * <p>
     * The returned value will contain the resulting ciphertext,
     * as well as any parameters needed during decryption (such as an initialization vector).
     * <p>
     * The return value will be in a format that can be passed to the
     * {@link #decrypt(SecretKey, byte[])} for decryption.
     */
    public abstract byte[] encrypt(SecretKey key, byte[] plainText) throws InvalidKeyException, InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException, IOException;

    /**
     * Decrypts the given ciphertext with the given key.
     * <p>
     * The ciphertext must be in a format that was returned by
     * {@link #encrypt(SecretKey, byte[])}.
     */
    public abstract byte[] decrypt(SecretKey key, byte[] cipherText) throws IllegalBlockSizeException, BadPaddingException, IOException, InvalidKeyException, InvalidAlgorithmParameterException;
}
