/*
 * $Id: //depot/prod/newport_news/main/modules/crypto-util/src/main/java/com/netapp/crypto/SymmetricCipherFacade.java#1 $
 *
 * Copyright (c) 2010 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.crypto;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.cert.CertificateException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;

import net.jcip.annotations.ThreadSafe;
import org.apache.commons.codec.binary.Base64InputStream;
import org.apache.commons.codec.binary.Base64OutputStream;
import org.apache.commons.io.IOUtils;

import com.netapp.crypto.KeyRotator.KeyEntry;

/**
 * Facade for encrypting/decrypting using a symmetric cipher.
 * <p>
 * You just need to provide a {@link KeyStoreWrapper}, and this class will do all the rest, including:
 * <ol>
 * <li>Generating keys</li>
 * <li>Using secure algorithms</li>
 * <li>Transparently handling backwards compatibility if a more secure algorithm needs to be used.</li>
 * <li>Support key rotation (See {@link #generateNewKey()} to generate a new encryption key)</li>
 * </ol>
 */
@ThreadSafe
public class SymmetricCipherFacade {
    /**
     * The default charset to use in the String based convenience methods.
     */
    public static final String DEFAULT_CHARSET = "UTF-8";

    /**
     * The currently supported cipher delegates (mapped by their id).
     * <p>
     * Each of these could potentially be used for decryption,
     * but only one of these is used for encryption
     * (the {@link #encryptionCipherDelegate}).
     * <p>
     * This is how backwards compatibility will be supported when
     * the algorithm needs to change in the future.
     * <p>
     * Each ciphertext is prepended with the delegate id.
     * So, when a new delegate needs to be added in the future,
     * it will be transparent to callers of {@link SymmetricCipherFacade}.
     * Decryption of existing ciphertexts can still occur by
     * using the old delegate.  But new encryptions can be handled
     * by the new delegate.
     */
    private final Map<Integer, SymmetricCipherDelegate> cipherDelegates;

    /**
     * The delegate used for new encryptions.
     */
    private final SymmetricCipherDelegate encryptionCipherDelegate;

    /**
     * Used to obtain encryption/decryption keys for specific cipher delegates.
     * <p>
     * This list is lazily populated on first delegate use.
     */
    private final Map<SymmetricCipherDelegate, KeyRotator> keyRotators;

    /**
     * Where the {@link javax.crypto.SecretKey}s are stored.
     * The callable should always return the same {@link KeyStoreWrapper}.
     * A callable is used so that the construction of the {@link KeyStoreWrapper}
     * can be delayed until it is actually used.
     */
    private Callable<KeyStoreWrapper> keyStoreWrapperProvider;

    /**
     * A Callable that always returns the same object.
     * We could just use Callables.returning(object) from guava,
     * but I didn't want to introduce a dependency on guava
     * purely for this functionality.
     */
    private static class SingletonCallable<T> implements Callable<T> {

        private final T singleton;

        public SingletonCallable(T singleton) {
            this.singleton = singleton;
        }

        @Override
        public T call() throws Exception {
            return this.singleton;
        }
    }

    /**
     * Construct a new {@link SymmetricCipherFacade} backed by the given {@link KeyStoreWrapper}.
     */
    public SymmetricCipherFacade(KeyStoreWrapper keyStoreWrapper) {
        this(new SingletonCallable<KeyStoreWrapper>(keyStoreWrapper));
    }

    /**
     * Construct a new {@link SymmetricCipherFacade} backed by the {@link KeyStoreWrapper}
     * returned by the given callable.  The callable should always return the same {@link KeyStoreWrapper}.
     * <p>
     * This method is provided so that the construction of the {@link KeyStoreWrapper}
     * can be delayed until it is actually used.
     */
    public SymmetricCipherFacade(Callable<KeyStoreWrapper> keyStoreWrapperProvider) {
        if (keyStoreWrapperProvider == null) {
            throw new NullPointerException("keyStoreWrapperProvider");
        }
        this.keyStoreWrapperProvider = keyStoreWrapperProvider;

        this.cipherDelegates = new HashMap<Integer, SymmetricCipherDelegate>();
        this.keyRotators = new HashMap<SymmetricCipherDelegate, KeyRotator>();

        List<SymmetricCipherDelegate> cipherDelegateLists = createCipherDelegates(new Sha1PrngSecureRandomFactory());
        this.encryptionCipherDelegate = cipherDelegateLists.get(cipherDelegateLists.size() - 1);

        for (final SymmetricCipherDelegate cipherDelegate : cipherDelegateLists) {

            this.cipherDelegates.put(cipherDelegate.getId(), cipherDelegate);
        }
    }

    /**
     * Creates and returns the list of delegates used for encryption/decryption.
     * <p>
     * The LAST entry in this list will be used for encryption.
     */
    private List<SymmetricCipherDelegate> createCipherDelegates(Callable<SecureRandom> secureRandomFactory) {
        List<SymmetricCipherDelegate> cipherDelegateList = new ArrayList<SymmetricCipherDelegate>();
        /*
         * Currently we only have one delegate.
         * If this ever becomes 'weak', and/or we have requirements to use a different
         * algorithm, then we can create another.
         *
         * However, we should ALWAYS keep this delegate to support decrypting
         * any ciphertexts created with it.
         */
        cipherDelegateList.add(new Aes128CbcPkcs5PaddingCipherDelegate(secureRandomFactory));
        return cipherDelegateList;
    }

    /**
     * Return the {@link KeyRotator} for the given {@link SymmetricCipherDelegate}.
     * <p>
     * Lazily creates the {@link KeyRotator} if it doesn't already exist.
     */
    private synchronized KeyRotator getKeyRotator(SymmetricCipherDelegate cipherDelegate) {
        KeyRotator keyRotator = this.keyRotators.get(cipherDelegate);
        if (keyRotator == null) {
            try {
                keyRotator = new KeyRotator(this.keyStoreWrapperProvider.call(), cipherDelegate);
                this.keyRotators.put(cipherDelegate, keyRotator);
            } catch (Exception e) {
                throw new CryptoException(e);
            }
        }
        return keyRotator;
    }

    /**
     * Convenience method to handle encryption with strings using the {@link #DEFAULT_CHARSET} charset.
     *
     * @see #encrypt(byte[])
     */
    public String encrypt(String plainText) {
        return encrypt(plainText, DEFAULT_CHARSET);
    }

    /**
     * Convenience method to handle encryption with strings using the given charset.
     *
     * @see #encrypt(byte[])
     */
    public String encrypt(String plainText, String charsetName) {
        if (plainText == null) {
            return null;
        }
        try {
            return new String(encrypt(plainText.getBytes(charsetName)), charsetName);
        } catch (UnsupportedEncodingException e) {
            throw new CryptoException(e);
        }
    }

    /**
     * Encrypts the given plainText with the current encryption key and cipher.
     * <p>
     *
     * Currently, encryption process works as follows:
     *
     * <ol>
     * <li>Randomly generate a {@link Aes128CbcPkcs5PaddingCipherDelegate#IV_BIT_LENGTH} bit Initialization Vector (IV) using a secure pseudo random number generator</li>
     * <li>Encrypt the plainText with the {@link Aes128CbcPkcs5PaddingCipherDelegate#TRANSFORMATION} algorithm using the current encryption key and IV</li>
     * <li>Prepend the encrypted data with
     *     <ul>
     *         <li>the {@value java.lang.Integer#SIZE} bit encryption cipher delegate id ({@link Aes128CbcPkcs5PaddingCipherDelegate#ID})
     *             (This is an internal identifier used to select the appropriate decryption algorithm), and</li>
     *         <li>the {@value java.lang.Long#SIZE} bit key index (highest byte first), and</li>
     *         <li>the {@link Aes128CbcPkcs5PaddingCipherDelegate#IV_BIT_LENGTH} bit Initialization Vector (IV)</li>
     *     </ul>
     * <li>Base64 encode the result</li>
     * </ol>
     * <p>
     * Or, in pseudocode:
     * <pre>
     * key = getCurrentKey();
     * keyIndex = getCurrentKeyIndex();
     * iv = generateIV();
     * return base64Encode(encryptionCipherDelegateId + keyIndex + iv + encrypt(key, iv, plaintext));
     * </pre>
     * <p>
     * The result can be decrypted by {@link #decrypt(byte[])}.
     */
    public byte[] encrypt(byte[] plainText) {

        if (plainText == null) {
            return null;
        }

        try {
            KeyEntry encryptionKeyEntry = getKeyRotator(this.encryptionCipherDelegate).getCurrentKey();

            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            DataOutputStream dos = new DataOutputStream(new Base64OutputStream(baos, true, -1, null));

            dos.writeInt(this.encryptionCipherDelegate.getId());
            dos.writeLong(encryptionKeyEntry.getIndex());
            dos.write(this.encryptionCipherDelegate.encrypt(encryptionKeyEntry.getKey(), plainText));
            dos.close();
            return baos.toByteArray();
        } catch (RuntimeException e) {
            throw e;
        } catch (Exception e) {
            throw new CryptoException(e);
        }
    }

    /**
     * Convenience method to handle decryption with strings using the {@link #DEFAULT_CHARSET} charset.
     *
     * @see #decrypt(byte[])
     */
    public String decrypt(String cipherText) {
        return decrypt(cipherText, DEFAULT_CHARSET);
    }

    /**
     * Convenience method to handle decryption with strings using the given charset.
     *
     * @see #decrypt(byte[])
     */
    public String decrypt(String cipherText, String charsetName) {

        if (cipherText == null) {
            return null;
        }

        try {
            return new String(decrypt(cipherText.getBytes(charsetName)), charsetName);
        } catch (UnsupportedEncodingException e) {
            throw new CryptoException(e);
        }
    }

    /**
     * Decrypts the given cipherText (in the format returned by {@link #encrypt(byte[])})
     * <p>
     *
     * The decryption process works as follows:
     *
     * <ol>
     * <li>Base64 decode the cipherText</li>
     * <li>Pluck off
     *     <ul>
     *         <li>the {@value java.lang.Integer#SIZE} bit decryption cipher delegate id (highest byte first)</li>
     *         <li>the {@value java.lang.Long#SIZE} bit key index (highest byte first), and</li>
     *         <li>the {@link Aes128CbcPkcs5PaddingCipherDelegate#IV_BIT_LENGTH} bit Initialization Vector (IV)</li>
     *     </ul>
     * <li>Lookup the cipher delegate used for encryption from cipher delegate id</li>
     * <li>Lookup the key used for encryption from the {@link KeyRotator} by the key index</li>
     * <li>Decrypt the plainText with the {@link Aes128CbcPkcs5PaddingCipherDelegate#TRANSFORMATION} algorithm using the key and IV</li>
     * <li>Return the plaintext result</li>
     * </ol>
     */
    public byte[] decrypt(byte[] cipherText) {

        if (cipherText == null) {
            return null;
        }

        DataInputStream dis = null;
        try {
            dis = new DataInputStream(new Base64InputStream(new ByteArrayInputStream(cipherText)));
            int cipherDelegateId = dis.readInt();

            SymmetricCipherDelegate decryptionCipherDelegate = this.cipherDelegates.get(cipherDelegateId);

            if (decryptionCipherDelegate == null) {
                throw new CryptoException(String.format("Unknown cipher delegate %d.  Ciphertext either was not created with SymmetricCipherFacade, or was created with a later version of SymmetricCipherFacade.", cipherDelegateId));
            }

            long keyIndex = dis.readLong();

            KeyEntry decryptionKeyEntry = getKeyRotator(decryptionCipherDelegate).getKey(keyIndex);

            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            IOUtils.copy(dis, baos);
            byte[] encryptedData = baos.toByteArray();

            byte[] plainText = decryptionCipherDelegate.decrypt(decryptionKeyEntry.getKey(), encryptedData);
            return plainText;
        } catch (RuntimeException e) {
            throw e;
        } catch (Exception e) {
            throw new CryptoException(e);
        } finally {
            IOUtils.closeQuietly(dis);
        }
    }

    /**
     * Generate the next key.
     * <p>
     * This new key will automatically become the new encryption key.
     * <p>
     * All ciphertexts previously encrypted with previous keys will still be able to be decrypted,
     * since the previous keys are not removed, and the ciphertext identifies which key
     * was used to encrypt it.
     */
    public void generateNewKey() throws NoSuchAlgorithmException, KeyStoreException, CertificateException, IOException {
        getKeyRotator(this.encryptionCipherDelegate).generateNextKey();
    }
}
