/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.originator.handler;

import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.table.TableResultParserDefault;
import com.netapp.oci.platform.originator.ObjState;
import com.netapp.oci.platform.originator.OriginatorUpdate;
import com.netapp.oci.platform.originator.OriginatorUpdateException;
import com.netapp.oci.platform.originator.handler.ColumnAttr;
import com.netapp.oci.platform.originator.handler.HandlerResult;
import com.netapp.oci.platform.originator.handler.XmlHandler;
import com.netapp.oci.platform.originator.key.KeyCacheException;
import com.netapp.oci.platform.originator.key.UuidKey;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.*;

/**
 * @author Al Velella
 *         Date: Sep 22, 2011
 *         Time: 12:00:37 PM
 */
public class VolumeHandler extends XmlHandler {

    public static final String TAG = "Volume";
    public static final String TABLE_NAME = "netapp_model.volume";
    /**
     * The key attribute for volume is InstanceUUID, not UUID.   On a volume move, UUID will change, but
     * instanceUUID will not.  Want to preserve the DB entry and objId in this case.
     */
    private static final String KEY_ATTR = "InstanceUUID";

    public VolumeHandler() {
        super(TAG, TABLE_NAME);
    }

    @Override
    public String getKeyAttr() {
        return KEY_ATTR;
    }

    @Override
    public HandlerResult parseProperties(OriginatorUpdate originatorUpdate, Map<String, String> properties,
                                         Map<UuidKey, ObjState> keyToObjState)
            throws OriginatorUpdateException, KeyCacheException {
        Long vserverId = originatorUpdate.peekObjectId();
        Long clusterId = originatorUpdate.peekObjectId(1);
        Long aggregateId = originatorUpdate.getObjectIdByTagname("Aggregate", properties.get("AggregateUUID"));
        Long exportPolicyId = originatorUpdate.getObjectIdByTagname("ExportPolicy", properties.get("ExportPolicyUUID"));
        Long snapshotPolicyId = originatorUpdate.getObjectIdByTagname("SnapshotPolicy", properties.get("SnapshotPolicyUUID"));
        Long sisPolicyId = originatorUpdate.getObjectIdByTagname("SisPolicy", properties.get("SisPolicyUUID"));
        Long storageServiceId = originatorUpdate.getObjectIdByTagname("StorageService", properties.get("StorageServiceUUID"));
        Long qosPolicyGroupId = originatorUpdate.getObjectIdByTagname("QosPolicyGroup", properties.get("QosPolicyGroupUUID"));

        return new HandlerResult(originatorUpdate, this, properties, KEY_ATTR, keyToObjState)
                .column("Name", String.class, EnumSet.of(ColumnAttr.REQUIRED))
                .column("UUID", String.class)
                .column("ResourceKey", String.class, EnumSet.of(ColumnAttr.REQUIRED, ColumnAttr.STRING_512))
                .column("ClusterId", clusterId)
                .column("VserverId", vserverId)
                .column("AggregateId", aggregateId)
                .column("ExportPolicyId", exportPolicyId)
                .column("SnapshotPolicyId", snapshotPolicyId)
                .column("SisPolicyId", sisPolicyId)
                .column("StorageServiceId", storageServiceId)
                .column("QosPolicyGroupId", qosPolicyGroupId)
                .column("Size", long.class)
                .column("SizeTotal", long.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("SizeUsed", long.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("SizeUsedPercent", int.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("SizeAvail", long.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("SizeAvailPercent", int.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("IsVserverRoot", boolean.class)
                .column("State", Enum.class)
                .column("StateRaw", String.class)
                .column("JunctionPath", String.class)
                .column("IsJunctionActive", boolean.class)
                .column("Style", Enum.class)
                .column("StyleRaw", String.class)
                .column("DerivedStyle", Enum.class)
                .column("IsSisVolume", boolean.class)
                .column("IsDataProtectionMirror", boolean.class)
                .column("IsLoadSharingMirror", boolean.class)
                .column("IsMoveMirror", boolean.class)
                .column("IsReplicaVolume", boolean.class)
                .column("IsSpaceGuaranteeEnabled", boolean.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("VMAlignSector", int.class)
                .column("VMAlignSuffix", String.class)
                .column("OverwriteReserve", long.class)
                .column("OverwriteReserveRequired", long.class)
                .column("OverwriteReserveUsed", long.class)
                .column("OverwriteReserveAvail", long.class)
                .column("OverwriteReserveActualUsed", long.class)
                .column("SnapshotReserveSize", long.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("PercentageSnapshotReserve", int.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("PercentageSnapshotReserveUsed", int.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("SizeAvailableForSnapshot", long.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("PercentageFractionalReserve", int.class)
                .column("DeduplicationSpaceSaved", long.class)
                .column("CompressionSpaceSaved", long.class)
                .column("SecurityStyle", Enum.class)
                .column("SecurityStyleRaw", String.class)
                .column("HybridCacheEligibility", Enum.class)
                .column("HybridCacheEligibilityRaw", String.class)
                .column("SizeUsedBySnapshots", long.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("InodeFilesUsed", long.class)
                .column("InodeFilesTotal", long.class)
                .column("InodeBlockType", Enum.class)
                .column("InodeBlockTypeRaw", String.class)
                .column("QuotaCommitted", long.class)
                .column("QuotaOverCommitted", long.class)
                .column("QuotaStatus", Enum.class)
                .column("QuotaStatusRaw", String.class)
                .column("FlexCacheMinReserve", long.class)
                .column("SpaceGuarantee", Enum.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("SpaceGuaranteeRaw", String.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("IsSnapshotCloneDependencyEnabled", boolean.class)
                .column("IsI2pEnabled", boolean.class)
                .column("IsAutoSnapshotsEnabled", boolean.class)
                .column("IsSnapDirAccessEnabled", boolean.class)
                .column("LanguageCode", String.class)
                .column("VolType", Enum.class)
                .column("VolTypeRaw", String.class)
                .column("SpaceMgmtOptionTryFirst", Enum.class)
                .column("SpaceMgmtOptionTryFirstRaw", String.class)
                .column("AutoSizeMode", Enum.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("AutoSizeModeRaw", String.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("AutoSizeMaximumSize", long.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("AutoSizeIncrementSize", long.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("IsAtimeUpdateEnabled", boolean.class)
                .column("IsCreateUcodeEnabled", boolean.class)
                .column("IsConvertUcodeEnabled", boolean.class)
                .column("CloneSplitEstimate", long.class)
                .column("CloneChildCount", int.class)
                .column("SnapshotCount", int.class, EnumSet.of(ColumnAttr.NULL_PERSISTS_LAST_VALUE))
                .column("IsSnapshotAutoDeleteEnabled", boolean.class)
                .column("SnapshotAutoDeleteCommitment", Enum.class)
                .column("SnapshotAutoDeleteCommitmentRaw", String.class)
                .column("SnapshotAutoDeleteDeleteOrder", Enum.class)
                .column("SnapshotAutoDeleteDeleteOrderRaw", String.class)
                .column("SnapshotAutoDeleteDeferDelete", Enum.class)
                .column("SnapshotAutoDeleteDeferDeleteRaw", String.class)
                .column("SnapshotAutoDeleteTargetFreeSpace", int.class)
                .column("SnapshotAutoDeleteTrigger", Enum.class)
                .column("SnapshotAutoDeleteTriggerRaw", String.class)
                .column("SnapshotAutoDeletePrefix", String.class)
                .column("SnapshotAutoDeleteDestroyList", String.class)
                .column("OldestSnapshotTimestamp", Timestamp.class)
                .column("SisStatus", Enum.class)
                .column("SisStatusRaw", String.class)
                .column("SisState", Enum.class)
                .column("SisStateRaw", String.class)
                .column("SisProgress", String.class)
                .column("SisType", Enum.class)
                .column("SisTypeRaw", String.class)
                .column("SisSchedule", String.class)
                .column("SisLastOpBeginTimestamp", Timestamp.class)
                .column("SisLastOpEndTimestamp", Timestamp.class)
                .column("SisLastOpError", String.class)
                .column("SisLastOpState", Enum.class)
                .column("SisLastOpStateRaw", String.class)
                .column("SisLastOpSize", long.class)
                .column("IsSisCompressionEnabled", boolean.class)
                .column("IsSisInlineCompressionEnabled", boolean.class)
                .column("PercentageDeduplicationSpaceSaved", int.class)
                .column("PercentageCompressionSpaceSaved", int.class)
                .column("IsSnapDiffEnabled", boolean.class)
                .column("IsManagedByService", boolean.class)
                .column("MaxDataConstituentSize", long.class)
                .column("MaxNamespaceConstituentSize", long.class)
                .column("ConstituentRole", Enum.class)
                .column("ConstituentRoleRaw", String.class)
                .column("IsConstituent", boolean.class)
                .column("SecurityUserID", String.class)
                .column("SecurityGroupID", String.class)
                .column("SecurityPermissions", String.class)
                .column("VolumeBlocksFootPrint", String.class)
                .column("CachingPolicy", Enum.class)
                .column("CachingPolicyRaw", String.class);
    }

    @Override
    public boolean isObjectOffline(OriginatorUpdate originatorUpdate, Map<String, String> objectProperties) {
        final String stateAttr = "State";
//        return (isObjectOffline(stateAttr, "OFFLINE", objectProperties) || isObjectOffline(stateAttr, "RESTRICTED", objectProperties))
//                && !originatorUpdate.isDelete(VserverHandler.TAG, originatorUpdate.peekObjectUuid());
        return isObjectOffline(stateAttr, "OFFLINE", objectProperties) || isObjectOffline(stateAttr, "RESTRICTED", objectProperties);
    }
    
    @Override
    public String alternateKeyToPrimaryKey(Connection connection, OriginatorUpdate originatorUpdate,
            Map<String, String> map) throws OriginatorUpdateException {
        Long originId = originatorUpdate.getOriginId();
        Long vServerId = originatorUpdate.getObjectIdByTagname("Vserver", map.get("vServerUUID"));
        String volumeName = map.get("name");
        List<String> results = Collections.<String> emptyList();

        try {
            StringBuilder sb = new StringBuilder();
            sb.append("SELECT ").append(getKeyAttr());
            sb.append(" FROM netapp_model.volume WHERE originid = ? ");
            sb.append(" AND vServerId = ? AND name = ?");
            
            PreparedStatementWrapper<String> preparedStatement = PreparedStatementWrapper.create(connection, sb.toString(), TableResultParserDefault.STRING);
            preparedStatement.setLong(1, originId);
            preparedStatement.setLong(2, vServerId);
            preparedStatement.setString(3, volumeName);
            
            results = preparedStatement.executeQueryEx();
        }
        catch (SQLException | TableException e) {
            throw new OriginatorUpdateException(String.format(
                    "Failed to retrieve volume uuid for originId: %d, vServerId: %d and volumeName: %s", originId,
                    vServerId, volumeName), e);
        }
        
        return (results != null && !results.isEmpty()) ? results.get(0) : null;
    }
}
