/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.hostadapter.interfaces.impl;


import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.hostadapter.interfaces.local.HostadapterPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.hostadapter.interfaces.remote.HostadapterPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.hostadapter.interfaces.data.HostadapterMeasurements;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import com.onaro.sanscreen.util.Constant;
import org.apache.log4j.Logger;
import org.jboss.ejb3.annotation.LocalBinding;
import org.jboss.ejb3.annotation.RemoteBinding;
import org.jboss.ejb3.annotation.RemoteBindings;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author dpadmana
 */
@Stateless
@RemoteBindings({
    @RemoteBinding(jndiBinding = HostadapterPerformanceQueryRemote.JNDI_NAME),
    @RemoteBinding(jndiBinding = Constant.HTTPS_JNDI_BINDING_PREFIX + HostadapterPerformanceQueryRemote.JNDI_NAME)})
@LocalBinding(jndiBinding = HostadapterPerformanceQueryLocal.JNDI_NAME)
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class HostadapterPerformanceQueryBean implements HostadapterPerformanceQueryRemote, HostadapterPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final Logger logger = Logger.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<HostadapterMeasurements>>> getSamples(Set<Long> hostadapterIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + hostadapterIds.size() + "hostadapter samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<HostadapterMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<HostadapterMeasurements>>>();

            Set<Long> remainingIds = new HashSet<Long>(hostadapterIds);
            Map<Long, Set<Long>> clusterIdToHostadapterIds = NetAppModel.INSTANCE.getClusterIdByHostadapterIds(connection, hostadapterIds);

            for (Map.Entry<Long, Set<Long>> entry : clusterIdToHostadapterIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, ids, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find hostadapter with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get hostadapter samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query hostadapter samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public Map<Long, List<SampleInfo<HostadapterMeasurements>>> getSamples(Long containerId, Set<Long> hostadapterIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + hostadapterIds.size() + "hostadapter samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return getSamples(connection, containerId, hostadapterIds, fromTime, toTime);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get hostadapter samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query hostadapter samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public List<SampleInfo<HostadapterMeasurements>> getSamples(Long containerId, Long hostadapterId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(hostadapterId), fromTime, toTime).get(hostadapterId);

    }

    private Map<Long, List<SampleInfo<HostadapterMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> hostadapterIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getHostadapterPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(hostadapterIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<HostadapterMeasurements>> getLatestSample(Long id, Set<Long> hostadapterIds) throws NetappInterfaceException {
        logger.info("Query " + hostadapterIds.size() + "hostadapter samples for :" + id);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getHostadapterPerformanceComponentsFactory().createPersistenceFacade(connection, id)
                            .getLatestSample(hostadapterIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get hostadapter samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<HostadapterMeasurements>>> getSummaries(Set<Long> hostadapterIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + hostadapterIds.size() + "hostadapter summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<HostadapterMeasurements>>>  idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(hostadapterIds);
            Map<Long, Set<Long>> clusterIdToHostadapterIds = NetAppModel.INSTANCE.getClusterIdByHostadapterIds(connection, hostadapterIds);

            for (Map.Entry<Long, Set<Long>> entry : clusterIdToHostadapterIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<HostadapterMeasurements>> summaries = ComponentsFactoryRepository.getHostadapterPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find hostadapter with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get hostadapter summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<HostadapterMeasurements>>> getSummaries(Long clusterId, Set<Long> hostadapterIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + hostadapterIds.size() + "hostadapter summaries from " + fromTime + " to " + toTime
                + " for cluster id " + clusterId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<HostadapterMeasurements>>>  idToSummaries = new HashMap<>();
                List<Summary<HostadapterMeasurements>> summaries = ComponentsFactoryRepository.getHostadapterPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, clusterId)
                        .getSummaries(new TimeSpan(fromTime, toTime), hostadapterIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get hostadapter summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
      public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
          logger.info("Query if " + objId + "hostadapter has counters from " + fromTime + " to " + toTime);

          Connection connection = null;
          try {
              connection = ConnectionHelper.newConnection(dataSource);
              return ComponentsFactoryRepository.getHostadapterPerformanceComponentsFactory()
                      .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
          } catch (Exception e) {
              throw new NetappInterfaceException("Failed to assert if hostadapter has counters ", e);
          } finally {
              ConnectionHelper.close(connection);
          }
      }
}
