
-- Create 2 quota tables, one for tree quotas and one for user quotas (includes user and group quotas)
-- Tree Quotas
CREATE TABLE netapp_model.qtree_quota (
    objid BIGINT(20) NOT NULL COMMENT 'Locally unique object identifier.  ZAPIs:  quota-report-iter.quota',
    originid BIGINT(20) NOT NULL,
    uuid VARCHAR(255) NOT NULL COMMENT 'UUID of the quota. Equal to the {VolumeUUID}.{QtreeUUID}.{QuotaType}.{QuotaTarget}. Not returned from ZAPI',
    resourceKey VARCHAR(512) NOT NULL COMMENT 'DERIVED. A structured key that may be used to uniquely identify this resource globally.',
    clusterId BIGINT(20) NOT NULL,
    vserverId BIGINT(20) NOT NULL,
    volumeId BIGINT(20) NOT NULL COMMENT 'Reference to the volume to which this quota belongs to',
    qtreeId BIGINT(20) DEFAULT NULL COMMENT 'Reference to the qtree to which this quota applies',
    quotaTarget VARCHAR(512) DEFAULT NULL COMMENT 'For an explicit quota, this value is specified in the /etc/quotas file and the domain in the QUOTA_TARGET_DOMAIN directive is in effect. Multiple targets are comma separated. For a derived quota, the field is blank.',
    quotaType ENUM ('USER', 'GROUP', 'TREE', 'NOT_MAPPED') DEFAULT NULL COMMENT 'Type of quota',
    diskLimit BIGINT(20) DEFAULT NULL COMMENT 'Maximum amount of disk space, in kilobytes, allowed for the quota target (hard disk space limit)',
    diskLimitUnlimited BOOL NOT NULL DEFAULT TRUE COMMENT 'True if there is no disk limit',
    diskUsed BIGINT(20) DEFAULT NULL COMMENT 'Current amount of disk space, in kilobytes, used by the quota target',
    diskUsedUnlimited BOOL NOT NULL DEFAULT TRUE COMMENT 'True if there is no limit fot the disk used',
    fileLimit BIGINT(20) DEFAULT NULL COMMENT 'Maximum number of files allowed for the quota target',
    fileLimitUnlimited BOOL NOT NULL DEFAULT TRUE COMMENT 'True if there is no file limit',
    fileUsed BIGINT(20) DEFAULT NULL COMMENT 'Current number of files used by the quota target',
    fileUsedUnlimited BOOL NOT NULL DEFAULT TRUE COMMENT 'True if there is no limit for file used',
    softDiskLimit BIGINT(20) DEFAULT NULL COMMENT 'Soft disk space limit, in kilobytes, for the quota target',
    softDiskLimitUnlimited BOOL NOT NULL DEFAULT TRUE COMMENT 'True if there is no soft disk limit',
    softFileLimit BIGINT(20) DEFAULT NULL COMMENT 'Soft file limit, in number of files, for the quota target',
    softFileLimitUnlimited BOOL NOT NULL DEFAULT TRUE COMMENT 'True if there is no soft file limit',
    threshold BIGINT(20) DEFAULT NULL COMMENT 'Disk space threshold, in kilobytes, for the quota target',
    thresholdUnlimited BOOL NOT NULL DEFAULT TRUE COMMENT 'True if there is no threshold limit',
    checksum INT NOT NULL,
    changeDetailChecksum INT NOT NULL,
    updatetime BIGINT(20) NOT NULL,
    KEY updatetime_index (updatetime),
    KEY originId_index (originid),
    CONSTRAINT FOREIGN KEY FK_clusterId (clusterId) REFERENCES netapp_model.cluster (objid),
    CONSTRAINT FOREIGN KEY FK_vserverId (vserverId) REFERENCES netapp_model.vserver (objid),
    CONSTRAINT FOREIGN KEY FK_volumeId (volumeId) REFERENCES netapp_model.volume (objid),
    CONSTRAINT FOREIGN KEY FK_qtreeId (qtreeId) REFERENCES netapp_model.qtree (objid),
    PRIMARY KEY (objid)
    ) ENGINE=INNODB;
    
-- User Quotas    
CREATE TABLE netapp_model.user_quota (
    objid BIGINT(20) NOT NULL COMMENT 'Locally unique object identifier.  ZAPIs:  quota-report-iter.quota',
    originid BIGINT(20) NOT NULL,
    uuid VARCHAR(255) NOT NULL COMMENT 'UUID of the quota. Equal to the {VolumeUUID}.{QtreeUUID}.{QuotaType}.{QuotaTarget}. Not returned from ZAPI',
    resourceKey VARCHAR(512) NOT NULL COMMENT 'DERIVED. A structured key that may be used to uniquely identify this resource globally.',
    clusterId BIGINT(20) NOT NULL,
    vserverId BIGINT(20) NOT NULL,
    volumeId BIGINT(20) NOT NULL COMMENT 'Reference to the volume to which this quota belongs to',
    qtreeId BIGINT(20) DEFAULT NULL COMMENT 'Reference to the qtree to which this quota applies',
    quotaTarget VARCHAR(512) DEFAULT NULL COMMENT 'For an explicit quota, this value is specified in the /etc/quotas file and the domain in the QUOTA_TARGET_DOMAIN directive is in effect. Multiple targets are comma separated. For a derived quota, the field is blank.',
    quotaType ENUM ('USER', 'GROUP', 'TREE', 'NOT_MAPPED') DEFAULT NULL COMMENT 'Type of quota',    
    diskLimit BIGINT(20) DEFAULT NULL COMMENT 'Maximum amount of disk space, in kilobytes, allowed for the quota target (hard disk space limit)',
    diskLimitUnlimited BOOL NOT NULL DEFAULT TRUE COMMENT 'True if there is no disk limit',
    diskUsed BIGINT(20) DEFAULT NULL COMMENT 'Current amount of disk space, in kilobytes, used by the quota target',
    diskUsedUnlimited BOOL NOT NULL DEFAULT TRUE COMMENT 'True if there is no limit fot the disk used',
    fileLimit BIGINT(20) DEFAULT NULL COMMENT 'Maximum number of files allowed for the quota target',
    fileLimitUnlimited BOOL NOT NULL DEFAULT TRUE COMMENT 'True if there is no file limit',
    fileUsed BIGINT(20) DEFAULT NULL COMMENT 'Current number of files used by the quota target',
    fileUsedUnlimited BOOL NOT NULL DEFAULT TRUE COMMENT 'True if there is no limit for file used',
    softDiskLimit BIGINT(20) DEFAULT NULL COMMENT 'Soft disk space limit, in kilobytes, for the quota target',
    softDiskLimitUnlimited BOOL NOT NULL DEFAULT TRUE COMMENT 'True if there is no soft disk limit',
    softFileLimit BIGINT(20) DEFAULT NULL COMMENT 'Soft file limit, in number of files, for the quota target',
    softFileLimitUnlimited BOOL NOT NULL DEFAULT TRUE COMMENT 'True if there is no soft file limit',
    threshold BIGINT(20) DEFAULT NULL COMMENT 'Disk space threshold, in kilobytes, for the quota target',
    thresholdUnlimited BOOL NOT NULL DEFAULT TRUE COMMENT 'True if there is no threshold limit',
    checksum INT NOT NULL,
    changeDetailChecksum INT NOT NULL,
    updatetime BIGINT(20) NOT NULL,
    KEY updatetime_index (updatetime),
    KEY originId_index (originid),
    CONSTRAINT FOREIGN KEY FK_clusterId (clusterId) REFERENCES netapp_model.cluster (objid),
    CONSTRAINT FOREIGN KEY FK_vserverId (vserverId) REFERENCES netapp_model.vserver (objid),
    CONSTRAINT FOREIGN KEY FK_volumeId (volumeId) REFERENCES netapp_model.volume (objid),
    CONSTRAINT FOREIGN KEY FK_qtreeId (qtreeId) REFERENCES netapp_model.qtree (objid),
    PRIMARY KEY (objid)
    ) ENGINE=INNODB;    

-- Sort the quotas from the original Quota table into the new typed quota tables 
INSERT INTO netapp_model.qtree_quota (SELECT * FROM netapp_model.quota WHERE quotaType = 'TREE'); 
INSERT INTO netapp_model.user_quota (SELECT * FROM netapp_model.quota WHERE quotaType <> 'TREE');


-- Rename the quotaId column of quota_user to userQuotaId
-- Create the new column
ALTER TABLE netapp_model.quota_user ADD COLUMN userQuotaId BIGINT(20) 
    NOT NULL 
    COMMENT 'Reference to the userquota to which this user belongs to' 
    AFTER vserverId;
ALTER TABLE netapp_model.quota_user ADD CONSTRAINT FOREIGN KEY FK_userQuotaId (userQuotaId) REFERENCES netapp_model.user_quota (objid);
-- Populate new column from old column
UPDATE netapp_model.quota_user SET userQuotaId = quotaId;
-- Remove the old constraint and column
ALTER TABLE netapp_model.quota_user DROP FOREIGN KEY quota_user_ibfk_3;
ALTER TABLE netapp_model.quota_user DROP COLUMN quotaId;

-- Delete the original Quota table
DROP TABLE netapp_model.quota;
