/*
 * Copyright (c) 2009 NetApp
 * All rights reserved
 */
package com.netapp.oci.platform.framework.performance.util;

import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.param.PluginParam;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.framework.performance.PersistenceStrategy;
import com.netapp.oci.platform.framework.performance.data.CurrentSampleStatus;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.framework.performance.tables.CurrentSampleStatusTable;
import com.netapp.oci.platform.framework.performance.tables.CurrentSampleTable;
import com.netapp.oci.platform.framework.performance.tables.PerformanceTableUtils;
import com.netapp.oci.platform.framework.performance.tables.SampleTable;
import com.onaro.commons.types.Period;
import org.apache.log4j.Logger;

import java.sql.Connection;
import java.util.*;

/**
 * Abstract implementation of sample persistence strategy
 *
 * @author Anton Sergeev
 */
public abstract class AbstractPersistenceStrategy<C> implements PersistenceStrategy<C> {
    /**
     * Logger instance
     */
    private Logger logger;

    /**
     * Param data
     */
    private PluginParam param;

    /**
     * Key of the container of samples: switch or storage or virtual center
     */
    private String sampleContainerKey;

    /**
     * Id of the container of samples: switch or storage or virtual center
     * Id is here for convenience only, not all containers are registered with SANscreen under certain id,
     * for example virtual center doesn't have id so this field may be <code>null</code>
     */
    private Long sampleContainerId;

        /**
     * Map of current sample table name to synchronization object
     * We need this object to synchronize access to current sample table
     */
    private static final Map<String, Object> currentSampleTableToSynchronizationObject = new HashMap<String, Object>();

    /**
     * Abstract method to get instance of sample table
     *
     * @param connection - database connection
     * @return sample table
     */
    protected abstract SampleTable<C> getSampleTable(Connection connection);

    /**
     * Abstract method to get instance of current sample table
     *
     * @param connection - database connection
     * @return current sample table
     */
    protected abstract CurrentSampleTable<C> getCurrentSampleTable(Connection connection);

    /**
     * Performs preparation step of the persistence strategy.
     * This method will be executed synchronously.
     *
     * @param connection - database connection
     * @param sampleTime - sample time
     * @throws PerformanceException if failed to prepare data
     */
    protected abstract void prepare(Connection connection, Long sampleTime) throws PerformanceException;

    /**
     * Completes the process of sample persistence
     *
     * @param connection - database connection
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if failed to complete execution of the strategy
     */
    protected abstract void complete(Connection connection) throws PerformanceException;

    /**
     * @param connection - database connection
     * @throws PerformanceException if failed to complete execution of the strategy
     */
    private void updateCurrent(Connection connection, Collection<Sample<C>> samples) throws PerformanceException {
        try {
            CurrentSampleTable<C> currentSampleTable = getCurrentSampleTable(connection);
            synchronized (getCurrentSampleTableSynchronizationObject(currentSampleTable)) {

                Sample<C> firstSample = samples.iterator().next();
                long time = firstSample.getTime();
                String sampleTimeMessage = "sample timestamp: " + time + " sample date: " + new Date(time);

                // samples collection might have samples for the same id but with different time
                // these samples are produced by raw sample converter as "first report" samples and they have "time - 1"
                // as their time. we need to filter them out or we can get duplicate key exception in the current sample tables
                Collection<Sample<C>> filteredSamples = new LinkedList<Sample<C>>();
                for (Sample<C> sample : samples) {
                    if (sample.getTime() == time) {
                        filteredSamples.add(sample);
                    }
                }
                getLogger().info("PersistenceStrategy: Insert current samples size: " + filteredSamples.size() + " " + sampleTimeMessage);
                currentSampleTable.insert(filteredSamples);
                new CurrentSampleStatusTable(connection, currentSampleTable.getDatabase()).insert(new CurrentSampleStatus(getPerformancePackageId(), sampleContainerKey, sampleContainerId, time));
            }
        } catch (TableException e) {
            throw new PerformanceException("Failed to insert samples into current table", e);
        }
    }

    private Object getCurrentSampleTableSynchronizationObject(CurrentSampleTable<C> currentSampleTable) {
        synchronized (currentSampleTableToSynchronizationObject) {
            // initialize synchronization object for current sample table
            Object synchObject = currentSampleTableToSynchronizationObject.get(currentSampleTable.getFullName());
            if (synchObject == null) {
                synchObject = new Object();
                currentSampleTableToSynchronizationObject.put(currentSampleTable.getFullName(), synchObject);
            }
            return synchObject;
        }
    }


    @SuppressWarnings("unchecked")
    protected AbstractPersistenceStrategy(Logger logger, PluginParam param, String sampleContainerKey, Long sampleContainerId) {
        this.logger = logger;
        this.param = param;
        this.sampleContainerKey = sampleContainerKey;
        this.sampleContainerId = sampleContainerId;
    }

    public Logger getLogger() {
        return logger;
    }

    public void execute(Connection connection, Collection<Sample<C>> samples) throws PerformanceException {
        if (samples == null || samples.isEmpty())
            throw new IllegalArgumentException("Sample list must not be empty");

        long execStartTime = System.currentTimeMillis();
        long summaryElapseTime = 0;
        long currentElapseTime = 0;
        final boolean updateCurrentSample = PerformanceTableUtils.getUpdateCurrentSample(param);

        synchronized (PersistenceState.getPackageSyncObject(getPerformancePackageId())) {
            boolean sampleInserted = false;
            Sample<C> sample = samples.iterator().next();
            long time = sample.getTime();
            if (getLogger().isDebugEnabled()) {
                getLogger().debug("PersistenceStrategy: Insert samples size: " + samples.size() + ", First sample: " + sample);
            }   

            SampleTable<C> sampleTable = getSampleTable(connection);
            Long lastSampleTime = getLastSampleTime(sampleTable);
            if (lastSampleTime != null) {
                GregorianCalendar calendar = new GregorianCalendar();

                calendar.setTimeInMillis(lastSampleTime.longValue());
                int lastReportHourOfDay = calendar.get(GregorianCalendar.HOUR_OF_DAY);
                calendar.setTimeInMillis(time);
                int reportHourOfDay = calendar.get(GregorianCalendar.HOUR_OF_DAY);

                if (getLogger().isDebugEnabled()) {
                    getLogger().debug(
                            "PersistenceStrategy: lastReportHourOfDay: "
                                    + lastReportHourOfDay
                                    + ", reportHourOfDay: "
                                    + reportHourOfDay
                                    + ", Sample class: "
                                    + (sample != null && sample.getSampleData() != null ? sample.getSampleData()
                                            .getClass() : null) + ", id: " + (sample != null ? sample.getId() : null));
                }
                
                // if must use "!=" and not "<" 23:00 is bigger then 00:00
                if (lastReportHourOfDay != reportHourOfDay || (time - lastSampleTime.longValue()) > Period.HOUR.milliseconds()) {
                    long startTime = System.currentTimeMillis();
                    try {
                        getLogger().info("PersistenceStrategy: Start summary");
                        Long sampleTime = null;
                        
                        if (lastSampleTime.longValue() > time) {
                            // The samples from different data sources are not received in chronological (time) order
                            sampleTime = Long.valueOf(time);
                            sampleTable.insert(samples);
                            sampleInserted = true;
                        }
                        
                        prepare(connection, sampleTime);
                        complete(connection);
                        summaryElapseTime = System.currentTimeMillis() - startTime;
                        getLogger().info("PersistenceStrategy: End summary in " + summaryElapseTime + " millseconds");
                    } 
                    catch (TableException e) {
                        throw new PerformanceException("Failed to insert samples", e);
                    }
                    catch (PerformanceException e) {
                        throw new PerformanceException("Failed to summarize " + sampleTable.getName(), e);
                    }
                }
            }

            try {
                if (!sampleInserted) {
                    sampleTable.insert(samples);
                }
            } catch (TableException e) {
                throw new PerformanceException("Failed to insert samples", e);
            }

            if (lastSampleTime == null || time > lastSampleTime.longValue()) {
              PersistenceState.setLastSampleTime(getPerformancePackageId(), time);
            }
            
            // save latest samples in atomic reference
            if (updateCurrentSample) {
                long startTime = System.currentTimeMillis();
                getLogger().info("PersistenceStrategy: Start current sample.");
                updateCurrent(connection, samples);
                currentElapseTime = System.currentTimeMillis() - startTime;
            }
        }

        long totalElapse = System.currentTimeMillis() - execStartTime;
        if (totalElapse > 20000) {
            getLogger().warn("method 'execute(...)' in class " + this.getClass() + " used (ms) " + totalElapse);
            getLogger().warn("summary used " + summaryElapseTime);
            if (updateCurrentSample) {
                getLogger().warn("current sample used " + currentElapseTime);
            }
        }
    }

    /**
     * Gets last report time for sample table
     *
     * @param sampleTable - sample table instance
     * @return the last report time for a given storageKey,  return null if the table does not exist or the table is empty
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if failed to get last report time from the sample table
     */
    private synchronized Long getLastSampleTime(SampleTable<C> sampleTable) throws PerformanceException {
        try {
            Long lastReportTime = PersistenceState.getLastSampleTime(getPerformancePackageId());
            if (lastReportTime == null) {
                if (sampleTable.isTableExist()) {
                    TimeSpan timeSpan = sampleTable.getTableTimeSpan();
                    if (timeSpan != null) {
                        lastReportTime = Long.valueOf(timeSpan.getToTime());
                    }
                }
            }

            return lastReportTime;
        } catch (TableException e) {
            throw new PerformanceException("Failed to get last report time from " + sampleTable.getFullName(), e);
        }
    }

    public String getSampleContainerKey() {
        return sampleContainerKey;
    }

    public Long getSampleContainerId() {
        return sampleContainerId;
    }

    protected TimeSpan getHourTimeSpan(SampleTable<C> sampleTable) throws TableException {
        TimeSpan tableTimeSpan = sampleTable.getTableTimeSpan();

        GregorianCalendar calendar = new GregorianCalendar();

        // reset minutes, seconds, milliseconds for from depending on the period
        calendar.setTimeInMillis(tableTimeSpan.getFromTime());
        calendar.set(Calendar.MINUTE, 0);
        calendar.set(Calendar.SECOND, 0);
        calendar.set(Calendar.MILLISECOND, 0);

        return new TimeSpan(calendar.getTimeInMillis(), calendar.getTimeInMillis() + Period.HOUR.milliseconds() - 1);
    }
}
