/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.framework.performance.util;

import com.netapp.oci.platform.common.table.Table;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.framework.performance.tables.SampleTable;
import com.netapp.oci.platform.framework.performance.tables.SummaryTable;
import com.netapp.oci.platform.framework.performance.TableManagement;
import org.apache.log4j.Logger;

import java.sql.Connection;
import java.util.HashSet;
import java.util.Set;

/**
 * Abstract implementation of the table management interface.
 * This implementation is in most cases what you need for your performance package.
 * It manages sample.
 *
 * @author Anton Sergeev
 */
public abstract class AbstractTableManagement<C> implements TableManagement {
    /**
     * Caching all the created tables, avoid executing SQL 'SHOW TABLES like ..'
     */
    private static final Set<String> createdTableNames = new HashSet<String>();

    protected Logger logger;

    protected AbstractTableManagement(Logger logger) {
        this.logger = logger;
    }

    /**
     * Abstract method to get instance of sample table
     *
     * @param connection - database connection
     * @return sample table
     */
    protected abstract SampleTable<C> getSampleTable(Connection connection);

    /**
     * Adapter method to create more tables if performance package relies on more tables
     * @param connection - database connection
     * @throws TableException if failed to create more tables
     */
    protected void createMoreTables(Connection connection) throws TableException {
    }

    /**
     * Adapter method to drop more tables if performance package relies on more tables
     * @param connection - database connection
     * @throws com.netapp.oci.platform.common.table.TableException if failed to drop more tables
     */
    protected void dropMoreTables(Connection connection) throws TableException {
    }

    /**
     * Adapter method to find out if more tables required by package exist
     * @param connection - database connection
     * @return indicator if tables exist
     * @throws com.netapp.oci.platform.common.table.TableException if failed to find out if more tables
     */
    protected boolean moreTablesExist(Connection connection) throws TableException {
        return true;
    }

    public void createTables(Connection connection) throws PerformanceException {
        SampleTable<C> sampleTable = getSampleTable(connection);

        if (isMarkCreated(sampleTable))
            return;

        try {
            if (!tablesExist(connection)) {
                logger.info("Table Management: Database tables don't exist - creating them.");
                sampleTable.createTable();
                logger.info("Table Management: Created table " + sampleTable.getFullName());

                createMoreTables(connection);
            }
            markCreated(sampleTable);
        } catch (TableException ex) {
            clearCaches();
            throw new PerformanceException("Failed to create performance tables", ex);
        }
    }

    public void dropTables(Connection connection) throws PerformanceException {
        try {
            SampleTable<C> sampleTable = getSampleTable(connection);
            sampleTable.dropTable();
            logger.info("Table Management: Dropped table " + sampleTable.getFullName());
            dropMoreTables(connection);
        } catch (TableException ex) {
            throw new PerformanceException("Failed to drop performance tables", ex);
        } finally {
            clearCaches();
        }
    }

    public static synchronized boolean isMarkCreated(Table table) {
        return createdTableNames.contains(table.getFullName());
    }

    public static synchronized void markCreated(Table table) {
        createdTableNames.add(table.getFullName());
    }

    public static synchronized void unmarkCreated(SampleTable sampleTable) {
        createdTableNames.remove(sampleTable.getFullName());
    }
    
    public static synchronized void clearCaches() {
        createdTableNames.clear();
    }

    private boolean tablesExist(Connection connection) throws TableException {
        SampleTable<C> sampleTable = getSampleTable(connection);
        logger.info("Checking if " + sampleTable.getFullName() + " exist");
        boolean result = sampleTable.isTableExist();
        result &= moreTablesExist(connection);

        return result;
    }
}
