/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.job;

import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import org.apache.log4j.Logger;
import org.jboss.ejb3.annotation.ResourceAdapter;
import org.quartz.Job;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;

import javax.annotation.Resource;
import javax.annotation.security.RunAs;
import javax.ejb.ActivationConfigProperty;
import javax.ejb.MessageDriven;
import javax.sql.DataSource;
import java.sql.Connection;

/**
 * @author Al Velella
 *         Date: 2/7/13
 *         Time: 3:10 PM
 */

/**
 * This job exists to workaround a jboss 6 issue where idle connection pools don't shutdown/start properly. We
 * workaround this by mandating a minimum pool size of 1. However, this causes a bad connection to be maintained
 * in the pool if mysql has shot the connection after a quiet period. Thus, this job pings the database every six hours
 * and attempts to keep the last connection alive.
 */

@MessageDriven(activationConfig = {
        @ActivationConfigProperty(propertyName = "cronTrigger", propertyValue = "0 0 0/6 * * ?"),
        @ActivationConfigProperty(propertyName = "jobName", propertyValue = "ConnectionPoolPingJob")})
@ResourceAdapter("quartz-ra.rar")
@RunAs("admin")
public class ConnectionPoolPingJob implements Job {

    private static final Logger logger = Logger.getLogger(ConnectionPoolPingJob.class);

    @Resource(name = "jdbc/OnaroDataSource",
            mappedName = "java:/onaro_pool",
            type = DataSource.class,
            authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSourceOnaroPool;

    @Resource(name = "jdbc/OnaroDataSource",
            mappedName = "java:/onaro_pool_acq",
            type = DataSource.class,
            authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSourceAcqOnaroPool;

    @Override
    public void execute(JobExecutionContext jobExecutionContext) throws JobExecutionException {
        if (logger.isDebugEnabled()) {
            logger.debug("Starting Connection Pool Ping Job");
        }
        ping(dataSourceOnaroPool);
        ping(dataSourceAcqOnaroPool);
    }

    private void ping(DataSource dataSource) {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            PreparedStatementWrapper.executeFunctionQuery(connection, "SELECT COUNT(*) FROM management.version", -1L);
        } catch (TableException e) {
            logger.error("Ping error", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
