/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.originator.db;

import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.originator.key.UuidKey;
import com.netapp.oci.platform.originator.handler.ColumnAttr;
import com.netapp.oci.platform.originator.handler.ColumnData;
import com.netapp.oci.platform.originator.handler.RowData;
import org.apache.log4j.Logger;

import java.sql.Connection;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

/**
 * @author Al Velella
 *         Date: Sep 22, 2011
 *         Time: 3:02:22 PM
 */
public abstract class AbstractBatchOperation implements DbOperation {

    private Logger logger = Logger.getLogger("com.netapp.oci.platform.originator.db");

    protected String sql;
    protected ColumnData columnData;
    protected boolean executeBatch;
    protected List<RowData> rowDatas;

    public AbstractBatchOperation(ColumnData columnData, boolean executeBatch) {
        this.columnData = columnData;
        this.executeBatch = executeBatch;
        this.rowDatas = new ArrayList<RowData>();
    }

    public ColumnData getColumnData() {
        return columnData;
    }

    public int getOperationsCount() {
        return rowDatas.size();
    }

    public void execute(Connection connection, long time) throws TableException {
        boolean debug = logger.isDebugEnabled();

        PreparedStatementWrapper<Object> statement = PreparedStatementWrapper.createForUpdate(connection, sql);

        if (executeBatch) {
            for (RowData rowData : rowDatas) {
                fillSqlStatement(statement, rowData, time);
                if (debug) {
                    logger.info(statement + ";");
                }
                statement.addBatchEx();
            }
            statement.executeBatchEx();
        }
        else {
            if (debug) {
                logger.info(statement + ";");
            }
            statement.executeUpdateEx();
        }
    }

    protected abstract void fillSqlStatement(PreparedStatementWrapper<Object> statement, RowData rowData, long time) throws TableException;

    protected void append(PreparedStatementWrapper<Object> statement, RowData rowData, ColumnAttr ignoreAttribute) throws TableException {
        List<EnumSet<ColumnAttr>> attributes = columnData.getAttributes();
        List<Class<?>> types = columnData.getTypes();
        List<Object> values = rowData.getValues();
        for (int i = 0; i < values.size(); i++) {
            appendValue(statement, ignoreAttribute, attributes, types, values, i);
        }
    }

    protected void appendBaseValues(PreparedStatementWrapper<Object> statement, RowData rowData, ColumnAttr ignoreAttribute) throws TableException {
        List<EnumSet<ColumnAttr>> attributes = columnData.getBaseAttributes();
        List<Class<?>> types = columnData.getBaseTypes();
        List<Object> values = rowData.getBaseValues();
        for (int i = 0; i < values.size(); i++) {
            appendValue(statement, ignoreAttribute, attributes, types, values, i);
        }
    }

    private void appendValue(PreparedStatementWrapper<Object> statement, ColumnAttr ignoreAttribute, List<EnumSet<ColumnAttr>> attributes, List<Class<?>> types, List<Object> values, int i) throws TableException {
        if (!attributes.get(i).contains(ignoreAttribute)) {
            if (types.get(i) == String.class) {
                statement.setNext((String) values.get(i));
            } else if (types.get(i) == long.class) {
                statement.setNext((Long) values.get(i));
            } else if (types.get(i) == Enum.class) {
                statement.setNext((String) values.get(i));
            } else if (types.get(i) == int.class) {
                statement.setNext((Integer) values.get(i));
            } else if (types.get(i) == boolean.class) {
                statement.setNext((Boolean) values.get(i));
            } else if (types.get(i) == Timestamp.class) {
                statement.setNext((Timestamp)values.get(i));
            } else if (types.get(i) == UuidKey.class) {
                statement.setNext(values.get(i).toString());
            } else {
                throw new RuntimeException ("Unrecognized column type: " + types.get(i).toString());
            }
        }
    }

    public void addRow(RowData rowData) {
        rowDatas.add(rowData);
    }

}
