/**
 * NetApp 2014
 */

package com.netapp.oci.statsutils;

import java.lang.IllegalArgumentException;
import java.lang.StringBuilder;

/**
 *  This class provides an abstraction from having to understand the MySQL
 *  sharding techniques that MEGA has deployed.  If new statistic types are
 *  added and new sharding techniques are used, this shared library should
 *  be updated.
 * 
 */ 
public class StatisticUtility {

    /**
     * The character used for table name seperation.
     */
    private static final String TABLE_SEPERATOR = "_";

    /**
     * This is the prefix that is used for the cooked data samples.
     */
    private static final String SAMPLE_TABLE = "sample";

    /**
     * This is the prefix that is used for the summary statistic data 
     * samples.
     */
    private static final String SUMMARY_TABLE = "summary";

    /**
     * Raw tables use a combination of the sample table and the a 
     * raw keyword.
     */
    private static final String RAW_TABLE = "raw";

    /**
     * Aggregated tables use this prefix
     */
    private static final String AGGREGATE_TABLE = "aggregate";


    /**
     * This method will return the statistical table name for the corresponding
     * resource type, statistic type and parameters.  Some knowledge of ONTAP 
     * containment is required.  As an example, the caller needs to understand
     * that a disk is not contained by a vserver so no need to pass that in.
     *
     * The returned parameter can be used directly against the "netapp_performance" 
     * database to look up the object.
     *
     * Shard Ids used per object:
     *
     *     o - aggregate : clusterId
     *     o - disk      : clusterId
     *     o - volume    : vserverId
     *     o - lun       : vserverId
     *
     * Resource types to stat table available:
     *
     *     o - node        : sample_node, sample_node_raw, summary_node
     *     o - vserver     : sample_vserver, sample_vserver_raw, summary_vserver
     *     o - processor   : sample_processor, sample_processor_raw, summary_processor
     *     o - networkport : sample_networkport, sample_networkport_raw, summary_networkport
     *     o - networklif  : sample_networklif, sample_networklif_raw, summary_networklif
     *     o - fcplif      : sample_fcplif, sample_fcplif_raw, summary_fcplif
     *     o - fcpport     : sample_fcpport, sample_fcpport_raw, summary_fcpport
     *     o - iscsilif    : sample_iscsilif, sample_iscsilif_raw, summary_iscsilif
     *     o - aggregate   : sample_aggregate_X, sample_aggregate_raw_X, summary_aggregate_X
     *     o - disk        : sample_disk_X, sample_disk_raw_X, summary_disk_X
     *     o - volume      : sample_volume_X, sample_volume_raw_X, summary_volume_X
     *     o - lun         : sample_lun_X, sample_lun_raw_X, summary_lun_X
     *
     *
     * @param rtype - The resource type that you want to get the table for
     * @param stype - The type of statistic that you are looking for.
     * @param clusterId - The database id of the cluster
     * @param nodeId - The database id of the node
     * @param vserverId - The database id of the vserver
     *
     * @throws IllegalParameterException when a parameter is invalid.  Pay 
     * attention to the message that comes back to understand what parameter
     * is missing.
     *
     */
    public static String getStatisticTableName( ResourceType rtype,
                                                StatisticType stype,
                                                Long clusterId,
                                                Long nodeId,
                                                Long vserverId )
                                                throws IllegalArgumentException {

        // Make sure that the rtype and stype are not null and are
        // not "unknown"
        //
        if ( rtype == null ) {
            throw new IllegalArgumentException( "Resource type cannot be null" );
        }
        if ( rtype.equals(ResourceType.UNKNOWN) ) {
            throw new IllegalArgumentException( "Resource type cannot be UNKNOWN" );
        }
        if ( stype == null ) {
            throw new IllegalArgumentException( "Statistic type cannot be null" );
        }
        if ( stype.equals(StatisticType.UNKNOWN) ) {
            throw new IllegalArgumentException( "Statistic type cannot be UNKNOWN" );
        }
                                                
        StringBuilder statisticTable = new StringBuilder();
        // Start to build up the table by looking at the statistic type
        //
        if ( stype.equals(StatisticType.COOKED) || stype.equals(StatisticType.RAW) ) {
            statisticTable.append(SAMPLE_TABLE + TABLE_SEPERATOR);
        } else if ( stype.equals(StatisticType.SUMMARY) ) {
            statisticTable.append(SUMMARY_TABLE + TABLE_SEPERATOR);
        } else {
            statisticTable.append(AGGREGATE_TABLE + TABLE_SEPERATOR);
        }

        // Now add the resource type to it.
        //
        statisticTable.append(rtype.getName());

        // Now if this is a raw table request, next format requires the raw suffix
        //
        if ( stype.equals(StatisticType.RAW) || stype.equals(StatisticType.AGGREGATE_RAW) ) {
            statisticTable.append(TABLE_SEPERATOR + RAW_TABLE);
        }

        // Now if this is an aggregate, disk, volume or lun, we have more processing
        //
        if ( rtype.equals(ResourceType.DISK) || 
             rtype.equals(ResourceType.AGGREGATE) || 
             rtype.equals(ResourceType.VOLUME) || 
             rtype.equals(ResourceType.LUN) ) {

            // Dont forget separator
            statisticTable.append(TABLE_SEPERATOR);

            // Would use a switch statement here, but due to the multi value of the
            // enumeration java compiler is complaining.
            //
            if ( rtype.equals(ResourceType.DISK) ) {
                // Disk requires the cluster Id, make sure it is not null
                if ( clusterId == null ) {
                    throw new IllegalArgumentException("Resource type of DISK requires cluster Id to be specified");
                }
                statisticTable.append(clusterId);
            } else if ( rtype.equals(ResourceType.AGGREGATE) ) {
                // Aggregate requires the cluster Id, make sure it is not null
                if ( clusterId == null ) {
                    throw new IllegalArgumentException("Resource type of AGGREGATE requires cluster Id to be specified");
                }
                statisticTable.append(clusterId);
            } else if ( rtype.equals(ResourceType.VOLUME) ) {
                // Volume requires the vserver Id, make sure it is not null
                if ( vserverId == null ) {
                    throw new IllegalArgumentException("Resource type of VOLUME requires vserver Id to be specified");
                }
                statisticTable.append(vserverId);
            } else if ( rtype.equals(ResourceType.LUN) ) {
                // Lun requires the vserver Id, make sure it is not null
                if ( vserverId == null ) {
                    throw new IllegalArgumentException("Resource type of LUN requires vserver Id to be specified");
                }
                statisticTable.append(vserverId);
            }
        }

        return statisticTable.toString();

    }


}

