package com.onaro.commons.exec;

import org.apache.log4j.Logger;

import java.io.*;

/**
 * Date: Aug 31, 2006
 * Time: 1:40:12 PM
 */
public class Exec {

    static transient final Logger logger = Logger.getLogger("com.onaro.commons.exec");//$NON-NLS-1$

    /**
     * a specified system command.
     */
    private String command;

    /**
     * the working directory of the subprocess, or
     * <tt>null</tt> if the subprocess should inherit
     * the working directory of the current process.
     */
    private File dir;

    /**
     * array of strings, each element of which
     * has environment variable settings in the format
     * <i>name</i>=<i>value</i>, or
     * <tt>null</tt> if the subprocess should inherit
     * the environment of the current process.
     */
    private String[] envp;

    private int exitCode;

    private String error;

    private String output;

    private Exception exception;

    public Exec(String command) {
        this(command, null, null);
    }


    public Exec(String command, File dir) {
        this(command, dir, null);
    }

    /**
     * Executes the specified string command in a separate process with the
     * specified environment and working directory.
     * <p/>
     * <p>This is a convenience method.  An invocation of the form
     * <tt>exec(command, envp, dir)</tt>
     * behaves in exactly the same way as the invocation
     * where <code>cmdarray</code> is an array of all the tokens in
     * <code>command</code>.
     * <p/>
     * <p>More precisely, the <code>command</code> string is broken
     * into tokens using a {@link java.util.StringTokenizer} created by the call
     * <code>new {@link java.util.StringTokenizer}(command)</code> with no
     * further modification of the character categories.  The tokens
     * produced by the tokenizer are then placed in the new string
     * array <code>cmdarray</code>, in the same order.
     *
     * @param command a specified system command.
     * @param envp    array of strings, each element of which
     *                has environment variable settings in the format
     *                <i>name</i>=<i>value</i>, or
     *                <tt>null</tt> if the subprocess should inherit
     *                the environment of the current process.
     * @param dir     the working directory of the subprocess, or
     *                <tt>null</tt> if the subprocess should inherit
     *                the working directory of the current process.
     * @return A new {@link Process} object for managing the subprocess
     * @throws SecurityException        If a security manager exists and its
     *                                  {@link SecurityManager#checkExec checkExec}
     *                                  method doesn't allow creation of the subprocess
     * @throws IOException              If an I/O error occurs
     * @throws NullPointerException     If <code>command</code> is <code>null</code>,
     *                                  or one of the elements of <code>envp</code> is <code>null</code>
     * @throws IllegalArgumentException If <code>command</code> is empty
     * @see ProcessBuilder
     * @since 1.3
     */
    public Exec(String command, File dir, String[] envp) {
        this.command = command;
        this.dir = dir;
        this.envp = envp;
    }

    public void run(long timout) {
        exception = null;
        output = null;
        error = null;
        exitCode = Integer.MIN_VALUE;

        ProcessOutputHandler stdout = new ProcessOutputHandler();
        ProcessOutputHandler stderr = new ProcessOutputHandler();
        try {
            final Process process = Runtime.getRuntime().exec(command, envp, dir);

            if (timout > -1) {
                final long killTime = System.currentTimeMillis() + timout;
                new Thread(new Runnable() {
                    public void run() {
                        try {
                            while (System.currentTimeMillis() < killTime && exitCode == Integer.MIN_VALUE) {
                                Thread.sleep(1000);
                            }
                            if (exitCode == Integer.MIN_VALUE) {
                                logger.info("destroy process");//$NON-NLS-1$
                                process.destroy();
                            }
                        } catch (InterruptedException e) {
                            if (exitCode == Integer.MIN_VALUE) {
                                logger.info("destroy process", e);//$NON-NLS-1$
                                process.destroy();
                            }
                        }
                    }
                }).start();
            }

            Thread currentThread = Thread.currentThread();
            stdout.start(currentThread, process.getInputStream(), "stdout");//$NON-NLS-1$
            stderr.start(currentThread, process.getErrorStream(), "stderr");//$NON-NLS-1$

            exitCode = process.waitFor();
            logger.debug("exitCode: " + exitCode);//$NON-NLS-1$

        } catch (Exception e) {
            exception = e;
            error = "Failed to exec: " + command;//$NON-NLS-1$
            logger.error("Failed to exec: " + command + " - " + e.getMessage(), e);//$NON-NLS-1$ //$NON-NLS-2$
            return;
        }

        error = stderr.getOutput();
        output = stdout.getOutput();

        exception = stderr.exception;
        if (exception == null)
            exception = stdout.exception;
    }

    public String getCommand() {
        return command;
    }

    public int getExitCode() {
        return exitCode;
    }

    public String getError() {
        return error;
    }

    public String getOutput() {
        return output;
    }

    public Exception getException() {
        return exception;
    }

    boolean isError() {
        return exception != null && error != null && error.length() > 0 && exitCode != Integer.MIN_VALUE;
    }

    class ProcessOutputHandler {

        LineNumberReader reader;

        StringBuilder output = new StringBuilder();

        Exception exception = null;

        Thread currentThread;

        String streamName;

        public void start(Thread currentThread, InputStream in, String streamName) {
            try {
                this.currentThread = currentThread;
                reader = new LineNumberReader(new InputStreamReader(in));
                this.streamName = streamName;
            } catch (Exception e) {
                exception = e;
                output.append(streamName).append(" Failed to create LineNumberReader:").append(e.getMessage());//$NON-NLS-1$
                logger.error(streamName + " Exception while execute " + command + " - " + e.getMessage(), e);//$NON-NLS-1$ //$NON-NLS-2$
                return;
            }

            String lineSeparator = System.getProperty("line.separator");//$NON-NLS-1$
            String line;
            try {
                while ((line = reader.readLine()) != null) {
                    output.append(line).append(lineSeparator);
                    if (logger.isDebugEnabled()) {
                        logger.debug(streamName + ": " + line);//$NON-NLS-1$
                    }
                }
                logger.debug(streamName + ": done");//$NON-NLS-1$
            } catch (IOException e) {
                logger.error("Exception while execute " + command + " - " + e.getMessage(), e);//$NON-NLS-1$ //$NON-NLS-2$
                exception = e;
                currentThread.interrupt();
            }
        }

        boolean getException() {
            return exception != null;
        }

        public String getOutput() {
            return output.toString();
        }
    }

    public static void main(String[] args) {
        if (args.length == 0)
            System.out.println("command missing line argument!");//$NON-NLS-1$

        String command = args[0];
        int timeout = 60000;
        if (args.length == 2) {
            timeout = Integer.parseInt(args[1]);
        }

        System.out.println("command:" + command);//$NON-NLS-1$
        System.out.println("timeout:" + timeout);//$NON-NLS-1$

        Exec exec = new Exec(command);
        exec.run(timeout);
    }
}
