package com.onaro.commons.prefs;

import java.util.prefs.NodeChangeEvent;
import java.io.NotSerializableException;

/**
 * Encapsulates an event emitted by a <code>PreferencesExt</code> node to indicate
 * that a child of that node has been added or removed.<p>
 *
 * Note, that although <code>NodeExtChangeEvent</code> inherits Serializable
 * interface from java.util.EventObject, it is not intended to be Serializable.
 * Appropriate serialization methods are implemented to throw
 * <code>NotSerializableException</code>.
 */
public class NodeExtChangeEvent extends java.util.EventObject {

    /**
     * The encapsulated event.
     */
    private NodeChangeEvent event;


    /**
     * Constructs a new <code>NodeExtChangeEvent</code> instance encapsulating
     * a <code>NodeChangeEvent</code>.
     *
     * @param event the encapsulated event
     */
    public NodeExtChangeEvent(NodeChangeEvent event) {
        super(event.getSource());
        this.event = event;
    }


    /**
     * Returns the parent of the node that was added or removed.
     *
     * @return  The parent Preferences node whose child was added or removed
     */
    public PreferencesExt getParent() {
        return PreferencesExt.factory.get(event.getParent());
    }


    /**
     * Returns the node that was added or removed.
     *
     * @return  The node that was added or removed.
     */
    public PreferencesExt getChild() {
        return PreferencesExt.factory.get(event.getChild());
    }


    /**
     * Throws NotSerializableException, since NodeChangeEvent objects are not
     * intended to be serializable.
     *
     * @param out
     * @throws java.io.NotSerializableException
     */
    private void writeObject(java.io.ObjectOutputStream out)
    throws NotSerializableException {
        throw new NotSerializableException("Not serializable."); //$NON-NLS-1$
    }


    /**
     * Throws NotSerializableException, since NodeChangeEvent objects are not
     * intended to be serializable.
     *
     * @param in
     * @throws java.io.NotSerializableException
     */
    private void readObject(java.io.ObjectInputStream in)
    throws NotSerializableException {
        throw new NotSerializableException("Not serializable."); //$NON-NLS-1$
    }
}
