package com.onaro.commons.prefs;

import java.util.prefs.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.LinkedList;
import java.util.Iterator;
import java.io.*;


/**
 * An implementation of the <code>PreferencesExt</code> using Java Preferences
 * as its backend, delegating all method calls to it.
 */
public class PreferencesExtImpl extends PreferencesExt {

    /**
     * The backend preferences node.
     */
    private Preferences backEndNode;


    /**
     * Constucts a <code>PreferencesExtImpl</code> object that delegates calls
     * to the specified Java Preferences node.
     *
     * @param prefsNode the backend preferences node
     */
    public PreferencesExtImpl(Preferences prefsNode) {
        this.backEndNode = prefsNode;
    }

    /**
     * Used for verifying the keys for string values don't mix with keys of split
     * values.
     */
    private static final Pattern NUMERIC_KEY_SUFFIX = Pattern.compile("\\.[0-9]+\\z"); //$NON-NLS-1$

    private String produceBoundedLengthKey(String key) {
        if (key.length() <= Preferences.MAX_KEY_LENGTH)
            return key;
        else {
            StringBuilder boundedLengthKey = new StringBuilder(key.substring(0, Preferences.MAX_KEY_LENGTH - 20));
            return boundedLengthKey.append(key.substring(Preferences.MAX_KEY_LENGTH - 20).hashCode()).toString();
        }
    }

    /**
     * Associates the specified value with the specified key in this
     * preference node. Allows unlimited value length by splitting the value
     * into multiple keys.
     *
     * @param key key with which the specified value is to be associated.
     * @param value value to be associated with the specified key.
     * @throws NullPointerException if key or value is <tt>null</tt>.
     * @throws IllegalArgumentException if <tt>key.length()</tt> exceeds
     *       <tt>MAX_KEY_LENGTH</tt> or if the key ends with a dot and a number
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     */
    public void put(String key, String value) {
        /**
         * Verify that the key is legal.
         */
        Matcher matcher = NUMERIC_KEY_SUFFIX.matcher(key);
        if (matcher.find()) {
            throw new IllegalArgumentException("The suffix of key '" + key + "' matches suffix used for split values"); //$NON-NLS-1$ //$NON-NLS-2$
        }

        /**
         * Remove previous chunks.
         */
        try {
            String keys[] = keys();
            for (int i = 0; i < keys.length; i++) {
                String currentKey = keys[i];
                matcher = NUMERIC_KEY_SUFFIX.matcher(currentKey);
                if (matcher.find()) {
                    remove(currentKey);
                }
            }
        } catch (BackingStoreException e) {
            logger.warn("Failed to get the key list from " + name()); //$NON-NLS-1$
        }

        /**
         * Store new values.
         */
        if (value.length() > Preferences.MAX_VALUE_LENGTH) {
	    int chunkNumber = 0;
	    int length = Math.min(value.length(), Preferences.MAX_VALUE_LENGTH);
	    int begin = 0;
	    while (length > 0) {
		int end = begin + length;
		String valueChunk = value.substring(begin, end);
		backEndNode.put(chunkNumber == 0 ? produceBoundedLengthKey(key) : produceBoundedLengthKey(key) + "." + chunkNumber, valueChunk); //$NON-NLS-1$

		++chunkNumber;
		begin += length;
		length = Math.min(value.length() - end, Preferences.MAX_VALUE_LENGTH);
	    }
        }
        else {
            backEndNode.put(produceBoundedLengthKey(key), value);
        }
    }


    /**
     * Returns the value associated with the specified key in this preference
     * node.  Returns the specified default if there is no value associated
     * with the key, or the backing store is inaccessible.
     *
     * <p>Some implementations may store default values in their backing
     * stores.  If there is no value associated with the specified key
     * but there is such a <i>stored default</i>, it is returned in
     * preference to the specified default.
     *
     * @param key key whose associated value is to be returned.
     * @param def the value to be returned in the event that this
     *        preference node has no value associated with <tt>key</tt>.
     * @return the value associated with <tt>key</tt>, or <tt>def</tt>
     *         if no value is associated with <tt>key</tt>, or the backing
     *         store is inaccessible.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.  (A
     *         <tt>null</tt> value for <tt>def</tt> <i>is</i> permitted.)
     */
    public String get(String key, String def) {
        return getString(key, def);
    }


    /**
     * Returns the value associated with the specified key in this preference
     * node. It supports getting values that are longer than {@link Preferences#MAX_VALUE_LENGTH}.
     * @param key key whose associated value is to be returned.
     * @return the value associated with <tt>key</tt>
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.  (A
     *         <tt>null</tt> value for <tt>def</tt> <i>is</i> permitted.)
     * @throws PreferenceNotFoundException if the key wasn't found in the preferences
     *         or the preferences backing store wasn't available
     */
    public String get(String key) {
        String val = getString(key, "DEFAULT"); //$NON-NLS-1$

        if (val.equals("DEFAULT") && getString(key, "DEFAULT_TWO").equals("DEFAULT_TWO")) //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
            throw new PreferenceNotFoundException(backEndNode, key);

        return val;
    }

    private String getString(String key, String def) {
        String val = backEndNode.get(produceBoundedLengthKey(key), def);
        if (val == def || val.equals(def)) return def;

        StringBuilder sb = new StringBuilder(val);
        for (int i = 1; i < 100; ++i) {
            String chunk = backEndNode.get(produceBoundedLengthKey(key) + "." + i, null); //$NON-NLS-1$
            if (chunk == null) {
                break;
            }
            sb.append(chunk);
        }
        return sb.toString();
    }

    /**
     * Removes the value associated with the specified key in this preference
     * node, if any.
     *
     * <p>If this implementation supports <i>stored defaults</i>, and there is
     * such a default for the specified preference, the stored default will be
     * "exposed" by this call, in the sense that it will be returned
     * by a succeeding call to <tt>get</tt>.
     *
     * @param key key whose mapping is to be removed from the preference node.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     */
    public void remove(String key) {
        backEndNode.remove(key);
    }


    /**
     * Removes all of the preferences (key-value associations) in this
     * preference node.  This call has no effect on any descendants
     * of this node.
     *
     * <p>If this implementation supports <i>stored defaults</i>, and this
     * node in the preferences hierarchy contains any such defaults,
     * the stored defaults will be "exposed" by this call, in the sense that
     * they will be returned by succeeding calls to <tt>get</tt>.
     *
     * @throws BackingStoreException if this operation cannot be completed
     *         due to a failure in the backing store, or inability to
     *         communicate with it.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see #removeNode()
     */
    public void clear() throws BackingStoreException {
        backEndNode.clear();
    }


    /**
     * Associates a string representing the specified int value with the
     * specified key in this preference node.  The associated string is the
     * one that would be returned if the int value were passed to
     * {@link Integer#toString(int)}.  This method is intended for use in
     * conjunction with {@link #getInt}.
     *
     * @param key key with which the string form of value is to be associated.
     * @param value value whose string form is to be associated with key.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @throws IllegalArgumentException if <tt>key.length()</tt> exceeds
     *         <tt>MAX_KEY_LENGTH</tt>.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see #getInt(String,int)
     */
    public void putInt(String key, int value) {
        backEndNode.putInt(produceBoundedLengthKey(key), value);
    }


    /**
     * Returns the int value represented by the string associated with the
     * specified key in this preference node.  The string is converted to
     * an integer as by {@link Integer#parseInt(String)}.  Returns the
     * specified default if there is no value associated with the key,
     * the backing store is inaccessible, or if
     * <tt>Integer.parseInt(String)</tt> would throw a {@link
     * NumberFormatException} if the associated value were passed.  This
     * method is intended for use in conjunction with {@link #putInt}.
     *
     * <p>If the implementation supports <i>stored defaults</i> and such a
     * default exists, is accessible, and could be converted to an int
     * with <tt>Integer.parseInt</tt>, this int is returned in preference to
     * the specified default.
     *
     * @param key key whose associated value is to be returned as an int.
     * @param def the value to be returned in the event that this
     *        preference node has no value associated with <tt>key</tt>
     *        or the associated value cannot be interpreted as an int,
     *        or the backing store is inaccessible.
     * @return the int value represented by the string associated with
     *         <tt>key</tt> in this preference node, or <tt>def</tt> if the
     *         associated value does not exist or cannot be interpreted as
     *         an int.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @see #putInt(String,int)
     * @see #get(String,String)
     */
    public int getInt(String key, int def) {
        return backEndNode.getInt(produceBoundedLengthKey(key), def);
    }


    /**
     * Returns the int value represented by the string associated with the
     * specified key in this preference node.  The string is converted to
     * an integer as by {@link Integer#parseInt(String)}.
     *
     * @param key key whose associated value is to be returned as an int.
     * @return the int value represented by the string associated with
     *         <tt>key</tt> in this preference node
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @throws PreferenceNotFoundException if this preference node has no value
     *         associated with <tt>key</tt> or the associated value cannot be
     *         interpreted as an int, or the backing store is inaccessible
     */
    public int getInt(String key) {
        int val = backEndNode.getInt(produceBoundedLengthKey(key), -1);

        if (val == -1 && backEndNode.getInt(produceBoundedLengthKey(key), -2) == -2)
            throw new PreferenceNotFoundException(backEndNode, key);

        return val;
    }


    /**
     * Associates a string representing the specified long value with the
     * specified key in this preference node.  The associated string is the
     * one that would be returned if the long value were passed to
     * {@link Long#toString(long)}.  This method is intended for use in
     * conjunction with {@link #getLong}.
     *
     * @param key key with which the string form of value is to be associated.
     * @param value value whose string form is to be associated with key.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @throws IllegalArgumentException if <tt>key.length()</tt> exceeds
     *         <tt>MAX_KEY_LENGTH</tt>.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see #getLong(String,long)
     */
    public void putLong(String key, long value) {
        backEndNode.putLong(produceBoundedLengthKey(key), value);
    }


    /**
     * Returns the long value represented by the string associated with the
     * specified key in this preference node.  The string is converted to
     * a long as by {@link Long#parseLong(String)}.  Returns the
     * specified default if there is no value associated with the key,
     * the backing store is inaccessible, or if
     * <tt>Long.parseLong(String)</tt> would throw a {@link
     * NumberFormatException} if the associated value were passed.  This
     * method is intended for use in conjunction with {@link #putLong}.
     *
     * <p>If the implementation supports <i>stored defaults</i> and such a
     * default exists, is accessible, and could be converted to a long
     * with <tt>Long.parseLong</tt>, this long is returned in preference to
     * the specified default.
     *
     * @param key key whose associated value is to be returned as a long.
     * @param def the value to be returned in the event that this
     *        preference node has no value associated with <tt>key</tt>
     *        or the associated value cannot be interpreted as a long,
     *        or the backing store is inaccessible.
     * @return the long value represented by the string associated with
     *         <tt>key</tt> in this preference node, or <tt>def</tt> if the
     *         associated value does not exist or cannot be interpreted as
     *         a long.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @see #putLong(String,long)
     * @see #get(String,String)
     */
    public long getLong(String key, long def) {
        return backEndNode.getLong(produceBoundedLengthKey(key), def);
    }


    /**
     * Returns the long value represented by the string associated with the
     * specified key in this preference node.  The string is converted to
     * a long as by {@link Long#parseLong(String)}.
     *
     * @param key key whose associated value is to be returned as a long.
     * @return the long value represented by the string associated with
     *         <tt>key</tt> in this preference node
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @throws PreferenceNotFoundException if this preference node has no value
     *         associated with <tt>key</tt> or the associated value cannot be
     *         interpreted as an long, or the backing store is inaccessible
     */
    public long getLong(String key) {
        long val = backEndNode.getLong(produceBoundedLengthKey(key), -1L);

        if (val == -1L && backEndNode.getLong(produceBoundedLengthKey(key), -2L) == -2L)
            throw new PreferenceNotFoundException(backEndNode, key);

        return val;
    }


    /**
     * Associates a string representing the specified boolean value with the
     * specified key in this preference node.  The associated string is
     * <tt>"true"</tt> if the value is true, and <tt>"false"</tt> if it is
     * false.  This method is intended for use in conjunction with
     * {@link #getBoolean}.
     *
     * @param key key with which the string form of value is to be associated.
     * @param value value whose string form is to be associated with key.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @throws IllegalArgumentException if <tt>key.length()</tt> exceeds
     *         <tt>MAX_KEY_LENGTH</tt>.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see #getBoolean(String,boolean)
     * @see #get(String,String)
     */
    public void putBoolean(String key, boolean value) {
        backEndNode.putBoolean(produceBoundedLengthKey(key), value);
    }


    /**
     * Returns the boolean value represented by the string associated with the
     * specified key in this preference node.  Valid strings
     * are <tt>"true"</tt>, which represents true, and <tt>"false"</tt>, which
     * represents false.  Case is ignored, so, for example, <tt>"TRUE"</tt>
     * and <tt>"False"</tt> are also valid.  This method is intended for use in
     * conjunction with {@link #putBoolean}.
     *
     * <p>Returns the specified default if there is no value
     * associated with the key, the backing store is inaccessible, or if the
     * associated value is something other than <tt>"true"</tt> or
     * <tt>"false"</tt>, ignoring case.
     *
     * <p>If the implementation supports <i>stored defaults</i> and such a
     * default exists and is accessible, it is used in preference to the
     * specified default, unless the stored default is something other than
     * <tt>"true"</tt> or <tt>"false"</tt>, ignoring case, in which case the
     * specified default is used.
     *
     * @param key key whose associated value is to be returned as a boolean.
     * @param def the value to be returned in the event that this
     *        preference node has no value associated with <tt>key</tt>
     *        or the associated value cannot be interpreted as a boolean,
     *        or the backing store is inaccessible.
     * @return the boolean value represented by the string associated with
     *         <tt>key</tt> in this preference node, or <tt>def</tt> if the
     *         associated value does not exist or cannot be interpreted as
     *         a boolean.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @see #get(String,String)
     * @see #putBoolean(String,boolean)
     */
    public boolean getBoolean(String key, boolean def) {
        return backEndNode.getBoolean(produceBoundedLengthKey(key), def);
    }


    /**
     * Returns the boolean value represented by the string associated with the
     * specified key in this preference node.  Valid strings
     * are <tt>"true"</tt>, which represents true, and <tt>"false"</tt>, which
     * represents false.  Case is ignored, so, for example, <tt>"TRUE"</tt>
     * and <tt>"False"</tt> are also valid.  This method is intended for use in
     * conjunction with {@link #putBoolean}.
     *
     * @param key key whose associated value is to be returned as a boolean.
     * @return the boolean value represented by the string associated with
     *         <tt>key</tt> in this preference node
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @throws PreferenceNotFoundException if this preference node has no value
     *         associated with <tt>key</tt> or the associated value cannot be
     *         interpreted as an boolean, or the backing store is inaccessible
     */
    public boolean getBoolean(String key) {
        boolean val = backEndNode.getBoolean(produceBoundedLengthKey(key), false);

        if (val == false && backEndNode.getBoolean(produceBoundedLengthKey(key), true))
            throw new PreferenceNotFoundException(backEndNode, key);

        return val;
    }


    /**
     * Associates a string representing the specified float value with the
     * specified key in this preference node.  The associated string is the
     * one that would be returned if the float value were passed to
     * {@link Float#toString(float)}.  This method is intended for use in
     * conjunction with {@link #getFloat}.
     *
     * @param key key with which the string form of value is to be associated.
     * @param value value whose string form is to be associated with key.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @throws IllegalArgumentException if <tt>key.length()</tt> exceeds
     *         <tt>MAX_KEY_LENGTH</tt>.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see #getFloat(String,float)
     */
    public void putFloat(String key, float value) {
        backEndNode.putFloat(produceBoundedLengthKey(key), value);
    }


    /**
     * Returns the float value represented by the string associated with the
     * specified key in this preference node.  The string is converted to an
     * integer as by {@link Float#parseFloat(String)}.  Returns the specified
     * default if there is no value associated with the key, the backing store
     * is inaccessible, or if <tt>Float.parseFloat(String)</tt> would throw a
     * {@link NumberFormatException} if the associated value were passed.
     * This method is intended for use in conjunction with {@link #putFloat}.
     *
     * <p>If the implementation supports <i>stored defaults</i> and such a
     * default exists, is accessible, and could be converted to a float
     * with <tt>Float.parseFloat</tt>, this float is returned in preference to
     * the specified default.
     *
     * @param key key whose associated value is to be returned as a float.
     * @param def the value to be returned in the event that this
     *        preference node has no value associated with <tt>key</tt>
     *        or the associated value cannot be interpreted as a float,
     *        or the backing store is inaccessible.
     * @return the float value represented by the string associated with
     *         <tt>key</tt> in this preference node, or <tt>def</tt> if the
     *         associated value does not exist or cannot be interpreted as
     *         a float.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @see #putFloat(String,float)
     * @see #get(String,String)
     */
    public float getFloat(String key, float def) {
        return backEndNode.getFloat(produceBoundedLengthKey(key), def);
    }


    /**
     * Returns the float value represented by the string associated with the
     * specified key in this preference node.  The string is converted to an
     * integer as by {@link Float#parseFloat(String)}.
     *
     * @param key key whose associated value is to be returned as a float.
     * @return the float value represented by the string associated with
     *         <tt>key</tt> in this preference node
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @throws PreferenceNotFoundException if this preference node has no value
     *         associated with <tt>key</tt> or the associated value cannot be
     *         interpreted as an float, or the backing store is inaccessible
     */
    public float getFloat(String key) {
        float val = backEndNode.getFloat(produceBoundedLengthKey(key), -1.0F);

        if (val == -1.0F && backEndNode.getFloat(produceBoundedLengthKey(key), -2.0F) == -2.0F)
            throw new PreferenceNotFoundException(backEndNode, key);

        return val;
    }


    /**
     * Associates a string representing the specified double value with the
     * specified key in this preference node.  The associated string is the
     * one that would be returned if the double value were passed to
     * {@link Double#toString(double)}.  This method is intended for use in
     * conjunction with {@link #getDouble}.
     *
     * @param key key with which the string form of value is to be associated.
     * @param value value whose string form is to be associated with key.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @throws IllegalArgumentException if <tt>key.length()</tt> exceeds
     *         <tt>MAX_KEY_LENGTH</tt>.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see #getDouble(String,double)
     */
    public void putDouble(String key, double value) {
        backEndNode.putDouble(produceBoundedLengthKey(key), value);
    }


    /**
     * Returns the double value represented by the string associated with the
     * specified key in this preference node.  The string is converted to an
     * integer as by {@link Double#parseDouble(String)}.  Returns the specified
     * default if there is no value associated with the key, the backing store
     * is inaccessible, or if <tt>Double.parseDouble(String)</tt> would throw a
     * {@link NumberFormatException} if the associated value were passed.
     * This method is intended for use in conjunction with {@link #putDouble}.
     *
     * <p>If the implementation supports <i>stored defaults</i> and such a
     * default exists, is accessible, and could be converted to a double
     * with <tt>Double.parseDouble</tt>, this double is returned in preference
     * to the specified default.
     *
     * @param key key whose associated value is to be returned as a double.
     * @param def the value to be returned in the event that this
     *        preference node has no value associated with <tt>key</tt>
     *        or the associated value cannot be interpreted as a double,
     *        or the backing store is inaccessible.
     * @return the double value represented by the string associated with
     *         <tt>key</tt> in this preference node, or <tt>def</tt> if the
     *         associated value does not exist or cannot be interpreted as
     *         a double.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @see #putDouble(String,double)
     * @see #get(String,String)
     */
    public double getDouble(String key, double def) {
        return backEndNode.getDouble(produceBoundedLengthKey(key), def);
    }


    /**
     * Returns the double value represented by the string associated with the
     * specified key in this preference node.  The string is converted to an
     * integer as by {@link Double#parseDouble(String)}.
     *
     * @param key key whose associated value is to be returned as a double.
     * @return the double value represented by the string associated with
     *         <tt>key</tt> in this preference node
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.
     * @throws PreferenceNotFoundException if this preference node has no value
     *         associated with <tt>key</tt> or the associated value cannot be
     *         interpreted as an double, or the backing store is inaccessible
     */
    public double getDouble(String key) {
        double val = backEndNode.getDouble(produceBoundedLengthKey(key), -1.0D);

        if (val == -1.0D && backEndNode.getDouble(produceBoundedLengthKey(key), -2.0D) == -2.0D)
            throw new PreferenceNotFoundException(backEndNode, key);

        return val;
    }


    /**
     * Associates a string representing the specified byte array with the
     * specified key in this preference node.  The associated string is
     * the <i>Base64</i> encoding of the byte array, as defined in <a
     * href=http://www.ietf.org/rfc/rfc2045.txt>RFC 2045</a>, Section 6.8,
     * with one minor change: the string will consist solely of characters
     * from the <i>Base64 Alphabet</i>; it will not contain any newline
     * characters.  Note that the maximum length of the byte array is limited
     * to three quarters of <tt>MAX_VALUE_LENGTH</tt> so that the length
     * of the Base64 encoded String does not exceed <tt>MAX_VALUE_LENGTH</tt>.
     * This method is intended for use in conjunction with
     * {@link #getByteArray}.
     *
     * @param key key with which the string form of value is to be associated.
     * @param value value whose string form is to be associated with key.
     * @throws NullPointerException if key or value is <tt>null</tt>.
     * @throws IllegalArgumentException if key.length() exceeds MAX_KEY_LENGTH
     *         or if value.length exceeds MAX_VALUE_LENGTH*3/4.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see #getByteArray(String,byte[])
     * @see #get(String,String)
     */
    public void putByteArray(String key, byte[] value) {
        /**
         * Verify that the key is legal.
         */
        Matcher matcher = NUMERIC_KEY_SUFFIX.matcher(key);
        if (matcher.find()) {
            throw new IllegalArgumentException("The suffix of key '" + key + "' matches suffix used for split values"); //$NON-NLS-1$ //$NON-NLS-2$
        }

        /**
         * Remove previous chunks.
         */
        try {
            String keys[] = keys();
            for (int i = 0; i < keys.length; i++) {
                String currentKey = keys[i];
                matcher = NUMERIC_KEY_SUFFIX.matcher(currentKey);
                if (matcher.find()) {
                    remove(currentKey);
                }
            }
        } catch (BackingStoreException e) {
            logger.warn("Failed to get the key list from " + name()); //$NON-NLS-1$
        }

        /**
         * Store new values.
         */
        int maxlen = (int)(Preferences.MAX_VALUE_LENGTH*3/4);
        if (value.length > maxlen) {
            int chunkNumber = 0;
            int length = Math.min(value.length, maxlen);
            int begin = 0;
            while (length > 0) {
                int end = begin + length;
                byte valueChunk[] = new byte[end-begin];
                System.arraycopy(value,begin,valueChunk,0,end-begin);
                backEndNode.putByteArray(chunkNumber == 0 ? produceBoundedLengthKey(key) : produceBoundedLengthKey(key) + "." + chunkNumber, valueChunk); //$NON-NLS-1$

                ++chunkNumber;
                begin += length;
                length = Math.min(value.length - end, maxlen);
            }
        }
        else {
            backEndNode.putByteArray(produceBoundedLengthKey(key), value);
        }
    }


    /**
     * Associates a string representing the specified object with the
     * specified key in this preference node.
     * the bytes of the serialized objects are writen using the {@link #putByteArray(String, byte[])} method
     *
     * @param key key with which the string form of value is to be associated.
     * @param value the object to save.
     * @throws NullPointerException if key or value is <tt>null</tt>.
     * @throws IllegalArgumentException if key.length() exceeds MAX_KEY_LENGTH
     *         or if value.length exceeds MAX_VALUE_LENGTH*3/4.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see #getObject(String)
     */
    public void putObject(String key, Object value) throws IllegalArgumentException{
        try {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            ObjectOutputStream ois = new ObjectOutputStream(baos);
            ois.writeObject(value);
            ois.flush();
            putByteArray(key, baos.toByteArray());
        } catch (IOException e) {
            throw new IllegalArgumentException("Failed to write value in preferences",e); //$NON-NLS-1$
        }
    }


    /**
     * Returns the byte array value represented by the string associated with
     * the specified key in this preference node.  Valid strings are
     * <i>Base64</i> encoded binary data, as defined in <a
     * href=http://www.ietf.org/rfc/rfc2045.txt>RFC 2045</a>, Section 6.8,
     * with one minor change: the string must consist solely of characters
     * from the <i>Base64 Alphabet</i>; no newline characters or
     * extraneous characters are permitted.  This method is intended for use
     * in conjunction with {@link #putByteArray}.
     *
     * <p>Returns the specified default if there is no value
     * associated with the key, the backing store is inaccessible, or if the
     * associated value is not a valid Base64 encoded byte array
     * (as defined above).
     *
     * <p>If the implementation supports <i>stored defaults</i> and such a
     * default exists and is accessible, it is used in preference to the
     * specified default, unless the stored default is not a valid Base64
     * encoded byte array (as defined above), in which case the
     * specified default is used.
     *
     * @param key key whose associated value is to be returned as a byte array.
     * @param def the value to be returned in the event that this
     *        preference node has no value associated with <tt>key</tt>
     *        or the associated value cannot be interpreted as a byte array,
     *        or the backing store is inaccessible.
     * @return the byte array value represented by the string associated with
     *         <tt>key</tt> in this preference node, or <tt>def</tt> if the
     *         associated value does not exist or cannot be interpreted as
     *         a byte array.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>.  (A
     *         <tt>null</tt> value for <tt>def</tt> <i>is</i> permitted.)
     * @see #get(String,String)
     * @see #putByteArray(String,byte[])
     */
    public byte[] getByteArray(String key, byte[] def) {
        return backEndNode.getByteArray(produceBoundedLengthKey(key), def);
    }


    /**
     * Returns the byte array value represented by the string associated with
     * the specified key in this preference node.  Valid strings are
     * <i>Base64</i> encoded binary data, as defined in <a
     * href=http://www.ietf.org/rfc/rfc2045.txt>RFC 2045</a>, Section 6.8,
     * with one minor change: the string must consist solely of characters
     * from the <i>Base64 Alphabet</i>; no newline characters or
     * extraneous characters are permitted.  This method is intended for use
     * in conjunction with {@link #putByteArray}.
     *
     * @param key key whose associated value is to be returned as a byte array.
     * @return the byte array value represented by the string associated with
     *         <tt>key</tt> in this preference node
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>
     * @throws PreferenceNotFoundException if this preference node has no value
     *         associated with <tt>key</tt> or the associated value cannot be
     *         interpreted as a byte array, or the backing store is inaccessible
     */
    public byte[] getByteArray(String key) {
        byte val[] = backEndNode.getByteArray(produceBoundedLengthKey(key), null);
        LinkedList reslst = new LinkedList();

        if (val == null) {
            byte otherDefault[] = new byte[0];
            byte otherVal[] = backEndNode.getByteArray(produceBoundedLengthKey(key), otherDefault);

            if (otherVal != null && otherVal.length == 0)
                throw new PreferenceNotFoundException(backEndNode, key);
        }
        reslst.add(val);
        int size = val.length;

        for (int i = 1; i < 100; ++i) {
            byte chunk[] = backEndNode.getByteArray(produceBoundedLengthKey(key) + "." + i, null); //$NON-NLS-1$
            if (chunk == null) {
                break;
            }
            reslst.add(chunk);
            size += chunk.length;
        }

        byte res[] = new byte[size];
        int respos = 0;
        for (Iterator itr = reslst.iterator(); itr.hasNext() ; ) {
            byte item[] = (byte[])itr.next();
            System.arraycopy(item,0, res, respos, item.length);
            respos += item.length;
        }

        return res;
    }

    /**
     * Returns the object value represented by the string associated with
     * the specified key in this preference node.
     * first the bytes of the serialized objected are read using {@link #getBoolean(String)} and then de-serialized
     *
     * @param key key whose associated value is to be returned as an object
     * @return the object value represented by the string associated with
     *         <tt>key</tt> in this preference node
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @throws NullPointerException if <tt>key</tt> is <tt>null</tt>
     * @throws PreferenceNotFoundException if this preference node has no value
     *         associated with <tt>key</tt> or the associated value cannot be
     *         interpreted as a object, or the backing store is inaccessible
     */
    public Object getObject(String key) throws PreferenceNotFoundException {
        try {
            byte[] data = getByteArray(key);
            ByteArrayInputStream bais = new ByteArrayInputStream(data);
            ObjectInputStream ois = new ObjectInputStream(bais);
            return ois.readObject();
        } catch (IOException e) {
            logger.warn("Failed to read object from the preferances" + " - " + e.getMessage(),e); //$NON-NLS-1$ //$NON-NLS-2$
            throw new PreferenceNotFoundException(backEndNode, key);
        } catch (ClassNotFoundException e) {
            logger.warn("Failed to read object from the preferances" + " - " + e.getMessage(), e); //$NON-NLS-1$ //$NON-NLS-2$
            throw new PreferenceNotFoundException(backEndNode, key);
        }
    }


    /**
     * Returns all of the keys that have an associated value in this
     * preference node.  (The returned array will be of size zero if
     * this node has no preferences.)
     *
     * <p>If the implementation supports <i>stored defaults</i> and there
     * are any such defaults at this node that have not been overridden,
     * by explicit preferences, the defaults are returned in the array in
     * addition to any explicit preferences.
     *
     * @return an array of the keys that have an associated value in this
     *         preference node.
     * @throws BackingStoreException if this operation cannot be completed
     *         due to a failure in the backing store, or inability to
     *         communicate with it.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     */
    public String[] keys() throws BackingStoreException {
        return backEndNode.keys();
    }


    /**
     * Returns the names of the children of this preference node, relative to
     * this node.  (The returned array will be of size zero if this node has
     * no children.)
     *
     * @return the names of the children of this preference node.
     * @throws BackingStoreException if this operation cannot be completed
     *         due to a failure in the backing store, or inability to
     *         communicate with it.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     */
    public String[] childrenNames() throws BackingStoreException {
        return backEndNode.childrenNames();
    }


    /**
     * Returns the parent of this preference node, or <tt>null</tt> if this is
     * the root.
     *
     * @return the parent of this preference node.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     */
    public PreferencesExt parent() {
        return factory.get(backEndNode.parent());
    }


    /**
     * Returns the named preference node in the same tree as this node,
     * creating it and any of its ancestors if they do not already exist.
     * Accepts a relative or absolute path name.  Relative path names
     * (which do not begin with the slash character <tt>('/')</tt>) are
     * interpreted relative to this preference node.
     *
     * <p>If the returned node did not exist prior to this call, this node and
     * any ancestors that were created by this call are not guaranteed
     * to become permanent until the <tt>flush</tt> method is called on
     * the returned node (or one of its ancestors or descendants).
     *
     * @param pathName the path name of the preference node to return.
     * @return the specified preference node.
     * @throws IllegalArgumentException if the path name is invalid (i.e.,
     *         it contains multiple consecutive slash characters, or ends
     *         with a slash character and is more than one character long).
     * @throws NullPointerException if path name is <tt>null</tt>.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see #flush()
     */
    public PreferencesExt node(String pathName) {
        return factory.get(backEndNode.node(pathName));
    }


    /**
     * Returns true if the named preference node exists in the same tree
     * as this node.  Relative path names (which do not begin with the slash
     * character <tt>('/')</tt>) are interpreted relative to this preference
     * node.
     *
     * <p>If this node (or an ancestor) has already been removed with the
     * {@link #removeNode()} method, it <i>is</i> legal to invoke this method,
     * but only with the path name <tt>""</tt>; the invocation will return
     * <tt>false</tt>.  Thus, the idiom <tt>p.nodeExists("")</tt> may be
     * used to test whether <tt>p</tt> has been removed.
     *
     * @param pathName the path name of the node whose existence
     *        is to be checked.
     * @return true if the specified node exists.
     * @throws BackingStoreException if this operation cannot be completed
     *         due to a failure in the backing store, or inability to
     *         communicate with it.
     * @throws IllegalArgumentException if the path name is invalid (i.e.,
     *         it contains multiple consecutive slash characters, or ends
     *         with a slash character and is more than one character long).
     * @throws NullPointerException if path name is <tt>null</tt>.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method and
     *         <tt>pathName</tt> is not the empty string (<tt>""</tt>).
     */
    public boolean nodeExists(String pathName) throws BackingStoreException {
        return backEndNode.nodeExists(pathName);
    }


    /**
     * Removes this preference node and all of its descendants, invalidating
     * any preferences contained in the removed nodes.  Once a node has been
     * removed, attempting any method other than {@link #name()},
     * {@link #absolutePath()}, {@link #isUserNode()}, {@link #flush()} or
     * {@link #node(String) nodeExists("")} on the corresponding
     * <tt>Preferences</tt> instance will fail with an
     * <tt>IllegalStateException</tt>.  (The methods defined on {@link Object}
     * can still be invoked on a node after it has been removed; they will not
     * throw <tt>IllegalStateException</tt>.)
     *
     * <p>The removal is not guaranteed to be persistent until the
     * <tt>flush</tt> method is called on this node (or an ancestor).
     *
     * <p>If this implementation supports <i>stored defaults</i>, removing a
     * node exposes any stored defaults at or below this node.  Thus, a
     * subsequent call to <tt>nodeExists</tt> on this node's path name may
     * return <tt>true</tt>, and a subsequent call to <tt>node</tt> on this
     * path name may may return a (different) <tt>Preferences</tt> instance
     * representing a non-empty collection of preferences and/or children.
     *
     * @throws IllegalStateException if this node (or an ancestor) has already
     *         been removed with the {@link #removeNode()} method.
     * @throws UnsupportedOperationException if this method is invoked on
     *         the root node.
     * @see #flush()
     */
    public void removeNode() throws BackingStoreException {
        backEndNode.removeNode();
    }


    /**
     * Returns this preference node's name, relative to its parent.
     *
     * @return this preference node's name, relative to its parent.
     */
    public String name() {
        return backEndNode.name();
    }


    /**
     * Returns this preference node's absolute path name.
     *
     * @return this preference node's absolute path name.
     */
    public String absolutePath() {
        return backEndNode.absolutePath();
    }


    /**
     * Returns <tt>true</tt> if this preference node is in the user
     * preference tree, <tt>false</tt> if it's in the system preference tree.
     *
     * @return <tt>true</tt> if this preference node is in the user
     *         preference tree, <tt>false</tt> if it's in the system
     *         preference tree.
     */
    public boolean isUserNode() {
        return backEndNode.isUserNode();
    }


    /**
     * Returns a string representation of this preferences node,
     * as if computed by the expression:<tt>(this.isUserNode() ? "User" :
     * "System") + " Preference Node: " + this.absolutePath()</tt>.
     */
    public String toString() {
        return backEndNode.toString();
    }


    /**
     * Forces any changes in the contents of this preference node and its
     * descendants to the persistent store.  Once this method returns
     * successfully, it is safe to assume that all changes made in the
     * subtree rooted at this node prior to the method invocation have become
     * permanent.
     *
     * <p>Implementations are free to flush changes into the persistent store
     * at any time.  They do not need to wait for this method to be called.
     *
     * <p>When a flush occurs on a newly created node, it is made persistent,
     * as are any ancestors (and descendants) that have yet to be made
     * persistent.  Note however that any preference value changes in
     * ancestors are <i>not</i> guaranteed to be made persistent.
     *
     * @throws BackingStoreException if this operation cannot be completed
     *         due to a failure in the backing store, or inability to
     *         communicate with it.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see    #sync()
     */
    public void flush() throws BackingStoreException {
        backEndNode.flush();
    }


    /**
     * Ensures that future reads from this preference node and its
     * descendants reflect any changes that were committed to the persistent
     * store (from any VM) prior to the <tt>sync</tt> invocation.  As a
     * side-effect, forces any changes in the contents of this preference node
     * and its descendants to the persistent store, as if the <tt>flush</tt>
     * method had been invoked on this node.
     *
     * @throws BackingStoreException if this operation cannot be completed
     *         due to a failure in the backing store, or inability to
     *         communicate with it.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see    #flush()
     */
    public void sync() throws BackingStoreException {
        backEndNode.sync();
    }


    /**
     * Registers the specified listener to receive <i>preference change
     * events</i> for this preference node.  A preference change event is
     * generated when a preference is added to this node, removed from this
     * node, or when the value associated with a preference is changed.
     * (Preference change events are <i>not</i> generated by the {@link
     * #removeNode()} method, which generates a <i>node change event</i>.
     * Preference change events <i>are</i> generated by the <tt>clear</tt>
     * method.)
     *
     * <p>Events are only guaranteed for changes made within the same JVM
     * as the registered listener, though some implementations may generate
     * events for changes made outside this JVM.  Events may be generated
     * before the changes have been made persistent.  Events are not generated
     * when preferences are modified in descendants of this node; a caller
     * desiring such events must register with each descendant.
     *
     * @param pcl The preference change listener to add.
     * @throws NullPointerException if <tt>pcl</tt> is null.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see #removePreferenceChangeListener(PreferenceExtChangeListener)
     * @see #addNodeChangeListener(NodeExtChangeListener)
     * @todo implement this
     */
    public void addPreferenceChangeListener(PreferenceExtChangeListener pcl) {
        throw new UnsupportedOperationException("Not implemented yet"); //$NON-NLS-1$
    }


    /**
     * Removes the specified preference change listener, so it no longer
     * receives preference change events.
     *
     * @param pcl The preference change listener to remove.
     * @throws IllegalArgumentException if <tt>pcl</tt> was not a registered
     *         preference change listener on this node.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see #addPreferenceChangeListener(PreferenceExtChangeListener)
     * @todo implement this
     */
    public void removePreferenceChangeListener(PreferenceExtChangeListener pcl) {
        throw new UnsupportedOperationException("Not implemented yet"); //$NON-NLS-1$
    }


    /**
     * Registers the specified listener to receive <i>node change events</i>
     * for this node.  A node change event is generated when a child node is
     * added to or removed from this node.  (A single {@link #removeNode()}
     * invocation results in multiple <i>node change events</i>, one for every
     * node in the subtree rooted at the removed node.)
     *
     * <p>Events are only guaranteed for changes made within the same JVM
     * as the registered listener, though some implementations may generate
     * events for changes made outside this JVM.  Events may be generated
     * before the changes have become permanent.  Events are not generated
     * when indirect descendants of this node are added or removed; a
     * caller desiring such events must register with each descendant.
     *
     * <p>Few guarantees can be made regarding node creation.  Because nodes
     * are created implicitly upon access, it may not be feasible for an
     * implementation to determine whether a child node existed in the backing
     * store prior to access (for example, because the backing store is
     * unreachable or cached information is out of date).  Under these
     * circumstances, implementations are neither required to generate node
     * change events nor prohibited from doing so.
     *
     * @param ncl The <tt>NodeChangeListener</tt> to add.
     * @throws NullPointerException if <tt>ncl</tt> is null.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see #removeNodeChangeListener(NodeExtChangeListener)
     * @see #addPreferenceChangeListener(PreferenceExtChangeListener)
     * @todo implement this
     */
    public void addNodeChangeListener(NodeExtChangeListener ncl) {
        throw new UnsupportedOperationException("Not implemented yet"); //$NON-NLS-1$
    }


    /**
     * Removes the specified <tt>NodeChangeListener</tt>, so it no longer
     * receives change events.
     *
     * @param ncl The <tt>NodeChangeListener</tt> to remove.
     * @throws IllegalArgumentException if <tt>ncl</tt> was not a registered
     *         <tt>NodeChangeListener</tt> on this node.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see #addNodeChangeListener(NodeExtChangeListener)
     * @todo implement this
     */
    public void removeNodeChangeListener(NodeExtChangeListener ncl) {
        throw new UnsupportedOperationException("Not implemented yet"); //$NON-NLS-1$
    }


    /**
     * Emits on the specified output stream an XML document representing all
     * of the preferences contained in this node (but not its descendants).
     * This XML document is, in effect, an offline backup of the node.
     *
     * <p>The XML document will have the following DOCTYPE declaration:
     * <pre>
     * &lt;!DOCTYPE preferences SYSTEM "http://java.sun.com/dtd/preferences.dtd"&gt;
     * </pre>
     * The UTF-8 character encoding will be used.
     *
     * <p>This method is an exception to the general rule that the results of
     * concurrently executing multiple methods in this class yields
     * results equivalent to some serial execution.  If the preferences
     * at this node are modified concurrently with an invocation of this
     * method, the exported preferences comprise a "fuzzy snapshot" of the
     * preferences contained in the node; some of the concurrent modifications
     * may be reflected in the exported data while others may not.
     *
     * @param os the output stream on which to emit the XML document.
     * @throws IOException if writing to the specified output stream
     *         results in an <tt>IOException</tt>.
     * @throws BackingStoreException if preference data cannot be read from
     *         backing store.
     * @see    #importPreferences(InputStream)
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     */
    public void exportNode(OutputStream os) throws IOException, BackingStoreException {
        backEndNode.exportNode(os);
    }


    /**
     * Emits an XML document representing all of the preferences contained
     * in this node and all of its descendants.  This XML document is, in
     * effect, an offline backup of the subtree rooted at the node.
     *
     * <p>The XML document will have the following DOCTYPE declaration:
     * <pre>
     * &lt;!DOCTYPE preferences SYSTEM "http://java.sun.com/dtd/preferences.dtd"&gt;
     * </pre>
     * The UTF-8 character encoding will be used.
     *
     * <p>This method is an exception to the general rule that the results of
     * concurrently executing multiple methods in this class yields
     * results equivalent to some serial execution.  If the preferences
     * or nodes in the subtree rooted at this node are modified concurrently
     * with an invocation of this method, the exported preferences comprise a
     * "fuzzy snapshot" of the subtree; some of the concurrent modifications
     * may be reflected in the exported data while others may not.
     *
     * @param os the output stream on which to emit the XML document.
     * @throws IOException if writing to the specified output stream
     *         results in an <tt>IOException</tt>.
     * @throws BackingStoreException if preference data cannot be read from
     *         backing store.
     * @throws IllegalStateException if this node (or an ancestor) has been
     *         removed with the {@link #removeNode()} method.
     * @see    #importPreferences(InputStream)
     * @see    #exportNode(OutputStream)
     */
    public void exportSubtree(OutputStream os) throws IOException, BackingStoreException {
        backEndNode.exportSubtree(os);
    }

}
