package com.onaro.sanscreen.acquisition.cli;

import com.onaro.sanscreen.acquisition.cli.commands.Command;
import com.onaro.sanscreen.acquisition.cli.commands.CommandType;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.Map;

/**
 * Implements the interactive use case (login once into a session and then simply execute command after another)
 *
 * User and password are provided only once when entering the interactive more
 */
public class InteractiveMode {
    public static void main(final String[] args) {
        Runtime.getRuntime().addShutdownHook(new RunWhenShuttingDown());

        final CLI cli = new CLI();

        try {
            // First parse the argument... they have to include user and password
            final Map<String, String> arguments = cli.parseArguments(args, 0);

            if (arguments.size() != 2) {
                throw new IllegalArgumentException("2 arguments (user and password are required)");
            }

            cli.checkMandatoryArguments(arguments, true, "user", "password");

            final String user = arguments.get("user");
            final String password = arguments.get("password");

            // Login (once) into the server
            cli.login(user, password);
        }
        catch (IllegalArgumentException e) {
            if (e != null) {
                System.out.println(e.getMessage());
                System.out.println();
            }

            System.out.println("Usage:");
            System.out.println("======");
            System.out.println("acqcli_interactive -user=<user> -password=<password>");

            System.exit(1);
        }
        catch (final Exception e) {
            System.err.println("Error: " + e.getMessage());
            System.exit(1);
        }

        System.out.println("Entering interactive mode... (type 'q' to exit)");
        System.out.println();

        final BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));

        // Now start the loop of the command line parsing...
        boolean stopped = false;
        while (!stopped) {
            try {
                System.out.print(">> ");

                final String line = reader.readLine().trim();

                if (line.length() == 0) {
                    continue;
                }

                // Check if we are stopped
                if (line.equals("q") || line.equalsIgnoreCase("quit")) {
                    break;
                }

                final String[] params = line.split("\\s+");

                // First argument has to be a command
                if (params.length < 1) {
                    throw new IllegalArgumentException("No arguments provided");
                }

                final Command command = CommandType.fromCommandString(params[0]).getCommand();
                final Map<String, String> arguments = cli.parseArguments(params, 1);

                cli.run(command, arguments);
                System.out.println();
            }
            catch (IllegalArgumentException iae) {
                iae.printStackTrace();

                showHelp();
            }
            catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    public static void showHelp() {
        System.out.println("Usage:");
        System.out.println("======");
        System.out.println("<command> [-argname=value ...]");

        final String formatString = CommandLineMode.showCommandHelp();

        // Manually add the additional option to exit interactive mode
        System.out.format(formatString, "q | quit", "Exit");
    }

    private static class RunWhenShuttingDown extends Thread {
        @Override
        public void run() {
            System.out.println();
            System.out.println("Goodbye...");
        }
    }
}
