package com.onaro.sanscreen.acquisition.cli.commands;

import com.onaro.sanscreen.acquisition.cli.CLI;
import com.onaro.sanscreen.acquisition.sessions.AcquisitionFacadeSession;
import com.onaro.sanscreen.acquisition.sessions.AcquisitionUnitException;
import com.onaro.sanscreen.server.dataobject.acquisition.AcquisitionUnitNew;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceDefaults;
import com.onaro.util.Properties2XMLSerializer;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import java.io.*;
import java.util.*;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

@SuppressWarnings({"deprecation"})
public class AddDsCommand implements Command {
    @Override
    public String usageString() {
        return "add [-au=<auname>] -type=<type> -name=<dsname> [-<propname1>=<propvalue1> ...]";
    }

    @Override
    public String usageDescription() {
        return "Add a new data source to the system";
    }


    @Override
    public void run(final CLI cli, final Map<String, String> args) throws Exception {
        if (!args.containsKey("type")) {
            args.put("type", "91");      // We are OCIE
        }
        cli.checkMandatoryArguments(args, true, "name");

        int dsType = Integer.parseInt(args.remove("type"));
        final String name = args.remove("name");
        final String auName = args.containsKey("au") ? args.remove("au") : "local";

        // Look up the vendor model for this type...
        final AcquisitionFacadeSession facadeSession = cli.getSession(AcquisitionFacadeSession.JNDI_NAME);

        final AcquisitionUnitNew acqUnit = facadeSession.getAcquisitionUnit(auName);
        if (acqUnit == null) {
            throw new IllegalArgumentException("Could not find acq unit " + auName);
        }


        // If we are creating a data source from a recording ZIP, we use properties from the ZIP
        final String RECORDING = "recording";
        final String NETAPPFOUNDATION = "netappfoundation";
        if (args.containsKey(RECORDING)) {
            final String recordingFileName = args.get(RECORDING);
            args.remove(RECORDING);
            final File recordingFile = new File(recordingFileName);
            if (!recordingFile.exists()) {
                throw new IllegalArgumentException("Data source recording " + recordingFileName + " does not exist.");
            }
            args.put("transport.simulate", "true");
            final String recordingFilePath = recordingFile.getParent();
            args.put("transport.simulate.dir", recordingFilePath);
            final String recordingFileShortName = recordingFile.getName();
            args.put("transport.simulate.file", recordingFileShortName);
            final ZipFile recordingZipFile = new ZipFile(recordingFile);
            ZipEntry recordedConfigFile = recordingZipFile.getEntry("root\\configuration");
            if (recordedConfigFile == null) {
                recordedConfigFile = recordingZipFile.getEntry("root/configuration");
                if (recordedConfigFile == null) {
                    throw new IllegalArgumentException("Data source recording " + recordingFileName + " is invalid, does not contain root/configuration.");
                }
            }
            InputStream configInputStream = null;
            final Properties recordedProperties = new Properties();
            try {
                configInputStream = recordingZipFile.getInputStream(recordedConfigFile);
                recordedProperties.loadFromXML(configInputStream);

            } catch (Exception e) {
                throw new IllegalArgumentException("Failed to read " + recordingFileName + ", caught " + e);
            } finally {
                if (configInputStream != null) {
                    try {
                        configInputStream.close();
                    } catch (Exception ignore) {}
                }
            }
            for (final String propName : recordedProperties.stringPropertyNames()) {
                // Recorded properties override user input properties
                if (propName.equals("system.datasource.type")) {
                    dsType = Integer.parseInt(recordedProperties.getProperty(propName));
                    continue;
                }
                if (propName.startsWith("system.")) {
                    continue;
                }
                final String value = recordedProperties.getProperty(propName);
                final String fullyQualifiedPropName = NETAPPFOUNDATION + '.' + propName;
                args.put(fullyQualifiedPropName, value);

            }
            final String packageName = recordedProperties.getProperty("system.package.name");
            // If this is a foundation recording then we must disable performance to avoid failure due to missing performance in recording
            if (NETAPPFOUNDATION.equals(packageName)) {
                args.put("netappstorageperformance.enabled", "false");
            }
        }

        // Get properties... (iterate over all remaining properties and add them in)
        final Properties userProperties = new Properties();
        for (final Map.Entry<String, String> argEntry : args.entrySet()) {
            userProperties.setProperty(argEntry.getKey(), argEntry.getValue());
        }

        final com.onaro.sanscreen.server.dataobject.acquisition.DataSourceVendorModel[] vendorModels = facadeSession.findAllVmByDataSourceType(dsType);
        if (vendorModels == null || vendorModels.length == 0) {
            throw new IllegalArgumentException("Data source type " + dsType + " is not recognized");
        }

        final String vendor = vendorModels[0].getVendorName();
        final String model = vendorModels[0].getModelName();



        // Make sure all mandatory properties are present if they don't have a default value
        final List<DSConfigAttribute> mandatoryAttributes = getDataSourceAttributes(facadeSession, dsType, true);
        for (final DSConfigAttribute mandatoryAttribute : mandatoryAttributes) {
            if (null == mandatoryAttribute.getDefaultValue() && !userProperties.containsKey(mandatoryAttribute.getName())) {

                String errorMessage = "Missing required attribute \"" + mandatoryAttribute.getName() + "\"\n" +
                        "Data source type " + dsType + " requires the following attributes.\n";

                for (final DSConfigAttribute mandatoryAttribute2 : mandatoryAttributes) {

                    errorMessage += mandatoryAttribute2.getName() + "(" + mandatoryAttribute2.getType() + ") : " + mandatoryAttribute2.getDescription();
                    errorMessage += "\n";
                }

                throw new IllegalArgumentException(errorMessage);

            }
        }

        // And now we're ready to add the data source
        final long dsId = facadeSession.addDS(auName, name, vendor, model, dsType, 1, false, userProperties);
        facadeSession.waitForDataSourceStart(dsId);

        System.out.println("Data source (with ID " + dsId + ") added.");
    }

    /**
     * Get a list of DSConfigAttribute for a given datasource type by parsing the XML properties
     *
     * @param datasourceTypeID The type id of the desired datasource
     *
     * @return A list of DSConfigAttribute objects for this datasource type
     */
    @SuppressWarnings("unchecked")
    public List<DSConfigAttribute> getDataSourceAttributes(final AcquisitionFacadeSession facadeSession, final int datasourceTypeID, final boolean mandatoryOnly) throws Exception {

        // Create a list of DSConfigAttribute objects to hold the default attributes
        final List<DSConfigAttribute> dataSourceConfigAttributes = new ArrayList<DSConfigAttribute>();

        try {
            // Iterate over the DataSourceDefaults
            for (final DataSourceDefaults ds : facadeSession.getDsDefaults()) {

                // If this DataSourceDefault is the one we are interested in...
                if (ds.getDataSourceType() == datasourceTypeID) {

                    // Parse the serialized XML string into a list that can be iterated over
                    final Properties defaultProperties = Properties2XMLSerializer.deserializeProperties(ds.getProperties());
                    final String xmlProperties = (String) defaultProperties.get("____");
                    final Element root = parseDeployDescriptor((new ByteArrayInputStream(xmlProperties.getBytes())), ds.getName());
                    final List<Element> propertyElements = (List<Element>) root.getChildren("property");

                    // Iterate over each property
                    for (final Element e : propertyElements) {

                        // Create a DSConfigAttribute object
                        final DSConfigAttribute configAttribute = new DSConfigAttribute();

                        // Populate the DSConfigAttribute with values from the properties
                        // that were de-serialized from the XML
                        configAttribute.setName(e.getAttributeValue("name"));
                        configAttribute.setDescription((e.getAttributeValue("description")));
                        configAttribute.setHidden(Boolean.valueOf(e.getAttributeValue("hidden")));
                        configAttribute.setMandatory(Boolean.valueOf(e.getAttributeValue("mandatory")));
                        configAttribute.setType(e.getAttributeValue("type"));
                        final String defaultValue = e.getAttributeValue("value");
                        if (defaultValue != null && !defaultValue.isEmpty()) {
                            configAttribute.setDefaultValue(defaultValue);
                        }

                        // Add this DSConfigAttribute to the list of them
                        if ((mandatoryOnly && configAttribute.isMandatory()) || !mandatoryOnly) {
                            dataSourceConfigAttributes.add(configAttribute);
                        }
                    }
                }
            }
        }
        catch (AcquisitionUnitException ex) {
            throw new Exception("Cannot retrieve default properties");
        }

        return dataSourceConfigAttributes;
    }

    /**
     * Parse the deployment descriptor
     *
     * @param deployeeFile The deployee file
     * @param path         The path to the file
     *
     * @return XML Element
     */
    protected Element parseDeployDescriptor(final InputStream deployeeFile, final String path) throws Exception {
        try {
            // Parse the  XML
            final SAXBuilder builder = new SAXBuilder(com.sun.org.apache.xerces.internal.parsers.SAXParser.class.getName());
            final Document doc = builder.build(deployeeFile);

            return doc.getRootElement();
        }
        catch (JDOMException e) {
            throw new Exception("Failed to parse  XML to retrieve properties " + path, e);
        }
        catch (IOException e) {
            throw new Exception("Failed to parse  XML to retrieve properties " + path, e);
        }
    }

    private class DSConfigAttribute {
        private String name;
        private boolean mandatory = false;
        private String description;
        private String type;
        private String minValue;
        private String maxValue;
        private String enumType;
        private String defaultValue;
        private boolean hidden = false;

        @SuppressWarnings({"UnusedDeclaration"})
        public boolean isHidden() {
            return hidden;
        }

        public void setHidden(final boolean hidden) {
            this.hidden = hidden;
        }

        @SuppressWarnings({"UnusedDeclaration"})
        public String getEnumType() {

            return enumType;
        }

        @SuppressWarnings({"UnusedDeclaration"})
        public void setEnumType(final String enumType) {
            this.enumType = enumType;
        }

        public String getType() {
            return type;
        }

        public void setType(final String type) {
            this.type = type;
        }

        @SuppressWarnings({"UnusedDeclaration"})
        public String getMinValue() {
            return minValue;
        }

        @SuppressWarnings({"UnusedDeclaration"})
        public void setMinValue(final String minValue) {
            this.minValue = minValue;
        }

        @SuppressWarnings({"UnusedDeclaration"})
        public String getMaxValue() {
            return maxValue;
        }

        @SuppressWarnings({"UnusedDeclaration"})
        public void setMaxValue(final String maxValue) {
            this.maxValue = maxValue;
        }

        public String getName() {
            return name;
        }

        public void setName(final String name) {
            this.name = name;
        }

        public boolean isMandatory() {
            return mandatory;
        }

        public void setMandatory(final boolean mandatory) {
            this.mandatory = mandatory;
        }

        public String getDescription() {
            return description;
        }

        public void setDescription(final String description) {
            this.description = description;
        }

        @SuppressWarnings({"UnusedDeclaration"})
        public String getDefaultValue() {
            return defaultValue;
        }

        public void setDefaultValue(final String defaultValue) {
            this.defaultValue = defaultValue;
        }

    }

}
