/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

 package com.onaro.sanscreen.acquisition.cli.commands;

import com.onaro.sanscreen.acquisition.cli.CLI;
import com.onaro.sanscreen.acquisition.sessions.AcquisitionFacadeSession;
import com.onaro.sanscreen.server.dataobject.acquisition.AcquisitionUnitNew;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceNew;

import java.beans.XMLEncoder;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.util.Map;
import java.util.Properties;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

public class UpdateDsPropertyCommand implements Command {
    @Override
    public String usageString() {
        return "update -id=<dsid> -<attribute>=<value> [...]";
    }

    @Override
    public String usageDescription() {
        return "Update a data source attribute value";
    }


    @Override
    public void run(final CLI cli, final Map<String, String> args) throws Exception {
        cli.checkMandatoryArguments(args, true, "id");
        final long dsid = Long.parseLong(args.remove("id"));

        final AcquisitionFacadeSession facadeSession = cli.getSession(AcquisitionFacadeSession.JNDI_NAME);

        // Get the data source (identified by the ID)
        final DataSourceNew dataSource = facadeSession.getDsById(dsid);
        final Properties currentProperties = dataSource.getAttrs();

        // Get the acquisition unit (to know its name)
        final AcquisitionUnitNew acqUnit = facadeSession.getAcquisitionUnitsByID().get(dataSource.getAuId());

        // If we are creating a data source from a recording ZIP, we use properties from the ZIP
        final String RECORDING = "recording";
        final String NETAPPFOUNDATION = "netappfoundation";
        if (args.containsKey(RECORDING)) {
            final String recordingFileName = args.get(RECORDING);
            args.remove(RECORDING);
            final File recordingFile = new File(recordingFileName);
            if (!recordingFile.exists()) {
                throw new IllegalArgumentException("Data source recording " + recordingFileName + " does not exist.");
            }
            args.put("transport.simulate", "true");
            final String recordingFilePath = recordingFile.getParent();
            args.put("transport.simulate.dir", recordingFilePath);
            final String recordingFileShortName = recordingFile.getName();
            args.put("transport.simulate.file", recordingFileShortName);
            final ZipFile recordingZipFile = new ZipFile(recordingFile);
            ZipEntry recordedConfigFile = recordingZipFile.getEntry("root\\configuration");
            if (recordedConfigFile == null) {
                recordedConfigFile = recordingZipFile.getEntry("root/configuration");
                if (recordedConfigFile == null) {
                    throw new IllegalArgumentException("Data source recording " + recordingFileName + " is invalid, does not contain root/configuration.");
                }
            }
            InputStream configInputStream = null;
            final Properties recordedProperties = new Properties();
            try {
                configInputStream = recordingZipFile.getInputStream(recordedConfigFile);
                recordedProperties.loadFromXML(configInputStream);

            } catch (Exception e) {
                throw new IllegalArgumentException("Failed to read " + recordingFileName + ", caught " + e);
            } finally {
                if (configInputStream != null) {
                    try {
                        configInputStream.close();
                    } catch (Exception ignore) {}
                }
            }
            for (final String propName : recordedProperties.stringPropertyNames()) {
                // Recorded properties override user input properties
                if (propName.equals("system.datasource.type")) {
                    continue;
                }
                if (propName.startsWith("system.")) {
                    continue;
                }
                final String value = recordedProperties.getProperty(propName);
                final String fullyQualifiedPropName = NETAPPFOUNDATION + '.' + propName;
                args.put(fullyQualifiedPropName, value);

            }
            final String packageName = recordedProperties.getProperty("system.package.name");
            // If this is a foundation recording then we must disable performance to avoid failure due to missing performance in recording
            if (NETAPPFOUNDATION.equals(packageName)) {
                args.put("netappstorageperformance.enabled", "false");
            }
        }

        for (final Map.Entry<String, String> argEntry : args.entrySet()) {
            currentProperties.setProperty(argEntry.getKey(), argEntry.getValue());
        }

        facadeSession.updateProp(acqUnit.getName(), dsid, currentProperties);

        System.out.println("Data source updated");
    }
}