package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;


import java.math.BigInteger;
import java.util.Collection;
import java.util.Map;

import com.netapp.oci.netapp.performance.disk.interfaces.data.DiskMeasurements;
import com.netapp.oci.netapp.performance.disk.interfaces.data.DiskRawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster.ClusterKey;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

public class DiskMetricsBuilder extends ArchiveMetricsBuilderBase<Cluster.ClusterKey, DiskMeasurements, DiskRawMeasurements, DiskMetricsBuilder.DiskCounter>{
	
	
	public static enum DiskCounter implements ArchiveCounterMetaData {
		total_transfers,
		user_reads,
		user_writes,
		cp_reads,
		disk_busy,
		base_for_disk_busy,
		cp_read_latency,
		cp_read_blocks,
		guaranteed_reads,
		guaranteed_writes,
		user_read_latency,
		user_read_blocks,
		user_write_latency,
		user_write_blocks;
		
		 
		 private String onTapCounterName;
		 private int onTapCounterIndex;
		 
		 private DiskCounter(final String onTapCounterName, final int onTapCounterIndex) {
	            this.onTapCounterName = onTapCounterName;
	            this.onTapCounterIndex = onTapCounterIndex;
	        }

	        private DiskCounter() {
	            this.onTapCounterName = this.name();
	            this.onTapCounterIndex = -1;
	        }
		 
		 
		 @Override
		 public String getOnTapCounterName() {
			 return onTapCounterName;
		 }

		 @Override
		 public int getOnTapCounterIndex() {
			 return onTapCounterIndex;
		 }
	}
	
	 public DiskMetricsBuilder(NetAppOCIEPerformancePackage reportingPackage) {
		super(reportingPackage, "disk", DiskCounter.class);
	}

	@Override
	protected ClusterKey getParentInventoryUUIDKey(
			NetAppDefaultStorageContext context, String perfInstanceUUID,
			String nodeName, Map<String, Object> counterRow) {
		return context.getKey();
	}
	
	@Override
	protected String getPerformanceUUID(final NetAppDefaultStorageContext context, final String nodeName,
            final Map<String, Object> counterRow) {
		String diskName = (String)counterRow.get("instance_name");
		Map<String, String> diskNameToUUIDMap = context.getDiskNameToUUIDMap();
		if(diskName != null){
			String diskUUID = diskNameToUUIDMap.get(diskName);
			if(diskUUID != null){
				return diskUUID;
			}
		}
		return null;
		
	}
	
	@Override
	public void reportPerformance(
			ClusterKey clusterKey,
			Map<ClusterKey, Collection<RawSample<DiskMeasurements>>> measurements)
			throws DataSourceErrorException {
		if (measurements == null) {
            return;
        }
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<DiskMeasurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<DiskMeasurements>> nodeMeasurements = measurementEntry.getValue();
            reportingPackage.reportDiskPerformance(key, nodeMeasurements);
        }	
		
	}

	@Override
	protected DiskCounter[] getCountersMetaData(
			NetAppDefaultStorageContext context) {
		return DiskCounter.values();
	}

	@Override
	protected DiskMeasurements normalizeMetrics(
			ArchiveMetricsScratchPad<DiskCounter> currentScratchPad,
			NetAppDefaultStorageContext context) {
		DiskMeasurements m = new DiskMeasurements();
		m.setDiskBusy(currentScratchPad.getPercentFromDeltas(DiskCounter.disk_busy, DiskCounter.base_for_disk_busy));
		m.setCpReads(currentScratchPad.getCounterRatePerSec(DiskCounter.cp_reads));
		m.setUserReads(currentScratchPad.getCounterRatePerSec(DiskCounter.user_reads));
		m.setUserWrites(currentScratchPad.getCounterRatePerSec(DiskCounter.user_writes));
		m.setTotalTransfers(currentScratchPad.getCounterRatePerSec(DiskCounter.total_transfers));
		m.setBaseForDiskBusy(currentScratchPad.getCounterDeltaFinal(DiskCounter.base_for_disk_busy));
		m.setCpReadLatency(currentScratchPad.getLatencyPerOp(DiskCounter.cp_read_latency, DiskCounter.cp_reads));
		m.setCpReadBlocks(currentScratchPad.getCounterRatePerSec(DiskCounter.cp_read_blocks));
		m.setGuaranteedReads(currentScratchPad.getCounterRatePerSec(DiskCounter.guaranteed_reads));
		m.setGuaranteedWrites(currentScratchPad.getCounterRatePerSec(DiskCounter.guaranteed_writes));
		m.setUserReadBlocks(currentScratchPad.getCounterRatePerSec(DiskCounter.user_read_blocks));
		m.setUserWriteBlocks(currentScratchPad.getCounterRatePerSec(DiskCounter.user_write_blocks));
		m.setUserReadLatency(currentScratchPad.getLatencyPerOp(DiskCounter.user_read_latency, DiskCounter.user_reads));
		m.setUserWriteLatency(currentScratchPad.getLatencyPerOp(DiskCounter.user_write_latency, DiskCounter.user_writes));
		//Synthetic Counters
		m.setOpmDiskThroughput(MetricsRowScratchPad.add(m.getUserReadBlocks(), m.getUserWriteBlocks(), m.getCpReadBlocks()));
		m.setTotalOps(MetricsRowScratchPad.add(m.getUserReads(), m.getUserWrites(), m.getCpReads()));
		final float totalOpsDelta = MetricsRowScratchPad.add(currentScratchPad.getCounterDelta(DiskCounter.user_reads), currentScratchPad.getCounterDelta(DiskCounter.user_writes), 
				currentScratchPad.getCounterDelta(DiskCounter.cp_reads));
		m.setOpmDiskLatency(getLatency(currentScratchPad, DiskCounter.user_read_latency, DiskCounter.user_write_latency, DiskCounter.cp_read_latency, totalOpsDelta));
		
		//TODO: What is the status on these counters?
		m.setGuaranteedReadBlocks(0f);
        m.setGuaranteedReadLatency(0f);
        m.setGuaranteedWriteBlocks(0f);
        m.setGuaranteedWriteLatency(0f);
		return m;
	}

	@Override
	protected DiskRawMeasurements rawMetrics(
			ArchiveMetricsScratchPad<DiskCounter> currentScratchPad,
			NetAppDefaultStorageContext context) {
		Map<DiskCounter,BigInteger> counterValues = currentScratchPad.getCounterValues();
		DiskRawMeasurements m = new DiskRawMeasurements();
		m.setBaseForDiskBusy(counterValues.get(DiskCounter.base_for_disk_busy));
		m.setCpReadBlocks(counterValues.get(DiskCounter.cp_read_blocks));
		m.setUserWrites(counterValues.get(DiskCounter.user_writes));
		m.setUserReads(counterValues.get(DiskCounter.user_reads));
		m.setTotalTransfers(counterValues.get(DiskCounter.total_transfers));
		m.setDiskBusy(counterValues.get(DiskCounter.disk_busy));
		m.setCpReadLatency(counterValues.get(DiskCounter.cp_read_latency));
		m.setUserWriteBlocks(counterValues.get(DiskCounter.user_write_blocks));
		m.setGuarenteedReads(counterValues.get(DiskCounter.guaranteed_reads));
		m.setGuarenteedWrites(counterValues.get(DiskCounter.guaranteed_writes));
		m.setCpReads(counterValues.get(DiskCounter.cp_reads));
		m.setUserReadBlocks(counterValues.get(DiskCounter.user_read_blocks));
		m.setUserWriteBlocks(counterValues.get(DiskCounter.user_write_blocks));
		m.setUserReadLatency(counterValues.get(DiskCounter.user_read_latency));
		m.setUserWriteLatency(counterValues.get(DiskCounter.user_write_latency));
		//TODO: status on these?
		m.setGuarenteedReadBlocks(null);
        m.setGuarenteedReadLatency(null);
        m.setGuarenteedWriteLatency(null);
		//Synthetic Counter
        m.setOpmDiskThroughput(MetricsRowScratchPad.add(m.getUserReadBlocks(), m.getUserWriteBlocks(), m.getCpReadBlocks()));
        m.setTotalOps(MetricsRowScratchPad.add(m.getUserReads(), m.getUserWrites(), m.getCpReads()));
        m.setOpmDiskLatency(MetricsRowScratchPad.add(m.getUserReadLatency(), m.getUserWriteLatency(), m.getCpReadLatency()));
		return m;
	}
	
	public void reportRawPerformance(
			ClusterKey clusterKey,
			Map<ClusterKey, Collection<RawSample<DiskRawMeasurements>>> rawMeasurements)
			throws DataSourceErrorException {
		for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<DiskRawMeasurements>>> measurementEntry : rawMeasurements.entrySet()) { 
			final Cluster.ClusterKey key = measurementEntry.getKey(); 
			final Collection<RawSample<DiskRawMeasurements>> diskMeasurements = measurementEntry.getValue(); 
			reportingPackage.reportDiskRawPerformance(key, diskMeasurements); 
		} 
	}
	

	public float getLatency(ArchiveMetricsScratchPad<DiskCounter> currentScratchPad, DiskCounter userReadLatency, 
			DiskCounter userWriteLatency, DiskCounter cpReadLatency, float totalOps){
		
		long totalLatency = MetricsRowScratchPad.add(currentScratchPad.getCounterDelta(userReadLatency),
                currentScratchPad.getCounterDelta(userWriteLatency), currentScratchPad.getCounterDelta(cpReadLatency));

		float weightedLatency = 0;
		if(totalLatency == MetricsRowSample.NULL_LONG || totalOps == MetricsRowSample.NULL_FLOAT || totalOps == 0){
			return MetricsRowSample.NULL_FLOAT;
		}

		weightedLatency = (float) ((double)totalLatency / (double)totalOps) ;
		return weightedLatency;

	}
	
}
