package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;


import java.math.BigInteger;
import java.util.Collection;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.util.StringUtils;

import com.netapp.oci.netapp.performance.hostadapter.interfaces.data.HostadapterMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster.ClusterKey;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

/**
 * hostadapter counter values are fetched from archiver for FS.1 or higher Ontap counter collection..
 * @author DPADMANA
 */
public class HostadapterMetricsBuilder extends ArchiveMetricsBuilderBase<ClusterKey, HostadapterMeasurements,HostadapterMeasurements, HostadapterMetricsBuilder.HostadapterCounter> {
	private static final Logger logger = LoggerFactory.getLogger(HostadapterMetricsBuilder.class);

	
	public static enum HostadapterCounter implements ArchiveCounterMetaData {
	   	bytes_read,
    	bytes_written,
    	max_link_data_rate;
 
		private final String onTapCounterName;
		private final int onTapCounterIndex;

		private HostadapterCounter() {
			this.onTapCounterName = this.name();
			this.onTapCounterIndex = -1;
		}

		@Override
		public String getOnTapCounterName() {
			return onTapCounterName;
		}

		@Override
		public int getOnTapCounterIndex() {
			return onTapCounterIndex;
		}

	}

	public HostadapterMetricsBuilder(NetAppOCIEPerformancePackage reportingPackage) {
		super(reportingPackage, "hostadapter", HostadapterCounter.class);
	}

	@Override
	protected HostadapterCounter[] getCountersMetaData(
	NetAppDefaultStorageContext context) {
		return HostadapterCounter.values();
	}

	@Override
	protected HostadapterMeasurements normalizeMetrics(ArchiveMetricsScratchPad<HostadapterCounter> currentScratchPad, NetAppDefaultStorageContext context) {
		HostadapterMeasurements hostadapterMeasurements = new HostadapterMeasurements();
		
		hostadapterMeasurements.setBytesRead(currentScratchPad.getCounterRatePerSec(HostadapterCounter.bytes_read));
		hostadapterMeasurements.setBytesWritten(currentScratchPad.getCounterRatePerSec(HostadapterCounter.bytes_written));
		
		//Since max link data rate for hostadapter is constant and this data is not available as configuration, we need the raw value here.
		BigInteger maxLinkDataRate = currentScratchPad.getCounterValues().get(HostadapterCounter.max_link_data_rate);
		hostadapterMeasurements.setMaxLinkDataRate((maxLinkDataRate != null && maxLinkDataRate.signum() >= 0) ? maxLinkDataRate.longValue(): MetricsRowSample.NULL_LONG);

		return hostadapterMeasurements;
	}

	
	@Override
	protected ClusterKey getParentInventoryUUIDKey(NetAppDefaultStorageContext context, String perfInstanceUUID,String nodeName, Map<String, Object> counterRow) {
		return context.getKey();
	}

	@Override
	public void reportPerformance(ClusterKey clusterKey,Map<ClusterKey, Collection<RawSample<HostadapterMeasurements>>> measurements)throws DataSourceErrorException {
		
		if (measurements == null) {
		         return;
		}
	
		for (final Map.Entry<ClusterKey, Collection<RawSample<HostadapterMeasurements>>> measurementEntry : measurements.entrySet()) {
			final Collection<RawSample<HostadapterMeasurements>> hostadapterMeasurements = measurementEntry.getValue();
			reportingPackage.reportHostadapterPerformance(clusterKey, hostadapterMeasurements);
		}		
	}

		
	@Override
	protected String getPerformanceUUID(final NetAppDefaultStorageContext context, final String nodeName,
            final Map<String, Object> counterRow){
		String instanceName = (String)counterRow.get("instance_name");
		if (instanceName == null){
			logger.warn("Could not get performance uuid as instance_name from counter raw row is null.");
			return null;
		}
		return new StringBuilder(nodeName).append(":kernel:").append(instanceName).toString();
	}

    @Override
    protected String getInventoryUUIDFromPerformanceUUID(final NetAppDefaultStorageContext context, final String performanceInstanceUUID, final String nodeName,
			final Map<String, Object> counterRow) {

		Map<String, String> nodeNameToUUID = context.getNodeNameToUUIDMap();
		String adapterName = (String) counterRow.get("instance_name");

		if (!StringUtils.isEmpty(nodeName) && !StringUtils.isEmpty(adapterName)) {
			String nodeUUID = nodeNameToUUID.get(nodeName);
			if (!StringUtils.isEmpty(nodeUUID)) {
				// nodeuuid.adapterName
				return new StringBuilder(nodeUUID).append(".").append(adapterName).toString();
			} 
		}
		logger.warn((new StringBuilder(
				"Cannot formulate Inventory UUID as one of the required field is null [ node name = "))
				.append(nodeName).append("adapterName = ").append(adapterName)
				.append("nodeUUId = ").append(nodeNameToUUID.get(nodeName))
				.toString());
		return null;
	}

	@Override
	public void reportRawPerformance(
			ClusterKey clusterKey,
			Map<ClusterKey, Collection<RawSample<HostadapterMeasurements>>> rawMeasurements)
			throws DataSourceErrorException {
	}

	@Override
	protected HostadapterMeasurements rawMetrics(
			ArchiveMetricsScratchPad<HostadapterCounter> currentScratchPad,
			NetAppDefaultStorageContext context) {
		return null;
	}
   
}