package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import java.math.BigInteger;
import java.util.Collection;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.netapp.performance.lun.interfaces.data.LunMeasurements;
import com.netapp.oci.netapp.performance.lun.interfaces.data.LunRawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster.ClusterKey;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.VServer;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.VServer.VServerKey;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

/**
 * Build the raw and cooked LUN measurement from archiver.
 *
 */
public class LunMetricsBuilder extends ArchiveMetricsBuilderBase<VServer.VServerKey,
	LunMeasurements, LunRawMeasurements, LunMetricsBuilder.LunCounter>{
	
	private static final Logger logger = LoggerFactory.getLogger(LunMetricsBuilder.class);
	
	public static enum LunCounter implements ArchiveCounterMetaData {
        avg_latency,
        avg_read_latency,
        avg_write_latency,
        other_ops,
        read_data,
        read_ops,
        read_partial_blocks,
        total_ops,
        write_data,
        write_ops,
        write_partial_blocks;
        
		private final String onTapCounterName;
		private final int onTapCounterIndex;

		private LunCounter(final String onTapCounterName, final int onTapCounterIndex) {
			this.onTapCounterName = onTapCounterName;
			this.onTapCounterIndex = onTapCounterIndex;
		}

		private LunCounter() {
			this.onTapCounterName = this.name();
			this.onTapCounterIndex = -1;
		}


		@Override
		public String getOnTapCounterName() {
			return onTapCounterName;
		}

		@Override
		public int getOnTapCounterIndex() {
			return onTapCounterIndex;
		}
	}

	public LunMetricsBuilder(NetAppOCIEPerformancePackage reportingPackage) {
		super(reportingPackage, "lun", LunCounter.class);
	}
	
	@Override
	protected VServerKey getParentInventoryUUIDKey(NetAppDefaultStorageContext context, 
		String perfInstanceUUID, String nodeName, Map<String, Object> counterRow) {
		
		String vServerUUID = (String)counterRow.get(VSERVER_UUID);
		if (vServerUUID == null) {
            logger.warn(PackageContext.getLogPrefix() + " - Unable to find " + VSERVER_UUID + " for LUN in archiver file.");
            return null;
		}
        return new VServer.VServerKey(context.getKey().getID(Cluster.ClusterKey.Field.cluster_uuid), vServerUUID);
	}

	@Override
	protected LunCounter[] getCountersMetaData(NetAppDefaultStorageContext context) {
		return LunCounter.values();
	}
	
	@Override
	public void reportPerformance(ClusterKey clusterKey, Map<VServerKey, 
		Collection<RawSample<LunMeasurements>>> measurements) throws DataSourceErrorException {
		
		if (measurements != null) {
			for (VServerKey vserverKey : measurements.keySet()) {
				reportingPackage.reportLunPerformance(clusterKey, vserverKey, measurements.get(vserverKey));
			}
		} else {
			logger.warn(PackageContext.getLogPrefix() + " - no LUN sample reported for cluster " + clusterKey);
		}
	}

	@Override
	protected LunMeasurements normalizeMetrics(ArchiveMetricsScratchPad<LunCounter> currentScratchPad,
		NetAppDefaultStorageContext context) {

		LunMeasurements m = new LunMeasurements();
		m.setAvgLatency(currentScratchPad.getLatencyPerOp(LunCounter.avg_latency, LunCounter.total_ops));
		m.setAvgReadLatency(currentScratchPad.getLatencyPerOp(LunCounter.avg_read_latency, LunCounter.read_ops));
		m.setAvgWriteLatency(currentScratchPad.getLatencyPerOp(LunCounter.avg_write_latency, LunCounter.write_ops));
		m.setOtherOps(currentScratchPad.getCounterRatePerSec(LunCounter.other_ops));
		m.setReadData(currentScratchPad.getCounterRatePerSec(LunCounter.read_data));
		m.setReadOps(currentScratchPad.getCounterRatePerSec(LunCounter.read_ops));
		m.setReadPartialBlocks(currentScratchPad.getPercentFromDeltas(LunCounter.read_partial_blocks, LunCounter.read_ops));
		m.setWriteData(currentScratchPad.getCounterRatePerSec(LunCounter.write_data));
		m.setWriteOps(currentScratchPad.getCounterRatePerSec(LunCounter.write_ops));
		m.setWritePartialBlocks(currentScratchPad.getPercentFromDeltas(LunCounter.write_partial_blocks, LunCounter.write_ops));
		m.setTotalOps(currentScratchPad.getCounterRatePerSec(LunCounter.total_ops));
		
		// set synthetic counter
		m.setOpmLunThroughput(MetricsRowScratchPad.add(m.getReadData(), m.getWriteData()));
		return m;
	}

	@Override
	protected LunRawMeasurements rawMetrics(ArchiveMetricsScratchPad<LunCounter> currentScratchPad,
		NetAppDefaultStorageContext context) {

		final LunRawMeasurements m = new LunRawMeasurements();
        Map<LunCounter,BigInteger> counterValue = currentScratchPad.getCounterValues();
        
        m.setAvgLatency(counterValue.get(LunCounter.avg_latency));
        m.setAvgReadLatency(counterValue.get(LunCounter.avg_read_latency));
        m.setAvgWriteLatency(counterValue.get(LunCounter.avg_write_latency));
        m.setOtherOps(counterValue.get(LunCounter.other_ops));
        m.setReadData(counterValue.get(LunCounter.read_data));
        m.setReadOps(counterValue.get(LunCounter.read_ops));
        m.setReadPartialBlocks(counterValue.get(LunCounter.read_partial_blocks));
        m.setTotalOps(counterValue.get(LunCounter.total_ops));
        m.setWriteData(counterValue.get(LunCounter.write_data));
        m.setWriteOps(counterValue.get(LunCounter.write_ops));
        m.setWritePartialBlocks(counterValue.get(LunCounter.write_partial_blocks));

		// set synthetic counter
        m.setOpmLunThroughput(MetricsRowScratchPad.add(m.getReadData(), m.getWriteData()));
		return m;
	}

	@Override
	public void reportRawPerformance(ClusterKey clusterKey,	Map<VServerKey, 
		Collection<RawSample<LunRawMeasurements>>> rawMeasurements)	throws DataSourceErrorException {
		
		if (rawMeasurements != null) {
			for (VServerKey vserverKey : rawMeasurements.keySet()) {
				reportingPackage.reportLunRawPerformance(clusterKey, vserverKey, rawMeasurements.get(vserverKey));
			}
		} else {
			logger.warn(PackageContext.getLogPrefix() + " - no LUN raw sample reported for cluster " + clusterKey);
		}
	}

}
