package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import com.netapp.oci.netapp.performance.nfs.interfaces.data.Nfsv41Measurements;
import com.netapp.oci.netapp.performance.nfs.interfaces.data.Nfsv41RawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation.BuilderUtil;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;
import java.util.Collection;
import java.util.Map;

/**
 * Builder class for processing and reporting NFSv41 metrics from archiver
 */
public class Nfsv41MetricsBuilder extends ArchiveMetricsBuilderBase<Cluster.ClusterKey, Nfsv41Measurements,Nfsv41RawMeasurements,
        Nfsv41MetricsBuilder.NFSv41Counter>   {

    private static final Logger logger = LoggerFactory.getLogger(Nfsv4MetricsBuilder.class);

    public static enum NFSv41Counter implements ArchiveCounterMetaData {

        read_avg_latency,
        read_total,
        nfs41_read_throughput,
        nfs41_throughput,
        write_avg_latency,
        write_total,
        nfs41_write_throughput,
        nfsv4_1_ops,
        latency,
        total_ops,
        instance_uuid;


        private final String onTapCounterName;
        private final int onTapCounterIndex;

        private NFSv41Counter(final String onTapCounterName, final int onTapCounterIndex) {
            this.onTapCounterName = onTapCounterName;
            this.onTapCounterIndex = onTapCounterIndex;
        }

        private NFSv41Counter() {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = -1;
        }

        private NFSv41Counter(int onTapCounterIndex) {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = onTapCounterIndex;
        }

        @Override
        public String getOnTapCounterName() {
            return onTapCounterName;
        }

        @Override
        public int getOnTapCounterIndex() {
            return onTapCounterIndex;
        }

    }

    /**
     * constructor
     *
     * @param reportingPackage
     */

    public Nfsv41MetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "nfsv4_1", NFSv41Counter.class);
    }


    /**
     * Retrieve the aggregate UUID fromm the performance archive
     *
     * @param context    The foundation context
     * @param nodeName   Node name from where the archiver was retrieved
     * @param counterRow Raw counter data form the performance archive
     * @return Performance Instance UUID
     */
    @Override
    protected String getPerformanceUUID(final NetAppDefaultStorageContext context,
                                        final String nodeName, final Map<String, Object> counterRow) {
        String instanceUUID = (String)counterRow.get(NFSv41Counter.instance_uuid.name()); // Format : vserverID(23):processor(233)
        String vserverId = BuilderUtil.extractIdFromParens(instanceUUID, "vserverID");
        String vServerUUID = null;

        if(vserverId != null  && !vserverId.isEmpty()) {
            Map<String, String> vserverIDToUUIDMap = context.getNfsv41VserverIDToUUIDMap();
            vServerUUID = vserverIDToUUIDMap.get(vserverId);

            if (vServerUUID == null) {
                logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServerId for NFSv41 " + nodeName);
            }
        } else {
            logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServerId for NFSv41 " + nodeName+ " with vserverID "+vserverId);
        }
        // Performance uuid is the vserverUUID
        return vServerUUID;
    }

    @Override
    protected Cluster.ClusterKey getParentInventoryUUIDKey(final NetAppDefaultStorageContext context,
                                                           final String perfInstanceUUID, final String nodeName,
                                                           final Map<String, Object> counterRow) {
        return context.getKey();
    }


    @Override
    public void reportPerformance(final Cluster.ClusterKey clusterKey, final Map<Cluster.ClusterKey,
            Collection<RawSample<Nfsv41Measurements>>> measurements) throws DataSourceErrorException {
        if (measurements == null) {
            return;
        }
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<Nfsv41Measurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<Nfsv41Measurements>> nfsv41Measurements = measurementEntry.getValue();
            reportingPackage.reportNfsv41Performance(key, nfsv41Measurements);
        }
    }

    @Override
    public NFSv41Counter[] getCountersMetaData(final NetAppDefaultStorageContext context) {
        return NFSv41Counter.values();
    }


    @Override
    protected Nfsv41Measurements normalizeMetrics(final ArchiveMetricsScratchPad<NFSv41Counter> currentScratchPad,
                                                 final NetAppDefaultStorageContext context) {
        final Nfsv41Measurements nfsv41Measurements = new Nfsv41Measurements();
        nfsv41Measurements.setNfsv41Ops(currentScratchPad.getCounterRatePerSec(NFSv41Counter.nfsv4_1_ops));
        nfsv41Measurements.setNfsv41ReadThroughput(currentScratchPad.getCounterRatePerSec(NFSv41Counter.nfs41_read_throughput));
        nfsv41Measurements.setNfsv41Throughput(currentScratchPad.getCounterRatePerSec(NFSv41Counter.nfs41_throughput));
        nfsv41Measurements.setNfsv41WriteThroughput(currentScratchPad.getCounterRatePerSec(NFSv41Counter.nfs41_write_throughput));
        nfsv41Measurements.setReadTotal(currentScratchPad.getCounterDeltaFinal(NFSv41Counter.read_total));
        nfsv41Measurements.setWriteTotal(currentScratchPad.getCounterDeltaFinal(NFSv41Counter.write_total));
        nfsv41Measurements.setReadAvgLatency(currentScratchPad.getLatencyPerOp(NFSv41Counter.read_avg_latency, getDenominatorForNumerator(NFSv41Counter.read_avg_latency)));
        nfsv41Measurements.setWriteAvgLatency(currentScratchPad.getLatencyPerOp(NFSv41Counter.write_avg_latency, getDenominatorForNumerator(NFSv41Counter.write_avg_latency)));
        nfsv41Measurements.setLatency(currentScratchPad.getLatencyPerOp(NFSv41Counter.write_avg_latency, getDenominatorForNumerator(NFSv41Counter.write_avg_latency)));
        nfsv41Measurements.setTotalOps(currentScratchPad.getCounterRatePerSec(NFSv41Counter.total_ops));

        return  nfsv41Measurements;
    }

    NFSv41Counter getDenominatorForNumerator(final NFSv41Counter numeratorName) {
        final NFSv41Counter denominatorName;
        switch (numeratorName) {
            case read_avg_latency:
                denominatorName = NFSv41Counter.read_total;
                break;
            case write_avg_latency:
                denominatorName = NFSv41Counter.write_total;
                break;
            case latency:
                denominatorName = NFSv41Counter.total_ops;
                break;
            default:
                denominatorName = null;
        }
        return denominatorName;
    }

    @Override
    public void reportRawPerformance(Cluster.ClusterKey clusterKey, Map<Cluster.ClusterKey, Collection<RawSample<Nfsv41RawMeasurements>>> rawMeasurements)
            throws DataSourceErrorException {
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<Nfsv41RawMeasurements>>> measurementEntry : rawMeasurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<Nfsv41RawMeasurements>> nfsv41RawMeasurements = measurementEntry.getValue();
            reportingPackage.reportNfsv41RawPerformance(clusterKey, nfsv41RawMeasurements);
        }
    }

    @Override
    protected Nfsv41RawMeasurements rawMetrics(ArchiveMetricsScratchPad<NFSv41Counter> currentScratchPad, NetAppDefaultStorageContext context) {
        final Nfsv41RawMeasurements m = new Nfsv41RawMeasurements();
        Map<NFSv41Counter,BigInteger> counterValues = currentScratchPad.getCounterValues();
        m.setNfs41ReadThroughput(counterValues.get(NFSv41Counter.nfs41_read_throughput));
        m.setNfs41Throughput(counterValues.get(NFSv41Counter.nfs41_throughput));
        m.setNfs41WriteThroughput(counterValues.get(NFSv41Counter.nfs41_write_throughput));
        m.setNfsv41Ops(counterValues.get(NFSv41Counter.nfsv4_1_ops));
        m.setReadAvgLatency(counterValues.get(NFSv41Counter.read_avg_latency));
        m.setReadTotal(counterValues.get(NFSv41Counter.read_total));
        m.setWriteAvgLatency(counterValues.get(NFSv41Counter.write_avg_latency));
        m.setWriteTotal(counterValues.get(NFSv41Counter.write_total));
        m.setLatency(counterValues.get(NFSv41Counter.latency));
        m.setTotalOps(counterValues.get(NFSv41Counter.total_ops));
        return m;
    }

}

