package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import com.netapp.oci.netapp.performance.nfs.interfaces.data.Nfsv4Measurements;
import com.netapp.oci.netapp.performance.nfs.interfaces.data.Nfsv4RawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation.BuilderUtil;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;
import java.util.Collection;
import java.util.Map;

/**
 * Builder class for processing and reporting NFSv4 metrics from archiver
 */
public class Nfsv4MetricsBuilder extends ArchiveMetricsBuilderBase<Cluster.ClusterKey, Nfsv4Measurements, Nfsv4RawMeasurements,
        Nfsv4MetricsBuilder.NFSv4Counter>   {

    private static final Logger logger = LoggerFactory.getLogger(Nfsv4MetricsBuilder.class);

    public static enum NFSv4Counter implements ArchiveCounterMetaData {

        read_avg_latency,
        read_total,
        nfs4_read_throughput,
        nfs4_throughput,
        write_avg_latency,
        write_total,
        nfs4_write_throughput,
        nfsv4_ops,
        latency,
        total_ops,
        instance_uuid;

        private final String onTapCounterName;
        private final int onTapCounterIndex;

        private NFSv4Counter(final String onTapCounterName, final int onTapCounterIndex) {
            this.onTapCounterName = onTapCounterName;
            this.onTapCounterIndex = onTapCounterIndex;
        }

        private NFSv4Counter() {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = -1;
        }

        private NFSv4Counter(int onTapCounterIndex) {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = onTapCounterIndex;
        }

        @Override
        public String getOnTapCounterName() {
            return onTapCounterName;
        }

        @Override
        public int getOnTapCounterIndex() {
            return onTapCounterIndex;
        }

    }

    /**
     * constructor
     *
     * @param reportingPackage
     */

    public Nfsv4MetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "nfsv4", NFSv4Counter.class);
    }


    /**
     * Retrieve the vserver_id fromm the performance archive
     *
     * @param context    The foundation context
     * @param nodeName   Node name from where the archiver was retrieved
     * @param counterRow Raw counter data form the performance archive
     * @return Performance Instance UUID
     */
    @Override
    protected String getPerformanceUUID(final NetAppDefaultStorageContext context,
                                        final String nodeName, final Map<String, Object> counterRow) {
        String instanceUUID = (String)counterRow.get(NFSv4Counter.instance_uuid.name()); // Format : vserverID(23):processor(233)
        String vserverId = BuilderUtil.extractIdFromParens(instanceUUID, "vserverID");
        String vServerUUID = null;

        if(vserverId != null && !vserverId.isEmpty()) {
            Map<String, String> vserverIDToUUIDMap = context.getNfsv4VserverIDToUUIDMap();
            vServerUUID = vserverIDToUUIDMap.get(vserverId);

            if (vServerUUID == null) {
                logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServerId for NFSv4 " + nodeName);
            }
        } else {
            logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServerId for NFSv4 " + nodeName+ " with vserverID "+vserverId);
        }
        // performance UUID will be the vServerUUID
        return vServerUUID;
    }

    @Override
    protected Cluster.ClusterKey getParentInventoryUUIDKey(final NetAppDefaultStorageContext context,
                                                           final String perfInstanceUUID, final String nodeName,
                                                           final Map<String, Object> counterRow) {
        return context.getKey();
    }

    @Override
    public void reportPerformance(final Cluster.ClusterKey clusterKey, final Map<Cluster.ClusterKey,
            Collection<RawSample<Nfsv4Measurements>>> measurements) throws DataSourceErrorException {
        if (measurements == null) {
            return;
        }
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<Nfsv4Measurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<Nfsv4Measurements>> NFSv4Measurements = measurementEntry.getValue();
            reportingPackage.reportNfsv4Performance(key, NFSv4Measurements);
        }
    }

    @Override
    public NFSv4Counter[] getCountersMetaData(final NetAppDefaultStorageContext context) {
        return NFSv4Counter.values();
    }


    @Override
    protected Nfsv4Measurements normalizeMetrics(final ArchiveMetricsScratchPad<NFSv4Counter> currentScratchPad,
                                                 final NetAppDefaultStorageContext context) {
        final Nfsv4Measurements nfsv4Measurements = new Nfsv4Measurements();
        nfsv4Measurements.setNfsv4Ops(currentScratchPad.getCounterRatePerSec(NFSv4Counter.nfsv4_ops));
        nfsv4Measurements.setNfsv4ReadThroughput(currentScratchPad.getCounterRatePerSec(NFSv4Counter.nfs4_read_throughput));
        nfsv4Measurements.setNfsv4Throughput(currentScratchPad.getCounterRatePerSec(NFSv4Counter.nfs4_throughput));
        nfsv4Measurements.setNfsv4WriteThroughput(currentScratchPad.getCounterRatePerSec(NFSv4Counter.nfs4_write_throughput));
        nfsv4Measurements.setReadTotal(currentScratchPad.getCounterDeltaFinal(NFSv4Counter.read_total));
        nfsv4Measurements.setWriteTotal(currentScratchPad.getCounterDeltaFinal(NFSv4Counter.write_total));
        nfsv4Measurements.setReadAvgLatency(currentScratchPad.getLatencyPerOp(NFSv4Counter.read_avg_latency, getDenominatorForNumerator(NFSv4Counter.read_avg_latency)));
        nfsv4Measurements.setWriteAvgLatency(currentScratchPad.getLatencyPerOp(NFSv4Counter.write_avg_latency, getDenominatorForNumerator(NFSv4Counter.write_avg_latency)));
        nfsv4Measurements.setLatency(currentScratchPad.getLatencyPerOp(NFSv4Counter.latency, getDenominatorForNumerator(NFSv4Counter.latency)));
        nfsv4Measurements.setTotal_ops(currentScratchPad.getCounterRatePerSec(NFSv4Counter.total_ops));

        return  nfsv4Measurements;
    }

    NFSv4Counter getDenominatorForNumerator(final NFSv4Counter numeratorName) {
        final NFSv4Counter denominatorName;
        switch (numeratorName) {
            case read_avg_latency:
                denominatorName = NFSv4Counter.read_total;
                break;
            case write_avg_latency:
                denominatorName = NFSv4Counter.write_total;
                break;
            case latency:
                denominatorName = NFSv4Counter.total_ops;
                break;

            default:
                denominatorName = null;
        }
        return denominatorName;
    }

    @Override
    public void reportRawPerformance(Cluster.ClusterKey clusterKey, Map<Cluster.ClusterKey, Collection<RawSample<Nfsv4RawMeasurements>>> rawMeasurements)
            throws DataSourceErrorException {
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<Nfsv4RawMeasurements>>> measurementEntry : rawMeasurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<Nfsv4RawMeasurements>> nfsv4RawMeasurements = measurementEntry.getValue();
            reportingPackage.reportNfsv4RawPerformance(clusterKey, nfsv4RawMeasurements);
        }
    }

    @Override
    protected Nfsv4RawMeasurements rawMetrics(ArchiveMetricsScratchPad<NFSv4Counter> currentScratchPad, NetAppDefaultStorageContext context) {
        final Nfsv4RawMeasurements m = new Nfsv4RawMeasurements();
        Map<NFSv4Counter,BigInteger> counterValues = currentScratchPad.getCounterValues();
        m.setNfs4ReadThroughput(counterValues.get(NFSv4Counter.nfs4_read_throughput));
        m.setNfs4Throughput(counterValues.get(NFSv4Counter.nfs4_throughput));
        m.setNfs4WriteThroughput(counterValues.get(NFSv4Counter.nfs4_write_throughput));
        m.setNfsv4Ops(counterValues.get(NFSv4Counter.nfsv4_ops));
        m.setReadAvgLatency(counterValues.get(NFSv4Counter.read_avg_latency));
        m.setReadTotal(counterValues.get(NFSv4Counter.read_total));
        m.setWriteAvgLatency(counterValues.get(NFSv4Counter.write_avg_latency));
        m.setWriteTotal(counterValues.get(NFSv4Counter.write_total));
        m.setLatency(counterValues.get(NFSv4Counter.latency));
        m.setTotalOps(counterValues.get(NFSv4Counter.total_ops));
        return m;
    }

}

