package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.netapp.performance.processor.interfaces.data.ProcessorMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster.ClusterKey;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

public class ProcessorMetricsBuilder extends ArchiveMetricsBuilderBase<Cluster.ClusterKey,
ProcessorMeasurements, ProcessorMeasurements, ProcessorMetricsBuilder.ProcessorCounter>{
	
	  private static final Logger logger = LoggerFactory.getLogger(ProcessorMetricsBuilder.class);
	 public static enum ProcessorCounter implements ArchiveCounterMetaData {
		 processor_busy,
		 processor_elapsed_time,
		 idle("domain_busy", 0),
		 kahuna("domain_busy", 1),
		 storage("domain_busy", 2),
		 exempt("domain_busy", 3),
		 raid("domain_busy", 4),
		 raid_exempt("domain_busy", 5),
		 target("domain_busy", 6),
		 dnscache("domain_busy", 7),
		 cifs("domain_busy", 8),
		 wafl_exempt("domain_busy", 9),
		 wafl_xcleaner("domain_busy", 10),
		 sm_exempt("domain_busy", 11),
		 protocol("domain_busy", 12),
		 nwk_exclusive("domain_busy", 13),
		 nwk_exempt("domain_busy", 14),
		 nwk_legacy("domain_busy", 15),
		 hostOS("domain_busy", 16),
		 ssan_exempt("domain_busy", 17);
		
		private final String onTapCounterName;
        private final int onTapCounterIndex;
        
        private ProcessorCounter(final String onTapCounterName, final int onTapCounterIndex) {
            this.onTapCounterName = onTapCounterName;
            this.onTapCounterIndex = onTapCounterIndex;
        }

        private ProcessorCounter() {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = -1;
        }


		@Override
		public String getOnTapCounterName() {
			return onTapCounterName;
		}

		@Override
		public int getOnTapCounterIndex() {
			return onTapCounterIndex;
		}
	 }
	 private ArrayList<ProcessorCounter> nBladeDomains;
	 private ArrayList<ProcessorCounter> dBladeDomains;
	 
	 public ProcessorMetricsBuilder(NetAppOCIEPerformancePackage reportingPackage) {
			super(reportingPackage, "processor", ProcessorCounter.class);
			nBladeDomains = new ArrayList<ProcessorCounter>(Arrays.asList(ProcessorCounter.cifs, ProcessorCounter.dnscache, ProcessorCounter.nwk_exclusive, 
								ProcessorCounter.nwk_exempt, ProcessorCounter.nwk_legacy, ProcessorCounter.protocol, ProcessorCounter.target));
			dBladeDomains = new ArrayList<ProcessorCounter>(Arrays.asList(ProcessorCounter.exempt, ProcessorCounter.kahuna, ProcessorCounter.raid, ProcessorCounter.sm_exempt,
								ProcessorCounter.storage, ProcessorCounter.wafl_exempt, ProcessorCounter.wafl_xcleaner));
			
		}

	

	@Override
	protected ProcessorCounter[] getCountersMetaData(
			NetAppDefaultStorageContext context) {
		return ProcessorCounter.values();
	}

	@Override
	protected ProcessorMeasurements normalizeMetrics(
			ArchiveMetricsScratchPad<ProcessorCounter> currentScratchPad,
			NetAppDefaultStorageContext context) {
		ProcessorMeasurements m = new ProcessorMeasurements();
		m.setCifsBusy(currentScratchPad.getPercentFromDeltas(ProcessorCounter.cifs, ProcessorCounter.processor_elapsed_time));
		m.setStorageBusy(currentScratchPad.getPercentFromDeltas(ProcessorCounter.storage, ProcessorCounter.processor_elapsed_time));
		m.setKahunaBusy(currentScratchPad.getPercentFromDeltas(ProcessorCounter.kahuna, ProcessorCounter.processor_elapsed_time));
		m.setNwkExclusiveBusy(currentScratchPad.getPercentFromDeltas(ProcessorCounter.nwk_exclusive, ProcessorCounter.processor_elapsed_time));
		m.setNwkLegacyBusy(currentScratchPad.getPercentFromDeltas(ProcessorCounter.nwk_legacy, ProcessorCounter.processor_elapsed_time));
		m.setProtocolBusy(currentScratchPad.getPercentFromDeltas(ProcessorCounter.protocol, ProcessorCounter.processor_elapsed_time));
		m.setRaidBusy(currentScratchPad.getPercentFromDeltas(ProcessorCounter.raid, ProcessorCounter.processor_elapsed_time));
		m.setTargetBusy(currentScratchPad.getPercentFromDeltas(ProcessorCounter.target, ProcessorCounter.processor_elapsed_time));
		m.setProcessorElapsedTime(currentScratchPad.getCounterDeltaFinal(ProcessorCounter.processor_elapsed_time));
		m.setProcessorBusy(currentScratchPad.getPercentFromDeltas(ProcessorCounter.processor_busy, ProcessorCounter.processor_elapsed_time));
		m.setNBladeBusy(getDomainsBusyPercent(currentScratchPad, nBladeDomains));
		m.setDBladeBusy(getDomainsBusyPercent(currentScratchPad, dBladeDomains));
		return m;
	}
	
	@Override
	 protected String getPerformanceUUID(final NetAppDefaultStorageContext context, final String nodeName,
             final Map<String, Object> counterRow){
		String processorName = (String)counterRow.get("instance_name");

        if(processorName != null || processorName.isEmpty()){
        	return new StringBuilder().append(nodeName).append(":kernel:").append(processorName).toString();
        }
        
        logger.warn("{} - Failed to get instance_name for processor {}", new Object[] {PackageContext.getLogPrefix()});
		return null;
        
	}
	
	@Override
	protected String getInventoryUUIDFromPerformanceUUID(final NetAppDefaultStorageContext context,
                final String performanceInstanceUUID, String nodeName, Map<String, Object> counterRow) {
		String processorName = (String)counterRow.get("instance_name");

        final String nodeUUID = context.getNodeNameToUUIDMap().get(nodeName);
        if(processorName != null || processorName.isEmpty()){
        	final String processorNum = processorName.substring("processor".length(), processorName.length());
        	return nodeUUID + "." + processorNum;
        }
        
		logger.warn("{} - Failed to find UUID for processor {}", new Object[] {PackageContext.getLogPrefix()});
		return null;
	}
	
	@Override
	protected ClusterKey getParentInventoryUUIDKey(
			NetAppDefaultStorageContext context, String perfInstanceUUID,
			String nodeName, Map<String, Object> counterRow) {
		return context.getKey(); 
	}

	@Override
	public void reportPerformance(
			ClusterKey clusterKey,
			Map<ClusterKey, Collection<RawSample<ProcessorMeasurements>>> measurements)
			throws DataSourceErrorException {
		if (measurements == null) {
            return;
        }
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<ProcessorMeasurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<ProcessorMeasurements>> processorMeasurements = measurementEntry.getValue();
            reportingPackage.reportProcessorPerformance(key, processorMeasurements);
        }	
	}

	@Override
	public void reportRawPerformance(
			ClusterKey clusterKey,
			Map<ClusterKey, Collection<RawSample<ProcessorMeasurements>>> rawMeasurements)
			throws DataSourceErrorException {
		
	}

	@Override
	protected ProcessorMeasurements rawMetrics(
			ArchiveMetricsScratchPad<ProcessorCounter> currentScratchPad,
			NetAppDefaultStorageContext context) {
		return null;
	}
	
	 private long getSumOfDeltas(ArchiveMetricsScratchPad<ProcessorCounter> currentScratchPad, final List<ProcessorCounter> counters){
		 long sumOfDeltas = 0;
		 for(ProcessorCounter counter : counters){
			 final Long domainDelta = currentScratchPad.getCounterDelta(counter);
			 if(domainDelta != MetricsRowSample.NULL_LONG){
				 sumOfDeltas += domainDelta;
			 }
			 
		 }
		 
		 return sumOfDeltas;
	 }
	 
	 private float getDomainsBusyPercent(ArchiveMetricsScratchPad<ProcessorCounter> currentScratchPad, final List<ProcessorCounter> counters){

		 final Long deltaProcessorTime = currentScratchPad.getCounterDelta(ProcessorCounter.processor_elapsed_time);
		 final long totalDomainBusyDelta = getSumOfDeltas(currentScratchPad, counters);
		 final float percentBusy = (deltaProcessorTime == 0)
				 ? MetricsRowSample.NULL_FLOAT
						 : (deltaProcessorTime > 0 && totalDomainBusyDelta > 0)
						 ? (float) (100d * (double) totalDomainBusyDelta / (double) deltaProcessorTime)
								 : MetricsRowSample.NULL_FLOAT;

		return percentBusy;

	 }

	
}
