package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import com.netapp.oci.netapp.performance.qosservicecenter.interfaces.data.QosServiceCenterMeasurements;
import com.netapp.oci.netapp.performance.qosworkloaddblade.interfaces.data.QosWorkloadQueueDBladeMeasurements;
import com.netapp.oci.netapp.performance.qosworkloadnblade.interfaces.data.QosWorkloadQueueNBladeMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.parser.CounterTable;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster.ClusterKey;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

import java.math.BigInteger;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Builder class for processing and reporting service center metrics. Utilization for resources are retrieved from the
 * resource object in the opm archive. The wait time for delay centers are computed by summarizing the delay center wait
 * time in the worload_queue_dblade objects.
 */
public class QosServiceCenterMetricsBuilder extends ArchiveMetricsBuilderBase<Cluster.ClusterKey,
        QosServiceCenterMeasurements, QosServiceCenterMeasurements, QosServiceCenterMetricsBuilder.ResourceCounter> {

    /**
     * ZERO sample to be returned on wrap reset.
     */
    private static final QosServiceCenterMeasurements ZERO_SAMPLE = new QosServiceCenterMeasurements();

    /**
     * Counters from the resource object,
     */
    public static enum ResourceCounter implements ArchiveCounterMetaData {
        /**
         * The percent of the resource that is used.
         */
        busy_percent,
        /**
         * Elapsed time since boot
         */
        cpu_elapsed_time,
        /**
         * The number of instances (processors, disks, and so on) in the resource.
         */
        num_instances,;

        @Override
        public String getOnTapCounterName() {
            return toString();
        }

        @Override
        public int getOnTapCounterIndex() {
            return -1;
        }

    }

    public QosServiceCenterMetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "resource", ResourceCounter.class);
    }

    /**
     * Returns the cluster key. Parent is the cluster.
     */
    @Override
    protected Cluster.ClusterKey getParentInventoryUUIDKey(final NetAppDefaultStorageContext context,
                                                           final String perfInstanceUUID, String nodeName, Map<String, Object> counterRow) {
        return context.getKey();
    }

    /**
     * Format of the inventory UUID - cluster_uuid.performance_uuid i.e cluster_uuid.node_name:kernel:resource_name
     */
    @Override
    protected String getInventoryUUIDFromPerformanceUUID(final NetAppDefaultStorageContext context,
                                                         final String performanceInstanceUUID, String nodeName, Map<String, Object> counterRow) {
        return context.getKey().getID(Cluster.ClusterKey.Field.uuid) + '.' + performanceInstanceUUID;
    }

    @Override
    public ResourceCounter[] getCountersMetaData(final NetAppDefaultStorageContext context) {
        return ResourceCounter.values();
    }

    /**
     * Calculate busy percent for resources - ((busy_percent / cpu_elapsed_time) * 100) / num_instances
     */
    @Override
    protected QosServiceCenterMeasurements normalizeMetrics(final ArchiveMetricsScratchPad<ResourceCounter> currentScratchPad,
                                                            final NetAppDefaultStorageContext context) {
        final BigInteger numOfInstancesBigInt = currentScratchPad.getCounterValues().get(ResourceCounter.num_instances);
        Long numOfInstances = MetricsRowSample.NULL_LONG;
        if (numOfInstancesBigInt != null) {
            numOfInstances = numOfInstancesBigInt.longValue();
        }

        QosServiceCenterMeasurements qosServiceCenterMeasurements = new QosServiceCenterMeasurements(
                MetricsRowSample.NULL_LONG,         // Wait time not collected directly , aggregated from workload queue dblade and nblade
                getBusyPercent(currentScratchPad),
                currentScratchPad.getCounterDeltaFinal(ResourceCounter.cpu_elapsed_time),
                numOfInstances,
                currentScratchPad.getCollectionPeriod()
        );
        if(qosServiceCenterMeasurements.getUtilization() != 0 &&
                qosServiceCenterMeasurements.getUtilization() != MetricsRowSample.NULL_FLOAT) {
            currentScratchPad.setZeroSuppressDetected(false);
        } else {
            currentScratchPad.setZeroSuppressDetected(true);
        }
        return qosServiceCenterMeasurements;
    }

    /**
     * Calculate busy percent for resources - ((busy_percent / cpu_elapsed_time) * 100) / num_instances
     */
    private float getBusyPercent(final ArchiveMetricsScratchPad<ResourceCounter> currentScratchPad) {
        final BigInteger numOfInstances = currentScratchPad.getCounterValues().get(ResourceCounter.num_instances);
        if (numOfInstances == null || numOfInstances.compareTo(BigInteger.ZERO) <= 0) {
            return MetricsRowSample.NULL_FLOAT;
        }
        final float percentAllInstance = currentScratchPad.getPercentFromDeltas(ResourceCounter.busy_percent,
                ResourceCounter.cpu_elapsed_time);
        final float normalizedPercent = percentAllInstance / numOfInstances.intValue();
        currentScratchPad.setFinalValueForDebug(ResourceCounter.busy_percent, normalizedPercent);
        return normalizedPercent;
    }
    
    /**
     * Process resource counters and then add delay center counters.
     */
    public Map<Cluster.ClusterKey, Collection<RawSample<QosServiceCenterMeasurements>>> processCounterData(
            final NetAppDefaultStorageContext context, final List<CounterTable> clusterCounterData,
            final Map<Cluster.ClusterKey, Collection<RawSample<QosWorkloadQueueDBladeMeasurements>>> qosDBladeMetrics,
            final Map<Cluster.ClusterKey, Collection<RawSample<QosWorkloadQueueNBladeMeasurements>>> qosNBladeMetrics)
            throws DataSourceErrorException {

        final Map<Cluster.ClusterKey, Collection<RawSample<QosServiceCenterMeasurements>>> clusterKeyCollectionMap =
                super.processCounterData(context, clusterCounterData);
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<QosServiceCenterMeasurements>>> clusterToSamples : clusterKeyCollectionMap.entrySet()) {
            addDelayCenters(context, clusterToSamples.getValue(),
                    qosDBladeMetrics.get(clusterToSamples.getKey()),
                    qosNBladeMetrics.get(clusterToSamples.getKey()));
        }
        return clusterKeyCollectionMap;
    }

    /**
     * Calculate delay center wait time for the following delay centes -
     *
     * DELAY_CENTER_WAFL_SUSP_CP - sum of workload_queue_dblade.delay_wafl_susp_cp_wait_time for all workloads in the node.
     * DELAY_CENTER_DISK_IO - sum of workload_queue_dblade.delay_disk_io_wait_time for all workloads in the node.
     * DELAY_CENTER_NVLOG_TRANSFER - sum of workload_queue_dblade.delay_nvlog_transfer_wait_time for all workloads in the node.
     * DELAY_CENTER_CLUSTER_INTERCONNECT - sum of workload_queue_nblade.delay_cluster_interconnect_wait_time for all
     *                                     workloads in the node.
     * DELAY_CENTER_QOS_LIMIT - sum of workload_queue_nblade.delay_qos_limit_wait_time for all workloads in the node.
     * DELAY_CENTER_NETWORK - sum of workload_queue_nblade.delay_network_wait_time for all workloads in the node.
     */
    private void addDelayCenters(final NetAppDefaultStorageContext context,
                                 final Collection<RawSample<QosServiceCenterMeasurements>> samplesFromCluster,
                                 final Collection<RawSample<QosWorkloadQueueDBladeMeasurements>> qosDBladeMetrics,
                                 final Collection<RawSample<QosWorkloadQueueNBladeMeasurements>> qosNBladeMetrics) {
        final Map<String, RawSample<QosServiceCenterMeasurements>>
                delayCenterWAFLCPByUUID = new HashMap<String, RawSample<QosServiceCenterMeasurements>>();
        final Map<String, RawSample<QosServiceCenterMeasurements>>
                delayCenterDiskIOByUUID = new HashMap<String, RawSample<QosServiceCenterMeasurements>>();
        final Map<String, RawSample<QosServiceCenterMeasurements>> 
        		delayNVLogTransferByUUID = new HashMap<String, RawSample<QosServiceCenterMeasurements>>();

        for (final RawSample<QosWorkloadQueueDBladeMeasurements> dBladeSample : qosDBladeMetrics) {
            final QosWorkloadQueueDBladeMeasurements dBladeMeasurement = dBladeSample.getCounters();
            final String workloadNodeRelUUID = dBladeSample.getKey();
            final long collectionPeriod = dBladeMeasurement.getCollectionPeriod();
            final long delayWAFLSuspCPWaitTime = dBladeMeasurement.getDelayWAFLSuspCPWaitTime();
            if (delayWAFLSuspCPWaitTime > 0) {
                addDelayCenterWaitTime(context, delayCenterWAFLCPByUUID, workloadNodeRelUUID,
                        "DELAY_CENTER_WAFL_SUSP_CP", delayWAFLSuspCPWaitTime, collectionPeriod);
            }
            final long delayDiskIOWaitTime = dBladeMeasurement.getDelayDiskIOWaitTime();
            if (delayDiskIOWaitTime > 0) {
                addDelayCenterWaitTime(context, delayCenterDiskIOByUUID, workloadNodeRelUUID,
                        "DELAY_CENTER_DISK_IO", delayDiskIOWaitTime, collectionPeriod);
            }
            final long delayNVLogTransferWaitTime = dBladeMeasurement.getDelayNvlogTransferWaitTime();
            if (delayNVLogTransferWaitTime > 0){
            	addDelayCenterWaitTime(context, delayNVLogTransferByUUID, workloadNodeRelUUID,
            			"DELAY_CENTER_NVLOG_TRANSFER", delayNVLogTransferWaitTime, collectionPeriod);
            }
        }

        final Map<String, RawSample<QosServiceCenterMeasurements>>
                delayCenterClusterInterconnectByUUID = new HashMap<String, RawSample<QosServiceCenterMeasurements>>();
        final Map<String, RawSample<QosServiceCenterMeasurements>>
                delayCenterQOSLimitByUUID = new HashMap<String, RawSample<QosServiceCenterMeasurements>>();
        final Map<String, RawSample<QosServiceCenterMeasurements>>
                delayCenterNetworkByUUID = new HashMap<String, RawSample<QosServiceCenterMeasurements>>();

        for (final RawSample<QosWorkloadQueueNBladeMeasurements> nBladeSample : qosNBladeMetrics) {
            final QosWorkloadQueueNBladeMeasurements nBladeMeasurement = nBladeSample.getCounters();
            final String workloadNodeRelUUID = nBladeSample.getKey();
            final long collectionPeriod = nBladeMeasurement.getCollectionPeriod();
            final long delayClusterInterconnectWaitTime = nBladeMeasurement.getDelayClusterInterconnectWaitTime();
            if (delayClusterInterconnectWaitTime > 0) {
                addDelayCenterWaitTime(context, delayCenterClusterInterconnectByUUID, workloadNodeRelUUID,
                        "DELAY_CENTER_CLUSTER_INTERCONNECT", delayClusterInterconnectWaitTime, collectionPeriod);
            }
            final long delayNetworkWaitTime = nBladeMeasurement.getDelayNetworkWaitTime();
            if (delayNetworkWaitTime > 0) {
                addDelayCenterWaitTime(context, delayCenterNetworkByUUID, workloadNodeRelUUID,
                        "DELAY_CENTER_NETWORK", delayNetworkWaitTime, collectionPeriod);
            }
            final long delayQosLimitWaitTime = nBladeMeasurement.getDelayQosLimitWaitTime();
            if (delayQosLimitWaitTime > 0) {
                addDelayCenterWaitTime(context, delayCenterQOSLimitByUUID, workloadNodeRelUUID,
                        "DELAY_CENTER_QOS_LIMIT", delayQosLimitWaitTime, collectionPeriod);
            }
        }
        //Add to sample from cluster to report to server.
        samplesFromCluster.addAll(delayCenterClusterInterconnectByUUID.values());
        samplesFromCluster.addAll(delayCenterDiskIOByUUID.values());
        samplesFromCluster.addAll(delayCenterNetworkByUUID.values());
        samplesFromCluster.addAll(delayCenterQOSLimitByUUID.values());
        samplesFromCluster.addAll(delayCenterWAFLCPByUUID.values());
        samplesFromCluster.addAll(delayNVLogTransferByUUID.values());
    }

    private void addDelayCenterWaitTime(final NetAppDefaultStorageContext context,
                                        final Map<String, RawSample<QosServiceCenterMeasurements>> delayCenterByUUID,
                                        final String workloadNodeRelUUID, final String delayCenterName,
                                        final long waitTime, final long collectionPeriod) {
        final String delayCenterUUID = createDelayCenterUUID(context, workloadNodeRelUUID, delayCenterName);
        if(delayCenterUUID == null) {
            return;
        }

        RawSample<QosServiceCenterMeasurements>
                delayCenter = delayCenterByUUID.get(delayCenterUUID);
        if (delayCenter == null) {
            delayCenter = new RawSample<QosServiceCenterMeasurements>(delayCenterUUID,
                    new QosServiceCenterMeasurements(0, MetricsRowSample.NULL_FLOAT, MetricsRowSample.NULL_LONG,
                            MetricsRowSample.NULL_LONG, 0));
            delayCenterByUUID.put(delayCenterUUID, delayCenter);
        }
        final QosServiceCenterMeasurements delayCenterCounters = delayCenter.getCounters();
        delayCenterCounters.setWaitTime(delayCenterCounters.getWaitTime() + waitTime);
        delayCenterCounters.setCollectionPeriod(collectionPeriod);
    }

    /**
     * Create delay center UUID - cluster_uuid.node_name:kernel:delay_center_name
     */
    private String createDelayCenterUUID(final NetAppDefaultStorageContext context, final String workloadNodeRelUUID, final String delayCenterName) {
        final Map<String, String> qosWorkloadConstituentMap = context.getQosWorkloadConstituentMap().get(workloadNodeRelUUID);
        if (qosWorkloadConstituentMap != null) {
            return new StringBuilder(context.getKey().getOcieUUID()).append(".")
                    .append(qosWorkloadConstituentMap.get("WorkloadConstituentNodeName"))
                    .append(":kernel:").append(delayCenterName).toString();
        } else {
            return null;
        }
    }

    @Override
    public void reportPerformance(Cluster.ClusterKey clusterKey, final Map<Cluster.ClusterKey,
            Collection<RawSample<QosServiceCenterMeasurements>>> measurements) throws DataSourceErrorException {
        if(measurements == null) {
            return;
        }
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<QosServiceCenterMeasurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<QosServiceCenterMeasurements>> resourceMeasurements = measurementEntry.getValue();
            reportingPackage.reportQosServiceCenterPerformance(key, resourceMeasurements);
        }
    }

    /**
     * Constructs the performance uuid for resource counters - node_name:kernel:instance_name
     *
     * @param context    The foundation context
     * @param nodeName Node name from where the archiver was retrieved
     * @param counterRow Raw counter data form the performance archive
     * @return Performance Instance UUID
     */
    @Override
    protected String getPerformanceUUID(final NetAppDefaultStorageContext context,
                                        final String nodeName, final Map<String, Object> counterRow) {
        final StringBuilder performanceUUID = new StringBuilder(nodeName);
        performanceUUID.append(":kernel:").append(counterRow.get("instance_name"));
        return performanceUUID.toString();
    }

    /**
     * Return a zero sample.
     *
     * @param currentScratchPad Scratch Pad containing the current and the previous values.
     * @param context           in case other data from the context is required to normalize data
     * @return
     */
    @Override
    protected QosServiceCenterMeasurements getWrapResetSample(final ArchiveMetricsScratchPad<ResourceCounter> currentScratchPad,
                                                              final NetAppDefaultStorageContext context) {
    	ZERO_SAMPLE.setCollectionPeriod(currentScratchPad.getCollectionPeriod());
        return  ZERO_SAMPLE;
    }

	@Override
	protected QosServiceCenterMeasurements rawMetrics(
			ArchiveMetricsScratchPad<ResourceCounter> currentScratchPad,
			NetAppDefaultStorageContext context) {
		return null;
	}

	@Override
	public void reportRawPerformance(
			ClusterKey clusterKey,
			Map<ClusterKey, Collection<RawSample<QosServiceCenterMeasurements>>> rawMeasurements)
			throws DataSourceErrorException {
		// Do nothing no raw data available
	}
}
