package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import com.netapp.oci.netapp.performance.qosworkloaddetail.interfaces.data.QosWorkloadDetailMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

import java.util.Collection;
import java.util.Map;

/**
*  This class is to retrieve workload_detail performance data from archive which starts from FS.1.
 */
public class QosWorkloadDetailMetricsBuilder extends ArchiveMetricsBuilderBase<Cluster.ClusterKey,
        QosWorkloadDetailMeasurements, QosWorkloadDetailMeasurements, QosWorkloadDetailMetricsBuilder.WorkloadDetailCounter> {
    /**
     * ZERO sample to be returned on wrap reset.
     */
    private static final QosWorkloadDetailMeasurements ZERO_SAMPLE = new QosWorkloadDetailMeasurements();
    private static final String NAME_OTHER = "OTHER";
    private static final String ZERO_UUID = "00000000-0000-0000-0000-000000000000";

    public static enum WorkloadDetailCounter implements ArchiveCounterMetaData {
        service_time,
        wait_time;

        private final String onTapCounterName;
        private final int onTapCounterIndex;

        private WorkloadDetailCounter(final String onTapCounterName, final int onTapCounterIndex) {
            this.onTapCounterName = onTapCounterName;
            this.onTapCounterIndex = onTapCounterIndex;
        }

        private WorkloadDetailCounter() {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = -1;
        }

        WorkloadDetailCounter(int onTapCounterIndex) {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = onTapCounterIndex;
        }

        @Override
        public String getOnTapCounterName() {
            return onTapCounterName;
        }

        @Override
        public int getOnTapCounterIndex() {
            return onTapCounterIndex;
        }
    }

    /**
     * Retrieve the workload UUID form the performance archive and append the node uuid.
     *
     * @param context    The foundation context
     * @param nodeName   Node name from where the archiver was retrieved
     * @param counterRow Raw counter data form the performance archive
     * @return Performance Instance UUID
     */
    @Override
    protected String getPerformanceUUID(final NetAppDefaultStorageContext context,
                                        final String nodeName, final Map<String, Object> counterRow) {
        final String workloadUUID = super.getPerformanceUUID(context, nodeName, counterRow);
        final String workloadName = context.getQosVolumeWorkloadUUIDtoNameMap().get(workloadUUID);
        final String aggregateUUID = super.getUUIDFromHiLo(counterRow, "archive_aggr_uuid_hi", "archive_aggr_uuid_lo");
        String aggregateName = context.getAggregateUUIDToNameMap().get(aggregateUUID);
        if(aggregateUUID.equals(ZERO_UUID))   {  //when  archive_aggr_uuid_hi=0 and archive_aggr_uuid_lo=0
            aggregateName = NAME_OTHER;
        }
        final Number diskTypeNum = (Number)counterRow.get("archive_disk_type");
        String diskType= "HDD";
        if(diskTypeNum.intValue()==1)  {
             diskType = "SSD";
        }
        StringBuilder builder = new StringBuilder();
        builder.append(context.getKey().getOcieUUID()).append('.').append(nodeName).append(":kernel:").append(workloadName).append(".DISK_").append(diskType)
                .append('_').append(aggregateName);
        if(!aggregateName.equals(NAME_OTHER)) {
           builder.append('_').append(aggregateUUID);
        }

        return builder.toString();
    }

    public QosWorkloadDetailMetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "workload_aggr", WorkloadDetailCounter.class);
    }

    @Override
    protected Cluster.ClusterKey getParentInventoryUUIDKey(final NetAppDefaultStorageContext context,
                                                           final String perfInstanceUUID, String nodeName, Map<String, Object> counterRow) {
        return context.getKey();
    }

    @Override
    public void reportPerformance(Cluster.ClusterKey clusterKey, final Map<Cluster.ClusterKey,
            Collection<RawSample<QosWorkloadDetailMeasurements>>> measurements) throws DataSourceErrorException {
        if (measurements == null) {
            return;
        }
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<QosWorkloadDetailMeasurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<QosWorkloadDetailMeasurements>> qosWorkloadDetailMeasurements = measurementEntry.getValue();
            reportingPackage.reportQosWorkloadDetailPerformance(key, qosWorkloadDetailMeasurements);
        }
    }

    /**
     * Return a ZERO sample on a wrap & reset.
     *
     * @param currentScratchPad Scratch Pad containing the current and the previous values.
     * @param context           in case other data from the context is required to normalize data
     * @return
     */
    @Override
    protected QosWorkloadDetailMeasurements getWrapResetSample(final ArchiveMetricsScratchPad<WorkloadDetailCounter> currentScratchPad,
                                                                    final NetAppDefaultStorageContext context) {
        ZERO_SAMPLE.setCollectionPeriod(currentScratchPad.getCollectionPeriod());
        return ZERO_SAMPLE;
    }

    @Override
    protected QosWorkloadDetailMeasurements normalizeMetrics(final ArchiveMetricsScratchPad<WorkloadDetailCounter> currentScratchPad,
                                                                  final NetAppDefaultStorageContext context) {
        final QosWorkloadDetailMeasurements qosWorkloadDetailMeasurements = new QosWorkloadDetailMeasurements(currentScratchPad.getCounterDeltaFinal(WorkloadDetailCounter.service_time),
                currentScratchPad.getCounterDeltaFinal(WorkloadDetailCounter.wait_time), currentScratchPad.getCollectionPeriod());

        if (qosWorkloadDetailMeasurements.getServiceTime() == 0 &&
                qosWorkloadDetailMeasurements.getWaitTime() == 0 ) {
            currentScratchPad.setZeroSuppressDetected(true);
        } else {
            currentScratchPad.setZeroSuppressDetected(false);
        }

        return qosWorkloadDetailMeasurements;
    }

    @Override
    public WorkloadDetailCounter[] getCountersMetaData(final NetAppDefaultStorageContext context) {
        return WorkloadDetailCounter.values();
    }

    @Override
    protected QosWorkloadDetailMeasurements rawMetrics(
            ArchiveMetricsScratchPad<WorkloadDetailCounter> currentScratchPad,
            NetAppDefaultStorageContext context) {
        return null;
    }

    @Override
    public void reportRawPerformance(
            Cluster.ClusterKey clusterKey,
            Map<Cluster.ClusterKey, Collection<RawSample<QosWorkloadDetailMeasurements>>> rawMeasurements)
            throws DataSourceErrorException {
        // Do nothing no raw data available
    }
}
