package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;


import com.netapp.oci.netapp.performance.qosworkload.interfaces.data.QosVolumeWorkloadMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

import java.util.*;

/**
 * Builder class for computing cluster level workload stats. The data for cluster level workload stats will be aggregated
 * from the workload_queue_nblade and workload_queue_dblade objects
 */
public class QosWorkloadMetricsBuilder {
    private final NetAppOCIEPerformancePackage netAppOCIEPerformancePackage;

    public QosWorkloadMetricsBuilder(final NetAppOCIEPerformancePackage netAppOCIEPerformancePackage) {
        this.netAppOCIEPerformancePackage = netAppOCIEPerformancePackage;
    }

    /**
     * Aggregate the data from the workload_queue_nblade and workload_queue_dblade objects to create cluster level
     * workload measurements.
     *
     * latency - Sum of total latency (read, write, other) for a workload across all nodes / Sum of total ops (read,
     *           write, other) for a workload across all nodes
     * read latency - Sum of read latency for a workload across all nodes / Sum of read ops for a workload across all
     *                 nodes
     * write latency - Sum of write latency for a workload across all nodes / Sum of write ops for a workload across all
     *                 nodes
     * ops - Sum of total ops (read, write, other) for a workload across all nodes / collection period
     * read ops - Sum of read ops for a workload across all nodes / collection period
     * write ops - Sum of write ops for a workload across all nodes / collection period
     * total data - Sum of read and write data for a workload across all nodes / collection period
     * read data - Sum of read data for a workload across all nodes / collection period
     * write data - Sum of write data for a workload across all nodes / collection period
     */
    public Map<Cluster.ClusterKey, Collection<RawSample<QosVolumeWorkloadMeasurements>>> processCounterData(
            final NetAppDefaultStorageContext context,
            final Map<String, ArchiveMetricsScratchPad<QosWorkloadQueueDBladeMetricsBuilder.WorkloadQueueDBladeCounter>> dBladeScratchPadsByUUID,
            final Map<String, ArchiveMetricsScratchPad<QosWorkloadQueueNBladeMetricsBuilder.WorkloadQueueNBladeCounter>> nBladeScratchPadsByUUID) {

        final List<RawSample<QosVolumeWorkloadMeasurements>> rawSamples = new LinkedList<RawSample<QosVolumeWorkloadMeasurements>>();
        final Map<Cluster.ClusterKey, Collection<RawSample<QosVolumeWorkloadMeasurements>>> allMeasurementsToReport =
                new HashMap<Cluster.ClusterKey, Collection<RawSample<QosVolumeWorkloadMeasurements>>>(1);
        allMeasurementsToReport.put(context.getKey(), rawSamples);

        final Map<String, QosWorkloadMetricsScratchPad> workloadScratchPadsByUUID =
                new HashMap<String, QosWorkloadMetricsScratchPad>(nBladeScratchPadsByUUID.size());

        for (final Map.Entry<String, ArchiveMetricsScratchPad<QosWorkloadQueueDBladeMetricsBuilder.WorkloadQueueDBladeCounter>> dBladeScratchPadByUUID : dBladeScratchPadsByUUID.entrySet()) {
            final ArchiveMetricsScratchPad<QosWorkloadQueueDBladeMetricsBuilder.WorkloadQueueDBladeCounter> scratchPad = dBladeScratchPadByUUID.getValue();
            // Check if previous poll data is not available return an empty
            // sample.
            if (!scratchPad.isPreviousPollValuesAvailable()) {
                continue;
            }
            final QosWorkloadMetricsScratchPad workloadMetricsScratchPad = getWorkloadMetricsScratchPad(
                    workloadScratchPadsByUUID, dBladeScratchPadByUUID.getKey(),
                    scratchPad.getCollectionPeriod());
            if (workloadMetricsScratchPad == null) {
                continue;
            }

            final long cacheMissRate = scratchPad
                    .getCounterDelta(QosWorkloadQueueDBladeMetricsBuilder.WorkloadQueueDBladeCounter.cache_miss_rate);
            final long cacheMissRateBase = scratchPad
                    .getCounterDelta(QosWorkloadQueueDBladeMetricsBuilder.WorkloadQueueDBladeCounter.cache_miss_rate_base);

            workloadMetricsScratchPad.addCacheMissRate(cacheMissRate);
            workloadMetricsScratchPad.addCacheMissRateBase(cacheMissRateBase);
        }
        
        for (final Map.Entry<String, ArchiveMetricsScratchPad<QosWorkloadQueueNBladeMetricsBuilder.WorkloadQueueNBladeCounter>>
                nBladeScratchPadByUUID : nBladeScratchPadsByUUID.entrySet()) {
            final ArchiveMetricsScratchPad<QosWorkloadQueueNBladeMetricsBuilder.WorkloadQueueNBladeCounter> scratchPad
                    = nBladeScratchPadByUUID.getValue();
            //Check if previous poll data is not available return an empty sample.
            if(!scratchPad.isPreviousPollValuesAvailable()) {
                continue;
            }
            final QosWorkloadMetricsScratchPad workloadMetricsScratchPad = getWorkloadMetricsScratchPad(workloadScratchPadsByUUID,
                    nBladeScratchPadByUUID.getKey(), scratchPad.getCollectionPeriod());
            if (workloadMetricsScratchPad == null) {
                continue;
            }


            final long readLatency = scratchPad.getCounterDelta(QosWorkloadQueueNBladeMetricsBuilder.WorkloadQueueNBladeCounter.read_latency);
            final long writeLatency = scratchPad.getCounterDelta(QosWorkloadQueueNBladeMetricsBuilder.WorkloadQueueNBladeCounter.write_latency);
            final long otherLatency = scratchPad.getCounterDelta(QosWorkloadQueueNBladeMetricsBuilder.WorkloadQueueNBladeCounter.other_latency);
            final long readOps = scratchPad.getCounterDelta(QosWorkloadQueueNBladeMetricsBuilder.WorkloadQueueNBladeCounter.read_iops);
            final long writeOps = scratchPad.getCounterDelta(QosWorkloadQueueNBladeMetricsBuilder.WorkloadQueueNBladeCounter.write_iops);
            final long otherOps = scratchPad.getCounterDelta(QosWorkloadQueueNBladeMetricsBuilder.WorkloadQueueNBladeCounter.other_iops);
            final long readData = scratchPad.getCounterDelta(QosWorkloadQueueNBladeMetricsBuilder.WorkloadQueueNBladeCounter.read_data);
            final long writeData = scratchPad.getCounterDelta(QosWorkloadQueueNBladeMetricsBuilder.WorkloadQueueNBladeCounter.write_data);

            workloadMetricsScratchPad.addReadLatency(readLatency);
            workloadMetricsScratchPad.addWriteLatency(writeLatency);
            workloadMetricsScratchPad.addOtherLatency(otherLatency);
            workloadMetricsScratchPad.addReadOps(readOps);
            workloadMetricsScratchPad.addWriteOps(writeOps);
            workloadMetricsScratchPad.addOtherOps(otherOps);
            workloadMetricsScratchPad.addReadData(readData);
            workloadMetricsScratchPad.addWriteData(writeData);
        }

        for (final Map.Entry<String, QosWorkloadMetricsScratchPad> workloadScratchPadByUUID : workloadScratchPadsByUUID.entrySet()) {
            final QosWorkloadMetricsScratchPad workloadScratchPad = workloadScratchPadByUUID.getValue();
            if (workloadScratchPad.isZeroSuppressDetected()) {
                continue;
            }
            final String inventoryUUID = context.getKey().getID(Cluster.ClusterKey.Field.uuid) + '.' + workloadScratchPadByUUID.getKey();
            final QosVolumeWorkloadMeasurements qosVolumeWorkloadMeasurements = new QosVolumeWorkloadMeasurements()
                                                                                    .setReadOps(workloadScratchPad.getReadOpsPerSecond())
                                                                                    .setWriteOps(workloadScratchPad.getWriteOpsPerSecond())
                                                                                    .setOps(workloadScratchPad.getTotalOpsPerSecond())
                                                                                    .setReadLatency(workloadScratchPad.getReadLatencyPerReadOps())
                                                                                    .setWriteLatency(workloadScratchPad.getWriteLatencyPerWriteOps())
                                                                                    .setLatency(workloadScratchPad.getTotalLatencyPerOps())
                                                                                    .setReadData(workloadScratchPad.getReadDataPerSecond())
                                                                                    .setWriteData(workloadScratchPad.getWriteDataPerSecond())
                                                                                    .setTotalData(workloadScratchPad.getTotalDataPerSecond())
                                                                                    .setCacheMissRate(workloadScratchPad.getCacheMissRatePercent())
                                                                                    .setCacheMissRateBase(workloadScratchPad.getCacheMissRateBase());
            rawSamples.add(new RawSample<QosVolumeWorkloadMeasurements>(inventoryUUID, qosVolumeWorkloadMeasurements));
        }
        return allMeasurementsToReport;
    }

    /**
     * Retrieve the workload uuid from the node workload uuid (node_uuid.workload_uuid) and create a
     * QosWorkloadMetricsScratchPad if not already created.
     */
    private QosWorkloadMetricsScratchPad getWorkloadMetricsScratchPad(
            final Map<String, QosWorkloadMetricsScratchPad> workloadScratchPadByUUID,
            final String nodeWorkloadUUID, final long collectionPeriod) {

        final String nodeWorkloadUUIDBroken[] = nodeWorkloadUUID.split("\\.");
        if (nodeWorkloadUUIDBroken.length < 2) {
            return null;
        }
        final String workloadUUID = nodeWorkloadUUIDBroken[1];
        QosWorkloadMetricsScratchPad workloadMetricsScratchPad = workloadScratchPadByUUID.get(workloadUUID);
        if (workloadMetricsScratchPad == null) {
            workloadMetricsScratchPad = new QosWorkloadMetricsScratchPad(collectionPeriod);
            workloadScratchPadByUUID.put(workloadUUID, workloadMetricsScratchPad);
        }
        //Use the longest collection period.
        if(collectionPeriod > workloadMetricsScratchPad.getCollectionPeriod()) {
            workloadMetricsScratchPad.setCollectionPeriod(collectionPeriod);
        }
        return workloadMetricsScratchPad;
    }

    public void reportPerformance(final Map<Cluster.ClusterKey, Collection<RawSample<QosVolumeWorkloadMeasurements>>> qosWorkloadMetrics)
            throws DataSourceErrorException {
        if(qosWorkloadMetrics == null) {
            return;
        }
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<QosVolumeWorkloadMeasurements>>> measurementEntry : qosWorkloadMetrics.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<QosVolumeWorkloadMeasurements>> workloadMeasurements = measurementEntry.getValue();
            netAppOCIEPerformancePackage.reportQosVolumeWorkloadPerformance(key, workloadMeasurements);
        }
    }

    /**
     * QosWorkloadMetricsScratchPad to hold workload raw values while summarizing and normalize the values.
     */
    private static class QosWorkloadMetricsScratchPad {
        private long readLatency;
        private long writeLatency;
        private long otherLatency;
        private long readOps;
        private long writeOps;
        private long otherOps;
        private long readData;
        private long writeData;
        private long cacheMissRate;
        private long cacheMissRateBase;
        private long collectionPeriod;

        private QosWorkloadMetricsScratchPad(final long collectionPeriod) {
            this.collectionPeriod = collectionPeriod;
        }
        
        private void addCacheMissRate(final long cacheMissRate) {
            this.cacheMissRate = (cacheMissRate == MetricsRowSample.NULL_LONG) ? this.cacheMissRate : this.cacheMissRate + cacheMissRate;
        }
        
        private void addCacheMissRateBase(final long cacheMissRateBase) {
            this.cacheMissRateBase = (cacheMissRateBase == MetricsRowSample.NULL_LONG) ? this.cacheMissRateBase : this.cacheMissRateBase + cacheMissRateBase;
        }

        private long getCacheMissRateBase() {
            return cacheMissRateBase;
        }

        private float getCacheMissRatePercent(){
            final float percent; 
            if (cacheMissRateBase == 0 || cacheMissRate == 0){
                percent = 0;
            }else if(cacheMissRateBase > 0 && cacheMissRate > 0){
                percent = (float) (100d * (double) cacheMissRate / (double) cacheMissRateBase);
            }else{
                percent = MetricsRowSample.NULL_FLOAT;
            }
            return percent;            
        }

        private long getCollectionPeriod() {
            return collectionPeriod;
        }

        private void setCollectionPeriod(final long collectionPeriod) {
            this.collectionPeriod = collectionPeriod;
        }

        private float getReadLatencyPerReadOps() {
            return getLatencyPerOps(this.readLatency, this.readOps);
        }

        private void addReadLatency(final long readLatency) {
            this.readLatency = (readLatency == MetricsRowSample.NULL_LONG) ? this.readLatency : this.readLatency + readLatency;
        }

        private float getWriteLatencyPerWriteOps() {
            return getLatencyPerOps(this.writeLatency, this.writeOps);
        }

        private void addWriteLatency(final long writeLatency) {
            this.writeLatency = (writeLatency == MetricsRowSample.NULL_LONG) ? this.writeLatency : this.writeLatency + writeLatency;
        }

        private float getTotalLatencyPerOps() {
            return getLatencyPerOps(readLatency + writeLatency + otherLatency, readOps + writeOps + otherOps);
        }

        private void addOtherLatency(final long otherLatency) {
            this.otherLatency = (otherLatency == MetricsRowSample.NULL_LONG) ? this.otherLatency : this.otherLatency + otherLatency;
        }

        private float getReadOpsPerSecond() {
            return getRatePerSecond(readOps);
        }

        private void addReadOps(final long readOps) {
            this.readOps = (readOps == MetricsRowSample.NULL_LONG) ? this.readOps : this.readOps + readOps;
        }

        private float getWriteOpsPerSecond() {
            return getRatePerSecond(writeOps);
        }

        private void addWriteOps(final long writeOps) {
            this.writeOps = (writeOps == MetricsRowSample.NULL_LONG) ? this.writeOps : this.writeOps + writeOps;
        }

        private float getTotalOpsPerSecond() {
            return getRatePerSecond(readOps + writeOps + otherOps);
        }

        private void addOtherOps(final long otherOps) {
            this.otherOps = (otherOps == MetricsRowSample.NULL_LONG) ? this.otherOps : this.otherOps + otherOps;
        }

        private float getReadDataPerSecond() {
            return getRatePerSecond(this.readData);
        }

        private void addReadData(final long readData) {
            this.readData = (readData == MetricsRowSample.NULL_LONG) ? this.readData : this.readData + readData;
        }

        private float getWriteDataPerSecond() {
            return getRatePerSecond(writeData);
        }

        private void addWriteData(final long writeData) {
            this.writeData = (writeData == MetricsRowSample.NULL_LONG) ? this.writeData : this.writeData + writeData;
        }

        public float getTotalDataPerSecond() {
            return getRatePerSecond(this.readData + this.writeData);
        }

        private float getRatePerSecond(final long counterDelta) {
            if (counterDelta == MetricsRowSample.NULL_LONG || this.collectionPeriod == MetricsRowSample.NULL_LONG) {
                return MetricsRowSample.NULL_FLOAT;
            }
            final double deltaPerMSDouble = ((double) counterDelta) / ((double) (collectionPeriod));
            final double deltaPerSDouble = deltaPerMSDouble * 1000d;
            return (float) deltaPerSDouble;
        }

        private float getLatencyPerOps(final long latency, final long ops) {
            final float latencyPerOp;
            if (ops > 0) {
                if (latency == MetricsRowSample.NULL_LONG) {
                    latencyPerOp = MetricsRowSample.NULL_FLOAT;
                } else {
                    final double deltaLatencyPerOpDouble = ((double) latency) / ((double) ops);
                    latencyPerOp = (float) deltaLatencyPerOpDouble;
                }
            } else if (ops == 0) {
                latencyPerOp = 0;
            } else {
                latencyPerOp = MetricsRowSample.NULL_FLOAT;
            }
            return latencyPerOp;
        }

        public boolean isZeroSuppressDetected() {
            return !(readOps != 0 || writeOps != 0 || otherOps != 0 || readLatency != 0 || writeLatency != 0
                    || otherLatency != 0 || readData != 0 || writeData != 0 || cacheMissRateBase != 0
                    || cacheMissRate != 0);
        }
    }
}
