package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;


import com.netapp.oci.netapp.performance.qosworkloaddblade.interfaces.data.QosWorkloadQueueDBladeMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster.ClusterKey;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

import java.util.Collection;
import java.util.Map;

/**
 * Builder class for processing and reporting workload queue dblade metrics.
 */
public class QosWorkloadQueueDBladeMetricsBuilder extends ArchiveMetricsBuilderBase<Cluster.ClusterKey,
        QosWorkloadQueueDBladeMeasurements, QosWorkloadQueueDBladeMeasurements, QosWorkloadQueueDBladeMetricsBuilder.WorkloadQueueDBladeCounter> {

    /**
     * ZERO sample to be returned on wrap reset.
     */
    private static final QosWorkloadQueueDBladeMeasurements ZERO_SAMPLE = new QosWorkloadQueueDBladeMeasurements();

    public static enum WorkloadQueueDBladeCounter implements ArchiveCounterMetaData {
        /**
         * The raw CPU Dblade residence time. This is the CPU service time plus the wait time for operations done for
         * this workload in the Dblade. Units are microseconds.
         */
        cpu_dblade_residence_time,
        /**
         * The raw CPU Dblade utilization service time. This is the CPU service time for operations done for this
         * workload in the Dblade. Units are microseconds.
         */
        cpu_dblade_util_service_time,
        /**
         * The raw CPU Dblade background service time. This is the CPU service time for background operations done for
         * this workload in the Dblade. Represents service time directly tallied for the workload.Units are
         * microseconds.
         */
        cpu_dblade_background_service_time(0),
        /**
         * The raw CPU Dblade background service time. Wafl_cp represents service time contributed by the WAFL_CP
         * workload on behalf of this workload. Units are microseconds.
         */
        cpu_dblade_background_wafl_cp_service_time("cpu_dblade_background_service_time", 1),
        /**
         * The raw CPU Dblade background service time. Background_delete represents service time contributed by the
         * BACKGROUND_DELETE workload on behalf of this workload. Units are microseconds.
         */
        cpu_dblade_background_background_delete_service_time("cpu_dblade_background_service_time", 2),
        /**
         * The raw CPU Dblade background service time. Readahead_x_stripe represents service time contributed by the
         * RA_STRIPE_X workload on behalf of this workload. Units are microseconds.
         */
        cpu_dblade_background_ra_stripe_x_service_time("cpu_dblade_background_service_time", 3),
        /**
         * The raw Disk HDD background service time. This is the amount of time HDDs spend servicing requests by this
         * workload. Represents service time directly tallied for the workload. Units are microseconds.
         */
        disk_hdd_background_service_time(0),
        /**
         * The raw Disk HDD background service time. This is the amount of time HDDs spend servicing requests by this
         * workload. Wafl_cp represents service time contributed by the WAFL_CP workload on behalf of this workload.
         * Units are microseconds.
         */
        disk_hdd_background_wafl_cp_service_time("disk_hdd_background_service_time", 1),
        /**
         * The raw Disk HDD background service time. This is the amount of time HDDs spend servicing requests by this
         * workload. Background_delete represents service time contributed by the BACKGROUND_DELETE workload on behalf
         * of this workload. Units are microseconds.
         */
        disk_hdd_background_background_delete_service_time("disk_hdd_background_service_time", 2),
        /**
         * The raw Disk HDD background service time. This is the amount of time HDDs spend servicing requests by this
         * workload. Readahead_x_stripe represents service time contributed by the RA_STRIPE_X workload on behalf of
         * this workload. Units are microseconds.
         */
        disk_hdd_background_ra_stripe_x_service_time("disk_hdd_background_service_time", 3),
        /**
         * The raw disk SSD background service time. This is the amount of time SSDs spend servicing requests by this
         * workload. Represents service time directly tallied for the workload. Units are microseconds.
         */
        disk_ssd_background_service_time(0),
        /**
         * The raw disk SSD background service time. This is the amount of time SSDs spend servicing requests by this
         * workload.  Wafl_cp represents service time contributed by the WAFL_CP workload on behalf of this workload.
         * Units are microseconds.
         */
        disk_ssd_background_wafl_cp_service_time("disk_ssd_background_service_time", 1),
        /**
         * The raw disk SSD background service time. This is the amount of time SSDs spend servicing requests by this
         * workload. Background_delete represents service time contributed by the BACKGROUND_DELETE workload on behalf
         * of this workload.Units are microseconds.
         */
        disk_ssd_background_background_delete_service_time("disk_ssd_background_service_time", 2),
        /**
         * The raw disk SSD background service time. This is the amount of time SSDs spend servicing requests by this
         * workload. Readahead_x_stripe represents service time contributed by the RA_STRIPE_X workload on behalf of
         * this workload.Units are microseconds.
         */
        disk_ssd_background_ra_stripe_x_service_time("disk_ssd_background_service_time", 3),
        /**
         * The raw WAFL suspended cp wait time. This is the time spent in WAFL waiting for a CP to complete for this
         * workload. Units are microseconds.
         */
        delay_wafl_susp_cp_wait_time,
        /**
         * The raw disk IO wait time. This is the time spent waiting for a read from disk to complete for this workload.
         * Units are microseconds.
         */
        delay_disk_io_wait_time,
        
        /**
         * This is the time spent waiting for a dma transfer to local nvram and remote nvram for writes. 
         * Units are microseconds.
         */
        delay_nvlog_transfer_wait_time,
        
        /**
         * This is the total number of WAFL read requests. It is used as a base counter (or denominator) for the 
         * cache_miss_rate counter. This counter is exactly similar to the read_io_type_base counter from the workload object.
         */
        cache_miss_rate_base,
        
        /**
         * This is the percentage of cache misses. The sum of read_io_type DISK and HYA_HDD.
         */
        cache_miss_rate;


        private final String onTapCounterName;
        private final int onTapCounterIndex;

        private WorkloadQueueDBladeCounter(final String onTapCounterName, final int onTapCounterIndex) {
            this.onTapCounterName = onTapCounterName;
            this.onTapCounterIndex = onTapCounterIndex;
        }

        private WorkloadQueueDBladeCounter() {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = -1;
        }

        WorkloadQueueDBladeCounter(int onTapCounterIndex) {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = onTapCounterIndex;
        }

        @Override
        public String getOnTapCounterName() {
            return onTapCounterName;
        }

        @Override
        public int getOnTapCounterIndex() {
            return onTapCounterIndex;
        }
    }

    /**
     * Retrieve the workload UUID form the performance archive and append the node uuid.
     *
     * @param context    The foundation context
     * @param nodeName   Node name from where the archiver was retrieved
     * @param counterRow Raw counter data form the performance archive
     * @return Performance Instance UUID
     */
    @Override
    protected String getPerformanceUUID(final NetAppDefaultStorageContext context,
                                        final String nodeName, final Map<String, Object> counterRow) {
        final String nodeUUID = context.getNodeNameToUUIDMap().get(nodeName);
        return new StringBuilder().append(nodeUUID).append('.')
                .append(super.getPerformanceUUID(context, nodeName, counterRow)).toString();
    }

    public QosWorkloadQueueDBladeMetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "workload_queue_dblade", WorkloadQueueDBladeCounter.class);
    }

    @Override
    protected Cluster.ClusterKey getParentInventoryUUIDKey(final NetAppDefaultStorageContext context,
                                                           final String perfInstanceUUID, String nodeName, Map<String, Object> counterRow) {
        return context.getKey();
    }

    @Override
    public void reportPerformance(Cluster.ClusterKey clusterKey, final Map<Cluster.ClusterKey,
            Collection<RawSample<QosWorkloadQueueDBladeMeasurements>>> measurements) throws DataSourceErrorException {
        if (measurements == null) {
            return;
        }
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<QosWorkloadQueueDBladeMeasurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<QosWorkloadQueueDBladeMeasurements>> resourceMeasurements = measurementEntry.getValue();
            reportingPackage.reportQosWorkloadQueueDBladePerformance(key, resourceMeasurements);
        }
    }

    @Override
    public WorkloadQueueDBladeCounter[] getCountersMetaData(final NetAppDefaultStorageContext context) {
        return WorkloadQueueDBladeCounter.values();
    }

    /**
     * Return a ZERO sample on a wrap & reset.
     *
     * @param currentScratchPad Scratch Pad containing the current and the previous values.
     * @param context           in case other data from the context is required to normalize data
     * @return
     */
    @Override
    protected QosWorkloadQueueDBladeMeasurements getWrapResetSample(final ArchiveMetricsScratchPad<WorkloadQueueDBladeCounter> currentScratchPad,
                                                                    final NetAppDefaultStorageContext context) {
        return ZERO_SAMPLE.setCollectionPeriod(currentScratchPad.getCollectionPeriod());
    }

    @Override
    protected QosWorkloadQueueDBladeMeasurements normalizeMetrics(final ArchiveMetricsScratchPad<WorkloadQueueDBladeCounter> currentScratchPad,
                                                                  final NetAppDefaultStorageContext context) {
        final QosWorkloadQueueDBladeMeasurements qosWorkloadQueueDBladeMeasurements = new QosWorkloadQueueDBladeMeasurements()
                .setCpuDBladeResidenceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.cpu_dblade_residence_time))
                .setCpuDBladeUtilServiceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.cpu_dblade_util_service_time))
                .setCpuDBladeBackgroundServiceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.cpu_dblade_background_service_time))
                .setCpuDBladeBackgroundWAFLCPServiceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.cpu_dblade_background_wafl_cp_service_time))
                .setCpuDBladeBackgroundBackgroundDeleteServiceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.cpu_dblade_background_background_delete_service_time))
                .setCpuDBladeBackgroundRAStripXServiceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.cpu_dblade_background_ra_stripe_x_service_time))
                .setDiskHDDBackgroundServiceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.disk_hdd_background_service_time))
                .setDiskHDDBackgroundWAFLCPServiceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.disk_hdd_background_wafl_cp_service_time))
                .setDiskHDDBackgroundBackgroundDeleteServiceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.disk_hdd_background_background_delete_service_time))
                .setDiskHDDBackgroundRAStripeXServiceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.disk_hdd_background_ra_stripe_x_service_time))
                .setDiskSSDBackgroundServiceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.disk_ssd_background_service_time))
                .setDiskSSDBackgroundWAFLCPServiceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.disk_ssd_background_wafl_cp_service_time))
                .setDiskSSDBackgroundBackgroundDeleteServiceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.disk_ssd_background_background_delete_service_time))
                .setDiskSSDBackgroundRAStripeXServiceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.disk_ssd_background_ra_stripe_x_service_time))
                .setDelayWAFLSuspCPWaitTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.delay_wafl_susp_cp_wait_time))
                .setDelayDiskIOWaitTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.delay_disk_io_wait_time))
                .setDelayNvlogTransferWaitTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.delay_nvlog_transfer_wait_time))
                .setCacheMissRate(currentScratchPad.getPercentFromDeltas(WorkloadQueueDBladeCounter.cache_miss_rate, WorkloadQueueDBladeCounter.cache_miss_rate_base))
                .setCacheMissRateBase(currentScratchPad.getCounterDeltaFinal(WorkloadQueueDBladeCounter.cache_miss_rate_base))
                .setCollectionPeriod(currentScratchPad.getCollectionPeriod());
        
        if (qosWorkloadQueueDBladeMeasurements.getCpuDBladeResidenceTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getCpuDBladeUtilServiceTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getCpuDBladeBackgroundServiceTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getCpuDBladeBackgroundWAFLCPServiceTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getCpuDBladeBackgroundBackgroundDeleteServiceTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getCpuDBladeBackgroundRAStripXServiceTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getDiskHDDBackgroundServiceTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getDiskHDDBackgroundWAFLCPServiceTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getDiskHDDBackgroundBackgroundDeleteServiceTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getDiskHDDBackgroundRAStripeXServiceTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getDiskSSDBackgroundServiceTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getDiskSSDBackgroundWAFLCPServiceTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getDiskSSDBackgroundBackgroundDeleteServiceTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getDiskSSDBackgroundRAStripeXServiceTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getDelayWAFLSuspCPWaitTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getDelayDiskIOWaitTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getDelayNvlogTransferWaitTime() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getCacheMissRateBase() == 0 &&
                qosWorkloadQueueDBladeMeasurements.getCacheMissRate() == 0) {
            currentScratchPad.setZeroSuppressDetected(true);
        } else {
            currentScratchPad.setZeroSuppressDetected(false);
        }

        return qosWorkloadQueueDBladeMeasurements;
    }

	@Override
	protected QosWorkloadQueueDBladeMeasurements rawMetrics(
			ArchiveMetricsScratchPad<WorkloadQueueDBladeCounter> currentScratchPad,
			NetAppDefaultStorageContext context) {
		return null;
	}

	@Override
	public void reportRawPerformance(
			ClusterKey clusterKey,
			Map<ClusterKey, Collection<RawSample<QosWorkloadQueueDBladeMeasurements>>> rawMeasurements)
			throws DataSourceErrorException {
		// Do nothing no raw data available
	}
}
