package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;


import com.netapp.oci.netapp.performance.qosworkloadnblade.interfaces.data.QosWorkloadQueueNBladeMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster.ClusterKey;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

import java.util.Collection;
import java.util.Map;

/**
 * Builder class for processing and reporting workload queue nblade metrics.
 */
public class QosWorkloadQueueNBladeMetricsBuilder extends ArchiveMetricsBuilderBase<Cluster.ClusterKey,
        QosWorkloadQueueNBladeMeasurements, QosWorkloadQueueNBladeMeasurements, QosWorkloadQueueNBladeMetricsBuilder.WorkloadQueueNBladeCounter> {

    /**
     * ZERO sample to be returned on wrap reset.
     */
    private static final QosWorkloadQueueNBladeMeasurements ZERO_SAMPLE = new QosWorkloadQueueNBladeMeasurements();

    public static enum WorkloadQueueNBladeCounter implements ArchiveCounterMetaData {
        /**
         * This is the rate of workload read operations that completed during the measurement interval.
         */
        read_iops("iops", 0),
        /**
         * This is the rate of workload write operations that completed during the measurement interval.
         */
        write_iops("iops", 1),
        /**
         * This is the rate of workload other operations that completed during the measurement interval.
         */
        other_iops("iops", 2),
        /**
         * This is the average response time for read requests that were initiated by the workload.
         */
        read_latency("total_latency", 0),
        /**
         * This is the average response time for write requests that were initiated by the workload.
         */
        write_latency("total_latency", 1),
        /**
         * This is the average response time for other requests that were initiated by the workload.
         */
        other_latency("total_latency", 2),
        /**
         * This is rate of data read per second by the workload during the measurement interval.
         */
        read_data,
        /**
         * This is rate of data write per second by the workload during the measurement interval.
         */
        write_data,
        /**
         * The raw CPU Nblade residence time. Units are microseconds.
         */
        cpu_nblade_residence_time,
        /**
         * The raw CPU Nblade util service time. Units are microseconds.
         */
        cpu_nblade_util_service_time,
        /**
         * The raw Cluster interconnect wait time. Units are microseconds.
         */
        delay_cluster_interconnect_wait_time,
        /**
         * The raw QoS wait time. Units are microseconds.
         */
        delay_qos_limit_wait_time,
        /**
         * The raw Network wait time. Units are microseconds.
         */
        delay_network_wait_time;

        private final String onTapCounterName;
        private final int onTapCounterIndex;

        private WorkloadQueueNBladeCounter(final String onTapCounterName, final int onTapCounterIndex) {
            this.onTapCounterName = onTapCounterName;
            this.onTapCounterIndex = onTapCounterIndex;
        }

        private WorkloadQueueNBladeCounter() {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = -1;
        }

        @Override
        public String getOnTapCounterName() {
            return onTapCounterName;
        }

        @Override
        public int getOnTapCounterIndex() {
            return onTapCounterIndex;
        }
    }

    /**
     * Retrieve the workload UUID form the performance archive and append the node uuid.
     *
     * @param context    The foundation context
     * @param nodeName   Node name from where the archiver was retrieved
     * @param counterRow Raw counter data form the performance archive
     * @return Performance Instance UUID
     */
    @Override
    protected String getPerformanceUUID(final NetAppDefaultStorageContext context,
                                        final String nodeName, final Map<String, Object> counterRow) {
        final String nodeUUID = context.getNodeNameToUUIDMap().get(nodeName);
        return new StringBuilder().append(nodeUUID).append('.')
                .append(super.getPerformanceUUID(context, nodeName, counterRow)).toString();
    }

    public QosWorkloadQueueNBladeMetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "workload_queue_nblade", WorkloadQueueNBladeCounter.class);
    }

    @Override
    protected Cluster.ClusterKey getParentInventoryUUIDKey(final NetAppDefaultStorageContext context,
                                                           final String perfInstanceUUID, String nodeName, Map<String, Object> counterRow) {
        return context.getKey();
    }

    @Override
    public void reportPerformance(Cluster.ClusterKey clusterKey, final Map<Cluster.ClusterKey,
            Collection<RawSample<QosWorkloadQueueNBladeMeasurements>>> measurements) throws DataSourceErrorException {
        if (measurements == null) {
            return;
        }
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<QosWorkloadQueueNBladeMeasurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<QosWorkloadQueueNBladeMeasurements>> resourceMeasurements = measurementEntry.getValue();
            reportingPackage.reportQosWorkloadQueueNBladePerformance(key, resourceMeasurements);
        }
    }

    @Override
    public WorkloadQueueNBladeCounter[] getCountersMetaData(final NetAppDefaultStorageContext context) {
        return WorkloadQueueNBladeCounter.values();
    }

    @Override
    protected QosWorkloadQueueNBladeMeasurements normalizeMetrics(final ArchiveMetricsScratchPad<WorkloadQueueNBladeCounter> currentScratchPad,
                                                                  final NetAppDefaultStorageContext context) {
        final QosWorkloadQueueNBladeMeasurements qosWorkloadQueueNBladeMeasurements = new QosWorkloadQueueNBladeMeasurements()
                .setReadOps(currentScratchPad.getCounterRatePerSec(WorkloadQueueNBladeCounter.read_iops))
                .setWriteOps(currentScratchPad.getCounterRatePerSec(WorkloadQueueNBladeCounter.write_iops))
                .setOtherOps(currentScratchPad.getCounterRatePerSec(WorkloadQueueNBladeCounter.other_iops))
                .setReadLatency(currentScratchPad.getLatencyPerOp(WorkloadQueueNBladeCounter.read_latency, WorkloadQueueNBladeCounter.read_iops))
                .setWriteLatency(currentScratchPad.getLatencyPerOp(WorkloadQueueNBladeCounter.write_latency, WorkloadQueueNBladeCounter.write_iops))
                .setOtherLatency(currentScratchPad.getLatencyPerOp(WorkloadQueueNBladeCounter.other_latency, WorkloadQueueNBladeCounter.other_iops))
                .setReadData(currentScratchPad.getCounterRatePerSec(WorkloadQueueNBladeCounter.read_data))
                .setWriteData(currentScratchPad.getCounterRatePerSec(WorkloadQueueNBladeCounter.write_data))
                .setCpuNBladeResidenceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueNBladeCounter.cpu_nblade_residence_time))
                .setCpuNBladeUtilServiceTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueNBladeCounter.cpu_nblade_util_service_time))
                .setDelayClusterInterconnectWaitTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueNBladeCounter.delay_cluster_interconnect_wait_time))
                .setDelayQosLimitWaitTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueNBladeCounter.delay_qos_limit_wait_time))
                .setDelayNetworkWaitTime(currentScratchPad.getCounterDeltaFinal(WorkloadQueueNBladeCounter.delay_network_wait_time))
                .setCollectionPeriod(currentScratchPad.getCollectionPeriod());
        if (qosWorkloadQueueNBladeMeasurements.getReadOps() != 0 ||
                qosWorkloadQueueNBladeMeasurements.getWriteOps() != 0 ||
                qosWorkloadQueueNBladeMeasurements.getOtherOps() != 0 ||
                qosWorkloadQueueNBladeMeasurements.getReadLatency() != 0 ||
                qosWorkloadQueueNBladeMeasurements.getWriteLatency() != 0 ||
                qosWorkloadQueueNBladeMeasurements.getOtherLatency() != 0 ||
                qosWorkloadQueueNBladeMeasurements.getReadData() != 0 ||
                qosWorkloadQueueNBladeMeasurements.getWriteData() != 0 ||
                qosWorkloadQueueNBladeMeasurements.getCpuNBladeResidenceTime() != 0 ||
                qosWorkloadQueueNBladeMeasurements.getCpuNBladeUtilServiceTime() != 0 ||
                qosWorkloadQueueNBladeMeasurements.getDelayClusterInterconnectWaitTime() != 0 ||
                qosWorkloadQueueNBladeMeasurements.getDelayQosLimitWaitTime() != 0 ||
                qosWorkloadQueueNBladeMeasurements.getDelayNetworkWaitTime() != 0) {
            currentScratchPad.setZeroSuppressDetected(false);
        } else {
            currentScratchPad.setZeroSuppressDetected(true);
        }

        return qosWorkloadQueueNBladeMeasurements;
    }

    @Override
    protected QosWorkloadQueueNBladeMeasurements getWrapResetSample(final ArchiveMetricsScratchPad<WorkloadQueueNBladeCounter> currentScratchPad,
                                                                    final NetAppDefaultStorageContext context) {
        return ZERO_SAMPLE.setCollectionPeriod(currentScratchPad.getCollectionPeriod());
    }

	@Override
	protected QosWorkloadQueueNBladeMeasurements rawMetrics(
			ArchiveMetricsScratchPad<WorkloadQueueNBladeCounter> currentScratchPad,
			NetAppDefaultStorageContext context) {
		return null;
	}

	@Override
	public void reportRawPerformance(
			ClusterKey clusterKey,
			Map<ClusterKey, Collection<RawSample<QosWorkloadQueueNBladeMeasurements>>> rawMeasurements)
			throws DataSourceErrorException {
		// Do nothing no raw data available		
	}

}
