package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.common;

import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.config.NetAppOCIEPerformanceConfiguration;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorType;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;

import org.apache.commons.io.FileUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * Downloads archived cmapl files from a cluster to a local temporary location until they are parsed and then cleaned up 
 *
 */
public class ArchiveFileDownloader extends AbstractFileDownloader implements FileDownloader { 

    private final String DEFAULT_TIMEOUT_PERCENT_OF_POLL_TIME = "80";
    private final String DEFAULT_MAX_RETRY_COUNT = "5";
    private final String DEFAULT_RETRY_WAIT_TIME = "2000";
    
    private final static Logger LOG = LoggerFactory.getLogger(ArchiveFileDownloader.class);
    private final int timeoutPercentOfPolltime = Integer.parseInt(ServiceManager.getInstance().getProperty("ARCHIVER_DOWNLOAD_TIMEOUT_PERCENT_OF_POLL_TIME", DEFAULT_TIMEOUT_PERCENT_OF_POLL_TIME).trim());
    private final int maxRetries = Integer.parseInt(ServiceManager.getInstance().getProperty("MAX_RETRY_COUNT", DEFAULT_MAX_RETRY_COUNT).trim());
    private final int retryWaitTime = Integer.parseInt(ServiceManager.getInstance().getProperty("RETRY_WAIT_TIME", DEFAULT_RETRY_WAIT_TIME).trim());
    

    /**
     * main method for downloading the CMAPL archived files from OnTap. It takes care of building the http url that is distinct for each
     * cluster and node. Downloading uses multiple threads where all tasks are added to a task list and executed in a batch fashion.
     *
     * @param archiveDataSourceInfoList List of archive file info from the perf-archive-datastore-get-iter ZAPI
     * @param downloadDir Root download directory
     * @param thePackage Performance package configuration
     * @throws DataSourceErrorException
     */
    @Override
    public void downloadArchiveFiles(final List<ArchiveDataSourceInfo> archiveDataSourceInfoList, final String downloadDirectory,
                                     final NetAppOCIEPerformanceConfiguration configuration) throws DataSourceErrorException {
        
        final int timeout = configuration.getPollTimeMillis() * timeoutPercentOfPolltime / 100;
        boolean done = false;
        int numberOfAttempts = 0;
        Exception lastRetriedFailure = null;
        Map<String, DownloadTask> downloadTasksToSubmit = new HashMap<String, ArchiveFileDownloader.DownloadTask>();
        final long start = System.currentTimeMillis();
        long timeConsumed = 0;
        
        // An Executor that uses a single worker thread operating off an unbounded queue. (Note however that if this single thread terminates due to a failure 
        // during execution prior to shutdown, a new one will take its place if needed to execute subsequent tasks.) Tasks are guaranteed to execute sequentially, and 
        // no more than one task will be active at any given time. Unlike the otherwise equivalent newFixedThreadPool(1) the returned executor is guaranteed not to be 
        // reconfigurable to use additional threads.
        ExecutorService executorService = Executors.newSingleThreadExecutor();

        for (final ArchiveDataSourceInfo archiveDataSourceInfo : archiveDataSourceInfoList) {
            downloadTasksToSubmit.put(archiveDataSourceInfo.getNodeName(), new DownloadTask(configuration, archiveDataSourceInfo, downloadDirectory, false));
        }
        
        do{
            // check if we have successfully collected all files
            if(downloadTasksToSubmit.size() == 0){
                done = true;
                LOG.debug("{}: Finished collection of archiver files from cluster in {} milliseconds.", configuration.getDataSourceName(), timeConsumed);
            }
            // check if we have reached the maximum allowed retries.
            else if(numberOfAttempts >= maxRetries){
                final String napeMessage = (lastRetriedFailure == null) ? "" : lastRetriedFailure.getLocalizedMessage();
                final String logMessage = "Communication problem with the cluster: " + configuration.getIP()
                        + ", Failed to download Archive files. Error: '" + napeMessage + "' on try " + numberOfAttempts + " out of " + maxRetries;
                throw new DataSourceErrorException(DataSourceErrorType.CONNECTING, logMessage, logMessage, lastRetriedFailure);
            }
            // collect files from cluster
            else{
                numberOfAttempts++;
                
                if(numberOfAttempts > 1){
                    LOG.info("{}: Retrying archive file download. Try #{} out of #{} maxRetries.", new Object[] {configuration.getDataSourceName(), numberOfAttempts, maxRetries});
                    // the apache server on ontap needs some time to restart 
                    // so we wait before we try the failed tasks again
                    try {
                        Thread.sleep(retryWaitTime);
                    } catch (InterruptedException e) {
                        String errorMessage = "Retrying archive file downloads failed after "+ numberOfAttempts +" attempts";
                        throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, errorMessage, errorMessage, e);
                    }
                }
                
                List<Future<DownloadTask>> futures = new ArrayList<Future<DownloadTask>>(downloadTasksToSubmit.size());
                for (final Map.Entry<String, DownloadTask> downloadTask : downloadTasksToSubmit.entrySet()) {    
                    futures.add(executorService.submit(downloadTask.getValue()));
                }
                try{
                    for(Future<DownloadTask> future : futures){
                        DownloadTask downloadTaskResult;
                        try {
                            downloadTaskResult = future.get(timeout - timeConsumed, TimeUnit.MILLISECONDS);
                            if(downloadTaskResult.isDownloadSuccess){
                                // if download is successful then remove downloadTask from to-submit list so that we retry only the failed ones, if any
                                downloadTasksToSubmit.remove(downloadTaskResult.archiveDataSourceInfo.getNodeName());                              
                            }else{
                                lastRetriedFailure = downloadTaskResult.downloadFailureCause;
                                // clean up stale files under the node directory where download failed
                                cleanDownloadedFiles(downloadDirectory + File.separator + downloadTaskResult.archiveDataSourceInfo.getNodeName());
                                LOG.warn("{}: Archiver file collection for node '{}' failed.", configuration.getDataSourceName(), downloadTaskResult.archiveDataSourceInfo.getNodeName());
                            }
                        }catch (InterruptedException ie){
                            future.cancel(true);
                            String errorMessage;
                            if(future.isCancelled()){
                                errorMessage = configuration.getDataSourceName() + ": Failed downloading archived files (" + ie.getMessage() + ")";
                            }else{
                                errorMessage = configuration.getDataSourceName() + ": Failed downloading archived files (" + ie.getMessage() + ") and an in-progress download could not be cancelled.";
                            }; 
                            lastRetriedFailure = ie;
                            LOG.warn("{} {}", errorMessage, ie.getCause());
                        }catch (ExecutionException | CancellationException ee){
                            final String errorMessage = configuration.getDataSourceName() + ": Failed downloading archived files (" + ee.getMessage() + ")";
                            lastRetriedFailure = ee;
                            LOG.warn("{} {}", errorMessage, ee.getCause());
                        }catch (TimeoutException te) {
                            future.cancel(true);
                            String errorMessage;
                            if(future.isCancelled()){
                                errorMessage = configuration.getDataSourceName() + ": Download wait timed out after: "+ timeout+ " milliseconds.";
                            }else{
                                errorMessage = configuration.getDataSourceName() + ": Download wait timed out after: "+ timeout+ " milliseconds and an in-progress download could not be cancelled.";
                            }
                            lastRetriedFailure = new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, errorMessage, errorMessage, te);
                            LOG.warn("{} {}", errorMessage, te.getCause());
                        }finally{
                            timeConsumed =  System.currentTimeMillis() - start;
                        }
                    }
                }finally{
                    for(Future<DownloadTask> future : futures){
                        if(!future.isDone()){
                            future.cancel(true);
                        }
                    }
                }
            }
        } // do-while ends here.
        while(!done);
        
        executorService.shutdown();
    }
    
    final static class DownloadTask implements Callable<DownloadTask> {
        private final ArchiveDataSourceInfo archiveDataSourceInfo;
        private final String downloadDirectory;
        private final NetAppOCIEPerformanceConfiguration configuration;
        private boolean isDownloadSuccess;
        private Exception downloadFailureCause;

        public DownloadTask(final NetAppOCIEPerformanceConfiguration configuration, final ArchiveDataSourceInfo archiveDataSourceInfo, final String downloadDirectory, boolean isDownLoadSuccess) {
            this.configuration = configuration;
            this.archiveDataSourceInfo = archiveDataSourceInfo;
            this.downloadDirectory = downloadDirectory;
            this.isDownloadSuccess = isDownLoadSuccess;
        }
        
        public DownloadTask setIsDownloadSuccess(boolean isDownLoadSuccess){
            this.isDownloadSuccess = isDownLoadSuccess;
            return this;
        }
        
        @Override
        public DownloadTask call() {
            try {
                Map<URL, String> downloadDetails = getDownloadDetails();
                for (final Map.Entry<URL, String> downloadDetail : downloadDetails.entrySet()){
                    downloadFile(downloadDetail.getKey(), downloadDetail.getValue());
                }
                return setIsDownloadSuccess(true);
            } catch (DataSourceErrorException e) {
                LOG.warn("{}: Unable to download archive file from node : {} to path :{} ({})", new Object[] {configuration.getDataSourceName(), archiveDataSourceInfo.getNodeName(), downloadDirectory, e.getMessage()});
                return setIsDownloadSuccess(false);
            }
        }
        
        public Map<URL, String> getDownloadDetails() throws DataSourceErrorException{
            final String clusterIp;
            String nodeName;
            String path;
            Set<String> fileNames;
            final StringBuilder url = new StringBuilder();
            final StringBuilder destinationPath = new StringBuilder();
            Map<URL, String> downloadDetailsMap = new HashMap<URL, String>();
            clusterIp = configuration.getIP();
            URL targetURL;
            // construct download url for each node and its files ( a .gz file and its meta file)
            /* sample download url:
            https://10.72.232.150/spi/sys-mcc-siteA-F3250-N1/etc/log/stats/ccma/kernel/opm/000003_000060_1392021892644_0256000.ccma.gz
             */
            path = archiveDataSourceInfo.getPath();
            nodeName = archiveDataSourceInfo.getNodeName();
            fileNames = archiveDataSourceInfo.getFileNames();
            
            if(path != null && nodeName != null){
            
                if (path.startsWith("/mroot/")){
                    path = path.substring(path.indexOf("/etc/"));
                }
    
                for (final String fileName : fileNames) {
                    
                    if("HTTPS".equals(configuration.getConnectionType())){
                        url.append("https://");
                    }else{
                        url.append("http://");
                    }
                    url.append(clusterIp).append(':').append(configuration.getPort()).append("/spi/").append(nodeName).append(path).append("/").append(fileName); 
    
                    // create destination directories if they do not already exist
                    createDownloadDirectory(destinationPath.append(this.downloadDirectory).append(File.separator).append(nodeName).toString(), configuration);    
    
                    try {
                        targetURL = new URL(url.toString());
                    } catch (MalformedURLException mURLExe) {
                        final String message = "Unable to construct a valid SPI URL. Invalid URL " + url.toString() +
                                ". MalformedURLException : " + mURLExe.getMessage();
                        throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, message, message, mURLExe);
                    }
                    destinationPath.append(File.separator).append(fileName);
                    LOG.debug("{} :Submitting download task for {} to destination path : {}", new Object[] {configuration.getDataSourceName(), url, destinationPath});
                    downloadDetailsMap.put(targetURL, destinationPath.toString());
    
                    // clear string buffers for reuse
                    url.setLength(0);
                    destinationPath.setLength(0);
                }
                
            }else{
                
                final String message = new StringBuilder()
                        .append(configuration.getDataSourceName())
                        .append(": One or more properties are Null in ArchiveDataSourceInfo. nodeName= ")
                        .append(nodeName).append(" path= ").append(path)
                        .toString();
                throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, message, message);
            }
            return downloadDetailsMap;
        }

        public void downloadFile(final URL url, final String writeToFile) throws DataSourceErrorException {

            BufferedInputStream bis = null;
            BufferedOutputStream bos = null;
            int bytesRead;
            long time = System.currentTimeMillis();

            try {
                
                LOG.debug("{}: Begin downloading : {} to path : {}", new Object[] {configuration.getDataSourceName(), url, writeToFile});
                final HttpURLConnection httpURLConnection = ServiceManager.getInstance().getCommunicationManager().
                        openAuthenticatedHttpConnection(url, configuration.getUser(), configuration.getPassword());

                bis = new BufferedInputStream(httpURLConnection.getInputStream());
                bos = new BufferedOutputStream(new FileOutputStream(new File(writeToFile)));
                while ((bytesRead = bis.read()) != -1) {
                    bos.write(bytesRead);
                }
                bos.flush();
                   long completedIn = System.currentTimeMillis() - time;
                   LOG.debug("{}: Completed downloading : {} to path : {} in {} milliseceonds", new Object[] {configuration.getDataSourceName(), url, writeToFile, String.valueOf(completedIn)});


            } catch (MalformedURLException mue) {
                final String errorMessage = configuration.getDataSourceName() + ": MalformedURLException Failed downloading archived file:" + url.toString()+" to path :" + writeToFile + "(" + mue.getMessage() + ")";
                downloadFailureCause = mue;
                throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, errorMessage, errorMessage, mue);
            } catch (IOException ioe) {
                final String errorMessage = configuration.getDataSourceName() + ": IOException, Failed downloading archived file:" + url.toString()+" to path :" + writeToFile + "(" + ioe.getMessage() + ")";
                downloadFailureCause = ioe;
                throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, errorMessage, errorMessage, ioe);
            } catch (FrameworkException frameExe) {
                final String errorMessage = configuration.getDataSourceName() + ": FrameworkException, Failed downloading archived file:" + url.toString()+" to path :" + writeToFile + "(" + frameExe.getMessage() + ")";
                downloadFailureCause = frameExe;
                throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, errorMessage, errorMessage, frameExe);
            } catch (Exception exp){
                final String errorMessage = configuration.getDataSourceName() + ": Failed downloading archived file:" + url.toString()+" to path :" + writeToFile + "(" + exp.getMessage() + ")";
                downloadFailureCause = exp;
                throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, errorMessage, errorMessage, exp);
            }
            
            finally {
                if (bis != null) {
                    try {
                        bis.close();
                    } catch (IOException ignore){
                        
                    }
                }
                if (bos != null) {
                    try {
                        bos.close();
                    } catch (IOException ignore) {
                    
                    }
                }
            }
        }
        
        public void createDownloadDirectory(String directoryPath, NetAppOCIEPerformanceConfiguration configuration) throws DataSourceErrorException{
            final File file = new File(directoryPath);
            if (!file.exists()){
                final boolean isMkDirsSuccess = file.mkdirs();
                if (!isMkDirsSuccess){
                    final String message = configuration.getDataSourceName() + ": Failed to create archive download destination directories. Dir Path :" + file.getAbsolutePath();
                    throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, message, message);
                }
            }
            // file exists but not a directory
            else if(file.exists() && !file.isDirectory()){
                // delete that file and create a directory with same
                LOG.debug("A file by name {} already exists but is not a directory.",file.getAbsolutePath());
                if(file.delete()){
                    final boolean dirSuccess = file.mkdirs();
                    if (!dirSuccess){
                        final String message = configuration.getDataSourceName() + ": Deleted existing file but failed to recreate archive download destination directories. Dir Path :" + file.getAbsolutePath();
                        throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, message, message);
                    }
                }else{ 
                    final String message = configuration.getDataSourceName() + "A file by name " + file.getAbsolutePath() + " already exists and cannot be deleted."; 
                    throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, message, message);
                }
                
            }
            // if directory exists but does not have write permissions
            if(file.exists() && !file.canWrite()){
                final String message = configuration.getDataSourceName() + ": Archiver files download directory does not have write access permissions. Dir Path :" + file.getAbsolutePath(); 
                throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, message, message);
            }
        }
    }
}
