package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation;

import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.EnumWrapper;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.EnumWrapperList;
import org.jdom.Element;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Util class used in various builders for reading/navigating the zapi response.  
 */
public class BuilderUtil {

    public static List<String> getListFromChildren(final Element elt, final String targetChildName, final String targetGrandchildName) {
        return getListFromChildren(elt, targetChildName, targetGrandchildName, 100);
    }

    public static List<String> getListFromChildren(final Element elt, final String targetChildName, final String targetGrandchildName, final int lengthHint) {
        final List<String> valList = new ArrayList<String>(lengthHint);
        if (elt != null) {
            @SuppressWarnings({"unchecked"})
            final List<Element> childElts = elt.getChildren(targetChildName);
            for (final Element childElt : childElts) {
                final Element grandChildElt = childElt.getChild(targetGrandchildName);
                if (grandChildElt != null) {
                    final String val = grandChildElt.getTextTrim();
                    valList.add(val);
                }
            }
        }
        return valList;
    }

    public static List<String> getListFromGrandChildren(final Element elt, final String containingChildName, final String repeatingGrandChildName) {
        return getListFromGrandChildren(elt, containingChildName, repeatingGrandChildName, 100);
    }

    public static List<String> getListFromGrandChildren(final Element elt, final String containingChildName, final String repeatingGrandChildName, final int lengthHint) {
        final List<String> valList = new ArrayList<String>(lengthHint);
        if (null != elt) {
            final Element childElement = elt.getChild(containingChildName);
            if (null != childElement) {
                @SuppressWarnings({"unchecked"})
                final List<Element> repeatingGrandChildElements = childElement.getChildren(repeatingGrandChildName);
                for (final Element grandChildElement : repeatingGrandChildElements) {
                    final String val = grandChildElement.getTextTrim();
                    valList.add(val);
                }
            }
        }
        return valList;
    }

    public static Long parseLongFromChild(final Element elt, final String targetChildName) {
        final String rawVal = getChildTextTrim(elt, targetChildName);
        return parseLongFromString(rawVal);
    }

    public static Long parseLongFromChildDefault(final Element elt, final String targetChildName, final long defaultVal) {
        final Long returnCandidate = parseLongFromChild(elt, targetChildName);
        return (returnCandidate == null) ? defaultVal : returnCandidate;
    }

    public static Long parseLongFromString(final String val) {
        Long returnVal = null;
        try {
            returnVal = Long.parseLong(val);
        }
        catch (NumberFormatException nfe) {
            // Ignore
        }
        return returnVal;
    }

    public static Boolean parseBooleanFromChild(final Element elt, final String targetChildName) {
        final String rawVal = getChildTextTrim(elt, targetChildName);
        return parseBooleanFromString(rawVal);
    }

    public static Boolean parseBooleanFromChildDefaultFalse(final Element elt, final String targetChildName) {
        final Boolean returnCandidate = parseBooleanFromChild(elt, targetChildName);
        return (returnCandidate == null) ? Boolean.FALSE : returnCandidate;
    }

    public static Boolean parseBooleanFromString(final String rawVal) {
        return (rawVal == null) ? null : Boolean.parseBoolean(rawVal);
    }

    public static Integer parseIntFromChild(final Element elt, final String targetChildName) {
        final String rawVal = getChildTextTrim(elt, targetChildName);
        return parseIntFromString(rawVal);
    }

    public static Integer parseIntFromString(final String rawVal) {
        Integer returnVal = null;
        // Parse rawVal only if likely to succeed
        if ((rawVal != null) && !rawVal.isEmpty()) {
            try {
                returnVal = Integer.parseInt(rawVal);
            }
            catch (NumberFormatException nfe) {
                // Ignore
            }
        }
        return returnVal;
    }

    public static <T extends Enum<T>> T getEnumByValue(final Class<T> enumClass, final Element elt, final String targetChildName, final T defaultVal) {
        final String rawVal = getChildTextTrim(elt, targetChildName);
        return getEnumByValue(enumClass, rawVal, defaultVal);
    }

    public static <T extends Enum<T>> T getEnumByValue(final Class<T> enumClass, final String rawValue, final T defaultVal) {
        if (rawValue == null || rawValue.isEmpty()) {
            return null;        // Undefined
        }
        T returnVal = defaultVal;
        final EnumSet<T> enumSet = EnumSet.allOf(enumClass);
        for (final T candidate : enumSet) {
            if (candidate.name().equalsIgnoreCase(rawValue)) {
                returnVal = candidate;
                break;
            }
        }
        return returnVal;
    }

    public static <T extends Enum<T>> EnumWrapper<T> getEnumWrapperByValue(final Class<T> enumClass, final Element elt, final String targetChildName, final T defaultVal) {
        final String rawVal = getChildTextTrim(elt, targetChildName);
        return getEnumWrapperByValue(enumClass, rawVal, defaultVal);
    }

    public static <T extends Enum<T>> EnumWrapper<T> getEnumWrapperByValue(final Class<T> enumClass, final String rawValue, final T defaultVal) {
        if (rawValue == null) {
            return new EnumWrapper<T>(rawValue, null);        // Undefined
        }
        if (rawValue.isEmpty()) {
            return new EnumWrapper<T>(rawValue, defaultVal);
        }
        T returnVal = defaultVal;
        final EnumSet<T> enumSet = EnumSet.allOf(enumClass);
        for (final T candidate : enumSet) {
            if (candidate.name().equalsIgnoreCase(rawValue)) {
                returnVal = candidate;
                break;
            }
        }
        return new EnumWrapper<T>(rawValue, returnVal);
    }

    @SuppressWarnings("unchecked")
    public static <T extends Enum<T>> EnumWrapperList<T> getEnumWrapperListByValue(final Class<T> enumClass,
            final Element elt, final String targetListName, final String targetChildName) {
        final EnumWrapperList<T> returnVal = new EnumWrapperList<T>();
        final Element childList = elt.getChild(targetListName);
        if (childList != null) {
            for (final Element child : (List<Element>)childList.getChildren(targetChildName)) {
                final String rawVal = child.getTextTrim();
                returnVal.add(getEnumWrapperByValue(enumClass, rawVal, null));
            }
        }
        return returnVal;
    }

    public static String getChildText(final Element elt, final String targetChildName) {
        if (elt == null) {
            return null;
        }

        final String val = elt.getChildText(targetChildName);
        return (val == null || val.isEmpty()) ? null : val;
    }

    public static String getChildTextDefaultEmpty(final Element elt, final String targetChildName) {
        final String candidateText = getChildText(elt, targetChildName);
        return (candidateText == null) ? "" : candidateText;
    }

    public static String getChildTextTrim(final Element elt, final String targetChildName) {
        if (elt == null) {
            return null;
        }

        final String val = elt.getChildTextTrim(targetChildName);
        return (val == null || val.isEmpty()) ? null : val;
    }

    public static String getChildTextTrimDefaultEmpty(final Element elt, final String targetChildName) {
        final String candidateText = getChildTextTrim(elt, targetChildName);
        return (candidateText == null) ? "" : candidateText;
    }

    // Within the model we have 3 different WWN formats that must be normalized:
    // 1) 50:0a:09:80:89:da:c0:f4
    // 2) 500a09820002f3cc
    // 3) 5:00a:098300:02f3cb

    // The Normal format is
    // 50:0A:09:80:89:DA:C0:F4

    // 8 groups of 2 digits like: 50:0a:09:80:89:da:c0:f4
    private static final Pattern WWN_PATTERN_1 = Pattern.compile("^([\\da-fA-F]{2}+):([\\da-fA-F]{2}+):([\\da-fA-F]{2}+):([\\da-fA-F]{2}+):([\\da-fA-F]{2}+):([\\da-fA-F]{2}+):([\\da-fA-F]{2}+):([\\da-fA-F]{2}+)$");

    // 16 digits no groups like: 500a09820002f3cc
    private static final Pattern WWN_PATTERN_2 = Pattern.compile("^([\\da-fA-F]{16}+)$");

    // 1, 3, 6, 6, like: 5:00a:098300:02f3cb
    private static final Pattern WWN_PATTERN_3 = Pattern.compile("^([\\da-fA-F]):([\\da-fA-F]{3}+):([\\da-fA-F]{6}+):([\\da-fA-F]{6}+)$");

    private static String normalizeWWN1(final String wwn1) {
        return wwn1.toUpperCase();
    }

    private static String normalizeWWN2(final String wwn2) {
        final StringBuilder wwnBuilder = new StringBuilder(24);
        wwnBuilder.append(wwn2.substring(0, 2));
        wwnBuilder.append(":");
        wwnBuilder.append(wwn2.substring(2, 4));
        wwnBuilder.append(":");
        wwnBuilder.append(wwn2.substring(4, 6));
        wwnBuilder.append(":");
        wwnBuilder.append(wwn2.substring(6, 8));
        wwnBuilder.append(":");
        wwnBuilder.append(wwn2.substring(8, 10));
        wwnBuilder.append(":");
        wwnBuilder.append(wwn2.substring(10, 12));
        wwnBuilder.append(":");
        wwnBuilder.append(wwn2.substring(12, 14));
        wwnBuilder.append(":");
        wwnBuilder.append(wwn2.substring(14, 16));
        return wwnBuilder.toString().toUpperCase();
    }

    /**
     * Normalize a WWN like 5:00a:098300:02f3cb to 50:0A:09:83:00:02:F3:CB
     *
     * @param wwn3 A WWN similar to: 5:00a:098300:02f3cb
     *
     * @return A WWN similar to 50:0A:09:83:00:02:F3:CB
     */
    private static String normalizeWWN3(final String wwn3) {
        final StringBuilder wwnBuilder = new StringBuilder(24);
        wwnBuilder.append(wwn3.substring(0, 1));
        wwnBuilder.append(wwn3.substring(2, 3));
        wwnBuilder.append(":");
        wwnBuilder.append(wwn3.substring(3, 5));
        wwnBuilder.append(":");
        wwnBuilder.append(wwn3.substring(6, 8));
        wwnBuilder.append(":");
        wwnBuilder.append(wwn3.substring(8, 10));
        wwnBuilder.append(":");
        wwnBuilder.append(wwn3.substring(10, 12));
        wwnBuilder.append(":");
        wwnBuilder.append(wwn3.substring(13, 15));
        wwnBuilder.append(":");
        wwnBuilder.append(wwn3.substring(15, 17));
        wwnBuilder.append(":");
        wwnBuilder.append(wwn3.substring(17, 19));
        return wwnBuilder.toString().toUpperCase();
    }

    /**
     * Best effort to normalize WWNs to regular format.
     *
     * @param rawWWN raw WWN to normalize
     *
     * @return if the raw WWN is in a recongnized format, then the normalized WWN is returned.
     *         If the raw WWN format is not recognized, the raw WWN is returned.
     */
    public static String normalizeWWN(final String rawWWN) {
        if (rawWWN != null) {
            final Matcher wwn1Matcher = WWN_PATTERN_1.matcher(rawWWN);
            if (wwn1Matcher.lookingAt()) {
                return normalizeWWN1(rawWWN);
            }
            final Matcher wwn2Matcher = WWN_PATTERN_2.matcher(rawWWN);
            if (wwn2Matcher.lookingAt()) {
                return normalizeWWN2(rawWWN);
            }
            final Matcher wwn3Matcher = WWN_PATTERN_3.matcher(rawWWN);
            if (wwn3Matcher.lookingAt()) {
                return normalizeWWN3(rawWWN);
            }
        }
        // Unrecognized format, return original value
        return rawWWN;
    }

    public static String getListAsCSVString(final List<String> list) {
        if (list == null || list.size() == 0) {
            return null;
        }
        final StringBuilder csvBuilder = new StringBuilder();
        for (final String val : list) {
            if (csvBuilder.length() > 0) {
                csvBuilder.append(',');
            }
            csvBuilder.append(val);
        }
        return csvBuilder.toString();
    }


    /**
     * Extracts the value from strExtract string.
     * E.g vserverID(234);processor(34)
     * srcString = vserverID(234);processor(34) and  strToExtract=vserverID
     * return value = 234
     * Internal Regex Pattern : (;strToExtract|\\AstrToExtract).(\\d+)
     * @param srcString
     * @param strToExtract
     * @return
     */
    public static String extractIdFromParens(String srcString, String strToExtract) {

        if(srcString == null) {
            return null;
        }

        final StringBuilder regexPattern = new StringBuilder();
        regexPattern.append("(;").append(strToExtract).append("|\\A").append(strToExtract).append(")");
        regexPattern.append(".(\\d+)");

        Pattern p = Pattern.compile(regexPattern.toString(), Pattern.CASE_INSENSITIVE | Pattern.DOTALL);
        Matcher m = p.matcher(srcString);
        return (m.find()) ?  m.group(2) : null;
    }
}
