package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation;

import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.cmode.*;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.ISCSISecurityEntry;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.ISCSISessionConnection;
import org.apache.log4j.Logger;
import org.jdom.Element;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

public class ISCSIBuilder extends  BuilderBase<CModeCluster> {

    @SuppressWarnings({"UnusedDeclaration"})
    private static final Logger logger = Logger.getLogger(ISCSIBuilder.class);
    private static final Map<String, String> initiatorNodeNameToAliasName = new HashMap<String, String>();

    public void buildModel(final CModeCluster cluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException {

        // Clear out map before starting
        initiatorNodeNameToAliasName.clear();           // TODO maps in builder should be in transform
        new ISCSIServiceBuilder().buildModel(cluster, zapiConnection);
        new ISCSIPortalGroupBuilder().buildModel(cluster, zapiConnection);
        new ISCSIPortalGroupToLIFBuilder().buildModel(cluster, zapiConnection);
        // Initiators need to be discovered before Sessions so we can use the initiatorNodeNameToAliasName Map
        new ISCSIInitiatorBuilder().buildModel(cluster, zapiConnection);
        new ISCSISessionBuilder().buildModel(cluster, zapiConnection);
        new ISCSISecurityEntryBuilder().buildModel(cluster, zapiConnection);

        // Clear out the map when finished
        initiatorNodeNameToAliasName.clear();
    }

    public static class ISCSIPortalGroupToLIFBuilder extends IterBuilderBase {

        public ISCSIPortalGroupToLIFBuilder() {
            super("iscsi-interface-get-iter", "iscsi-interface-list-entry-info", getDefaultMaxRecords(), new LinkedHashMap<String[], String[]>() {{
                put(new String[]{"iscsi-interface-list-entry-info"}, new String[]{"interface-name", "ip-address", "ip-port", "tpgroup-name", "tpgroup-tag", "vserver"});
            }});
        }

        protected void buildIterInstance(final CModeCluster cluster, final Element iscsiIFEntryElt) {
            final String interfaceName = BuilderUtil.getChildTextTrim(iscsiIFEntryElt, "interface-name");
            final String ipAddressString = BuilderUtil.getChildTextTrim(iscsiIFEntryElt, "ip-address");
            final String ipPortString = BuilderUtil.getChildTextTrim(iscsiIFEntryElt, "ip-port");
            final String tpGroupName = BuilderUtil.getChildTextTrim(iscsiIFEntryElt, "tpgroup-name");
            final Integer tpGroupTag = BuilderUtil.parseIntFromChild(iscsiIFEntryElt, "tpgroup-tag");
            final String vServerName = BuilderUtil.getChildTextTrim(iscsiIFEntryElt, "vserver");
            final CModeISCSIPortalGroupToLIF cModeISCSIPortalGroupToLIF = cluster.createISCSIPortalGroupToLIF(vServerName, interfaceName, tpGroupName);
            cModeISCSIPortalGroupToLIF.setIpAddress(ipAddressString);
            cModeISCSIPortalGroupToLIF.setPortalGroupTag(tpGroupTag);
            cModeISCSIPortalGroupToLIF.setIpPort(ipPortString);     // JSD TODO Integer?
            cluster.getISCSIPortalGroupToLIFs().add(cModeISCSIPortalGroupToLIF);
        }
    }

    public static class ISCSIPortalGroupBuilder extends IterBuilderBase {

        public ISCSIPortalGroupBuilder() {
            super("iscsi-tpgroup-get-iter", "iscsi-tpgroup-list-entry-info", getDefaultMaxRecords(), new LinkedHashMap<String[], String[]>() {{
                put(new String[]{"iscsi-tpgroup-list-entry-info"}, new String[]{"tpgroup-name", "tpgroup-tag", "vserver"});
            }});
        }

        @Override
        protected void buildIterInstance(final CModeCluster cluster, final Element tpGroupElt) {
            final String vServerName = BuilderUtil.getChildTextTrim(tpGroupElt, "vserver");
            final String portalGroupName = BuilderUtil.getChildTextTrim(tpGroupElt, "tpgroup-name");
            final Integer portalGroupTag = BuilderUtil.parseIntFromChild(tpGroupElt, "tpgroup-tag");
            final CModeISCSIPortalGroup cModeISCSIPortalGroup = cluster.createISCSIPortalGroup(vServerName, portalGroupName);
            cModeISCSIPortalGroup.setTag(portalGroupTag);
            cluster.getISCSIPortalGroups().add(cModeISCSIPortalGroup);
        }
    }

    public static class ISCSISessionBuilder extends IterBuilderBase {

        public ISCSISessionBuilder() {
            super("iscsi-session-get-iter", "iscsi-session-list-entry-info", getDefaultMaxRecords(), new LinkedHashMap<String[], String[]>() {{
                put(new String[]{"iscsi-session-list-entry-info"}, new String[]{"vserver", "tpgroup-name", "target-session-id", "max-connections", "initiator-nodename", "isid"});
                put(new String[]{"iscsi-session-list-entry-info", "iscsi-session-connection-list-entries", "iscsi-session-connection-list-entry-info"}, new String[]{"authentication-method", "connection-id", "interface-name", "local-ip-address", "local-ip-port", "remote-ip-address"});
            }});

        }

        @Override
        protected void buildIterInstance(final CModeCluster cluster, final Element iscsiSessionElt) {
            final String vServerName = BuilderUtil.getChildTextTrim(iscsiSessionElt, "vserver");
            final String portalGroupName = BuilderUtil.getChildTextTrim(iscsiSessionElt, "tpgroup-name");
            final Integer targetSessionID = BuilderUtil.parseIntFromChild(iscsiSessionElt, "target-session-id");
            final Integer maxConnections = BuilderUtil.parseIntFromChild(iscsiSessionElt, "max-connections");
            final String initiatorNodeName = BuilderUtil.getChildTextTrim(iscsiSessionElt, "initiator-nodename");
            final String initiatorAliasName = initiatorNodeNameToAliasName.get(initiatorNodeName);  // TODO in transform!
            final String initiatorSessionID = BuilderUtil.getChildTextTrim(iscsiSessionElt, "isid");
            final CModeISCSISession cModeISCSISession = cluster.createISCSISession(vServerName, portalGroupName, targetSessionID);
            cModeISCSISession.setMaxConnections(maxConnections);
            cModeISCSISession.setInitiatorNodeName(initiatorNodeName);
            cModeISCSISession.setInitiatorAliasName(initiatorAliasName);
            cModeISCSISession.setInitiatorSessionID(initiatorSessionID);

            final Element connectionListEntries = iscsiSessionElt.getChild("iscsi-session-connection-list-entries");
            if (connectionListEntries == null) {
                return;
            }
            @SuppressWarnings({"unchecked"})
            final List<Element> connectionEntries = connectionListEntries.getChildren("iscsi-session-connection-list-entry-info");
            for (final Element connectionEntry : connectionEntries) {
                final ISCSISessionConnection.ISCSISessionAuthenticationStyle authenticationMethod = BuilderUtil.getEnumByValue(ISCSISessionConnection.ISCSISessionAuthenticationStyle.class, connectionEntry, "authentication-method", ISCSISessionConnection.ISCSISessionAuthenticationStyle.NOT_MAPPED);
                final Integer connectionID = BuilderUtil.parseIntFromChild(connectionEntry, "connection-id");
                final String interfaceName = BuilderUtil.getChildTextTrim(connectionEntry, "interface-name");
                final String localIPAddress = BuilderUtil.getChildTextTrim(connectionEntry, "local-ip-address");
                final String localIPPort = BuilderUtil.getChildTextTrim(connectionEntry, "local-ip-port");     // JSD TODO Integer?
                final String remoteIPAddress = BuilderUtil.getChildTextTrim(connectionEntry, "remote-ip-address");
                final CModeISCSISessionConnection cModeISCSISessionConnection = cModeISCSISession.createCModeISCSISessionConnection(connectionID);
                cModeISCSISessionConnection.setAuthenticationMethod(authenticationMethod);
                cModeISCSISessionConnection.setLocalIPAddress(localIPAddress);
                cModeISCSISessionConnection.setLocalIPPort(localIPPort);
                cModeISCSISessionConnection.setRemoteIPAddress(remoteIPAddress);
                cModeISCSISessionConnection.setInterfaceName(interfaceName);
                cModeISCSISession.getCModeISCSISessionConnections().add(cModeISCSISessionConnection);
            }
            cluster.getISCSISessions().add(cModeISCSISession);
        }
    }

    public static class ISCSIInitiatorBuilder extends IterBuilderBase {
        public ISCSIInitiatorBuilder() {
            super("iscsi-initiator-get-iter", "iscsi-initiator-list-entry-info", getDefaultMaxRecords(), new LinkedHashMap<String[], String[]>() {{
                put(new String[]{"iscsi-initiator-list-entry-info"}, new String[]{"initiator-nodename", "initiator-aliasname"});
            }});
        }

        @Override
        protected void buildIterInstance(final CModeCluster cluster, final Element initiatorInfoElt) {
            final String initiatorNodeName = BuilderUtil.getChildTextTrim(initiatorInfoElt, "initiator-nodename");
            final String rawInitiatorAliasName = BuilderUtil.getChildTextTrim(initiatorInfoElt, "initiator-aliasname");
            final String initiatorAliasName = (null != rawInitiatorAliasName) ? rawInitiatorAliasName : "";      // JSD TODO default
            initiatorNodeNameToAliasName.put(initiatorNodeName, initiatorAliasName);
        }
    }

    public static class ISCSISecurityEntryBuilder extends IterBuilderBase {
        public ISCSISecurityEntryBuilder() {
            super("iscsi-initiator-auth-get-iter", "iscsi-security-entry-info", getDefaultMaxRecords(), new LinkedHashMap<String[], String[]>() {{
                put(new String[]{"iscsi-security-entry-info"}, new String[]{"vserver", "initiator", "auth-type", "user-name", "outbound-user-name"});
            }});
        }

        @Override
        protected void buildIterInstance(final CModeCluster cluster, final Element iscsiSecurityElt) {
            final String initiator = BuilderUtil.getChildTextTrim(iscsiSecurityElt, "initiator");
            final String vServerName = BuilderUtil.getChildTextTrim(iscsiSecurityElt, "vserver");
            final ISCSISecurityEntry.ISCSISecurityEntryAuthenticationType authenticationType =
                    BuilderUtil.getEnumByValue(ISCSISecurityEntry.ISCSISecurityEntryAuthenticationType.class, iscsiSecurityElt,
                            "auth-type", ISCSISecurityEntry.ISCSISecurityEntryAuthenticationType.NOT_MAPPED);
            final String username = BuilderUtil.getChildTextTrimDefaultEmpty(iscsiSecurityElt, "user-name");
            final String outboundUsername = BuilderUtil.getChildTextTrimDefaultEmpty(iscsiSecurityElt, "outbound-user-name");

            if (vServerName == null) {
                // Older ONTAP versions don't identify the vServer, so there is no point to continue
                if (logger.isDebugEnabled()) {
                    logger.debug("Ignoring ISCSI security entry with no vServer");
                }
                return;
            }

            final CModeISCSISecurityEntry cModeISCSISecurityEntry = cluster.createISCSISecurityEntry(vServerName, initiator);
            cModeISCSISecurityEntry.setAuthenticationType(authenticationType);
            cModeISCSISecurityEntry.setUsername(username);
            cModeISCSISecurityEntry.setOutboundUsername(outboundUsername);
            cluster.getISCSISecurityEntries().add(cModeISCSISecurityEntry);
        }
    }

    public class ISCSIServiceBuilder extends IterBuilderBase {
        public ISCSIServiceBuilder() {
            super("iscsi-service-get-iter", "iscsi-service-info", getDefaultMaxRecords(), new LinkedHashMap<String[], String[]>() {{
                put(new String[]{"iscsi-service-info"}, new String[]{"alias-name", "is-available", "node-name", "vserver",});
            }});
        }

        @Override
        protected void buildIterInstance(final CModeCluster cluster, final Element iscsiServiceElt) {
            final String iqn = BuilderUtil.getChildTextTrim(iscsiServiceElt, "node-name");
            final String vServerName = BuilderUtil.getChildTextTrim(iscsiServiceElt, "vserver");
            final String aliasName = BuilderUtil.getChildTextTrim(iscsiServiceElt, "alias-name");
            final Boolean isAvailable = BuilderUtil.parseBooleanFromChild(iscsiServiceElt, "is-available");
            final CModeISCSIService iscsiService = cluster.createCModeISCSIService(iqn);
            iscsiService.setVServerName(vServerName);
            iscsiService.setAliasName(aliasName);
            iscsiService.setAvailable(isAvailable);
            cluster.getIscsiServices().add(iscsiService);
        }
    }
}                                                                                               
