package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation;

import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPINames;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIUtils;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZapiCommand;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileWriter;
import netapp.manage.NaElement;
import org.jdom.Element;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

/**
 * Base clase for iterator based foundation builders.   Handles iterator tag management and allows derived
 * classes to parse the repeated XML element.
 */
public abstract class IterBuilderBaseImpl<M extends BuilderModel> extends BuilderBase<M> {

    private static final Logger LOG = LoggerFactory.getLogger(IterBuilderBaseImpl.class);

    private final String zapiCommand;
    private final String repeatingInstanceEltName;
    private int maxRecords;
    private final NaElement desiredAttributes;
    private final NaElement[] additionalElements;

    private static int defaultMaxRecords = 1000;
    private boolean forceDeviceCall;

    protected abstract void buildIterInstance(final M model, final Element iterationElt);

    public final String getZapiCommand()
    {
      return zapiCommand;
    }

  public static int getDefaultMaxRecords() {
        return defaultMaxRecords;
    }

    public static void setDefaultMaxRecords(final int defaultMaxRecords) {
        IterBuilderBaseImpl.defaultMaxRecords = defaultMaxRecords;
    }

  public boolean isForceDeviceCall()
  {
    return forceDeviceCall;
  }

  public void setForceDeviceCall(boolean forceDeviceCall)
  {
    this.forceDeviceCall = forceDeviceCall;
  }

  /**
     * Create a new builder for a specific iterator based ZAPI
     *
     * @param zapiCommand              the iterator based ZAPI to invoke
     * @param repeatingInstanceEltName the name of the top level element that iterates (appears more than once, repeats)
     * @param maxRecords               the requested maximum record chunk size.   Usually defaults to 1000 instances (DEFAULT_MAX_RECORDS)
     * @param desiredAttributes        a Map where the Key represents a path to an Element of interest, and the value represents List of desired child elements.
     *                                 The Key is an Array of Strings that represents the path to the XML Element containing desired attributes.
     *                                 The Value is an Array of Strings that represents the child elements of interest.
     * @param additionalElements       optional Ordered Map of Key, Value pairs used as additional parameters to the ZAPI.  It is important that
     *                                 this Map be ordered so that the toString of the request is consistent, and may be used as a deterministic key for lookup
     *                                 during recording playback.
     */
    protected IterBuilderBaseImpl(final String zapiCommand, final String repeatingInstanceEltName, final int maxRecords,
                                  final NaElement desiredAttributes, final NaElement... additionalElements) {
        this.zapiCommand = zapiCommand;
        this.repeatingInstanceEltName = repeatingInstanceEltName;
        this.maxRecords = maxRecords;
        this.desiredAttributes = desiredAttributes;
        this.additionalElements = additionalElements;
    }

  public int getMaxRecords()
  {
    return maxRecords;
  }

  public void setMaxRecords(int maxRecords)
  {
    this.maxRecords = maxRecords;
  }

  @Override
    public void buildModel(final M model, final ZAPIConnection zapiConnection) throws DataSourceErrorException
    {
        // ZAPI iterator support keeping track of nextTag, invoking command multiple times until all results returned
        final String logPrefix = PackageContext.getLogPrefix();
        String nextTag = null;      // keeps track of next tag to request when iterating.   Not specified on the first call.
        int iterCount = 0;
        final List<Integer> iterChunkCounts = new ArrayList<Integer>(100);
        final long startTime = System.currentTimeMillis();
        long totalZapiTime = 0;
        int totalZapiInvocations = 0;
        int totalElements = 0;

        Integer maxRecordsOverride = null;

        do {
            // Build XML request
            final NaElement request = ZAPIUtils.elem(zapiCommand);
            if (additionalElements != null) {
                for (final NaElement element : additionalElements) {
                    request.addChildElem(element);
                }
            }

            if (maxRecordsOverride != null) {
                LOG.info("Override max records: {} {}", zapiCommand, maxRecordsOverride);
            }

            request.addChildElem(ZAPIUtils.elem(ZAPINames.MAX_RECORDS, (maxRecordsOverride == null) ? maxRecords : maxRecordsOverride));
            if (nextTag != null) {
                // Not the first call, and there are more records to iterate.   Specify where iteration left off
                request.addChildElem(ZAPIUtils.elem(ZAPINames.TAG, nextTag));
            }
            if (desiredAttributes != null) {
                request.addChildElem(desiredAttributes);
            }

            // Get next chunk ot iteration elements
            final ZAPIConnection.ZapiResponse response = zapiConnection.execute( new ZapiCommand( request, iterCount, forceDeviceCall ) );

            //handle response
            iterCount++;
            totalZapiTime += response.getTotalTime();
            totalZapiInvocations += response.getInvocationCount();
            final Integer overrideCandidate = response.getMaxRecordsOverride();
            if (overrideCandidate != null)  {
                // A success after a failure will NOT populate maxRecordsOverride, but we STILL
                // want to keep the override.  If we set it back to null, we will again get the default!
                maxRecordsOverride = overrideCandidate;
            }

            final Element resultsElt = response.getElement();
            if(resultsElt == null || resultsElt.getChild("attributes-list") == null){
            	iterChunkCounts.add( 0 );
                break;
            }
            // Now iterate through returned elements
            @SuppressWarnings({"unchecked"})
            final List<Element> iterationElts = resultsElt.getChild("attributes-list").getChildren(repeatingInstanceEltName);
            final int chunkSizeReturned = iterationElts.size();
            iterChunkCounts.add( chunkSizeReturned );
            totalElements += chunkSizeReturned;
            for (final Element iterationElt : iterationElts) {
                buildIterInstance(model, iterationElt);
            }
            // Are there more records to iterate?
            final Element nextTagElt = resultsElt.getChild("next-tag");   // won't exist if no more records
            if (nextTagElt == null) {
                // No more elements to iterate
                nextTag = null;
            }
            else {
                /* Remove whitespace, including newlines, in the nextTag used
                 * in the next request, in order to simplify simulation when the
                 * recording platform is not the same as the simulator platform.
                 */
                nextTag = nextTagElt.getTextTrim().replaceAll("\\s", "");
            }
        }
        while (nextTag != null);

        final long consumedTime = System.currentTimeMillis() - startTime;

        final String fqZapiCommand = ("perf-object-instance-list-info-iter".equals(zapiCommand)) ? zapiCommand + " "
                + getAdditionalElementValue("objectname") : zapiCommand;
        LOG.debug("{} - {}: max records: {}, iterations: {}, total clock time: {}ms, total zapi time: {}ms, total zapi invocations {}, total elements: {}, buckets: {}",
                new Object[] { logPrefix, fqZapiCommand,
                ((maxRecordsOverride == null) ? maxRecords : maxRecordsOverride),
                iterCount, consumedTime, totalZapiTime, totalZapiInvocations,
                totalElements, getCSVStringFromIntList(iterChunkCounts) });

        try {
            final DSFileWriter responseTimesWriter = zapiConnection.getResponseTimesWriter();
            if( responseTimesWriter !=  null )
            {
              responseTimesWriter.writeln( ", ", PackageContext.getPackage().getConfiguration().getDataSourceName(),
                      fqZapiCommand, new Date( startTime ),
                      ((maxRecordsOverride == null) ? maxRecords : maxRecordsOverride),
                      iterCount, consumedTime, totalZapiTime, totalZapiInvocations,
                      totalElements, getCSVStringFromIntList( iterChunkCounts ) );
            }
        }
        catch (IOException ioe) {
            LOG.warn("{} - Failed to record ZAPI response time for: {}",
                    logPrefix, zapiCommand);
        }
    }

    protected String getAdditionalElementValue(final String name) {
        if (additionalElements != null) {
            for (final NaElement element : additionalElements) {
                if (name.equals(element.getName())) {
                    return element.getContent();
                }
            }
        }
        return null;
    }

    protected String getCSVStringFromIntList(final List<Integer> list) {
        final int listLen = list.size();
        if (listLen == 0) {
            return "";
        }
        final StringBuilder builder = new StringBuilder(listLen * 6);  // estimate for digits plus comma plus space for each entry
        builder.append(list.get(0));
        if (listLen > 1) {
            for (final Integer elt : list.subList(1, listLen)) {
                builder.append(", ");
                builder.append(elt);
            }
        }
        return builder.toString();
    }

}
