package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappstorageperformance;

import com.netapp.oci.netapp.performance.iscsilif.interfaces.data.IscsiLifMeasurements;
import com.netapp.oci.netapp.performance.iscsilif.interfaces.data.IscsiLifRawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetAppISCSILIFMetrics;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.VServer;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.apache.log4j.Logger;

import java.util.*;

public class ISCSILIFMetricsBuilder extends MetricsBuilderBase<VServer.VServerKey,
                                                               IscsiLifMeasurements,
                                                               IscsiLifRawMeasurements,
                                                               NetAppISCSILIFMetrics,
                                                               NetAppISCSILIFMetrics.ISCSILIFCounter,
                                                               NetAppISCSILIFMetrics.ISCSILIFCounter> {

    private static final Logger logger = Logger.getLogger(ISCSILIFMetricsBuilder.class);
    private final Map<String, String> iscsiLifPerfUUIDToVServerUUID = new HashMap<String, String>();
    private final Map<String, String> iscsiLifPerfUUIDToLifFoundationUUID = new TreeMap<String, String>();

    public ISCSILIFMetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "iscsi_lif", NetAppISCSILIFMetrics.ISCSILIFCounter.class);
    }

    @Override
    List<String> getTargetPerfUUIDs(final NetAppDefaultStorageContext context, final ZAPIConnection zapiConnection) {
        iscsiLifPerfUUIDToVServerUUID.clear();
        iscsiLifPerfUUIDToLifFoundationUUID.clear();
        final Map<String, String> networkLifNameToFoundationUUIDMap = context.getNetworkLifNameToFoundationUUIDMap();
        final Map<String, String> networkLifNameToVServerUUIDMap = context.getNetworkLifNameToVServerUUIDMap();

        // Perf Names are formatted VServerName:LIFName
        // Need to pull out the names so we can correlate the UUIDs
        if (context.isLessThanSierraNevada()) {
            // We need to discover the performance UUID for the LIF dynamically, it is not provided by
            // foundation.
            final Map<String, String> perfUUIDToNameMap = getPerfUUIDToNameMap(zapiConnection);
            for (final Map.Entry<String, String> entry : perfUUIDToNameMap.entrySet()) {
                final String lifPerfUUID = entry.getKey();
                final String lifPerfNameRaw = entry.getValue();
                if (lifPerfNameRaw == null) {
                    logger.warn(PackageContext.getLogPrefix() + " - Failed find name for iscsi lif " + lifPerfUUID);
                    continue;
                }
                final String lifPerfName = lifPerfNameRaw.replace('.', ':');

                if (!networkLifNameToFoundationUUIDMap.containsKey(lifPerfName)) {
                    logger.warn(PackageContext.getLogPrefix() + " - Failed to find UUID for iscsi lif " + lifPerfName);
                } else {
                    final String lifFoundationUUID = networkLifNameToFoundationUUIDMap.get(lifPerfName);
                    iscsiLifPerfUUIDToLifFoundationUUID.put(lifPerfUUID, lifFoundationUUID);
                }

                if (!networkLifNameToVServerUUIDMap.containsKey(lifPerfName)) {
                    logger.warn(PackageContext.getLogPrefix() + " - Failed to find vserver for iscsi lif " + lifPerfName);
                } else {
                    final String vServerUUID = networkLifNameToVServerUUIDMap.get(lifPerfName);
                    iscsiLifPerfUUIDToVServerUUID.put(lifPerfUUID, vServerUUID);
                }

            }
        } else {
            for (final Map.Entry<String, String> lifEntry : networkLifNameToFoundationUUIDMap.entrySet()) {
                final String lifName = lifEntry.getKey();
                final String lifPerfUUID = lifEntry.getValue();
                iscsiLifPerfUUIDToLifFoundationUUID.put(lifPerfUUID, lifPerfUUID);
                final String vServerUUID = networkLifNameToVServerUUIDMap.get(lifName);
                iscsiLifPerfUUIDToVServerUUID.put(lifPerfUUID, vServerUUID);
            }
        }
        return new ArrayList<String>(iscsiLifPerfUUIDToLifFoundationUUID.keySet());
    }

    /**
     * Required for 8.1, need to discovery the performance UUIDs dynamically
     * @param zapiConnection connection used to discover performance UUIDs
     * @return map of Lif performance UUID to Lif Name
     */
    private Map<String, String> getPerfUUIDToNameMap(final ZAPIConnection zapiConnection) {
        final Map<String, String> perfUUIDToNameMap = new HashMap<String, String>();
        try {
            perfUUIDToNameMap.putAll(getPerfInstanceUUIDs(targetObjectName, zapiConnection));
        }
        catch (DataSourceErrorException ex) {
            logger.error(PackageContext.getLogPrefix() + " - Failed to discover iscsi lif performance UUIDs: " + ex.getLocalizedMessage(), ex);
        }
        return perfUUIDToNameMap;
    }

    @Override
    VServer.VServerKey getParentInventoryUUIDKey(final NetAppDefaultStorageContext context, final String perfInstanceUUID) {
        final Map<String, String> networkLifNameToVServerUUIDMap = context.getNetworkLifNameToVServerUUIDMap();

        final String vServerUUID = iscsiLifPerfUUIDToVServerUUID.get(perfInstanceUUID);
        if (vServerUUID == null) {
            logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServer for iscsi lif " + perfInstanceUUID);
            return null;
        }
        return new VServer.VServerKey(context.getKey().getOcieUUID(), vServerUUID);
    }

    @Override
    String getIventoryUUIDFromPerfUUID(final NetAppDefaultStorageContext context, final String perfInstanceUUID) {
        return iscsiLifPerfUUIDToLifFoundationUUID.get(perfInstanceUUID);
    }

    @Override
    NetAppISCSILIFMetrics createMetricsFromValues(final MetricsRowScratchPad<NetAppISCSILIFMetrics.ISCSILIFCounter> currentValues, final MetricsRowScratchPad<NetAppISCSILIFMetrics.ISCSILIFCounter> previousValues) {
        return new NetAppISCSILIFMetrics(currentValues, previousValues);
    }

    @Override
    void reportPerformance(Cluster.ClusterKey clusterKey, final Map<VServer.VServerKey, Collection<RawSample<IscsiLifMeasurements>>> measurements) throws DataSourceErrorException {
        for (final Map.Entry<VServer.VServerKey, Collection<RawSample<IscsiLifMeasurements>>> measurementEntry : measurements.entrySet()) {
            final VServer.VServerKey key = measurementEntry.getKey();
            final Collection<RawSample<IscsiLifMeasurements>> iscsiLifMeasurements = measurementEntry.getValue();
            storagePerformanceReporter.reportIscsiLifPerformance(clusterKey, key, iscsiLifMeasurements);
        }
    }

    @Override
    void reportRawPerformance(Cluster.ClusterKey clusterKey, final Map<VServer.VServerKey, Collection<RawSample<IscsiLifRawMeasurements>>> measurements) throws DataSourceErrorException {
        for (final Map.Entry<VServer.VServerKey, Collection<RawSample<IscsiLifRawMeasurements>>> measurementEntry : measurements.entrySet()) {
            final VServer.VServerKey key = measurementEntry.getKey();
            final Collection<RawSample<IscsiLifRawMeasurements>> iscsiLifMeasurements = measurementEntry.getValue();
            storagePerformanceReporter.reportIscsiLifRawPerformance(clusterKey, key, iscsiLifMeasurements);
        }
    }
}
