package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappstorageperformance;

import com.netapp.oci.netapp.performance.networklif.interfaces.data.NetworkLifMeasurements;
import com.netapp.oci.netapp.performance.networklif.interfaces.data.NetworkLifRawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappLifMetrics;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.VServer;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.jdom.Element;

import java.util.*;

public class LifMetricsBuilder extends MetricsBuilderBase<VServer.VServerKey,
                                                          NetworkLifMeasurements,
                                                          NetworkLifRawMeasurements,
                                                          NetappLifMetrics,
                                                          NetappLifMetrics.LifCounter,
                                                          NetappLifMetrics.LifCounter> {

    private static final Logger logger = LoggerFactory.getLogger(LifMetricsBuilder.class);
    private final Map<String, String> lifPerfUUIDToVServerUUID = new HashMap<String, String>();
    private final Map<String, String> lifPerfUUIDToLifFoundationUUID = new TreeMap<String, String>();

    public LifMetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "lif", NetappLifMetrics.LifCounter.class);
    }

    @Override
    List<String> getTargetPerfUUIDs(final NetAppDefaultStorageContext context, final ZAPIConnection zapiConnection) {
        lifPerfUUIDToVServerUUID.clear();
        lifPerfUUIDToLifFoundationUUID.clear();
        // We need to discover the performance UUID for the LIF dynamically, it is not provided by
        // foundation.
        final Map<String, String> perfUUIDToNameMap = getPerfUUIDToNameMap(zapiConnection);
        return new ArrayList<String>(perfUUIDToNameMap.keySet());
    }

    /**
     * Required for 8.1, need to discovery the performance UUIDs dynamically
     * @param zapiConnection connection used to discover performance UUIDs
     * @return map of Lif performance UUID to Lif Name
     */
    private Map<String, String> getPerfUUIDToNameMap(final ZAPIConnection zapiConnection) {
        final Map<String, String> perfUUIDToNameMap = new HashMap<String, String>();
        try {
            perfUUIDToNameMap.putAll(getPerfInstanceUUIDs(targetObjectName, zapiConnection));
        }
        catch (DataSourceErrorException ex) {
            this.clearPreviousValues();
            logger.info(PackageContext.getLogPrefix() + " - Cleared counter values cached from previous poll for " + this.getClass().getName());
            logger.error(PackageContext.getLogPrefix() + " - Failed to discover lif performance UUIDs: " + ex.getLocalizedMessage(), ex);
        }
        return perfUUIDToNameMap;
    }

    @Override
    VServer.VServerKey getParentInventoryUUIDKey(final NetAppDefaultStorageContext context, final String perfInstanceUUID) {
        final String vServerUUID = lifPerfUUIDToVServerUUID.get(perfInstanceUUID);
        if (vServerUUID == null) {
            logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServer for network lif " + perfInstanceUUID);
            return null;
        }
        return new VServer.VServerKey(context.getKey().getID(Cluster.ClusterKey.Field.cluster_uuid), vServerUUID);
    }

    
    @Override
    String getIventoryUUIDFromPerfUUID(final NetAppDefaultStorageContext context, final String perfInstanceUUID) {
        throw new UnsupportedOperationException();
    }
    
    @Override
    /**
     * Lif CM objects does not use standard lif UUIDs. This method correlates inventory and perf UUIDs using the target lif name 
     * and parent vserver name. Returns inventory UUID of the target lif object.
     */
    protected String getIventoryUUIDFromPerfUUID(NetAppDefaultStorageContext context, String perfInstanceUUID, Element instanceDataElt) {
        final Map<String, String> networkLifNameToFoundationUUIDMap = context.getNetworkLifNameToFoundationUUIDMap();
        final Map<String, String> networkLifNameToVServerUUIDMap = context.getNetworkLifNameToVServerUUIDMap();
        // Lif CM objects does not use standard lif uuids, hence MEGA has to workaround by correlating stats and foundation objects using lif names
        // But there could be duplicate lif names across vServers. To overcome this the foundation lif names are formatted VServerName:LIFName
        // Need to pull out the 'vserver_name' perf counter so we can correlate the UUIDs
        String vserverName = null;
        final Element countersElt = instanceDataElt.getChild("counters");
        final String lifName = instanceDataElt.getChildTextTrim("name");
        if (countersElt == null) {
            logger.warn(PackageContext.getLogPrefix() + " - perf-object-get-instance " + lifName
                    + " failed to retrieve any counters");
            return null;
        }
        @SuppressWarnings({"unchecked"})
        final List<Element> counterDataElts = countersElt.getChildren("counter-data");
        for (final Element counterDataElt : counterDataElts) {
            final String counterName = counterDataElt.getChildTextTrim("name");
            if("vserver_name".equals(counterName)){
                final String counterValue = counterDataElt.getChildTextTrim("value");
                if (counterValue != null) {
                    vserverName = counterValue;
                    break;
                }
            }
        }
        
        String lifFoundationUUID = null;
        String vserverUUID = null;
        // Perf Names in older ontap versions (SN) are already formatted VServerName:LIFName
        if(context.isLessThanFullSteam() && lifName != null){ 
            if(networkLifNameToFoundationUUIDMap.containsKey(lifName)){ 
                lifFoundationUUID = networkLifNameToFoundationUUIDMap.get(lifName);
            }
            if(networkLifNameToVServerUUIDMap.containsKey(lifName)){
                vserverUUID = networkLifNameToVServerUUIDMap.get(lifName);
            }
        }else if(vserverName != null && lifName != null){
            // Perf names in fullsteam are NOT formatted VServerName:LIFName. We need to build this string here.
            final StringBuilder perfLifName = new StringBuilder();
            perfLifName.append(vserverName);
            perfLifName.append(':');
            perfLifName.append(lifName);
            if(networkLifNameToFoundationUUIDMap.containsKey(perfLifName.toString())){
                lifFoundationUUID = networkLifNameToFoundationUUIDMap.get(perfLifName.toString());
            }
            if(networkLifNameToVServerUUIDMap.containsKey(perfLifName.toString())){
                vserverUUID = networkLifNameToVServerUUIDMap.get(perfLifName.toString());
            }
        }
        
        if(lifFoundationUUID == null){
            logger.warn("{} - Failed to find UUID for network lif {}", new Object[] {PackageContext.getLogPrefix(), lifName});
        }else{
            lifPerfUUIDToLifFoundationUUID.put(perfInstanceUUID, lifFoundationUUID);
        }
        
        if(vserverUUID == null){
            logger.warn("{} - Failed to find vserver for network lif {}", new Object[] {PackageContext.getLogPrefix(), lifName});
        }else{
            lifPerfUUIDToVServerUUID.put(perfInstanceUUID, vserverUUID);
        }

        return lifFoundationUUID;
    }

    @Override
    NetappLifMetrics createMetricsFromValues(final MetricsRowScratchPad<NetappLifMetrics.LifCounter> currentValues, final MetricsRowScratchPad<NetappLifMetrics.LifCounter> previousValues) {
        return new NetappLifMetrics(currentValues, previousValues);
    }

    @Override
    void reportPerformance(Cluster.ClusterKey clusterKey, final Map<VServer.VServerKey, Collection<RawSample<NetworkLifMeasurements>>> measurements) throws DataSourceErrorException {
        for (final Map.Entry<VServer.VServerKey, Collection<RawSample<NetworkLifMeasurements>>> measurementEntry : measurements.entrySet()) {
            final VServer.VServerKey key = measurementEntry.getKey();
            final Collection<RawSample<NetworkLifMeasurements>> lifMeasurements = measurementEntry.getValue();
            storagePerformanceReporter.reportLifPerformance(clusterKey, key, lifMeasurements);
        }
    }
    
    @Override 
    void reportRawPerformance(Cluster.ClusterKey clusterKey, final Map<VServer.VServerKey, Collection<RawSample<NetworkLifRawMeasurements>>> rawMeasurements) throws DataSourceErrorException { 
        for (final Map.Entry<VServer.VServerKey, Collection<RawSample<NetworkLifRawMeasurements>>> measurementEntry : rawMeasurements.entrySet()) { 
            final VServer.VServerKey key = measurementEntry.getKey(); 
            final Collection<RawSample<NetworkLifRawMeasurements>> lifMeasurements = measurementEntry.getValue(); 
            storagePerformanceReporter.reportLifRawPerformance(clusterKey, key, lifMeasurements); 
        } 
    } 
    
}
