package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappstorageperformance;

import com.netapp.oci.netapp.performance.qosservicecenter.interfaces.data.QosServiceCenterMeasurements;
import com.netapp.oci.netapp.performance.qosworkload.interfaces.data.QosVolumeWorkloadMeasurements;
import com.netapp.oci.netapp.performance.qosworkloaddetail.interfaces.data.QosWorkloadDetailMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation.BuilderUtil;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation.PerfInstanceBuilder;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.EnumWrapper;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceOutputFileType;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.framework.datasource.config.PackageConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileWriter;
import com.onaro.sanscreen.acquisition.framework.recordings.RecordingSession;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppStoragePerformanceReporter;
import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.util.common.ParseUtils;
import org.apache.log4j.Logger;
import org.jdom.Element;

import java.io.File;
import java.io.IOException;
import java.math.BigInteger;
import java.util.*;

/**
 * Base class to get, transform & normalize, and report to server sets of performance metrics (defined in derived classes)
 *
 * @param <K> the foundation UUID parent Key to identify performance measurements
 * @param <S> the normalized performance Sample to report to the server
 * @param <R> the raw performance Sample to report to the server
 * @param <M> the Metrics type that maps counter names to normalization recipes, and contains MetricsRowScratchPad for recording
 *            initial, intermediate, and final reported metric values for current and previous polls
 * @param <Z> An Enumeration, defined by MetricsRowSample derived classes, that identifies the set of counter/column names requested from ZAPI
 * @param <O> An Enumeration, defined by MetricsRowSample derived classes, that identifies the set of counter/column names sent to OCIE
 */
public abstract class MetricsBuilderBase<K extends Key, S, R, M extends MetricsRowSample<S, R, Z, O>, Z extends Enum<Z>, O extends Enum<O>> {

    private static final Logger logger = Logger.getLogger(MetricsBuilderBase.class);

    // The String key is the performance object UUID.
    // Needs to be translated to foundation UUID for reporting to server
    private final Map<String, MetricsRowScratchPad<Z>> perfUUIDToPreviousValues;

    final String logPrefix;
    final boolean metricsDebugEnabled;
    final RecordingSession recordingSession;
    final PackageConfiguration configuration;

    protected final NetAppOCIEPerformancePackage netAppOCIEPerformancePackage;
    protected final NetAppStoragePerformanceReporter storagePerformanceReporter;

    protected String[] filteredCounterNames = null;

    /**
     * Token identifying specific enumeration class for access to class instance methods
     */
    final Class<Z> metricEnumClass;

    /**
     * Target zapi object name to query
     */
    protected final String targetObjectName;

    /**
     * CSV file to assist in with metrics debugging
     */
    private DSFileWriter metricsDebugWriter = null;

    /**
     * Create a new Metrics builder instance
     *
     * @param reportingPackage the performance package
     * @param targetObjectName the ZAPI target object name (disk, volume, lun, lif...)
     * @param metricEnumClass  the enum defining target counter names
     */
    MetricsBuilderBase(final NetAppOCIEPerformancePackage reportingPackage, final String targetObjectName, final Class<Z> metricEnumClass) {
        this.perfUUIDToPreviousValues = new TreeMap<>();
        this.logPrefix = reportingPackage.getLogPrefix();
        this.metricsDebugEnabled = reportingPackage.isMetricsDebugEnabled();
        this.recordingSession = reportingPackage.getRecordingSession();
        this.configuration = reportingPackage.getConfiguration();
        this.targetObjectName = targetObjectName;
        this.metricEnumClass = metricEnumClass;
        this.storagePerformanceReporter = reportingPackage;
        this.netAppOCIEPerformancePackage = reportingPackage;
    }

    /**
     * Create a new Metrics builder instance, without a target object name. The target object will be retrieved by
     * calling getObjectName.
     *
     * @param reportingPackage the performance package
     * @param metricEnumClass  the enum defining target counter names
     */
    MetricsBuilderBase(NetAppOCIEPerformancePackage reportingPackage, final Class<Z> metricEnumClass) {
        this(reportingPackage, "", metricEnumClass);
    }

    public String getLogPrefix() {
        return logPrefix;
    }

    public RecordingSession getRecordingSession() {
        return recordingSession;
    }

    public PackageConfiguration getConfiguration() {
        return configuration;
    }

    public NetAppOCIEPerformancePackage getNetAppOCIEPerformancePackage() {
        return netAppOCIEPerformancePackage;
    }
    
    protected String getObjectName(final NetAppDefaultStorageContext context){
        return this.targetObjectName;
    }

    /**
     * Get the List of Lists of target performance UUIDs. perf-object-get-instances takes a list of
     * up to 100 target performance UUIDs.  Inner lists should be no larger than 100 elements.
     *
     * @param context        the foundation context used to determine the performance UUID keys
     * @param zapiConnection the connection to the ZAPI server
     *
     * @return the List of Lists of target performance UUIDs
     */
    abstract List<String> getTargetPerfUUIDs(final NetAppDefaultStorageContext context, final ZAPIConnection zapiConnection);

    /**
     * Get the names of the counters to acquire for the target instances
     *
     * @return the names of the counters to acquire for the target instances
     *         TODO: another version that takes a String parameter representing the Cluster version.
     *         Based on the version, return a filtered set of counter names known to be supported
     *         by the version.
     */
    private String[] getTargetCounterNames() {
        return (filteredCounterNames != null) ? filteredCounterNames : getAllCounterNames();
    }

    /**
     * Get all counter enumeration names.   This set is expected to be filtered based on cluster version
     * to return a set of counters supported by the version
     *
     * @return all counter enumeration names
     */
    String[] getAllCounterNames() {
        final Z[] enumVals = metricEnumClass.getEnumConstants();
        final String[] counterNames = new String[enumVals.length];
        for (final Z z : enumVals) {
            counterNames[z.ordinal()] = z.name();
        }
        return counterNames;
    }

    /**
     * Get the parent inventory/foundation UUID key to use to report the samples under
     *
     * @param context          the foundation context used to determine the parent foundation UUID key
     * @param perfInstanceUUID the performance instance UUID to find the parent for
     *
     * @return the parent inventory/foundation UUID key to use to report the samples under
     */
    abstract K getParentInventoryUUIDKey(final NetAppDefaultStorageContext context, final String perfInstanceUUID);

    /**
     * Get the inventory/foundation UUID key for the givent performance UUID key
     *
     * @param context          the foundation context used to determine the foundation UUID key
     * @param perfInstanceUUID the performance UUID to find the foundation UUID for
     *
     * @return the inventory/foundation UUID key for the givent performance UUID key
     */
    abstract String getIventoryUUIDFromPerfUUID(final NetAppDefaultStorageContext context, final String perfInstanceUUID);

    /**
     * Wrap the given time stamp and scratchpad values in the appropriate Metrics class that provides
     * tranformation/normalization recipes
     *
     * @param currentValues  current poll's scratchpad values to encapsulate
     * @param previousValues previous poll's scratchpad values (if available) for comparison
     *
     * @return the given timestamp and raw values wrapped in the appropriate Measurements class
     */
    abstract M createMetricsFromValues(final MetricsRowScratchPad<Z> currentValues, final MetricsRowScratchPad<Z> previousValues);

    /**
     * Normalize the current sample by comparing to the previous sample and converting counters to rates, etc
     *
     * @param currentMetrics the current set of metrics
     *
     * @return the normalized raw sample
     */
    final RawSample<S> getNormalizedSample(final M currentMetrics, final NetAppDefaultStorageContext context) {
        final String inventoryUUID = currentMetrics.getCurrentValues().getInventoryUUID();
        final S measurements = currentMetrics.normalizeMetrics(context);
        if (currentMetrics.isZeroSuppressDetected()) {
            // Make no report at all if zero suppression has been detected.
            return null;
        }
        else if (currentMetrics.isWrapResetDetected()) {
            // detected Wrap/Reset, contract with server requires reporting the UUID with a null payload.

            //OPM requirement: avoid reporting gaps in QoS performance tables. reporting 0 instead of null (DE5069)
            //ONTAP indicates that when resets are observed in Qos counters, we may assume that the reason is due to
            //memory reclamation ("aged out") which frees the memory of counters that don't change.
            //That is why we can interpret the reset as "no change" and report rates of 0
            if(measurements instanceof QosServiceCenterMeasurements){
                return (RawSample<S>)new RawSample<QosServiceCenterMeasurements>(inventoryUUID, new QosServiceCenterMeasurements(0L, 0.0f, 0L, 0L, currentMetrics.getCurrentValues().getCollectionPeriod()));
            }
            else if(measurements instanceof QosVolumeWorkloadMeasurements)  {
                return (RawSample<S>)new RawSample<QosVolumeWorkloadMeasurements>(inventoryUUID, new QosVolumeWorkloadMeasurements());
            }
            else if(measurements instanceof QosWorkloadDetailMeasurements) {
                return (RawSample<S>)new RawSample<QosWorkloadDetailMeasurements>(inventoryUUID, new QosWorkloadDetailMeasurements(0L, 0L, currentMetrics.getCurrentValues().getCollectionPeriod()));
            }
            else {
                return new RawSample<S>(inventoryUUID, null);
            }
        }
        else {
            return new RawSample<S>(inventoryUUID, measurements);
        }
    }
    

    /**
     * Get the raw counters out of the current metric.
     * @param currentMetrics
     * @param context
     * @return
     */
    final RawSample<R> getRawSample(final M currentMetrics, final NetAppDefaultStorageContext context) {
        final String inventoryUUID = currentMetrics.getCurrentValues().getInventoryUUID();
        final R measurements = currentMetrics.rawMetrics(context);
        return new RawSample<R>(inventoryUUID, measurements);
    }

    /**
     * Report the given samples to the server
     *
     * @param clusterKey the clusterKey measurements belongs to
     * @param measurements the samples to report
     *
     * @throws DataSourceErrorException when there is a problem reporting samples to the server
     */
    abstract void reportPerformance(Cluster.ClusterKey clusterKey, final Map<K, Collection<RawSample<S>>> measurements) throws DataSourceErrorException;
    
    /**
     * Report the raw counters to the the server
     * 
     * @param clusterKey the clusterKey measurements belongs to
     * @param rawMeasurements
     * @throws DataSourceErrorException
     */
    void reportRawPerformance(Cluster.ClusterKey clusterKey, final Map<K, Collection<RawSample<R>>> rawMeasurements) throws DataSourceErrorException {}
    
    /**
     * Get and report normalized performance measurements
     *
     * @param context        the foundation context used to determine the performance UUID keys
     * @param zapiConnection the connection to the ZAPI server
     *
     * @throws DataSourceErrorException when there is a problem acquiring or reporting samples to the server
     */
    public void getAndReportPerformance(final NetAppDefaultStorageContext context, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        final Map<K, Collection<RawSample<S>>> metricsToReport = getMetricsAndReportRawPerformance(context, zapiConnection);
        reportPerformance(context.getKey(), metricsToReport);
    }

    protected Map<K, Collection<RawSample<S>>> getMetricsAndReportRawPerformance(final NetAppDefaultStorageContext context, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        final Map<K, Collection<RawSample<S>>> metricsToReport = new HashMap<K, Collection<RawSample<S>>>();
        final Map<K, Collection<RawSample<R>>> rawMetricsToReport = new HashMap<K, Collection<RawSample<R>>>();
        boolean cookedPersistenceEnabled = context.isCookedPersistenceEnabled();
        boolean rawPersistenceEnabled = context.isRawPersistenceEnabled();

        if (cookedPersistenceEnabled || rawPersistenceEnabled) {
            final Map<Long, Map<String, M>> instanceValues = getCounterValues(context, zapiConnection);

            // Clear out all previous entries, the new entries will become the previous
            perfUUIDToPreviousValues.clear();
            for (final Map.Entry<Long, Map<String, M>> iterResponseEntry : instanceValues.entrySet()) {
                final Map<String, M> instanceValueEntryMap = iterResponseEntry.getValue();
                for (final Map.Entry<String, M> instanceValueEntry : instanceValueEntryMap.entrySet()) {
                    final String performanceUUID = instanceValueEntry.getKey();
                    final M currentMetrics = instanceValueEntry.getValue();
                    final MetricsRowScratchPad<Z> currentValues = currentMetrics.getCurrentValues();
                    final MetricsRowScratchPad<Z> previousValues = currentMetrics.getPreviousValues();

                    // The current values become the previous values for the next poll
                    this.perfUUIDToPreviousValues.put(performanceUUID, currentValues);

                    // Can't report values without a parent
                    final K parentKey = getParentInventoryUUIDKey(context, performanceUUID);
                    if (parentKey == null) {
                        logger.warn(PackageContext.getLogPrefix() + " - Failed to find parent for performance target " + performanceUUID);
                        continue;
                    }

                    if (cookedPersistenceEnabled) {
                        if (!metricsToReport.containsKey(parentKey)) {
                            // First instance for this parent
                            metricsToReport.put(parentKey, new LinkedList<RawSample<S>>());
                        }

                        if (previousValues == null) {
                            // logger.info(PackageContext.getLogPrefix() + " - No previous sample for " + getObjectName(context) + " " + performanceUUID);
                            writeMetricsDebugCSV(context, currentValues);
                            continue;           // first poll, nothing to report
                        }

                        // Compare current values to previous
                        final RawSample<S> metricsPayloadForReport = getNormalizedSample(currentMetrics, context);

                        // Now we finally have the final value for debug CSV
                        writeMetricsDebugCSV(context, currentValues);
                        final Collection<RawSample<S>> samplesForReport = metricsToReport.get(parentKey);
                        if (metricsPayloadForReport != null) {
                            samplesForReport.add(metricsPayloadForReport);
                        }
                    }

                    if (rawPersistenceEnabled) {
                        if(!rawMetricsToReport.containsKey(parentKey)) {
                            rawMetricsToReport.put(parentKey, new LinkedList<RawSample<R>>());
                        }

                        final Collection<RawSample<R>> countersForReport = rawMetricsToReport.get(parentKey);
                        if (countersForReport != null) {
                            countersForReport.add(getRawSample(currentMetrics, context));
                        }
                    }
                }
            }

            if (isMetricsDebugEnabled()) {
                closeMetricDebug();
            }

            // Report raw metrics
            if (rawPersistenceEnabled) {
                reportRawPerformance(context.getKey(), rawMetricsToReport);
            }
        }
        //Add to global scratchpad
        this.netAppOCIEPerformancePackage.getGlobalScratchPad().put(getObjectName(context), this.getPerfUUIDToPreviousValues());
        return metricsToReport;
    }

    /**
     * Invoke Zapi perf-object-get-instances to get the desired counters for the desired targets
     *
     * @param context        the foundation context used to determine the performance UUID keys
     * @param zapiConnection the connection to the ZAPI server
     *
     * @return A Map of performance UUID keys to array of raw long values for each UUID, indexed by timestamp.
     *
     * @throws DataSourceErrorException when there is a problem acquiring performance measurements
     */
    protected Map<Long, Map<String, M>> getCounterValues(final NetAppDefaultStorageContext context, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        final String[] targetCounterNames = getTargetCounterNames();  // TODO get counter targets based on ONTAP version
        final List<String> targetPerfUUIDs = getTargetPerfUUIDs(context, zapiConnection);
        final List<List<String>> chunkedTargetPerfUUIDs = getListOfTargetIDLists(context, targetPerfUUIDs);
        if (targetPerfUUIDs == null || targetPerfUUIDs.isEmpty()) {
            // probably sub class failed to implement correctly or at all
            if (logger.isDebugEnabled()) {
                logger.info(PackageContext.getLogPrefix() + " - " + getObjectName(context) + " 0 performance targets");
            }
            return Collections.emptyMap();
        }
        final int maxChunkSize = chunkedTargetPerfUUIDs.get(0).size();  // max size is size of first chunk
        final Map<Long, Map<String, M>> returnMetrics = new TreeMap<Long, Map<String, M>>();
        final long pollStartTime = System.currentTimeMillis();  // Need to keep track of local time to calculate accurate time deltas
        final long serverPollStartTime = netAppOCIEPerformancePackage.getSampleTime();
        int iterCount = 0;
        final List<Integer> iterChunckCounts = new ArrayList<Integer>(100);  // size hint - 100 chunks total, NOT size of each chunk
        long totalZapiTime = 0;
        int totalZapiInvocations = 0;
        int totalElements = 0;

        // The set of instances we want, chunked into groups
        for (final List<String> targetIDChunk : chunkedTargetPerfUUIDs) {
            final int requestedIDCount = targetIDChunk.size();
            if (requestedIDCount == 0) {
                break;
            }

            // Map of instance name to counter value for each counter name received at a specific time (single iterator response)
            final Map<String, M> uuidToMetrics = new LinkedHashMap<String, M>();
            final long zapiStartTime = System.currentTimeMillis();
            final ZAPIConnection.ZapiResponse zapiResponse = getMetricsResponse(targetIDChunk, targetCounterNames, getObjectName(context), zapiConnection);
            returnMetrics.put(zapiResponse.getReceivedTimeStamp(), uuidToMetrics);
            iterCount++;
            totalZapiTime += zapiResponse.getTotalTime();
            totalZapiInvocations += zapiResponse.getInvocationCount();

            final int numPerfInstances = processPerfInstancesElt(zapiResponse.getElement(), context, uuidToMetrics,
                                                                 serverPollStartTime, pollStartTime, zapiStartTime,
                                                                 zapiResponse.getReceivedTimeStamp());
            iterChunckCounts.add(numPerfInstances);
            totalElements += numPerfInstances;
        }
        final long consumedTime = System.currentTimeMillis() - pollStartTime;
        final String zapiCommand = "perf-object-get-instances " + getObjectName(context);
        if (logger.isDebugEnabled()) {
            logger.debug(PackageContext.getLogPrefix() + " - " + zapiCommand + ": max records: " + maxChunkSize
                         + ", iterations: " + iterCount + ", total clock time: " + consumedTime + " ms, total zapi time: "
                         + totalZapiTime + " ms, total zapi invocations " + totalZapiInvocations + ", total elements: "
                         + totalElements + ", buckets: " + getCSVStringFromIntList(iterChunckCounts));
        }
        try {
            final DSFileWriter responseTimesWriter = zapiConnection.getResponseTimesWriter();
            responseTimesWriter.writeln(PackageContext.getPackage().getConfiguration().getDataSourceName() + ", "
                                        + zapiCommand + ", " + new Date(pollStartTime) + ", " + maxChunkSize + ", " + iterCount + ", "
                                        + consumedTime + ", " + totalZapiTime + ", " + totalZapiInvocations + ", " + totalElements + ", "
                                        + getCSVStringFromIntList(iterChunckCounts));
        }
        catch (IOException ioe) {
            logger.warn(PackageContext.getLogPrefix() + " - Failed to record ZAPI response time for: " + zapiCommand);
        }

        return returnMetrics;
    }

    /**
     * Process the specified ZAPI perf response.
     *
     * @param perfInstancesElt       a single XML element of a ZAPI performance response
     * @param context                foundation/inventory data for UUID lookups
     * @param uuidToMetrics          map of UUID to MetricsRowSample for duplicate detection
     * @param serverPollStartTime    cached server time of poll start
     * @param pollStartTime          local acq time of server start
     * @param zapiStartTime          local acq time of ZAPI start
     * @param zapiResponseTime       local time of zapi response
     *
     * @return The number of perf instances processed.
     */
    @SuppressWarnings("unchecked")
    int processPerfInstancesElt(final Element perfInstancesElt, final NetAppDefaultStorageContext context, final Map<String, M> uuidToMetrics,
                                final long serverPollStartTime, final long pollStartTime, final long zapiStartTime, final long zapiResponseTime) {
               
        if(perfInstancesElt == null || perfInstancesElt.getChild("instances") == null){
            return 0;
        }
        int totalElements = 0;
        final List<Element> instanceDataElts = perfInstancesElt.getChild("instances").getChildren("instance-data");
        for (final Element instanceDataElt : instanceDataElts) {
            final String performanceUUID = instanceDataElt.getChildTextTrim("uuid");
            final String inventoryUUID = getIventoryUUIDFromPerfUUID(context, performanceUUID, instanceDataElt);
            if(inventoryUUID == null){
                logger.warn(PackageContext.getLogPrefix() + " - Unable to correlate inventory and performance("+ performanceUUID +") uuids for an instance of object :" + instanceDataElt.getChildTextTrim("name"));
                continue;
            }
            final MetricsRowScratchPad<Z> previousValues = this.perfUUIDToPreviousValues.get(performanceUUID);

            // Get the values out of the response and into a scratch pad
            final MetricsRowScratchPad<Z> currentValues = getMetricsScratchPad(metricEnumClass, performanceUUID, inventoryUUID, serverPollStartTime,
                    pollStartTime, zapiStartTime, zapiResponseTime, previousValues);
            final Element countersElt = instanceDataElt.getChild("counters");
            if (countersElt == null) {
                final String instanceName = instanceDataElt.getChildTextTrim("name");
                logger.warn(PackageContext.getLogPrefix() + " - perf-object-get-instance " + instanceName + " failed to retrieve any counters");
                continue;
            }

            @SuppressWarnings({"unchecked"})
            final List<Element> counterDataElts = countersElt.getChildren("counter-data");
            for (final Element counterDataElt : counterDataElts) {
                final String counterNameDebug = counterDataElt.getChildTextTrim("name");
                final EnumWrapper<Z> enumWrapper = BuilderUtil.getEnumWrapperByValue(metricEnumClass, counterNameDebug, null);
                final Z counterName = enumWrapper.getEnumValue();
                if (counterName == null) {
                    if (logger.isDebugEnabled()) {
                        logger.debug(PackageContext.getLogPrefix() + " - perf-object-get-instance no defined enum for " + getObjectName(context) + " "
                                     + metricEnumClass.getSimpleName() + " " + counterNameDebug + " " + enumWrapper.getRawValue());
                    }
                    continue;
                }
                final String rawValue = counterDataElt.getChildTextTrim("value");
                BigInteger counterValue = MetricsRowSample.NULL_BIG_INT;
                if (rawValue != null) {
                    // Lots of values are 0, it is worth optimizing
                    if ("0".equals(rawValue)) {
                        counterValue = BigInteger.ZERO;
                    } else {
                        try {
                            counterValue = new BigInteger(rawValue);
                        }
                        catch (NumberFormatException nfe) {
                            // Ignore, store raw value in the raw values map
                            currentValues.getRawValues().put(counterName, rawValue);
                        }
                    }
                }
                currentValues.getCounterValues().put(counterName, counterValue);
            }

            // Workaround IBG-1145/Burt ID 673719 - perf-object-get-instances returns duplicate instances
            // If two sets of counters are returned for the same UUID, combine them into a single set of
            // counters by adding counters together.
            if (uuidToMetrics.containsKey(performanceUUID)) {
                logger.warn(PackageContext.getLogPrefix() + " - Found duplicate instance-data element for "
                            + getObjectName(context) + " UUID " + performanceUUID);
                // Add values from both instances together
                final MetricsRowScratchPad<Z> existingValues = uuidToMetrics.get(performanceUUID).getCurrentValues();
                for (final Z key : metricEnumClass.getEnumConstants()) {
                    final BigInteger counterValue1 = existingValues.getCounterValues().get(key);
                    final BigInteger counterValue2 = currentValues.getCounterValues().get(key);
                    final BigInteger combinedValue;
                    if (MetricsRowScratchPad.isNull(counterValue1)) {
                        combinedValue = counterValue2;
                    }
                    else if (MetricsRowScratchPad.isNull(counterValue2)) {
                        combinedValue = counterValue1;
                    }
                    else {
                        combinedValue = counterValue1.add(counterValue2);
                    }
                    if (combinedValue != null) {
                        currentValues.getCounterValues().put(key, combinedValue);
                    }
                }
            }

            // Now package the scratch pad with the Metrics recipes used to normalize them in preparation for report.
            final M currentMetrics = createMetricsFromValues(currentValues, previousValues);
            final M oldMetrics = uuidToMetrics.put(performanceUUID, currentMetrics);
            // Don't count duplicates twice!
            if (oldMetrics == null) {
                totalElements++;
            }
        }
        return totalElements;
    }

    protected String getIventoryUUIDFromPerfUUID(NetAppDefaultStorageContext context, String performanceUUID, Element instanceDataElt) {
        return getIventoryUUIDFromPerfUUID(context, performanceUUID);
    }

    private ZAPIConnection.ZapiResponse getMetricsResponse(final List<String> targetIDs, final String[] targetCounterNames, final String objectName, final ZAPIConnection zapiConnection) throws DataSourceErrorException {

        if (targetIDs.size() == 0) {
            return null;
        }
        final Map<String, String> args = new LinkedHashMap<String, String>();
        args.put("objectname", objectName);

        // The set of counters we want
        final ZAPIConnection.ZapiParamArray zapiCounterNameArray = new ZAPIConnection.ZapiParamArray("counters", "counter", Arrays.asList(targetCounterNames));

        final ZAPIConnection.ZapiParamArray[] zapiParamArrays = {new ZAPIConnection.ZapiParamArray("instance-uuids", "instance-uuid", targetIDs),    // instance names
                                                                 zapiCounterNameArray,  // counter names
        };

        final ZAPIConnection.ZapiResponse zapiResponse = zapiConnection.invokeCommandArrayArgs(
                "perf-object-get-instances", args, Arrays.asList(zapiParamArrays));
        final Element perfObjectInstanceListInfoElt = zapiResponse.getElement();
        if (perfObjectInstanceListInfoElt != null && perfObjectInstanceListInfoElt.getChild("instances") == null) {
            String reason = perfObjectInstanceListInfoElt.getAttributeValue("reason");
            if (reason != null && reason.contains("non-existent counters are not allowed")) {
                // Incorrect counter list: "avg_other_latency,avg_read_latency,avg_write_latency,total_ops", non-existent counters are not allowed.
                final int quoteStart = reason.indexOf('"');
                if (quoteStart < 0) {
                    return zapiResponse;
                }
                final int quoteEnd = reason.indexOf('"', quoteStart + 1);
                if (quoteEnd < 0) {
                    return zapiResponse;
                }
                final String badCountersList = reason.substring(quoteStart + 1, quoteEnd);
                logger.info(PackageContext.getLogPrefix() + " Retry without unsupported counters: " + objectName + " " + badCountersList);
                final String[] badCounters = badCountersList.split(",");
                final Set<String> badCounterSet = new HashSet<String>();
                badCounterSet.addAll(Arrays.asList(badCounters));
                final String[] filteredCounters = new String[targetCounterNames.length - badCounters.length];
                int i = 0;
                for (final String candidateCounterName : targetCounterNames) {
                    if (badCounterSet.contains(candidateCounterName)) {
                        continue;
                    }
                    filteredCounters[i++] = candidateCounterName;
                }
                filteredCounterNames = filteredCounters;
                return getMetricsResponse(targetIDs, filteredCounters, objectName, zapiConnection);
            }
        }
        return zapiResponse;
    }

    /**
     * Convert a list of integers to a single string of comma separated ints for logging/debugging.
     *
     * @param list list of integers to convert to single string of comma separated ints
     *
     * @return a single string of comma separated ints
     */
    protected static String getCSVStringFromIntList(final List<Integer> list) {
        return ParseUtils.toStringList(list);
    }

    /**
     * Discover the set of performance UUIDs for the given target object name
     *
     * @param objectName     the target object name to discover instance UUIDS for
     * @param zapiConnection connection to cluster zapi server used for discovery
     *
     * @return the set of performance UUIDs for the given target object name
     *
     * @throws DataSourceErrorException when an acquisition error occurs
     */
    static Map<String, String> getPerfInstanceUUIDs(final String objectName, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        final PerfInstanceBuilder perfInstanceBuilder = new PerfInstanceBuilder(objectName);
        perfInstanceBuilder.buildModel(null, zapiConnection);
        return perfInstanceBuilder.getUUIDNameMap();
    }

    public int getMaxChunkSize(final NetAppDefaultStorageContext context) {
        return (context.isLessThanSierraNevada())
               ? 100
               : 500;
    }

    public List<List<String>> getListOfTargetIDLists(final NetAppDefaultStorageContext context, final Collection<String> idList) {
        return getListOfTargetIDListsLengthN(idList, getMaxChunkSize(context));
    }

    public static List<List<String>> getListOfTargetIDListsLengthN(final Collection<String> idList, final int n) {
        final Set<String> allUUIDs = new HashSet<String>();
        final int idCount = idList.size();
        final int idListsCount = (idCount / n) + 1;
        final List<List<String>> idLists = new ArrayList<List<String>>(idListsCount);
        List<String> listOfLengthN = new ArrayList<String>(n);
        idLists.add(listOfLengthN);
        if (idCount > 0) {
            for (final String id : idList) {
                // Previous chunk of length n is full, create a new chunck of length n
                if (listOfLengthN.size() == n) {
                    listOfLengthN = new ArrayList<String>(n);
                    idLists.add(listOfLengthN);
                }
                if (allUUIDs.contains(id)) {
                    logger.warn(PackageContext.getLogPrefix() + " - Skipping duplicate perf UUID: " + id);
                }
                else {
                    allUUIDs.add(id);
                    listOfLengthN.add(id);
                }
            }
        }
        return idLists;
    }

    private boolean isMetricsDebugEnabled() {
        return metricsDebugEnabled;
    }

    private DSFileWriter getDSMetricsDebugWriter(final NetAppDefaultStorageContext context, final MetricsRowScratchPad<Z> currentValues) {
        if (metricsDebugWriter != null) {
            return metricsDebugWriter;              // Already initialized
        }
        final String metricsName = metricEnumClass.getSimpleName();
        try {
            final com.onaro.sanscreen.acquisition.framework.datasource.Package perfPackage = PackageContext.getPackage();
            final PackageConfiguration perfConfig = perfPackage.getConfiguration();
            metricsDebugWriter = perfPackage.getRecordingSession().getCurrentSample().createTextualOutputFile(perfConfig, DataSourceOutputFileType.perfdata, metricsName, "csv");
            final File tempFile = new File(metricsDebugWriter.getFileName());
            PackageContext.getPackage().getRecordingSession().getCurrentSample().addFileDescription(tempFile.getName(), metricsName, "");
            final String columnHeaders = currentValues.getDebugColumnHeaders(getObjectName(context));
            metricsDebugWriter.write(columnHeaders);
            metricsDebugWriter.flush();
            return metricsDebugWriter;
        }
        catch (IOException e) {
            logger.error(PackageContext.getLogPrefix() + " Failed to create a debug file for performance data", e);
            return null;
        }
    }

    private void writeMetricsDebugCSV(final NetAppDefaultStorageContext context, final MetricsRowScratchPad<Z> currentValues) {
        if (!isMetricsDebugEnabled()) {
            return;
        }
        final String metricsDebugCSV = currentValues.toDebugCSV();
        try {
            final DSFileWriter metricsDebugWriter = getDSMetricsDebugWriter(context, currentValues);
            metricsDebugWriter.write(metricsDebugCSV);
            metricsDebugWriter.flush();
        }
        catch (IOException ioe) {
            logger.warn(PackageContext.getLogPrefix() + " - Failed to debug metrics: " + ioe, ioe);
        }
    }

    private void closeMetricDebug() {
        if (!isMetricsDebugEnabled()) {
            return;
        }
        if (metricsDebugWriter == null) {
            return;
        }
        try {
            metricsDebugWriter.close();
        }
        catch (IOException ignore) {
        }
        metricsDebugWriter = null;
    }

    protected MetricsRowScratchPad<Z> getMetricsScratchPad(final Class<Z> enumClass, final String perfInstanceUUID, final String inventoryUUID,
                                                           final long serverPollStartTimeStamp, final long pollStartTimeStamp, final long zapiStartTimeStamp,
                                                           final long zapiResponseTimestamp, final MetricsRowScratchPad<Z> previousValues) {

        return new MetricsRowScratchPad<Z>(enumClass, perfInstanceUUID, inventoryUUID, serverPollStartTimeStamp,
                                           pollStartTimeStamp, previousValues, isMetricsDebugEnabled());
    }

    Map<String, MetricsRowScratchPad<Z>> getPerfUUIDToPreviousValues() {
        return Collections.unmodifiableMap(perfUUIDToPreviousValues);
    }
    
    public void clearPreviousValues(){
        perfUUIDToPreviousValues.clear();
    }
}
