package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappstorageperformance;


import com.netapp.oci.netapp.performance.qosservicecenter.interfaces.data.QosServiceCenterMeasurements;
import com.netapp.oci.netapp.performance.qosworkloaddetail.interfaces.data.QosWorkloadDetailMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappQosServiceCenterMetrics;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappQosWorkloadDetailMetrics;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

import org.apache.log4j.Logger;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.LinkedList;
import java.util.EnumMap;


public class QosServiceCenterMetricsBuilder extends MetricsBuilderBase<Cluster.ClusterKey,
                                                                       QosServiceCenterMeasurements,
                                                                       QosServiceCenterMeasurements,
                                                                       NetappQosServiceCenterMetrics,
                                                                       NetappQosServiceCenterMetrics.ServiceCenterCounter,
                                                                       NetappQosServiceCenterMetrics.OCIEServiceCenterCounter> {

    private static final Logger logger = Logger.getLogger(QosServiceCenterMetricsBuilder.class);
    /**
     * WARNING: Do not call getAndReportPerformance on qosWorkloadDetailMetricsBuilder, will result in Infinite recursion.
     */
    private final QosWorkloadDetailMetricsBuilder qosWorkloadDetailMetricsBuilder;


    public QosServiceCenterMetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage,
                                          final QosWorkloadDetailMetricsBuilder qosWorkloadDetailMetricsBuilder) {

        super(reportingPackage, "resource", NetappQosServiceCenterMetrics.ServiceCenterCounter.class);

        // Metrics for Non-Delay Service Centers are collected directly from ZAPI using target object-name "resource"
        // Metrics for Delay Service Centers are summations of child Workload Details metrics.
        // This is a new use case, going forward assert:
        //  - if performance collection becomes multi-threaded, workload details collection must complete before
        //    service center collection begins
        //  - data shared from workload details to service center should not be modified, it needs to be used
        //    in subsequent poll.   Do not modify metric values, data immutability is not yet enforced!
        //    Either make immutable or make defensive copy to share.
        this.qosWorkloadDetailMetricsBuilder = qosWorkloadDetailMetricsBuilder;
    }

    @Override
    List<String> getTargetPerfUUIDs(final NetAppDefaultStorageContext context, final ZAPIConnection zapiConnection) {
        return context.getNonDelayCenterResourceUUIDs();
    }

    @Override
    Cluster.ClusterKey getParentInventoryUUIDKey(final NetAppDefaultStorageContext context, final String perfInstanceUUID) {
        return context.getKey();
    }

    @Override
    String getIventoryUUIDFromPerfUUID(final NetAppDefaultStorageContext context, final String perfInstanceUUID) {
        return context.getKey().getID(Cluster.ClusterKey.Field.uuid) + '.' + perfInstanceUUID;
    }

    @Override
    NetappQosServiceCenterMetrics createMetricsFromValues(final MetricsRowScratchPad<NetappQosServiceCenterMetrics.ServiceCenterCounter> currentValues, final MetricsRowScratchPad<NetappQosServiceCenterMetrics.ServiceCenterCounter> previousValues) {
        return new NetappQosServiceCenterMetrics(currentValues, previousValues);
    }

    @Override
    void reportPerformance(Cluster.ClusterKey clusterKey, final Map<Cluster.ClusterKey, Collection<RawSample<QosServiceCenterMeasurements>>> measurements) throws DataSourceErrorException {
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<QosServiceCenterMeasurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<QosServiceCenterMeasurements>> resourceMeasurements = measurementEntry.getValue();
            storagePerformanceReporter.reportQosServiceCenterPerformance(key, resourceMeasurements);
        }
    }
    
    /**
     * returns metrics for Delay Service Centers resources as summations of child Workload Details metrics(waitTimes).
     * @param context
     * @return collection of QosServiceCenterMeasurements
     */
    public Collection<RawSample<QosServiceCenterMeasurements>> getDelayCenterMetrics(final NetAppDefaultStorageContext context){

        final Collection<RawSample<QosServiceCenterMeasurements>> delayCenterSamples = new LinkedList<RawSample<QosServiceCenterMeasurements>>();

        // First, aggregate workload detail metrics for Delay center resources when there is a previous poll
        final Map<String, Set<String>> resourceToChildWorkloadDetails = context.getQosServiceCenterUUIDtoWorkloadDetailUUIDs();
        if (this.getPerfUUIDToPreviousValues().size() > 0) {
            final Map<String,  QosWorkloadDetailMeasurements> workloadDetailMeasurementsMap =  
                    qosWorkloadDetailMetricsBuilder.getQosWorkloadDetailMeasurements();
            
            for (final Map.Entry<String, Set<String>> resourceToChildWorkloadDetailsEntry : resourceToChildWorkloadDetails.entrySet()) {
                final String resourceUUID = this.getIventoryUUIDFromPerfUUID(context, resourceToChildWorkloadDetailsEntry.getKey());
                final Set<String> childWorkloadDetailUUIDs = resourceToChildWorkloadDetailsEntry.getValue();
                long sumOfDetailWaitTime = 0l; 
                long detailMaxCollectionPeriod = 0l; 
                
                for (final String workloadDetailUUID : childWorkloadDetailUUIDs) {
                    final String workloadDetailInventoryUUID = qosWorkloadDetailMetricsBuilder.getIventoryUUIDFromPerfUUID(context, workloadDetailUUID);
                    if (workloadDetailMeasurementsMap.containsKey(workloadDetailInventoryUUID)) {
                        final QosWorkloadDetailMeasurements workloadDetailMeasurement = workloadDetailMeasurementsMap.get(workloadDetailInventoryUUID); 
                        
                        final long detailWaitTime = workloadDetailMeasurement.getWaitTime();
                        if (detailWaitTime == 0 || detailWaitTime == MetricsRowSample.NULL_LONG) {
                            continue;
                        }
                        // aggregate wait times from all associated workload details
                        sumOfDetailWaitTime += detailWaitTime;
                        
                        // each workload detail measurement has its own collection period. report the one that took highest time for collection
                        final long detailCollectionPeriod = workloadDetailMeasurement.getCollectionPeriod(); 
                        if (detailCollectionPeriod > detailMaxCollectionPeriod){
                            detailMaxCollectionPeriod = detailCollectionPeriod;
                        }
                    }
                }

                if (sumOfDetailWaitTime > 0l) {
                    // Data to report!
                    final QosServiceCenterMeasurements serviceCenterMeasurements =
                                new QosServiceCenterMeasurements(sumOfDetailWaitTime, MetricsRowSample.NULL_FLOAT,
                                        MetricsRowSample.NULL_LONG, MetricsRowSample.NULL_LONG, detailMaxCollectionPeriod);
                    final RawSample<QosServiceCenterMeasurements> serviceCenterSample = new RawSample<QosServiceCenterMeasurements>(resourceUUID, serviceCenterMeasurements);
                        delayCenterSamples.add(serviceCenterSample);
                } else {
                    if (logger.isDebugEnabled()) {
                        logger.debug("volume workload: " + resourceUUID + " is invalid (no previous and/or no valid child workload details)");
                    }
                }
            }
        }
        return delayCenterSamples;
    }

    @Override
    public void getAndReportPerformance(final NetAppDefaultStorageContext context, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        final Cluster.ClusterKey clusterKey = context.getKey();
        
        final Map<Cluster.ClusterKey, Collection<RawSample<QosServiceCenterMeasurements>>> metricsToReport = new HashMap<Cluster.ClusterKey, Collection<RawSample<QosServiceCenterMeasurements>>>();
        final Collection<RawSample<QosServiceCenterMeasurements>> delayCenterSamples = getDelayCenterMetrics(context);

        // Next, get (from ZAPI) metrics for non-Delay center resources
        metricsToReport.putAll(super.getMetricsAndReportRawPerformance(context, zapiConnection));

        // Add our derived counters
        if (!metricsToReport.containsKey(clusterKey)) {
            // It is possible that all service centers are delay centers,
            // if so initialize the collection here
            metricsToReport.put(clusterKey, new LinkedList<RawSample<QosServiceCenterMeasurements>>());
        }
        metricsToReport.get(clusterKey).addAll(delayCenterSamples);

        // Report the combined set of metrics
        reportPerformance(clusterKey, metricsToReport);
    }

}
