package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappstorageperformance;

import com.netapp.oci.netapp.performance.qosworkloadconstituent.interfaces.data.QosWorkloadConstituentMeasurements;
import com.netapp.oci.netapp.performance.qosworkloaddetail.interfaces.data.QosWorkloadDetailMeasurements;
import com.netapp.oci.netapp.performance.qosworkloadnblade.interfaces.data.QosWorkloadQueueNBladeMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

import java.util.*;
import java.util.Map.Entry;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Builder class for processing and reporting workload queue nblade metrics derived from Qos workload detail 
 * and Qos workload constituent measurements.
 */
public class QosWorkloadQueueNBladeMetricsBuilder {

    private static final Logger LOG = LoggerFactory
            .getLogger(QosWorkloadQueueNBladeMetricsBuilder.class);

    private final QosWorkloadDetailMetricsBuilder qosWorkloadDetailMetricsBuilder;
    private final QosWorkloadConstituentMetricsBuilder qosWorkloadConstituentMetricsBuilder;

    protected final NetAppOCIEPerformancePackage reportingPackage; 

    public QosWorkloadQueueNBladeMetricsBuilder(
            final NetAppOCIEPerformancePackage reportingPackage,
            final QosWorkloadDetailMetricsBuilder qosWorkloadDetailMetricsBuilder,
            final QosWorkloadConstituentMetricsBuilder qosWorkloadConstituentMetricsBuilder) {

        this.reportingPackage = reportingPackage;
        this.qosWorkloadDetailMetricsBuilder = qosWorkloadDetailMetricsBuilder;
        this.qosWorkloadConstituentMetricsBuilder = qosWorkloadConstituentMetricsBuilder;

    }

    public void getAndReportPerformance(NetAppDefaultStorageContext context) throws DataSourceErrorException {
        Map<Cluster.ClusterKey,Collection<RawSample<QosWorkloadQueueNBladeMeasurements>>> metricsToReport = new HashMap<Cluster.ClusterKey, Collection<RawSample<QosWorkloadQueueNBladeMeasurements>>>();
        LinkedList<RawSample<QosWorkloadQueueNBladeMeasurements>> measurementList = getNbladePerformanceMetrics(context);
        metricsToReport.put(context.getKey(), measurementList);
        reportPerformance(metricsToReport);
    }
    
    /**
     * process Qos workload detail and Qos workload constituent measurements to derive Qos workload queue nblade metrics
     */
    public LinkedList<RawSample<QosWorkloadQueueNBladeMeasurements>> getNbladePerformanceMetrics(NetAppDefaultStorageContext context){
        LinkedList<RawSample<QosWorkloadQueueNBladeMeasurements>> measurementList = new LinkedList<RawSample<QosWorkloadQueueNBladeMeasurements>>();
        
        Map<String, QosWorkloadDetailMeasurements> qosWorkloadDetailMeasurementsMap = qosWorkloadDetailMetricsBuilder.getQosWorkloadDetailMeasurements();
        Map<String, QosWorkloadConstituentMeasurements> qosWorkloadConstituentMeasurementsMap = qosWorkloadConstituentMetricsBuilder.getQosWorkloadConstituentMeasurements();
        Map<String, String> nodeNameToUUID = context.getNodeNameToUUIDMap();
        final Map<String, Map<String, String>> qosWorkloadNodeRelationshipMap = context.getQosWorkloadConstituentMap();

        for(Entry<String, Map<String, String>> eachNodeWorkloadRelationship : qosWorkloadNodeRelationshipMap.entrySet()){
            String nodeName = eachNodeWorkloadRelationship.getValue().get("WorkloadConstituentNodeName");
            String workloadAndClusterUUID = eachNodeWorkloadRelationship.getValue().get("WorkloadUUID");
            String workloadName = eachNodeWorkloadRelationship.getValue().get("WorkloadName");
            String workloadUUID = workloadAndClusterUUID.substring(workloadAndClusterUUID.indexOf(".")+1, workloadAndClusterUUID.length());
            String constituentInventoryUUID = nodeNameToUUID.get(nodeName).concat(".").concat(workloadUUID);

            QosWorkloadQueueNBladeMeasurements eachNBladeMeasurements = new QosWorkloadQueueNBladeMeasurements();
            // counters derived from workload constituent
            QosWorkloadConstituentMeasurements eachWorkloadConstMeasurement = qosWorkloadConstituentMeasurementsMap.get(constituentInventoryUUID);
            if(eachWorkloadConstMeasurement != null){
                eachNBladeMeasurements.setReadOps(eachWorkloadConstMeasurement.getReadOps())
                .setWriteOps(eachWorkloadConstMeasurement.getWriteOps())
                .setOtherOps(eachWorkloadConstMeasurement.getOps()-(eachWorkloadConstMeasurement.getReadOps() + eachWorkloadConstMeasurement.getWriteOps()))
                .setReadLatency(eachWorkloadConstMeasurement.getReadLatency())
                .setWriteLatency(eachWorkloadConstMeasurement.getWriteLatency())
                .setOtherLatency(eachWorkloadConstMeasurement.getOtherLatency())
                .setReadData(eachWorkloadConstMeasurement.getReadData())
                .setWriteData(eachWorkloadConstMeasurement.getWriteData());
                setMaxCollectionPeriod(eachNBladeMeasurements, eachWorkloadConstMeasurement, null);
            }

            String workloadDetailUUID = manufactureWorkloadDetailUUID(context, nodeName, workloadName);
            // workload_detail[node, CPU_nblade]
            String cpuNBladeKey = workloadDetailUUID.concat("CPU_nblade");
            QosWorkloadDetailMeasurements qosWorkloadDetailMeasurement = qosWorkloadDetailMeasurementsMap.get(cpuNBladeKey);     
            if(qosWorkloadDetailMeasurement != null){
                eachNBladeMeasurements.setCpuNBladeResidenceTime(qosWorkloadDetailMeasurement.getServiceTime() + qosWorkloadDetailMeasurement.getWaitTime())
                .setCpuNBladeUtilServiceTime(qosWorkloadDetailMeasurement.getServiceTime());
                setMaxCollectionPeriod(eachNBladeMeasurements, null, qosWorkloadDetailMeasurement);
            }

            // workload_detail[node, DELAY_CENTER_CLUSTER_INTERCONNECT]
            String dcClusterInterconnectKey = workloadDetailUUID.concat("DELAY_CENTER_CLUSTER_INTERCONNECT"); 
            qosWorkloadDetailMeasurement = qosWorkloadDetailMeasurementsMap.get(dcClusterInterconnectKey);
            if(qosWorkloadDetailMeasurement != null){
                eachNBladeMeasurements.setDelayClusterInterconnectWaitTime(qosWorkloadDetailMeasurement.getWaitTime());
                setMaxCollectionPeriod(eachNBladeMeasurements, null, qosWorkloadDetailMeasurement);
            }

            // workload_detail[node, DELAY_CENTER_QOS_LIMIT]
            String dcQosLimit = workloadDetailUUID.concat("DELAY_CENTER_QOS_LIMIT"); 
            qosWorkloadDetailMeasurement = qosWorkloadDetailMeasurementsMap.get(dcQosLimit);
            if(qosWorkloadDetailMeasurement != null){
                eachNBladeMeasurements.setDelayQosLimitWaitTime(qosWorkloadDetailMeasurement.getWaitTime());
                setMaxCollectionPeriod(eachNBladeMeasurements, null, qosWorkloadDetailMeasurement);
            }

            // workload_detail[node, DELAY_CENTER_NETWORK]
            String dcNetwork = workloadDetailUUID.concat("DELAY_CENTER_NETWORK"); 
            qosWorkloadDetailMeasurement = qosWorkloadDetailMeasurementsMap.get(dcNetwork);
            if(qosWorkloadDetailMeasurement != null){
                eachNBladeMeasurements.setDelayNetworkWaitTime(qosWorkloadDetailMeasurement.getWaitTime());
                setMaxCollectionPeriod(eachNBladeMeasurements, null, qosWorkloadDetailMeasurement);
            }

            if(isZeroSupressDetected(eachNBladeMeasurements)){
                continue;
            }else{
                String key = new StringBuilder(nodeNameToUUID.get(nodeName)).append('.').append(workloadUUID).toString();
                RawSample<QosWorkloadQueueNBladeMeasurements> eachRawSample = new RawSample<QosWorkloadQueueNBladeMeasurements>(key, eachNBladeMeasurements);
                measurementList.add(eachRawSample);
            }
        }
        return measurementList;
    }

    void reportPerformance(final Map<Cluster.ClusterKey, Collection<RawSample<QosWorkloadQueueNBladeMeasurements>>> measurements)
            throws DataSourceErrorException {
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<QosWorkloadQueueNBladeMeasurements>>> measurementEntry : measurements
                .entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<QosWorkloadQueueNBladeMeasurements>> workloadQueueNBladeMeasurements = measurementEntry
                    .getValue();
            reportingPackage.reportQosWorkloadQueueNBladePerformance(
                    key, workloadQueueNBladeMeasurements);
            
        }
    }

    public String manufactureWorkloadDetailUUID(NetAppDefaultStorageContext context, String nodeName, String workloadName) {
        String clusterUUID = context.getKey().getOcieUUID();
        String workloadUUID = new StringBuilder(clusterUUID).append('.').append(nodeName)
                .append(":kernel:").append(workloadName).append('.').toString();
        return workloadUUID;
    }

    private boolean isZeroSupressDetected(QosWorkloadQueueNBladeMeasurements nBladeMeasurement) {
        return (nBladeMeasurement.getReadOps() == 0f)
                && (nBladeMeasurement.getWriteOps() == 0f)
                && (nBladeMeasurement.getOtherOps() == 0f)
                && (nBladeMeasurement.getReadLatency() == 0f)
                && (nBladeMeasurement.getWriteLatency() == 0f)
                && (nBladeMeasurement.getOtherLatency() == 0f)
                && (nBladeMeasurement.getReadData() == 0f)
                && (nBladeMeasurement.getWriteData() == 0f)
                && (nBladeMeasurement.getCpuNBladeResidenceTime() == 0L)
                && (nBladeMeasurement.getCpuNBladeUtilServiceTime() == 0L)
                && (nBladeMeasurement.getDelayClusterInterconnectWaitTime() == 0L)
                && (nBladeMeasurement.getDelayQosLimitWaitTime() == 0L)
                && (nBladeMeasurement.getDelayNetworkWaitTime() == 0L);
    }
    
    /**
     * this makes sure the collection period for resultant dblade measurement is always the maximum of collection period associated to the source 
     * workload detail and source workload constituent measurement. 
     * this also takes care of wrap/reset measurements by not overriding the dblade collection period to zero  
     * @param nBladeMeasurement
     * @param constituentMeasurement
     * @param detailMeasurement
     */
    private void setMaxCollectionPeriod(QosWorkloadQueueNBladeMeasurements nBladeMeasurement, QosWorkloadConstituentMeasurements constituentMeasurement ,QosWorkloadDetailMeasurements detailMeasurement){
        if(constituentMeasurement != null){
            final long constituentCollectionPeriod = constituentMeasurement.getCollectionPeriod(); 
            if(constituentCollectionPeriod > nBladeMeasurement.getCollectionPeriod()){
                nBladeMeasurement.setCollectionPeriod(constituentCollectionPeriod);
            }
        }
        if(detailMeasurement != null){
            final long detailCollectionPeriod = detailMeasurement.getCollectionPeriod();
            if(detailCollectionPeriod > nBladeMeasurement.getCollectionPeriod()){
                nBladeMeasurement.setCollectionPeriod(detailCollectionPeriod);
            }
        }
    }

}