package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics;

import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.apache.log4j.Logger;

import java.math.BigInteger;

/**
 * Base class for handling sets of counters for a given performance target uuid (node, volume, disk...) from a specific poll,
 * calculating time shift estimates of counter values at poll start, and calculating rates from deltas of counter values and time between polls.
 *
 * The common case is ZAPI, Intermediate, and OCIE values all share the same set of columns,
 * and a single MetricsRowScratchPad contains all 3 stages (Z == I == O).
 *
 * In cases where ZAPI column names return CSV histograms, or OCIE requires custom metric processing
 * (NBlade, DBlade processor metric sums), unique columns will be defined for each stage
 *
 * @param <S> A Measurement container type, defined by Server, that defines the set of reported values for a performance target.
 *              Derived classes must return this type for reporting to the server
 *              This class is a container for counter/columns defined by O for a specific row.
 * @param <R> It is like <S> but contains the raw counters coming back in zapi
 * @param <Z> An Enumeration, defined by derived classes, that identifies the set of requested ZAPI counter/column names
 * @param <O> An Enumeration, defined by derived classes, that identifies the set of final counter/column names reported to OCIE server
 *            Common case is O == Z, occasionally  O is orthogonal to Z with custom calculations
 *
 */
public abstract class MetricsRowSample<S, R, Z extends Enum<Z>, O extends Enum<O>> {

    @SuppressWarnings({"UnusedDeclaration"})
    private static final Logger logger = Logger.getLogger(MetricsRowSample.class);

    // Flags defined by server indicate undefined values
    public static final int NULL_INT = Integer.MIN_VALUE;
    public static final long NULL_LONG = Long.MIN_VALUE;
    public static final float NULL_FLOAT = Float.MIN_VALUE;                             // This is a very small POSITIVE value.
    public static final BigInteger NULL_BIG_INT = BigInteger.valueOf(Long.MIN_VALUE);   // This is just BigInteger alias for NULL_LONG, for comparison to other BigIntegers.
                                                                                        // BigInteger is used for intermediate calculations, the esult is reported as long.
                                                                                        // NULL_BIG_INT is used to identify/flag invalid intermediate BigInteger values. 
                                                                                        // Not a primitive so use .equals NOT ==

    // Container for timestamps, raw values, and time shifted estimates from (based on current/prev poll) if available, for a specific target UUID
    protected final MetricsRowScratchPad<Z> currentValues;

    // Previous values, may be null, or may contain raw values but no estimates
    protected final MetricsRowScratchPad<Z> previousValues;

    /**
     * Token identifying specific enumeration class <Z> for access to class instance methods and EnumMap construction
     */
    final Class<Z> enumClass;


    /**
     * Create a new container for raw polled counters, intermediate time shift estimates, and final reported values.
     *
     * @param enumClass     the class of the enumeration used to represent counter names, for access to class methods
     * @param currentValues the values acquired during the poll
     * @param previousValues previous values for poll start estimation and final value calculation
     */
    public MetricsRowSample(final Class<Z> enumClass, final MetricsRowScratchPad<Z> currentValues, final MetricsRowScratchPad<Z> previousValues) {
        this.enumClass = enumClass;
        this.currentValues = currentValues;
        this.previousValues = previousValues;
    }

    /**
     * Get the ScratchPad of current values
     * @return the ScratchPad of current values
     */
    public MetricsRowScratchPad<Z> getCurrentValues() {
        return currentValues;
    }

    /**
     * Get the ScratchPad of previous values
     * @return the ScratchPad of previous values
     */
    public MetricsRowScratchPad<Z> getPreviousValues() {
        return previousValues;
    }


    /**
     * Get the normalized integer metric value for the given enum (metric identifyer), set of current values, and set of previous values.
     * Some normalizations require more than one counter value, so the entire sets for current and previous are required.
     *
     * @param counterName     the name/id/enum of the counter to normalize
     *
     * @return the normalized counter value, -1 if normalization was not possible
     */
    abstract float getNormalizedFloatMetric(final O counterName);


    /**
     * Normalized this metrics instance against the previous metrics instance, and wrap in the
     * measurements class expected by the server
     * @param context in case other data from the context is required to normalize data
     * @return normalized metric values wrapped in the class expected by the server
     */
    public abstract S normalizeMetrics(final NetAppDefaultStorageContext context);

    /**
     * Get the raw counters from the metric
     * @param context in case other data is needed
     * @return
     */
    public R rawMetrics(final NetAppDefaultStorageContext context){return null;}


    /**
     * Get the flag indicating wrap/reset detected.  This causes the uuid to be reported (confirming existinance of uuid),
     * however to values are reported.   A wrap/reset in any column invalidates the entire row of values.
     * @return the flag indicating wrap/reset detected.
     */
    public boolean isWrapResetDetected() {
        // JSD todo check ZAPI, Intermediate, final

        // Poll Start Estimates values are based on 3 consecutive samples.
        // Wrap/Reset may have occurred between polls 1 and 2 (previous) or between 2 and 3 (current)
        return currentValues.isWrapResetDetected() || (previousValues != null && previousValues.isWrapResetDetected());
    }

    public boolean isZeroSuppressDetected() {
        return currentValues.isZeroSuppressDetected();
    }

    /**
     * Given the numeratorName, return the corresponding denominatorName for calculating percents and ratios
     *
     * @param numeratorName the numeratorName to find the denominatorName for
     *
     * @return the corresponding denominatorName for calculating percents and ratios
     */
    Z getDenominatorForNumerator(final Z numeratorName) {
        return null;
    }

    /**
     * @return <tt>true</tt> if the specified value is <tt>null</tt> or represents the undefined BigInteger value,
     *         <tt>false</tt> otherwise.
     */
    public static boolean isNull(final BigInteger value) {
        return (value == null) || NULL_BIG_INT.equals(value);
    }
    
}
