package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics;

import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import org.apache.log4j.Logger;

import java.math.BigInteger;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.List;

/**
 * For a single target performance UUID and corresponding foundation UUID, and a set of counter/column names <C>,
 * this class serves as a container for results of a Single step of the transformation pipeline from raw string value
 * to final normalized value reported to server.
 *
 * Steps serviced by this class include:
 *      Raw values from Cluster ZAPI response <Z>
 *      Intermediate calculations (included in debug CSV) <I>
 *      Final values reported to OCIE server <E>
 *
 * Every step of the calculation pipeline is based on monotonically increasing counters.
 * A single step of the calculation pipeline consists of calculating the delta between the
 * current and previous counter values, setting wrap/reset and zero suppression detection
 * where appropriate, and applying a recipe the the delta between consecutive polls
 * (e.g. calculating rate of change, or percent busy)
 *
 * The common case is ZAPI, Intermediate, and OCIE values all share the same set of columns,
 * and a single MetricsRowScratchPad contains all 3 stages.
 *
 * In cases where ZAPI column names return CSV histograms, or OCIE requires custom metric processing
 * (NBlade, DBlade processor metric sums), unique columns will be defined for each stage
 *
 * @param <C> the enumeration that defines the Counter/Column names for this step of translation.
 * For each column, we keep track of current value, previous value, delta value, recipe applied (to delta),
 * and final calculated value.
 */
public class MetricsRowScratchPad<C extends Enum<C>> {

    @SuppressWarnings({"UnusedDeclaration"})
    private static final Logger logger = Logger.getLogger(MetricsRowScratchPad.class);

    /**
     * The performance UUID for this set of metric values (used to get values from ZAPI)
     */
    protected final String perfInstanceUUID;

    /**
     * The inventory UUID for this set of metric values (used to report values to OCIE)
     */
    protected final String inventoryUUID;

    /**
     * Enum Type token for EnumSet, EnumMap construction
     */
    protected final Class<C> enumClass;

    /**
     * If metrics debug is enabled, then the final values will be held in the finalValues map.
     */
    protected final boolean metricsDebugEnabled;

    /**
     * For each requested counter name, the raw value from ZAPI. The map only contains raw values that are not numbers
     * (Strings). All numbers are converted to BigInteger's and stored in counterValues.
     */
    protected final EnumMap<C, String> rawValues;

    /**
     * For each requested counter name, the parsed (big) integer value.   Not always valid for all requested names (e.g. csv histograms)
     */
    protected final EnumMap<C, BigInteger> counterValues;

    /**
     * For each Metric, the final value after time shifting raw values and applying normalization recipe
     */
    protected final EnumMap<C, Number> finalValues;

    /**
     * If any counter wrap or reset is detected, the sample should not be reported
     */
    protected boolean wrapResetDetected = false;

    /**
     * If counter reported as exactly "0", assume ZAPI has applied zero-supression.
     * For
     */
    protected boolean zeroSuppressDetected = false;


    /**
     * The Server's time stamp when the poll was started
     */
    protected final long serverPollStartTimeStamp;

    /**
     * The local acquisition time stamp when acquisition invoked the ZAPI to get this counter instance. In case of a
     * chunked ZAPI request this is the timestamp of the first ZAPI request. For the counters retrieved from the archive
     * this is the timestamp associated with the counter in the archive file.
     */
    protected final long pollStartTimeStamp;


    /**
     * Reference to previous
     */
   protected MetricsRowScratchPad<C> previousValues;

    public MetricsRowScratchPad(final Class<C> enumClass, final String perfInstanceUUID, final String inventoryUUID,
                                final long serverPollStartTimeStamp, final long pollStartTimeStamp,
                                final MetricsRowScratchPad<C> previousValues, boolean isMetricsDebugEnabled) {
        this.enumClass = enumClass;
        this.metricsDebugEnabled = isMetricsDebugEnabled;
        this.perfInstanceUUID = (perfInstanceUUID == null) ? "" : perfInstanceUUID;
        this.inventoryUUID = (inventoryUUID == null) ? "" : inventoryUUID;
        this.rawValues = new EnumMap<C, String>(enumClass);
        this.counterValues = new EnumMap<C, BigInteger>(enumClass);
        this.finalValues = new EnumMap<C, Number>(enumClass);
        this.serverPollStartTimeStamp = serverPollStartTimeStamp;
        this.pollStartTimeStamp = pollStartTimeStamp;
        this.previousValues = previousValues;
        if (this.previousValues != null) {
            this.previousValues.previousValues = null;   // Only keep one previous poll
        }
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getPerfInstanceUUID() {
        return perfInstanceUUID;
    }

    public String getInventoryUUID() {
        return inventoryUUID;
    }

    /**
     * A String representation counters/Columns <C> in this scratchpad.
     * @return A String representation counters/Columns <C> in this scratchpad.
     */
    @SuppressWarnings({"UnusedDeclaration"})
    public String getMetricType() {
        return enumClass.getSimpleName();
    }

    public EnumMap<C, String> getRawValues() {
        return rawValues;
    }

    public EnumMap<C, BigInteger> getCounterValues() {
        return counterValues;
    }

    public long getServerPollStartTimeStamp() {
        return serverPollStartTimeStamp;
    }

    public long getPollStartTimeStamp() {
        return pollStartTimeStamp;
    }

    public MetricsRowScratchPad<C> getPreviousScratchPad() {
        return previousValues;
    }

    public BigInteger addToCounter(final C counterName, final BigInteger valueToAdd) {
        final BigInteger bigInteger = counterValues.get(counterName);
        if(valueToAdd == null || MetricsRowSample.NULL_BIG_INT.equals(valueToAdd)) {
            //Nothing to add
            return (bigInteger == null) ? MetricsRowSample.NULL_BIG_INT : bigInteger;
        }
        if(bigInteger == null || MetricsRowSample.NULL_BIG_INT.equals(bigInteger)) {
            counterValues.put(counterName, valueToAdd);
            return valueToAdd;
        }
        final BigInteger result = bigInteger.add(valueToAdd);
        counterValues.put(counterName, result);
        return result;
    }

    public void addToScratchPad(MetricsRowScratchPad<C> scratchPad) {
    	for (C key : scratchPad.getCounterValues().keySet()) {
    		addToCounter(key, scratchPad.getCounterValues().get(key));
    	}
    }

    void setWrapResetDetected(final C counterName, final BigInteger previous, final BigInteger current) {
        final com.onaro.sanscreen.acquisition.framework.datasource.Package thePackage = PackageContext.getPackage();
        final String dsName = thePackage.getConfiguration().getDataSourceName();
        logger.warn("Performance Counter Wrap/Reset " + dsName + " " + enumClass.getSimpleName() + " "  + perfInstanceUUID + " " + counterName + " " + previous + " " + current);
        this.wrapResetDetected = true;
    }

    void setTimestampWrapResetDetected(final C counterName, final long previous, final long current) {
        logger.warn("Performance Counter Timestamp Wrap/Reset " + perfInstanceUUID + " " + enumClass.getSimpleName() + " " + counterName + " " + previous + " " + current);
        this.wrapResetDetected = true;
    }

    public boolean isWrapResetDetected() {
        return wrapResetDetected;
    }

    public boolean isZeroSuppressDetected() {
        return zeroSuppressDetected;
    }

    public void setZeroSuppressDetected(final boolean zeroSuppressDetected) {
        this.zeroSuppressDetected = zeroSuppressDetected;
    }

    /**
     * Given the numeratorName, return the corresponding denominatorName for calculating percents and ratios
     *
     * @param numeratorName the numeratorName to find the denominatorName for
     *
     * @return the corresponding denominatorName for calculating percents and ratios
     */
    C getDenominatorForNumerator(@SuppressWarnings({"UnusedDeclaration"}) final C numeratorName) {
        return null;
    }

    /**
     * Get the difference between two polls if possible, -1 if counter either value unknown or counter wrap/reset
     *
     * @param current  the current poll value (should be greater than previous poll)
     * @param previous the previous poll value
     *
     * @return the difference between two polls if possible, -1 if counter either value unknown or counter wrap/reset
     */
    public static long deltaLong(final long current, final long previous) {
        return ((current == MetricsRowSample.NULL_LONG || previous == MetricsRowSample.NULL_LONG) || (current < previous))
                    ? MetricsRowSample.NULL_LONG
                    : current - previous;
    }

    /**
     * Add long a to long b
     * @param a A value to add
     * @param b Another value to add
     * @return the sum of a and b
     */
    public static long addLong(final long a, final long b) {
        if (MetricsRowSample.NULL_LONG == a) {
            return MetricsRowSample.NULL_LONG;
        }
        if (MetricsRowSample.NULL_LONG == b) {
            return MetricsRowSample.NULL_LONG;
        }
        return a + b;
    }

    /**
     * Subtract long b from long a
     * @param a A long value
     * @param b Another long value
     * @return the difference of a and b
     */
    public static long subtractLong(final long a, final long b) {
        if (MetricsRowSample.NULL_LONG == a) {
            return MetricsRowSample.NULL_LONG;
        }
        if (MetricsRowSample.NULL_LONG == b) {
            return MetricsRowSample.NULL_LONG;
        }
        return a - b;
    }

    /**
     * Add float a to float b
     * @param a A value to add
     * @param b Another value to add
     * @return the sum of a and b
     */
    public static float ocieFloatAdd(final float a, final float b) {
        if (MetricsRowSample.NULL_FLOAT == a) {
            return MetricsRowSample.NULL_FLOAT;
        }
        if (MetricsRowSample.NULL_FLOAT == b) {
            return MetricsRowSample.NULL_FLOAT;
        }
        return a + b;
    }

    /**
     * Subtract float b from float a
     * @param a The value to subtract from
     * @param b The value to be subtracted
     * @return the difference between a and b
     */
    public static float ocieFloatSubtract(final float a, final float b) {
        if (MetricsRowSample.NULL_FLOAT == a) {
            return MetricsRowSample.NULL_FLOAT;
        }
        if (MetricsRowSample.NULL_FLOAT == b) {
            return a;
        }
        final float delta = a - b;
        return Math.max(0.0f, delta);
    }

    /**
     * Safely calculate the delta between two BigInteger values
     * @param current the current value
     * @param previous the previous value
     * @return the delta between two values if possible, otherwise NULL_LONG
     */
    public long deltaBigInteger(final C counterName, final BigInteger current, final BigInteger previous) {
        if (current == null || previous == null) {
            return MetricsRowSample.NULL_LONG;
        }
        if (MetricsRowSample.NULL_BIG_INT.equals(current) || MetricsRowSample.NULL_BIG_INT.equals(previous)) {
            return MetricsRowSample.NULL_LONG;
        }
        if (-1 == current.compareTo(previous)) {
            setWrapResetDetected(counterName, previous, current);
            return MetricsRowSample.NULL_LONG;
        }
        return current.subtract(previous).longValue();
    }

    /**
     * Indicates whether the BigInteger is null - either actually null or the flag for null
     * @param value the BigInteger to test
     * @return <tt>true</tt> if the specified value is <tt>null</tt> or represents the undefined BigInteger value,
     *         <tt>false</tt> otherwise.
     */
    public static boolean isNull(final BigInteger value) {
        return (value == null) || MetricsRowSample.NULL_BIG_INT.equals(value);
    }

    protected BigInteger getPreviousValue(final C counterName) {
        if (previousValues == null) {
            return null;
        } else {
            return previousValues.getCounterValues().get(counterName);
        }
    }

    /**
     * Get the difference between two polls if possible, NULL_LONG if counter either value unknown or counter wrap/reset.
     * When wrap/reset detected, the sample sets isWrapResetDetected to true so the sample will not be reported.
     *
     * This method does not write to the debug CSV. Use this method, if the delta values are used in the computation of
     * the final counter value
     *
     * @param counterName     the name/id/enum of the counter to normalize
     *
     * @return the difference between two polls if possible, NULL_LONG if counter either value unknown or counter wrap/reset
     */
    public final long getCounterDelta(final C counterName) {
        final BigInteger currentValue = getCounterValues().get(counterName);
        if(previousValues == null){
        	return MetricsRowSample.NULL_LONG;
        }
        
        final BigInteger previousValue = previousValues.getCounterValues().get(counterName);
        if (previousValue == null || MetricsRowSample.NULL_BIG_INT.equals(previousValue) || currentValue == null || MetricsRowSample.NULL_BIG_INT.equals(currentValue)) {
            // Wrap/Reset causes row to be dropped.   Return null flag but don't set wrap/reset flag
            return MetricsRowSample.NULL_LONG;
        }
        // deltaBigInteger detects counter wrap/reset
        // This probably never happens for the Poll Start Estimate values.
        // Instead, wrap/reset is first detected in the deltas of raw counter values.
        return deltaBigInteger(counterName, currentValue, previousValue);
    }

    /**
     * This method is a wrapper around getCounterDelta. In addition to calculating the delta value, this method also
     * writes to the debug CSV.This method should be called when the final counter is the delta.
     *
     * @param counterName
     * @return
     */
    public final long getCounterDeltaFinal(final C counterName) {
        long counterDelta = getCounterDelta(counterName);
        setFinalValueForDebug(counterName, counterDelta);
        return counterDelta;
    }

    /**
     * Gets the rate of change for the counter identified by counterName.  Rate of change
     * is calculated as (delta counter / delta time).
     * @param counterName the counter to return rate of change for
     * @return the rate of change for the counter identified by counterName.
     */
    public final float getCounterRatePerSec(final C counterName) {
        // All recipes are applied to Time Shifted Poll Start Estimates, not raw counter values
        final long deltaTime = deltaLong(getPollStartTimeStamp(), previousValues.getPollStartTimeStamp());
        if (deltaTime == MetricsRowSample.NULL_LONG) {
            setTimestampWrapResetDetected(counterName, previousValues.getPollStartTimeStamp(), getPollStartTimeStamp());               // Timestamps out of order, should not happen
            setFinalValueForDebug(counterName, MetricsRowSample.NULL_FLOAT);
            return MetricsRowSample.NULL_FLOAT;
        }

        return getDeltaLongPerSecFromDeltaMs(getCounterValues().get(counterName),
                                             previousValues.getCounterValues().get(counterName), deltaTime, counterName);
    }

    /**
     * get the rate of change in long values given the current and previous values, and the delta time in ms
     *
     * @param currentValue  the counter value from the current poll
     * @param previousValue the counter value from the previous poll
     * @param deltaTimeMs      the time in ms between polls
     * @param counterName      for debugging
     *
     * @return the rate of change in long values given the current and previous values
     */
    private final float getDeltaLongPerSecFromDeltaMs(final BigInteger currentValue, final BigInteger previousValue, final long deltaTimeMs, final C counterName) {
        if (previousValue == null || MetricsRowSample.NULL_BIG_INT.equals(previousValue) || currentValue == null || MetricsRowSample.NULL_BIG_INT.equals(currentValue)) {
            // When previous value was not collected or was invalid due to wrap/reset
            setFinalValueForDebug(counterName, MetricsRowSample.NULL_FLOAT);
            return MetricsRowSample.NULL_FLOAT;
        }
        final long rawDelta = deltaBigInteger(counterName, currentValue, previousValue);
        if (rawDelta == MetricsRowSample.NULL_LONG) {
            return MetricsRowSample.NULL_FLOAT;
        }
        final double deltaPerMSDouble = ((double) rawDelta) / ((double) (deltaTimeMs));
        final double deltaPerSDouble = deltaPerMSDouble * 1000d;
        setFinalValueForDebug(counterName, (float) deltaPerSDouble);
        return (float) deltaPerSDouble;
    }

    /**
     * Get the operation latency for the given latency counter id, current and previous raw metric values
     * Operation latency is defined as the ratio of the change in operation latency count (amount of time
     * spent waiting on operations between the current and previous poll) and the change in operation count
     * (operations performed between the current and previous poll)
     *
     * @param latencyCounterName the name/id/enum of the counter to get operation latency
     *
     * @return the operation latency for the given latency counter id, current and previous raw metric values
     */
    public final float getLatencyPerOp(final C latencyCounterName, final C denominatorName) {
        final long deltaLatency = getCounterDelta(latencyCounterName);
        final long deltaOps = getCounterDelta(denominatorName);
        final float latencyPerOp;
        if (deltaOps > 0) {
            if (deltaLatency == MetricsRowSample.NULL_LONG) {
                latencyPerOp = MetricsRowSample.NULL_FLOAT;
            }
            else {
                final double deltaLatencyPerOpDouble = ((double) deltaLatency) / ((double) deltaOps);
                latencyPerOp = (float) deltaLatencyPerOpDouble;
            }
        }
        else if (deltaOps == 0) {
            latencyPerOp = 0;
        }
        else {
            latencyPerOp = MetricsRowSample.NULL_FLOAT;
        }
        setFinalValueForDebug(latencyCounterName, latencyPerOp);
        return latencyPerOp;
    }

    /**
     * Calculate the ratio of rate of change of the given numeratorName counter with respect to the
     * defined denominator (see: getDenominatorForNumerator()
     * @param numeratorName the name of the numerator
     * @return the ratio of rate of change of the given numeratorName counter with respect to the
     * defined denominator
     */
    public final float getPercentFromDeltas(final C numeratorName, final C denominatorName) {
        final long deltaNumerator = getCounterDelta(numeratorName);
        final long deltaDenominator = getCounterDelta(denominatorName);

        final float percent = (deltaDenominator == 0 || deltaNumerator == 0)
                              ? 0
                              : (deltaDenominator > 0 && deltaNumerator > 0)
                                    ? (float) (100d * (double) deltaNumerator / (double) deltaDenominator)
                                    : MetricsRowSample.NULL_FLOAT;
        setFinalValueForDebug(numeratorName, percent);
        return percent;
    }

    /**
     * Calculate the sum of Big Integers passed to the method, ignoring the MetricsRowSample.NULL_BIG_INTs in the list.
     * @param numbers - the array of Big Integers
     * @return sum of the BigIntegers
     */
    public final static BigInteger add(final BigInteger... numbers){
    	BigInteger sum = BigInteger.ZERO;
    	for(BigInteger number : numbers){
    		if(!MetricsRowSample.NULL_BIG_INT.equals(number) && number != null){
    			sum = sum.add(number);
    		}
    	}
    	return sum;
    }

    /**
     * Calculate the sum of floats passed to the method, ignoring the MetricsRowSample.NULL_FLOATs in the list.
     * @param numbers - the array of floats
     * @return sum of the floats
     */
    public final static float add(final float... numbers){
    	float sum = 0.0f;
    	for(float number : numbers){
    		if(number != MetricsRowSample.NULL_FLOAT){
    			sum += number;
    		}
    	}
    	return sum;
    }

    /**
     *  calculate weighted average  e.g. (latency1*op1 + latecy2*op2 + latency3*op3) /(op1+op2+op3)
     *  @param latencyList
     *  @param opList
     *  @return weighted_average
     */
    public final static float averageLatency(final List<Float> latencyList, final List<Float> opList) {
        if(latencyList.size() !=  opList.size()) {
            return MetricsRowSample.NULL_FLOAT;
        }

        float sum_numerator = 0.0f;
        float sum_denominator = 0.0f;

        for(int i=0; i< latencyList.size(); i++)  {
            if(latencyList.get(i) == MetricsRowSample.NULL_FLOAT || opList.get(i) == MetricsRowSample.NULL_FLOAT){
                continue;
            }
            sum_denominator = add(sum_denominator, opList.get(i));
            sum_numerator = add(sum_numerator, latencyList.get(i) * opList.get(i));
        }

        if(sum_denominator==0)  {
            return MetricsRowSample.NULL_FLOAT;
        }

        return (float)((double)sum_numerator / (double)sum_denominator);
    }


    /**
     * Calculate the sum of longs passed to the method, ignoring the MetricsRowSample.NULL_Longs in the list.
     * @param numbers - the array of longs
     * @return sum of the longs
     */

    public final static long add(final long... numbers){
    	long sum = 0l;
    	for(long number : numbers){
    		if(number != MetricsRowSample.NULL_LONG){
    			sum += number;
    		}
    	}
    	return sum;
    }

    @Override
    public String toString() {
        return "MetricsScratchPad: " + enumClass.getSimpleName() + ", " + inventoryUUID;
    }

    // Infrastructure for generating debug CSVs

    public void setFinalValueForDebug(final C metricEnum, final Number finalMetricValue) {
        if (metricsDebugEnabled) {
            finalValues.put(metricEnum, finalMetricValue);
        }
    }

    public String getStaticDebugColumnHeaders() {
        final StringBuilder debugBuilder = new StringBuilder();
        debugBuilder.append("\nfoundation UUID,performance UUID,,server poll start,,au poll start,prev au poll start,delta au poll start,,");
        return debugBuilder.toString();
    }

    public String getDebugColumnHeaders(final String targetObjectName) {
        final StringBuilder debugBuilder = new StringBuilder("# ").append(targetObjectName);
        debugBuilder.append(getStaticDebugColumnHeaders());
        for (final C counterName : EnumSet.allOf(enumClass)) {
            debugBuilder.append("curr raw ").append(counterName).append(",prev raw ").append(counterName).append(",delta raw ").append(counterName);
            debugBuilder.append(",final ").append(counterName).append(",,");
        }
        debugBuilder.append('\n');
        return debugBuilder.toString();
    }

    public String getStaticDebugColumns() {
        final Long lastAcqPollStartTimeStamp;
        final Long acqDeltaBetweenPolls;
        if (previousValues != null) {
            lastAcqPollStartTimeStamp = previousValues.getPollStartTimeStamp();
            acqDeltaBetweenPolls = pollStartTimeStamp - lastAcqPollStartTimeStamp;
        } else {
            lastAcqPollStartTimeStamp = null;
            acqDeltaBetweenPolls = null;
        }

        final StringBuilder debugBuilder = new StringBuilder(inventoryUUID).append(',').append(perfInstanceUUID).append(",,");
        debugBuilder.append(getDebugString(serverPollStartTimeStamp)).append(",,");
        debugBuilder.append(getDebugString(pollStartTimeStamp)).append(',').append(getDebugString(lastAcqPollStartTimeStamp)).append(',').append(getDebugString(acqDeltaBetweenPolls)).append(",,");
        return debugBuilder.toString();
    }

    public String toDebugCSV() {
        final StringBuilder debugBuilder = new StringBuilder(getStaticDebugColumns());
        for (final C metricName : EnumSet.allOf(enumClass)) {
            debugBuilder.append(getCounterDebugCSV(metricName));
        }
        debugBuilder.append('\n');
        return debugBuilder.toString();
    }

    protected String getDebugString(final Number number) {
        final String returnVal;
        if (null == number) {
            returnVal = "";
        } else if (MetricsRowSample.NULL_BIG_INT.equals(number)) {
            returnVal = "NULL_BIG_INT";
        } else if (number.equals(MetricsRowSample.NULL_LONG)) {
            returnVal = "NULL_LONG";
        } else if (number.equals(MetricsRowSample.NULL_INT)) {
            returnVal = "NULL_INT";
        } else if (number.equals(MetricsRowSample.NULL_FLOAT)) {
            returnVal = "NULL_FLOAT";
        } else {
            returnVal = String.valueOf(number);
        }
        // logger.error
        return returnVal;
    }

    /**
     * Default to CSV debug string.
     * @param counterName the counter to debug
     * @return  default CSV debug string for the counter
     */
    protected String getCounterDebugCSV(final C counterName) {
        final StringBuilder debugBuilder = new StringBuilder();
        if(counterValues.get(counterName) != MetricsRowSample.NULL_BIG_INT){
        	debugBuilder.append(getDebugString(counterValues.get(counterName))).append(',');             // current counter
        }
        else {
        	debugBuilder.append(rawValues.get(counterName)).append(',');             // current counter
        }
        if (previousValues != null) {
            if (previousValues.counterValues.get(counterName) != MetricsRowSample.NULL_BIG_INT) {
                debugBuilder.append(getDebugString(getPreviousValue(counterName))).append(',');              // prev counter
            } else {
                debugBuilder.append(previousValues.rawValues.get(counterName)).append(',');             // current counter
            }
        }
        debugBuilder.append(previousValues != null ?
                getDebugString(getCounterDelta(counterName)) : "").append(',');                    // delta counter
        debugBuilder.append(getDebugString(finalValues.get(counterName))).append(",,");              // final value
        return debugBuilder.toString();
    }

    public boolean isPreviousPollValuesAvailable() {
        return (previousValues != null);
    }

    /**
     * The time elapsed between the previous and the current poll.
     * @return Collection period in milliseconds
     */
    public long getCollectionPeriod() {
        if (previousValues == null) {
            return MetricsRowSample.NULL_LONG;
        } else {
            return deltaLong(getPollStartTimeStamp(), previousValues.getPollStartTimeStamp());
        }
    }
}
