package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics;


import java.math.BigInteger;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.netapp.performance.qosworkloadconstituent.interfaces.data.QosWorkloadConstituentMeasurements;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

/**
 * This class provides simple transformation recipe to normalize the workload constituent metrics to OCIE expected units
 * (rate or ratio to other counter/column delta). The required Qos workload constituent metrics columns are orthogonal to 
 * the set of metrics requested from ZAPI,so specific logic is required.
 */
public class NetappQosWorkloadConstituentMetrics extends MetricsRowSample<QosWorkloadConstituentMeasurements, QosWorkloadConstituentMeasurements, NetappQosWorkloadConstituentMetrics.QosWorkloadConstituentCounter,
NetappQosWorkloadConstituentMetrics.QosWorkloadConstituentCounter> {
	private static final Logger LOG = LoggerFactory.getLogger(NetappQosWorkloadConstituentMetrics.class);

	/**
	 * Qos workload constituent counters with information on what kind of normalization should be applied on their raw values.
	 * normalizationType = 0  (default) counter holds a rated value. Rate of change is calculated as (delta counter / delta time).
	 * normalizationType = 1  counter holds latency value per operation count
	 * normalizationType = 2  counter holds read latency value per read operation count
	 * normalizationType = 3  counter holds write latency value per write operation count
	 */
	public static enum QosWorkloadConstituentCounter {
		dirty_buffers,
		freed_blocks,
		readahead_x_stripe,
		ops,
		read_ops,
		write_ops,
		latency(1),
		read_latency(2),
		write_latency(3),
		read_data,
		total_data,
		write_data,
		read_io_type_base,
		read_io_type;

		private final int normalizationType;

		private QosWorkloadConstituentCounter(int normalizationType) {
			this.normalizationType = normalizationType;
		}

		private QosWorkloadConstituentCounter() {
			this(0);
		}

		public boolean isCounterValueRated(){
			return normalizationType == 0;
		}

		public boolean isCounterValuePerOps(){
			return normalizationType == 1;
		}

		public boolean isCounterValuePerReadOps(){
			return normalizationType == 2;
		}

		public boolean isCounterValuePerWriteOps(){
			return normalizationType == 3;
		}

	}

	public NetappQosWorkloadConstituentMetrics(final MetricsRowScratchPad<QosWorkloadConstituentCounter> currentValues, final MetricsRowScratchPad<QosWorkloadConstituentCounter> previousValues) {
		super(QosWorkloadConstituentCounter.class, currentValues, previousValues);
	}

	@Override
	float getNormalizedFloatMetric(final QosWorkloadConstituentCounter counterName) {
		final float normalizedMetric;
		if(counterName.isCounterValueRated()){
			normalizedMetric = currentValues.getCounterRatePerSec(counterName);
		}else if(counterName.isCounterValuePerOps()){
			normalizedMetric = currentValues.getLatencyPerOp(counterName, QosWorkloadConstituentCounter.ops);
		}else if(counterName.isCounterValuePerReadOps()){
			normalizedMetric = currentValues.getLatencyPerOp(counterName, QosWorkloadConstituentCounter.read_ops);
		}else if(counterName.isCounterValuePerWriteOps()){
			normalizedMetric = currentValues.getLatencyPerOp(counterName, QosWorkloadConstituentCounter.write_ops);
		}else{
			normalizedMetric = NULL_FLOAT;
		}
		return normalizedMetric;
	}

	public float getLatencyPerOtherOps(MetricsRowScratchPad<QosWorkloadConstituentCounter> currentValues){
		long totalOps = currentValues.getCounterDelta(QosWorkloadConstituentCounter.ops);
		long readOps = currentValues.getCounterDelta(QosWorkloadConstituentCounter.read_ops);
		long writeOps = currentValues.getCounterDelta(QosWorkloadConstituentCounter.write_ops);
		long writeLatency = currentValues.getCounterDelta(QosWorkloadConstituentCounter.write_latency);
		long readLatency = currentValues.getCounterDelta(QosWorkloadConstituentCounter.read_latency);
		long totalLatency = currentValues.getCounterDelta(QosWorkloadConstituentCounter.latency);
		double denominator = ((double) (totalOps - (readOps + writeOps)));
		float latencyPerOp = 0;
		if(denominator > 0){
			double latencyPerOpDouble = ((double) (totalLatency - (readLatency + writeLatency)))/denominator ;
			latencyPerOp = (float) latencyPerOpDouble;
		}
		else if (denominator== 0) {
			latencyPerOp = 0;
		}
		return latencyPerOp;
	}

	@Override
	public QosWorkloadConstituentMeasurements normalizeMetrics(final NetAppDefaultStorageContext context) {

		final long dirtyBuffers = currentValues.getCounterDelta(QosWorkloadConstituentCounter.dirty_buffers);
		final long freedBlocks = currentValues.getCounterDelta(QosWorkloadConstituentCounter.freed_blocks);
		final long readAheadXStripe = currentValues.getCounterDelta(QosWorkloadConstituentCounter.readahead_x_stripe);
		final float ops = getNormalizedFloatMetric(QosWorkloadConstituentCounter.ops);
		final float readOps = getNormalizedFloatMetric(QosWorkloadConstituentCounter.read_ops);
		final float writeOps = getNormalizedFloatMetric(QosWorkloadConstituentCounter.write_ops);
		final float readData = getNormalizedFloatMetric(QosWorkloadConstituentCounter.read_data);
		final float writeData = getNormalizedFloatMetric(QosWorkloadConstituentCounter.write_data);
		final float totalData = getNormalizedFloatMetric(QosWorkloadConstituentCounter.total_data);
		final float readLatency = getNormalizedFloatMetric(QosWorkloadConstituentCounter.read_latency);
		final float writeLatency = getNormalizedFloatMetric(QosWorkloadConstituentCounter.write_latency);
		final float latency = getNormalizedFloatMetric(QosWorkloadConstituentCounter.latency);
		final float otherLatency = getLatencyPerOtherOps(currentValues);
		WorkloadConstituentRowMetricsScratchPad currentValue = (WorkloadConstituentRowMetricsScratchPad)currentValues;
		final long sumDiskAndHyaHDD = currentValue.getSumOfDiskAndHyaHDD();
		final long read_io_type_base = currentValues.getCounterDelta(QosWorkloadConstituentCounter.read_io_type_base);
		final long collectionPeriod = currentValues.getCollectionPeriod();

		if((dirtyBuffers == NULL_FLOAT || dirtyBuffers == 0f) && (freedBlocks == NULL_FLOAT || freedBlocks == 0f)
				&& (readAheadXStripe == NULL_FLOAT || readAheadXStripe == 0f)
				&& (ops == NULL_FLOAT || ops == 0f) && (readOps == NULL_FLOAT || readOps == 0f)
				&& (writeOps == NULL_FLOAT || writeOps == 0f) && (readData == NULL_FLOAT || readData == 0f) && (writeData == NULL_FLOAT || writeData == 0f) 
				&& (totalData == NULL_FLOAT || totalData == 0f) && (readLatency == NULL_FLOAT || readLatency == 0f) 
				&& (writeLatency == NULL_FLOAT || writeLatency == 0f) && (latency == NULL_FLOAT || latency == 0f) 
				&& (otherLatency == NULL_FLOAT || otherLatency == 0f)
				&& (read_io_type_base == NULL_LONG || read_io_type_base == 0l)
				&& (sumDiskAndHyaHDD == NULL_LONG || sumDiskAndHyaHDD == 0l)){
			currentValues.setZeroSuppressDetected(true);
		}

		final QosWorkloadConstituentMeasurements measurementObj = new QosWorkloadConstituentMeasurements()
		.setDirtyBuffers(dirtyBuffers)
		.setFreedBlocks(freedBlocks)
		.setReadAheadXStripe(readAheadXStripe)
		.setOps(ops)
		.setReadOps(readOps)
		.setWriteOps(writeOps)
		.setReadData(readData)
		.setWriteData(writeData)
		.setTotalData(totalData)
		.setReadLatency(readLatency)
		.setWriteLatency(writeLatency)
		.setLatency(latency)
		.setOtherLatency(otherLatency)
		.setReadIOTypeBase(read_io_type_base)
		.setSumDiskAndHyaHDD(sumDiskAndHyaHDD)
		.setCollectionPeriod(collectionPeriod);

		return measurementObj;
	}

	public static final class WorkloadConstituentRowMetricsScratchPad extends MetricsRowScratchPad<NetappQosWorkloadConstituentMetrics.QosWorkloadConstituentCounter> {

		private long sumOfDiskAndHyaHDD;
		private List<String> readIOColumns;


		public WorkloadConstituentRowMetricsScratchPad(String perfInstanceUUID, String inventoryUUID,
                                                       long serverPollStartTimeStamp, long pollStartTimeStamp,
                                                       MetricsRowScratchPad<QosWorkloadConstituentCounter> previousValues,
                                                       boolean metricsDebugEnabled, List<String> readIOColumns){
			super(QosWorkloadConstituentCounter.class, perfInstanceUUID, inventoryUUID, serverPollStartTimeStamp,
                    pollStartTimeStamp, previousValues, metricsDebugEnabled);
			this.readIOColumns = readIOColumns; 
		}

		public long[] getArrayCounterDelta(final Map<QosWorkloadConstituentCounter,String> zapiValues){


			String read_io_current_rawvalue = zapiValues.get(QosWorkloadConstituentCounter.read_io_type);
			String[] read_io_type_string = read_io_current_rawvalue.split(",");
			BigInteger[] read_io_current_values = new BigInteger[read_io_type_string.length]; 
			for(int i=0 ; i < read_io_type_string.length ; i++){
				try {
					read_io_current_values[i] = new BigInteger(read_io_type_string[i]);
				}
				catch (NumberFormatException e){
					read_io_current_values[i] = MetricsRowSample.NULL_BIG_INT;
				}
			}

			String read_io_previous_rawvalue = previousValues.rawValues.get(QosWorkloadConstituentCounter.read_io_type);
			String[] read_io_type_previous_string = read_io_previous_rawvalue.split(",");
			BigInteger[] read_io_previous_values = new BigInteger[read_io_type_previous_string.length];
			for(int i = 0 ; i < read_io_type_previous_string.length ; i++){
				try{
					read_io_previous_values[i] = new BigInteger(read_io_type_previous_string[i]);
				}
				catch(NumberFormatException e){
					read_io_previous_values[i] = MetricsRowSample.NULL_BIG_INT;
				}
			}

			long[] deltaReadIOTypeValues = new long[read_io_current_values.length];
			for(int i = 0 ; i < read_io_current_values.length ; i++){
				BigInteger current = read_io_current_values[i];
				BigInteger previous = read_io_previous_values[i];
				if (current == null || previous == null) {
					deltaReadIOTypeValues[i] =  MetricsRowSample.NULL_LONG;
				}
				if (MetricsRowSample.NULL_BIG_INT.equals(current) || MetricsRowSample.NULL_BIG_INT.equals(previous)) {
					deltaReadIOTypeValues[i] = MetricsRowSample.NULL_LONG;
				}
				if (-1 == current.compareTo(previous)) {
					deltaReadIOTypeValues[i] = MetricsRowSample.NULL_LONG;
					setWrapResetDetected(QosWorkloadConstituentCounter.read_io_type, previous , current);
				}

				deltaReadIOTypeValues[i] = current.subtract(previous).longValue();
			}

			return deltaReadIOTypeValues;
		}
		public long addDiskAndHyaHDD(long[] read_io_type, List<String> volumeWorkloadReadIOColumns){
		    if (read_io_type.length != volumeWorkloadReadIOColumns.size()){
		        LOG.warn("Discovered read_io_type length({}) != volumeWorkloadReadIOColumns size({})", read_io_type.length, volumeWorkloadReadIOColumns.size());
		        return MetricsRowSample.NULL_LONG;
		    }
			long disk = MetricsRowSample.NULL_LONG, hya_hdd = MetricsRowSample.NULL_LONG;
			int i = -1;
			for(final String readIOColumn : volumeWorkloadReadIOColumns){
				i++;
				final long deltaReadIOValue = read_io_type[i];
				if(readIOColumn.equals("disk")){
					disk = deltaReadIOValue;
				}
				else if(readIOColumn.equals("hya_hdd")){
					hya_hdd = deltaReadIOValue;
				}
			}


			return (disk != MetricsRowSample.NULL_LONG && hya_hdd != MetricsRowSample.NULL_LONG) ? (disk + hya_hdd) : MetricsRowSample.NULL_LONG ;
		}

		public boolean isPreviousPollValuesAvailable() {
			return (previousValues != null);
		}

		public long getSumOfDiskAndHyaHDD(){

			if (!this.isPreviousPollValuesAvailable()) {
				return MetricsRowSample.NULL_LONG;
			}

			long[] diff = getArrayCounterDelta(rawValues);
			this.sumOfDiskAndHyaHDD = addDiskAndHyaHDD(diff, readIOColumns);
			return sumOfDiskAndHyaHDD;

		}

		@Override
		protected String getCounterDebugCSV(final QosWorkloadConstituentCounter counterName) {
			final StringBuilder debugBuilder = new StringBuilder();
			if(counterName.name().equals("read_io_type")){
				String counter = rawValues.get(QosWorkloadConstituentCounter.read_io_type);
				if(counter != null){
					debugBuilder.append("\"").append(counter).append("\"");
				}
				debugBuilder.append(',');
				debugBuilder.append(getPreviousReadIOValue()).append(',');
				debugBuilder.append("").append(',');             // delta counter
				debugBuilder.append("").append(',');             // recipe
				debugBuilder.append("").append(",,");            // final value
				return debugBuilder.toString();
			}
			else{
				return super.getCounterDebugCSV(counterName);
			}

		}

		protected String getPreviousReadIOValue() {
			if (previousValues == null) {
				return null;
			} else {
				return  ((previousValues.rawValues.get(QosWorkloadConstituentCounter.read_io_type) != null) 
						? "\"".concat(previousValues.rawValues.get(QosWorkloadConstituentCounter.read_io_type)).concat("\"") 
								: ",") ;
			}
		}
	}
}
