package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.reattribution.impl;


import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.reattribution.data.QosWorkloadDetailData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.reattribution.interfaces.IReAttributionContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This is step one in re-attribution. Identify deferred workloads, group them by node & aggregate and identifies client
 * workload details for resources CPU_dblade_background, DISK_HDD, DISK_SSD, that has to be re-attributed and groups
 * them by node & aggregate.
 */
public class ReAttributeDeferredWorkServiceStep1 {

    private static final Logger logger = LoggerFactory.getLogger(ReAttributeDeferredWorkServiceStep1.class);


    /**
     * Work represented by _WAFL_CP.CPU_dblade in each node. Node Name -> Total Service Time
     */
    private final Map<String, Long> waflCPCPUdBladeByNode = new HashMap<String, Long>();
    /**
     * Work represented by _WAFL_CP.DISK_HDD_* in each aggregate. Aggregate UUID -> Total Service Time
     */
    private final Map<String, Long> waflCPDiskHDDByAggregate = new HashMap<String, Long>();
    /**
     * Work represented by _WAFL_CP.DISK_SSD_* in each aggregate. Aggregate UUID -> Total Service Time
     */
    private final Map<String, Long> waflCPDiskSSDByAggregate = new HashMap<String, Long>();
    /**
     * Work represented by _BACKGROUND_DELETE.CPU_dblade in each node. Node Name -> Total Service Time
     */
    private final Map<String, Long> backgroundDeleteCPUdBladeByNode = new HashMap<String, Long>();
    /**
     * Work represented by _BACKGROUND_DELETE.DISK_HDD_* in each aggregate. Aggregate UUID -> Total Service Time
     */
    private final Map<String, Long> backgroundDeleteDiskHDDByAggregate = new HashMap<String, Long>();
    /**
     * Work represented by _BACKGROUND_DELETE.DISK_SSD_ in each aggregate. Aggregate UUID -> Total Service Time
     */
    private final Map<String, Long> backgroundDeleteDiskSSDByAggregate = new HashMap<String, Long>();
    /**
     * /**
     * Work represented by _RA_STRIPE_X.CPU_dblade in each node. Node Name -> Total Service Time
     */
    private final Map<String, Long> raXStripeCPUdBladeByNode = new HashMap<String, Long>();
    /**
     * Work represented by _RA_STRIPE_X.DISK_HDD_* in each aggregate
     */
    private final Map<String, Long> raXStripeDiskHDDByAggregate = new HashMap<String, Long>();
    /**
     * Work represented by _RA_STRIPE_X.DISK_SSD_ in each aggregate. Aggregate UUID -> Total Service Time
     */
    private final Map<String, Long> raXStripeDiskSSDByAggregate = new HashMap<String, Long>();

    /**
     * CPU_dblade workload detail to be re-attributed. Node Name -> Workload UUID ->  QosWorkloadDetailData
     */
    private final Map<String, HashMap<String, QosWorkloadDetailData>> nodeToWorkloadToCPUDBladeBackgroundMap
            = new HashMap<String, HashMap<String, QosWorkloadDetailData>>();
    /**
     * DISK_HDD workload detail to be re-attributed. Aggregate UUID -> Workload UUID ->  QosWorkloadDetailData
     */
    private final Map<String, HashMap<String, QosWorkloadDetailData>> aggregateToWorkloadToDiskHDDMap
            = new HashMap<String, HashMap<String, QosWorkloadDetailData>>();
    /**
     * DISK_SSD workload detail to be re-attributed. Aggregate UUID -> Workload UUID ->  QosWorkloadDetailData
     */
    private final Map<String, HashMap<String, QosWorkloadDetailData>> aggregateToWorkloadToDiskSSDMap
            = new HashMap<String, HashMap<String, QosWorkloadDetailData>>();

    public ReAttributeDeferredWorkServiceStep1(IReAttributionContext reAttributionContext,
                                               List<QosWorkloadDetailData> qosWorkloadDetailDataList) {
        perform(reAttributionContext, qosWorkloadDetailDataList);

    }

    public Map<String, Long> getWaflCPCPUdBladeByNode() {
        return waflCPCPUdBladeByNode;
    }

    public Map<String, Long> getWaflCPDiskHDDByAggregate() {
        return waflCPDiskHDDByAggregate;
    }

    public Map<String, Long> getWaflCPDiskSSDByAggregate() {
        return waflCPDiskSSDByAggregate;
    }

    public Map<String, Long> getBackgroundDeleteCPUdBladeByNode() {
        return backgroundDeleteCPUdBladeByNode;
    }

    public Map<String, Long> getBackgroundDeleteDiskHDDByAggregate() {
        return backgroundDeleteDiskHDDByAggregate;
    }

    public Map<String, Long> getBackgroundDeleteDiskSSDByAggregate() {
        return backgroundDeleteDiskSSDByAggregate;
    }

    public Map<String, Long> getRaXStripeCPUdBladeByNode() {
        return raXStripeCPUdBladeByNode;
    }

    public Map<String, Long> getRaXStripeDiskHDDByAggregate() {
        return raXStripeDiskHDDByAggregate;
    }

    public Map<String, Long> getRaXStripeDiskSSDByAggregate() {
        return raXStripeDiskSSDByAggregate;
    }

    public Map<String, HashMap<String, QosWorkloadDetailData>> getNodeToWorkloadToCPUDBladeBackgroundMap() {
        return nodeToWorkloadToCPUDBladeBackgroundMap;
    }

    public Map<String, HashMap<String, QosWorkloadDetailData>> getAggregateToWorkloadToDiskHDDMap() {
        return aggregateToWorkloadToDiskHDDMap;
    }

    public Map<String, HashMap<String, QosWorkloadDetailData>> getAggregateToWorkloadToDiskSSDMap() {
        return aggregateToWorkloadToDiskSSDMap;
    }

    /**
     * Parse the workload detail UUID to determine the workload name and resource name. These are then used to group the
     * workload details.
     * <p/>
     * Workload UUID format - cluster_uuid.node_name:kernel:$group1.$group2
     *
     * @param reAttributionContext      Re-attribution context with the inventory data.
     * @param qosWorkloadDetailDataList List of Qos workload details to re-attribute.
     */
    private void perform(IReAttributionContext reAttributionContext,
                         List<QosWorkloadDetailData> qosWorkloadDetailDataList) {

        /*
         * Split the workload detail UUID string into two groups.
         * Group 1: Workload Name. Starts with the characters after :kernel: and ends with the characters before the
         *          following first dot "."
         * Group 2: Resource Name. The rest of the string after the dot "." identified in group 1
         *
         * cluster_uuid.node_name:kernel:$group1.$group2
         */
        final Pattern WORKLOAD__DETAIL_UUID_REGULAR_EXP = Pattern.compile("^.*" + ":kernel:" + "(.*?)" + "\\." + "(.*)$");
        for (final QosWorkloadDetailData workloadDetailData : qosWorkloadDetailDataList) {
            final String workloadDetailUUID = workloadDetailData.getUUID();
            final Map<IReAttributionContext.QosWorkloadDetailAttributes, String> qosWorkloadDetail =
                    reAttributionContext.getQosWorkloadDetailMap().get(workloadDetailUUID);
            if (qosWorkloadDetail != null) {
                String workloadDetailNodeName = qosWorkloadDetail.get(IReAttributionContext.QosWorkloadDetailAttributes.NODE_NAME);
                String workloadDetailAggregateUUID = qosWorkloadDetail.get(IReAttributionContext.QosWorkloadDetailAttributes.AGGREGATE_UUID);
                String workloadUUID = qosWorkloadDetail.get(IReAttributionContext.QosWorkloadDetailAttributes.WORKLOAD_UUID);

                final Matcher uuidMatcher = WORKLOAD__DETAIL_UUID_REGULAR_EXP.matcher(workloadDetailUUID);
                if (uuidMatcher.matches()) {
                    /**Identify deferred workload details. **/
                    String workloadName = uuidMatcher.group(1);
                    String resourceName = uuidMatcher.group(2);
                    groupWorkloadDetail(reAttributionContext, workloadDetailData, workloadDetailNodeName,
                            workloadDetailAggregateUUID, workloadUUID, workloadName, resourceName);
                }
            }
        }

        logger.debug("Results after ReAttributeDeferredWorkServiceStep1 {}", this);

    }

    /**
     * This method loops identifies the workload details and computes the following -
     * <p/>
     * Work represented (service time) by "_WAFL_CP.CPU_dblade" in each node.
     * Work represented (service time) by "_WAFL_CP.DISK_HDD_*" for each aggregate.
     * Work represented (service time) by "_WAFL_CP.DISK_SSD_*" for each aggregate.
     * Work represented (service time) by "_BACKGROUND_DELETE.CPU_dblade" in each node.
     * Work represented (service time) by "_BACKGROUND_DELETE.DISK_HDD_*" for each aggregate.
     * Work represented (service time) by "_BACKGROUND_DELETE.DISK_SSD_*" for each aggregate.
     * Work represented (service time) by "_RA_STRIPE_X.CPU_dblade" in each node.
     * Work represented (service time) by "_RA_STRIPE_X.DISK_HDD_*" for each aggregate.
     * Work represented (service time) by "_RA_STRIPE_X.DISK_SSD_*" for each aggregate
     * Identifies client workload details for resources CPU_dblade_background, DISK_HDD, DISK_SSD, that has to be
     * re-attributed and groups them by node and aggregate
     */
    private void groupWorkloadDetail(IReAttributionContext reAttributionContext,
                                     QosWorkloadDetailData workloadDetailData,
                                     String workloadDetailNodeName, String workloadDetailAggregateUUID,
                                     String workloadDetailWorkloadUUID, String workloadDetailWorkloadName,
                                     String workloadDetailResourceName) {
        final Map<String, Map<IReAttributionContext.AggregateAttributes, String>> qosWorkloadUUIDToAggregateMap
                = reAttributionContext.getQosVolumeWorkloadUUIDToAggregateUUIDMap();
        if ("_WAFL_CP".equals(workloadDetailWorkloadName)) {
            if ("CPU_dblade".equals(workloadDetailResourceName)) {
                //UUID format ${cluster_uuid}.${node_name}:kernel:_WAFL_CP.CPU_dblade
                groupDeferredWork(this.waflCPCPUdBladeByNode, workloadDetailNodeName, workloadDetailData);
            } else if ("CPU_dblade_background".equals(workloadDetailResourceName)) {
                //UUID format ${cluster_uuid}.${node_name}:kernel:_WAFL_CP.CPU_dblade_background
                groupDeferredWork(this.waflCPCPUdBladeByNode, workloadDetailNodeName, workloadDetailData);
            } else if (workloadDetailResourceName != null && workloadDetailResourceName.startsWith("DISK_HDD_")) {
                //UUID format ${cluster_uuid}.${node_name}:kernel:_WAFL_CP.DISK_HDD_${aggregate}_${aggregate uuid}
                groupDeferredWork(this.waflCPDiskHDDByAggregate, workloadDetailAggregateUUID, workloadDetailData);
            } else if (workloadDetailResourceName != null && workloadDetailResourceName.startsWith("DISK_SSD_")) {
                //UUID format ${cluster_uuid}.${node_name}:kernel:_WAFL_CP.DISK_SSD_${aggregate}_${aggregate uuid}
                groupDeferredWork(this.waflCPDiskSSDByAggregate, workloadDetailAggregateUUID, workloadDetailData);
            }
        } else if ("_BACKGROUND_DELETE".equals(workloadDetailWorkloadName)) {
            if ("CPU_dblade".equals(workloadDetailResourceName)) {
                //UUID format ${cluster_uuid}.${node_name}:kernel:_BACKGROUND_DELETE.CPU_dblade
                groupDeferredWork(this.backgroundDeleteCPUdBladeByNode, workloadDetailNodeName, workloadDetailData);
            } else if ("CPU_dblade_background".equals(workloadDetailResourceName)) {
                //UUID format ${cluster_uuid}.${node_name}:kernel:_BACKGROUND_DELETE.CPU_dblade_background
                groupDeferredWork(this.backgroundDeleteCPUdBladeByNode, workloadDetailNodeName, workloadDetailData);
            } else if (workloadDetailResourceName != null && workloadDetailResourceName.startsWith("DISK_HDD_")) {
                //UUID format ${cluster_uuid}.${node_name}:kernel:_BACKGROUND_DELETE.DISK_HDD_${aggregate}_${aggregate uuid}
                groupDeferredWork(this.backgroundDeleteDiskHDDByAggregate, workloadDetailAggregateUUID, workloadDetailData);
            } else if (workloadDetailResourceName != null && workloadDetailResourceName.startsWith("DISK_SSD_")) {
                //UUID format ${cluster_uuid}.${node_name}:kernel:_BACKGROUND_DELETE.DISK_SSD_${aggregate}_${aggregate uuid}
                groupDeferredWork(this.backgroundDeleteDiskSSDByAggregate, workloadDetailAggregateUUID, workloadDetailData);
            }
        } else if ("_RA_STRIPE_X".equals(workloadDetailWorkloadName)) {
            if ("CPU_dblade".equals(workloadDetailResourceName)) {
                //UUID format ${cluster_uuid}.${node_name}:kernel:_ RA_STRIPE_X.CPU_dblade
                groupDeferredWork(this.raXStripeCPUdBladeByNode, workloadDetailNodeName, workloadDetailData);
            } else if ("CPU_dblade_background".equals(workloadDetailResourceName)) {
                //UUID format ${cluster_uuid}.${node_name}:kernel:_ RA_STRIPE_X.CPU_dblade_background
                groupDeferredWork(this.raXStripeCPUdBladeByNode, workloadDetailNodeName, workloadDetailData);
            } else if (workloadDetailResourceName != null && workloadDetailResourceName.startsWith("DISK_HDD_")) {
                //UUID format ${cluster_uuid}.${node_name}:kernel:_RA_STRIPE_X.DISK_HDD_${aggregate}_${aggregate uuid}
                groupDeferredWork(this.raXStripeDiskHDDByAggregate, workloadDetailAggregateUUID, workloadDetailData);
            } else if (workloadDetailResourceName != null && workloadDetailResourceName.startsWith("DISK_SSD_")) {
                //UUID format ${cluster_uuid}.${node_name}:kernel:_RA_STRIPE_X.DISK_SSD_${aggregate}_${aggregate uuid}
                groupDeferredWork(this.raXStripeDiskSSDByAggregate, workloadDetailAggregateUUID, workloadDetailData);
            }
        }
        /****Identify originating client workload details i.e. workload details to re-attribute.****/
        else if (workloadDetailResourceName != null && workloadDetailResourceName.endsWith("CPU_dblade_background")
                && qosWorkloadUUIDToAggregateMap.containsKey(workloadDetailWorkloadUUID)) {
            //UUID format ${cluster_uuid}.${node_name}:kernel:${workload name}.CPU_dblade_background
            groupOriginatingWorkload(this.nodeToWorkloadToCPUDBladeBackgroundMap,
                    workloadDetailNodeName, workloadDetailWorkloadUUID, workloadDetailData);
        } else if (workloadDetailResourceName != null && workloadDetailResourceName.contains("DISK_HDD_")
                && qosWorkloadUUIDToAggregateMap.containsKey(workloadDetailWorkloadUUID)) {
            //UUID format ${cluster_uuid}.${node_name}:kernel:${workload name}.DISK_HDD_${aggregate}_${aggregate uuid}
            groupOriginatingWorkload(this.aggregateToWorkloadToDiskHDDMap,
                    workloadDetailAggregateUUID, workloadDetailWorkloadUUID, workloadDetailData);
        } else if (workloadDetailResourceName != null && workloadDetailResourceName.contains("DISK_SSD_")
                && qosWorkloadUUIDToAggregateMap.containsKey(workloadDetailWorkloadUUID)) {
            //UUID format ${cluster_uuid}.${node_name}:kernel:${workload name}.DISK_SSD_${aggregate}_${aggregate uuid}
            groupOriginatingWorkload(this.aggregateToWorkloadToDiskSSDMap,
                    workloadDetailAggregateUUID, workloadDetailWorkloadUUID, workloadDetailData);
        }
    }

    /**
     * Group the workload details by node name or aggregate uuid and workload uuid.
     */
    private void groupOriginatingWorkload(
            final Map<String, HashMap<String, QosWorkloadDetailData>> nodeOrAggregateToWorkloadToWorkloadDetailMap,
            final String nodeOrAggregate, final String workloadUUID, final QosWorkloadDetailData counters) {
        HashMap<String, QosWorkloadDetailData> workloadUUIDtoWorkloadDetailMap =
                nodeOrAggregateToWorkloadToWorkloadDetailMap.get(nodeOrAggregate);
        if (workloadUUIDtoWorkloadDetailMap == null) {
            workloadUUIDtoWorkloadDetailMap = new HashMap<String, QosWorkloadDetailData>();
            nodeOrAggregateToWorkloadToWorkloadDetailMap.put(nodeOrAggregate,
                    workloadUUIDtoWorkloadDetailMap);
        }
        workloadUUIDtoWorkloadDetailMap.put(workloadUUID, counters);
    }


    /**
     * Group the workload details by node name or aggregate uuid.
     */
    private void groupDeferredWork(final Map<String, Long> nodeOrAggregateToWorkloadDetailMap,
                                   final String nodeOrAggregate,
                                   final QosWorkloadDetailData workloadDetail) {
        Long work = nodeOrAggregateToWorkloadDetailMap.get(nodeOrAggregate);
        if (work == null) {
            work = (long) 0;
        }
        nodeOrAggregateToWorkloadDetailMap.put(nodeOrAggregate, (workloadDetail.getServiceTime() + work));
    }

    @Override
    public String toString() {
        final StringBuilder sb = new StringBuilder("ReAttributeDeferredWorkServiceStep1{");
        sb.append("waflCPCPUdBladeByNode=").append(waflCPCPUdBladeByNode);
        sb.append(", waflCPDiskHDDByAggregate=").append(waflCPDiskHDDByAggregate);
        sb.append(", waflCPDiskSSDByAggregate=").append(waflCPDiskSSDByAggregate);
        sb.append(", backgroundDeleteCPUdBladeByNode=").append(backgroundDeleteCPUdBladeByNode);
        sb.append(", backgroundDeleteDiskHDDByAggregate=").append(backgroundDeleteDiskHDDByAggregate);
        sb.append(", backgroundDeleteDiskSSDByAggregate=").append(backgroundDeleteDiskSSDByAggregate);
        sb.append(", raXStripeCPUdBladeByNode=").append(raXStripeCPUdBladeByNode);
        sb.append(", raXStripeDiskHDDByAggregate=").append(raXStripeDiskHDDByAggregate);
        sb.append(", raXStripeDiskSSDByAggregate=").append(raXStripeDiskSSDByAggregate);
        sb.append(", nodeToWorkloadToCPUDBladeBackgroundMap=").append(nodeToWorkloadToCPUDBladeBackgroundMap);
        sb.append(", aggregateToWorkloadToDiskHDDMap=").append(aggregateToWorkloadToDiskHDDMap);
        sb.append(", aggregateToWorkloadToDiskSSDMap=").append(aggregateToWorkloadToDiskSSDMap);
        sb.append('}');
        return sb.toString();
    }
}
