package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.reattribution.impl;

import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.reattribution.data.QosWorkloadConstituentData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.reattribution.interfaces.IReAttributionContext;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.reattribution.interfaces.IReAttributionContext.QosWorkloadNodeRelationshipAttributes;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 *  In this step we identify the deferred work in each node, update the DB, FB and RA counters for multi level re
 *  attribution and compute total DB, FB and RA in each node and aggregate.
 */
public class ReAttributeDeferredWorkServiceStep2 {

    private static final Logger logger = LoggerFactory.getLogger(ReAttributeDeferredWorkServiceStep2.class);

    private final Map<String, QosWorkloadConstituentData> waflCPByNode = new HashMap<String, QosWorkloadConstituentData>();
    private final Map<String, QosWorkloadConstituentData> backgroundDeleteByNode = new HashMap<String, QosWorkloadConstituentData>();
    private final Map<String, QosWorkloadConstituentData> raStripeXByNode = new HashMap<String, QosWorkloadConstituentData>();


    private final Map<String, Double> totalDirtyBuffersByNode = new HashMap<String, Double>();
    private final Map<String, Double> totalDirtyBuffersByAggregate = new HashMap<String, Double>();
    private final Map<String, Double> totalFreedBlocksByNode = new HashMap<String, Double>();
    private final Map<String, Double> totalFreedBlocksByAggregate = new HashMap<String, Double>();
    private final Map<String, Double> totalReadAheadXStripeByNode = new HashMap<String, Double>();
    private final Map<String, Double> totalReadAheadXStripeByAggregate = new HashMap<String, Double>();

    public ReAttributeDeferredWorkServiceStep2(IReAttributionContext reAttributionContext,
                                               List<QosWorkloadConstituentData> qosWorkloadConstituentDataList) {
        perform(reAttributionContext, qosWorkloadConstituentDataList);


    }

    public Map<String, Double> getTotalDirtyBuffersByNode() {
        return totalDirtyBuffersByNode;
    }

    public Map<String, Double> getTotalDirtyBuffersByAggregate() {
        return totalDirtyBuffersByAggregate;
    }

    public Map<String, Double> getTotalFreedBlocksByNode() {
        return totalFreedBlocksByNode;
    }

    public Map<String, Double> getTotalFreedBlocksByAggregate() {
        return totalFreedBlocksByAggregate;
    }

    public Map<String, Double> getTotalReadAheadXStripeByNode() {
        return totalReadAheadXStripeByNode;
    }

    public Map<String, Double> getTotalReadAheadXStripeByAggregate() {
        return totalReadAheadXStripeByAggregate;
    }

    private void perform(IReAttributionContext reAttributionContext,
                         List<QosWorkloadConstituentData> qosWorkloadConstituentDataList) {
        step2_1(reAttributionContext, qosWorkloadConstituentDataList);
        step2_2(reAttributionContext, qosWorkloadConstituentDataList);
        step2_3();
        step2_4(reAttributionContext, qosWorkloadConstituentDataList);

        logger.debug("Results after ReAttributeDeferredWorkServiceStep2 {}", this);
    }

    /**
     * Perform the following using workload constituents -
     * 1. Identify deferred workloads _WAFL_CP, _BACKGROUND_DELETE, _RA_STRIPE_X in each node.
     * 2. Calculate the total dirty buffers, free blocks and read ahead x stripe for each node from foreground workloads.
     */
    private void step2_1(IReAttributionContext reAttributionContext,
                         List<QosWorkloadConstituentData> qosWorkloadConstituentDataList) {
        final Map<String, Map<IReAttributionContext.QosWorkloadNodeRelationshipAttributes, String>>
                qosWorkloadNodeRelationshipMap = reAttributionContext.getQosWorkloadNodeRelationshipMap();
        for (QosWorkloadConstituentData qosWorkloadConstituentData : qosWorkloadConstituentDataList) {
            final Map<QosWorkloadNodeRelationshipAttributes, String> qosWorkloadConstituent =
                    qosWorkloadNodeRelationshipMap.get(qosWorkloadConstituentData.getUUID());
            if (qosWorkloadConstituent != null) {
                final String workloadName = qosWorkloadConstituent.get(QosWorkloadNodeRelationshipAttributes.WORKLOAD_NAME);
                final String workloadNodeName = qosWorkloadConstituent.get(QosWorkloadNodeRelationshipAttributes.NODE_NAME);
                switch (workloadName) {
                    case "_WAFL_CP":
                        this.waflCPByNode.put(workloadNodeName, qosWorkloadConstituentData);
                        break;
                    case "_BACKGROUND_DELETE":
                        this.backgroundDeleteByNode.put(workloadNodeName, qosWorkloadConstituentData);
                        break;
                    case "_RA_STRIPE_X":
                        this.raStripeXByNode.put(workloadNodeName, qosWorkloadConstituentData);
                        break;
                    default:
                        sumWorkloadCounters(totalDirtyBuffersByNode, workloadNodeName,
                                Double.valueOf(qosWorkloadConstituentData.getDirtyBuffers()));
                        sumWorkloadCounters(totalFreedBlocksByNode, workloadNodeName,
                                Double.valueOf(qosWorkloadConstituentData.getFreeBlocks()));
                        sumWorkloadCounters(totalReadAheadXStripeByNode, workloadNodeName,
                                Double.valueOf(qosWorkloadConstituentData.getReadAheadXStrips()));
                        break;
                }
            }
        }
    }

    /**
     * Multilevel Re Attribution - For each foreground workload add dirty buffers, free blocks and read ahead x stripe
     * contributions from _WAFL_CP,_BACKGROUND_DELETE and _RA_STRIPE_X.
     */
    private void step2_2(IReAttributionContext reAttributionContext,
                         List<QosWorkloadConstituentData> qosWorkloadConstituentDataList) {
        final Map<String, Map<IReAttributionContext.QosWorkloadNodeRelationshipAttributes, String>>
                qosWorkloadNodeRelationshipMap = reAttributionContext.getQosWorkloadNodeRelationshipMap();
        for (QosWorkloadConstituentData qosWorkloadConstituentData : qosWorkloadConstituentDataList) {
            final Map<QosWorkloadNodeRelationshipAttributes, String> qosWorkloadConstituent =
                    qosWorkloadNodeRelationshipMap.get(qosWorkloadConstituentData.getUUID());
            if (qosWorkloadConstituent != null) {
                final String workloadName = qosWorkloadConstituent.get(QosWorkloadNodeRelationshipAttributes.WORKLOAD_NAME);
                if ("_WAFL_CP".equals(workloadName) || "_BACKGROUND_DELETE".equals(workloadName)
                        || "_RA_STRIPE_X".equals(workloadName)) {
                    continue;
                }
                final String workloadNodeName = qosWorkloadConstituent.get(QosWorkloadNodeRelationshipAttributes.NODE_NAME);
                //Sum of DB, FB, RA in a node from all workloads except _WAFL_CP,_BACKGROUND_DELETE,_RA_STRIPE_X
                Double totalDBInNode = totalDirtyBuffersByNode.get(workloadNodeName);
                Double totalFBInNode = totalFreedBlocksByNode.get(workloadNodeName);
                Double totalRAInNode = totalReadAheadXStripeByNode.get(workloadNodeName);

                long workloadDB = qosWorkloadConstituentData.getDirtyBuffers();
                long workloadFB = qosWorkloadConstituentData.getFreeBlocks();
                long workloadRA = qosWorkloadConstituentData.getReadAheadXStrips();

                double dbDistributionRatio = 0d;
                if (workloadDB > 0 && totalDBInNode != null && totalDBInNode > 0) {
                    dbDistributionRatio = (double) workloadDB / totalDBInNode;
                }
                double fbDistributionRatio = 0d;
                if (workloadFB > 0 && totalFBInNode != null && totalFBInNode > 0) {
                    fbDistributionRatio = (double) workloadFB / totalFBInNode;
                }
                double raDistributionRatio = 0d;
                if (workloadRA > 0 && totalRAInNode != null && totalRAInNode > 0) {
                    raDistributionRatio = (double) workloadRA / totalRAInNode;
                }

                QosWorkloadConstituentData waflCP = waflCPByNode.get(workloadNodeName);
                QosWorkloadConstituentData backgroundDelete = backgroundDeleteByNode.get(workloadNodeName);
                QosWorkloadConstituentData readAheadXStripe = raStripeXByNode.get(workloadNodeName);

                if (waflCP != null) {
                    //Add dirty buffer, free blocks & read ahead strips contributions from _WAFL_CP
                    addDBFBRAContributions(qosWorkloadConstituentData, waflCP, dbDistributionRatio);
                }
                if (backgroundDelete != null) {
                    //Add dirty buffer, free blocks & read ahead strips contributions from _BACKGROUND_DELETE
                    addDBFBRAContributions(qosWorkloadConstituentData, backgroundDelete, fbDistributionRatio);
                }
                if (readAheadXStripe != null) {
                    //Add dirty buffer, free blocks & read ahead strips contributions from _RA_STRIPE_X
                    addDBFBRAContributions(qosWorkloadConstituentData, readAheadXStripe, raDistributionRatio);
                }
            }
        }
        logger.debug("Updated Workload Constituent {}" + qosWorkloadConstituentDataList);
    }

    /**
     * Add dirty buffers,free blocks and read ahead x stripe from _WAFL_CP,_BACKGROUND_DELETE and _RA_STRIPE_X to the
     * node level totals created in step 2_1.
     */
    private void step2_3() {
        addToNodeTotal(waflCPByNode);
        addToNodeTotal(backgroundDeleteByNode);
        addToNodeTotal(raStripeXByNode);
    }

    /**
     * Add dirty buffers,free blocks and read ahead x stripe from a background workload to the node level totals
     * created in step 2_1
     *
     * @param backgroundWorkloadByNode Background workload grouped by node.
     */
    private void addToNodeTotal(final Map<String, QosWorkloadConstituentData> backgroundWorkloadByNode) {
        for (Map.Entry<String, QosWorkloadConstituentData> workloadConstituentDataEntry : backgroundWorkloadByNode.entrySet()) {
            sumWorkloadCounters(totalDirtyBuffersByNode,
                    workloadConstituentDataEntry.getKey(),
                    Double.valueOf(workloadConstituentDataEntry.getValue().getDirtyBuffers()));
            sumWorkloadCounters(totalFreedBlocksByNode,
                    workloadConstituentDataEntry.getKey(),
                    Double.valueOf(workloadConstituentDataEntry.getValue().getFreeBlocks()));
            sumWorkloadCounters(totalReadAheadXStripeByNode,
                    workloadConstituentDataEntry.getKey(),
                    Double.valueOf(workloadConstituentDataEntry.getValue().getReadAheadXStrips()));
        }
    }

    /**
     * Use the total dirty buffers, total free blocks and total read ahead x stripe computed in step 2_2 and compute the
     * sum of dirty buffers, free blocks and read ahead x stripe for each aggregate.
     */
    private void step2_4(IReAttributionContext reAttributionContext,
                         List<QosWorkloadConstituentData> qosWorkloadConstituentDataList) {
        final Map<String, Map<IReAttributionContext.QosWorkloadNodeRelationshipAttributes, String>>
                qosWorkloadNodeRelationshipMap = reAttributionContext.getQosWorkloadNodeRelationshipMap();
        for (QosWorkloadConstituentData qosWorkloadConstituentData : qosWorkloadConstituentDataList) {
            final Map<QosWorkloadNodeRelationshipAttributes, String> qosWorkloadConstituent =
                    qosWorkloadNodeRelationshipMap.get(qosWorkloadConstituentData.getUUID());
            if (qosWorkloadConstituent != null) {

                final String workloadUUID = qosWorkloadConstituent.get(QosWorkloadNodeRelationshipAttributes.WORKLOAD_UUID);
                final Map<IReAttributionContext.AggregateAttributes, String> aggregate = reAttributionContext.getQosVolumeWorkloadUUIDToAggregateUUIDMap().get(workloadUUID);

                if (aggregate != null) {
                    final String aggregateUUID = aggregate.get(IReAttributionContext.AggregateAttributes.AGGREGATE_UUID);
                    sumWorkloadCounters(totalDirtyBuffersByAggregate, aggregateUUID,
                            qosWorkloadConstituentData.getTotalDirtyBuffers());
                    sumWorkloadCounters(totalFreedBlocksByAggregate, aggregateUUID,
                            qosWorkloadConstituentData.getTotalFreeBlocks());
                    sumWorkloadCounters(totalReadAheadXStripeByAggregate, aggregateUUID,
                            qosWorkloadConstituentData.getTotalReadAheadXStrips());
                }
            }
        }
    }


    /**
     * Add dirty buffer, free blocks & read ahead strips contributions from background workloads to the foreground
     * workloads.
     */
    private void addDBFBRAContributions(QosWorkloadConstituentData foregroundWork,
                                        QosWorkloadConstituentData backgroundWork, double distributionRatio) {
        foregroundWork.addTotalDirtyBuffers(backgroundWork.getDirtyBuffers() * distributionRatio);
        foregroundWork.addTotalFreeBlocks(backgroundWork.getFreeBlocks() * distributionRatio);
        foregroundWork.addTotalReadAheadXStrips(backgroundWork.getReadAheadXStrips() * distributionRatio);
    }


    private void sumWorkloadCounters(final Map<String, Double> sumByMap, final String sumByKey, final Double counter) {
        Double counterTotal = sumByMap.get(sumByKey);
        counterTotal = (counterTotal == null) ? counter : (counterTotal + counter);
        sumByMap.put(sumByKey, counterTotal);
    }

    @Override
    public String toString() {
        final StringBuilder sb = new StringBuilder("ReAttributeDeferredWorkServiceStep2{");
        sb.append("waflCPByNode=").append(waflCPByNode);
        sb.append(", backgroundDeleteByNode=").append(backgroundDeleteByNode);
        sb.append(", raStripeXByNode=").append(raStripeXByNode);
        sb.append(", totalDirtyBuffersByNode=").append(totalDirtyBuffersByNode);
        sb.append(", totalDirtyBuffersByAggregate=").append(totalDirtyBuffersByAggregate);
        sb.append(", totalFreedBlocksByNode=").append(totalFreedBlocksByNode);
        sb.append(", totalFreedBlocksByAggregate=").append(totalFreedBlocksByAggregate);
        sb.append(", totalReadAheadXStripeByNode=").append(totalReadAheadXStripeByNode);
        sb.append(", totalReadAheadXStripeByAggregate=").append(totalReadAheadXStripeByAggregate);
        sb.append('}');
        return sb.toString();
    }
}
