package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.synthetic.builders;

import com.netapp.oci.netapp.performance.nfs.interfaces.data.NfsMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.GlobalScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.synthetic.mergers.NFSMetricsMerger;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.synthetic.mergers.NFSMetricsMergerSN;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster.ClusterKey;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

import java.util.Collection;
import java.util.Map;

public class NFSMetricsBuilder extends SyntheticBuilderBase<NfsMeasurements, NfsMeasurements, Cluster.ClusterKey, NFSMetricsBuilder.NFSCounter> {

    public static enum NFSCounter {
        opm_nfs_ops,
        opm_nfs_read_ops,
        opm_nfs_write_ops,
        opm_nfs_throughput,
        opm_nfs_read_throughput,
        opm_nfs_write_throughput,
        opm_nfs_read_latency,
        opm_nfs_write_latency,
        opm_nfs_latency,
    }

    public NFSMetricsBuilder(NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage);
    }

    @Override
    protected Cluster.ClusterKey getParentInventoryUUIDKey(NetAppDefaultStorageContext context,
                                                           String inventoryUUID,
                                                           MetricsRowScratchPad<NFSCounter> metricsRowScratchPad) {
        return context.getKey();
    }

    @Override
    protected Map<String, MetricsRowScratchPad<NFSCounter>> generateMetricsRowScratchPads(
            NetAppDefaultStorageContext context, ZAPIConnection zapiConnection, GlobalScratchPad globalScratchPad) {
        if (context.isLessThanFullSteam()) {
            return globalScratchPad.merge(new NFSMetricsMergerSN(), "nfsv3", "nfsv4", "nfsv4_1");
        }
        else {
            return globalScratchPad.merge(new NFSMetricsMerger(), "nfsv3", "nfsv4", "nfsv4_1");
        }
    }


    @Override
    protected NfsMeasurements normalizeMetrics(MetricsRowScratchPad<NFSCounter> currentScratchPad,
                                               NetAppDefaultStorageContext context) {
        NfsMeasurements measurements = new NfsMeasurements();
        measurements.setOpmNfsThroughput(getNormalizedFloatMetric(NFSCounter.opm_nfs_throughput, currentScratchPad));
        measurements.setOpmNfsReadThroughput(getNormalizedFloatMetric(NFSCounter.opm_nfs_read_throughput, currentScratchPad));
        measurements.setOpmNfsWriteThroughput(getNormalizedFloatMetric(NFSCounter.opm_nfs_write_throughput, currentScratchPad));
        measurements.setOpmNfsOps(getNormalizedFloatMetric(NFSCounter.opm_nfs_ops, currentScratchPad));
        measurements.setOpmNfsReadOps(getNormalizedFloatMetric(NFSCounter.opm_nfs_read_ops, currentScratchPad));
        measurements.setOpmNfsWriteOps(getNormalizedFloatMetric(NFSCounter.opm_nfs_write_ops, currentScratchPad));
        measurements.setOpmNfsReadLatency(getNormalizedFloatMetric(NFSCounter.opm_nfs_read_latency, currentScratchPad));
        measurements.setOpmNfsWriteLatency(getNormalizedFloatMetric(NFSCounter.opm_nfs_write_latency, currentScratchPad));
        measurements.setOpmNfsLatency(getNormalizedFloatMetric(NFSCounter.opm_nfs_latency, currentScratchPad));
        return measurements;
    }

    private float getNormalizedFloatMetric(NFSCounter counterName, MetricsRowScratchPad<NFSCounter> currentValues) {
        final float normalizedMetric;

        switch (counterName) {
            case opm_nfs_ops:
            case opm_nfs_read_ops:
            case opm_nfs_write_ops:
            case opm_nfs_read_throughput:
            case opm_nfs_throughput:
            case opm_nfs_write_throughput:
                normalizedMetric = currentValues.getCounterRatePerSec(counterName);
                break;
            case opm_nfs_read_latency:
                normalizedMetric = currentValues.getLatencyPerOp(counterName, NFSCounter.opm_nfs_read_ops);
                break;
            case opm_nfs_write_latency:
                normalizedMetric = currentValues.getLatencyPerOp(counterName, NFSCounter.opm_nfs_write_ops);
                break;
            case opm_nfs_latency:
                normalizedMetric = currentValues.getLatencyPerOp(counterName, NFSCounter.opm_nfs_ops);
                break;
            default:
                normalizedMetric = MetricsRowSample.NULL_FLOAT;
                break;
        }
        return normalizedMetric;

    }

    @Override
    public void reportPerformance(Cluster.ClusterKey clusterKey,
                                  Map<Cluster.ClusterKey, Collection<RawSample<NfsMeasurements>>> measurements)
            throws DataSourceErrorException {
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<NfsMeasurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<NfsMeasurements>> nfsMeasurements = measurementEntry.getValue();
            reportingPackage.reportNfsPerformance(key, nfsMeasurements);
        }
    }

	@Override
	protected NfsMeasurements rawMetrics(
			MetricsRowScratchPad<NFSCounter> currentScratchPad,
			NetAppDefaultStorageContext context) {
		// no raw counter available
		return null;
	}

	@Override
	public void reportRawPerformance(
			ClusterKey clusterKey,
			Map<ClusterKey, Collection<RawSample<NfsMeasurements>>> rawMeasurements)
			throws DataSourceErrorException {
		// nothing to report
	}
}