package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transform.ocie;

import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.cmode.CModeAggregate;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.cmode.CModeCluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.*;

import java.util.List;
import java.util.Map;

/**
 * Generate workload details, based on the workloads and resources returned by the cluster.
 */
public class QosWorkloadDetailTransform {

    private static final String[] DYNAMIC_WORKLOAD_DETAIL_RESOURCES_PREFIXES = new String[]{"DISK_HDD_", "DISK_SSD_"};
    private static final String[] STATIC_WORKLOAD_DETAIL_RESOURCES = new String[]{"DISK_HDD_OTHER", "DISK_SSD_OTHER"};


    /**
     * For FS & above QosWorkloadDetail will be generated only for the internal workloads and for the following
     * resources -
     * DISK_HDD_*
     * DISK_HDD_OTHER
     * DISK_SSD_*
     * DISK_SSD_OTHER
     */
    static void generateQosWorkloadDetailsForFSAndAbove(final CModeCluster cModeCluster, final Cluster cluster,
                                                        final Map<String, QosServiceCenter.QosResourceKey> qosServiceCenterUUIDToKey,
                                                        final Map<String, String> partnerNodeNames) {
        for (final QosVolumeWorkload qosVolumeWorkload : cluster.getQosVolumeWorkloads().values()) {
            if (qosVolumeWorkload.getVolumeKey() == null) {
                createAddStaticWorkloadDetailsForInternalWorkload(cluster, qosVolumeWorkload, qosServiceCenterUUIDToKey);
                for (final CModeAggregate cModeAggregate : cModeCluster.getAggregates().values()) {
                    createAndAddDBladeAggregateWorkloadDetails(cluster, qosVolumeWorkload, cModeAggregate,
                            qosServiceCenterUUIDToKey, partnerNodeNames, DYNAMIC_WORKLOAD_DETAIL_RESOURCES_PREFIXES);
                }
            }
        }
    }

    /**
     * For a internal workload create the "DISK_HDD_OTHER" & DISK_SSD_OTHER workload details, one for each node in the cluster.
     */
    private static void createAddStaticWorkloadDetailsForInternalWorkload(final Cluster cluster,
                                                                          final QosVolumeWorkload qosVolumeWorkload,
                                                                          final Map<String, QosServiceCenter.QosResourceKey> qosServiceCenterUUIDToKey) {
        final String volumeWorkloadName = qosVolumeWorkload.getName();
        final QosVolumeWorkload.QosVolumeWorkloadKey qosVolumeWorkloadKey = qosVolumeWorkload.getKey();

        for (final Node node : cluster.getNodes().values()) {
            for (String resourceName : STATIC_WORKLOAD_DETAIL_RESOURCES) {
                final String workloadDetailName = volumeWorkloadName + "." + resourceName;
                final String nodeNamePrefix = node.getName() + ":kernel:";
                final String workloadDetailUUID = nodeNamePrefix + workloadDetailName;
                final String resourceUUID = nodeNamePrefix + resourceName;
                final QosServiceCenter.QosResourceKey qosResourceKey = qosServiceCenterUUIDToKey.get(resourceUUID);
                if (qosResourceKey != null) {
                    final QosWorkloadDetail workloadDetail = createQosWorkloadDetail(cluster, workloadDetailName, workloadDetailUUID,
                            qosVolumeWorkloadKey, qosResourceKey);
                    workloadDetail.setNodeNamePrefix(node.getName());
                    cluster.getQosWorkloadDetails().add(workloadDetail);
                }
            }
        }
    }

    /**
     * For the passed internal workload, create and add dBlade aggregate workload details. Two workload details will
     * be created for each aggregate in the custer, one for the node the aggregate is associated with and one
     * for the partner node.
     */
    private static void createAndAddDBladeAggregateWorkloadDetails(final Cluster cluster, final QosVolumeWorkload qosWorkload,
                                                                   final CModeAggregate cModeAggregate,
                                                                   final Map<String, QosServiceCenter.QosResourceKey> qosServiceCenterUUIDToKey,
                                                                   final Map<String, String> partnerNodeNames,
                                                                   final String[] dBladeDetailPrefixes) {
        final String aggregateName = cModeAggregate.getAggrName();
        final String aggregateUUID = cModeAggregate.getKey().getUUID();
        final List<String> aggregateNodeNames = NetAppOCIETransformCommon.getNodeNamesForAggregate(cModeAggregate, partnerNodeNames);
        for (final String nodeName : aggregateNodeNames) {
            for (final String dBladeDetailPrefix : dBladeDetailPrefixes) {
                final String workloadName = qosWorkload.getName();
                final String aggregateSuffix = aggregateName + '_' + aggregateUUID;
                final String nodePrefix = nodeName + ":kernel:";
                final String workloadDetailName = workloadName + '.' + dBladeDetailPrefix + aggregateSuffix;
                final String workloadDetailUUID = nodePrefix + workloadDetailName;
                final String serviceCenterUUID = nodePrefix + dBladeDetailPrefix + aggregateSuffix;
                final QosVolumeWorkload.QosVolumeWorkloadKey workloadKey = qosWorkload.getKey();

                final QosServiceCenter.QosResourceKey qosServiceCenterKey = qosServiceCenterUUIDToKey.get(serviceCenterUUID);
                if (qosServiceCenterKey != null) {
                    final QosWorkloadDetail workloadDetail = createQosWorkloadDetail(cluster, workloadDetailName, workloadDetailUUID,
                            workloadKey, qosServiceCenterKey);
                    workloadDetail.setNodeNamePrefix(nodeName);
                    workloadDetail.setAggregateUUIDSuffix(aggregateUUID);
                    cluster.getQosWorkloadDetails().add(workloadDetail);
                }
            }
        }
    }

    /**
     * Create and populate a QOS Workload detail.
     */
    private static QosWorkloadDetail createQosWorkloadDetail(final Cluster cluster, final String workloadDetailName,
                                                             final String workloadDetailUUID,
                                                             final QosVolumeWorkload.QosVolumeWorkloadKey volumeWorkloadKey,
                                                             final QosServiceCenter.QosResourceKey qosServiceCenterKey) {
        final QosWorkloadDetail qosWorkloadDetail = cluster.createQosWorkloadDetail(workloadDetailUUID);
        qosWorkloadDetail.setName(workloadDetailName);
        qosWorkloadDetail.setQosVolumeWorkloadKey(volumeWorkloadKey);
        qosWorkloadDetail.setQosResourceKey(qosServiceCenterKey);
        return qosWorkloadDetail;
    }
}

