package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi;

import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.config.NetAppOCIEDataSourceException;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.partialpoll.PartialPollContext;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.partialpoll.ZAPIPartialPollCache;
import com.onaro.sanscreen.acquisition.framework.datasource.*;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileWriter;
import com.onaro.sanscreen.acquisition.util.cli.TextSimulator;
import netapp.manage.*;
import org.apache.commons.lang3.StringUtils;
import org.jdom.Attribute;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.net.ssl.SSLHandshakeException;
import java.io.IOException;
import java.io.StringReader;
import java.net.SocketTimeoutException;
import java.util.*;
import java.util.concurrent.atomic.AtomicLong;

public abstract class ZAPIConnection {

  private static final Logger logger = LoggerFactory.getLogger(ZAPIConnection.class);

  //Keep an empty object in store.
  private static final Element EMPTY_ELEMENT = new Element ("EmptyElement");
  private static final ZapiResponse EMPTY_ZAPI_RESPONSE = new ZapiResponse(EMPTY_ELEMENT, 0, 0, 0, 0);

  private static final Integer MIN_RECORDS = 10;

  //todo should be private
  protected volatile String dsName;
  protected volatile Integer configuredBackoffMaxRecords;
  protected final AtomicLong totalZAPILatency = new AtomicLong(0L);
  protected volatile String targetIPAddressString;
  protected TextSimulator recording = null;
  protected TextSimulator simulator = null;
  protected final SortedSet<String> attributes = Collections.synchronizedSortedSet( new TreeSet<String>() );

  private PartialPollContext partialPollContext;

  private volatile NaServer server = null;

  public abstract DSFileWriter getResponseTimesWriter();

  public abstract void recordAttributes();

  public abstract void recordRequestedAttribute(String s);

  public abstract String getIP();

  public abstract void close() throws DataSourceErrorException;

  protected abstract boolean isDataSourceStopped();

  protected abstract int getMaxRetries();

  protected abstract boolean isSimulating();

  protected ZAPIConnection()
  {
  }

  protected ZAPIConnection(NaServer server)
  {
    this.server = server;
  }

  public String getTargetIPAddressString()
  {
    return targetIPAddressString;
  }

  public NaServer getServer()
  {
    return server;
  }

  public void setServer(NaServer server)
  {
    this.server = server;
  }

  public long getTotalZAPILatency() {
    return totalZAPILatency.get();
  }

  public PartialPollContext getPartialPollContext()
  {
    return partialPollContext;
  }

  public void setPartialPollContext(PartialPollContext partialPollContext)
  {
    this.partialPollContext = partialPollContext;
  }

  public ZapiResponse invokeCommandDesiredAttributes(final String command, final Map<String, String> parameters,
                                                     final ZapiDesiredAttributes desiredAttributes) throws DataSourceErrorException {
    return invokeElem(generateInput(command, parameters, desiredAttributes));
  }

  public ZapiResponse invokeCommand(final String command,
                                    final Map<String, String> parameters) throws DataSourceErrorException {
    return invokeCommandNestedArgs(command, parameters, null, null, null);
  }

  public ZapiResponse invokeElem(final NaElement input) throws DataSourceErrorException {
    return invokeElem(input, input.getChildByName(ZAPINames.DESIRED_ATTRIBUTES));
  }

  public ZapiResponse invokeCommandArrayArgs(final String command, final Map<String, String> parameters,
                                             final List<ZapiParamArray> zapiParamArrays) throws DataSourceErrorException {

    final StringBuilder attributePathBuilder = new StringBuilder(command);
    final NaElement input = new NaElement(command);
    if (parameters != null) {
      for (final Map.Entry<String, String> param : parameters.entrySet()) {
        attributePathBuilder.append('.').append(param.getValue());
        input.addNewChild(param.getKey(), param.getValue());
      }
    }
    final String attributePath = attributePathBuilder.append('.').toString();
    for (final ZapiParamArray zapiParamArray : zapiParamArrays) {
      final boolean isCountersArray = (!"instance-uuids".equals(zapiParamArray.parentEltName));
      final NaElement params = new NaElement(zapiParamArray.parentEltName);
      input.addChildElem(params);
      for (final String childParam : zapiParamArray.childEltValues) {
        params.addNewChild(zapiParamArray.childEltName, childParam);
        final String attribute = attributePath + childParam;
        if (isCountersArray) {
          attributes.add(attribute);
        }
      }
    }
    return invokeElem(getServer(), input);
  }

  public ZapiResponse invokeCommandNestedArgs(final String command, final Map<String, String> parameters,
                                              final String parentParamName, final String childParamName, final List<String> childParamValues)
          throws DataSourceErrorException {
    final NaElement input = new NaElement(command);

    if (parameters != null) {
      for (final Map.Entry<String, String> param : parameters.entrySet()) {
        input.addNewChild(param.getKey(), param.getValue());
      }
    }
    if (parentParamName != null) {
      final NaElement params = new NaElement(parentParamName);
      input.addChildElem(params);
      for (final String childParam : childParamValues) {
        params.addNewChild(childParamName, childParam);
      }
    }
    return invokeElem(server, input);
  }

  protected Element parseResponseXMLFromZAPIOutput(final String outputXML, final NaElement input) throws DataSourceErrorException, NaProtocolException, NaAPIFailedException
  {
    final String ip = getIP();
    final SAXBuilder builder = new SAXBuilder();
    builder.setExpandEntities(false);
    Element results = null;

    try {
      final Document responseDoc = builder.build(new StringReader(outputXML));
      results =  responseDoc.getRootElement();
    }catch (IOException e) {
      final EnhancedErrorDetail eed = new EnhancedErrorDetail("ioException.html", ip, "Communication problem with the cluster: " + ip);
      eed.addParam("IP", ip);
      eed.addParam("filerMessage", e.getMessage());
      eed.addParam("port", String.valueOf(getServer().getPort()));
      eed.addParam("command", input.getName());
      throw DataSourceErrorException.createWithEnhanced(DataSourceErrorType.CONNECTING, "Failed to connect to ONTAPI server", eed, e);
    }catch (JDOMException e) {
      throw new DataSourceErrorException(DataSourceErrorType.INCONSISTENT_DATA, "Failed to parse response", "Failed to parse response", e);
    }

    //checking from NaServer----This part of code is in NaServer's invokeElem method. Since we are not using it, we still want to have the checking.
    if(results!=null) {
            /*
               *  If the API succeeded, return the results now
            */
      Attribute status = results.getAttribute("status");
      if (status!=null && "passed".equals(status.getValue())) {
        return results;
      }
            /*
               *  Otherwise, something went wrong, and we need to build and
               *  throw an NaAPIFailedException.
            */
      int errno ;
      Attribute err = results.getAttribute("errno");
      if (null == err) {
                /*
                *  This is to handle the rare case when the errno is missing,
                *  which only occurs in Ontap 6.3, as far as I know.
                */
        errno = NaErrno.EINTERNALERROR;
      } else {
        try {
          errno = err.getIntValue();
        } catch (Exception e) {
          throw new NaProtocolException("Malformed errno code: " + err, e);
        }
      }
      Attribute reason = results.getAttribute("reason");
      if(null != reason) {
        throw new NaAPIFailedException(errno, reason.getValue());
      }
      else {
        throw new NaAPIFailedException(errno, null); //if reason attribute is null, we still need throw NaAPIFailedException
      }
    } //end of checking from NaServer
    else {
      return null;
    }

  }

  private static NaElement generateInput(final String command, final Map<String, String> parameters,
                                         final ZapiDesiredAttributes desiredAttributes) {
    final NaElement input = new NaElement(command);
    if (parameters != null) {
      for (final Map.Entry<String, String> param : parameters.entrySet()) {
        input.addNewChild(param.getKey(), param.getValue());
      }
    }
    if (desiredAttributes != null) {
      input.addChildElem(ZAPIUtils.convertDesiredAttributes(desiredAttributes.getDesiredAttributesMap()));
    }
    return input;
  }

  // Few error codes need to be inspected. Few error codes need to be inspected instead of failing acquisition.
  // This method tests all these required errors.
  public static boolean checkError(String ip, String zapiCommand, int errNo, String reason) throws DataSourceErrorException{
    final String logMessage = zapiCommand + "; errno: " + errNo + ", reason: " + reason;
    if (errNo == NaErrno.EAPIPRIVILEDGE && StringUtils.contains( reason, "Insufficient privileges" )) {
      final EnhancedErrorDetail eed = new EnhancedErrorDetail("authenticationProblem.html", ip, "Failed to authenticate with cluster: " + ip);
      eed.addParam("IP", ip);
      eed.addParam("filerMessage", reason);
      eed.addParam("command", zapiCommand);
      throw DataSourceErrorException.createWithEnhanced( DataSourceErrorType.INVALID_LOGIN_CREDENTIALS, logMessage, eed, new Throwable(reason));
    }else if (errNo == NaErrno.EAPIERROR && StringUtils.contains(reason, "database") && StringUtils.contains(reason, "is not open")) {
      // message is "database  is not open" with extra spaces, which may be fixed on ZAPI side some day.
      final EnhancedErrorDetail eed = new EnhancedErrorDetail("ioException.html", ip, "Communication problem with the cluster: " + ip);
      eed.addParam("IP", ip);
      eed.addParam("filerMessage", reason);
      eed.addParam("command", zapiCommand);
      throw DataSourceErrorException.createWithEnhanced(DataSourceErrorType.CONNECTING, logMessage, eed, new Throwable(reason));
    }else if (errNo == NaErrno.EONTAPI_ETIMEDOUT && StringUtils.contains(reason, "API did not finish on time")) {
      //  Throw exception to force a re-try
      logger.warn(ip + " ZAPI timed out on non-promoted ONTAP build: " + errNo + " " + reason);
      final EnhancedErrorDetail eed = new EnhancedErrorDetail("ioException.html", ip, "Communication problem with the cluster: " + ip);
      eed.addParam("IP", ip);
      eed.addParam("filerMessage", reason);
      eed.addParam("errorNumber", String.valueOf(errNo));
      eed.addParam("command", zapiCommand);
      throw new ZapiFailedToCompleteOnTime(eed, logMessage, new Throwable(logMessage));
    }else if (errNo == NaErrno.EFILER_NOT_HEALTHY && StringUtils.contains(reason, "Node is not healthy") || StringUtils.contains(reason, "initialize_session() returned NULL environment")) {
      // Often 13114 but not always
      final String errorMessage = ip + " " + zapiCommand + "; errno: " + errNo + ", reason: " + reason;
      final EnhancedErrorDetail eed = new EnhancedErrorDetail("ioException.html", ip, "Communication problem with the cluster: " + ip);
      eed.addParam("IP", ip);
      eed.addParam("filerMessage", reason);
      eed.addParam("errorNumber", String.valueOf(errNo));
      eed.addParam("command", zapiCommand);
      throw DataSourceErrorException.createWithEnhanced(DataSourceErrorType.RETRIEVE_DATA_FROM_DEVICE, logMessage, eed, new Throwable(errorMessage));
    }else{
      //Following the same logic as before OCIE team was following.
      //Throw an exception for few errors and ignore remaining all others.
      return false;
    }
  }

  protected ZapiResponse invokeElem(final NaServer server, final NaElement input) throws DataSourceErrorException
  {
    return invokeElem( server, new ZapiCommand( input ) );
  }

  protected ZapiResponse invokeElem(final NaServer server, final ZapiCommand command) throws DataSourceErrorException
  {
    NaElement input = command.getRequest();

    //Make sure the data source is still running before making the zapi call
    if (isDataSourceStopped()) {
      final String logAndGUIMessage = "Data source is stopped.";
      throw new DataSourceErrorException(DataSourceErrorType.RETRIEVE_DATA_FROM_DEVICE, logAndGUIMessage, logAndGUIMessage, null);
    }

    //Make sure the connection is valid to onTap or we are in a simulating mode
    if (server == null && !isSimulating()) {
      final String logAndGUIMessage = "Connection closed.";
      throw new DataSourceErrorException(DataSourceErrorType.RETRIEVE_DATA_FROM_DEVICE, logAndGUIMessage, logAndGUIMessage, null);
    }

    // Check if zapi is disabled
    Set<String> disabledZapis = ZAPIPropertiesManager.getDisableZapisSet();
    // Do not process further and return empty zapi response if zapi is disabled
    if (disabledZapis.contains(input.getName())){
      return EMPTY_ZAPI_RESPONSE;
    }

    final long start = System.currentTimeMillis();

    final String ip = getIP();
    final String inputString = input.toString();
    String outputXml = null;
    final int maxRetries = getMaxRetries();
    int invocations = 0;
    final long timeConsumed;
    Exception lastRetriedFailure = null;
    final String zapiCommand = input.getName();
    long receivedTimestamp = -1;

    boolean done = false;
    Element resultsElt = null;

    Integer backoffMaxRecords = null;
    try {
      do {

        if (invocations >= maxRetries) {

          final EnhancedErrorDetail eed = new EnhancedErrorDetail("ioException.html", ip, "Communication problem with the cluster: " + ip);
          final String napeMessage = (lastRetriedFailure == null) ? "" : lastRetriedFailure.getLocalizedMessage();
          final String logMessage = "Communication problem with the cluster: " + ip
                  + ", command: " + input.getName() + ", error: '" + napeMessage + "' on try " + invocations + " out of " + maxRetries;
          eed.addParam("IP", ip);
          eed.addParam("filerMessage", logMessage);
          eed.addParam("port", String.valueOf(server.getPort()));
          eed.addParam("command", zapiCommand);
          outputXml = "<!-- " + logMessage + " -->";
          throw DataSourceErrorException.createWithEnhanced(DataSourceErrorType.CONNECTING, logMessage, eed, lastRetriedFailure);
        }
        invocations++;
        try {
          if (isSimulating()) {
            // TextSimulator strips System.getProperty("line.separator"),
            // however NaElement.addNewChild appears to generate XML requests with
            // embedded "\n".   Need to strip these or simulation won't work.
            final String simulatedInputString = inputString.replaceAll("\n", "");
            outputXml = simulator.simulateResponse(simulatedInputString);
          }
          else {
            // fix of burt 831508: performance is slower
            receivedTimestamp = System.currentTimeMillis();

            //try using the cache first
            String zapiName = input.getName();
            try
            {
              //Try to read from the cache if the result is null then update it; Any other type of poll means to update everything
              if( !command.isIgnoreCache() && partialPollContext != null && partialPollContext.isGloballyEnabled() && partialPollContext.getPollType() == PartialPollContext.PollType.Partial)
                outputXml = ZAPIPartialPollCache.getInstance().read( partialPollContext, dsName, zapiName, inputString );
            }
            catch( Exception e )
            {
              //if anything goes wrong with creating the data from the partial poll log it and use the real zapi
              logger.error( "Partial polling failed for data source" + dsName + ". The data will be retrieved from the cluster instead", e );
            }

            //if the data was not returned from the cache call the ZAPI
            if( outputXml != null )
            {
              if( partialPollContext != null )
              {
                if( logger.isTraceEnabled() ) logger.trace( "Read from cache " + zapiName + " for " + dsName  );
                partialPollContext.addReadFromCache( zapiName );
              }
            }
            else
            {
              //read data from ontap
              if( logger.isTraceEnabled() )logger.trace( "Reading from device "  + zapiName + " for " + dsName  );
              String outputString = server.invokeElemForXML( input );
              if( partialPollContext != null )
                partialPollContext.addReadFromDevice( zapiName );

              //message result
              final int startPos = outputString.indexOf("<results");
              final int endPos = outputString.length() - 9;    // chop off closing </netapp>
              outputXml = outputString.substring(startPos, endPos);

              //save to data to the partial poll
              try
              {
                if( partialPollContext != null && partialPollContext.isGloballyEnabled() )
                {
                  boolean persisted = ZAPIPartialPollCache.getInstance().save( partialPollContext, dsName, zapiName, inputString, outputXml );

                  if( persisted && logger.isTraceEnabled() )
                      logger.trace( "Saved " + zapiName + " for " + dsName  );
                }
              }
              catch( Exception e )
              {
                //if anything goes wrong with creating the data from the partial poll log it and use the real zapi
                logger.error( "Partial polling failed for data source" + dsName + ". The data was not store in cache.", e );
              }
            }
          }

          // parse the XML
          try {
            if(outputXml != null){
              resultsElt = parseResponseXMLFromZAPIOutput(outputXml, input);
            }
            done = true;
          } catch (ZapiFailedToCompleteOnTime zfcot) {

            // We will retry if ZAPI failed to complete on time.
            // Look for max records in arguments and if it is present, reduce it
            final NaElement maxRecordsElt = input.getChildByName("max-records");
            if (maxRecordsElt != null) {
              final String maxRecordsString = maxRecordsElt.getContent();
              final Integer maxRecords = Integer.parseInt(maxRecordsString);
              final Integer reducedMaxRecords = maxRecords / 10;
              final Integer newMaxRecords = Math.max(MIN_RECORDS, reducedMaxRecords);
              maxRecordsElt.setContent(String.valueOf(newMaxRecords));
            }
            lastRetriedFailure = zfcot;
          }
        }

        // First catch exceptions that should be retried:
        catch (SocketTimeoutException ste) {
          // We will retry if socket timed out
          lastRetriedFailure = ste;
          final NaElement maxRecordsElt = input.getChildByName("max-records");
          if (maxRecordsElt != null) {
            final String maxRecordsString = maxRecordsElt.getContent();
            logger.warn(PackageContext.getLogPrefix() + " - " + dsName + " " + ip + " Socket timeout on " + zapiCommand + " try " + invocations + " out of " + maxRetries + ", max-records = " + maxRecordsString, ste);
            final Integer maxRecords = Integer.parseInt(maxRecordsString);
            if (maxRecords > configuredBackoffMaxRecords) {
              backoffMaxRecords = configuredBackoffMaxRecords;
              maxRecordsElt.setContent(String.valueOf(backoffMaxRecords));
              logger.warn(PackageContext.getLogPrefix() + " - " + dsName + " " + ip + " Reducing snapmirror-get-iter max-records to " + configuredBackoffMaxRecords);
            }
          }
        }
        catch (NaProtocolException e) {
          // Dmitry says sometimes we get intermittent NaProtocolException, and a retry will resolve the issue
          // We will retry on NaProtocolException
          logger.warn(PackageContext.getLogPrefix() + " - " + dsName + " " + ip + " - protocol error on " + zapiCommand + " try " + invocations + " out of " + maxRetries, e);
          lastRetriedFailure = e;
        }

        catch (SSLHandshakeException sslhe) {
          // SSLHandshakeException: Remote host closed connection during handshake
          // if (logger.isDebugEnabled()) {
          logger.warn(PackageContext.getLogPrefix() + " - " + dsName + " " + ip + " - SSL handshake error on " + zapiCommand + " try " + invocations + " out of "
                  + maxRetries + ", " + sslhe.getMessage(), sslhe);
          // }
          lastRetriedFailure = sslhe;
        }
        // Now catch exceptions that are not failures but not quite success.
        // Let clients deal with degenerate ZAPI response.
        catch (NaAPIFailedException e) {
          // Instance(s) for given object not found.
          // TODO: Need to check message not just return code
          if(checkError(ip, zapiCommand, e.getErrno(),e.getReason())){
            throw new NetAppOCIEDataSourceException(DataSourceErrorType.RETRIEVE_DATA_FROM_DEVICE, "ONTAP ERROR: " + e.getMessage(), null, e, e.getErrno());
          }else{
            done = true;
            logWarningIgnoreFailure(ip, zapiCommand + "; errno: " + e.getErrno() + ", reason: " + e.getReason());
          }
        }
        // Now catch exceptions that are hard failures and should not be retried
        catch (IOException e) {
          final EnhancedErrorDetail eed = new EnhancedErrorDetail("ioException.html", ip, "Communication problem with the cluster: " + ip);
          eed.addParam("IP", ip);
          eed.addParam("filerMessage", e.getMessage());
          eed.addParam("port", String.valueOf(server.getPort()));
          eed.addParam("command", zapiCommand);
          throw DataSourceErrorException.createWithEnhanced(DataSourceErrorType.CONNECTING, "Failed to connect to the cluster.", eed, e);
        }
        catch (NaAuthenticationException e) {
          final EnhancedErrorDetail eed = new EnhancedErrorDetail("authenticationProblem.html", ip, "Failed to log in to the cluster: " + ip);
          eed.addParam("IP", ip);
          eed.addParam("port", String.valueOf(server.getPort()));
          eed.addParam("filerMessage", e.getMessage());
          eed.addParam("command", zapiCommand);
          throw DataSourceErrorException.createWithEnhanced(DataSourceErrorType.INVALID_LOGIN_CREDENTIALS, "Failed to login to the cluster.", eed, e);
        }
      }
      while (!done);
    }
    finally {
      timeConsumed = System.currentTimeMillis() - start;
      totalZAPILatency.addAndGet(timeConsumed);
      // Don't try to record the request if the datasource has been stopped mid-poll because RecordingSession
      // is actively closing & deleting ds files.  We want to avoid IOException due to Stream closed.
      if (!isDataSourceStopped()) {
        try {
          if( recording != null )
          recording.recordRequest(inputString, outputXml, timeConsumed);
        }
        catch (DataSourceErrorException e) {
          logger.error(PackageContext.getLogPrefix() + " - failed to record request for output : " + outputXml, e);
        }
      }
      //If resultsElt is null, send an empty JDOM element. This prevents null pointers down the stack.
      if(resultsElt == null)
        resultsElt = EMPTY_ELEMENT;
    }
    return new ZapiResponse(resultsElt, timeConsumed, invocations, receivedTimestamp, backoffMaxRecords);
  }

  public ZapiResponse invokeElem(final NaElement input, final NaElement attributesElem) throws DataSourceErrorException
  {
    return invokeElem( new ZapiCommand( input ), attributesElem );
  }

  public ZapiResponse invokeElem(final ZapiCommand command, final NaElement attributesElem) throws DataSourceErrorException {
    // Save attributes element to attributes set for logging purposes
    saveAttributes( attributesElem, command.getRequest().getName() );
    // Invoke request
    return invokeElem(getServer(), command);
  }

  /**
   * Concatenate two strings with a separator in-between.
   */
  private static String concat(final String str1, final char separator, final String str2) {
    final StringBuilder buf = new StringBuilder();
    if (str1 != null) {
      buf.append(str1);
    }
    if ((str2 != null) && !str2.isEmpty()) {
      if (buf.length() > 0) {
        buf.append(separator);
      }
      buf.append(str2);
    }
    return buf.toString();
  }

  /**
   * Save the specified desired attributes element/child to {@link #attributes} set.
   *
   * @param element Desired attributes element/child.
   * @param path    Path to desired attributes element/child.
   */
  @SuppressWarnings("unchecked")
  private void saveAttributes(final NaElement element, final String path) {
    if (element != null) {
      if (element.hasChildren()) {
        for (final NaElement child : (List<NaElement>)element.getChildren()) {
          saveAttributes(child, concat(path, '.', child.getName()));
        }
      }
      else { // Leaf element
        if (path.length() > 0) {
          attributes.add(path);
        }
      }
    }
  }

  private void logWarningIgnoreFailure(final String ip, final String logMessage) {
    if (logger.isDebugEnabled()) {
      logger.warn(PackageContext.getLogPrefix() + " - " + dsName + " " + ip + " - Ignoring failed zapi: " + logMessage);
    }
  }

  public ZapiResponse execute(ZapiCommand zapiCommand) throws DataSourceErrorException
  {
    return invokeElem( zapiCommand, zapiCommand.getRequest().getChildByName(ZAPINames.DESIRED_ATTRIBUTES) );
  }

  public static final class ZapiFailedToCompleteOnTime extends DataSourceErrorException {
    private ZapiFailedToCompleteOnTime(final EnhancedErrorDetail enhanceErrorDetail,
                                       final String logMessage, final Throwable cause) {
      super(DataSourceErrorType.TIMEOUT_ERROR, enhanceErrorDetail, logMessage, cause);
    }
  }

  public static final class ZapiResponse {
    private final Element element;
    private final long totalTime;
    private final int invocations;
    private final long receivedTimestamp;
    private final Integer maxRecordsOverride;

    protected ZapiResponse(final Element element, final long totalTime, final int invocations, final long receivedTimestamp,
                         final Integer maxRecordsOverride) {
      this.element = element;
      this.totalTime = totalTime;
      this.invocations = invocations;
      this.receivedTimestamp = receivedTimestamp;
      this.maxRecordsOverride = maxRecordsOverride;
    }

    public Element getElement() {
      return element;
    }

    public long getTotalTime() {
      return totalTime;
    }

    public int getInvocationCount() {
      return invocations;
    }

    public long getReceivedTimeStamp() {
      return receivedTimestamp;
    }

    public Integer getMaxRecordsOverride() {
      return maxRecordsOverride;
    }
  }

  public static class ZapiDesiredAttributes {
    public static final String DESIRED_ATTRIBUTES = "desired-attributes";
    @SuppressWarnings({"unused declaration"})
    private final String instanceEltName;
    private final LinkedHashMap<String[], String[]> desiredAttributesMap;

    public ZapiDesiredAttributes(final String instanceEltName, final LinkedHashMap<String[], String[]> desiredAttributesMap) {
      this.instanceEltName = instanceEltName;
      this.desiredAttributesMap = desiredAttributesMap;
    }

    public LinkedHashMap<String[], String[]> getDesiredAttributesMap() {
      return new LinkedHashMap<String[], String[]>(desiredAttributesMap);
    }
  }

  public static class ZapiParamArray {
    final String parentEltName;
    final String childEltName;
    final List<String> childEltValues;

    public ZapiParamArray(final String parentEltName, final String childEltName, final List<String> childEltValues) {
      this.parentEltName = parentEltName;
      this.childEltName = childEltName;
      this.childEltValues = childEltValues;
    }
  }
}
