package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi;

import netapp.manage.NaElement;
import org.apache.commons.lang3.StringUtils;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

/**
 * Collection of static ZAPI-related utility methods.
 */
public class ZAPIUtils {

    public static NaElement elem(final String name) {
        return elem(name, null, (NaElement[])null);
    }

    /**
     * Create an XML element initialized with the specified name and child.
     * Needed to avoid ambiguity with {@link #elem(String, Object)}.
     *
     * @param name XML element name.
     * @param child XML element child.
     * @return The newly created XML element.
     */
    public static NaElement elem(final String name, final NaElement child) {
        return elem(name, null, new NaElement[] { child });
    }

    public static NaElement elem(final String name, final NaElement... children) {
        return elem(name, null, children);
    }

    public static NaElement elem(final String name, final Object content) {
        return elem(name, content != null ? content.toString() : null, (NaElement[])null);
    }

    /**
     * Build an XML element with the specified name, content and children.
     *
     * @param name     The XML node name.
     * @param content  The XML node content. May be <tt>null</tt>.
     * @param children Child elements. May be <tt>null</tt>.
     * @return A new XML element.
     */
    public static NaElement elem(final String name, final String content, final NaElement... children) {
        final NaElement elem = new NaElement(name, content);
        if (children != null) {
            for (final NaElement child : children) {
                elem.addChildElem(child);
            }
        }
        return elem;
    }

    /**
     * Convert the specified values to an array of XML nodes.
     * Simplifies syntax when defining child nodes of mixed types.
     *
     * @param children Arguments may be of type {@link netapp.manage.NaElement}, {@link netapp.manage.NaElement}[], {@link String}, {@link String}[]
     *                 or {@link java.util.Map}&lt;{@link String}, {@link String}&gt;.
     *                 Arguments of type {@link Object} are converted to {@link String}.
     *                 Arguments of type {@link Object}[] are converted to {@link String}[].
     * @return An array of XML nodes.
     */
    @SuppressWarnings("unchecked")
    public static NaElement[] children(final Object... children) {
        final List<NaElement> result = new ArrayList<NaElement>();
        for (final Object obj : children) {
            if (obj.getClass().isArray()) {
                if (obj instanceof NaElement[]) {
                    Collections.addAll(result, (NaElement[])obj);
                }
                else {
                    for (final Object child : (Object[])obj) {
                        result.add(new NaElement(child.toString()));
                    }
                }
            }
            else {
                if (obj instanceof NaElement) {
                    result.add((NaElement)obj);
                }
                else if (obj instanceof Map) {
                    final Map<String, Object> map = (Map<String, Object>)obj;
                    for (final Map.Entry<String, Object> entry : map.entrySet()) {
                        result.add(elem(entry.getKey(), entry.getValue()));
                    }
                }
                else {
                    result.add(new NaElement(obj.toString()));
                }
            }
        }
        return result.toArray(new NaElement[result.size()]);
    }

    /**
     * Convert the specified map of desired attributes to an XML element.
     *
     * @param desiredAttributesMap The map of desired attributes.
     * @return The XML element of type <tt>desired-attributes</tt>.
     */
    public static NaElement convertDesiredAttributes(final Map<String[], String[]> desiredAttributesMap) {
        final NaElement desiredAttrsElem = new NaElement(ZAPIConnection.ZapiDesiredAttributes.DESIRED_ATTRIBUTES);
        for (final Map.Entry<String[], String[]> desiredAttributesEntry : desiredAttributesMap.entrySet()) {
            final String[] elementPath = desiredAttributesEntry.getKey();
            NaElement parentElem = desiredAttrsElem;
            for (final String elementName : elementPath) {
                // Does this element exist yet?
                NaElement childElem = parentElem.getChildByName(elementName);
                if (childElem == null) {
                    childElem = new NaElement(elementName);
                    parentElem.addChildElem(childElem);
                }
                parentElem = childElem;
            }
            final String[] desiredAttributeNames = desiredAttributesEntry.getValue();
            for (final String desiredAttributeName : desiredAttributeNames) {
                final NaElement desiredAttributeNameElem = new NaElement(desiredAttributeName);
                parentElem.addChildElem(desiredAttributeNameElem);
            }
        }
        return desiredAttrsElem;
    }

  /**
   * Until method for creating pipe delimited strings used when query zapis by name
   *
   * @param zapiNames of zapi names
   *
   * @return a pipe delimited string of zapi names
   */
  public static String formattedZapiNameList(List<String> zapiNames)
  {
    return StringUtils.join( zapiNames, "|" );
  }
}
