package com.onaro.sanscreen.acquisition.diskmodelutils;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.DiskModelDescription;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DiskModelUtil;

import java.io.*;
import java.util.*;

public final class DiskModelMerger {

    private static final String MERGED_FILE_NAME_FORMAT = "merged_disk_models_%1$d.txt";

    private final Map<InputStream, File> sourceInputStreams = new LinkedHashMap<InputStream, java.io.File>();

    public DiskModelMerger(final String[] diskFileNames) throws Exception {
        this(getFilesFromNames(diskFileNames));
    }

    public DiskModelMerger(final File[] diskFiles) throws Exception {
        for (final File diskFile : diskFiles) {
            if (!diskFile.exists()) {
                throw new FileNotFoundException("file not found: " + diskFile.getAbsolutePath());
            }
            if (!diskFile.canRead()) {
                throw new FileNotFoundException("file can not be read: " + diskFile.getAbsolutePath());
            }
            sourceInputStreams.put(new FileInputStream(diskFile), diskFile);
        }
        if (sourceInputStreams.isEmpty()) {
            throw new Exception("no files specified");
        }
    }

    public DiskModelMerger(final InputStream[] diskFileStreams) throws Exception {
        for (final InputStream is : diskFileStreams) {
            if (is != null) {
                sourceInputStreams.put(is, null);
            }
        }
        if (sourceInputStreams.isEmpty()) {
            throw new Exception("no files specified");
        }
    }

    private static File[] getFilesFromNames(final String[] fileNames) {
        if (fileNames == null) {
            return new File[0];
        }
        final File[] sourceFiles = new File[fileNames.length];
        for (int i = 0; i < fileNames.length; i++) {
            sourceFiles[i] = new File(fileNames[i]);
        }
        return sourceFiles;
    }

    public static void main(final String[] args) {
        if (args.length < 1) {
            usage();
            System.exit(1);
        }

        try {
            new DiskModelMerger(args).merge(true);
        }
        catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    public Map<Integer, File> merge(final boolean suppressExceptions) throws IOException {

        final Map<Integer, File> results = new TreeMap<Integer, File>();

        final Map<String, DiskModelAttributes> allDiskModels = new HashMap<String, DiskModelAttributes>();
        final Map<String, Map<String, DiskModelAttributes>> allModelsByVendor = new TreeMap<String, Map<String, DiskModelAttributes>>();
        final Map<String, Set<String>> vendorsComments = new TreeMap<String, Set<String>>();

        for (Map.Entry<InputStream, File> entry : sourceInputStreams.entrySet()) {
            try {
                final Map<String, DiskModelAttributes> diskModelDescrs = parseDiskModelDescriptionStream(entry.getKey(), entry.getValue(), suppressExceptions);

                if (allDiskModels.size() == 0) {
                    System.out.println("Validating " + diskModelDescrs.size() + " new descriptions.");
                }
                else {
                    System.out.println("Validating " + diskModelDescrs.size() + " new descriptions against existing " + allDiskModels.size());
                }

                for (DiskModelAttributes modelDescr : diskModelDescrs.values()) {
                    if (allDiskModels.containsKey(modelDescr.key)) {
                        final DiskModelAttributes lastModelDescr = allDiskModels.get(modelDescr.key);
                        try {
                            modelDescr = mergeDiskModelDescriptions(lastModelDescr, modelDescr);
                        }
                        catch (IllegalArgumentException iae) {
                            if (suppressExceptions) {
                                System.err.println("ERROR: " + modelDescr.key + ": " + iae.getLocalizedMessage());
                                modelDescr = lastModelDescr;
                            }
                            else {
                                throw iae;
                            }
                        }
                    }

                    Map<String, DiskModelAttributes> vendorsDiskModels = allModelsByVendor.get(modelDescr.vendor);
                    if (vendorsDiskModels == null) {
                        vendorsDiskModels = new TreeMap<String, DiskModelAttributes>();
                        allModelsByVendor.put(modelDescr.vendor, vendorsDiskModels);
                    }

                    Set<String> vendorComments = vendorsComments.get(modelDescr.vendor);
                    if (vendorComments == null) {
                        vendorComments = new TreeSet<String>();
                        vendorsComments.put(modelDescr.vendor, vendorComments);
                    }
                    allDiskModels.put(modelDescr.key, modelDescr);
                    vendorsDiskModels.put(modelDescr.key, modelDescr);
                    diskModelDescrs.put(modelDescr.key, modelDescr);
                }

                // allModelsByStream.put(fileName, diskModelDescrs);
            }
            catch (FileNotFoundException fnfe) {
                if (suppressExceptions) {
                    System.err.println("ERROR: " + fnfe.getLocalizedMessage());
                }else{
                    throw fnfe;
                }
            }
            catch (IOException ioe) {
                if (suppressExceptions) {
                    System.err.println("ERROR: " + ioe.getLocalizedMessage());
                }else{
                    throw ioe;
                }
            }
        }

        for (final Map.Entry<Integer, String> entry : DiskModelUtil.getAllBanners().entrySet()) {
            final Integer version = entry.getKey();
            final String fileName = String.format(MERGED_FILE_NAME_FORMAT, version);
            results.put(version, new File(fileName));
            BufferedWriter writer = null;
            try {
                writer = new BufferedWriter(new FileWriter(fileName));
                writer.write(entry.getValue());
                for (final Map.Entry<String, Map<String, DiskModelAttributes>> vendorEntry : allModelsByVendor.entrySet()) {
                    final String vendorName = vendorEntry.getKey();
                    final String vendorHeader = getVendorHeader(vendorName);
                    final Map<String, DiskModelAttributes> vendorsDiskModels = vendorEntry.getValue();
                    writer.write(vendorHeader);
                    for (final DiskModelAttributes diskModel : vendorsDiskModels.values()) {
                        writer.write(diskModel.to60CSV());
                    }
                }
                System.out.println("Wrote " + allDiskModels.size() + " model descriptions.");
            }
            catch (IOException ioe) {
                throw new IOException("ERROR: Failed to create " + fileName + ": " + ioe.getLocalizedMessage(), ioe);
            }
            finally {
                if (writer != null) {
                    try {
                        writer.close();
                    }
                    catch (IOException ignore) {
                    }
                }
            }
        }
        return results;
    }

    public static void usage() {
        System.out.println("\nUsage: <command> disk_models_file [disk_models_file...]");
        System.out.println("\tArguments: are one or more files of disk model descriptions.");
        System.out.println("\tProduces: error messages, or merged disk model description files in 5.1 and 6.0 formats");
    }

    private static Map<String, DiskModelAttributes> parseDiskModelDescriptionStream(final InputStream inputStream, final File file, final boolean suppressExceptions) throws IOException, IllegalArgumentException {
        final Map<String, DiskModelAttributes> diskModels = new TreeMap<String, DiskModelAttributes>();
        final String fileName = file != null ? file.getAbsolutePath() : "stream";

        int lineNumber = 0;
        BufferedReader modelDescriptionReader = null;
        boolean afterBanner = false;
        boolean inVendorHeader = false;
        try {
            modelDescriptionReader = new BufferedReader(new InputStreamReader(inputStream, "utf-8"));

            ArrayList<String> comments = new ArrayList<String>(5);
            while (true) {
                final String modelDescriptionLine = modelDescriptionReader.readLine();
                if (null == modelDescriptionLine) {
                    break;
                }
                ++lineNumber;

                if (modelDescriptionLine.trim().equals("")) {
                    afterBanner = true;
                    inVendorHeader = false;
                    continue;
                }

                if (modelDescriptionLine.startsWith("##")) {
                    inVendorHeader = true;
                    continue;
                }

                if (modelDescriptionLine.startsWith("#")) {
                    if ((afterBanner) && (!inVendorHeader)) {
                        comments.add(modelDescriptionLine);
                    }
                    continue;
                }

                final int commaCount = countCommas(modelDescriptionLine);
                DiskModelAttributes modelAttributes;
                if (commaCount == 9) {
                    try {
                        modelAttributes = parse60DiskModelCSVAttributes(fileName, lineNumber, modelDescriptionLine);
                    }
                    catch (Exception e) {
                        System.err.println(e.getMessage());
                        continue;
                    }
                }
                else {
                    final String msg = "ERROR: " + fileName + " line " + lineNumber + " unexpected arg count: " + (commaCount + 1);
                    if (suppressExceptions) {
                        System.err.println(msg);
                        continue;
                    }
                    else {
                        throw new IllegalArgumentException(msg);
                    }
                }
                modelAttributes.comments.addAll(comments);
                comments = new ArrayList<String>(5);

                if (diskModels.containsKey(modelAttributes.key)) {
                    final DiskModelAttributes lastModelDescr = diskModels.get(modelAttributes.key);
                    try {
                        modelAttributes = mergeDiskModelDescriptions(lastModelDescr, modelAttributes);
                    }
                    catch (IllegalArgumentException iae) {
                        final String msg = "ERROR: " + fileName + " contains conflicting entries for: " + modelAttributes.key + ": " + iae.getLocalizedMessage();
                        if (suppressExceptions) {
                            System.err.println(msg);
                        }
                        else {
                            throw new IllegalArgumentException(msg, iae);
                        }
                    }

                }

                diskModels.put(modelAttributes.key.toUpperCase(), modelAttributes);
            }
        }
        finally {
            if (null != modelDescriptionReader) {
                try {
                    modelDescriptionReader.close();
                }
                catch (IOException ignore) {
                }
            }
        }
        System.out.println("Parsed " + diskModels.size() + " from " + fileName);
        if (!suppressExceptions && diskModels.size() == 0){
            throw new IllegalArgumentException("no disk models found");
        }
        return diskModels;
    }

    private static DiskModelAttributes parse60DiskModelCSVAttributes(final String streamName, final int lineNumber, final String csvDiskModelDescription) throws IllegalArgumentException {

        final int EXPECTED_ATTRIBUTE_COUNT = 10;
        final int MANDATORY_ATTRIBUTE_COUNT = 6;
        final String[] diskParameters = csvDiskModelDescription.split(",");
        if (diskParameters.length > 10) {
            final String errorMessage = "Failed to parse " + streamName + " line " + lineNumber + ", expected " + EXPECTED_ATTRIBUTE_COUNT + " attributes, found: " + diskParameters.length;
            throw new IllegalArgumentException(errorMessage);
        }

        for (int i = 0; i < 6; ++i) {
            if (diskParameters[i].trim().length() == 0) {
                final String errorMessage = "Failed to parse " + streamName + " line " + lineNumber + ", expected " + MANDATORY_ATTRIBUTE_COUNT + " non-empty attributes.";
                throw new IllegalArgumentException(errorMessage);
            }
        }
        try {
            final String modelKey = diskParameters[0].trim().toUpperCase();
            final String vendor = diskParameters[1].trim();
            final String model = diskParameters[2].trim().toUpperCase();

            final int speed = Integer.parseInt(diskParameters[3].trim());
            final Long capacity = Long.parseLong(diskParameters[4].trim());
            final String interfaceType = diskParameters[5].trim();

            final String seekTimeStr = (diskParameters.length > 6) ? diskParameters[6].trim() : "";
            final Float seekTime = (seekTimeStr.length() > 0) ? Float.parseFloat(seekTimeStr) : null;

            final String potentialTransferRateStr = (diskParameters.length > 7) ? diskParameters[7].trim() : "";
            final Float potentialTransferRate = (potentialTransferRateStr.length() > 0) ? Float.parseFloat(potentialTransferRateStr) : null;

            final String interfaceSpeedStr = (diskParameters.length > 8) ? diskParameters[8].trim() : "";
            final Float interfaceSpeed = (interfaceSpeedStr.length() > 0) ? Float.parseFloat(interfaceSpeedStr) : null;

            final String extraInfo = (diskParameters.length > 9) ? diskParameters[9].trim() : "";
            return new DiskModelAttributes(modelKey, vendor, model, speed, capacity, interfaceType, seekTime, potentialTransferRate, interfaceSpeed, extraInfo);
        }
        catch (NumberFormatException nfe) {
            final String errorMessage = "Failed to parse " + streamName + " line " + lineNumber + ": " + nfe.getLocalizedMessage();
            throw new IllegalArgumentException(errorMessage, nfe);
        }
    }

    private static int countCommas(final String searchTarget) {
        int c = 0;
        if (searchTarget != null) {
            for (int i = 0; i < searchTarget.length(); ++i) {
                if (',' == searchTarget.charAt(i)) {
                    ++c;
                }
            }
        }

        return c;
    }

    private static DiskModelAttributes mergeDiskModelDescriptions(final DiskModelAttributes origDescr, final DiskModelAttributes newDescr) {
        if (origDescr.equals(newDescr)) {
            origDescr.comments.addAll(newDescr.comments);
            return origDescr;
        }

        final List<String> errors = new ArrayList<String>();
        final DiskModelAttributes mergedDescr = new DiskModelAttributes(null);

        mergedDescr.key = origDescr.key;
        mergedDescr.model = origDescr.model;
        if (!origDescr.model.equals(newDescr.model)) {
            errors.add("model doesn't match: " + origDescr.model + ", " + newDescr.model);
        }

        mergedDescr.vendor = origDescr.vendor;
        if (!origDescr.vendor.equals(newDescr.vendor)) {
            errors.add("vendor doesn't match: " + origDescr.vendor + ", " + newDescr.vendor);
        }

        mergedDescr.rpm = origDescr.rpm;
        if (!origDescr.rpm.equals(newDescr.rpm)) {
            errors.add("rpm doesn't match: " + origDescr.rpm + ", " + newDescr.rpm);
        }

        mergedDescr.size = origDescr.size;
        if (!origDescr.size.equals(newDescr.size)) {
            errors.add("size doesn't match: " + origDescr.size + ", " + newDescr.size);
        }

        mergedDescr.interfaceType = origDescr.interfaceType;
        if (!origDescr.interfaceType.equals(newDescr.interfaceType)) {
            errors.add("interfaceType doesn't match: " + origDescr.interfaceType + ", " + newDescr.interfaceType);
        }

        if (origDescr.interfaceSpeed != null) {
            if ((newDescr.interfaceSpeed != null) && (!origDescr.interfaceSpeed.equals(newDescr.interfaceSpeed))) {
                errors.add("interfaceSpeed doesn't match: " + origDescr.interfaceSpeed + ", " + newDescr.interfaceSpeed);
            }
            else {
                mergedDescr.interfaceSpeed = origDescr.interfaceSpeed;
            }
        }
        else if (newDescr.interfaceSpeed != null) {
            mergedDescr.interfaceSpeed = newDescr.interfaceSpeed;
        }

        if (origDescr.seekTime != null) {
            if ((newDescr.seekTime != null) && (!origDescr.seekTime.equals(newDescr.seekTime))) {
                errors.add("interfaceSpeed doesn't match: " + origDescr.seekTime + ", " + newDescr.seekTime);
            }
            else {
                mergedDescr.seekTime = origDescr.seekTime;
            }
        }
        else if (newDescr.seekTime != null) {
            mergedDescr.seekTime = newDescr.seekTime;
        }

        if (origDescr.potentialTransferRate != null) {
            if ((newDescr.potentialTransferRate != null) && (!origDescr.potentialTransferRate.equals(newDescr.potentialTransferRate))) {

                errors.add("potentialTransferRate doesn't match: " + origDescr.potentialTransferRate + ", " + newDescr.potentialTransferRate);
            }
            else {
                mergedDescr.potentialTransferRate = origDescr.potentialTransferRate;
            }
        }
        else if (newDescr.potentialTransferRate != null) {
            mergedDescr.potentialTransferRate = newDescr.potentialTransferRate;
        }

        mergedDescr.extraInfo = origDescr.extraInfo.trim();
        if (newDescr.extraInfo != null && newDescr.extraInfo.length() > 0) {
            if (mergedDescr.extraInfo.length() > 0) {
                mergedDescr.extraInfo += " ";
            }
            mergedDescr.extraInfo += newDescr.extraInfo;
        }

        mergedDescr.comments.addAll(origDescr.comments);
        mergedDescr.comments.addAll(newDescr.comments);

        if (errors.size() > 0) {
            final StringBuilder errorBuffer = new StringBuilder(errors.size() * 80);
            for (final String error : errors) {
                if (errorBuffer.length() > 0) {
                    errorBuffer.append(", ");
                }
                errorBuffer.append(error);
            }
            throw new IllegalArgumentException(errorBuffer.toString());
        }

        return mergedDescr;
    }

    private static String getVendorHeader(final String vendor) {
        final int vendorLen = vendor.length();
        final StringBuilder stringBuffer = new StringBuilder();
        stringBuffer.append("\n## ");
        for (int i = 0; i < vendorLen; ++i) {
            stringBuffer.append("-");
        }
        stringBuffer.append("\n#  ");
        stringBuffer.append(vendor);
        stringBuffer.append("\n#  ");
        for (int i = 0; i < vendorLen; ++i) {
            stringBuffer.append("-");
        }
        stringBuffer.append("\n");
        return stringBuffer.toString();
    }

    private static final class DiskModelAttributes {
        private String key;
        private String vendor;
        private String model;
        private Integer rpm;
        private Long size;
        private String interfaceType;
        private Float seekTime;
        private Float potentialTransferRate;
        private Float interfaceSpeed;
        private String extraInfo;
        // private String streamName;
        // private Integer lineNum;
        final private ArrayList<String> comments = new ArrayList<String>();

        private DiskModelAttributes(final String key, final String vendor, final String model, final Integer rpm, final Long size, final String interfaceType,
                                    final Float seekTime, final Float potentialTransferRate, final Float interfaceSpeed, final String extraInfo) {
            this.key = key;
            this.vendor = vendor;
            this.model = model;
            this.rpm = rpm;
            this.size = size;
            this.seekTime = seekTime;
            this.potentialTransferRate = potentialTransferRate;
            this.interfaceSpeed = interfaceSpeed;
            this.interfaceType = interfaceType;
            this.extraInfo = extraInfo;

            sanitizeAttributes();
        }

        private DiskModelAttributes(final DiskModelDescription diskModelDescr) {
            if (diskModelDescr != null) {
                this.key = diskModelDescr.getKey();
                this.vendor = diskModelDescr.getVendor();
                this.model = diskModelDescr.getModel();
                this.rpm = diskModelDescr.getRpm();
                this.size = diskModelDescr.getSizeGiB();
                this.seekTime = diskModelDescr.getSeekTime();
                this.potentialTransferRate = diskModelDescr.getPotentialTransferRate();
                this.interfaceSpeed = diskModelDescr.getInterfaceSpeedInternal();
                this.interfaceType = diskModelDescr.getInterfaceType();
                this.extraInfo = diskModelDescr.getExtraInfo();
            }
        }

        public int hashCode() {
            int result = 19;
            result += this.model.hashCode();
            result += this.vendor.hashCode();
            result += this.rpm;
            result += this.size.hashCode();
            result += this.interfaceType.hashCode();
            if (this.interfaceSpeed != null) {
                result += this.interfaceSpeed.hashCode();
            }
            if (this.seekTime != null) {
                result += Float.floatToIntBits(this.seekTime);
            }
            if (this.potentialTransferRate != null) {
                result += Float.floatToIntBits(this.potentialTransferRate);
            }
            if (this.extraInfo != null) {
                result += this.extraInfo.hashCode();
            }
            return result;
        }

        public boolean equals(final Object o) {
            if (o == this) {
                return true;
            }
            if (!(o instanceof DiskModelAttributes)) {
                return false;
            }
            final DiskModelAttributes dmd = (DiskModelAttributes) o;
            if (!this.model.equals(dmd.model)) {
                return false;
            }
            if (!this.vendor.equals(dmd.vendor)) {
                return false;
            }
            if (!this.rpm.equals(dmd.rpm)) {
                return false;
            }
            if (!this.size.equals(dmd.size)) {
                return false;
            }
            if (!this.interfaceType.equals(dmd.interfaceType)) {
                return false;
            }
            if (this.interfaceSpeed == null) {
                if (dmd.interfaceSpeed != null) {
                    return false;
                }
            }
            else {
                if (dmd.interfaceSpeed == null) {
                    return false;
                }
                if (!this.interfaceSpeed.equals(dmd.interfaceSpeed)) {
                    return false;
                }
            }
            if (this.seekTime == null) {
                if (dmd.seekTime != null) {
                    return false;
                }
            }
            else {
                if (dmd.seekTime == null) {
                    return false;
                }
                if (!this.seekTime.equals(dmd.seekTime)) {
                    return false;
                }
            }
            if (this.potentialTransferRate == null) {
                if (dmd.potentialTransferRate != null) {
                    return false;
                }
            }
            else {
                if (dmd.potentialTransferRate == null) {
                    return false;
                }
                if (!this.potentialTransferRate.equals(dmd.potentialTransferRate)) {
                    return false;
                }
            }
            if (this.extraInfo == null) {
                if (dmd.extraInfo != null) {
                    return false;
                }
            }
            else {
                if (dmd.extraInfo == null) {
                    return false;
                }
                if (!this.extraInfo.equals(dmd.extraInfo)) {
                    return false;
                }
            }
            return true;
        }

        public String to60CSV() {
            final StringBuilder csvBuffer = new StringBuilder(255);
            for (final String comment : this.comments) {
                csvBuffer.append(comment).append("\n");
            }
            csvBuffer.append(this.key.toUpperCase()).append(",");
            csvBuffer.append(this.vendor).append(",");
            csvBuffer.append(this.model.toUpperCase()).append(",");
            csvBuffer.append(this.rpm).append(",");
            csvBuffer.append(this.size).append(",");
            csvBuffer.append(this.interfaceType).append(",");
            csvBuffer.append((null == this.seekTime) ? "" : this.seekTime).append(",");
            csvBuffer.append((null == this.potentialTransferRate) ? "" : this.potentialTransferRate).append(",");
            csvBuffer.append((null == this.interfaceSpeed) ? "" : this.interfaceSpeed).append(",");
            csvBuffer.append((null == this.extraInfo) ? "" : this.extraInfo);
            csvBuffer.append("\n");
            return csvBuffer.toString();
        }

        private void sanitizeAttributes() {

            final Map<String, DiskType> diskTypeCorrectionsMap = new HashMap<String, DiskType>();
            diskTypeCorrectionsMap.put("SCSI", DiskType.FC);
            diskTypeCorrectionsMap.put("PATA", DiskType.ATA);

            final Map<DiskType, Float> defaultInterfaceSpeeds = new HashMap<DiskType, Float>();
            defaultInterfaceSpeeds.put(DiskType.ATA, 100f);        // Ultra ATA/100
            defaultInterfaceSpeeds.put(DiskType.FATA, 100f);       // ATA is the bottle neck
            defaultInterfaceSpeeds.put(DiskType.SATA, 187.5f);
            defaultInterfaceSpeeds.put(DiskType.SATA2, 375f);
            defaultInterfaceSpeeds.put(DiskType.SATA3, 750f);
            defaultInterfaceSpeeds.put(DiskType.SAS, 300f);
            defaultInterfaceSpeeds.put(DiskType.FC, 106.25f);
            defaultInterfaceSpeeds.put(DiskType.SSD, 600f);

            final EnumSet<DiskType> allDiskTypes = EnumSet.allOf(DiskType.class);
            DiskType diskType = null;
            for (final DiskType candidateType : allDiskTypes) {
                if (candidateType.name().equalsIgnoreCase(this.interfaceType)) {
                    diskType = candidateType;
                    break;
                }
            }
            if (diskType == null) {
                if (diskTypeCorrectionsMap.containsKey(interfaceType)) {
                    diskType = diskTypeCorrectionsMap.get(interfaceType);
                    interfaceType = diskType.name();
                }
                else {
                }
            }

            if (interfaceSpeed == null) {
                interfaceSpeed = defaultInterfaceSpeeds.get(diskType);
            }

            final Map<Float, Float> interfaceSpeedCorrectionsMap = new LinkedHashMap<Float, Float>();
            assert diskType != null;
            switch (diskType) {
                case ATA:
                    //            ATA
                    //            ---
                    //            ATA PIO Mode 0           26.4 Mbit/s     3.3 MB/s
                    //            ATA PIO Mode 1           41.6 Mbit/s     5.2 MB/s
                    //            ATA PIO Mode 2           66.4 Mbit/s     8.3 MB/s
                    //            ATA PIO Mode 3           88.8 Mbit/s    11.1 MB/s
                    //            ATA PIO Mode 4          133.3 Mbit/s    16.7 MB/s
                    interfaceSpeedCorrectionsMap.put(26.4f, 3.3f);
                    interfaceSpeedCorrectionsMap.put(41.6f, 5.2f);
                    interfaceSpeedCorrectionsMap.put(66.4f, 8.3f);
                    interfaceSpeedCorrectionsMap.put(88.8f, 11.1f);
                    interfaceSpeedCorrectionsMap.put(133.3f, 16.7f);

                    interfaceSpeedCorrectionsMap.put(3000f, 100f);
                    interfaceSpeedCorrectionsMap.put(5400f, 100f);
                    interfaceSpeedCorrectionsMap.put(7200f, 100f);

                    break;
                case FATA:  // bottle necked by ATA
                    interfaceSpeedCorrectionsMap.put(26.4f, 3.3f);
                    interfaceSpeedCorrectionsMap.put(41.6f, 5.2f);
                    interfaceSpeedCorrectionsMap.put(66.4f, 8.3f);
                    interfaceSpeedCorrectionsMap.put(88.8f, 11.1f);
                    interfaceSpeedCorrectionsMap.put(133.3f, 16.7f);
                    interfaceSpeedCorrectionsMap.put(2000f, 100f);
                    break;
                case SATA:
                    interfaceSpeedCorrectionsMap.put(1500f, 187.5f);
                    interfaceSpeedCorrectionsMap.put(3000f, 375f);
                    interfaceSpeedCorrectionsMap.put(6000f, 750f);
                    interfaceSpeedCorrectionsMap.put(4000f, 375f);
                    interfaceSpeedCorrectionsMap.put(7200f, 187.5f);
                    break;
                case SATA2:
                    interfaceSpeedCorrectionsMap.put(3000f, 375f);
                    interfaceSpeedCorrectionsMap.put(6000f, 750f);
                    break;
                case SATA3:
                    interfaceSpeedCorrectionsMap.put(6000f, 750f);
                    break;
                case FC:
                    interfaceSpeedCorrectionsMap.put(1000f, 106.25f);
                    interfaceSpeedCorrectionsMap.put(2000f, 212.5f);
                    interfaceSpeedCorrectionsMap.put(4000f, 425f);
                    interfaceSpeedCorrectionsMap.put(8000f, 850f);

                    interfaceSpeedCorrectionsMap.put(200f, 212.5f);
                    interfaceSpeedCorrectionsMap.put(10000f, 106.25f);
                    interfaceSpeedCorrectionsMap.put(15000f, 106.25f);
                    interfaceSpeedCorrectionsMap.put(3000f, 106.25f);
                    interfaceSpeedCorrectionsMap.put(160f, 106.25f);
                    interfaceSpeedCorrectionsMap.put(320f, 212.5f);
                    interfaceSpeedCorrectionsMap.put(7200f, 106.25f);
                    interfaceSpeedCorrectionsMap.put(6000f, 425f);
                    break;
                case SAS:
                    interfaceSpeedCorrectionsMap.put(2400f, 300f);
                    interfaceSpeedCorrectionsMap.put(4800f, 600f);

                    interfaceSpeedCorrectionsMap.put(2000f, 300f);
                    interfaceSpeedCorrectionsMap.put(3000f, 300f);

                    interfaceSpeedCorrectionsMap.put(4000f, 300f);
                    interfaceSpeedCorrectionsMap.put(5000f, 300f);
                    interfaceSpeedCorrectionsMap.put(6000f, 600f);
                    break;
                case SSD:
                    interfaceSpeedCorrectionsMap.put(3000f, 600f);
                    interfaceSpeedCorrectionsMap.put(4000f, 600f);
                    rpm = 1000000;
                    break;
                default:
                    break;
            }

            if (interfaceSpeedCorrectionsMap.containsKey(interfaceSpeed)) {
                interfaceSpeed = interfaceSpeedCorrectionsMap.get(interfaceSpeed);
            }

            if (DiskType.SATA.equals(diskType)) {
                if (interfaceSpeed >= 750f) {
                    diskType = DiskType.SATA3;
                }
                else if (interfaceSpeed >= 375f) {
                    diskType = DiskType.SATA2;
                }
            }

            //System.out.println(vendor + " " + model + " " + interfaceType + " " + size + " " + rpm + " " + interfaceSpeed + " " + potentialTransferRate);
            if (potentialTransferRate != null) {
                if (potentialTransferRate.equals(80.0f) || potentialTransferRate > interfaceSpeed) {
                    potentialTransferRate = null;
                    potentialTransferRate = figureOutPTR(diskType.name(), interfaceSpeed, rpm);
                }
            }
            else {
                potentialTransferRate = figureOutPTR(diskType.name(), interfaceSpeed, rpm);
            }
        }
    }

    private static float figureOutPTR(final String diskType, final float interfaceSpeed, final int rpm) {
        // Potential Transfer Rate in MB/s (bytes not bits)
        //
        // Estimate of the maximum bandwidth a specific disk is capable of producing, independent of
        // the interface type.
        //
        // Historically, we often confused PTR with interface speed and/or RPM, or specified a value in Mbits/sec
        // instead of MB/s, or simply used a default value of 80 if PTR was not defined.
        //
        // PTR is used primarily to calculate disk utilization when the device does not itself export
        // a metric of disk utilization.
        //
        // Issue 23731 exposed bad disk utilization for SSD disks due to a bad (default) PTR of 80.
        // Issue 23902 notes that the PTR is wrong (wrong value, wrong units) or undefined for most
        // disk types.
        //
        // This method is an attempt to estimate a more accurate PTR for disks based interface type/speed,
        // RPM, and capacity.
        //
        // Numbers are based roughly on measured disk performance available at:
        //
        // http://www.tomshardware.com/charts/hdd-charts-2012/IOMeter-2006.07.27-Streaming-Reads,2928.html
        // http://www.tomshardware.com/charts/hdd-charts-2012/IOMeter-2006.07.27-Streaming-Writes,2929.html
        // http://www.techarp.com/showarticle.aspx?artno=511&pgno=4

        // Estimate a base PTR based on interface type/speed

        float potentialTransferRate = 80;
        if ("ATA".equals(diskType)) {
            potentialTransferRate = 70;
        }
        else if ("FATA".equals(diskType)) {
            potentialTransferRate = 70;
        }
        else if ("SATA".equals(diskType)) {
            potentialTransferRate = 80;
        }
        else if ("SATA2".equals(diskType)) {
            potentialTransferRate = 150;
        }
        else if ("SATA3".equals(diskType)) {
            potentialTransferRate = 200;
        }
        else if ("FC".equals(diskType)) {
            // Don't have FC2, FC4, FC8 defined, need to take interface speed into account
            potentialTransferRate = 90;         // up to 106.25
            if (interfaceSpeed > 106.25 && interfaceSpeed <= 212.5) {
                potentialTransferRate = 120;    // up to 212.5
            }
            else if (interfaceSpeed > 212.5 && interfaceSpeed <= 425) {
                potentialTransferRate = 170;    // up to 425
            }
            else if (interfaceSpeed > 425) {
                potentialTransferRate = 240;    // up to 850
            }
        }
        else if ("SAS".equals(diskType)) {
            // Don't have SAS2 defined, need to take interface speed into account
            potentialTransferRate = 150;        // up to 300
            if (interfaceSpeed > 300) {
                potentialTransferRate = 200;    // up to 600
            }
        }
        else if ("SSD".equals(diskType)) {
            potentialTransferRate = 500;        // always
        }

        // Adjust estimate based on RPM
        if (rpm <= 5400) {
            potentialTransferRate -= 5;
        }
        else if (rpm <= 7200) {
            // don't adjust
        }
        else if (rpm <= 10000) {
            potentialTransferRate = potentialTransferRate + 10;
        }
        else if (rpm <= 15000) {
            potentialTransferRate = potentialTransferRate + 20;
        }
        else {
            // SSDs should have RPM of 100000 and PTR 500
            potentialTransferRate = 500;
        }

        // Charts of measured disk performance did not indicate a obvious
        // correlation between capacity and transfer rate
        return potentialTransferRate;
    }

    //please do not modify without communicating to server/dwh team.
    public enum DiskType {
        OTHER,
        ATA,
        FATA,
        SATA,
        SATA2,
        SATA3,
        FC,
        SAS,
        SSD,
    }
}
