package com.onaro.sanscreen.acquisition.foundation.datasource;

import com.netapp.acquisition.aquisitionserverside.interfaces.remote.AcquisitionServerSession;
import com.onaro.sanscreen.acquisition.foundation.datasource.config.FoundationConfiguration;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.*;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import com.onaro.sanscreen.acquisition.schema.odata2.OData2Root;
import com.onaro.sanscreen.acquisition.sessions.AcquisitionUnitException;
import com.onaro.sanscreen.util.Schema;
import org.apache.log4j.Logger;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * Main listener for foundation events. This object contains the logic needed to link the foundation sample
 * to the SANscreen server / product.
 * <p/>
 * Replacing this set of interfaces allows the framework to run a data source completely detached from SANscreen
 * (see {@link com.onaro.sanscreen.acquisition.foundation.standalone.FoundationStandaloneListener})
 */
public class FoundationListener implements FoundationUpdatesListener, DataSourceSampleListener, OriginatorProvider, DataSourceLifeCycleListener {
    private final static Logger logger = Logger.getLogger(FoundationListener.class);

    private final FoundationConfiguration configuration;
    private final AuditSession auditSession;

    private boolean sendOnlyUpdates = true;

    private final Map<String, Long> originatorCache = new HashMap<String, Long>(10);

    public FoundationListener(final FoundationConfiguration configuration) {
        this.configuration = configuration;
        sendOnlyUpdates = ServiceManager.getInstance().getProperty("SEND_ONLY_UPDATES", Boolean.TRUE.toString()).equalsIgnoreCase(Boolean.TRUE.toString());

        auditSession = new AuditSession(configuration);
    }

    public void dataSourceStarting(final BaseDataSource dataSource) throws FrameworkException {
    }

    public void dataSourceStopping(final BaseDataSource dataSource) throws FrameworkException {
    }

    public void dataSourceStatusChanged(final BaseDataSource dataSource, final DataSourceStatus oldStatus, final DataSourceStatus newStatus, final String message) {
    }

    public void startingSample(final BaseDataSource dataSource, final PollingPackage pollType) {
        auditSession.createSession("Acquiring device");
    }

    public void sampleSuccessfull(final BaseDataSource baseDataSource, final PollingPackage aPackage) throws FrameworkException {
        // This acquisition is completely successful. This is the time to handle upgrades from non MOID to MOID implementations
        if (configuration.isMOID())
            handleMoidUpgrade();

        auditSession.updateFinalStatus(AcquisitionServerSession.AuditType.SUCCESS, "Success");
    }

    /**
     * Retires all originators reported by the datasource in the past that
     * are not reported on the current successful poll.
     *
     * @param aPackage foundation package being run
     * @throws FrameworkException for communication issues.
     */
    private void retireUnreportedOriginators(final PollingPackage aPackage) throws FrameworkException {
        if (!configuration.isMOID()
            || !configuration.isRetireUnreportedOriginators()
            || aPackage.getDataSource().isStopped())
                return;

        final Set<Long> activeOriginatorIDs = ((BaseFoundationPackage) aPackage).getCurrentlyReportedOriginatorIds();

        // Now, remove all originators for this data source that we did not successfully report
        try {
            ServiceManager.getInstance().
                    getCommunicationManager().
                    getAcquisitionFacadeSession().
                    retireUnreportedOriginators(activeOriginatorIDs, aPackage.getDataSource().getDataSourceID());
        }
        catch (RemoteException e) {
            throw new FrameworkException(e.getMessage(), e);
        }
        catch (AcquisitionUnitException e) {
            throw new FrameworkException(e.getMessage(), e);
        }
    }


    public void finishedSample(final BaseDataSource dataSource, final PollingPackage aPackage) throws FrameworkException {
        final boolean success = auditSession.closeSession(aPackage);

        if (success) {
            retireUnreportedOriginators(aPackage);
        }
    }

    public void sampleFailed(final BaseDataSource dataSource, final PollingPackage aPackage, final DataSourceErrorException e) {
        auditSession.updateFinalStatus(e);
    }

    public void sampleFailed(final BaseDataSource dataSource, final PollingPackage aPackage, final Throwable e) {
        logger.error(aPackage.getLogPrefix() + " - Unhandled exception: " + e.getMessage(), e);

        //noinspection ThrowableInstanceNeverThrown
        auditSession.updateFinalStatus(new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR,
                "Package " + aPackage.getPackageName() + " Failed with error:" + e.getMessage(),
                aPackage.getLogPrefix() + " - Unhandled exception: " + e.getMessage(),
                e));
    }

    /**
     * Returns an originator ID for the given key.
     * <p/>
     * This method will try to use the originator cache first to minimize
     * network calls to the server. If an originator ID could not be found in the cache for the given originatorKey,
     * the server will be queried and the cache updated.
     * <p/>
     * Calls into the AcquisitionFacadeRMI to retrieve an OriginatorID from the server
     * <p/>
     * 1) If datasourceID is null, an exception will be thrown in the server<br>
     * 2) If datasourceID is non-null, and devicekey is null, and a datasource row exists in the database, the
     * datasourceID will be returned as the originatorId<br>
     * 3) If datasourceID is non-null, device key is null, and no datasource row exists in the database, null will be
     * returned<br>
     * 4) If datasourceID is non-null and devicekey is non-null, an existing datasourceID will be returned if present,
     * otherwise a new originatorID will be returned, and entered into the database.<br>
     *
     * @param datasourceID
     * @param originatorKey
     * @param originatorDisplayName
     * @throws FrameworkException
     */
    public long getOriginatorID(final Long datasourceID, final String originatorKey, final String originatorDisplayName) throws FrameworkException {
        // we use the server's acquisitionFacadeSession method "getOriginatorID" to also update the originator's last access time
        // When we use the cached originator id, we skip the last access update. This should be updated at the end of
        // of the datasource run, but for now, we will disable the cache and force it to always update here.
        // TODO : impelement the proper update of this at the end of the datasource poll cycle. See bug 15498.
        Long originatorID = null;
//        Long originatorID = originatorCache.get(originatorKey);
//        if (originatorID == null) {
        try {
            originatorID = ServiceManager.getInstance().getCommunicationManager().getAcquisitionFacadeSession().getOriginatorID(datasourceID, originatorKey, originatorDisplayName);
        }
        catch (FrameworkException e) {
            throw new FrameworkException("Failed to get originator ID from server", e);
        }
        catch (AcquisitionUnitException e) {
            throw new FrameworkException("Failed to get originator ID from server", e);
        }

        originatorCache.put(originatorKey, originatorID);
//        }

        return originatorID;
    }

    @Override
    public long getOriginatorID(final long datasourceID, final String deviceKey, final String originatorName, final boolean performDuplicateDatasourceCheck) throws FrameworkException {
        // we use the server's acquisitionFacadeSession method "getOriginatorID" to also update the originator's last access time
        // When we use the cached originator id, we skip the last access update. This should be updated at the end of
        // of the datasource run, but for now, we will disable the cache and force it to always update here.
        // TODO : impelement the proper update of this at the end of the datasource poll cycle. See bug 15498.
        Long originatorID = null;
//        Long originatorID = originatorCache.get(originatorKey);
//        if (originatorID == null) {
        try {
            originatorID = ServiceManager.getInstance().getCommunicationManager().getAcquisitionFacadeSession().getOriginatorID(datasourceID, deviceKey, originatorName, performDuplicateDatasourceCheck);
        }
        catch (FrameworkException e) {
            throw new FrameworkException("Failed to get originator ID from server", e);
        }
        catch (AcquisitionUnitException e) {
            throw new FrameworkException("Failed to get originator ID from server", e);
        }

        originatorCache.put(deviceKey, originatorID);
//        }

        return originatorID;
    }

    public void deleteOriginatorID(final Long originatorID, final String originatorKey) throws FrameworkException {
        if (logger.isDebugEnabled())
            logger.debug("deleteOriginatorID() called. Using originatorID: [" + originatorID + "] Originator Key:[" + originatorKey + "]");

        originatorCache.remove(originatorKey);

        try {
            ServiceManager.getInstance().getCommunicationManager().getAcquisitionFacadeSession().deleteOriginatorID(originatorID, originatorKey);
        }
        catch (RemoteException e) {
            throw new FrameworkException("Failed to delete originator ID from server", e);
        }
        catch (FrameworkException e) {
            throw new FrameworkException("Failed to delete originator ID from server", e);
        }
        catch (AcquisitionUnitException e) {
            throw new FrameworkException("Failed to delete originator ID from server", e);
        }
    }

    /**
     * Called by the datasource the first time it polls to set the Management URL for the datasource itself.
     *
     * @param manageURL
     */
    public void reportManageURL(final BaseDataSource dataSource, final String manageURL) {
        try {
            ServiceManager.getInstance().getCommunicationManager().getAcquisitionFacadeSession().updateManageURL(dataSource.getDataSourceID(), manageURL);
        }
        catch (FrameworkException e) {
            logger.error(dataSource.getDataSourceName() + " - Failed to update originator audit", e);

            // Fall through...
        }
        catch (AcquisitionUnitException e) {
            logger.error(dataSource.getDataSourceName() + " - Failed to update originator audit", e);

            // Fall through...
        }
    }


    public void reportOData2(final BaseDataSource dataSource,
                            final String originaorKey,
                            final String origintorDisplayName,
                            final OData2Root odataRoot,
                            final String odataXML,
                            final boolean changed,
                            final Serializable context) throws FrameworkException {


        assert auditSession.isOpen();

        if (logger.isDebugEnabled())
            logger.debug(dataSource.getDataSourceName() + " - Closing audit 'Sending data to server'");

        // Check if data source was stopped. If it was, skip the update...
        if (!dataSource.getStatus().equals(DataSourceStatus.Aqcuiring) && !dataSource.getStatus().equals(DataSourceStatus.StandBy)) {
            logger.warn(dataSource.getDataSourceName() + " - Data source is not currently in Acquiring state (" + dataSource.getStatus() + ") skipping server update.");
            return;
        }

        if (changed || !sendOnlyUpdates) {
            try {
                // Send the new XML back to the server
                ServiceManager.getInstance().getCommunicationManager().getAcquisitionFacadeSession().update(odataXML, Schema.UNKNOWN);

                if (logger.isDebugEnabled())
                    logger.debug(dataSource.getDataSourceName() + " - Reported update to server");
            }
            catch (AcquisitionUnitException e) {
                throw new FrameworkException("Cannot update server (" + e.getMessage() + ")", e);
            }

            if (origintorDisplayName != null)
                auditSession.updateProgress("Sent " + origintorDisplayName + " data to server");
        }

        // Update the originator last access time
        try {
            // Send the new XML back to the server
            ServiceManager.getInstance().getCommunicationManager().getAcquisitionFacadeSession().updateOriginatorLastAccessTime(odataRoot.getOriginator().getOriginatorID());

            if (logger.isDebugEnabled())
                logger.debug(dataSource.getDataSourceName() + " - Update to originator access time");
        }
        catch (AcquisitionUnitException e) {
            throw new FrameworkException("Cannot update server (" + e.getMessage() + ")", e);
        }

    }

    public void recordingSessionDone(final BaseDataSource dataSource,
                                     final PollingPackage aPackage,
                                     final File recordingSession,
                                     final boolean shouldReport) throws FrameworkException {
        try {
            if (recordingSession != null && shouldReport)  //don't send to server if foundation indicated not to report 
                ServiceManager.getInstance().getLogsManager().sendZipToServer(ServiceManager.getInstance().getAcquisitionUnitManager().getAcquisitionUnit().getName(), recordingSession);
        }
        catch (IOException e) {
            throw new FrameworkException("Failed to send ZIP file: " + e.getMessage(), e);
        }
    }

    /**
     * When upgrading non-MOID data sources to MOID, we want to delete the old MOID (identified by originator ID null). We
     * want to do this only if the new MOID implementation successfully acquired ALL devices (to minimize the number of changes
     * generated).
     * <p/>
     * If the originator ID for the null key does not exist, this method will do nothing.
     *
     * @throws FrameworkException
     */
    private void handleMoidUpgrade() throws FrameworkException {
        assert configuration.isMOID();

        if (logger.isDebugEnabled())
            logger.debug(configuration.getDataSourceName() + " - handleMoidUpgrade() called - datasourceId: [" + configuration.getDataSourceID() + "]");

        deleteOriginatorID(configuration.getDataSourceID(), null);
    }

    public synchronized void cleanOriginatorCache() {
        originatorCache.clear();
    }

    public final String[] getOriginatorCache() {
        final String[] cache = new String[originatorCache.size()];

        int i = 0;
        for (final Map.Entry<String, Long> entry : originatorCache.entrySet())
            cache[i++] = entry.getKey() + " - " + entry.getValue();

        return cache;
    }
}
