package com.onaro.sanscreen.acquisition.foundation.datasource.dependant.recording;

import com.onaro.sanscreen.acquisition.foundation.datasource.dependant.BaseFoundationDependantPackage;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import org.apache.log4j.Logger;

import java.io.File;
import java.io.IOException;

/**
 * This class attempts to help guide a foundation dependent package through the appropriate states for
 * waiting on the first run of foundation and on taking dependent recordings with appropriate number of samples.
 *
 * <p>One important realization here is that the states related to "foundation" are only valid immediately after
 * package initialization and before foundation has completed once.   Once foundation has run once, these states
 * should no longer be encountered.   Calls to set them will be no-ops.   States past that point relate entirely
 * to the dependent package taking recordings and managing samples, or not as the case may be.
 *
 * <p>From a recording perspective, foundation dependent recordings only contain N samples (polls), after which recording
 * is shutdown (unlike foundation where every poll is recorded).   Recordings are then taken once a day (by default) with
 * the same amount of samples.   Starting/stopping the data source is designed to induce a new recording as well.  Finally,
 * one may trigger an out-of-band recording via the MBean interface.
 */
public class FoundationDependentRecordingHandler {

    private static final Logger logger = Logger.getLogger(FoundationDependentRecordingHandler.class);

    private static final long AUTOMATED_RECORDING_DELTA_MILLISECONDS =  1000 * 60 * 60 * 24 *
                Long.parseLong(ServiceManager.getInstance().getProperty("FOUNDATION_DEPENDENT_RECORDING_FREQUENCY_DAYS", "1"));

    /**
     * jConsole user may request recordings.
     * This enum illustrates the states of request processing.
     */
    private static enum STATE {
        NO_REQUESTS("no requests for recording"),
        // user has requested recording and foundation not started
        WAIT_FOR_FOUNDATION_TO_START("waiting for foundation sample to start"),
        // user has requested recording and foundation started
        WAIT_FOR_FOUNDATION_TO_COMPLETE("waiting for foundation sample to complete"),
        // foundation ran and was successful.
        FOUNDATION_SUCCESS ("foundation success"),
        // foundation ran and resulted in failure.
        FOUNDATION_FAILURE ("foundation failed"),
        // user has requested recording and foundation has completed a full sample
        WAIT_FOR_POLL("waiting for poll to start collecting"),
        // polls being recorded
        ACCUMULATING_SAMPLES("accumulating samples"),
        DONE("done collecting samples");

        private final String desc;

        private STATE(final String desc) {
            this.desc = desc;
        }

        public String toString() {
            return desc;
        }
    }

    private STATE state = STATE.NO_REQUESTS;

    private int numberOfRequestedRecordings = 1;

    private BaseFoundationDependantPackage thePackage;

    private long lastRecordingEndTime = -1;

    /**
     * Create a new recording handler state machine.
     *
     * @param recordingPackage  the package we are handling.
     */
    public FoundationDependentRecordingHandler(final BaseFoundationDependantPackage recordingPackage) {
        this.thePackage = recordingPackage;
    }

    /**
     * Number of requested recording samples used in recordings. (used for MBean exposure).
     * @return   the number of samples in any recording.
     */
    public synchronized int getNumberOfRequestedRecordings() {
        return state == STATE.NO_REQUESTS ? 0 : numberOfRequestedRecordings;
    }

    /**
     * @return false in case we are collecting first complete foundation sample, otherwise - true.
     */
    public synchronized boolean skipNotChangedReportOdata() {
        return state != STATE.WAIT_FOR_FOUNDATION_TO_COMPLETE;
    }

    /**
     * Call back when foundation is starting a sample.   If foundation has already completed successfully once before, this operation
     * is a no-op.   Note that the machine may be returned to the state of waiting on foundation by calling requestRecording.
     *
     * <p>If valid, the state machine will enter a state where it awaits the completion of the foundation poll.
     */
    public synchronized void startingFoundationSample() {
        if (state != STATE.WAIT_FOR_FOUNDATION_TO_START)
            return;

        state = STATE.WAIT_FOR_FOUNDATION_TO_COMPLETE;
        logger.debug(thePackage.getLogPrefix() + " - change recording state to WAIT_FOR_FOUNDATION_TO_COMPLETE");
    }

    /**
     * Call back when foundation has finished as ample.   If foundation has already completed successfully once before, this operation
     * is a no-op.  Note that the machine may be returned to the state of waiting on foundation by calling requestRecording.
     *
     * <p>When valid, if foundation was indicated to be a success, the state machine will move to a state where
     * foundation dependent polls are unlocked and may proceed to setup recordings and generate data.
     *
     * <p>When valid, if foundation failed, the state machine will return to a state where it awaits the NEXT foundation
     * poll.  The dependent package will NOT be unlocked to poll or take any new recordings in this case.   Again, note
     * this situation only applies to foundation failures on its first run.   Once foundation is successful once,
     * these states no longer apply.
     */
    public synchronized void finishFoundationSample() {
        if (state != STATE.FOUNDATION_SUCCESS && state != STATE.FOUNDATION_FAILURE)
            return;

        //if foundation succeeded, we will move this wait_for_poll, which will trigger performance polls and recordings
        //if necessary.  If failed, return to waiting_for_foundation_to_start.
        if (state == STATE.FOUNDATION_FAILURE) {
            state = STATE.WAIT_FOR_FOUNDATION_TO_START;
            logger.debug(thePackage.getLogPrefix() + " - change recording state to WAIT_FOR_FOUNDATION_TO_START");
        }
        else {
            state = STATE.WAIT_FOR_POLL;
            logger.debug(thePackage.getLogPrefix() + " - change recording state to WAIT_FOR_POLL");
        }
    }

    /**
     * Indicate foundation run was successful.  If foundation has already completed successfully once before, this operation
     * is a no-op.  Note that the machine may be returned to the state of waiting on foundation by calling requestRecording.
     * 
     * <p>Note this method should be called on partial success as well.
     */
    public synchronized void setFoundationSuccess() {
       if (state != STATE.WAIT_FOR_FOUNDATION_TO_COMPLETE)
            return;

        state = STATE.FOUNDATION_SUCCESS;
        logger.debug (thePackage.getLogPrefix() + " - change recording state to FOUNDATION_SUCCESS");
    }

    /**
     * Indicate foundation run was successful.  If foundation has already completed successfully once before, this operation
     * is a no-op.   Note that the machine may be returned to the state of waiting on foundation by calling requestRecording.
     */
    public synchronized void setFoundationFailed() {
       if (state != STATE.WAIT_FOR_FOUNDATION_TO_COMPLETE)
            return;

        state = STATE.FOUNDATION_FAILURE;
        logger.debug (thePackage.getLogPrefix() + " - change recording state to FOUNDATION_FAILURE");
    }

    /**
     * Request a recording via the state machine.  This is valid only for certain initial criteria, and will
     * throw a runtime exception if not met.   From a state perspective, this will return the state machine to a
     * state where it is again waiting on foundation to run before it can be polled.
     *
     * This may also be called when somebody initiates a recording via the MBean interface.   This will only succeed
     * in the cases where previous recordings have succeeded and we are not already in a foundation run.
     *
     * @param recordingSamples  the number of samples in the recording to take.
     */
    public void requestRecording(final int recordingSamples) {
        synchronized (this) {
            // we can let user press the button until we recieved a ping from foundation
            if (!(state == STATE.NO_REQUESTS || state == STATE.WAIT_FOR_FOUNDATION_TO_START)) {
                throw new RuntimeException("Can not accept request in state : " + state.toString());
            }

            // Can't ask for less than one recording
            if (recordingSamples < 1)
                throw new RuntimeException("Only accept requests for at least 1 sample");

            numberOfRequestedRecordings = recordingSamples;
            state = STATE.WAIT_FOR_FOUNDATION_TO_START;
            logger.debug(thePackage.getLogPrefix() + " - change recording state to WAIT_FOR_FOUNDATION_TO_START");
        }

    }

    /**
     * Get internal state as a string.
     *
     * @return    the internal state.
     */
    public String getRecordingStatus() {
        return state.toString();
    }

    /**
     * Called on startSample call back for dependent package.   Prepares things for recording if required and
     * appropriate.   If a recording is appropriate, this will actually write to the package configuration option
     * to flag that we are not recording.   This can then be read by the package implementation to determine what
     * to do if necessary.
     *
     * <p>Recordings are triggered immediately after foundation has fully completed successfully the first time
     * (or after recording reset is triggered) or after one full day (by default) since last recording completed.
     *
     * <p>When recording, the state is flagged as accumlating samples.   Once N samples have been gathered,
     * the recording state is shutdown and further samples (polls) are not recorded.   This call is a logical no-op in those
     * cases.
     *
     */
    public synchronized void prepareRecordingForSample() {        
        if (state == STATE.WAIT_FOR_POLL) {
            prepareForRecording(false);
        }
        else if (state == STATE.ACCUMULATING_SAMPLES) {
            logger.debug(thePackage.getLogPrefix() + " - Recording Handler creates new sample while in state ACCUMULATING_SAMPLES");
        }
        else if (state == STATE.DONE || state == STATE.NO_REQUESTS) {
            if (shouldInitiateAutomatedRecording()) {
                prepareForRecording(true);
            }
        }

        //if we are not recording now, this gets cleaned up in stopRecordingSessionIfNeeded if configuration is not
        //set to recording.
        thePackage.getRecordingSession().createSample();
    }

    /**
     * Internal method to determine if an actual recording should be initialized.   This is true if no recording
     * has already been taken, or if 1 day (by default) has passed since last recording.
     *
     * @return   true if new recording should be initiated.
     */
    private synchronized boolean shouldInitiateAutomatedRecording() {
        if (lastRecordingEndTime == -1) {
            return true;
        }
        final long now = System.currentTimeMillis();
        final long delta = now - lastRecordingEndTime;
        if (delta <= 0) {
            logger.warn(thePackage.getLogPrefix() + " Delta time since last recording was skewed.  Resetting.");
            lastRecordingEndTime = now;
            return false;
        }
        else {
            return delta >= AUTOMATED_RECORDING_DELTA_MILLISECONDS;
        }
    }

    /**
     * Internal method to setup configuration for recording.   This method actually sets the bit on the
     * package configuration to flag we are now recording.   If the passed boolean is true, we also reset the
     * recording session (resets and clears all previous samples).
     *
     *
     * @param resetSession    true if the session should be reset w/r to previous samples.
     */
    private void prepareForRecording(boolean resetSession) {
        thePackage.setRecording(true);
        if (resetSession) {
            thePackage.getRecordingSession().reset();
        }
        thePackage.getRecordingSession().setMaxRecordingSamples(numberOfRequestedRecordings);

        state = STATE.ACCUMULATING_SAMPLES;
        logger.debug(thePackage.getLogPrefix() + " - change recording state to ACCUMULATING_SAMPLES");
    }

    /**
     * Are we still waiting for foundation to complete successfully at least once?   If true, no foundation
     * dependent polls should proceed.   Note when this state machine gets reset to take another recording, we
     * will return to a state of waiting for foundation.
     *
     * Once foundation runs once, this will always return false.
     *
     * @return true if we are waiting for a complete foundation sample to start recording
     */
    public synchronized boolean isWaitingForFoundationInitialization() {
        return state == STATE.WAIT_FOR_FOUNDATION_TO_COMPLETE
               || state == STATE.WAIT_FOR_FOUNDATION_TO_START
               || state == STATE.FOUNDATION_SUCCESS
               || state == STATE.FOUNDATION_FAILURE;
    }

    /**
     * Called after dependant package poll finished.   If we aren't recording, this will clear the most recent sample.
     * If we are recording and the max number of samples (polls) has been recorded, this will trigger the state machine
     * to finalize the recording and put us in a state where we are no longer recording.
     *
     */
    public synchronized void stopRecordingSessionIfNeeded() {
        if (!thePackage.isRecording()) {
            thePackage.getRecordingSession().reset();
        }
        else if (state == STATE.ACCUMULATING_SAMPLES && thePackage.getRecordingSession().hasMaxRecordingSamples()) {
            stopRecordingSession();
        }
    }

    /**
     * Stop recording and finalize any current recording samples.   Puts in a state where subsequent polls are not
     * recorded.
     */
    public void stopRecordingSession() {
        state = STATE.DONE;
        logger.debug(thePackage.getLogPrefix() + " - change recording state to DONE");

        thePackage.setRecording(false); 
        lastRecordingEndTime = System.currentTimeMillis();

        final File recordingSession;
        try {
            recordingSession = thePackage.getRecordingSession().buildRecordingFile();
            thePackage.getDataSource().recordingSessionDone(thePackage, recordingSession, true);
        }
        catch (IOException e) {
            logger.error(thePackage.getLogPrefix() + " - failed to finalize recording file " + e.getMessage(), e);
        }
        finally {
            state = STATE.NO_REQUESTS;
            logger.debug(thePackage.getLogPrefix() + " - change recording state to NO_REQUESTS");
        }
    }
}
