package com.onaro.sanscreen.acquisition.framework;

import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import org.apache.log4j.Logger;
import org.apache.log4j.helpers.LogLog;

import java.io.*;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;

/**
 * Main starting point for the acquisition
 */
public class Main {
    private static Logger logger;

    private static Main intstance = new Main();

    private boolean stopping = false;
    private boolean stopped = false;

    int exitLevel = -1;

    static {
        LogLog.setQuietMode(true);
    }

    public boolean isStopped() {
        return stopped;
    }

    public boolean isStopping() {
        return stopping;
    }

    public void stop() {
        stopping = true;
        Thread.currentThread().interrupt();
    }

    /**
     * Creates a new instance of Main
     */
    private Main() {
    }

    public static Main getInstance() {
        return intstance;
    }

    /**
     * The main function. This is where the acquisition is initialized, waited upon, and stopped.
     *
     * @throws FrameworkException When there was an error while running
     */
    public void run() throws FrameworkException {
        try {
            stopping = false;

            logger.info("Main - Starting acquisition...");

            try {
                ServiceManager.getInstance().start();
            }
            catch (FrameworkException e) {
                logger.error("Main - failed to start services: " + e.getMessage(), e);
                exitLevel = 1;
                throw e;
            }

            logger.info("Main - Acquisition is up and running!");

            // Wait until a stopping signal will be detected
            while (!isStopping()) {
                try {
                    Thread.sleep(1000);
                }
                catch (InterruptedException e) {
                    // Fall through...
                }
            }

            logger.info("Main - Stopping acquisition...");

            // Stop all services...
            ServiceManager.getInstance().stop();

            logger.info("Main - Stopped!");

            exitLevel = 0;
        }
        finally {
            stopped = true;
        }
    }

    /**
     * Main method.
     * <p/>
     * <p/>Usage: <b>java com.onaro.sanscreen.acquisition.frameworkMain [start|stop]</b>
     * <p/>
     * <p/>When mode is <code>start</code>, this method will initialize and start the acquisition.
     * <p/>When mode is <code>stopping</code>, this will stopping the running acquition service (if there is one)
     * <p/>
     * The stopping is done by connecting to the existing acq process using a locally opened port. This port is typically
     * 12123 but can be changed using the system property <code>sanscreen.acq.stop.port</code>
     *
     * @param args
     */
    public static void main(String args[]) {
        
        // Set the logger. We can do this only here after the JVM is initiailized...
        logger = Logger.getLogger(Main.class);


        final int exitCode;

        final int stopServiceListeningPort = Integer.getInteger("sanscreen.acq.stop.port", 12123);

        if (args.length == 0 || args[0].equalsIgnoreCase("start")) {
            exitCode = Main.startUp(stopServiceListeningPort);
        }
        else if (args[0].equalsIgnoreCase("stop")) {
            exitCode = Main.shutDown(stopServiceListeningPort);
        }
        else {
            System.out.println("Invalid option '" + args[0] + "'");
            exitCode = -1;
        }

        System.exit(exitCode);
    }

    /**
     * Start acquisition on this machine.
     *
     * @param stopServiceListeningPort
     */
    private static int startUp(final int stopServiceListeningPort) {
        int exitLevel;

        EndRequestListener shutDownListener = null;

        try {

            final Main theMain = new Main();

            // Now start a listener thread for the shutdown port
            shutDownListener = new EndRequestListener(theMain, stopServiceListeningPort);
            shutDownListener.start();

            // Add a shutdown hook that will notify us when system is shutting down without our request!!!
            Runtime.getRuntime().addShutdownHook(new Thread() {
                public void run() {
                    logger.info("Shutting down by external request");
                }
            });

            theMain.run();

            if (theMain.isStopped()) {
                shutDownListener = null;
            }

            exitLevel = theMain.getExitLevel();
        }
        catch (Throwable e) {
            logger.error("Main - " + e.getMessage(), e);

            exitLevel = 2;

            // Fall through...
        }
        finally {
            if (shutDownListener != null) {
                shutDownListener.shutdown();
            }
        }

        return exitLevel;
    }

    /**
     * Returns indicates what is the exit level. As long as the acquisition is not about to exit the value is -1. Exiting
     * normaly is 0, and any value>0 is an error.
     *
     * @return indicates what is the exit level. As long as the acquisition is not about to exit the value is -1. Exiting
     */
    public int getExitLevel() {
        return exitLevel;
    }
    
    /**
     * Shut down a running acquisition on this machine.
     * <p/>
     * Connect using a socket to the 'acquisition shutdown port' to signal the acquisition to shut down.
     *
     * @param stopServiceListeningPort
     */
    private static int shutDown(final int stopServiceListeningPort) {
        return EndRequestListener.sendStopCommand(stopServiceListeningPort);
    }

    /**
     * Listener thread for shutdown requests using a port. This is used when trying to shutdown acuisition using
     * "net stopping acquisition". This translates into calling Main with the <code>stopping</code> argument.
     * <p/>
     * <p/>The stopping method will connect to this port and this connection will trigger the stopping of the acquisition.
     */
    private static class EndRequestListener extends Thread {
        private boolean stop = false;
        private final Main theMain;
        private final int stopServiceListeningPort;

        public EndRequestListener(final Main theMain, final int stopServiceListeningPort) {
            this.theMain = theMain;
            this.stopServiceListeningPort = stopServiceListeningPort;
        }

        /**
         * Signals the process to stop and wait until the exit code is returned
         *
         * @param stopPort
         *
         * @return Exit code for the process
         */
        public static int sendStopCommand(final int stopPort) {
            Socket socket = null;
            String response = null;
            try {
                // Internally connect to the server socket to realse the accept method...
                socket = new Socket(InetAddress.getByAddress(new byte[]{127, 0, 0, 1}), stopPort);

                final BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream(), "utf-8"));
                writer.write("Stop");
                writer.newLine();
                writer.flush();
                socket.shutdownOutput();

                final BufferedReader lnr = new BufferedReader(new InputStreamReader(socket.getInputStream(), "utf-8"));
                response = lnr.readLine();
                return Integer.parseInt(response);
            }
            catch (IOException e) {
                logger.error("Main - Failed to signal acquisition service to stop...", e);
                return 3;
            }
            catch (NumberFormatException e) {
                // DE2571 Burt 648415 - ACQ log throws EEROR during shutdown - "Main - Failed to parse exit code from null"
                //
                // Known error condition under linux due to:
                // On Windows, "Stop" sender is in same process space as acq service via service wrapper.
                // On Linux, "Stop" sender is invoked in a separate JVM
                // This along with slightly different socket timing/listening/closing behavior
                // causes the reader to read null.
                //
                // Specifically, on Linux, the Stop listener never sends a response, the socket is closed
                // and we read null.
                //
                // See linux comment in run()
                //
                // For now, suppress the error message for this case.
                // TODO: better solution
                if (response != null) {
                    logger.error("Main - Failed to parse exit code from " + response);
                }
                return 4;
            }
            finally {
                if (socket != null) {
                    try {
                        socket.close();
                    }
                    catch (IOException e) {
                        // Fall through...
                    }
                }
            }
        }

        public void shutdown() {
            stop = true;

            sendStopCommand(stopServiceListeningPort);
        }

        public void run() {
            try {
                final ServerSocket shutdownSocket = new ServerSocket(stopServiceListeningPort, 1, InetAddress.getByAddress(new byte[]{127, 0, 0, 1}));

                while (true) {
                    final Socket socket = shutdownSocket.accept();

                    if (stop)
                        return;

                    // The socket should have the text 'stop'....
                    final BufferedReader lnr = new BufferedReader(new InputStreamReader(socket.getInputStream(), "utf-8"));
                    final String msg = lnr.readLine();
                    if (msg == null || !msg.equals("Stop")) {
                        logger.warn("Main - received message other than Stop in the service stop listening socket [" + msg + "]");

                        socket.close();
                        continue;
                    }

                    // If we're here, then we got a shut down request...
                    logger.info("Main - Acquisition stop requested externally, pending....");

                    // Signal the acquisition to stopping...
                    theMain.stop();

                    // And wait until it is stopped
                    while (!theMain.isStopped()) {
                        try {
                            Thread.sleep(500);
                        }
                        catch (InterruptedException e1) {
                            // Fall through...
                        }
                    }

                    // DE2571 On linux, the JVM often exits before get here and return a status code.

                    // Now that the acquisition has stopped, write back the exit code in the stream to the one initiating the
                    // stopping command.
                    final BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream(), "utf-8"));
                    writer.write(Integer.toString(theMain.getExitLevel()));
                    writer.newLine();
                    writer.flush();

                    socket.shutdownOutput();
                }
            }
            catch (IOException e) {
                logger.error("Main - Can't wait for acquisition stop requests!", e);

                // Fall through...
            }
            finally {
                logger.info("Main - Stop listener thread exited!");
            }
        }
    }
}
