package com.onaro.sanscreen.acquisition.framework.datasource;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import java.util.LinkedList;
import java.util.List;


/**
 * This exception class aggregates a list of {@link DataSourceErrorException} objects that were generated during the data source
 * sample. This is especially usefull for data sources implementing the Multiple Originator ID algorithm. Since each
 * device is reported under a different originator ID, there can be multiple errors generated during the sample.
 */
public class AggregatedDataSourceErrorsException extends DataSourceErrorException {
    private List<DataSourceErrorException> exceptions = new LinkedList<DataSourceErrorException>();
    private int totalNumberOfOriginators;

    /**
     * Constructor.
     * This constructor is mandatory to guarantee the correct display of the final message.
     * The final message is the one that displays  that indicates the number of datasources that
     * failed and the ones that success
     *
     * @param totalNumberOfOriginators
     */
    public AggregatedDataSourceErrorsException(final int totalNumberOfOriginators) {
        this.totalNumberOfOriginators = totalNumberOfOriginators;
    }



    /**
     * Returns TRUE if this obj. was used to collect exception generated by the data source.
     */
    public boolean containsEnhancedErrorDetail() {
        return (!exceptions.isEmpty()); // Negative statement.
    }

    /**
     * Adds a new exception to the buffer.
     *
     * @param e
     * @param deviceKey
     */
    public void addException(final DataSourceErrorException e, final String deviceKey) {
        e.setDeviceKey(deviceKey);
        exceptions.add(e);

        invalidateCrcCache();
    }

    /**
     * Returns consolidated message string
     *
     * @return Returns consolidated message string
     */
    public String getMessage() {
        final StringBuilder sb = new StringBuilder(160);
        final int totalSuccessCount = getNumberOfSuccessfulOriginators();

        sb.append(totalSuccessCount).append(" out of ").append(totalNumberOfOriginators).append(" devices acquired successfully.");

        return sb.toString();
    }

    /**
     * Returns the list of exceptions gather at this point.
     *
     * @return list of exceptions gather so far.
     */
    public List<DataSourceErrorException> getExceptions() {
        return exceptions;
    }

    /**
     * Returns true if no exceptions where reported, false otherwise.
     */
    public boolean hasExceptions() {
        return !exceptions.isEmpty();
    }

    public DataSourceErrorType getType() {
        // If all of them are failing, report it as failure instead of partial success (take the first one's type)
        if (getNumberOfSuccessfulOriginators() == 0)
            return DataSourceErrorType.FULL_FAILURE;

        return DataSourceErrorType.PARTIAL_SUCCESS;
    }

    /**
     * Returns saves the total number of originators that are going to participate in this processing.
     * In order to know how many originators terminated successfully we are going to substract the total number of
     * originators (given at Constructor) and the number of Exceptions that we buffered up.
     *
     * @return the total number of originators that are going to participate in this processing.
     */
    public int getTotalNumberOfOriginators() {
        return totalNumberOfOriginators;
    }

    /**
     * How many of the originators succeeded in this aggregation?   Defined as difference between total originators
     * and number of exceptions.
     *
     * @return  number of successful originators.
     */
    public int getNumberOfSuccessfulOriginators() {
        final int t= totalNumberOfOriginators - exceptions.size();
        return t < 0 ? 0 : t;
    }

    /**
     * Reports the error details to a string builder. This is used to be able to construct the exception as a string and
     * then later be identified compared to previously reported errors.
     *
     * @param sb StringBuilder to put result in
     */
    public void printToBuffer(final StringBuilder sb) {
        sb.append(getMessage());

        // Now report each of the contained exceptions
        int i = 1;
        for (final DataSourceErrorException e : exceptions) {
            // Report this error to the log first (the type of report is taken from the error type - this will allow automatic scripting to distinguish between
            // different types of errors using the log level).
            final StringBuilder msg = new StringBuilder(500);
            msg.append("\t Error ").append(i).append(" out of ").append(exceptions.size()).append(": ");
            msg.append(" [").append(e.getType().getMessage()).append("] - ").append(e.getMessage()).append((e.getLogMessage() != null ? " (" + e.getLogMessage() + ")" : ""));

            sb.append(msg);
            e.printToBuffer(sb);
            sb.append(crlf);
            i++;
        }
    }

    public void printToLog(final Logger logger, final String dataSourceName) {
        // Report the general message
        logger.error(dataSourceName + " - " + getMessage());

        // Now report each of the contained exceptions
        int i = 1;
        for (final DataSourceErrorException e : exceptions) {
            // Report this error to the log first (the type of report is taken from the error type - this will allow automatic scripting to distinguish between
            // different types of errors using the log level).
            final StringBuilder msg = new StringBuilder(500);
            msg.append("\t Error ").append(i).append(" out of ").append(exceptions.size()).append(": ");
            msg.append(dataSourceName).append(" [").append(e.getType().getMessage()).append("] - ").append(e.getMessage()).append((e.getLogMessage() != null ? " (" + e.getLogMessage() + ")" : ""));
            final Exception expToLog = (e.getType().getReportLevel() == Level.ERROR || e.getType().getReportLevel() == Level.FATAL) ? e : null;
            logger.log(e.getType().getReportLevel(), msg, expToLog);
            i++;
        }
    }
}
