package com.onaro.sanscreen.acquisition.framework.datasource;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.DataSourceAuditDetail;
import com.netapp.acquisition.aquisitionserverside.interfaces.remote.AcquisitionServerSession;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.config.PackageConfiguration;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import org.apache.log4j.Logger;

import java.util.LinkedList;
import java.util.List;

/**
 * An audit session is the status representation of a single DS package poll. Each times a package wakes up and starts
 * polling a device, it opens a new audit session and this remains active and gets updated until the poll is done and
 * the audit session is closed.
 */
@SuppressWarnings({"ConstantConditions"})
public final class AuditSession {
    private final static Logger logger = Logger.getLogger(AuditSession.class);

    private final String packageName;
    private final String logPrefix;
    private final PackageConfiguration configuration;

    private long auditId = -1;

    private AcquisitionServerSession.AuditType finalType = null;
    private String finalStatus = null;
    private DataSourceErrorException finalError = null;

    public AuditSession(final PackageConfiguration configuration) {
        this.packageName = configuration.getAuditName();
        this.configuration = configuration;
        this.logPrefix = configuration.getDataSourceName() + "-" + packageName;
    }

    /**
     * Creates a new audit session (including the changes in the DB)
     *
     * @param msg Initial message to set for audit
     */
    public final void createSession(final String msg) {
        if (logger.isDebugEnabled())
            logger.debug("createAudit() - (auditId:" + auditId + ") (msg:" + msg + ")");

        try {
            finalType = null;
            finalStatus = null;
            finalError = null;

            assert (auditId != -1);
            auditId = ServiceManager.getInstance().getCommunicationManager().getAcquisitionServerSide().createDsAudit(configuration.getDataSourceID(), packageName, msg);
        }
        catch (FrameworkException e) {
            logger.error(logPrefix + " - Failed to open audit record for data source - processing (auditId:" + auditId + ") +(msg:" + msg + ")", e);

            // Fall through...
        }
        catch (Throwable e) {
            // Unhandled exceptions related with audit should not cause the acqisition to crash
            logger.error(logPrefix + " - Unhandled exception (auditID " + auditId + "): " + e.getMessage(), e);

            // Fall through...
        }
    }

    /**
     * Update the status of the audit session (audit must be previously created by {@link #createSession(String)})
     *
     * <p/>This method is used to report a progress update on how the poll is going. For example, if the DS is acquiring
     * different devices during its poll cycle, it would be helpful to get an update on which device is currently being
     * polled.
     *
     * @param msg New audit status
     */
    public final void updateProgress(final String msg) {
        assert auditId != -1;

        if (logger.isDebugEnabled())
            logger.debug("updateAudit() - input (auditId:" + auditId + ") (msg:" + msg + ")");

        try {
            ServiceManager.getInstance().getCommunicationManager().getAcquisitionServerSide().updateDsAudit(auditId, AcquisitionServerSession.AuditType.ACQUIRING, msg);
        }
        catch (FrameworkException e) {
            logger.error(logPrefix + " - Failed to update audit record for data source - processing (auditId:" + auditId + ") (msg:" + msg + ")", e);

            // Fall through...
        }
        catch (Throwable e) {
            // Unhandled exceptions related with audit should not cause the acquisition to crash
            logger.error(logPrefix + " - Unhandled exception (auditID " + auditId + "): " + e.getMessage(), e);

            // Fall through...
        }
    }

    /**
     * Update the audit session with an error that was caught during the audit session.
     *
     * <p/>This information will only be sent to the server when the session is actually closed
     *
     * @param exception Error caught
     */
    public final void updateFinalStatus(final DataSourceErrorException exception) {
        finalError = exception;

        String guiMessage = exception.getType().getMessage();
        if (exception.getMessage() != null)
            guiMessage += ": " + exception.getMessage();

        // Write to the data source audit micro-view
        updateFinalStatus(AcquisitionServerSession.AuditType.ERROR, guiMessage);
    }

    /**
     * Update the audit session with a success message
     *
     * <p/>This information will only be sent to the server when the session is actually closed
     *
     * @param type   of audit
     * @param status Last status for package
     */
    public final void updateFinalStatus(final AcquisitionServerSession.AuditType type, final String status) {
        finalType = type;
        finalStatus = status;
    }

    /**
     * Closes the audit session (audit must be previously created by {@link #createSession(String)})
     *
     * <p/>This will also update the DB with the last status reported for this session and any extended error details
     * that go with it.
     *
     * <p/>This method also tries to minimize the amount of server calls made by trying to detect repeating messages and
     * avoid sending this information to the server. This is especially important when dealing with the extended error
     * details.
     *
     * @param pack handle to the package for which we are closing this audit.
     *
     * @return true if the poll was a success, false otherwise 
     */
    public final boolean closeSession(final Package pack) {
        if (logger.isDebugEnabled())
            logger.debug("closeAudit() - (auditId:" + auditId + ")");

        assert auditId != -1;

        try {
            List<DataSourceAuditDetail> messages = null;
            long crc = -1;

            if (finalError != null && finalError.containsEnhancedErrorDetail()) {
                // Extract the enhanced details from the exception and report it
                messages = new LinkedList<DataSourceAuditDetail>();

                final List<DataSourceErrorException> exceptionsToHandle = new LinkedList<DataSourceErrorException>();
                // First figure out how many exceptions we have to report...
                if (finalError instanceof AggregatedDataSourceErrorsException)
                    exceptionsToHandle.addAll(((AggregatedDataSourceErrorsException) finalError).getExceptions());
                else
                    exceptionsToHandle.add(finalError);

                // For each exception, report its details to the server
                for (final DataSourceErrorException dse : exceptionsToHandle) {
                    if (!dse.containsEnhancedErrorDetail())
                        continue;

                    if (dse.getDeviceKey() == null)
                        dse.setDeviceKey("General Device");

                    messages.add(new DataSourceAuditDetail(auditId,
                                                           dse.getDeviceKey(),
                                                           dse.getType().getReportLevel().toString(),
                                                           dse.getMessage(),
                                                           dse.getEnhancedErrorDetail()));
                }

                crc = finalError.getExceptionCRC();
            }

            if (finalType == null) {
                if (pack.getDataSource().isStopped()) {
                    if (logger.isDebugEnabled()) {
                        logger.debug(logPrefix + " - DataSource was stopped during poll and audit details need cleaned up.");
                    }
                    updateFinalStatus(AcquisitionServerSession.AuditType.ERROR, "DataSource was stopped during poll.");
                }
                else {
                    throw new FrameworkException ("Could not determine final Audit type!");
                }
            }

            ServiceManager.getInstance().getCommunicationManager().getAcquisitionServerSide().closeDsAudit(auditId, finalType, finalStatus, messages, crc);

            // report success
            if (finalError == null && finalStatus != null && finalStatus.equals("Success")) {
                return true;
            }

        }
        catch (FrameworkException e) {
            logger.error(logPrefix + " - Failed to close audit record for package  - processing (auditId:" + auditId + ")", e);

            // Fall through...
        }
        catch (Throwable e) {
            // Unhandled exceptions related with audit should not cause the acqisition to crash
            logger.error(logPrefix + " - Unhandled exception (auditID " + auditId + "): " + e.getMessage(), e);

            // Fall through...
        }
        finally {
            auditId = -1;
            if (logger.isDebugEnabled())
                logger.debug("closeAudit() - resetting auditId:" + auditId + ")");

            finalError = null;
            finalStatus = null;
            finalType = null;
        }

        return false;

    }

    /**
     * @return true if the session is open
     */
    public final boolean isOpen() {
        return auditId != -1;
    }
}
