package com.onaro.sanscreen.acquisition.framework.datasource;

import com.netapp.oci.profiler.ProfilerService;
import com.netapp.oci.profiler.ProfilerServiceException;
import com.netapp.oci.profiler.ProfilerServiceFactory;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.config.DataSourceConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.config.PackageLibraryMBean;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileWriter;
import com.onaro.sanscreen.acquisition.framework.mgmt.DataSourceManager;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import com.onaro.sanscreen.acquisition.sessions.AcquisitionUnitException;
import com.onaro.util.CyclicArray;
import org.apache.log4j.Logger;

import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import java.io.File;
import java.io.IOException;
import java.lang.management.ManagementFactory;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Base class for all data sources. This class provides utility methods for handling log files, packing, marshaling
 * the XMLs communicating to the server, etc...
 */
public class BaseDataSource implements BaseDataSourceMBean {
    private final static Logger logger = Logger.getLogger(BaseDataSource.class);

    private final List<DataSourceLifeCycleListener> lifeCycleListeners = new LinkedList<DataSourceLifeCycleListener>();
    private final Map<String, List<DataSourceSampleListener>> sampleListeners = new LinkedHashMap<String, List<DataSourceSampleListener>>();

    private final DataSourceConfiguration configuration;

    private final Map<String, Package> packages = new LinkedHashMap<String, Package>(10);
    private final Map<String, ObjectName> libraryMBeans = new ConcurrentHashMap<String, ObjectName>();

    private volatile DataSourceStatus status = DataSourceStatus.Stopped;

    // Track status change history (for management purposes)
    private final CyclicArray<String> statusHistory = new CyclicArray<String>(30);

    // Keeps track of the types of polling the data source is currently performing
    private final Set<PollingPackage> currentlyPolling = new HashSet<PollingPackage>();

    /**
     * Constructs a (empty) data source instance.
     *
     * Note: this will not register any package implementation into this data source. Use {@link #registerPackage(Package)}
     * to add Package implementations.
     *
     * @param configuration Description of the data source's configuration
     */
    public BaseDataSource(final DataSourceConfiguration configuration) {
        this.configuration = configuration;
    }

    public final DataSourceConfiguration getConfiguration() {
        return configuration;
    }

    public final void addDataSourceLifeCycleListener(final DataSourceLifeCycleListener listener) {
        if (status.isStarted())
            throw new IllegalStateException("Cannot add listener after data source is started");

        lifeCycleListeners.add(listener);
    }

    public final void addDataSourceSampleListener(final String packageName, final DataSourceSampleListener listener) {
        if (status.isStarted())
            throw new IllegalStateException("Cannot add listener after data source is started");

        List<DataSourceSampleListener> listenersForPackage = sampleListeners.get(packageName);
        if (listenersForPackage == null) {
            listenersForPackage = new LinkedList<DataSourceSampleListener>();
            sampleListeners.put(packageName, listenersForPackage);
        }

        listenersForPackage.add(listener);
    }

    public final void addDataSourceSampleListener(final String packageName, final int index, final DataSourceSampleListener listener) {
        if (status.isStarted())
            throw new IllegalStateException("Cannot add listener after data source is started");

        List<DataSourceSampleListener> listenersForPackage = sampleListeners.get(packageName);
        if (listenersForPackage == null) {
            listenersForPackage = new LinkedList<DataSourceSampleListener>();
            sampleListeners.put(packageName, listenersForPackage);
        }

        listenersForPackage.add(index, listener);
    }

    /**
     * Initialize data source.
     *
     * <p/>This method will call all the life cycle listeners that registered for these events. It assumes that as long
     * as nobody threw an exception, the data source should continue starting!
     *
     * @return true if initialization was successful, false otherwise (data source will not start sampling)
     */
    @SuppressWarnings({"ThrowableInstanceNeverThrown"})
    public final boolean start() throws FrameworkException {
        status = DataSourceStatus.Starting;

        registerMBeans();

        // Register this poll type as an MBean for management
        for (final DataSourceLifeCycleListener dataSourceLifeCycleListener : lifeCycleListeners)
            dataSourceLifeCycleListener.dataSourceStarting(this);

        // Start all packages
        for (final Package aPackage : packages.values()) {
            if (logger.isDebugEnabled())
                logger.debug(getDataSourceName() + " - Starting package " + aPackage.getPackageName());

            // TODO: Change interfaces and code to support error handling for non-polling packages!
            if (!(aPackage instanceof PollingPackage))
                throw new FrameworkException("Non polling packages are not supported yet!");

            final PollingPackage pollingPackage = (PollingPackage) aPackage;

            try {
                pollingPackage.start();
            }
            catch (FrameworkException e) {
                reportError(new DataSourceErrorException(DataSourceErrorType.INITIALIZING, "Data source failed to start", e.getMessage(), e), pollingPackage);
                return false;
            }
            catch (DataSourceErrorException e) {
                reportError(new DataSourceErrorException(DataSourceErrorType.INITIALIZING, "Data source failed to start", e.getMessage(), e), pollingPackage);
                return false;
            }
        }

        changeStatus(DataSourceStatus.StandBy, "Stand-by");
        return true;
    }

    /**
     * Stop the data source. When possible, perform a graceful shutdown, close all open communication with any device
     * and then return.
     *
     * <p/>This method will call all the life cycle listeners that registered for these events.
     */
    public final void stop() {
        final ClassLoader previousClassLoader = Thread.currentThread().getContextClassLoader();

        try {
            Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());

            changeStatus(DataSourceStatus.Stopping, "Stopping");

            for (final DataSourceLifeCycleListener dataSourceLifeCycleListener : lifeCycleListeners) {
                try {
                    dataSourceLifeCycleListener.dataSourceStopping(this);
                }
                catch (FrameworkException e) {
                    logger.error(getDataSourceName() + " - Data source failed to initialize: " + e.getMessage(), e);

                    // Fall through...
                }
            }

            // Stop all packages
            for (final Package aPackage : packages.values()) {
                try {
                    if (logger.isDebugEnabled())
                        logger.debug(getDataSourceName() + " - Starting package " + aPackage.getPackageName());

                    aPackage.stop();
                }
                catch (DataSourceErrorException e) {
                    logger.error(getDataSourceName() + " - Failed to stop package " + aPackage.getPackageName(), e);
                }
                catch (Throwable e) {
                    logger.error(getDataSourceName() + " - Failed to stop package " + aPackage.getPackageName(), e);
                }
            }

            unregisterMBeans();

            changeStatus(DataSourceStatus.Stopped, "Stopped");
        }
        finally {
            Thread.currentThread().setContextClassLoader(previousClassLoader);
        }
    }

    /**
     * Perform a single acquisition cycle.
     *
     * <p/>This method is called when the {@link DataSourceManager} when the data source is scheduled to run and a
     * thread is allocated for it. A check is made to make sure we are not already running this data source with the
     * same poll type, and then the method performs all the initialization and context switching neccessary for a
     * data source to run like switching to the right class loader, etc...
     * 
     * <p/>After initialization is over, the <code>doRun</code> method is called.
     *
     * @param packageName The name of the package to poll
     */
    public final void run(final String packageName) {
        final PollingPackage aPackage = getPackage(packageName);
        run(aPackage);
    }

    /**
     * Perform a single acquisition cycle.
     * 
     * <p/>This method is called when the {@link DataSourceManager} when the data source is scheduled to run and a
     * thread is allocated for it. A check is made to make sure we are not already running this data source with the
     * same poll type, and then the method performs all the initialization and context switching neccessary for a
     * data source to run like switching to the right class loader, etc...
     *
     * <p/>After initialization is over, the <code>doRun</code> method is called.
     *
     * @param aPackage The type of poll to run
     */
    public final void run(final PollingPackage aPackage)
    {
      run( aPackage, null );
    }

    @SuppressWarnings({"ConstantConditions"})
    public final void run(final PollingPackage aPackage, final Properties properties) {
        final ProfilerService profilerService = ProfilerServiceFactory.getInstance();
        final String profilerPollName = aPackage.getPackageName() + "." + getDataSourceName();
        synchronized (this) {
            if (currentlyPolling.contains(aPackage)) {
                if (logger.isDebugEnabled())
                    logger.debug(getDataSourceName() + " - Data source is already performing a poll of " + aPackage.getPackageName());

                return;
            }

            currentlyPolling.add(aPackage);
        }

        final long starttime = System.currentTimeMillis();
        try {
            profilerService.log("Package Poll Start", profilerPollName);
        } catch (ProfilerServiceException pse) {
            logger.warn(getDataSourceName() + " - Failed to profile poll start for " + aPackage.getPackageName());
        }

        try {
            startingAcquisition(aPackage);

            final ClassLoader previousClassLoader = Thread.currentThread().getContextClassLoader();

            try {
                // Set the right class loader in the thread's context
                Thread.currentThread().setContextClassLoader(aPackage.getConfiguration().getMetaData().getPackageClassLoader());

                //set thread local context for this thread with this package.
                PackageContext.initialize(aPackage);

                if (logger.isDebugEnabled())
                    logger.debug(getDataSourceName() + " [" + aPackage.getPackageName() + "] - Started acquisition");

                // Now call the data source's implementation to perform the acquisition
                aPackage.pollNow( properties );

                final List<DataSourceSampleListener> listenersForPackage = sampleListeners.get(aPackage.getPackageName());
                if (listenersForPackage != null) {
                    for (final DataSourceSampleListener sampleListener : listenersForPackage)
                        sampleListener.sampleSuccessfull(this, aPackage);
                }
            }
            catch (DataSourceErrorException e) {
                reportError(e, aPackage);
            }
            catch (FrameworkException fe) {
                // Handle duplicate data source
                final Throwable cause = fe.getCause();
                if (isDuplicateDatasourceException(cause)) {
                    handleDuplicateDatasource(aPackage, (AcquisitionUnitException)cause);
                } else {
                    handleThrowable(aPackage, fe);
                }
            }
            catch (Throwable t) {
                handleThrowable(aPackage, t);
            }
            finally {
                // Restore the previous class loader for this thread
                Thread.currentThread().setContextClassLoader(previousClassLoader);
                PackageContext.remove();
            }

            finishedAcquisition(aPackage);
        }
        finally {
            logger.info(aPackage.getLogPrefix() + " - Finished acquisition (" + Long.toString(System.currentTimeMillis() - starttime) + " msecs)");
            try {
                profilerService.log("Package Poll Finish", profilerPollName);
            } catch (ProfilerServiceException pse) {
                logger.warn(getDataSourceName() + " - Failed to profile poll end for " + aPackage.getPackageName());
            }

            // At the very end of the poll, enable other threads to poll this poll type again.
            synchronized (this) {
                currentlyPolling.remove(aPackage);
            }
        }
    }

    private boolean isDuplicateDatasourceException(final Throwable t) {
        if (!(t instanceof AcquisitionUnitException)) {
            return false;
        }
        final AcquisitionUnitException aue = (AcquisitionUnitException) t;
        final String aueMessage = aue.getMessage();
        return aueMessage.startsWith("Another datasource was found acquiring the same device");
    }

    private void handleThrowable(final PollingPackage aPackage, final Throwable t) {
        logger.fatal(aPackage.getLogPrefix() + " - Runtime exception: " + t.getMessage(), t);
        final EnhancedErrorDetail eed = new EnhancedErrorDetail("general_error.html", getDataSourceName(), t.getMessage());
        eed.addParam("runtime.detail", t.getClass().getCanonicalName() + " - " +  t.getMessage());
        final DataSourceErrorException errorException = DataSourceErrorException.createWithEnhanced(DataSourceErrorType.INTERNAL_ERROR, t.getMessage(), eed, t);
        // Note: Generally, we should not catch throwable at all but since we are working within a thread pool here,
        // not catching throwable will cause the thread pool to handle it incorrectly and we will have an inconsistent
        // behavior.
        reportError(errorException, aPackage);

        // TODO: In case of Throwable other than RuntimeException, maybe cause the acquisition to restart???
    }

    private void handleDuplicateDatasource(final PollingPackage aPackage, final AcquisitionUnitException aue) {
        logger.fatal(aPackage.getLogPrefix() + " - Runtime exception: " + aue.getMessage(), aue);
        final EnhancedErrorDetail eed = new EnhancedErrorDetail("duplicate_data_source.html", getDataSourceName(), aue.getMessage());
        eed.addParam("runtime.detail", aue.getMessage());
        final DataSourceErrorException errorException = DataSourceErrorException.createWithEnhanced(DataSourceErrorType.INTERNAL_ERROR, aue.getMessage(), eed, aue);
        // Note: Generally, we should not catch throwable at all but since we are working within a thread pool here,
        // not catching throwable will cause the thread pool to handle it incorrectly and we will have an inconsistent
        // behavior.
        reportError(errorException, aPackage);
    }

    public final <P extends Package> P getPackage(final String packageName) {
        //noinspection unchecked
        final P aPackage = (P) packages.get(packageName);
        assert aPackage != null;

        return aPackage;
    }

    public final Map<String, Package> getPackages() {
        return Collections.unmodifiableMap(packages);
    }

    public final void registerPackage(final Package aPackage) {
        assert status == DataSourceStatus.Stopped;
        assert !packages.containsKey(aPackage.getPackageName());

        packages.put(aPackage.getPackageName(), aPackage);
    }

    /**
     * Call back allowing libraries of specific datasources (e.g. process framework, ssh utils, etc.) to
     * register datasources under their specific datasource and package.  Since libraries are technically
     * "per package", this context must be
     *
     * @param library library to register
     * @param p package under which library is running.
     */
    public final void registerLibraryMBean(final PackageLibraryMBean library, final Package p) {
        try {
            final ObjectName on = formatPackageLibraryObjectName(library, p);
            if (ManagementFactory.getPlatformMBeanServer().isRegistered(on)) return;
            ManagementFactory.getPlatformMBeanServer().registerMBean(library, on);
            libraryMBeans.put(library.getLibraryName(), on);
        }
        catch (Exception e) {
            logger.error(getDataSourceName() + " - Failed to register library MBean: " + e.getMessage(), e);

            // Fall through...
        }
    }


    /**
     * Register the data source and its packages as MBeans
     */
    private void registerMBeans() {
        // No MBeans in standalone mode
        if (ServiceManager.getInstance().isStandalone())
            return;
        try {

            unregisterMBeans(); //clean up previous registrations.
            
            ManagementFactory.getPlatformMBeanServer().registerMBean(this, formatDataSourceObjectName());

            for (final Package aPackage : packages.values())
                ManagementFactory.getPlatformMBeanServer().registerMBean(aPackage, formatPackageObjectName(aPackage));
        }
        catch (Exception e) {
            logger.error(getDataSourceName() + " - Failed to register MBeans: " + e.getMessage(), e);

            // Fall through...
        }
    }

    /**
     * Unregister all the MBeans created by {@link #registerMBeans()}
     */
    private void unregisterMBeans() {
        // No MBeans in standalone mode
        if (ServiceManager.getInstance().isStandalone())
            return;        
        try {
            ManagementFactory.getPlatformMBeanServer().unregisterMBean(formatDataSourceObjectName());

            for (final Package aPackage : packages.values())
                ManagementFactory.getPlatformMBeanServer().unregisterMBean(formatPackageObjectName(aPackage));

            for (final ObjectName on : libraryMBeans.values())
                ManagementFactory.getPlatformMBeanServer().unregisterMBean(on);
        }
        catch (Exception e) {
            if (logger.isDebugEnabled()) {
                logger.debug(getDataSourceName() + " - Failed to un-register MBeans: " + e.getMessage());
            }
            // Fall through...
        }
    }

    /**
     * Format an MBean object name for this poll type
     *
     * @param aPackage The package reporting this event
     *
     * @return Object name for this package.
     *
     * @throws MalformedObjectNameException if bad object name.
     */
    private ObjectName formatPackageObjectName(final Package aPackage) throws MalformedObjectNameException {
        final Hashtable<String, String> props = new Hashtable<String, String>();
        props.put("name", aPackage.getPackageName());
        props.put("type", getFormattedDataSourceMBeanName());

        return new ObjectName("DataSources", props);
    }

    private ObjectName formatPackageLibraryObjectName(final PackageLibraryMBean library, final Package p) throws MalformedObjectNameException {
        final Hashtable<String, String> props = new Hashtable<String, String>();
        props.put("name", p.getPackageName() + "." + library.getLibraryName());
        props.put("type", getFormattedDataSourceMBeanName());

        return new ObjectName("DataSources", props);
    }

    /**
     * Format an MBean object name for this data source
     *
     * @return Object Name for the data source.
     *
     * @throws MalformedObjectNameException if bad object name.
     */
    private ObjectName formatDataSourceObjectName() throws MalformedObjectNameException {
        final Hashtable<String, String> props = new Hashtable<String, String>();
        props.put("name", "datasource");
        props.put("type", getFormattedDataSourceMBeanName());

        return new ObjectName("DataSources", props);
    }

    /**
     *  Returns the user entered data source name filtered for MBean object name usage as much as possible.
     * Return is the name quoted with certain wild-card characters escaped (?, *).
     *
     * @return the data source name formatted for object name usage.
     */
    private String getFormattedDataSourceMBeanName () {
        final String name =  getDataSourceName();
        final StringBuilder formattedName = new StringBuilder();
        for (final char c : name.toCharArray()) {
            switch (c) {
                case '?':
                case '*':
                    formattedName.append("\\");
                    formattedName.append(c);
                    break;
                default:
                    formattedName.append(c);
            }
        }
        return "\"" + formattedName.toString() + "\"";
    }

    /**
     * @return true if the data source is stopped or stopping
     */
    public final boolean isStopped() {
        return this.status.equals(DataSourceStatus.Stopping) || this.status.equals(DataSourceStatus.Stopped);
    }

    public final String getDataSourceName() {
        return configuration.getDataSourceName();
    }

    /**
     * Returns the data source ID. This is the objid of the data source entry in the data_source table in the
     * database.
     */
    public final long getDataSourceID() {
        return configuration.getDataSourceID();
    }

    public final DataSourceStatus getStatus() {
        return status;
    }

    /**
     * Used for JMX management
     */
    public final String getStatusString() {
        return status.toString();
    }

    /**
     * Report back to the listener that the data source is starting acquisition. This will open the audit record for
     * this sample.
     *
     * @param aPackage The package reporting this event
     */
    protected final void startingAcquisition(final PollingPackage aPackage) {
        // Internally change the status of the data source back to stand-by. We don't send the update to the server because
        // we don't want the status in the GUI to change from the last sample.
        if (logger.isDebugEnabled())
            logger.debug(getDataSourceName() + " - Data source internally changed status from '" + this.status + "' to '" + DataSourceStatus.Aqcuiring + "'");

        this.status = DataSourceStatus.Aqcuiring;
        addToStatusHistory(DataSourceStatus.Aqcuiring.toString());

        final List<DataSourceSampleListener> listenersForPackage = sampleListeners.get(aPackage.getPackageName());
        if (listenersForPackage != null) {
            for (final DataSourceSampleListener sampleListener : listenersForPackage)
                sampleListener.startingSample(this, aPackage);
        }

        try {
            // Add a special log4j appender for this sample. This appender will write all of this sample log entries into a
            // special log file that will be wrapped into the ZIP file of this sample.
            if (aPackage.getConfiguration().isRecording()) {
                final DSFileWriter sampleLogFile = aPackage.getRecordingSession().getRootSample().createTextualOutputFile(aPackage.getConfiguration(),
                                                                                                                          DataSourceOutputFileType.log,
                                                                                                                          "sample",
                                                                                                                          "log");

                ServiceManager.getInstance().getLogsManager().addSampleAppender(Thread.currentThread().getName(), sampleLogFile);
            }
        }
        catch (IOException e) {
            logger.error(aPackage.getLogPrefix() + " - Failed to create private sample log file", e);
            // Fall through...
        }
    }

    /**
     * Report back to the listener that the data source has finished an acquisition cycle
     *
     * @param aPackage The package reporting this event
     */
    protected final void finishedAcquisition(final PollingPackage aPackage) {
        if (!this.status.equals(DataSourceStatus.StandBy) && !isStopped()) {
            // Internally change the status of the data source back to stand-by. We don't send the update to the server because
            // we don't want the status in the GUI to change from the last sample.
            if (logger.isDebugEnabled())
                logger.debug(getDataSourceName() + " - Data source internally changed status from '" + this.status + "' to '" + DataSourceStatus.StandBy + "'");

            this.status = DataSourceStatus.StandBy;
            addToStatusHistory("Finished acquisition");
        }

        if (aPackage.getConfiguration().isRecording()) {
            // Remove the private sample log appender
            ServiceManager.getInstance().getLogsManager().removeSampleAppender(Thread.currentThread().getName());
        }

        // Now let the framework really close the audit record for this sample
        final List<DataSourceSampleListener> listenersForPackage = sampleListeners.get(aPackage.getPackageName());
        if (listenersForPackage != null) {
            for (final DataSourceSampleListener sampleListener : listenersForPackage) {
                try {
                    sampleListener.finishedSample(this, aPackage);
                }
                catch (FrameworkException e) {
                    logger.error(getDataSourceName() + " - failed to report sample finished...", e);
                }
            }
        }
    }

    public void recordingSessionDone(final PollingPackage aPackage,
                                       final File recordingSession,
                                       final boolean shouldReport) {
        final List<DataSourceSampleListener> listenersForPackage = sampleListeners.get(aPackage.getPackageName());
        if (listenersForPackage != null) {
            for (final DataSourceSampleListener sampleListener : listenersForPackage) {
                try {
                    sampleListener.recordingSessionDone(this, aPackage, recordingSession, shouldReport);
                }
                catch (FrameworkException e) {
                    logger.error(getDataSourceName() + " - failed to report sample recording session...", e);
                }
            }
        }
    }

    /**
     * Report back to the listener any change in the data source status
     *
     * @param newStatus new datasource status.
     *
     * @param message Change status message.
     */
    protected final void changeStatus(final DataSourceStatus newStatus, final String message) {
        if (logger.isDebugEnabled())
            logger.debug(getDataSourceName() + " - Data source changed status from '" + this.status + "' to '" + newStatus + "' (" + message + ")");

        final DataSourceStatus oldStatus = this.status;

        this.status = newStatus;
        addToStatusHistory(newStatus.toString() + " : " + message);

        for (final DataSourceLifeCycleListener listener : lifeCycleListeners)
            listener.dataSourceStatusChanged(this, oldStatus, newStatus, message);
    }

    /**
     * Used to report an error for a package
     *
     * <p/>Called whenever we catch a data source error exception. This is the type of exception that the data source will
     * throws when encountering an error condition.
     *
     * <p/>This will also take care of reporting the error details to the server to be later displayed in the GUI.
     *
     * @param error The exception object
     * @param aPackage The package reporting this event
     */
    public void reportError(final DataSourceErrorException error, final PollingPackage aPackage) {
        // If we're stopping right now, ignore this report!
        if (isStopped())
            return;

        if (logger.isDebugEnabled())
            logger.debug(getDataSourceName() + " - Data source changed status from " + this.status + " to " + DataSourceStatus.StandBy);
        //populate the exception with package instance
        if (error instanceof AggregatedDataSourceErrorsException) {

            for (final DataSourceErrorException exc : ((AggregatedDataSourceErrorsException) error).getExceptions()) {
                exc.setPackage(aPackage);
            }
        }
        else
            error.setPackage(aPackage);

        status = DataSourceStatus.StandBy;

        // Let the exception object print itself to the log
        error.printToLog(logger, getDataSourceName());

        addToStatusHistory("[" + aPackage.getPackageName() + "] - ds error: " + String.valueOf(error.getMessage()));

        final List<DataSourceSampleListener> listenersForPackage = sampleListeners.get(aPackage.getPackageName());
        if (listenersForPackage != null) {
            for (final DataSourceSampleListener listener : listenersForPackage)
                listener.sampleFailed(this, aPackage, error);
        }
    }

    /**
     * Used to report an error for a package
     *
     * <p/>Called whenever we catch a data source error exception. This is the type of exception that the data source will
     * throws when encountering an error condition.
     *
     * <p/>This will also take care of reporting the error details to the server to be later displayed in the GUI.
     *
     * @param error The exception object
     * @param aPackage The package reporting this event
     */
    public void reportError(final Throwable error, final PollingPackage aPackage) {
        // If we're stopping right now, ignore this report!
        if (isStopped())
            return;

        if (logger.isDebugEnabled())
            logger.debug(getDataSourceName() + " - Data source changed status from " + this.status + " to " + DataSourceStatus.StandBy);

        status = DataSourceStatus.StandBy;

        addToStatusHistory("[" + aPackage.getPackageName() + "] - framework exception: " + error.getMessage());

        // Log the error
        logger.error(getDataSourceName() + "[" + aPackage.getPackageName() + "] - " + error.getMessage(), error);

        final List<DataSourceSampleListener> listenersForPackage = sampleListeners.get(aPackage.getPackageName());
        if (listenersForPackage != null) {
            // TODO: This should be done also for non polling packages (maybe transferred to LifeCycle listener)
            for (final DataSourceSampleListener listener : listenersForPackage)
                listener.sampleFailed(this, aPackage, error);
        }
    }

    public String toString() {
        return getDataSourceName() + ":" + getDataSourceID() + ":" + configuration.getMetaDataName();
    }

    public final void addToStatusHistory(final String status) {
        this.statusHistory.add(new Date().toString() + " - " + status);
    }

    public final String[] getStatusHistory() {
        final String[] history = new String[statusHistory.size()];
        return statusHistory.toArray(history);
    }
}
