package com.onaro.sanscreen.acquisition.framework.datasource;

import com.onaro.util.CrcCalculator;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import java.util.HashMap;
import java.util.Map;


/**
 * Base class for all data source errors<br><br>
 *
 * <p/>This class contains enough information for the framework to format a proper error message and report it back to
 * the GUI.
 */
public class DataSourceErrorException extends Exception {
    protected final static String crlf = System.getProperty("line.seperator");

    private EnhancedErrorDetail enhancedErrorDetail;

    private DataSourceErrorType type;
    private String logMessage;
    private String deviceKey;
    private long exceptionCRCValue = -1;
    private boolean isValidCRC = false;

    /**
     * Empty constructor.
     *
     * @see AggregatedDataSourceErrorsException
     */
    protected DataSourceErrorException() {
    }

    /**
     * Factory Method. Created in order to prevent ambiguos references with the old constructors.
     * In this way we can keep the well known order of paramameters that we are used to work with, without conflicts.
     * For example, we have to make a difference in the following case, otherwise the compiler will complain about ambigous-references:
     *
     * <blockquote>
     * DataSourceErrorException(DataSourceErrorType.RETRIEVE_DATA_FROM_DEVICE, "Hello", null);
     * </blockquote>
     *
     * In the above case, either of these constructors apply for the call:
     *
     * <blockquote>
     * DataSourceErrorException(DataSourceErrorType, String, String);
     * DataSourceErrorException(DataSourceErrorType, String, EnhancedErrorDetail);
     * </blockquote>
     */
    public static DataSourceErrorException createWithEnhanced(final DataSourceErrorType type, final String logMessage, final EnhancedErrorDetail enhanceErrorDetail) {
        return new DataSourceErrorException(type, enhanceErrorDetail, logMessage, null);
    }

    /**
     * Factory Method. Created in order to prevent ambiguos references with the old constructors.
     * In this way we can keep the well known order of paramameters that we are used to work with, without conflicts.
     * For example, we have to make a difference in the following case, otherwise the compiler will complain about ambigous-references:
     *
     * <blockquote>
     * DataSourceErrorException(DataSourceErrorType.RETRIEVE_DATA_FROM_DEVICE, "Hello", null);
     * </blockquote>
     *
     * In the above case, either of these constructors apply for the call:
     *
     * <blockquote>
     * DataSourceErrorException(DataSourceErrorType, String, String);
     * DataSourceErrorException(DataSourceErrorType, String, EnhancedErrorDetail);
     * </blockquote>
     */
    public static DataSourceErrorException createWithEnhanced(final DataSourceErrorType type, final String logMessage, final EnhancedErrorDetail enhanceErrorDetail, final Throwable cause) {
        return new DataSourceErrorException(type, enhanceErrorDetail, logMessage, cause);
    }

    /**
     * Factory Method. Created only to be consistent with the new factory methods that creates {@link EnhancedErrorDetail }.
     */
    public static DataSourceErrorException createWithoutEnhanced(final DataSourceErrorType type, final String guiMessage, final String logMessage) {
        return new DataSourceErrorException(type, guiMessage, logMessage);
    }

    /**
     * Factory Method. Created only to be consistent with the new factory methods that creates {@link EnhancedErrorDetail }.
     */
    public static DataSourceErrorException createWithoutEnhanced(final DataSourceErrorType type, final String guiMessage, final String logMessage, final Throwable cause) {
        return new DataSourceErrorException(type, guiMessage, logMessage, cause);
    }

    /**
     * Constructs an error message with all the mandatory data
     *
     * @param type
     * @param guiMessage
     * @param logMessage
     */
    public DataSourceErrorException(final DataSourceErrorType type, final String guiMessage, final String logMessage) {
        super(guiMessage);
        this.type = type;
        this.logMessage = logMessage;
        enhancedErrorDetail = new EnhancedErrorDetail("defaultMessage.html", null, guiMessage);
    }

    public DataSourceErrorException(final DataSourceErrorType type, final String guiMessage, final String logMessage, final Throwable cause) {
        super(guiMessage, cause);
        this.type = type;
        this.logMessage = logMessage;
        enhancedErrorDetail = new EnhancedErrorDetail("defaultMessage.html", null, guiMessage);
    }

    /**
     * @return <code>true</code> if this error contains enhanced error information
     */
    public boolean containsEnhancedErrorDetail() {
        return (enhancedErrorDetail != null);
    }

    public String getEnhancedErrorDetail() {
        if (enhancedErrorDetail == null)
            return null;

        return enhancedErrorDetail.getMessage();
    }

    public DataSourceErrorType getType() {
        return type;
    }

    public String getMessage() {
        if (enhancedErrorDetail != null)
            return enhancedErrorDetail.getGuiMessage();

        return super.getMessage();
    }

    /**
     * This message is sent to the log file. It usually contains more detailed / internal information than the message
     * sent to the screen.
     */
    public String getLogMessage() {
        if (logMessage == null)
            return getMessage();

        if (deviceKey == null)
            return logMessage;

        return "[Device name " + deviceKey + "]: " + logMessage;
    }

    /**
     * Reports the error details to a string builder. This is used to be able to construct the exception as a string and
     * then later be identified compared to previously reported errors.
     *
     * @param sb StringBuilder to put result in
     */

    public void printToBuffer(final StringBuilder sb) {
        // Append the header of the exception
        sb.append(" [").append(getType().getMessage()).append("] - ").append(getMessage()).append((getLogMessage() != null ? " (" + getLogMessage() + ")" : "")).append(crlf);

        // Append extended error details (if available)
        if (enhancedErrorDetail != null)
            sb.append(enhancedErrorDetail.getMessage()).append(crlf);

        printStackStraceToBuffer(sb, this);
    }

    /**
     * Given an exception object, goes through the entire stack trace and print it
     *
     * @param sb
     * @param e
     */
    public static void printStackStraceToBuffer(final StringBuilder sb, final Throwable e) {
        // And finally, append the call stack of this error
        for (StackTraceElement stackTraceElement : e.getStackTrace())
            sb.append("\t").append(stackTraceElement.toString()).append(crlf);

        Throwable cause = e.getCause();
        while (cause != null) {
            sb.append("Caused By: ").append(cause.toString()).append(crlf);

            for (StackTraceElement stackTraceElement : e.getStackTrace())
                sb.append("\t").append(stackTraceElement.toString()).append(crlf);

            cause = cause.getCause();
        }
    }

    /**
     * Calculate crc for the exception
     *
     * @return
     */
    public long getExceptionCRC() {
        if (!isValidCRC) {
            final StringBuilder sb = new StringBuilder(1000);
            printToBuffer(sb);

            exceptionCRCValue = CrcCalculator.calculateCRC(sb);
            isValidCRC = true;
        }

        return exceptionCRCValue;
    }

    protected boolean isValidCRC() {
        return isValidCRC;
    }

    /**
     * Setting the validCRC to false will force calculating the exception crc value.
     */
    protected void invalidateCrcCache() {
        isValidCRC = false;
    }

    /**
     * Reports the error details to the log...
     *
     * @param logger Logger to use
     */
    public void printToLog(final Logger logger, final String dataSourceName) {
        // Report this error to the log first (the type of report is taken from the error type - this will allow automatic scripting to distinguish between
        // different types of errors using the log level).
        final Exception expToLog = (getType().getReportLevel() == Level.ERROR || getType().getReportLevel() == Level.FATAL) ? this : null;
        logger.log(getType().getReportLevel(), dataSourceName + " [" + getType().getMessage() + "] - " + getMessage() + (getLogMessage() != null ? " (" + getLogMessage() + ")" : ""), expToLog);
    }

    /**
     * Creates a instance of this class based on a given {@link EnhancedErrorDetail } and {@link Throwable}.
     * references.
     */
    protected DataSourceErrorException(final DataSourceErrorType type,
                                     final EnhancedErrorDetail enhanceErrorDetail,
                                     final String logMessage,
                                     final Throwable cause) {
        super(enhanceErrorDetail.getGuiMessage(), cause);
        this.type = type;
        this.logMessage = logMessage;
        this.enhancedErrorDetail = enhanceErrorDetail;
        this.deviceKey = enhanceErrorDetail.getDeviceName();
    }

 
    protected void setEnhancedErrorDetail(final EnhancedErrorDetail enhancedErrorDetail) {
        this.enhancedErrorDetail = enhancedErrorDetail;
    }

    protected void setPackage(final PollingPackage pakage) {
        if (enhancedErrorDetail != null)
            enhancedErrorDetail.getTemplateConverter().setThePackage(pakage);

    }

    protected void setType(DataSourceErrorType type) {
        this.type = type;
    }

    protected void setLogMessage(String logMessage) {
        this.logMessage = logMessage;
    }

    public void setDeviceKey(final String deviceKey) {
        this.deviceKey = deviceKey;
    }

    public String getDeviceKey() {
        return deviceKey;
    }

    public Map<String, String> getTemplateConverterParameters() {
        if (enhancedErrorDetail == null) {
            return new HashMap<String, String>();
        }
        final EnhancedErrorDetail.TemplateConverter templateConverter = enhancedErrorDetail.getTemplateConverter();
        if (templateConverter == null) {
            return new HashMap<String, String>();
        }
        return templateConverter.getParameters();
    }
}
