package com.onaro.sanscreen.acquisition.framework.datasource;

import org.apache.log4j.Level;

/**
 * Enumeration of all different error types a data source could generate.
 *
 * <p/>The purpose of this Enum is to provide a uniform error message formatting (as much as possible). All error messages
 * will begin with the same prefix, and then contained a more detailed and specific description.
 *
 * <p/>Each type has a log level attached to it. The basic rule is that an error that is persistent and we have to fix something
 * in order to solve it, should be reported as a log error. Other events are warnings.
 */
public enum DataSourceErrorType {
    INITIALIZING("Error initializing", Level.ERROR),
    CONNECTING("Error connecting", Level.ERROR),
    INVALID_LOGIN_CREDENTIALS("Invalid login credentials", Level.ERROR),
    VERSION_MISMATCH("Version mismatch", Level.ERROR),
    RETRIEVE_DATA_FROM_DEVICE("Error retrieving data", Level.ERROR),
    UNEXPECTED_DATA_FROM_DEVICE("Unexpected data retrieved", Level.ERROR),
    INCONSISTENT_DATA("Inconsistent data retrieved", Level.ERROR),
    EXECUTING_EXTERNAL_RECORDING_UTILITY("Failed to execute external utility", Level.ERROR),
    INTERNAL_ERROR("Internal error", Level.ERROR),
    TIMEOUT_ERROR("Timeout", Level.ERROR),
    LICENSE_ERROR("Licensing problem", Level.ERROR),
    FAILED_TO_SEND_TO_SERVER("Failed to update server", Level.ERROR),
    PACKAGE_DEPENDENCY("Package dependency", Level.WARN),
    CONFIGURATION("Configuration error", Level.WARN),
    PACKAGE_SKIPPED("Package skipped", Level.WARN),
    /**
     * This type is used in conjunction with Multiple OID error handling.
     * @see AggregatedDataSourceErrorsException
     */
    PARTIAL_SUCCESS("Partial Success", Level.ERROR),
    FULL_FAILURE("All devices failed", Level.ERROR),
    PROCESS_TIMEOUT("External process run timed out", Level.ERROR),
    PROCESS_DATASOURCE_STOP("External process destroyed due to a DataSource stop.", Level.WARN);

    private String message;
    private Level reportLevel;

    DataSourceErrorType(final String message, final Level reportLevel) {
        this.message = message;
        this.reportLevel = reportLevel;
    }

    public String getMessage() {
        return message;
    }

    public Level getReportLevel() {
        return reportLevel;
    }
}
