package com.onaro.sanscreen.acquisition.framework.datasource.config;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Map;
import java.util.Properties;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSInputStream;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSSessionFileReader;
import com.onaro.sanscreen.acquisition.framework.datasource.metadata.DataSourceMetaData;
import com.onaro.sanscreen.acquisition.framework.datasource.metadata.PackageMetaData;
import com.onaro.sanscreen.acquisition.framework.mgmt.DeployManager;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Factory class to instantiate configuration classes based on the different use cases in the system
 */
public class ConfigurationFactory {

    private static final Logger LOG = LoggerFactory.getLogger(ConfigurationFactory.class);

    private final String dataSourceName;
    private final long dataSourceID;
    private final boolean offlineTraining;

    public ConfigurationFactory(final String dataSourceName, final long dataSourceID) {
        this.dataSourceName = dataSourceName;
        this.dataSourceID = dataSourceID;
        this.offlineTraining = Boolean.parseBoolean(System.getProperty("offline.training", Boolean.FALSE.toString()));
    }

    /**
     * Create a data source configuration using the normal creating process (as done from the GUI). This will automatically
     * detect for each package if it needs to simulate or not.
     *
     * @param dataSourceType the data source type
     * @param properties     the data source properties
     * @return A configuration object
     * @throws com.onaro.sanscreen.acquisition.framework.FrameworkException
     *          If could not find data source type
     */
    public DataSourceConfiguration create(final long dataSourceType, final Properties properties) throws FrameworkException {

        final DataSourceMetaData metaData = lookupMetaData(dataSourceType);

        // Finally, create the DS configuration object
        final DataSourceConfiguration configuration = new DataSourceConfiguration(metaData, dataSourceName, dataSourceID);


        // Now decide how to create this datasource (normal or simulation).
        final boolean simulating = prepareForSimulation(properties, metaData);

        // Iterate over the packages and create a configuration for each one
        for (final PackageMetaData packageMetaData : metaData.getPackages()) {
            final Properties packageProperties = filterPropertiesForPackage(packageMetaData.getName(), properties);

            // Check "enabled" property
            final boolean packageEnabled = Boolean.parseBoolean(packageProperties.getProperty("enabled", Boolean.toString(true)));
            if (!packageEnabled) {
                continue;
            }

            final PackageConfiguration<?> packageConfiguration;
            if (offlineTraining || simulating) {
                packageConfiguration = createPackageForGuiSimulation(configuration,
                        packageMetaData.getName(),
                        packageProperties,
                        dataSourceType);

                if (packageConfiguration == null) {
                    continue;
                }
            } else {
                packageConfiguration = createPackageNormally(configuration,
                        packageMetaData.getName(),
                        packageProperties,
                        dataSourceType);
            }

            configuration.registerPackageConfiguraiton(packageConfiguration);
        }

        return configuration;
    }

    /**
     * Setup simulation properties, so in case of simulation that all packages should point to the same file

     * @param properties configuration properties
     * @param metaData used to enumerate packages
     * @return true if datasource should be running in simulating mode
     */
    private boolean prepareForSimulation(final Properties properties, final DataSourceMetaData metaData) {
        boolean simulating = false;
        String simulationFolder = null;
        String simulationFile = null;

        for (final Map.Entry<Object, Object> entry : properties.entrySet()) {
            if (((String) entry.getKey()).endsWith("transport.simulate") && Boolean.parseBoolean((String) entry.getValue())) {
                simulating = true;
                continue;
            }
            if (((String) entry.getKey()).endsWith("transport.simulate.dir")) {
                simulationFolder = (String) entry.getValue();
                continue;
            }
            if (((String) entry.getKey()).endsWith("transport.simulate.file")) {
                simulationFile = (String) entry.getValue();
            }
        }

        if (simulating) {
            for (final PackageMetaData packageMetaData : metaData.getPackages()) {
                // Setup simulation properties
                properties.setProperty(packageMetaData.getName() + ".transport.simulate", Boolean.TRUE.toString());
                properties.setProperty(packageMetaData.getName() + ".transport.simulate.dir", simulationFolder);
                properties.setProperty(packageMetaData.getName() + ".transport.simulate.file", simulationFile);
            }
        }
        return simulating;
    }

    /**
     * Creates a data source configuration (empty) that will be used when simulating from recording files
     *
     * @param dataSourceType the data source type
     * @return a {@link DataSourceConfiguration} object initialized with the packages indicated in the sample files
     * @throws FrameworkException throws when encountering framework level error
     */
    public DataSourceConfiguration createEmpty(final long dataSourceType) throws FrameworkException {
        // Get the first data source type and use it to obtain a meta-data object
        final DataSourceMetaData metaData = lookupMetaData(dataSourceType);
        return new DataSourceConfiguration(metaData, dataSourceName, dataSourceID);
    }

    /**
     * Create a configuration object based on a deployed data source and configuration properties.
     * This method will use the {@link DeployManager#lookup} to retrieve the appropriate meta data information for the data source type.
     *
     * @param configuration  the data source configuration
     * @param packageName    the package name
     * @param properties     the data source properties
     * @param datasourceType the data source type
     * @return A configuration object
     * @throws com.onaro.sanscreen.acquisition.framework.FrameworkException
     *          If could not find data source type
     */
    public PackageConfiguration createPackageNormally(final DataSourceConfiguration configuration,
                                                      final String packageName,
                                                      final Properties properties,
                                                      final long datasourceType) throws FrameworkException {
        final PackageMetaData packageMetaData = lookupPackageMetaData(datasourceType, packageName);

        final PackageConfiguration<?> packageConfiguration = packageMetaData.instantiateConfiguration(configuration,
                dataSourceName,
                dataSourceID,
                properties);

        // Indicate that we're NOT simulating
        packageConfiguration.setSimulationSession(null);

        return packageConfiguration;
    }

    /**
     * Create a configuration object based on a give deploy descriptor file. Used for standalone.
     * <p/>
     * <p/>This method is different because the deployee XML is provided to avoid the need for {@link DeployManager}
     * (which is not available when running as standalone). No class loaders will be used here since all the relevant
     * JARs are already loaded into the system class loader by the standalone batch file.
     *
     * @param properties       the data source properties
     * @param deployDescriptor Direct link to the deploy descriptor file
     * @return The configuration object
     * @throws FrameworkException  throws when encountering framework level error
     * @throws java.io.IOException throws when communication error happens
     */
    public DataSourceConfiguration createForStandalone(final Properties properties, final File deployDescriptor) throws FrameworkException, IOException {
        final DataSourceMetaData metaData = DataSourceMetaData.createWithClassLoadersDisabled(deployDescriptor.getPath(), new FileInputStream(deployDescriptor) );

        final DataSourceConfiguration configuration = new DataSourceConfiguration(metaData, dataSourceName, dataSourceID);

        // Iterate over the packages and create a configuration for each one
        for (final PackageMetaData packageMetaData : metaData.getPackages()) {
            final Properties packageProperties = filterPropertiesForPackage(packageMetaData.getName(), properties);

            final PackageConfiguration<?> packageConfiguration = packageMetaData.instantiateConfiguration(configuration,
                    dataSourceName,
                    dataSourceID,
                    packageProperties);

            // Indicate that we're NOT simulating
            packageConfiguration.setSimulationSession(null);

            configuration.registerPackageConfiguraiton(packageConfiguration);
        }

        return configuration;
    }

    public DataSourceConfiguration createForStandalone(final String packageName, final Properties properties, final File deployDescriptor) throws FrameworkException, IOException {


        final Properties packageProperties = filterPropertiesForPackage(packageName, properties);

        // Create datasource
        final DataSourceMetaData metaData = DataSourceMetaData.createWithClassLoadersDisabled(deployDescriptor.getPath(), new FileInputStream(deployDescriptor) );
        final DataSourceConfiguration configuration = new DataSourceConfiguration(metaData, dataSourceName, dataSourceID);

        // Create and register configuration for the only package defined in session
        final PackageMetaData packageMetaData = metaData.getPackage(packageName);
        final PackageConfiguration<?> packageConfiguration = packageMetaData.instantiateConfiguration(configuration,
                dataSourceName,
                dataSourceID,
                packageProperties);

        configuration.registerPackageConfiguraiton(packageConfiguration);
        // Indicate that we're NOT simulating
        packageConfiguration.setSimulationSession(null);
        return configuration;
    }


    /**
     * Create a configuration object that will simulate from a Sample File (ZIP) within the SANscreen GUI. To avoid mistakes,
     * this method verifies that the selected ZIP file conforms with the right DS type.
     * <p/>
     * <p/>Simulating from the GUI means that data source name, poll and type were already selected by the user. These
     * values will override whatever is in the ZIP file.
     *
     * @param configuration     the data source configuration
     * @param packageName       the package name
     * @param packageProperties Properties used to configure the data source in the GUI
     * @param datasourceType    the data source type
     * @return A configuration object for the package setup for simulation, <code>null</code> if the package cannot be simulated
     * @throws com.onaro.sanscreen.acquisition.framework.FrameworkException
     *          throws when framework level error occurs
     */
    private PackageConfiguration createPackageForGuiSimulation(final DataSourceConfiguration configuration,
                                                               final String packageName,
                                                               final Properties packageProperties,
                                                               final long datasourceType) throws FrameworkException {
        try {

            final PackageMetaData packageMetaData = lookupPackageMetaData(datasourceType, packageName);
            fillInDefaultValues(packageMetaData, packageProperties);

            final DSSessionFileReader session = loadSimulationFile(packageMetaData, dataSourceName, packageProperties);
            if (session == null) {
                return null;
            }

            final Properties effectiveProperties = session.getProperties();

            // Make sure package is enabled
            effectiveProperties.setProperty("enabled", Boolean.TRUE.toString());

            // Override poll value from zip to one set in GUI
            effectiveProperties.setProperty("poll", packageProperties.getProperty("poll"));

            // Override the filter from zip to one set in GUI
            final String exclude = packageProperties.getProperty("filter.exclude");
            if (exclude != null) {
                effectiveProperties.setProperty("filter.exclude", exclude);
            }

            // Make sure data source and simulation file are referring to the same type!
            if (datasourceType != session.getDataSourceType()) {
                throw new FrameworkException("Trying to load ZIP file created by data source type " + session.getDataSourceType() + " into data source type " + datasourceType);
            }

            final PackageConfiguration<?> packageConfiguration = packageMetaData.instantiateConfiguration(configuration,
                    dataSourceName,
                    dataSourceID,
                    effectiveProperties);

            packageConfiguration.setSimulationSession(session);

            return packageConfiguration;
        } catch (IOException e) {
            throw new FrameworkException("Failed to load simulation file", e);
        }
    }

    /**
     * Scans the meta data and adds default values for properties that are defined but missing in the configuration.
     *
     * @throws FrameworkException - exception
     */
    private void fillInDefaultValues(final PackageMetaData packageMetaData, final Properties packageProperties) throws FrameworkException {
        final Properties defaultProperties = packageMetaData.getDefaultProperties(true);

        final Enumeration<?> e = defaultProperties.propertyNames();
        while (e.hasMoreElements()) {
            final String propertyName = (String) e.nextElement();

            final String propertyDefaultValue = defaultProperties.getProperty(propertyName);

            // LIST properties define the list columns so they are always put in the properties
            if (propertyName.indexOf("LIST") > -1 && !packageProperties.containsKey(propertyName)) {
                packageProperties.setProperty(propertyName, propertyDefaultValue);
                continue;
            }

            if (!packageProperties.containsKey(propertyName) && propertyDefaultValue != null && propertyDefaultValue.length() > 0) {
                packageProperties.setProperty(propertyName, propertyDefaultValue);
            }
        }
    }

    /**
     * Instantiates a data source configuration for simulation from unit tests.
     * <p/>
     * <p/>Simulating from unit tests is different from GUI. In this case, we would like to take whatever is in the sample
     * file (ZIP) and use it to configure the data source. Data source type and name will override any other configuration.
     * <p/>
     * <p/>Note: This method will use the {@link com.onaro.sanscreen.acquisition.framework.mgmt.DeployManager} to find the right meta data for this configuration
     *
     * @param configuration        the data source configuration
     * @param session              the data source session file reader for recording
     * @param propertiesToOverride Use these properties to override any other properties loaded from the file
     * @return a package configuration object
     * @throws com.onaro.sanscreen.acquisition.framework.FrameworkException
     *          when a framework level error occurs
     */
    public PackageConfiguration createPackageForUnitTestSimulation(final DataSourceConfiguration configuration,
                                                                   final DSSessionFileReader session,
                                                                   final Properties propertiesToOverride) throws FrameworkException {
        // Setup the properties for this data source. We start with the properties in the simulation file and then override them
        // with whatever is neccessary to override.
        final Properties props = new Properties();
        props.putAll(session.getProperties());

        if (propertiesToOverride != null) {
            props.putAll(propertiesToOverride);
        }

        // fix old version
        props.remove("____");

        // Lookup the meta data using the type indicated in the simulation file
        final PackageMetaData packageMetaData = lookupPackageMetaData(session.getDataSourceType(), session.getPackageName());

        // And instantiate the data source (use the simulation file's data source name and ID)
        final PackageConfiguration<?> packageConfiguration = packageMetaData.instantiateConfiguration(configuration,
                dataSourceName,
                dataSourceID,
                props);

        // Set the simulation streams for this configuration!
        packageConfiguration.setSimulationSession(session);

        return packageConfiguration;
    }

    /**
     * Load the simulation file for this package.
     * <p/>
     * <p/>When a user selects a ZIP file for simulation in the GUI, the system will automatically add to the data source
     * properties information on how to locate the ZIP file (properties <code>transport.simulate</code>,
     * <code>transport.simulate.file</code> and <code>transport.simulate.dir</code>.
     * <p/>
     * <p/>When the system is set with jvm argument <code>offline.training</code> as <code>true</code>, then this code
     * will look for the ZIP files in a different way. The ZIP file directory will be pointed to via the jvm argument
     * <code>offline.training.scenario.dir</code>. The name of the ZIP file will be formatted like
     * <code>&lt;package name&gt;_&lt;ip address&gt;.zip</code>. Using this format we will allow the trainee to setup multiple data sources
     * of the same type.
     * <p/>
     * <p/>Offline training is designed to give the trainee a feeling that he is working on a real environment without
     * having to attach SANscreen to any real device. All data sources will work with simulation mode from a well-known
     * recording location (a.k.a scenario).
     *
     * @param packageMetaData   the package metadata info
     * @param dataSourceName    the data source name
     * @param packageProperties the package properties
     * @return Session file reader with the appropriate simulation file to be simulated by the package
     * @throws FrameworkException when a framework level error occurs
     * @throws IOException        when recording files related I/O error occurs
     */
    private DSSessionFileReader loadSimulationFile(final PackageMetaData packageMetaData, final String dataSourceName, final Properties packageProperties) throws FrameworkException, IOException {

        if (offlineTraining) {
            final String packageKey = dataSourceName + "/" + packageMetaData.getName();

            LOG.warn("{} - Running data source in training mode!", packageKey);

            final String defaultDir = new File(ServiceManager.getInstance().getLogsManager().getLogDirectory().getParentFile().getParentFile(), "training/scenario").getAbsolutePath();
            final File dir = new File(System.getProperty("offline.training.scenario.dir", defaultDir));
            if (!dir.exists()) {
                throw new FrameworkException("Offline training scenario for pacakge '" + packageKey + "': dir (" + dir.getAbsolutePath() + ") does not exist");
            }

            if (!dir.isDirectory()) {
                throw new FrameworkException("Offline training scenario for package '" + packageKey + "': dir (" + dir.getAbsolutePath() + ") is not a directory");
            }

            final String ip = packageProperties.getProperty("ip");
            if (ip == null || ip.length() == 0) {
                LOG.warn("{} - Offline training scenario for package '{}': file cannot be found (ip property not set)",
                        packageKey, packageKey);
                return null;
            }

            final File simulateFile = new File(dir, packageMetaData.getName() + "_" + ip + ".zip");
            if (!dir.exists()) {
                LOG.warn("{} - Offline training scenario for package '{}': file cannot be found ({})",
                        new Object[] { packageKey , packageKey, simulateFile.getAbsolutePath() });
                return null;
            }

            return new DSSessionFileReader(simulateFile);
        }

        // Read and remove the system properties from the collection
        final boolean transportSimulate = Boolean.parseBoolean((String) packageProperties.remove("transport.simulate"));
        final String transportSimulateFile = (String) packageProperties.remove("transport.simulate.file");
        final String transportSimulateDir = (String) packageProperties.remove("transport.simulate.dir");

        if (!transportSimulate) {
            throw new FrameworkException("Not in simulation mode (" + dataSourceName + " package '" + packageMetaData.getName() + "')");
        }

        if (transportSimulateFile == null || transportSimulateDir == null) {
            throw new FrameworkException("Inconsistent simulation configuration for data source " + dataSourceName + " package '" + packageMetaData.getName() + "'");
        }

        final File dir = new File(transportSimulateDir);
        if (!dir.exists() || !dir.isDirectory()) {
            throw new FrameworkException("Invalid configuration for simulation dir (does not exist or not a directory) - " + transportSimulateDir);
        }

        final File simulateFile = new File(dir, transportSimulateFile);

        final DSSessionFileReader sessionFileReader = new DSSessionFileReader(simulateFile);

        // If recording was created by the package we creating now - thats the session reader to return
        if (sessionFileReader.getPackageName().equals(packageMetaData.getName())) {
            return sessionFileReader;
        }

        // If we here it means that zip file is produced different package.
        // Currently we will handle only 2 use case: zip created by FoundationDependentPackage for "foundation" or "netappfoundation" simulation
        if (!packageMetaData.getName().equals("foundation") && !packageMetaData.getName().equals("netappfoundation")) {
            throw new FrameworkException("Invalid configuration for simulation (recording created by package we do not know how to simulate) - " + sessionFileReader.getPackageName());
        }

        // Now we need to extract foundation recording from inside a sample made by FoundationDependant package
        try {
            for (final DSSessionFileReader.Sample sample : sessionFileReader.getSamples()) {
                final DSSessionFileReader foundationReader = retrieveFoundationRecordingFromSample(sample);
                if (foundationReader != null) {
                    return foundationReader;
                }
            }

            //IBG-523: Recordings for datasources with disabled array performance fail when imported when it is enabled
            //If we reach this point then the simulation package does not contain any foundation information for the desired package.
            //Provide the user with enough information to determine the cause of the error. It may be as simple as disabling
            //Array Performance when running a simulation without performance information!
            throw new FrameworkException("Unable to load simulation package: " + packageMetaData.getName() +
                                         "! The provided simulation file does not contain enough information to reconstruct the desired package. " +
                                         "Check to confirm the datasource configuration is not attempting to run additional packages that are not" +
                                         " present in the simulation file; for example, array performance on a foundation-only simulation.");
        } catch (IOException e) {
            throw new FrameworkException("Invalid configuration for simulation (failed to retrieve foundation recording from first sample)");
        } finally {
            sessionFileReader.close();
        }
    }

    /**
     * Determine if passed recorded sample has an embedded foundation zip and retrieve it if so.   Returns NULL
     * if there is no foundation updated in this sample.
     *
     * @param sample the sample to check and retrieve foundation from.
     * @return the recording session representation of the foundation zip if present.  NULL if not in this sample.
     * @throws IOException for any problem dealing with the recording or retrieval.
     */
    public static DSSessionFileReader retrieveFoundationRecordingFromSample(final DSSessionFileReader.Sample sample) throws IOException {
        final String foundationRecordingFile = sample.getDescriptionProperty("foundation", "recording");
        if (foundationRecordingFile == null) {
            return null;
        }
        final File outputDir = ServiceManager.getInstance().getLogsManager().getAcqRecordingDir();
        DSSessionFileReader foundationZip = null;
        for (final DSInputStream stream : sample.getStreams()) {
            if (stream.getFileName().equals(foundationRecordingFile)) {
                //make a copy of this outside the zip file for use.
                final File outputZip = new File(outputDir, foundationRecordingFile);
                if (!outputZip.exists()) {
                    BufferedOutputStream out = null;
                    try {
                        out = new BufferedOutputStream(new FileOutputStream(outputZip));

                        final int bufferSize = 8 * 1024;
                        final byte[] buffer = new byte[bufferSize];
                        int count = stream.read(buffer);

                        while (count > 0) {
                            out.write(buffer, 0, count);
                            count = stream.read(buffer);
                        }
                        out.flush();

                    } catch (IOException ioe) {
                        // ignore
                    } finally {
                        if (out != null) {
                            try {
                                out.close();
                            } catch (IOException e) {

                                //ignore.
                            }
                        }
                    }
                }

                foundationZip = new DSSessionFileReader(outputZip);
            }
        }

        return foundationZip;
    }

    private DataSourceMetaData lookupMetaData(final long datasourceType) throws FrameworkException {
        // Use the deploy manager to lookup the meta data based on the data source type
        final DataSourceMetaData dataSourceMetaData = ServiceManager.getInstance().getDeployManager().lookup(datasourceType);
        if (dataSourceMetaData == null) {
            throw new FrameworkException("Could not find DS meta data for data source type " + datasourceType + " (check if data source type was deployed correctly)");
        }

        return dataSourceMetaData;
    }

    /*
     * The {@link DataSourceMetaData} contains the a list of all {@link PackageMetaData} objects this data source can
     * potentially run. This method will lookup the appropriate data source meta data and then inside the relevant package
     * meta data.
     */
    private PackageMetaData lookupPackageMetaData(final long datasourceType, final String packageName) throws FrameworkException {
        final PackageMetaData packageMetaData = lookupMetaData(datasourceType).getPackage(packageName);
        if (packageMetaData == null) {
            throw new FrameworkException("Could not find package meta data for data source type " + datasourceType + " package " + packageName + " (check datasource deploy descriptor file)");
        }

        return packageMetaData;
    }

    /**
     * Package properties start with the package name. When a data source is instantiated, all properties for all packages
     * are combined (as they come from the database). This method filters out all properties that do not belong to the given
     * package.
     *
     * @param packageName the package name
     * @param properties  the package properties
     * @return A new properties collection containing only properties for the given package
     */
    private Properties filterPropertiesForPackage(final String packageName, final Properties properties) {
        final Properties propsForPackage = new Properties();
        final String packagePrefix = packageName + ".";

        for (final Map.Entry<Object, Object> entry : properties.entrySet()) {
            final String key = (String) entry.getKey();
            if (key.startsWith(packagePrefix)) {
                // Add properties that have the package name as their prefix
                propsForPackage.put(key.substring(packagePrefix.length()), entry.getValue());
            } else if (key.startsWith("datasource.")) {
                // Add generic data source properties as well (like vendor, model, etc...)
                propsForPackage.put(key, entry.getValue());
            }
        }

        // All packages need visibility to the transport.simulate family of properties.
        // TODO: we would like to change the property prefix from "foundation." to "datasource.",
        // however when Client DataSourceAttributeWizardPage sets "datasource.transport.simulate"
        // instead of "foundation.transport.simulate", the property gets filtered somewhere before
        // we get here.   So for now, DataSourceAttributeWizardPage continues to provide the
        // "foundation." prefix, and we copy the values for each package.
        final String isSimulating = properties.getProperty("netappfoundation.transport.simulate");
        if (isSimulating != null) {
            propsForPackage.put("transport.simulate", isSimulating);
        }
        final String simulatingDir = properties.getProperty("netappfoundation.transport.simulate.dir");
        if (simulatingDir != null) {
            propsForPackage.put("transport.simulate.dir", simulatingDir);
        }
        final String simulatingFile = properties.getProperty("netappfoundation.transport.simulate.file");
        if (simulatingFile != null) {
            propsForPackage.put("transport.simulate.file", simulatingFile);
        }
        return propsForPackage;
    }
}
