package com.onaro.sanscreen.acquisition.framework.datasource.io;

import java.io.File;
import java.io.IOException;
import java.util.zip.CRC32;

/**
 * Extend DSFileWriter with CRC functionality for marshaling device models
 */
public class DSFileCRCWriter extends DSFileWriter implements DSFile {
    private final CRC32 allAttributesCRC = new CRC32();
    private final CRC32 requiredAttributesCRC = new CRC32();

    public DSFileCRCWriter(final String fileName, final File file) throws IOException {
        super(fileName, file);
    }


    /**
     * Marshall a required schema object attribute to the given writer, and update both CRCs
     * @param attributeName the name of the attribute to marshall
     * @param escapedAttributeValue the escaped value of the attribute to marshall
     * @throws IOException on IO error
     */
    public final void addRequiredAttributeAndCRC(final String attributeName, final String escapedAttributeValue) throws IOException {
        if (escapedAttributeValue == null) {
            return;
        }
        append(attributeName).append("=\"").append(escapedAttributeValue).append("\" ");
        final byte[] escapedAttributeBytes = escapedAttributeValue.getBytes();
        allAttributesCRC.update(escapedAttributeBytes);
        requiredAttributesCRC.update(escapedAttributeBytes);
    }

    /**
     * Marshall an optional schema object attribute to the given writer, and update CRC representing all attributes
     * @param attributeName the name of the attribute to marshall
     * @param escapedAttributeValue the escaped value of the attribute to marshall
     * @throws IOException on IO error
     */
    public final void addOptionalAttributeAndCRC(final String attributeName, final String escapedAttributeValue) throws IOException {
        if (escapedAttributeValue == null) {
            return;
        }
        append(attributeName).append("=\"").append(escapedAttributeValue).append("\" ");
        final byte[] escapedAttributeBytes = escapedAttributeValue.getBytes();
        allAttributesCRC.update(escapedAttributeBytes);
    }

    /**
     * Marshall an optional schema object attribute to the given writer, and update CRC representing all attributes
     * @param attributeName the name of the attribute to marshall
     * @param escapedAttributeValue the escaped value of the attribute to marshall
     * @throws IOException on IO error
     */
    public final void addAttributeNoCRC(final String attributeName, final String escapedAttributeValue) throws IOException {
        if (escapedAttributeValue == null) {
            return;
        }
        append(attributeName).append("=\"").append(escapedAttributeValue).append("\" ");
    }

    public CRC32 getAllAttributesCRC() {
        return allAttributesCRC;
    }

    public CRC32 getRequiredAttributesCRC() {
        return requiredAttributesCRC;
    }
}
