package com.onaro.sanscreen.acquisition.framework.datasource.io;

import com.onaro.sanscreen.acquisition.framework.common.SortedProperties;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceOutputFileType;
import com.onaro.sanscreen.acquisition.framework.datasource.config.PackageConfiguration;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import com.onaro.sanscreen.acquisition.framework.recordings.RecordingSample;

import java.io.*;
import java.util.Collection;
import java.util.Properties;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

public class DSSessionFileWriter extends DSSessionFileBase {
    private final RecordingSample rootSample;
    private final Collection<RecordingSample> samples;
    private final PackageConfiguration configuration;

    public DSSessionFileWriter(final RecordingSample rootSample, final Collection<RecordingSample> samples, final PackageConfiguration configuration) {
        this.rootSample = rootSample;
        this.samples = samples;
        this.configuration = configuration;
    }

    /**
     * Pack the entire collection of files generated during this sample into a single sample file (ZIP).
     *
     * @return The newly created sample file (ZIP)
     *
     * @throws java.io.IOException If any problem occured during the creation of the ZIP file
     */
    public File toFile() throws IOException {
        final File zipFile = new File(
                ServiceManager.getInstance().getLogsManager().getAcqRecordingDir(),
                RecordingSample.formatSampleFileName(configuration));

        ZipOutputStream zos = null;

        try {
            zos = new ZipOutputStream(new FileOutputStream(zipFile));
            zos.setLevel(9);

            // first zip "root" files
            zipFiles(zos, rootSample, rootFolder);
            zos.flush();

            // store configuration into root
            final byte[] configurationData = storeConfiguration(configuration);

            final ZipEntry fileEntry = new ZipEntry(rootFolder + File.separator + DataSourceOutputFileType.configuration.toString());
            zos.putNextEntry(fileEntry);
            zos.write(configurationData);
            zos.flush();

            // now store all the samples
            for (final RecordingSample sample : samples) {
                zipFiles(zos, sample, Long.toString(sample.getTimestamp()));
            }

            return zipFile;
        }
        finally {
            if (zos != null) {
                zos.flush();
                zos.finish();
                zos.close();
            }
        }
    }

    private static void zipFiles(final ZipOutputStream zos, final RecordingSample sample, final String folder) throws IOException {
        final byte[] buf = new byte[4096];
        int length;

        for (final DSFile dsFile : sample.getFileHandles()) {
            final File file = new File(dsFile.getFileName());
            if (!file.exists())
                continue;

            // Create a file input stream and a buffered input stream.
            final FileInputStream fis = new FileInputStream(file);

            // Create a Zip Entry and put it into the archive (no data yet).
            final String fileName = folder + File.separator + file.getName();
            final ZipEntry fileEntry = new ZipEntry(fileName);
            fileEntry.setTime(file.lastModified());

            zos.putNextEntry(fileEntry);

            BufferedInputStream bis = null;

            try {
                bis = new BufferedInputStream(fis);
                while ((length = bis.read(buf)) > -1)
                    zos.write(buf, 0, length);
            }
            finally {
                if (bis != null)
                    bis.close();
            }
        }

        // Now create an additional file which we will later delete
        final ByteArrayOutputStream os = new ByteArrayOutputStream(10000);
        sample.getDescription().storeToXML(os, "description");


        final ZipEntry fileEntry = new ZipEntry(folder + File.separator + descriptionFile);
        zos.putNextEntry(fileEntry);
        zos.write(os.toByteArray());
        zos.flush();
    }

    /**
     * Stores the package configuration into the sample file. This includes enough of the meta data for the framework to
     * later be able to reconstruct the right package with the right configuration for simulation.
     *
     * @param configuration package configuration
     *
     * @return A byte array containing the configuration
     */
    private static byte[] storeConfiguration(final PackageConfiguration configuration) throws IOException {
        final Properties propertiesToSave = new SortedProperties();

        propertiesToSave.put(dataSourceNameProperty, configuration.getDataSourceName());
        propertiesToSave.put(dataSourceIDProperty, Long.toString(configuration.getDataSourceID()));
        propertiesToSave.put(dataSourceTypeProperty, Long.toString(configuration.getDataSourceType()));
        propertiesToSave.put(packageNameProperty, configuration.getPackageName());

        propertiesToSave.putAll(configuration.getPropertiesForRecording());

        // Add licenses
//        configuration.getLicenses().store(propertiesToSave);

        // Now create an additional file which we will later delete
        final ByteArrayOutputStream os = new ByteArrayOutputStream(10000);
        propertiesToSave.storeToXML(os, configuration.getDataSourceName());

        return os.toByteArray();
    }
}
