package com.onaro.sanscreen.acquisition.framework.datasource.metadata;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * Each instance of this class represents an XML's tag named <properties> ...</properties> <br>
 * Each "valid" attribute of this class represent XML's attributes
 */
public class Attribute {
    private String name;
    private String value;
    private String description;
    private String label;
    private Type type;
    private String min;
    private String max;
    private Set<String> enums;  // Original name in XML is enum, but it has being changed because is a Java's keyword
    private String role;
    private boolean editable;
    private boolean encrypt;

    /**
     * Constructor,
     */
    Attribute(final String name,
              final String value,
              final String description,
              final String label,
              final String type,
              final String min,
              final String max,
              final String enums,
              final String role,
              final String editable,
              final String encrypt) throws FrameworkException {
        // Mandatory fields with data on them.
        this.name = getStringValue("name", name, true, null);
        this.value = getStringValue("value", value, true, null);

        // Optional fields with data on them.
        this.description = getStringValue("description", description, false, "");
        this.label = getStringValue("label", label, false, "");
        this.type = Type.fromString(type);
        this.min = getStringValue("min", min, false, "");
        this.max = getStringValue("max", max, false, "");
        this.enums = translateEnums(getStringValue("enum", enums, false, ""));
        this.role = getStringValue("role", role, false, "mandatory");
        this.editable = getBooleanValue("editable", editable, false, true);
        this.encrypt = getBooleanValue("encrypt", encrypt, false, false);
        if (this.encrypt && this.type != Type.STRING)
            throw new FrameworkException("Encrypted value can only be of type string. Attribute [" + this.name + "] is defined with type " + this.type);
    }

    /**
     * @return A set of enum values or null if not available
     */
    private Set<String> translateEnums(final String enumsStr) {
        if (enumsStr == null || enumsStr.length() == 0)
            return null;

        enums = new HashSet<String>(10);

        for (final StringTokenizer tok = new StringTokenizer(enumsStr, ";"); tok.hasMoreTokens();) {
            final String enumValue = tok.nextToken();
            enums.add(enumValue);
        }

        return enums;
    }

    /**
     * Constructor,
     */
    Attribute(final Map<String, String> attributes) throws FrameworkException {
        this(attributes.get("name"),
             attributes.get("value"),
             attributes.get("description"),
             attributes.get("label"),
             attributes.get("type"),
             attributes.get("min"),
             attributes.get("max"),
             attributes.get("enum"),
             attributes.get("role"),
             attributes.get("editable"),
             attributes.get("encrypt"));
    }

    public String getName() {
        return name;
    }

    public String getValue() {
        return value;
    }

    public String getDescription() {
        return description;
    }

    public String getLabel() {
        return label;
    }

    public Type getType() {
        return type;
    }

    public String getMin() {
        return min;
    }

    public String getMax() {
        return max;
    }

    public Set<String> getEnums() {
        return enums;
    }

    public boolean isMandatory() {
        return role.equals("mandatory");
    }

    public boolean isEditable() {
        return editable;
    }

    public boolean isEncrypt() {
        return encrypt;
    }

    public boolean isHidden() {
        return role.equals("hidden");
    }

    public boolean isSystemAttribute() {
        return name.startsWith("transport");
    }

    public static String decrypt(final String data) {
        try {
            return PBE.decrypt(data);
        }
        catch (Exception e) {
            return data;
        }
    }

    /**
     * Returns an Java's integer primitive  as a result of extracting the attribute name from <code>Element</code>.
     * If no value is found then <code>defaultValue</code> is returned.
     *
     * @param attrName
     * @param attrValue
     * @param isMandatory
     * @param defaultValueInCaseOfNull
     *
     * @throws com.onaro.sanscreen.acquisition.framework.FrameworkException
     *
     */
    private String getStringValue(final String attrName,
                                  final String attrValue,
                                  boolean isMandatory,
                                  final String defaultValueInCaseOfNull) throws FrameworkException {

        if (attrValue == null) {
            if (isMandatory)
                throw new FrameworkException("Missing value for a mandatroy attribute [" + attrName + "]");

            // Let's take the default value in case of null.
            return defaultValueInCaseOfNull;
        }

        return attrValue;
    }

    /**
     * Returns an Java's boolean primitive.
     * If <code>attrValue</code> is null then a 0 (zero) will be returned.
     *
     * @param attrName
     * @param attrValue
     * @param isMandatory
     * @param defaultValueInCaseOfNull
     *
     * @return boolean primitive value
     *
     * @throws com.onaro.sanscreen.acquisition.framework.FrameworkException
     *
     */
    private boolean getBooleanValue(final String attrName,
                                    final String attrValue,
                                    boolean isMandatory,
                                    boolean defaultValueInCaseOfNull) throws FrameworkException {

        boolean value;
        if (attrValue == null) {
            if (isMandatory)
                throw new FrameworkException("Missing value for a mandatroy attribute [" + attrName + "]");

            // Let's take the default value in case of null.
            value = defaultValueInCaseOfNull;
        }
        else {
            // If the <code>attrValue</code> is not true neither false then
            // calling Boolean.parBoolean(attrValue) will always going to be false.
            // If we can not match with TRUE or FALSE then the defaultValue is used.
            if (attrValue.equalsIgnoreCase(Boolean.TRUE.toString()))
                value = true;
            else if (attrValue.equalsIgnoreCase(Boolean.FALSE.toString()))
                value = false;
            else
                value = defaultValueInCaseOfNull;
        }

        // return the value as boolena primity.
        return value;
    }

    /**
     * Given a specific value, checks if this value can be successfully set in this attribute (type-wise)
     */
    public boolean checkValueType(final String valueToCheck) {
        if (valueToCheck == null || valueToCheck.length() == 0)
            return true;

        switch (getType()) {
            case INTEGER:
                try {
                    Integer.parseInt(valueToCheck);
                    return true;
                }
                catch (NumberFormatException e) {
                    return false;
                }

            case BOOLEAN:
                if (valueToCheck.equals(Boolean.TRUE.toString()) || valueToCheck.equals(Boolean.FALSE.toString()))
                    return true;

                return false;

            case FLOAT:
                try {
                    Float.parseFloat(valueToCheck);
                    return true;
                }
                catch (NumberFormatException e) {
                    return false;
                }

            case ENUM:
                return enums.contains(valueToCheck);

            case STRING:
            default:
                return true;
        }
    }

    /**
     * Marshal the attribute in an XML format. This is used when we need to deploy the configuration of the data source's
     * to the database.
     */
    public void toXML(final StringBuilder builder) {
        builder.append("<property ");
        builder.append("name=\"").append(name).append("\" ");
        builder.append("value=\"").append(encrypt ? "******" : value).append("\" ");

        if (description != null && description.length() > 0)
            builder.append("description=\"").append(description).append("\" ");

        if (label != null && label.length() > 0)
            builder.append("label=\"").append(label).append("\" ");

        if (type != null)
            builder.append("type=\"").append(type).append("\" ");

        if (min != null && min.length() > 0)
            builder.append("min=\"").append(min).append("\" ");

        if (max != null && max.length() > 0)
            builder.append("max=\"").append(max).append("\" ");

        if (enums != null && enums.size() > 0) {
            builder.append("enum=\"");

            for (final String enumValue : enums)
                builder.append(enumValue).append(";");

            builder.deleteCharAt(builder.length() - 1);

            builder.append("\" ");
        }

        if (!isMandatory())
            builder.append("mandatory=\"").append(false).append("\" ");
        else
            builder.append("mandatory=\"").append(true).append("\" ");

        if (!editable)
            builder.append("editable=\"").append(editable).append("\" ");

        if (encrypt)
            builder.append("password=\"").append(encrypt).append("\" ");

        if (isHidden())
            builder.append("hidden=\"").append("true").append("\" ");

        builder.append("/>");
    }

    public static enum Type {
        STRING("string"),
        INTEGER("integer"),
        FLOAT("float"),
        BOOLEAN("boolean"),
        ENUM("enum"),
        LIST("list");

        private String e;

        private Type(final String e) {
            this.e = e;
        }

        public static Type fromString(final String s) {
            if (s == null || s.length() == 0)
                return STRING;

            for (Type type : Type.values()) {
                if (type.e.equals(s))
                    return type;
            }

            throw new IllegalArgumentException("Invalid argument type '" + s + "'");
        }

        public String toString() {
            return e;
        }
    }
}
