package com.onaro.sanscreen.acquisition.framework.datasource.metadata;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import org.jdom.Element;

import java.io.Serializable;
import java.util.*;

/**
 * Implements the XML parsing and construction of "structures" that represent the attribute for
 * each "properties" xml's tag.
 */
class AttributesSection implements Serializable {
    private final static long serialVersionUID = 0L;

    private Map<String, Attribute> attributes = new LinkedHashMap<String, Attribute>();

    /**
     * If the XML doesn't have all the mandatory fields then the call will fail, inforcing the mandatory fields.
     */
    @SuppressWarnings({"unchecked"})
    public AttributesSection(final Element rootElement) throws FrameworkException {
        final Element properties = rootElement.getChild("properties");
        if (properties == null)
            throw new FrameworkException("Missing element 'properties' from package definition");

        for (final Element property : (List<Element>) properties.getChildren("property")) {
            // Creates a new Map, each time we are processing a new Instance tag.
            final Map<String, String> map = new TreeMap<String, String>();

            // Loading all the properties for this Instance.
            putPropertyAttributesInMap(property, map);
            String attributeType = map.get("type");
            if (attributeType != null && attributeType.equals("list")) {
                String listName = validatePropertyName(map.get("name"));
//                map.put("name", "LIST." + listName);

//                // LIST: we want to add the list as a property in the format LIST.listname
//                this.attributes.put("LIST." + listName, new Attribute(map));

                // COLUMNS: for each column add it as a property in the format LIST.listname.0.columnname
                for (final Element column : (List<Element>) property.getChild("columns").getChildren("property")) {
                    final Map<String, String> columnAttributeMap = new TreeMap<String, String>();
                    putPropertyAttributesInMap(column, columnAttributeMap);
                    String columnName = validatePropertyName(columnAttributeMap.get("name"));
                    columnAttributeMap.put("name", "LIST." + listName + ".0." + columnName);
                    this.attributes.put("LIST." + listName + ".0." + columnName, new Attribute(columnAttributeMap));
                }
                continue;
            }

            this.attributes.put(map.get("name"), new Attribute(map));
        }
    }

    private void putPropertyAttributesInMap(final Element property, final Map<String, String> map) throws FrameworkException {
        for (final org.jdom.Attribute attr : (List<org.jdom.Attribute>) property.getAttributes()) {
            String key = attr.getName();
            String value = attr.getValue();
            map.put(key, value);
        }
    }

    public Map<String, Attribute> getAttributes() {
        return attributes;
    }

    /**
     * Returns the structure <code>PropertyDesc</code> based on property's name.
     */
    public final Attribute getAttribute(String propertyName) {
        return attributes.get(propertyName);
    }

    /**
     * Returns all the atrributes found for this properties xml's tag.
     * The list of attribure's name include mandatory and not-mandatory fields.
     */
    public final Set<String> getAllAttributeNames() {
        return attributes.keySet();
    }

    public void setAttributeNotInDeployeeXML(final String name, final Attribute attribute) {
        attributes.put(name, attribute);
    }

    /**
     * Returns a list of Mandatory properties only.
     * The flag that indicates that a specific xml's tag is mandatory can be checked using the structure that
     * holds all the attribures of this property's name
     */
    public final List<String> getMandatoryAttributeNames() {
        List<String> list = new ArrayList<String>();
        for (Map.Entry<String, Attribute> entry : attributes.entrySet()) {
            final String attrName = entry.getKey();
            final Attribute attribute = entry.getValue();

            if (attribute.isMandatory() && !attribute.isHidden())
                list.add(attrName);
        }

        return list;
    }


    /**
     * Returns a list of Hidden properties only.
     * The flag that indicates that a specific xml's tag is hidden can be checked using the structure that
     * holds all the attribures of this property's name
     */
    public final List<String> getNotHiddenAttributeNames() {
        List<String> list = new ArrayList<String>();
        for (Map.Entry<String, Attribute> entry : attributes.entrySet()) {
            String attrName = entry.getKey();
            Attribute attribute = entry.getValue();

            if (!attribute.isHidden())
                list.add(attrName);
        }

        return list;
    }

    /**
     * Translates the attributes into the old format of encoded properties to allow the framework to put it in the
     * database.
     *
     * @param includeHiddens
     */
    public Properties getDefaultProperties(final boolean includeHiddens) throws FrameworkException {
        final Properties defaultProperties = new Properties();

        // First add all the default values
        for (final Attribute attribute : attributes.values()) {
            if (!includeHiddens && attribute.isHidden())
                continue;

            defaultProperties.setProperty(attribute.getName(), attribute.getValue());
        }

        return defaultProperties;
    }

    private String validatePropertyName(final String name) throws FrameworkException {
        if (name.matches("^([a-zA-Z] || \\.)"))
            throw new FrameworkException("property name format is illegal:" + name);
        return name;
    }
}
