package com.onaro.sanscreen.acquisition.framework.datasource.metadata;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.net.URL;
import java.util.jar.Attributes;
import java.util.jar.JarInputStream;
import java.util.jar.Manifest;

/**
 * Abstraction of retrieving information from system JAR file manifests.  Includes build time, versions, and so on.
 *
 * All information gathered from JAR files downloaded to AU from remote server.
 */
public class CustomManifestInfo {

    private static final Logger logger = LogManager.getLogger(CustomManifestInfo.class);

    private static final String ONARO_VERSION_CUSTOMER = "onaro-version-customer";
    private static final String ONARO_VERSION_TIME = "onaro-version-time";
    private static final String ONARO_VERSION_CHECKSUM = "onaro-version-checksum";

    public static final String UNAVAILABLE = "<unavailable>";
    private final String jarDisplayName;

    private String customer = CustomManifestInfo.UNAVAILABLE;
    private String time = CustomManifestInfo.UNAVAILABLE;
    private String checksum = CustomManifestInfo.UNAVAILABLE;

    /**
     * Create a new manifest info using the information in the passed JAR file URL.
     *
     * @param jar  the URL pointing to the JAR file from which manifest information is desired.
     * @param jarDisplayName the display name of the JAR file in question.
     */
    public CustomManifestInfo(final URL jar, final String jarDisplayName) {

        this.jarDisplayName = jarDisplayName;
        parse(jar);
    }

    /**
     * Parse manifest information from JAR at end of given URL.
     *
     * @param jar the jar URL.
     */
    private void parse(final URL jar) {
        JarInputStream jis = null;
        try {
            jis = new JarInputStream(jar.openStream());
            final Manifest jarManifest = jis.getManifest();
            final Attributes manifestAttributes = jarManifest.getMainAttributes();

            // If this JAR does not have the Onaro-specific manifest version properties
            // then it is probably a library JAR needed by the data source
            if (!manifestAttributes.containsKey(new Attributes.Name(ONARO_VERSION_CUSTOMER))) {
                return;
            }

            customer = (String) manifestAttributes.get(new Attributes.Name(ONARO_VERSION_CUSTOMER));

            if (manifestAttributes.containsKey(new Attributes.Name(ONARO_VERSION_TIME))) {
                time = (String) manifestAttributes.get(new Attributes.Name(ONARO_VERSION_TIME));
            }

            if (manifestAttributes.containsKey(new Attributes.Name(ONARO_VERSION_CHECKSUM))) {
                checksum = (String) manifestAttributes.get(new Attributes.Name(ONARO_VERSION_CHECKSUM));
            }

        }
        catch (final IOException ioe) {
            logger.warn("Could not parse MANIFEST for " + jarDisplayName + ".  Version data unavailable.");
            //not fatal - move on.
        }
        finally {
            if (jis != null) {
                try {
                    jis.close();
                }
                catch (Exception e) {
                    //fall through
                }
            }
        }
    }

    /**
     * Formatted version string for logging.
     *
     * @return   the formatted log string.
     */
    public String getVersionString() {
        return "name: " + jarDisplayName + ", time: " + time + ", checksum: " + checksum + ", customer: " + customer;
    }

    public String getJarDisplayName() {
        return jarDisplayName;
    }

    public String getCustomer() {
        return customer;
    }

    public String getTime() {
        return time;
    }

    public String getChecksum() {
        return checksum;
    }

}
