package com.onaro.sanscreen.acquisition.framework.datasource.metadata;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.config.DataSourceConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.config.PackageConfiguration;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;

import org.jdom.Element;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

/**
 * Consolidates all the implementation specific details in a data source meta data configuration.
 * This includes:<br>
 * 1. The implementation class / classes<br>
 * 2. The libraries to load<br>
 * 3. The class loading classLoadingPolicy<br>
 *
 * <p/>This class will be able to instantiate a data source based on this data
 */
class ImplementationSection implements Serializable {

    private static final Logger LOG = LoggerFactory.getLogger(ImplementationSection.class);

    private final static long serialVersionUID = 0L;

    /**
     * Package JARs are loaded in the package class loader. These JARs are common to all instances of the
     * package and will be shared with other packages via dependency.
     */
    private final List<String> packageJars = new ArrayList<String>();

    /**
     * Instance JARs are additional jars included for each instance of the package. These JARs are private to the package
     * itself and cannot be shared between packages (via dependency)
     */
    private final List<String> instanceJars = new ArrayList<String>();

    private String mainJar;
    private String version = null;
    private final String configurationClassName;
    private final String packageImplementationClass;

    private ClassLoader packageClassLoader = null;

    private final PackageMetaData packageMetaData;

    ImplementationSection(final PackageMetaData packageMetaData, final Element packageElement) throws FrameworkException {
        this.packageMetaData = packageMetaData;

        final Element implementationElement = packageElement.getChild("implementation");
        if (implementationElement == null) {
            throw new FrameworkException("Missing mandatory element 'implementation' under 'package' " + packageMetaData.getName());
        }

        // Check if this data source supports the new implementation....
        configurationClassName = implementationElement.getAttributeValue("configurationClass");
        if (configurationClassName == null) {
            throw new FrameworkException("Mandatory attribute 'configurationClass' is missing from 'implementation' section for package " + packageMetaData.getName());
        }

        // Check if this data source supports the new implementation....
        packageImplementationClass = implementationElement.getAttributeValue("packageImplementationClass");
        if (packageImplementationClass == null) {
            throw new FrameworkException("Mandatory attribute 'packageImplementationClass' is missing from 'implementation' section for package " + packageMetaData.getName());
        }

        readClassLoadingConfiguration(implementationElement);
    }

    /**
     * Called as a second stage initialization (after configuration is read and dependency order is established)
     *
     * @param disableClassLoaders true if class loader hierarchy should be disabled
     *
     * @throws FrameworkException On any error
     */
    void initializeClassLoaders(final boolean disableClassLoaders) throws FrameworkException {
        if (disableClassLoaders) {
            LOG.info("Package {} - Loading with class loaders disabled (standalone mode)", packageMetaData.getName());

            // Get the system class loader
            packageClassLoader = getRootLoader();
            return;
        }

        // Now that we have the full list of JARs to load, let's instantiate the package class loader
        packageClassLoader = instantiateClassLoader(packageMetaData.getName(), packageJars, getRootLoader());
    }

    /**
     * This is the root loader of the system, obtained from deploy manager.   It becomes the parent loader for all package
     * level loaders.   It holds the base package jars and certain system wide libraries.   There is one root loader
     * per AU.
     *
     * <p>In standalone situations (no deploy manager), this returns the system loader (loader of ServiceManager).
     *
     * @return the root loader.
     */
    private ClassLoader getRootLoader() {
        final boolean standAlone = ServiceManager.getInstance().isStandalone();
        return !standAlone
               ? ServiceManager.getInstance().getDeployManager().getRootLoader()
               : ServiceManager.getInstance().getClass().getClassLoader();
    }

    /**
     * Loads the package implementation class in the appropriate class loader.
     *
     * @return The package implementation class
     *
     * @throws FrameworkException     When the result class is not derivable from the <code>Package</code> interface
     * @throws ClassNotFoundException When class could not be loaded / found
     */
    Class<?> getPackageImplementationClass() throws FrameworkException, ClassNotFoundException {
        final Class<?> aClass = figureOutWhichClassLoaderToUse().loadClass(packageImplementationClass);
        if (aClass == null || !(com.onaro.sanscreen.acquisition.framework.datasource.Package.class.isAssignableFrom(aClass))) {
            throw new FrameworkException("Package configuration class " + packageImplementationClass + " does not derive from Package");
        }

        return aClass;
    }

    /**
     * Based on the class loading policy, return the right class loader to use
     *
     * @return The class loader most appropriate to use (package / instance / etc...)
     *
     * @throws FrameworkException On any error
     */
    private ClassLoader figureOutWhichClassLoaderToUse() throws FrameworkException {
        if (instanceJars.size() == 0) {
            // Reuse the data source type class loader
            return packageClassLoader;
        }

        // Create a new instance of a class loader just for this data source instance
        return instantiateClassLoader(packageMetaData.getName() + " (instance)", instanceJars, packageClassLoader);
    }

    /**
     * From the configuration, create a class loader that will represent this data source type.
     * Depending on the classLoadingPolicy, this will either create a special class loader just for this instance or reuse the
     * instance class loader or use the system class loader.
     *
     * <p/>This class loader will be loaded with all the JARs in the <code>jars</code> list.
     *
     * @param name   Name of class loader (used for debugging)
     * @param jars   List of JAR names to include in the class path
     * @param parent Parent class loader to associate with this new class loader
     *
     * @return A newly instantiated class loader
     *
     * @throws FrameworkException On any error
     */
    private ClassLoader instantiateClassLoader(final String name, final List<String> jars, final ClassLoader parent) throws FrameworkException {
        final URL[] urls = new URL[jars.size()];

        int j = 0;
        for (final String jarName : jars) {
            urls[j] = createURL(jarName);
            j++;

            LOG.info("{} - jar loaded : {}", packageMetaData.getName(), jarName);
        }

        return new DataSourceInstanceClassLoader(name, urls, parent);
    }

    /**
     * Instantiates a new package configuration object based on this meta data. This class will be loaded in the appropriate
     * package class loader and will later be used to instantiate the package implementation itself.
     *
     * @param configuration  Data source configuration (container)
     * @param dataSourceName Data source name
     * @param dataSourceID   Data source ID
     * @param properties     Properties for package
     *
     * @return A new instance of the package configuration as described in the <code>configurationClass</code> attribute
     *
     * @throws FrameworkException On any error
     */
    PackageConfiguration instantiateConfiguration(final DataSourceConfiguration configuration,
                                                  final String dataSourceName,
                                                  final long dataSourceID,
                                                  final Properties properties) throws FrameworkException {
        // Create a special class loader for this instance!
        final ClassLoader classLoaderToUse = figureOutWhichClassLoaderToUse();

        final Class<?> packageConfigurationClass;
        try {
            // Make sure that the implementation field contains a valid class name that derives from BaseDataSource
            packageConfigurationClass = classLoaderToUse.loadClass(configurationClassName);
            if (packageConfigurationClass == null || !(PackageConfiguration.class.isAssignableFrom(packageConfigurationClass))) {
                throw new FrameworkException("Package configuration class " + configurationClassName + " for data source " + dataSourceName + " does not derive from PackageConfiguration");
            }
        }
        catch (ClassNotFoundException e) {
            throw new FrameworkException("Could not load implementation class " + configurationClassName + " for data source " + dataSourceName, e);
        }

        //noinspection CatchGenericClass
        try {
            final Class<?>[] constructorSignature = {DataSourceConfiguration.class, PackageMetaData.class, String.class, long.class, Properties.class};
            final Constructor<?> constructor = packageConfigurationClass.getConstructor(constructorSignature);

            // Construct the data source instance
            final Object[] constructorValues = {configuration, packageMetaData, dataSourceName, dataSourceID, properties};
            return (PackageConfiguration) constructor.newInstance(constructorValues);
        }
        catch (Throwable e) {
            throw new FrameworkException("Failed to instantiate data source " + dataSourceName, e);
        }
    }

    String getVersion() {
        return version;
    }

    ClassLoader getPackageClassLoader() {
        return packageClassLoader;
    }

    /**
     * Format a URL pointing to a JAR based on the deploy directory location and the jar name
     *
     * @param jarName JAR name to create URL for
     *
     * @return A URL pointing to the JAR in the deploy directory
     *
     * @throws FrameworkException On any error
     */
    private URL createURL(final String jarName) throws FrameworkException {
        if (jarName.length() == 0) {
            throw new IllegalArgumentException("Invalid empty jar name");
        }

        try {
            // Figure out from the class serverDeployDirLocation the server serverDeployDirLocation
            final String serverDeployDir = System.getProperty("codebase") + "/datasources/";
            return new URL(serverDeployDir + jarName);
        }
        catch (MalformedURLException e) {
            throw new FrameworkException("Failed to load JAR " + jarName, e);
        }
    }

    /**
     * Reads the <code>libs</code> section in the XML to determine which JARs should be loaded for this
     * data source.
     *
     * @param implementationElement Element to parse
     *
     * @throws FrameworkException On any error
     */
    private void readClassLoadingConfiguration(final Element implementationElement) throws FrameworkException {
        final List<Element> packageJarElements = (List<Element>) implementationElement.getChildren("packageJar");
        for (final Element jarElement : packageJarElements) {
            final String jarName = jarElement.getAttributeValue("name");

            if (Boolean.valueOf(jarElement.getAttributeValue("main"))) {
                if (mainJar != null) {
                    throw new FrameworkException("More than one 'main' jars defined for package " + packageMetaData.getName());
                }

                mainJar = jarName;
                version = new CustomManifestInfo(createURL(mainJar), mainJar).getVersionString();

            }

            //Skip this check in standalone.  There is no deploy manager!
            if (!ServiceManager.getInstance().isStandalone()) {
                if (ServiceManager.getInstance().getDeployManager().isRootLoaded(jarName)) {
                    throw new FrameworkException("DataSource package " + this.packageImplementationClass.getClass().getSimpleName()
                                                 + " is trying to load " + jarName + " which was already in the root loader!");
                }
            }

            packageJars.add(jarName);
        }

        final List<Element> instanceJarElements = (List<Element>) implementationElement.getChildren("instanceJar");
        if (instanceJarElements.size() > 0) {
            for (final Element jarElement : instanceJarElements) {
                final String jarName = jarElement.getAttributeValue("name");

                if (Boolean.valueOf(jarElement.getAttributeValue("main"))) {
                    if (mainJar != null) {
                        throw new FrameworkException("More than one 'main' jars defined for package " + packageMetaData.getName());
                    }

                    mainJar = jarName;
                    version = new CustomManifestInfo(createURL(mainJar), mainJar).getVersionString();
                }

                instanceJars.add(jarName);
            }
        }

        // We should fail in case the package has no main jar.
        // But we allow package without main jar that has a dependency
        // i.e. we hope this package can get all it needs from parent's package main jar.
        if (mainJar == null) {
            throw new FrameworkException("Missing 'main' jar for package " + packageMetaData.getName());
        }
    }

}
