package com.onaro.sanscreen.acquisition.framework.datasource.metadata;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.config.DataSourceConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.config.PackageConfiguration;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;

import org.jdom.Element;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.Serializable;
import java.util.*;

/**
 * Represents the configuration part of a data source, including the configuration, class loading policy, implementation
 * etc..
 *
 * <p/>This class basically represents the Deployee.xml file
 */
public class PackageMetaData implements Serializable {

    private static final Logger LOG = LoggerFactory.getLogger(PackageMetaData.class);

    private static final long serialVersionUID = 0L;
    private final String packageName;

    private final ImplementationSection implementation;
    private final AttributesSection attributes;

    private final DataSourceMetaData dataSourceMetaData;

    /**
     * Creates a meta data class based on the given XML.
     *
     * @param dataSourceMetaData
     * @param packageRootElement Root element of package
     *
     * @return The meta data for this data source type
     *
     * @throws FrameworkException
     */
    static PackageMetaData create(final DataSourceMetaData dataSourceMetaData, final Element packageRootElement) throws FrameworkException {
        return new PackageMetaData(dataSourceMetaData, packageRootElement);
    }

    /**
     * Constructs the meta data object from the package section inside a deploy desciptor file. This method is called by
     * {@link PackageMetaData} when parsing the deploy descriptor file to create a meta data object for each package
     * implemented for this data source.
     *
     * @param dataSourceMetaData
     * @param packageElement     Root element describing package in the deploy descriptor file
     *
     * @throws FrameworkException
     */
    private PackageMetaData(final DataSourceMetaData dataSourceMetaData, final Element packageElement) throws FrameworkException {
        this.dataSourceMetaData = dataSourceMetaData;

        // Now get the data...
        packageName = packageElement.getAttributeValue("name");
        if (packageName == null) {
            throw new FrameworkException("Missing mandatory attribute 'Name' from 'Package' element");
        }

        this.implementation = new ImplementationSection(this, packageElement);
        this.attributes = new AttributesSection(packageElement);

        addSiteName();
        addServerVersion();
        addVersionInfo(); // the jar version
    }

    public final String getName() {
        return packageName;
    }

    /**
     * @return The version embedded in the main jar of this package
     */
    public final String getVersion() {
        return implementation.getVersion();
    }

    /**
     * Loads the package implementation class in the appropriate class loader.
     *
     * @throws FrameworkException
     * @throws ClassNotFoundException
     */
    public final Class<?> getPackageImplementationClass()
            throws FrameworkException, ClassNotFoundException
    {
        return implementation.getPackageImplementationClass();
    }

    /**
     * Based on this meta-data, instantiate the appropriate configuration object (within the right class loader) to deal
     * with the specifics of this package.
     *
     * @param configuration
     * @param dataSourceName
     * @param dataSourceID
     * @param properties
     *
     * @throws FrameworkException
     */
    public final PackageConfiguration instantiateConfiguration(final DataSourceConfiguration configuration, final String dataSourceName, final long dataSourceID, final Properties properties) throws FrameworkException {
        return implementation.instantiateConfiguration(configuration, dataSourceName, dataSourceID, properties);
    }

    public final Properties getDefaultProperties(final boolean includeHidden) throws FrameworkException {
        return attributes.getDefaultProperties(includeHidden);
    }

    public final Map<String, Attribute> getAttributes() {
        return attributes.getAttributes();
    }

    public final Attribute getAttribute(final String attributeName) {
        return attributes.getAttributes().get(attributeName);
    }

    @Override
    public final String toString() {
        return getName();
    }

    public final DataSourceMetaData getDataSourceMetaData() {
        return dataSourceMetaData;
    }

    public final ClassLoader getPackageClassLoader() {
        return implementation.getPackageClassLoader();
    }

    final void initializeClassLoaders(final boolean disableClassLoaders) throws FrameworkException {
        implementation.initializeClassLoaders(disableClassLoaders);
    }

    /**
     * Add the version information to the attributes of the meta data
     */
    private void addVersionInfo() {
        final String dsVersion = implementation.getVersion();
        if (dsVersion == null || dsVersion.length() == 0) {
            return;
        }

        try {
            final Attribute versionAttribute = new Attribute("datasource.version",
                    dsVersion, "Data source version:", "Data source version:",
                    "string", null, null, null, Boolean.FALSE.toString(),
                    Boolean.FALSE.toString(), Boolean.FALSE.toString());
            attributes.setAttributeNotInDeployeeXML(versionAttribute.getName(), versionAttribute);

            LOG.info("loaded package {} for {} [version={}]", new Object[] {
                    packageName, dataSourceMetaData.getName(), dsVersion });
        }
        catch (FrameworkException fe) {
            LOG.error("{} - Failed to add version attribute", getName(), fe);
        }
    }

    private void addSiteName() {
        try {
            final String siteName = ServiceManager.getInstance().getCommunicationManager().getSiteName();
            final Attribute versionAttribute = new Attribute("system.site.name",
                    siteName , "Customer site:", "Customer site:", "string",
                    null, null, null, Boolean.FALSE.toString(),
                    Boolean.FALSE.toString(), Boolean.FALSE.toString());
            attributes.setAttributeNotInDeployeeXML(versionAttribute.getName(), versionAttribute);
        }
        catch (FrameworkException fe) {
            LOG.error("{} - Failed to add site name attribute", getName(), fe);
        }
    }

    private void addServerVersion() {
        try {
            final String serverVersion = ServiceManager.getInstance().getCommunicationManager().getServerVersion();
            final Attribute versionAttribute = new Attribute("system.server.version",
                    serverVersion, "Server version:", "server version:", "string",
                    null, null, null, Boolean.FALSE.toString(),
                    Boolean.FALSE.toString(), Boolean.FALSE.toString());
            attributes.setAttributeNotInDeployeeXML(versionAttribute.getName(), versionAttribute);
        }
        catch (FrameworkException fe) {
            LOG.error("{} - Failed to add server version attribute", getName(), fe);
        }
    }

}
