package com.onaro.sanscreen.acquisition.framework.diskmodel;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.DiskModelDescription;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DiskModelUtil;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DiskModelWrapper;
import com.netapp.acquisition.aquisitionserverside.interfaces.remote.DiskModelRemote;
import com.onaro.sanscreen.acquisition.foundation.datasource.BaseFoundationPackage;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.*;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileWriter;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSInputStream;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import com.onaro.sanscreen.acquisition.framework.recordings.RecordingSample;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.odata2.Disk;
import com.onaro.sanscreen.acquisition.schema.odata2.Storage;
import com.onaro.util.Capacity;
import org.apache.log4j.Logger;

import java.io.File;
import java.io.IOException;
import java.util.*;

/**
 * Acquisition specific Disk Model description functionality
 */
public class AcqDiskModel {
    private static final Logger logger = Logger.getLogger(AcqDiskModel.class);

    // String constants to store/retrieve the disk model descriptions file to/from recording zips
    private static final String RECORDING_KEY_PREFIX = "diskmodel";
    private static final String RECORDING_KEY = "descriptions";

    private static final String DISK_DESCR_RECORDING = RECORDING_KEY_PREFIX + "-" + RECORDING_KEY;

    /**
     * The singleton <code>AcqDiskModel</code> instance
     */
    private static final AcqDiskModel acqDiskModel = new AcqDiskModel();

    /**
     * A cache of the merged map of master (SANscreen - default or latest from gate.onaro.com) and custom (user defined) disk model descriptions
     * The disk model is in the root loader and common to all datasources
     * The cache includes the default or latest disk models, plus any custom disk models
     * the time of the mose recent update is kept so a request to the server will only return a new disk model if it has been updated
     */
    private volatile DiskModelWrapper diskModel = new DiskModelWrapper(new TreeMap<String, DiskModelDescription>(), -1);

    /**
     * a cache of simulated disk models only
     */
    private Map<BaseDataSource, SortedMap<String, DiskModelDescription>> simulatedDiskModelDescriptionsCache = new HashMap<BaseDataSource, SortedMap<String, DiskModelDescription>>();

    /**
     * Private constructor, no initialization required.
     */
    private AcqDiskModel() {
    }

    /**
     * Gets the singleton <code>AcqDiskModel</code> instance
     *
     * @return the singleton <code>AcqDiskModel</code> instance
     */
    public static AcqDiskModel getInstance() {
        return AcqDiskModel.acqDiskModel;
    }

    /**
     * get the shared disk model map. If simulating, gets the disk model from the recording
     *
     * @return a SortedMap of disk models
     *
     * @throws AcqDiskModelException if the simulated disk model does not exist for the datasource
     */
    private SortedMap<String, DiskModelDescription> getDiskModelMap() throws AcqDiskModelException {
        final SortedMap<String, DiskModelDescription> diskModelMap;
        if (PackageContext.isSimulating()) {
            diskModelMap = simulatedDiskModelDescriptionsCache.get(PackageContext.getPackage().getDataSource());
            if (diskModelMap == null) {
                final String msg = "simulated disk model not available";
                throw new AcqDiskModelException(DataSourceErrorType.INTERNAL_ERROR, msg, msg);
            }
        }
        else {
            diskModelMap = diskModel.getDescriptions();
        }
        return diskModelMap;
    }

    /**
     * Lookup a disk model description for the given disk model ID.
     *
     * @param diskModelID the ID of the requested disk model description
     *
     * @return a disk model description for the given disk model ID.
     *
     * @throws AcqDiskModelException if there is no description for the given model ID.
     */
    public synchronized DiskModelDescription lookupDiskModelDescription(final String diskModelID) throws AcqDiskModelException {
        final String normalizedDiskModelID = diskModelID.toUpperCase();
        final SortedMap<String, DiskModelDescription> diskModelMap = getDiskModelMap();

        if (!diskModelMap.containsKey(normalizedDiskModelID)) {
            final DiskModelDescription dmdFromPartialMatch = lookupPartialDiskModel(normalizedDiskModelID);
            if (dmdFromPartialMatch != null) {
                return dmdFromPartialMatch;
            }

            // We couldn't find this disk model. Let's stop and let the user know. He can add this disk to disk_models file.
            final EnhancedErrorDetail eed =
                    new EnhancedErrorDetail("disk_model_parsing_error.html", "disk_model_parsing_error.html");
            eed.addParam("disk.model", diskModelID);
            eed.addParam("disk.model.filename2", "new_disk_models.txt");
            eed.addParam("disk.model.filename1", "disk_models.txt");
            throw AcqDiskModelException.createWithEnhanced(
                    DataSourceErrorType.INCONSISTENT_DATA, "New Disk Type detected: " + diskModelID, eed);
        }

        return diskModelMap.get(normalizedDiskModelID);
    }

    /**
     * Lookup a disk model but don't throw an exception if the model is not found...just return null
     *
     * @param diskModelID the disk model id (the key in the map)
     *
     * @return a disk model or null
     */
    public synchronized DiskModelDescription lookupDiskModelDescNoThrow(final String diskModelID) {
        final String normalizedDiskModelID = diskModelID.toUpperCase();

        try {
            final SortedMap<String, DiskModelDescription> diskModelMap = getDiskModelMap();
            if (!diskModelMap.containsKey(normalizedDiskModelID)) {
                final DiskModelDescription dmdFromPartialMatch = lookupPartialDiskModel(normalizedDiskModelID);
                if (dmdFromPartialMatch != null) {
                    return dmdFromPartialMatch;
                }
            }
            // Will return the disk model or null
            return diskModelMap.get(normalizedDiskModelID);
        }
        catch (AcqDiskModelException e) {
            return null;
        }
    }

    /**
     * Lookup disk model via sub-strings to accommodate older data sources
     * @param diskModelID Disk model ID from device
     * @return DiskModel or null
     */
    private synchronized DiskModelDescription lookupPartialDiskModel(final String diskModelID) throws AcqDiskModelException {
        final SortedMap<String, DiskModelDescription> diskModelMap = getDiskModelMap();
        // ICI-428 and others
        // Try sub string match.
        // Data Sources used to use sub strings
        // Example:  STT34007 CLAR400 => STT34007
        final String[] subStrings = diskModelID.split("\\s+");
        if (subStrings.length > 1) {
            if (diskModelMap.containsKey(subStrings[0])) {
                return diskModelMap.get(subStrings[0]);
            }
        }
        else {
            // Try first 8 characters if ID is longer than 8 + "CLAR"  for a total of at least 12 characters
            // Case: WD1002FBCLAR1000 => WD1002FB
            final int index = diskModelID.indexOf("CLAR");
            if (index > 0) {
                final String partialDiskModelID = diskModelID.substring(0, index);
                if (diskModelMap.containsKey(partialDiskModelID)) {
                    return diskModelMap.get(partialDiskModelID);
                }
            }
        }

        return null;
    }

    /**
     * Update the disk model attributes cache from the Master and Users lists on the SANscreen server, or
     * from the recording if simulated.   Add all disk model attributes to the current recording.
     *
     * @param foundationPackage the client foundation package
     *
     * @throws AcqDiskModelException    when there is a problem acquiring the disk model descriptions
     * @throws DataSourceErrorException when there is a problem storing the disk model descriptions to the recording
     */
    public void updateAndRecordDiskModelDescriptions(final BaseFoundationPackage foundationPackage)
            throws DataSourceErrorException, FrameworkException {

        final boolean isSimulating = foundationPackage.isSimulating();

        if (isSimulating && getAndRecordDiskModelsFromRecording(foundationPackage)) {
            return;
        }

        final SortedMap<String, DiskModelDescription> tempDiskModelDescriptions = new TreeMap<String, DiskModelDescription>();

        long updateTime = 0;
        // Download the disk model descriptions from the SANscreen server.
        // This includes default an custom disk models.
        final DiskModelRemote diskModelSession = ServiceManager.getInstance().getCommunicationManager().getDiskModelRemote();
        final DiskModelWrapper diskModelWrapper = diskModelSession.getDiskModelIfNewer(diskModel.getUpdateTimeStamp());
        if ((diskModelWrapper == null) || diskModelWrapper.getDescriptions().isEmpty()){
            if (diskModel.getDescriptions().isEmpty()) {
                // Yikes! We don't have any disk models!
                throw new FrameworkException("disk models could not be retrieved from the OCI server");
            }
        }
        else {
            tempDiskModelDescriptions.putAll(diskModelWrapper.getDescriptions());
            updateTime = diskModelWrapper.getUpdateTimeStamp();
        }

        // update the cache if it changed
        if (!tempDiskModelDescriptions.isEmpty()) {
            synchronized (this) {
                // if the cache was updated set it in a sync block
                diskModel = new DiskModelWrapper(Collections.unmodifiableSortedMap(tempDiskModelDescriptions), updateTime);
            }
        }

        if (foundationPackage.isRecording()) {
            addDiskModelsToRecording(foundationPackage, diskModel.getDescriptions());
        }
    }

    private boolean getAndRecordDiskModelsFromRecording(final BasePackage foundationPackage) throws DataSourceErrorException {
        final SortedMap<String, DiskModelDescription> simDiskModelDescriptions;
        try {
            simDiskModelDescriptions = getDiskModelsFromRecording(foundationPackage);
            if (simDiskModelDescriptions != null) {
                simulatedDiskModelDescriptionsCache.put(foundationPackage.getDataSource(), simDiskModelDescriptions);

                // record and return the disk models in
                if (foundationPackage.isRecording()) {
                    addDiskModelsToRecording(foundationPackage, simDiskModelDescriptions);
                }
                return true;
            }
        }
        catch (AcqDiskModelException e) {
            logger.warn("Unable to retrieve disk models from recording. Using server definitions.", e);
        }
        return false;
    }

    public void removeSimulatedDiskModelDescriptions(final BaseFoundationPackage foundationPackage) {
        simulatedDiskModelDescriptionsCache.remove(foundationPackage.getDataSource());
    }

    /**
     * gets disk models from the recording
     *
     * @param foundationPackage the foundation package
     *
     * @return A map of disk models in the recording or null
     *
     * @throws AcqDiskModelException the disk model file could not be parsed
     */
    private synchronized SortedMap<String, DiskModelDescription> getDiskModelsFromRecording(final BasePackage foundationPackage) throws AcqDiskModelException {
        if (!foundationPackage.isSimulating()) {
            return null;
        }
        SortedMap<String, DiskModelDescription> recordedDiskModelDescriptions = null;
        try {
            final DSInputStream modelDescriptionStream =
                    foundationPackage.getConfiguration().getSimulationSession().getCurrentSample().getStreamByDescription(
                            DataSourceOutputFileType.recording, RECORDING_KEY_PREFIX, RECORDING_KEY);

            if (null != modelDescriptionStream) {
                recordedDiskModelDescriptions = DiskModelUtil.parseDiskModelDescriptionStream(modelDescriptionStream, DISK_DESCR_RECORDING, true);
            }
        }
        catch (Exception e) {
            throw new AcqDiskModelException(DataSourceErrorType.INITIALIZING,
                                            "Error reading disk model configuration: " + DISK_DESCR_RECORDING, e.getMessage(), e);

        }
        return recordedDiskModelDescriptions;
    }

    /**
     * Adds all disk model configs to the recording file.
     *
     * @param foundationPackage     the client foundation package
     * @param diskModelDescriptions the map of disk models
     *
     * @throws DataSourceErrorException when there is a problem storing the disk model descriptions to the recording
     */
    private void addDiskModelsToRecording(final BasePackage foundationPackage, final Map<String, DiskModelDescription> diskModelDescriptions)
            throws DataSourceErrorException {

        // Iterate through all referenced disk models and store the model description in the recording.
        final RecordingSample recordingSample = foundationPackage.getRecordingSession().getCurrentSample();
        try {

            final DSFileWriter recordingWriter =
                    recordingSample.createTextualOutputFile(foundationPackage.getConfiguration(),
                                                            DataSourceOutputFileType.configuration, DISK_DESCR_RECORDING, DiskModelUtil.TXT);

            recordingSample.addFileDescription(getLocalFileName(recordingWriter.getFileName()), RECORDING_KEY_PREFIX, RECORDING_KEY);
            DiskModelUtil.writeUserDiskModelAttributes(recordingWriter, DiskModelUtil.getCurrentBanner(), diskModelDescriptions, true);
        }
        catch (IOException ioe) {
            throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR,
                                               "Failed to persist disk model configuration to recording.",
                                               ioe.getMessage(), ioe);
        }
    }

    /**
     * Creates an odata2 Disk object initialized using given disk model attributes.
     *
     * @param storage        the parent Storage to contain the Disk
     * @param diskKey        the unique identifier of the disk
     * @param diskModelDescr the disk model attributes used to initialize the Disk
     *
     * @return an odata2 Disk object initialized using given disk model attributes.
     *
     * @throws SchemaException when the given DiskModelDescription contains invalid values
     */
    public final Disk createDisk(final Storage storage, final String diskKey, final DiskModelDescription diskModelDescr)
            throws SchemaException {
        return createDisk(storage, diskKey, diskKey, diskModelDescr);
    }

    /**
     * Creates an odata2 Disk object initialized using given disk IDs and model attributes.
     *
     * @param storage        the parent Storage to contain the Disk
     * @param diskID         the unique identifier of the disk
     * @param diskName       the name of the disk
     * @param diskModelDescr the disk model attributes used to initialize the Disk
     *
     * @return an odata2 Disk object initialized using given disk IDs and model attributes.
     *
     * @throws SchemaException when the given DiskModelDescription contains invalid values
     */
    public final Disk createDisk(final Storage storage, final String diskID, final String diskName,
                                 final DiskModelDescription diskModelDescr) throws SchemaException {
        final Disk disk = storage.createDisk(diskID, diskName);
        try {
            disk.setDiskTypeEnum(Disk.DiskType.valueOf(diskModelDescr.getInterfaceType()));
        }
        catch (IllegalArgumentException iae) {
            // No Disk.Type enum defined for the given string, map to "OTHER"
            if (logger.isDebugEnabled()) {
                logger.debug("mapping disk type " + diskModelDescr.getInterfaceType()
                             + " to Disk.DiskType." + Disk.DiskType.OTHER);
            }
            disk.setDiskTypeEnum(Disk.DiskType.OTHER);
            disk.setType(diskModelDescr.getInterfaceType());
        }
        disk.setModel(diskModelDescr.getModel());
        disk.setVendor(diskModelDescr.getVendor());
        disk.setVendorID(diskModelDescr.getVendor());
        disk.setCapacity(Capacity.create(diskModelDescr.getSizeGiB(), Capacity.Unit.GB_1000));
        disk.setSeekTime(diskModelDescr.getSeekTimeOrDefault());
        if (Disk.DiskType.SSD.equals(disk.getDiskTypeEnum())) {
            // Hard code SSD values for now
            disk.setPotentialTransferRate(500);
            disk.setSpeed(1000000);
        }
        else {
            disk.setPotentialTransferRate(diskModelDescr.getPotentialTransferRateOrDefault());
            disk.setSpeed(diskModelDescr.getRpm());
        }
        return disk;
    }

    /**
     * Create a disk without filling in the parameters from a model
     *
     * @param storage  the parent Storage to contain the Disk
     * @param diskID   the unique identifier of the disk
     * @param diskName the name of the disk
     *
     * @return an odata2 Disk object initialized using given disk IDs
     *
     * @throws SchemaException
     */
    public final Disk createUnknownDisk(final Storage storage, final String diskID, final String diskName) throws SchemaException {
        return storage.createDisk(diskID, diskName);
    }

    /**
     * Create a disk without filling in the parameters from a model
     *
     * @param storage the parent Storage to contain the Disk
     * @param diskID  the unique identifier of the disk
     *
     * @return an odata2 Disk object initialized using given disk IDs
     *
     * @throws SchemaException
     */
    @SuppressWarnings("UnusedDeclaration")
    public final Disk createUnknownDisk(final Storage storage, final String diskID) throws SchemaException {
        return createUnknownDisk(storage, diskID, diskID);
    }

    /**
     * Removes the path prefix from the given file name, returning the "local" file name, i.e.
     * the name of the file as it would appear when listed from its immediate parent directory.
     *
     * @param fullyQualifiedFileName the fully qualified file name
     *
     * @return the file name with the path removed
     */
    private String getLocalFileName(final String fullyQualifiedFileName) {
        final File file = new File(fullyQualifiedFileName);
        return file.getName();
    }
}
