package com.onaro.sanscreen.acquisition.framework.mgmt;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.util.CyclicArray;

import java.util.*;
import java.util.concurrent.*;

import org.apache.log4j.Logger;

/**
 * This class handles asynchronous tasks running in the acquisition framework.
 * <p/>
 * <p/>There are many ocasions where framework services (or other objects) need to register one time or periodic tasks to
 * be executed. This class eliminates the need to create special thread for these tasks. It uses the global thread pool
 * handling all jobs in the acquisition framework and this reduces the overhead of creating threads by allowing the system
 * to reuse resources.
 */
public class AsynchronousTaskManager extends FrameworkService implements AsynchronousTaskManagerMBean {
    private static Logger logger = Logger.getLogger(AsynchronousTaskManager.class);

    private final TasksTimer timer;
    private final ExecutorService executor;
    private final SortedSet<Task> tasks = new TreeSet<Task>(new TaskComparator());

    private final CyclicArray<String> executionHistory = new CyclicArray<String>(30);

    private boolean traceEnabled = false;

    AsynchronousTaskManager(final ServiceStatusListener statusListener) {
        super("Asynchronous Task Manager", statusListener);

        // TODO: CHECK 180 is this is seconds
        executor = new ThreadPoolExecutor(0, 30, 180L, TimeUnit.SECONDS, new SynchronousQueue<Runnable>());

        timer = new TasksTimer(this);
    }

    protected void doStart() throws FrameworkException {
        timer.start();
    }

    protected void doStop() {
        timer.cancel();

        executor.shutdownNow();

        try {
            executor.awaitTermination(1, TimeUnit.SECONDS);
        }
        catch (InterruptedException e) {
            // Fall through...
        }
    }

    /**
     * Schedule a specific task to run at a given rate (millis)
     *
     * @param taskName    Name of task
     * @param logicToRun  Code to run when task is due
     * @param delayMillis Initial delay
     * @param rateMillis  Interval between runs
     *
     * @throws FrameworkException When task is scheduled more than once 
     */
    public void schedule(final String taskName, final Runnable logicToRun, final long rateMillis) throws FrameworkException {
        final Task task = new Task(taskName, logicToRun, rateMillis);

        if (logger.isDebugEnabled())
            logger.debug(getServiceName() + " - Task scheduled: " + taskName + " [" + rateMillis + "]");

        synchronized (tasks) {
            for (final Task aTask : tasks) {
                if (aTask.getName().equals(task.getName()))
                    throw new FrameworkException("Task " + task.getName() + " alraedy scheduled!");
            }

            tasks.add(task);
        }

        timer.tasksQueueWasUpdated();
    }

    public boolean unschedule(final String taskName) {
        boolean taskFound = false;

        Task task = null;
        synchronized (tasks) {
            // Find the task to cancel
            final Iterator<Task> it = tasks.iterator();
            while (it.hasNext()) {
                task = it.next();
                if (task.getName().equals(taskName)) {
                    it.remove();

                    taskFound = true;
                    break;
                }
            }
        }

        assert task != null;

        if (taskFound) {
            timer.tasksQueueWasUpdated();
            task.cancel();

            if (logger.isDebugEnabled())
                logger.debug(getServiceName() + " - Unscheduled task " + taskName);
        }
        else {
            logger.warn(getServiceName() + " - Unscheduled task failed: Couldn't find task " + taskName);
        }

        return taskFound;
    }

    /**
     * Retrieves a list of all the scheduled tasks in the task manager. Tasks that are canceled will be omitted.
     * This is used for management and monitoring purposes.
     */
    public String[] getScheduledTasks() {
        synchronized (tasks) {
            int i = 0;
            final String[] taskNames = new String[tasks.size()];

            for (final Task task : tasks)
                taskNames[i++] = "[" + new Date().toString() + "] - " + task.getName();

            return taskNames;
        }
    }

    public String[] getExecutionHistory() {
        final String[] history = new String[executionHistory.size()];
        return executionHistory.toArray(history);
    }

    public boolean isTraceEnabled() {
        return traceEnabled;
    }

    public void setTraceEnabled(boolean enable) {
        traceEnabled = enable;
    }

    private void reschedule(Task task) {
        synchronized (tasks) {
            // To keep the queue sorted by time, we need to delete and reinsert the task after the new time is calculated
            tasks.remove(task);
            task.updateNextRun();
            tasks.add(task);
        }

        timer.tasksQueueWasUpdated();
    }

    /**
     * This class handles the events for the tasks... It is working off the sorted tasks queue (first in queue is the task
     * to run first).
     *
     * HEADSUP: TODO Explain why we're doing this ugly mechanism instead of using java.concurrent 
     */
    private final class TasksTimer extends FrameworkThread<AsynchronousTaskManager> {
        private TasksTimer(final AsynchronousTaskManager parentService) {
            super("Asynchronous Manager Task Timer", parentService);
        }

        protected final boolean doRun() {
            while (!isStopping()) {
                final AsynchronousTaskManager.Task task;
                synchronized (tasks) {
                    // Get the first task in the queue (this is the task that should be executed first) (also check if
                    // the queue is empty)
                    task = (tasks.size() != 0) ? tasks.first() : null;
                }

                if (task == null) {
                    // No tasks: Sleep until there is something interesting coming (the thread will be interrupted when
                    // an event occurs of service is stopping)
                    try {
                        Thread.sleep(100000);
                    }
                    catch (InterruptedException e) {
                        // Fall through...
                    }

                    // We need to re-evaluate our stopping status and task queue
                    continue;
                }

                assert task != null;

                try {
                    // Sleep enough for the first task scheduled to run
                    Thread.sleep(Math.max(0, task.getNextRun() - System.currentTimeMillis()));

                    // If we're here - we woke up naturally (not interrupted)
                    final Future future = task.getFuture();

                    if (future != null && !future.isDone()) {
                        // Task is currently running. Skip this tick
                    }
                    else {
                        // Submit the logic for execution
                        task.execute();
                    }
                    
                    reschedule(task);
                }
                catch (InterruptedException e) {
                    // Fall through (re-evaluate our stopping status and task queue)
                }
            }

            return isStopping();
        }

        /**
         * Signal the tasks timer that something changed. This will cause the tasks timer to re-evaluate which tasks
         * should run first.
         */
        public final void tasksQueueWasUpdated() {
            interrupt();
        }
    }

    /**
     * Represents an asynchronous task running in the async task manager
     */
    private final class Task {
        private final static int MIN_INTERVAL = 100;

        private final String name;
        private final Runnable logicToRun;

        /**
         * Time between runs (millis)
         */
        private final long interval;

        /**
         * Time this task should run next (millis)
         */
        private long nextRun;

        /**
         * If executing, this contains the Future object representing the execution
         */
        private Future future = null;

        private Task(final String name, final Runnable logicToRun, final long interval) {
            this.name = name;
            this.logicToRun = logicToRun;

            if (interval <= 0)
                throw new IllegalArgumentException("Interval cannot be negative (or zero)");

            // Make sure the interval is not too small
            this.interval = Math.max(MIN_INTERVAL, interval);

            updateNextRun();
        }

        final void cancel() {
            if (future != null)
                future.cancel(true);
        }

        /**
         * Should only be called from ctor or reschedule!
         */
        final void updateNextRun() {
            nextRun = System.currentTimeMillis() + interval;
        }

        final String getName() {
            return name;
        }

        final long getNextRun() {
            return nextRun;
        }

        final Future getFuture() {
            return future;
        }

        final void execute() {
            // HEADSUP: There is no need to synchornize on future because there is only one thread (the timertask) thread
            // that initiates the change of this member!
            future = executor.submit(logicToRun);
            executionHistory.add("[" + new Date().toString() + "] " + getName());

            if (logger.isDebugEnabled() && isTraceEnabled())
                logger.debug(getServiceName() + " - Executing task " + getName());
        }
    }

    private static final class TaskComparator implements Comparator<Task> {
        public int compare(Task o1, Task o2) {
            int i = (int) (o1.nextRun - o2.nextRun);
            if (i != 0)
                return i;

            // Make sure comparator is unique by comparing names in case we have the same time.
            return o1.getName().compareTo(o2.getName());
        }
    }
}
