package com.onaro.sanscreen.acquisition.framework.mgmt;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionUnit;
import com.netapp.acquisition.aquisitionserverside.interfaces.remote.AcquisitionServerSession;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.Package;
import com.onaro.sanscreen.acquisition.framework.datasource.config.PackageConfiguration;
import org.apache.log4j.Logger;

import java.lang.ref.WeakReference;
import java.util.Properties;

/**
 * Handles events coming from the server (and GUI) and forwards these events to the appropriate data sources.
 */
public class DataSourceEventsManager extends FrameworkService implements DataSourceEventsManagerMBean {
    private static Logger logger = Logger.getLogger(DataSourceManager.class);

    private ForceAcqListenerThread eventsListener;

    private final WeakReference<CommunicationManager> communicationManager;
    private final WeakReference<AcquisitionUnitManager> acquisitionUnitManager;
    private final WeakReference<DataSourceManager> dataSourceManager;

    DataSourceEventsManager(final ServiceStatusListener statusListener,
                            final CommunicationManager communicationManager,
                            final AcquisitionUnitManager acquisitionUnitManager,
                            final DataSourceManager dataSourceManager) {
        super("Data Source Events Manager", statusListener);

        this.communicationManager = new WeakReference<CommunicationManager>(communicationManager);
        this.acquisitionUnitManager = new WeakReference<AcquisitionUnitManager>(acquisitionUnitManager);
        this.dataSourceManager = new WeakReference<DataSourceManager>(dataSourceManager);
    }

    protected void doStart() throws FrameworkException {
        eventsListener = new ForceAcqListenerThread(this);
        eventsListener.start();
    }

    protected void doStop() {
        eventsListener.cancel();
        eventsListener = null;
    }

    public Thread.State getForceAcqListenerThreadStatus() {
        return eventsListener.getState();
    }

    private AcquisitionUnit getAcquisitionUnit() throws FrameworkException {
        return acquisitionUnitManager.get().getAcquisitionUnit();
    }

    private static class ForceAcqListenerThread extends FrameworkThread<DataSourceEventsManager> {
        public ForceAcqListenerThread(final DataSourceEventsManager dataSourceEventsManager) {
            super("Force Acq Listener", dataSourceEventsManager);
        }

        @SuppressWarnings({"ConstantConditions"})
        public boolean doRun() {
            int numberOfExceptions = 0;
            while (!isStopping()) {
                boolean exception = false;

                try {
                    final Properties event = getCommunicationManager().getAcquisitionServerSide().listen(getParentService().getAcquisitionUnit().getId());

                    // No event was received, just continue pooling
                    if (event == null)
                        continue;

                    // Verify that we have the right acquisition unit!
                    final AcquisitionServerSession.Command command = AcquisitionServerSession.Command.valueOf(event.getProperty("command"));
                    final String dsName = event.getProperty("ds");
                    final String packageName = event.getProperty("package");
                                 
                    switch (command) {
                        case remove:
                        case update:
                            if (logger.isDebugEnabled())
                                logger.debug(getParentService().getServiceName() + " - DS configuration has changed.");

                            getDataSourceManager().synchronizeDataSourcesWithServer(true);
                            break;

                        case crawl:
                            if (logger.isDebugEnabled())
                                logger.debug(getParentService().getServiceName() + " - Received crawl message: " + event);

                            // If could not figure out the package, do nothing!
                            if (packageName == null || packageName.length() == 0)
                                continue;

                            // Signal the package to crawl (once!)
                            final DataSourceManager.DataSourceInfo dataSourceInfo = getDataSourceManager().getDataSourceInfo(dsName);

                            // Check if this package supports crawling
                            final Package thePackage = dataSourceInfo.getDataSourceInstance().getPackage(packageName);
                            if (!thePackage.getConfiguration().isCrawlSupported()) {
                                logger.info(getParentService().getServiceName() + " - Received crawl for package " + packageName + " that does not support crawling!");
                                continue;
                            }

                            final PackageConfiguration packageConfiguration = dataSourceInfo.getDataSourceInstance().getConfiguration().getPackageConfiguration(packageName);
                            packageConfiguration.turnOnCrawlingOnce();
                            break;

                        case force:
                            if (logger.isDebugEnabled())
                                logger.debug(getParentService().getServiceName() + " - Received event (force acq): " + event);

                            if (packageName == null || packageName.length() == 0) {
                                // Poll all packages
                                getDataSourceManager().pollNow( dsName, event );
                            }
                            else {
                                // Poll specific package
                                getDataSourceManager().pollNow(dsName, packageName, event);
                            }

                            break;

                        case licensingChange:
                            logger.info(getParentService().getServiceName() +" - Licensing change notification from server. "
                                        + "Preparing to restart all Services and Data sources.");
                            ServiceManager.getInstance().restart();
                            break;

                        default:
                            logger.error(getParentService().getServiceName() + " - Unrecognized event: " + event);
                    }
                }
                catch (Throwable e) {
                    getParentService().reportError("Failed to listen to acquisition events!", e);
                    exception = true;
                }
                finally {
                    //slow down persistent error.
                    if (exception) {
                        numberOfExceptions++;
                        try {
                            Thread.sleep(10000L);
                        }
                        catch (InterruptedException e) {
                            //fall through
                        }
                        if (numberOfExceptions > 360)
                            getParentService().reportFatal("Failed to communicate with server. Acquisition unit is shutting down", null);
                    }
                    else
                        numberOfExceptions = 0;

                }
            }

            return isStopping();
        }

        private CommunicationManager getCommunicationManager() {
            return getParentService().communicationManager.get();
        }

        private DataSourceManager getDataSourceManager() {
            return getParentService().dataSourceManager.get();
        }
    }
}
