package com.onaro.sanscreen.acquisition.framework.mgmt;

import org.apache.log4j.Logger;

import java.lang.ref.WeakReference;
import java.util.Map;

/**
 * Base class for all threads being used in the framework.
 * This base class handles the details of controlling the thread: start, stop, wait until stopped, handle uncaught
 * errors, etc...
 */
public abstract class FrameworkThread<P extends FrameworkService> extends Thread {
    private static Logger logger = Logger.getLogger(FrameworkThread.class);

    private final WeakReference<P> parentService;
    private volatile boolean stopping = false;

    /**
     * Constructs the framework thread with a pointer to it's parent {@link FrameworkService}
     *
     * @param threadName
     * @param parentService
     */
    public FrameworkThread(final String threadName, P parentService) {
        super();
        this.setName(threadName);

        this.parentService = new WeakReference<P>(parentService);
    }

    /**
     * Starts the thread (the thread's main function {@link #run} will be called when the actual thread is initialized
     */
    public synchronized void start() {
        stopping = false;

        super.start();
    }

    /**
     * Stop the thread and clean up.
     *
     * This method will return when the thread is actually stopped.
     */
    public void cancel() {
        if (isStopping())
            return;

        if (Thread.currentThread() == this) {
            // Sanity check: This system is designed to be stopped from a dedicated thread (other than the working threads).
            // If by any chance this thread tries to stop itself, this will cause a deadlock. In this case, it is safest to
            // stop the whole system.
            logger.fatal("Thread " + getName() + " tries to stop itself. This will cause a deadlock. System shuts down!!!");

            // Dump all the stack traces of all threads in the system!
            final StringBuilder sb = new StringBuilder(1000);
            sb.append("Stack trace:\n");

            for (final Map.Entry<Thread, StackTraceElement[]> entry : Thread.getAllStackTraces().entrySet()) {
                sb.append("\tThread: ").append(entry.getKey()).append(":\n");

                for (final StackTraceElement traceElement : entry.getValue())
                    sb.append("\t\t").append(traceElement.toString()).append("\n");
            }

            logger.fatal(sb.toString());

            // Now shutdown the system! We cannot continue the shutdown sequence because there is no way to guarantee what
            // will happen
            System.exit(2);
        }

        if (logger.isDebugEnabled())
            logger.debug(getParentService().getServiceName() + " - " + getThreadName() + " stopping...");

        // Signal the thread to stop
        stopping = true;
        interrupt();

        // Wait until it is reallly stopped.
        while (isAlive()) {
            try {
                Thread.sleep(200);
            }
            catch (InterruptedException e) {
                // Do nothing...
            }
        }

        if (logger.isDebugEnabled())
            logger.debug(getParentService().getServiceName() + " - " + getThreadName() + " stopped!");
    }

    /**
     * Returns true if the thread is supposed to be stopping. This should be used inside the implementation of the
     * {@link #doRun} method to exit the method.
     */
    public final boolean isStopping() {
        return stopping;
    }

    public final String getThreadName() {
        return getName();
    }

    public final P getParentService() {
        return parentService.get();
    }

    public final void run() {
        boolean threadExitedExpectedly = false;

        try {
            if (logger.isDebugEnabled())
                logger.debug(getParentService().getServiceName() + " - Thread " + getThreadName() + " started...");

            threadExitedExpectedly = doRun();
        }
        catch (Throwable e) {
            logger.fatal(getParentService().getServiceName() + " - " + e.getMessage(), e);

            // Fall through... This will cause the acquisition to stop (a fatal error will be reported)
            threadExitedExpectedly = false;
        }
        finally {
            if (logger.isDebugEnabled())
                logger.debug(getParentService().getServiceName() + " - Thread " + getThreadName() + " exiting...");

            // If we should not be stopping, then this is an unexpected exit!
            if (!isStopping() && !threadExitedExpectedly) {
                // Signal this thread as stopping. Since we're exiting the thread's main look, it will also die. This
                // will maximize the chances of this thread gracefully exiting.
                stopping = true;
                getParentService().reportFatal("Thread " + getThreadName() + " exited unexpectedly", null);
            }
        }
    }

    /**
     * The main method for the thread. This method should be overriden by derived classes to perform the thread's action.
     *
     * If the thread is expected to exit out of normal execution or if stopping, it should return true.
     * For any other reason, it should return false.
     *
     * @return true if thread exists expectedly, false otherwise
     */
    protected abstract boolean doRun();
}
