package com.onaro.sanscreen.acquisition.framework.mgmt.scheduler;

import com.onaro.sanscreen.acquisition.framework.mgmt.FrameworkService;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Timer;

class SchedulerTasksTimer<Type, Target, Service extends FrameworkService> {
    private Timer timer = new Timer(false);
    private List<TaskDescription> tasksInTimer = new LinkedList<TaskDescription>();

    public synchronized void schedule(final SchedulerTask<Type, Target, Service> task, final long delay, final long interval) {
        // The timer does not accept 0 interval. Set it to 1 msec instead of 0.
        timer.schedule(task, delay, Math.max(1, interval));

        tasksInTimer.add(new TaskDescription(task, delay, interval));
    }

    /**
     * Deletes all the tasks associated with given target
     *
     * @param target
     *
     * @return true if any task was found, false otherwise
     */
    public synchronized boolean unschedule(final Target target) {
        boolean found = false;

        // Find all occurences of this target in the timer queue
        for (final Iterator<TaskDescription> it = tasksInTimer.iterator(); it.hasNext();) {
            final SchedulerTask<Type, Target, Service> timerTask = it.next().task;
            if (timerTask.getTarget().equals(target)) {
                // Remove it from the queue
                it.remove();

                // And cancel it from the scheduler
                timerTask.cancel();

                found = true;
            }
        }

        return found;
    }

    /**
     * Looks for a task in the timer queue.
     *
     * @param target
     * @param type
     *
     * @return The task if found, null otherwise
     */
    public synchronized SchedulerTask<Type, Target, Service> find(final Target target, final Type type) {
        for (final TaskDescription timerTask : tasksInTimer) {
            if (timerTask.task.getTarget().equals(target) && timerTask.task.getType().equals(type)) {
                return timerTask.task;
            }
        }

        return null;
    }

    /**
     * Cancel all tasks in the timer
     */
    public synchronized void cancel() {
        timer.cancel();
        tasksInTimer.clear();
    }

    public synchronized int size() {
        return tasksInTimer.size();
    }

    public synchronized String[] describeStatus() {
        final String[] status = new String[tasksInTimer.size()];

        int i = 0;
        for (TaskDescription task : tasksInTimer) {
            status[i] = task.toString() ;
            i++;
        }

        return status;
    }

    class TaskDescription {
        final public SchedulerTask<Type, Target, Service> task ;
        final public long delay;
        final public long interval;

        public TaskDescription(final SchedulerTask<Type, Target, Service> task, final long delay, final long interval) {
            this.task = task;
            this.delay = delay;
            this.interval = interval;
        }

        @Override
        public String toString() {
            return task.toString() + ":" + Long.toString(delay/1000) + ":" + Long.toString(interval/1000) + task.formatExecutionHistory();
        }
    }
}
