package com.onaro.sanscreen.acquisition.framework.partialpoll;

import com.onaro.sanscreen.acquisition.framework.partialpoll.generated.*;
import org.apache.log4j.Logger;

import javax.xml.XMLConstants;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.Unmarshaller;
import javax.xml.datatype.XMLGregorianCalendar;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import java.io.File;
import java.io.InputStream;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

/**
 * Created with IntelliJ IDEA.
 * User: Jianwei
 * Date: 11/6/14
 * Time: 9:26 AM
 * To change this template use File | Settings | File Templates.
 */
public class PartialPollConfigurationRetriever
{
  private static Logger logger = Logger.getLogger( PartialPollConfigurationRetriever.class );
  private static final String configXML = "preferences/partial_poll_config.xml";
  private static final String XSD = "preferences/partial_poll_config.xsd";

  private XMLGregorianCalendar fullPollDailyTime;
  private Map<String, ZapiCommandListType> zapiCalls;

  private PartialPollConfigurationRetriever()
  {
    loadConfiguration();
  }

  //implementation of lazy initialization
  private static class LazyHolder
  {
    private static final PartialPollConfigurationRetriever INSTANCE = new PartialPollConfigurationRetriever();
  }

  public static PartialPollConfigurationRetriever getInstance()
  {
    return LazyHolder.INSTANCE;
  }

  public Map<String, ZapiCommandListType> getZapicalls()
  {
    return zapiCalls;
  }

  public void setZapiCalls(Map<String, ZapiCommandListType> zapiCalls)
  {
    this.zapiCalls = zapiCalls;
  }

  public XMLGregorianCalendar getFullPollDailyTime()
  {
    return fullPollDailyTime;
  }

  public void setFullPollDailyTime(XMLGregorianCalendar fullPollTime)
  {
    this.fullPollDailyTime = fullPollTime;
  }

  /**
   * This method is used to load partial poll configuration xml file
   */
  private void loadConfiguration()
  {
    try
    {
      //read the xsd for file validation
      final URL schemaURL = this.getClass().getClassLoader().getResource( XSD );
      if( schemaURL == null )
      {
        logger.info( "Cannot locate partial_poll_config.xsd." );
        throw new RuntimeException( "Cannot locate partial_poll_config.xsd." );
      }

      //Read the config file preferences/partial_poll_config.xml
      final InputStream partialPollConfig = this.getClass().getClassLoader().getResourceAsStream( configXML );
      if( partialPollConfig == null )
      {
        logger.info( "Cannot locate partial_poll_config.xml." );
        throw new RuntimeException( "Cannot locate partial_poll_config.xml." );
      }

      //Configure JAXB with the correct validation schema
      JAXBContext jaxbContext = JAXBContext.newInstance( ObjectFactory.class );
      SchemaFactory schemaFactory = SchemaFactory.newInstance( XMLConstants.W3C_XML_SCHEMA_NS_URI );
      Schema schema = schemaFactory.newSchema(schemaURL);
      Unmarshaller unmarshaller = jaxbContext.createUnmarshaller();
      unmarshaller.setSchema( schema );

      //un marshal the xml
      PartialPollConfiguration partialConfig = (PartialPollConfiguration) unmarshaller.unmarshal( partialPollConfig );

      //Set partial poll attributes
      fullPollDailyTime = partialConfig.getFullPollDailyTime();
      if( fullPollDailyTime == null ) throw new RuntimeException( "The fullPollDailyTime is not set correctly in the partial_poll_config.xml file. Should be in the form <fullPollDailyTime>12:00:00</fullPollDailyTime>, note white space causes issues." );

      //Process the Zapi commands Groups
      ZapiCommandGroupsType groups = partialConfig.getZapiCommandGroups();
      if( groups == null )
        logger.warn( "There is no zapi commands defined in PartialPollConfig.xml." );
      else
      {
        zapiCalls = new HashMap<>();
        for( ZapiCommandGroupType group : groups.getZapiCommandGroup() )
        {
          //Make sure we have a valid version
          String version = group.getVersion();
          if( fullPollDailyTime == null ) throw new RuntimeException( "The version is not set correctly for one of the groups in the partial_poll_config.xml file, note white space causes issues" );

          //Add list of zapis supported for this version
          ZapiCommandListType zapiList = group.getZapiCommands();
          zapiCalls.put( version, zapiList );
        }
      }
    }
    catch( Exception ex )
    {
      logger.error( "Could not configure partial polling error reading partial_poll_config.xml", ex );
      throw new RuntimeException( "Could not configure partial polling error reading partial_poll_config.xml", ex );
    }
  }
}
