package com.onaro.sanscreen.acquisition.framework.recordings;

import com.onaro.sanscreen.acquisition.framework.common.SortedProperties;
import com.onaro.sanscreen.acquisition.framework.datasource.BaseDataSource;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceOutputFileType;
import com.onaro.sanscreen.acquisition.framework.datasource.config.PackageConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.io.*;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import org.apache.log4j.Logger;

import java.io.*;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;

/**
 * Class that handles recording files that can be generated during a sample.
 */
public class RecordingSample {
    private final static Logger logger = Logger.getLogger(BaseDataSource.class);

    private final static DateFormat dateFormat = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss_SSS");

    private static long lastTimestamp = Long.MIN_VALUE;

    /**
     * List of files that were created during the acquisition. This list allows the framework to collect the data
     * generated during the smaple and decide weather to keep it or to delete it.
     */
    private final List<DSFile> fileHandles = new LinkedList<DSFile>();
    private final Properties description = new SortedProperties();
    private final long timestamp;

    RecordingSample() {
        this.timestamp = System.currentTimeMillis();
    }

    /**
     * Returns list of files that were created during the acquisition. This list allows the framework to collect the data
     * generated during the smaple and decide weather to keep it or to delete it.
     *
     * @return list of files that were created during the acquisition. This list allows the framework to collect the data
     */
    public List<DSFile> getFileHandles() {
        return fileHandles;
    }

    public long getTimestamp() {
        return timestamp;
    }

    public Properties getDescription() {
        return description;
    }

    public void recordObject(final PackageConfiguration configuration, final Serializable param, final String keyPrefix, final String key) throws IOException {
        recordObject(configuration, param, DataSourceOutputFileType.recording, keyPrefix, key);
    }

    public void recordObject(final PackageConfiguration configuration, final Serializable param, final DataSourceOutputFileType type, final String keyPrefix, final String key) throws IOException {
        final DSFileOutputStream paramOutputFile = createStreamWithDescription(configuration, type, keyPrefix, key);

        final ObjectOutputStream encoder = new ObjectOutputStream(paramOutputFile);
        encoder.writeObject(param);
        encoder.flush();
        encoder.close();
    }

    public void addFileDescription(final String fileName, final String keyPrefix, final String key) {
        getDescription().put(keyPrefix + "-" + key, fileName);
    }

    public DSFileOutputStream createRecordingStreamWithDescription(final PackageConfiguration configuration, final String keyPrefix, final String key) throws IOException {
        return createStreamWithDescription(configuration, DataSourceOutputFileType.recording, keyPrefix, key);
    }

    public DSFileOutputStream createStreamWithDescription(final PackageConfiguration configuration, final DataSourceOutputFileType type, final String keyPrefix, final String key) throws IOException {
        final DSFileOutputStream paramOutputFile = createBinaryOutputFile(configuration, type, keyPrefix, "dat");
        addFileDescription(new File(paramOutputFile.getFileName()).getName(), keyPrefix, key);
        return paramOutputFile;
    }

    public DSFileWriter createRecordingTextWithDescription(final PackageConfiguration configuration, final String keyPrefix, final String key) throws IOException {
        return createTextWithDescription(configuration, DataSourceOutputFileType.recording, keyPrefix, key);
    }

    public DSFileWriter createTextWithDescription(final PackageConfiguration configuration, final DataSourceOutputFileType type, final String keyPrefix, final String key) throws IOException {
        final DSFileWriter paramOutputFile = createTextualOutputFile(configuration, type, keyPrefix, "dat");
        addFileDescription(new File(paramOutputFile.getFileName()).getName(), keyPrefix, key);
        return paramOutputFile;
    }


    public void reset() {
        closeFiles();
        deleteFiles();
        fileHandles.clear();
        description.clear();
    }

    /**
     * Close all the files created during the sample. And return a list of names back to be sent to he server.
     *
     * <p/>Only files created by the framework methods will be closed before being deleted. Files added to the collection
     * using the {@link RecordingSample#addRecordingFile} will not be closed because
     * they were not created by the framework. It is the responsibility of the creator of the file to make sure the file
     * is flushed and closed before the DS sample is finished.
     *
     * <p/>Note: This method will also clear the set of files maintained by the base class to prepare for the next sample.
     */
    protected void closeFiles() {
        for (final DSFile fileToClose : fileHandles) {
            // Close only files that the framework openned. If the value is null, then this file was added using the addRecordingFile method.
            try {
                fileToClose.close();
                if (!new File(fileToClose.getFileName()).exists()) {
                    if (logger.isDebugEnabled())
                        logger.debug("RecordingSample - file does not exist : " + fileToClose.getFileName());
                }
            }
            catch (IOException e) {
                if (logger.isDebugEnabled())
                    logger.debug("RecordingSample - Failed to close file " + fileToClose.getFileName(), e);
            }
            catch (IllegalStateException e) {
                if (logger.isDebugEnabled())
                    logger.debug("RecordingSample - Failed to close file " + fileToClose.getFileName(), e);
            }
        }
    }

    /**
     * Delete all the files reported in <code>fileHandles</code>. Files will be closed first and then deleted from the file system.
     *
     * <p/>This method is used when the framework detected that a report is identical to the previous report by this datasource
     */
    protected void deleteFiles() {
        for (final DSFile fileToDelete : fileHandles) {
            final File file = new File(fileToDelete.getFileName());
            file.delete();
        }
    }

    /**
     * Adds a file to the collections of files generated by the current sample.
     *
     * <p/>This method is to be used when an external program generated the file and we just want to make sure it will
     * be included with the report of the datasource or deleted if no changed were detected.
     *
     * <p/>Whenever possible, use the {@link #createBinaryOutputFile}, {@link #createTextualOutputFile} } to
     * ensure that the framework will also close any file left open and maintain a consistent file naming convension.
     *
     * @param fileName File name to add
     *
     * @throws FileNotFoundException
     */
    public final synchronized void addRecordingFile(final String fileName) throws FileNotFoundException {
        fileHandles.add(new DSFileExternallyManaged(fileName));
    }

    public final synchronized void addRecordingFileWithDescritpion(final String fileName, final String keyPrefix, final String key) throws FileNotFoundException {
        addRecordingFile(fileName);
        addFileDescription(new File(fileName).getName(), keyPrefix, key);
    }

    /**
     * Adds a file to the collections of files generated by the current sample.
     *
     * <p/>This method is to be used when an external program generated the file and we just want to make sure it will
     * be included with the report of the datasource or deleted if no changed were detected.
     *
     * <p/>Whenever possible, use the {@link #createBinaryOutputFile}, {@link #createTextualOutputFile} } to
     * ensure that the framework will also close any file left open and maintain a consistent file naming convension.
     *
     * @param stream Stream to add
     */
    public final synchronized void addRecordingStream(final DSFile stream) {
        fileHandles.add(stream);
    }

    public final synchronized DSFileOutputStream createBinaryOutputFile(final PackageConfiguration configuration, final DataSourceOutputFileType type, final String custom, final String extension) throws IOException {
        final File file = new File(formatFileName(ServiceManager.getInstance().getLogsManager().getAcqRecordingDir(),
                                                  configuration,
                                                  type,
                                                  custom,
                                                  extension));

        final DSFileOutputStream dsfos = new DSFileOutputStream(file.getAbsolutePath(), file);
        fileHandles.add(dsfos);

        return dsfos;
    }

    public final DSFileOutputStream createBinaryRecordingFile(final PackageConfiguration configuration) throws IOException {
        return createBinaryOutputFile(configuration, DataSourceOutputFileType.recording, null, "dat");
    }

    public final synchronized DSFileWriter createTextualOutputFile(final PackageConfiguration configuration, final DataSourceOutputFileType type, final String custom, final String extension) throws IOException {
        final File file = new File(formatFileName(ServiceManager.getInstance().getLogsManager().getAcqRecordingDir(),
                                                  configuration,
                                                  type,
                                                  custom,
                                                  extension));

        final DSFileWriter dsfos = new DSFileWriter(file.getAbsolutePath(), file);
        fileHandles.add(dsfos);

        return dsfos;
    }

    public final synchronized DSFileCRCWriter createTextualOutputCRCFile(final PackageConfiguration configuration, final DataSourceOutputFileType type, final String custom, final String extension) throws IOException {
        final File file = new File(formatFileName(ServiceManager.getInstance().getLogsManager().getAcqRecordingDir(),
                                                  configuration,
                                                  type,
                                                  custom,
                                                  extension));

        final DSFileCRCWriter dsCRCfos = new DSFileCRCWriter(file.getAbsolutePath(), file);
        fileHandles.add(dsCRCfos);

        return dsCRCfos;
    }

    public final DSFileWriter createTextRecordingFile(final PackageConfiguration configuration) throws IOException {
        return createTextualOutputFile(configuration, DataSourceOutputFileType.recording, null, "txt");
    }

    public final DSFileWriter createTextRecordingFile(final PackageConfiguration configuration, final String custom) throws IOException {
        return createTextualOutputFile(configuration, DataSourceOutputFileType.recording, custom, "txt");
    }

    /**
     * Formats file names in a uniform way. All fileHandles include the data source name, met datasource name, a time stamp
     * and a special prefix to distinguish this file from other fileHandles on the same sample.
     * <p/>
     * <p/>Directory will always be the acq log directory (from the "ACQUISITION_LOG_DIR" in the preferences)
     * Format looks like:
     * workingdirectory/type_metadatasourcename_datasourcename_timestamp_custom.extension
     *
     * @param dir
     * @param configuration
     * @param type
     * @param custom is a way to distinguish between multiple recordings. Could be NULL
     * @param extension file extnsion
     *
     * @return formated file name
     */
    public static String formatFileName(final File dir,
                                                     final PackageConfiguration configuration,
                                                     final DataSourceOutputFileType type,
                                                     final String custom,
                                                     final String extension) {
        final StringBuilder sb = new StringBuilder(50);

        sb.append(dir).append(File.separator);
        sb.append(type).append("_");
        sb.append(configuration.getPackageName()).append("_");
        sb.append(configuration.getDataSourceName()).append("_");
        sb.append(Long.toString(configuration.getDataSourceType())).append("_");
        sb.append(getUniqueTimestamp());

        if (custom != null && custom.length() != 0) {
            sb.append("_").append(custom);
        }

        sb.append(".").append(extension);

        return sb.toString();
    }

    public static String formatDirName(final File dir,
                                       final PackageConfiguration configuration,
                                       final DataSourceOutputFileType type,
                                       final String custom) {
        final StringBuilder sb = new StringBuilder(50);

        sb.append(dir).append(File.separator);
        sb.append(type).append("_");
        sb.append(configuration.getPackageName()).append("_");
        sb.append(configuration.getDataSourceName()).append("_");
        sb.append(Long.toString(configuration.getDataSourceType())).append("_");
        sb.append(getUniqueTimestamp());

        if (custom != null && custom.length() != 0) {
            sb.append("_").append(custom);
        }

        return sb.toString();
    }

    public static String formatSampleFileName(final PackageConfiguration configuration) {
        return configuration.getPackageName() + "_" +
               configuration.getDataSourceName() + "_" +
               configuration.getDataSourceID() + "_" +
               getUniqueTimestamp() + ".zip";
    }

    /**
     * @return A unique timestamp string. Ensures that the returned timestamp
     * is always unique, even if this method is called multiple times during a
     * 1 ms period.
     */
    private synchronized static String getUniqueTimestamp() {
        lastTimestamp = Math.max(System.currentTimeMillis(), lastTimestamp + 1);
        return dateFormat.format(new Date(lastTimestamp));
    }

    public void setSampleStatus(final SampleStatus status) {
        description.setProperty("system.sample.status", status.name());
    }

    public void setSamplePartialSuccessInformation(final int numberOfOriginators, final int numberOfExceptions) {
        description.setProperty("system.sample.numberOfOriginators", Integer.toString(numberOfOriginators));
        description.setProperty("system.sample.numberOfExceptions", Integer.toString(numberOfExceptions));
    }

    public enum SampleStatus {
        SUCCESS, FAILED, PARTIAL_SUCCESS
    }
}
