package com.onaro.sanscreen.acquisition.netappfoundation.datasource.config;

import com.netapp.oci.platform.acq.interfaces.data.PackagePollCompleteStatus;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.BaseDataSource;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceLifeCycleListener;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceSampleListener;
import com.onaro.sanscreen.acquisition.framework.datasource.config.BasePollingPackageConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.config.DataSourceConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.metadata.PackageMetaData;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.*;
import com.onaro.sanscreen.util.Schema;
import com.onaro.util.DeviceFilter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.lang.reflect.Constructor;
import java.util.*;

/**
 * Represents the configuration of a data source (types, properties, etc...)
 */
public abstract class NetAppFoundationConfiguration<FP extends NetAppFoundationPackage> extends BasePollingPackageConfiguration<FP> {

    private static final Logger logger = LoggerFactory.getLogger( NetAppFoundationConfiguration.class );

    private final String ip;
    private final String user;
    private String password;


    private final DeviceFilter filter;
    private boolean retireUnreportedOriginators;

    private final Map<String, AcquisitionPollStatus> lastPollStatusByPackage = new HashMap<String, AcquisitionPollStatus>();


    /**
     * Instantiates this configuration object. This object is contructed by the {@link PackageMetaData#instantiateConfiguration} method
     *
     * @param dataSourceConfiguration
     * @param metaData
     * @param dataSourceName
     * @param dataSourceID
     * @param props
     *
     * @throws FrameworkException
     */
    public NetAppFoundationConfiguration(final DataSourceConfiguration dataSourceConfiguration, final PackageMetaData metaData, final String dataSourceName, final long dataSourceID, final Properties props) throws FrameworkException {
        //Note package "name" here is hard-coded since "Feature" basis doesn't support OCIE
        this(dataSourceConfiguration, metaData, dataSourceName, dataSourceID, "Inventory", props);
    }

    /**
     * Instantiates this configuration object. This object is contructed by the {@link PackageMetaData#instantiateConfiguration} method
     *
     * @param dataSourceConfiguration
     * @param metaData
     * @param dataSourceName
     * @param dataSourceID
     * @param packageAuditName
     * @param props
     *
     * @throws FrameworkException
     */
    public NetAppFoundationConfiguration(final DataSourceConfiguration dataSourceConfiguration, final PackageMetaData metaData, final String dataSourceName, final long dataSourceID, final String packageAuditName, final Properties props) throws FrameworkException {
        super(dataSourceConfiguration, metaData, dataSourceName, dataSourceID, packageAuditName, props, true);
        // get common properties
        String ip = properties.getProperty("ip");
        if (ip == null)
            throw new FrameworkException("Configuration for " + toString() + " is missing IP");
        this.ip = ip.trim();
        user = properties.getProperty("user");
        password = properties.getProperty("password");
        filter = new DeviceFilter(getProperty("filter.exclude", ""));
        retireUnreportedOriginators = true;
        props.remove("enabled");
    }

    /**
     * @return true if this is a MOID data source
     */
    public abstract boolean isMOID();

    public final String getIP() {
        return ip;
    }

    public final String getUser() {
        return user;
    }

    public final String getPassword() {
        return password;
    }

    public int getMaxRecordingSamples() {
        return 1;
    }

    public boolean isEnabled() {
        return true;
    }

    /**
     * @return true if the device identified by <code>id</code> is filtered out
     */
    public final boolean isFiltered(final String id) {
        return filter.contains(id);
    }

    public DeviceFilter getFilter() {
        return filter;
    }

    /**
     * If true, acquisition will delete all originators ever reported
     * by the datasource that are not reported in the current
     * *completely successful* poll.
     * If false, no originators will be deleted.
     * Default value is true.
     */
    public boolean isRetireUnreportedOriginators() {
        return true;
    }



    /**
     * Instantiate a new configuration object for this package based on the meta data described in this class.
     *
     * <p/>The configuration object was already created in the right class loader (see {@link PackageMetaData#instantiateConfiguration})
     * so we can simply instantiate a new Foundation package without having to worry about it.
     */
    public FP instantiatePackage(final BaseDataSource dataSource) throws FrameworkException {
        // Create a foundation listener for this package...
        final NetAppFoundationListener foundationListener = new NetAppFoundationListener(this);

        return instantiatePackage(dataSource, foundationListener, foundationListener, foundationListener, foundationListener, foundationListener);
    }

    public FP instantiatePackage(final BaseDataSource dataSource,
                                 final DataSourceLifeCycleListener dataSourceLifeCycleListener,
                                 final DataSourceSampleListener dataSourceSampleListener,
                                 final NetAppFoundationUpdatesListener foundationUpdatesListener,
                                 final NetAppOriginatorProvider originatorProvider,
                                 final NetAppClockSynchronizer clockSynchronizer) throws FrameworkException {
        try {
            final Class packageImplementationClass = getMetaData().getPackageImplementationClass();

            final Class[] constructorSignature = {BaseDataSource.class,
                                                  this.getClass(),
                                                  DataSourceLifeCycleListener.class,
                                                  DataSourceSampleListener.class,
                                                  NetAppFoundationUpdatesListener.class,
                                                  NetAppOriginatorProvider.class,
                                                  NetAppClockSynchronizer.class};

            final Constructor constructor = packageImplementationClass.getConstructor(constructorSignature);

            // Construct the data source instance
            final Object[] constructorValues = {dataSource, this, dataSourceLifeCycleListener, dataSourceSampleListener, foundationUpdatesListener, originatorProvider, clockSynchronizer};

            //noinspection unchecked
            return (FP) constructor.newInstance(constructorValues);
        }
        catch (Throwable e) {
            throw new FrameworkException("Failed to instantiate foundation package for data source " + getDataSourceName(), e);
        }
    }

    // for AcquisitionStatusRemote#packageCompleted
    public static class AcquisitionPollStatus {
        public static final Schema schema = Schema.NETAPP;
        private final long datasourceID;
        private final String datasourceName;
        private final String packageName;
        private int pollDuration;
        private final List<Map<PackagePollCompleteStatus.Error, Map<String, String>>> packagePollErrors = new ArrayList<Map<PackagePollCompleteStatus.Error, Map<String, String>>>();


        private long pollStartTime = -1;
        private long sampleTime = -1;
        private long totalZAPITime = 0;
        private long totalReportTime = 0;
        private boolean dataReported = false;   // Successful foundation always reports data.
                                                // Performance polls may be successful (i.e. didn't fail),
                                                // but still have no data to report.
                                                // This flag will identify the case where performance poll
                                                // was successful, but there was no data to report.

        public AcquisitionPollStatus(final long datasourceID, final String datasourceName, final String packageName, final int pollDuration) {
            this.datasourceID = datasourceID;
            this.datasourceName = datasourceName;
            this.packageName = packageName;
            this.pollDuration = pollDuration;
        }

        public Long getDatasourceID() {
            return datasourceID;
        }

        public String getDatasourceName() {
            return datasourceName;
        }

        public String getPackageName() {
            return packageName;
        }

        public void setPollDuration(final int pollDuration) {
            this.pollDuration = pollDuration;
        }

        public void setTotalZAPITime(final long totalZAPITime) {
            this.totalZAPITime = totalZAPITime;
        }

        public long getTotalZAPITime() {
            return totalZAPITime;
        }

        public void setTotalReportTime(final long totalReportTime) {
            this.totalReportTime = totalReportTime;
        }

        public int getPollDuration() {
            return pollDuration;
        }

        public void recordOCIEReportTime(final String reportAPI, final long reportTime) {
            synchronized(this) {
                totalReportTime += reportTime;
            }
        }

        public long getTotalReportTime() {
            return totalReportTime;
        }

        public long getPollStartTime() {
            return pollStartTime;
        }

        public void setPollStartTime(final long pollStartTime) {
            this.pollStartTime = pollStartTime;
        }

        public long getSampleTime() {
            return sampleTime;
        }

        public void setSampleTime(final long sampleTime) {
            this.sampleTime = sampleTime;
        }

        public boolean isDataReported() {
            return dataReported;
        }

        public void setDataReported(boolean dataReported) {
            this.dataReported = dataReported;
        }

        public void addDataSourceError(final Map<PackagePollCompleteStatus.Error, Map<String, String>> pollErrorDetails) {
            this.packagePollErrors.add(pollErrorDetails);
        }

        public PackagePollCompleteStatus getPackagePollCompleteStatus(final boolean success) {
            final PackagePollCompleteStatus packagePollCompleteStatus = new PackagePollCompleteStatus(datasourceName, datasourceID,
                    packageName, success, pollStartTime, sampleTime, pollDuration);

            for (final Map<PackagePollCompleteStatus.Error, Map<String, String>> packagePollError : packagePollErrors) {
                for (final Map.Entry<PackagePollCompleteStatus.Error, Map<String, String>> packagePollEntry : packagePollError.entrySet()) {
                    final PackagePollCompleteStatus.Error error = packagePollEntry.getKey();
                    final Map<String, String> errorProperties = packagePollEntry.getValue();
                    packagePollCompleteStatus.addError(error, errorProperties);
                }
            }
            return packagePollCompleteStatus;
        }
    }

    public void putAcquisitionPollStatus(final String packageName, final AcquisitionPollStatus pollStatus) {
        lastPollStatusByPackage.put(packageName, pollStatus);
    }

    public AcquisitionPollStatus getAcquisitionPollStatus(final String packageName) {
        return lastPollStatusByPackage.get(packageName);
    }
}
