package com.onaro.sanscreen.acquisition.netappfoundation.datasource.dependant;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.*;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import com.onaro.sanscreen.acquisition.framework.recordings.RecordingSample;
import com.onaro.sanscreen.acquisition.framework.recordings.RecordingSession;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.NetAppFoundationPackage;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.NetAppFoundationUpdatesListener;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.dependant.config.NetAppFoundationDependantConfiguration;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.dependant.recording.FoundationDependentRecordingHandler;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.StorageRoot;
import com.onaro.sanscreen.acquisition.schema.Root;
import org.apache.log4j.Logger;

import java.io.*;
import java.util.Properties;

/**
 * This is a base class for every package that depends on foundation data, such as Array or Switch Performance.
 * It makes foundation data available synchronization-free.
 * Also, this class takes care of managing mixed recordings: foundation and dependant package recordings are kept in the same RecordingSession
 */
public abstract class NetAppFoundationDependantPackage<C extends NetAppFoundationDependantConfiguration>
        extends BasePollingPackage<C> implements NetAppBaseFoundationDependantPackageMBean,
                                                 NetAppFoundationUpdatesListener,
                                                 DataSourceSampleListener,
                                                 DataSourceLifeCycleListener {

    private static Logger logger = Logger.getLogger(NetAppFoundationDependantPackage.class);

    private volatile File lastRecordedFoundationZipFile = null;

    private final FoundationDependentRecordingHandler recordingHandler = new FoundationDependentRecordingHandler(this);

    protected NetAppFoundationDependantPackage(final BaseDataSource dataSource,
                                             final C configuration,
                                             final DataSourceSampleListener dataSourceSampleListener) {
        super(dataSource, configuration, dataSourceSampleListener);

        final NetAppFoundationPackage foundationPackage = dataSource.getPackage(NetAppFoundationPackage.PACKAGE);
        if (foundationPackage == null)
            throw new IllegalStateException("Could not find foundation package");

        // This class will register itself to monitor foundation and dependant package events.
        dataSource.addDataSourceSampleListener(NetAppFoundationPackage.PACKAGE, this);
        foundationPackage.addUpdatesListener(this);

        dataSource.addDataSourceSampleListener(getPackageName(), this);
        dataSource.addDataSourceLifeCycleListener(this);
    }

    public void dataSourceStarting(final BaseDataSource dataSource) throws FrameworkException {

        //take a recording of foundation dependent packages on every start.
        int numberOfSamples = getConfiguration().getRecordingSamples();
        if (isSimulating())
            numberOfSamples = getConfiguration().getSimulationSession().getSamples().size();

        recordingHandler.requestRecording(numberOfSamples);
    }

    public void dataSourceStopping(final BaseDataSource dataSource) throws FrameworkException {
        recordingHandler.stopRecordingSession();
    }

    public void dataSourceStatusChanged(final BaseDataSource dataSource, final DataSourceStatus oldStatus, final DataSourceStatus newStatus, final String message) {
        // nothing to do
    }

    /**
     * This is for only for data sources from 6.0 and up   NEW SCHEMA.
     *
     * Called by data source when it finished creating the XML. This will also update the
     * status of the data source to "Success"
     *
     * @param dataSource
     * @param originatorKey
     * @param originatorDisplayName
     * @param netAppRoot The generated XML
     * @param changed Indicates if the framework believes this data is changed from the last report
     * @param context Context object containing device specific data being passed arround by foundation to other packages
     *
     * @throws FrameworkException when could not update data to the server
     */
    public void reportNetAppData(final BaseDataSource dataSource,
                              final String originatorKey,
                              final String originatorDisplayName,
                              final StorageRoot netAppRoot,
                              final File netappDataFile,
                              final boolean changed,
                              final Serializable context) throws FrameworkException {

        // We only need foundation data if it changed. But if we are recording we want to have it anyway.
        if (!changed && recordingHandler.skipNotChangedReportOdata()) {
            logger.debug(getLogPrefix() + " - skip non changed reportOdata");
            return;
        }

        //save this data to "lastUpdatesFromFoundation"---this can make sure performancePackage can always have the latest foundationData.
        logger.debug(getLogPrefix() + " - cache changed reportOdata for key " + originatorKey);
        // Process the foundation data now so the whole cluster can be garbage collected.
        synchronizeWithFoundation(new NetAppFoundationData(originatorKey, netAppRoot, changed, context));
    }


    public void reportManageURL(final BaseDataSource dataSource, final String manageURL) {
        // Do nothing - foundation uses it to update the management URL for this data source.
    }

    public void startingSample(final BaseDataSource dataSource, final PollingPackage pollType) {
        logger.debug(getLogPrefix() + " - process startingSample()");

        if (pollType.getPackageName().equals(NetAppFoundationPackage.PACKAGE))
            recordingHandler.startingFoundationSample();
        else if (pollType.getPackageName().equals(getPackageName())) {
            //This will trigger the creation of a sample for recording if appropriate.  Also will put in the
            //foundation data last received.   As an optimization, we only put this in if has changed since
            //the last sample, as applicable.

            //This addresses a theoretical race condition when finishSample were to be called on dependent package
            //before on foundation.  In that case, if the dependent thread started at the exact right moment, we
            //could have a null pointer here.   Ordering of the packages may prevent this, but keep it for safety.
            if (this.lastRecordedFoundationZipFile != null) {
                //update state machine and reset current sample.
                recordingHandler.prepareRecordingForSample();
                if (isRecording()) {
                    if (logger.isDebugEnabled()) {
                        logger.debug(getLogPrefix() + " - preparing new recording sample.");
                    }
                    final String previousFoundationCopy = findPreviousFoundationCopy(getRecordingSession());
                    final boolean isFirstSample = getRecordingSession().getSamples().size() == 1;
                    try {
                        final File localFoundationZipFile;
                        synchronized (this) {
                            localFoundationZipFile = this.lastRecordedFoundationZipFile;    //updated by foundation callback.
                        }

                        final String localZipCopyName = generateFoundationCopyName(localFoundationZipFile.getName());
                        if (isFirstSample || (!localZipCopyName.equals(previousFoundationCopy))) {
                            //add new foundation data!
                            final File foundationCopy = makeFoundationCopy(localFoundationZipFile);
                            final RecordingSample recordingSample = getRecordingSession().getCurrentSample();
                            recordingSample.addRecordingFileWithDescritpion(foundationCopy.getAbsolutePath(), "foundation", "recording");
                        }
                    }
                    catch (final IOException ioe) {
                        //cancel recording!   Somebody deleted the foundation zip file!
                        recordingHandler.stopRecordingSession();
                        logger.error(getLogPrefix() + " Failed to write foundation data to dependent recording.  Cancelling dependent recording!", ioe);
                    }

                }
            }
        }
        else {
            if (logger.isDebugEnabled()) {
                logger.debug(getLogPrefix() + " - skipping foundation dependent recording setup - no foundation zip file reference!");
            }
        }

    }

    /**
     * Find the last copy of foundation used in the current recording session.  Method assumes the package is
     * recording.
     *
     * <p>If this is the first sample, this method will return null.   Otherwise, loops through all older samples
     * and finds most recent copy of foundation recording used.
     *
     * @param currentSession the current recording session.
     *
     * @return File name of most recent foundation recording copy stored in the recording session.  NULL if nto found.
     */
    private String findPreviousFoundationCopy(final RecordingSession currentSession) {

        //search over samples and find latest sample entry with foundation marker.  Should not be null, except
        //on first copy.
        String previousFoundationRecordingCopy = null;
        for (final RecordingSample sample : currentSession.getSamples()) {   //list is ordered, first to last.
            if (sample.getDescription().getProperty("foundation-recording") != null) {
                previousFoundationRecordingCopy = sample.getDescription().getProperty("foundation-recording");
            }
        }

        return previousFoundationRecordingCopy;
    }

    /**
     * Copy give file into a new file with name as determined in {@link #generateFoundationCopyName}.  Returns
     * handle to the new copy.   To be used when needing a different copy of the file (e.g. foundation zip file)
     * to be placed in a recording.
     *
     * @param foundationZipFile zip file handle as originally reported by foundation.
     *
     * @return handle to the copy.
     *
     * @throws IOException in event of errors making the copy file.
     */
    private File makeFoundationCopy(final File foundationZipFile) throws IOException {
        final File parentDir = foundationZipFile.getParentFile();
        final String copyName = generateFoundationCopyName(foundationZipFile.getName());
        final File copy = new File(parentDir, copyName);
        BufferedInputStream in = null;
        BufferedOutputStream out = null;

        try {
            in = new BufferedInputStream(new FileInputStream(foundationZipFile));
            out = new BufferedOutputStream(new FileOutputStream(copy));

            final int bufferSize = 8 * 1024;
            final byte[] buffer = new byte[bufferSize];
            int count = in.read(buffer);
            while (count > 0) {
                out.write(buffer, 0, count);
                count = in.read(buffer);
            }
        }
        finally {
            if (in != null) {
                try {
                    in.close();
                }
                catch (Exception e) {
                    //ignore.
                }
            }

            if (out != null) {
                try {
                    out.close();
                }
                catch (Exception e) {
                    //ignore.
                }
            }
        }

        return copy;
    }

    /**
     * Name to use for copy of foundation update zip file.   Will be placed in dependent recording.    Assumes
     * passed filename is NOT null
     *
     * @param fileName the original file name.
     *
     * @return file name to use for the copy.
     */
    private String generateFoundationCopyName(final String fileName) {
        final int lastDotIndex = fileName.lastIndexOf(".");
        if (lastDotIndex != -1) {
            return fileName.substring(0, lastDotIndex) + "_copy."
                   + (lastDotIndex != fileName.length() - 1
                      ? fileName.substring(lastDotIndex + 1)
                      : "");
        }
        else {
            return fileName + "_copy";
        }
    }

    public void finishedSample(final BaseDataSource dataSource, final PollingPackage aPackage) throws FrameworkException {
        logger.debug(getLogPrefix() + " - process finishedSample()");
        if (aPackage.getPackageName().equals(NetAppFoundationPackage.PACKAGE)) {
            recordingHandler.finishFoundationSample();
        }
        else if (aPackage.getPackageName().equals(getPackageName())) {
            recordingHandler.stopRecordingSessionIfNeeded();
        }

    }

    public void sampleSuccessfull(final BaseDataSource dataSource, final PollingPackage aPackage) throws FrameworkException {
        if (aPackage.getPackageName().equals(NetAppFoundationPackage.PACKAGE)) {
            recordingHandler.setFoundationSuccess();
        }
    }

    public void sampleFailed(final BaseDataSource dataSource, final PollingPackage aPackage, final DataSourceErrorException e) {
        if (aPackage.getPackageName().equals(NetAppFoundationPackage.PACKAGE)) {
            //handle the case where we get a MOID aggregated error as well.
            if (e instanceof AggregatedDataSourceErrorsException) {
                final AggregatedDataSourceErrorsException adee = (AggregatedDataSourceErrorsException) e;
                if (adee.getType() == DataSourceErrorType.PARTIAL_SUCCESS) {
                    recordingHandler.setFoundationSuccess();
                    return;
                }
            }
            recordingHandler.setFoundationFailed();
        }
    }

    public void sampleFailed(final BaseDataSource dataSource, final PollingPackage aPackage, final Throwable e) {
        if (aPackage.getPackageName().equals(NetAppFoundationPackage.PACKAGE)) {
            recordingHandler.setFoundationFailed();
        }
    }

    public void recordingSessionDone(final BaseDataSource dataSource,
                                     final PollingPackage aPackage,
                                     final File recordingSession,
                                     final boolean shouldReport) throws FrameworkException {

        if (aPackage.getPackageName().equals(NetAppFoundationPackage.PACKAGE)) {
            synchronized (this) {
                if (logger.isDebugEnabled()) {
                    logger.debug(getLogPrefix() + " - update from foundation - recording made: " + recordingSession);
                }
                lastRecordedFoundationZipFile = recordingSession;
            }
        }

    }

    public void requestRecording(final int recordingSamples) {
        logger.debug(getLogPrefix() + " - process requestRecording()");
        recordingHandler.requestRecording(recordingSamples);
        if (!isSimulating()) {
            new Thread() {
                public void run() {
                    ServiceManager.getInstance().getDataSourceManager().pollNow(getDataSource().getDataSourceName(),
                                                                                NetAppFoundationPackage.PACKAGE);
                }
            }.start();
        }
    }

    public int getRecordingSamplesCollected() {
        return getRecordingSession().getSamples().size();
    }

    public int getRecordingSamplesRequested() {
        return recordingHandler.getNumberOfRequestedRecordings();
    }

    public String getRecordingStatus() {
        return recordingHandler.getRecordingStatus();
    }

    /**
     * Called at the end of the foundation Poll.
     * Should reset collected foundation data (and keep it in recording if needed)
     * Also should let actual dependent package to use foundation data for its own purpose
     * (like get a list of reported volumes)
     * @param lastUpdatesFromFoundation Foundation Data
     */
    private void synchronizeWithFoundation(NetAppFoundationData lastUpdatesFromFoundation) {
        // Notify actual dependent package (and record data)
        if (lastUpdatesFromFoundation!=null) {
            logger.debug(getLogPrefix() + " - call doUpdateFromFoundation for key " + lastUpdatesFromFoundation.getOriginatorKey());
            doUpdateFromFoundation(lastUpdatesFromFoundation.getOriginatorKey(), lastUpdatesFromFoundation.getOdataRoot(), lastUpdatesFromFoundation.isChanged(), lastUpdatesFromFoundation.getContext());
        }
        else {
            logger.info(getLogPrefix() + " - synchoronizeWithFoundation lastUpdatesFromFoundation is null. ");
        }
    }

  @Override
  public void pollNow() throws DataSourceErrorException, FrameworkException
  {
    pollNow( null );
  }

    final public void pollNow(Properties properties) throws DataSourceErrorException, FrameworkException {
        // If we are waiting for a complete foundation sample to start recording - we will skip this poll.
        if (recordingHandler.isWaitingForFoundationInitialization()) {
            final String msg = "Reporting from this package will be deferred until Service Insight runs.";
            throw new DataSourceErrorException(DataSourceErrorType.PACKAGE_DEPENDENCY, msg, msg);
        }

        // Call to actual dependent package implementation.
        doFoundationDependantPollNow();
    }

    protected final boolean isWaitingForFoundationInitialization() {
        return recordingHandler.isWaitingForFoundationInitialization();
    }


    //Actual Foundation Dependant package poll

    abstract protected void doFoundationDependantPollNow() throws DataSourceErrorException, FrameworkException;

    //This method lets actual foundation dependant package to go over foundation data in its own thread.

    protected abstract void doUpdateFromFoundation(final String originatorKey, final Root odataRoot, final boolean changed, final Serializable context);
}
