package com.onaro.sanscreen.acquisition.netappfoundation.standalone;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.BaseDataSource;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.config.ConfigurationFactory;
import com.onaro.sanscreen.acquisition.framework.datasource.config.DataSourceConfiguration;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.NetAppFoundationPackage;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.config.NetAppFoundationConfiguration;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.SimpleLayout;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.util.Properties;

/**
 * Standalone class to run foundation polls
 */
public class NetAppFoundationStandalone {
    private final static Logger logger = Logger.getLogger(NetAppFoundationStandalone.class);
    private final static ConsoleAppender consoleAppender = new ConsoleAppender(new SimpleLayout());

    protected final DataSourceConfiguration configuration;
    protected final BaseDataSource dataSource;
    protected final NetAppFoundationStandaloneListener foundationListener;
    protected final NetAppFoundationPackage foundationPackage;

    static
    {
        // Add a console appender for the log of this data source
        consoleAppender.setTarget("System.out");
        Logger.getRootLogger().addAppender(consoleAppender);
    }

    public NetAppFoundationStandalone(final File deployDescriptor, final Properties properties) throws IOException, FrameworkException {
        // Create the data source!
        final ConfigurationFactory factory = new ConfigurationFactory("ocie-standalone", 1L);

        // For standalone, we don't want performance to default to TRUE.
        // If the user did not specify, set performance enabled to false
        if (!properties.containsKey("netappstorageperformance.enabled")) {
            properties.put("netappstorageperformance.enabled", "false");
        }
        // Also set the default performance poll period to 10s (between polls)
        // for standalone mode.
        if (!properties.containsKey("netappstorageperformance.poll")) {
            properties.put("netappstorageperformance.poll", "10");
        }
        // Finally, set default set of perf samples to 3
        if (!properties.containsKey("netappstorageperformance.recording.samples")) {
            properties.put("netappstorageperformance.recording.samples", "3");
        }

        configuration = factory.createForStandalone(properties, deployDescriptor);
        dataSource = new BaseDataSource(configuration);

        // Create the foundation package
        final NetAppFoundationConfiguration foundationConfiguration = configuration.getPackageConfiguration(NetAppFoundationPackage.PACKAGE);
        foundationListener = new NetAppFoundationStandaloneListener();

        foundationPackage = foundationConfiguration.instantiatePackage(dataSource, foundationListener, foundationListener, foundationListener, foundationListener, foundationListener);
        dataSource.registerPackage(foundationPackage);
    }

    protected void start() throws FrameworkException {
        dataSource.start();
    }

    protected void stop() {
        dataSource.stop();
    }

    /**
     * Creates an instance of the target datasource, then run methods: start, run and stop
     *
     * @throws FrameworkException On any framework problem
     * @throws DataSourceErrorException On poll failure
     */
    public void run() throws FrameworkException, DataSourceErrorException {
        logger.info("Acquiring Configuration...");
        dataSource.run(NetAppFoundationPackage.PACKAGE);

        if (!foundationListener.isSuccess())
            throw new RuntimeException("Failed during initial configuration poll");
    }

    /**
     * Main class, run it as stand-alone application, usually for recording purposes.
     *
     * @param args Arguments
     */
    public static void main(final String[] args) {
        setup(NetAppFoundationStandalone.class, args);
    }

    /**
     * This is what main does - but if also has a parameter for which class should be instantiated when running the
     * standalone.
     *
     * This is used to allow reusing of this code - while still supporting different implementations.
     *
     * @param clazz Standalone class to instantiate
     * @param args Arguments
     */
    protected static void setup(final Class clazz, final String[] args) {
        if (args.length < 3) {
            System.out.println("Usage: standalone -netappfoundation.ip=<ip> -netappfoundation.user=<user> -netappfoundation.password<password>");
            return;
        }

        try {
            setLogLevel(Level.ERROR);

            ServiceManager.getInstance().startForStandalone();
            ServiceManager.getInstance().start();

            // final NetAppFoundationStandalone standalone = new NetAppFoundationStandalone();

            try {

                final File deployDescriptor = new File("netapp_ocie.xml");
                if (!deployDescriptor.exists())
                    throw new FileNotFoundException("Could not find netapp_ocie.xml (expected to be located at current directory");

                final Properties properties = createPropertiesFromCommandLineArgs(args);

                final String logLevelRaw = properties.getProperty("loglevel");
                final Level logLevel = Level.toLevel(logLevelRaw, Level.INFO);

                // Construct the standalone instance using reflection to allow other implementations
                final Class[] constructorSignature = {java.io.File.class, java.util.Properties.class};
                final Constructor constructor = clazz.getConstructor(constructorSignature);
                final Object[] constructorValues = {deployDescriptor, properties};
                final NetAppFoundationStandalone standalone = (NetAppFoundationStandalone)constructor.newInstance(constructorValues);

                standalone.start();

                setLogLevel(logLevel);
                standalone.run();

                setLogLevel(Level.ERROR);
                standalone.stop();
            }
            finally {
                ServiceManager.getInstance().stop();
            }
        }
        catch (DataSourceErrorException e) {
            logger.error("Data source failed: " + e.getMessage(), e);
        }
        catch (Exception e) {
            logger.error("Internal error: " + e.getMessage(), e);
        }
    }

    /**
     * Parses the command line to set the properties for this data source
     *
     * <p/>Assuming user's imput follows the format: -&lt;property-name&gt;=&lt;value&gt;
     *
     * <p/>example: standalone -ip=10.5.6.76 -name="ds-8000" -path="c:\program files\cli" ...
     *
     * @param args User's input command line.
     *
     * @return Properties that has been set with their values.
     */
    protected static Properties createPropertiesFromCommandLineArgs(final String[] args) throws FrameworkException {
        final Properties properties = new Properties();


        // Loops though all the User's input arguments.
        for (final String token : args) {
            // Option HELP is a special case, because it doesn't have a value.
            if (token.contains("-help")) {
                properties.setProperty("help", Boolean.TRUE.toString());
                continue;
            }

            // Finds the dash sign...
            final int dash = token.indexOf("-");
            if (dash == -1)
                throw new FrameworkException("Option entered: [" + token + "] do not have prefix '-' (dash) char.");

            // Finds the equal sign...
            final int equal = token.indexOf("=");
            if (equal == -1)
                throw new FrameworkException("Equal sign not found in option entered: [" + token + "]");

            // Gets name and value.
            final String optionName = token.substring(dash + 1, equal);
            final String optionValue = token.substring(equal + 1);

            properties.setProperty(optionName, optionValue);
        }

        return properties;
    }

    /**
     * Sets the log level of what's being written to the console
     *
     * @param logLevel New log level
     */
    protected static void setLogLevel(final Level logLevel) {
        consoleAppender.setThreshold(logLevel);
        Logger.getRootLogger().setLevel(logLevel);
    }
}
