package com.onaro.sanscreen.acquisition.netappschema.netappdata;

import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileCRCWriter;
import com.onaro.sanscreen.acquisition.netappschema.annotations.*;
import com.onaro.sanscreen.acquisition.schema.BaseCRCXMLNode;
import com.onaro.sanscreen.acquisition.schema.KeyBase;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;

import java.io.IOException;
import java.util.EnumMap;

@OCIETable("cifs_share")
@OCIEResourceType("cifs_share")
public class CIFSShare extends BaseCRCXMLNode<VServer> {

    public final static class CIFSShareKey extends KeyBase<CIFSShareKey.Field> {
        public static enum Field {
            @OCIEZapiSource(
                sourceZapi = "cluster-identity-get",
                zapiResponsePath = "netapp.results.attributes.cluster-identity-info.cluster-uuid"
            )
            cluster_uuid,

            @OCIEZapiSource(
                sourceZapi = "vserver-get-iter",
                zapiResponsePath = "netapp.results.attributes-list.vserver-info.uuid"
            )
            vserver_uuid,

            @OCIEZapiSource(
                sourceZapi = "cifs-share-get-iter",
                zapiResponsePath = "netapp.results.attributes-list.cifs-share.share-name"
            )
            share_name
        }

        public CIFSShareKey(final String clusterUUID, final String vServerUUID, final String shareName) {
            super("cifs_share",
                  new EnumMap<Field, String>(Field.class) {{
                        put(Field.cluster_uuid, clusterUUID);
                        put(Field.vserver_uuid, vServerUUID);
                        put(Field.share_name, shareName);
                  }});
        }

        @Override
        @OCIEUUID(uuidDescr = "vserver_uuid.share_name")
        public String getOcieUUID() {
            final String clusterUUID = getID(Field.cluster_uuid);
            final String vServerUUID = getID(Field.vserver_uuid);
            final String fqVServerUUID = VServer.VServerKey.getOCIEVServerUUID(clusterUUID, vServerUUID);
            final String shareName = getID(Field.share_name);
            return fqVServerUUID + '.' + shareName;
        }
    }

    private final CIFSShareKey key;

    @OCIEZapiSource(
        sourceZapi = "cifs-share-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.cifs-share.path"
    )
    @OCIEValue(
        ocieColumn = "path"
    )
    private String path;

    @OCIEZapiSource(
        sourceZapi = "cifs-share-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.cifs-share.comment"
    )
    @OCIEValue(
        ocieColumn = "comment"
    )
    private String comment;

    @OCIEZapiSource(
        sourceZapi = "cifs-share-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.cifs-share.path"
    )
    @OCIEValue(
        ocieColumn = "volumeId",
        notes = "derived from path"
    )
    private Volume.VolumeKey volumeKey;

    @OCIEZapiSource(
        sourceZapi = "cifs-share-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.cifs-share.path"
    )
    @OCIEValue(
        ocieColumn = "qtreeId",
        notes = "derived from path"
    )
    private QTree.QTreeKey qTreeKey;

    @OCIEZapiSource(
        sourceZapi = "cifs-share-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.cifs-share.share-properties.cifs-share-properties"
    )
    @OCIEValue(
        ocieColumn = "shareProperties",
        notes = "CSV list"
    )
    private String shareProperties;

    private final CIFSShareACLs cifsShareACLs;

    CIFSShare(final VServer parentVServer, final String shareName) {
        super(parentVServer, "CIFSShare");
        final VServer.VServerKey parentVServerKey = parentVServer.getKey();
        final String clusterUUID = parentVServerKey.getID(VServer.VServerKey.Field.cluster_uuid);
        final String vServerUUID = parentVServerKey.getID(VServer.VServerKey.Field.uuid);
        this.cifsShareACLs = new CIFSShareACLs(this);
        this.key = new CIFSShareKey(clusterUUID, vServerUUID, shareName);
    }

    @Override
    @SuppressWarnings("unchecked")
    public CIFSShareKey getKey() {
        return key;
    }

    public String getName() {
        return key.getID(CIFSShareKey.Field.share_name);
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getPath() {
        return path;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getComment() {
        return comment;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public Volume.VolumeKey getVolumeKey() {
        return volumeKey;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public QTree.QTreeKey getQtreeKey() {
        return qTreeKey;
    }

    public void setPath(final String path) {
        this.path = path;
    }

    public void setComment(final String comment) {
        this.comment = comment;
    }

    public void setVolumeKey(final Volume.VolumeKey volumeKey) {
        this.volumeKey = volumeKey;
    }

    public void setQTreeKey(final QTree.QTreeKey qTreeKey) {
        this.qTreeKey = qTreeKey;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getShareProperties() {
        return shareProperties;
    }

    public void setShareProperties(final String shareProperties) {
        this.shareProperties = shareProperties;
    }

    public CIFSShareACLs getCIFSShareACLs() {
        return cifsShareACLs;
    }

    public CIFSShareACL createCIFSShareACL(final String userOrGroup) {
        return new CIFSShareACL(this, userOrGroup);
    }

    public static enum ShareProperty {
        NOT_MAPPED,
        OPLOCKS,
        BROWSABLE,
        SHOWSNAPSHOT,
        CHANGENOTIFY,
        HOMEDIRECTORY,
        ATTRIBUTECACHE,
        BRANCHCACHE,
        CONTINUOUSLY_AVAILABLE,
        SHADOWCOPY,
        ACCESS_BASED_ENUMERATION,
        NAMESPACE_CACHING,
        ENCRYPT_DATA
    }



    protected void marshallAttributesToFileAndCRC(final MarshallingCache cache, final DSFileCRCWriter writer) throws IOException, SchemaException {
        super.marshallAttributesToFileAndCRC(cache, writer);
        addRequiredAttributeAndCRC(writer, "ResourceKey", key);
        addRequiredAttributeAndCRC(writer, "UUID", key.getOcieUUID());
        addRequiredAttributeAndCRC(writer, "VolumeUUID", Volume.VolumeKey.getVolumeInstanceUUIDOrNull(volumeKey));
        addRequiredAttributeAndCRC(writer, "QtreeUUID", KeyBase.getOcieUUIDOrNull(qTreeKey));
        addRequiredAttributeAndCRC(writer, "Name", key.getID(CIFSShareKey.Field.share_name));
        addRequiredAttributeAndCRC(writer, "Path", path);
        addRequiredAttributeAndCRC(writer, "Comment", comment);
        addRequiredAttributeAndCRC(writer, "ShareProperties", shareProperties);
    }


    @Override
    protected boolean hasChildren() {
        return !cifsShareACLs.isEmpty();

    }

    @SuppressWarnings("unchecked")
    @Override
    protected void marshallChildrenToFileAndCRC(final MarshallingCache cache, final DSFileCRCWriter writer, final int tabs) throws IOException, SchemaException {
        addAllChildrenUnderNodeAndCRC(cache, writer, "CIFSShareACLs", cifsShareACLs.values(), tabs);
    }
}
