package com.onaro.sanscreen.acquisition.netappschema.netappdata;

import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileCRCWriter;
import com.onaro.sanscreen.acquisition.netappschema.annotations.*;
import com.onaro.sanscreen.acquisition.schema.BaseCRCXMLNode;
import com.onaro.sanscreen.acquisition.schema.KeyBase;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;

import java.io.IOException;
import java.util.EnumMap;

@OCIETable("iscsi_session")
@OCIEResourceType("iscsi_session")
public class ISCSISession extends BaseCRCXMLNode<ISCSIPortalGroup> {

    public static final class ISCSISessionKey extends KeyBase<ISCSISessionKey.Field> {

        public static enum Field {
            @OCIEZapiSource(
                sourceZapi = "cluster-identity-get",
                zapiResponsePath = "netapp.results.attributes.cluster-identity-info.cluster-uuid"
            )
            cluster_uuid,

            @OCIEZapiSource(
                sourceZapi = "vserver-get-iter",
                zapiResponsePath = "netapp.results.attributes-list.vserver-info.uuid"
            )
            vserver_uuid,

            @OCIEZapiSource(
                sourceZapi = "iscsi-tpgroup-get-iter",
                zapiResponsePath = "netapp.results.attributes-list.iscsi-tpgroup-list-entry-info.tpgroup-name"
            )
            portal_group_name,

            @OCIEZapiSource(
                sourceZapi = "iscsi-session-get-iter",
                zapiResponsePath = "netapp.results.attributes-list.iscsi-session-list-entry-info.target-session-id"
            )
            target_session_id
        }

        public ISCSISessionKey(final String clusterUUID, final String vServerUUID, final String portalGroupName, final String targetSessionID) {
            super("iscsi_session", new EnumMap<Field, String>(Field.class) {{
                put(Field.cluster_uuid, clusterUUID);
                put(Field.vserver_uuid, vServerUUID);
                put(Field.portal_group_name, portalGroupName);
                put(Field.target_session_id, targetSessionID);
            }});
        }

        @Override
        @OCIEUUID(uuidDescr = "cluster_uuid.vserver_uuid.portal_group_name.target_session_id")
        public String getOcieUUID() {
            return getFQID();     // public UUID requires cluster UUID
        }
    }

    private final ISCSISessionKey key;

    @OCIEZapiSource(
        sourceZapi = "iscsi-session-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.iscsi-session-list-entry-info.target-session-id"
    )
    @OCIEValue(
        ocieColumn = "targetSessionID"
    )
    private Integer targetSessionID;

    @OCIEZapiSource(
        sourceZapi = "iscsi-session-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.iscsi-session-list-entry-info.initiator-nodename"
    )
    @OCIEValue(
        ocieColumn = "initiatorNodeName"
    )
    private String initiatorNodeName;

    @OCIEZapiSource(
        sourceZapi = "iscsi-initiator-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.iscsi-initiator-list-entry-info.initiator-aliasname"
    )
    @OCIEValue(
        ocieColumn = "initiatorAliasName"
    )
    private String initiatorAliasName;

    @OCIEZapiSource(
        sourceZapi = "iscsi-session-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.iscsi-session-list-entry-info.max-connections"
    )
    @OCIEValue(
        ocieColumn = "maxConnections"
    )
    private Integer maxConnections;

    @OCIEZapiSource(
        sourceZapi = "iscsi-session-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.iscsi-session-list-entry-info.isid"
    )
    @OCIEValue(
        ocieColumn = "initiatorSessionId"
    )
    private String initiatorSessionID;

    private final ISCSISessionConnections iscsiSessionConnections;

    ISCSISession(final ISCSIPortalGroup parent, final String clusterUUID, final VServer.VServerKey vServerKey, final String portalGroupName, final Integer targetSessionID) {
        super(parent, ISCSISession.class.getSimpleName());
        final String shortVServerUUID = vServerKey.getID(VServer.VServerKey.Field.uuid);
        this.key = new ISCSISessionKey(clusterUUID, shortVServerUUID, portalGroupName, String.valueOf(targetSessionID));
        this.iscsiSessionConnections = new ISCSISessionConnections(this);
    }

    @Override
    @SuppressWarnings({"unchecked"})
    public ISCSISessionKey getKey() {
        return key;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public Integer getTargetSessionID() {
        return targetSessionID;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getInitiatorNodeName() {
        return initiatorNodeName;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getInitiatorAliasName() {
        return initiatorAliasName;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public Integer getMaxConnections() {
        return maxConnections;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getInitiatorSessionID() {
        return initiatorSessionID;
    }

    public void setTargetSessionID(final Integer targetSessionID) {
        this.targetSessionID = targetSessionID;
    }

    public void setInitiatorNodeName(final String initiatorNodeName) {
        this.initiatorNodeName = initiatorNodeName;
    }

    public void setInitiatorAliasName(final String initiatorAliasName) {
        this.initiatorAliasName = initiatorAliasName;
    }

    public void setMaxConnections(final Integer maxConnections) {
        this.maxConnections = maxConnections;
    }

    public void setInitiatorSessionID(final String initiatorSessionID) {
        this.initiatorSessionID = initiatorSessionID;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public ISCSISessionConnections getIscsiSessionConnections() {
        return iscsiSessionConnections;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public ISCSISessionConnections getISCSISessionConnections() {
        return iscsiSessionConnections;
    }

    public ISCSISessionConnection createISCSISessionConnection(final Integer connectionID) {
        return new ISCSISessionConnection(this, connectionID);
    }

    @Override
    protected boolean hasChildren() {
        return !iscsiSessionConnections.isEmpty();
    }

    @Override
    protected void marshallChildrenToFileAndCRC(final MarshallingCache marshallingCache, final DSFileCRCWriter writer, final int tabs) throws IOException, SchemaException {
        super.marshallChildrenToFileAndCRC(marshallingCache, writer, tabs);

        addAllChildrenUnderNodeAndCRC(marshallingCache, writer, "ISCSISessionConnections", iscsiSessionConnections.values(), tabs);
    }

    @Override
    protected void marshallAttributesToFileAndCRC(final MarshallingCache marshallingCache, final DSFileCRCWriter writer) throws IOException, SchemaException {
        super.marshallAttributesToFileAndCRC(marshallingCache, writer);
        addRequiredAttributeAndCRC(writer, "ResourceKey", key);
        addRequiredAttributeAndCRC(writer, "UUID", key.getOcieUUID());
        addRequiredAttributeAndCRC(writer, "InitiatorSessionID", initiatorSessionID);
        addRequiredAttributeAndCRC(writer, "TargetSessionID", targetSessionID);
        addRequiredAttributeAndCRC(writer, "InitiatorNodeName", initiatorNodeName);
        addRequiredAttributeAndCRC(writer, "InitiatorAliasName", initiatorAliasName);
        addRequiredAttributeAndCRC(writer, "MaxConnections", maxConnections);
    }
}
