package com.onaro.sanscreen.acquisition.netappschema.netappdata;

import java.io.IOException;
import java.util.EnumMap;

import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileCRCWriter;
import com.onaro.sanscreen.acquisition.netappschema.annotations.OCIEResourceType;
import com.onaro.sanscreen.acquisition.netappschema.annotations.OCIETable;
import com.onaro.sanscreen.acquisition.netappschema.annotations.OCIEValue;
import com.onaro.sanscreen.acquisition.netappschema.annotations.OCIEZapiSource;
import com.onaro.sanscreen.acquisition.schema.*;

/**
 * Network ports that are assigned to a broadcast domain.  While broadcast
 * domains may be assigned multiple network ports, each network port is assigned
 * to only one broadcast domain.  Since a broadcast domain is owned by an
 * {@link NetworkIpSpace}, network ports are therefore implicitly owned by an IPSpace.
 */
@OCIETable("net_port_broadcast_domain")
@OCIEResourceType("net_port_broadcast_domain")
public class NetworkPortBroadcastDomain extends BaseCRCXMLNode<Cluster> {

    /**
     * The owning broadcast domain.
     *
     * @see <a href="https://wikid.netapp.com/w/SCON/Fullsteam/Simplicity/Simplicity_Networking_FS#Broadcast_Domain">Simplicity Networking FS/Broadcast Domain</a>
     * @see <a href="https://wikid.netapp.com/w/NGS_NPI/TOI/NSM/NSM_3.2/NSM32_Networking#Broadcast_Domain>NSM32 Networking/Broadcast_Domain</a>
     */
    @OCIEZapiSource(
            sourceZapi = "net-port-broadcast-domain-get-iter",
            zapiResponsePath = "netapp.results.attributes-list.net-port-broadcast-domain-info.broadcast-domain"
    )
    @OCIEValue(
            ocieColumn = "name"
    )
    private final String broadcastDomain;

    /**
     * Maximum Transfer Unit (MTU).
     */
    @OCIEZapiSource(
            sourceZapi = "net-port-broadcast-domain-get-iter",
            zapiResponsePath = "netapp.results.attributes-list.net-port-broadcast-domain-info.mtu"
    )
    @OCIEValue(
            ocieColumn = "mtu"
    )
    private Long mtu;

    /**
     * The owning {@link NetworkIpSpace}.
     */
    @OCIEZapiSource(
            sourceZapi = "net-port-broadcast-domain-get-iter",
            zapiResponsePath = "netapp.results.attributes-list.net-port-broadcast-domain-info.ipspace"
    )
    @OCIEValue(
            ocieColumn = "networkIpSpaceId"
    )
    private final NetworkIpSpace.NetworkIpSpaceKey networkIpSpaceKey;

    public static final class NetworkPortBroadcastDomainKey extends KeyBase<NetworkPortBroadcastDomainKey.Field> {

        /**
         * YYYYMMDD
         */
        private static final long serialVersionUID = 20140318L;

        public static enum Field {
            @OCIEZapiSource(
                    sourceZapi = "cluster-identity-get",
                    zapiResponsePath = "netapp.results.attributes.cluster-identity-info.cluster-uuid"
            )
            CLUSTER_UUID,

            @OCIEZapiSource(
                    sourceZapi = "net-port-broadcast-domain-get-iter",
                    zapiResponsePath = "netapp.results.attributes-list.net-port-broadcast-domain-info.broadcast-ipspace"
            )
            IP_SPACE_UUID,


            @OCIEZapiSource(
                    sourceZapi = "net-port-broadcast-domain-get-iter",
                    zapiResponsePath = "netapp.results.attributes-list.net-port-broadcast-domain-info.broadcast-domain"
            )
            BROADCAST_DOMAIN
        }

        @SuppressWarnings("serial")
        public NetworkPortBroadcastDomainKey(final String clusterUUID, final String ipSpaceUUID,
                                             final String broadcastDomainName) {
            super("network_port_broadcast_domain",
                    new EnumMap<Field, String>(Field.class) {
                        {
                            put(Field.CLUSTER_UUID, clusterUUID);
                            put(Field.IP_SPACE_UUID, ipSpaceUUID);
                            put(Field.BROADCAST_DOMAIN, broadcastDomainName);
                        }
                    });
        }

        @Override
        public String getOcieUUID() {
            return getFQID();
        }
    }

    private final NetworkPortBroadcastDomainKey broadcastDomainKey;

    public NetworkPortBroadcastDomain(final Cluster parent, final String broadcastDomainName,
                                      final NetworkIpSpace.NetworkIpSpaceKey networkIpSpaceKey) {
        super(parent, "NetworkPortBroadcastDomain");
        this.broadcastDomain = broadcastDomainName;
        this.networkIpSpaceKey = networkIpSpaceKey;
        this.broadcastDomainKey = new NetworkPortBroadcastDomainKey(parent.getClusterUUID(),
                networkIpSpaceKey.getOcieUUID(), broadcastDomainName);
    }

    @Override
    public NetworkPortBroadcastDomainKey getKey() {
        return broadcastDomainKey;
    }

    public String getBroadcastDomain() {
        return broadcastDomain;
    }

    public Long getMtu() {
        return mtu;
    }

    public NetworkPortBroadcastDomain setMtu(Long mtu) {
        this.mtu = mtu;
        return this;
    }

    public NetworkIpSpace.NetworkIpSpaceKey getNetworkIpSpaceKey() {
        return networkIpSpaceKey;
    }

    @Override
    protected void marshallAttributesToFileAndCRC(final MarshallingCache marshallingCache, final DSFileCRCWriter writer) throws IOException, SchemaException {
        super.marshallAttributesToFileAndCRC(marshallingCache, writer);
        addRequiredAttributeAndCRC(writer, "ResourceKey", broadcastDomainKey);
        addRequiredAttributeAndCRC(writer, "UUID", broadcastDomainKey.getOcieUUID());
        addRequiredAttributeAndCRC(writer, "Name", broadcastDomain);
        addRequiredAttributeAndCRC(writer, "MTU", mtu);
        addRequiredAttributeAndCRC(writer, "NetworkIpSpaceUUID", networkIpSpaceKey.getOcieUUID());
    }
}
