package com.onaro.sanscreen.acquisition.netappschema.netappdata;

import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileCRCWriter;
import com.onaro.sanscreen.acquisition.netappschema.annotations.*;
import com.onaro.sanscreen.acquisition.schema.BaseCRCXMLNode;
import com.onaro.sanscreen.acquisition.schema.KeyBase;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;

import java.io.IOException;
import java.util.EnumMap;

/**
 * This object represents a physical CPU inside of a Node
 */
@OCIETable("storage_service")
@OCIEResourceType("storage_service")
public class StorageService extends BaseCRCXMLNode<VServer> {

    public final static class StorageServiceKey extends KeyBase<StorageServiceKey.Field> {
        public static enum Field {
            @OCIEZapiSource(
                sourceZapi = "cluster-identity-get",
                zapiResponsePath = "netapp.results.attributes.cluster-identity-info.cluster-uuid"
            )
            cluster_uuid,

            @OCIEZapiSource(
                sourceZapi = "volume-get-iter",
                zapiResponsePath = "netapp.results.attributes-list.volume-attributes.volume-id-attributes.uuid"
            )
            volume_uuid,

            @OCIEZapiSource(
                sourceZapi = "volume-storage-service-get-iter",
                zapiResponsePath = "netapp.results.attributes-list.storage-service-info.storage-service-id"
            )
            uuid
        }

        public StorageServiceKey(final String clusterUUID, final String volumeUUID, final String uuid) {
            super("infinitevol_storage_service", new EnumMap<Field, String>(Field.class) {{
                put(Field.cluster_uuid, clusterUUID);
                put(Field.volume_uuid, volumeUUID);
                put(Field.uuid, uuid);
            }});
        }

        @Override
        @OCIEUUID(uuidDescr = "node_uuid.adapter_name")
        public String getOcieUUID() {
            return getID(Field.volume_uuid) + '.' + getID(Field.uuid);
        }
    }

    private final StorageServiceKey key;

    @OCIEZapiSource(
        sourceZapi = "volume-storage-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.storage-service-info.vserver"
    )
    @OCIEValue(
        ocieColumn = "vserverId"
    )
    private final VServer.VServerKey vServerKey;

    @OCIEZapiSource(
        sourceZapi = "volume-storage-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.storage-service-info.volume"
    )
    @OCIEValue(
        ocieColumn = "volumeId"
    )
    private final Volume.VolumeKey volumeKey;

    @OCIEZapiSource(
        sourceZapi = "volume-storage-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.storage-service-info.storage-service-id"
    )
    @OCIEValue(
        ocieColumn = "storageServiceIdentifier"
    )
    private final Long storageServiceID;

    @OCIEZapiSource(
        sourceZapi = "volume-storage-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.storage-service-info.storage-service-id"
    )
    @OCIEValue(
        ocieColumn = "storageServiceName"
    )
    private String storageServiceName;

    /*
    @OCIEZapiSource(
        sourceZapi = "volume-storage-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.storage-service-info.constituent-group-id"
    )
    private Long storageServiceConstituentGroupID;
    */


    @OCIEZapiSource(
        sourceZapi = "derived"
    )
    @OCIEValue(
        ocieColumn = "sizeUsed",
        notes = "Derived, sum of constituent volumes size used"
    )
    private long sizeUsed = 0L;

    @OCIEZapiSource(
        sourceZapi = "derived"
    )
    @OCIEValue(
        ocieColumn = "sizeTotal",
        notes = "Derived, sum of constituent volumes size total"
    )
    private long sizeTotal = 0L;

    @OCIEZapiSource(
        sourceZapi = "derived"
    )
    @OCIEValue(
        ocieColumn = "sizeUsedBySnapshots",
        notes = "Derived, sum of constituent volumes size used by snapshots"
    )
    private long sizeUsedBySnapshots = 0L;

    @OCIEZapiSource(
        sourceZapi = "derived"
    )
    @OCIEValue(
        ocieColumn = "snapshotReservedSize",
        notes = "Derived, sum of constituent volumes snapshot reserve size"
    )
    private long snapshotReservedSize = 0L;

    @OCIEZapiSource(
        sourceZapi = "derived"
    )
    @OCIEValue(
        ocieColumn = "deduplicationSpaceSaved",
        notes = "Derived, sum of constituent volumes deduplication space saved"
    )
    private long deduplicationSpaceSaved = 0L;

    @OCIEZapiSource(
        sourceZapi = "derived"
    )
    @OCIEValue(
        ocieColumn = "compressionSpaceSaved",
        notes = "Derived, sum of constituent volumes compression space saved"
    )
    private long compressionSpaceSaved = 0L;


    StorageService(final VServer parent, final Long serviceID, final VServer.VServerKey vServerKey, final Volume.VolumeKey volumeKey) {
        super(parent, "StorageService");
        this.key = new StorageServiceKey(parent.getKey().getID(VServer.VServerKey.Field.cluster_uuid),
                                         volumeKey.getVolumeInstanceUUID(),
                                         String.valueOf(serviceID));
        this.vServerKey = vServerKey;
        this.volumeKey = volumeKey;
        this.storageServiceID = serviceID;
    }

    @Override
    @SuppressWarnings({"unchecked"})
    public StorageServiceKey getKey() {
        return key;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getStorageServiceName() {
        return storageServiceName;
    }

    public void setStorageServiceName(final String storageServiceName) {
        this.storageServiceName = storageServiceName;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public Long getStorageServiceID() {
        return storageServiceID;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public VServer.VServerKey getVServerKey() {
        return vServerKey;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public Volume.VolumeKey getVolumeKey() {
        return volumeKey;
    }

    public void addToSizeUsed(final long constituentSizeUsed) {
        this.sizeUsed += constituentSizeUsed;
    }

    public void addToSizeTotal(final long constituentSizeTotal) {
        this.sizeTotal += constituentSizeTotal;
    }

    public void addToSizeUsedBySnapshots(final long constituentSizeUsedBySnapshots) {
        this.sizeUsedBySnapshots += constituentSizeUsedBySnapshots;
    }

    public void addToSnapshotReservedSize(final long constituentSnapshotReservedSize) {
        this.snapshotReservedSize += constituentSnapshotReservedSize;
    }

    public void addToDeduplicationSpaceSaved(final long constituentDeduplicationSpaceSaved) {
        this.deduplicationSpaceSaved += constituentDeduplicationSpaceSaved;
    }

    public void addToCompressionSpaceSaved(final long constituentCompressionSpaceSaved) {
        this.compressionSpaceSaved += constituentCompressionSpaceSaved;
    }



    @Override
    protected void marshallAttributesToFileAndCRC(final MarshallingCache marshallingCache, final DSFileCRCWriter writer) throws IOException, SchemaException {
        super.marshallAttributesToFileAndCRC(marshallingCache, writer);
        addRequiredAttributeAndCRC(writer, "ResourceKey", key);
        addRequiredAttributeAndCRC(writer, "UUID", key.getOcieUUID());
        addRequiredAttributeAndCRC(writer, "StorageServiceName", storageServiceName);
        addRequiredAttributeAndCRC(writer, "StorageServiceIdentifier", storageServiceID);
        addRequiredAttributeAndCRC(writer, "SizeUsed", sizeUsed);
        addRequiredAttributeAndCRC(writer, "SizeTotal", sizeTotal);
        addRequiredAttributeAndCRC(writer, "SizeUsedBySnapshots", sizeUsedBySnapshots);
        addRequiredAttributeAndCRC(writer, "SnapshotReserveSize", snapshotReservedSize);
        addRequiredAttributeAndCRC(writer, "DeduplicationSpaceSaved", deduplicationSpaceSaved);
        addRequiredAttributeAndCRC(writer, "CompressionSpaceSaved", compressionSpaceSaved);
    }
}
