package com.onaro.sanscreen.acquisition.netappstorageperformance.datasource;

import java.util.ArrayList;
import java.util.List;

/**
 * This Class Aggregates  performance volume counters  for specific  Interval
 */

public class NetAppVolumePerfCounters {
    // holds current timestamp of the sample or interval between samples.
    protected long interval;
    protected long numRead,
            numWrite,
            numReadByte,
            numWriteByte,
            readResponseTime,
            writeResponseTime,
            totalResponseTime;
    double readCacheHitRatio,
            writeCacheHitRatio,
            totalCacheHitRatio;
    long writePending;


    /**
     * @param interval
     * @param numRead
     * @param numWrite
     * @param numReadByte
     * @param numWriteByte
     * @param readResponseTime
     * @param writeResponseTime
     * @param totalResponseTime
     * @param readCacheHitRatio
     * @param writeCacheHitRatio
     * @param totalCacheHitRatio
     * @param writePending
     */

    public NetAppVolumePerfCounters(final long interval,
                              final long numRead,
                              final long numWrite,
                              final long numReadByte,
                              final long numWriteByte,
                              final long readResponseTime,
                              final long writeResponseTime,
                              final long totalResponseTime,
                              final double readCacheHitRatio,
                              final double writeCacheHitRatio,
                              final double totalCacheHitRatio,
                              final long writePending) {


        this.interval = interval;
        this.numRead = numRead;
        this.numWrite = numWrite;
        this.numReadByte = numReadByte;
        this.numWriteByte = numWriteByte;
        this.readResponseTime = readResponseTime;
        this.writeResponseTime = writeResponseTime;
        this.totalResponseTime = totalResponseTime;
        this.readCacheHitRatio = readCacheHitRatio;
        this.writeCacheHitRatio = writeCacheHitRatio;
        this.totalCacheHitRatio = totalCacheHitRatio;
        this.writePending = writePending;
    }

    /**
     * Get a new counter object containing diff of passed value (newer) with the current values (older).    Algorithm
     * accounts for wraps (negative differences) by setting that counter to -1 to flag a skipped poll value.
     *
     * @param newer the new counter set.
     * @return newly instantiated counter set with difference values.
     */
    public NetAppVolumePerfCounters diffFromNewer(final NetAppVolumePerfCounters newer) {

        //pattern below accounts for wrapping.   If difference is negative, -1 is flagged to indicate we won't know
        //the value on this poll.
        final long intervalDiff = newer.interval - interval; //should never be -1
        final long numReadDiff = subtractLong(numRead, newer.numRead);
        final long numWriteDiff = subtractLong(numWrite, newer.numWrite);
        final long numReadByteDiff = subtractLong(numReadByte, newer.numReadByte);
        final long numWriteByteDiff = subtractLong(numWriteByte, newer.numWriteByte);
        final long readResponseTimeDiff = subtractLong(readResponseTime, newer.readResponseTime);
        final long writeResponseTimeDiff = subtractLong(writeResponseTime, newer.writeResponseTime);
        final long totalResponseTimeDiff = subtractLong(totalResponseTime, newer.totalResponseTime);
        final double readCacheHitRatioDiff = subtractDouble(readCacheHitRatio, newer.readCacheHitRatio);
        final double writeCacheHitRatioDiff = subtractDouble(writeCacheHitRatio, newer.writeCacheHitRatio);
        final double totalCacheHitRatioDiff = subtractDouble(totalCacheHitRatio, newer.totalCacheHitRatio);
        final long writePendingDiff = subtractLong(writePending, newer.writePending);


        return new NetAppVolumePerfCounters(
                intervalDiff,
                numReadDiff, numWriteDiff,
                numReadByteDiff, numWriteByteDiff,
                readResponseTimeDiff, writeResponseTimeDiff, totalResponseTimeDiff,
                readCacheHitRatioDiff, writeCacheHitRatioDiff, totalCacheHitRatioDiff,
                writePendingDiff);
    }

    /**
     * Return value of c2 - c1.   Special logic treats -1 values (unsupported).  If either value is -1, -1 is
     * returned.
     *
     * <p>If the calculated subtraction results in a negative value, -1 is returned.   This accounts for wrapped
     * counters (c2 being the newer counter).
     *
     * @param older 1st value (older)
     * @param newer 2nd value (newer)
     * @return  c2 - c1. -1 in some cases as described above.
     */
    private long subtractLong(final long older, final long newer) {
        if (older == -1 || newer== -1) {
            return -1;
        }
        else {
            final long diff = newer - older;
            return diff < 0 ? -1 : diff;
        }
    }

    /**
     * See {@link #subtractLong(long, long)} , only with double type.
     *
     * @param older 1st value (older)
     * @param newer 2nd value (newer)
     * @return  c2 - c1. -1 in some cases as described above.
     */
    private double subtractDouble(final double older, final double newer) {
        if (older == -1 || newer == -1) {
            return -1;
        }
        else {
            final double diff = newer - older;
            return diff < 0 ? -1 : diff;
        }
    }


    /**
     * @param volumeMembersPref
     * @return
     */
    static public NetAppVolumePerfCounters aggregateVolumesPrefCounters(final List<NetAppVolumePerfCounters> volumeMembersPref) {
        long totalReadHitIO = 0;
        long totalWriteHitIO = 0;
        long interval = 0;
        long readIO = 0;
        long writeIO = 0;
        long numReadByte = 0;
        long numWriteByte = 0;
        long readResponseTime = 0;
        long writeResponseTime = 0;
        long totalResponseTime = 0;
        long writePending = 0;

        for (final NetAppVolumePerfCounters volumeMemberPerfCounter : volumeMembersPref) {
            interval = volumeMemberPerfCounter.getInterval();
            readIO += volumeMemberPerfCounter.getNumRead();
            writeIO += volumeMemberPerfCounter.getNumWrite();
            numReadByte += volumeMemberPerfCounter.getNumReadByte();
            numWriteByte += volumeMemberPerfCounter.getNumWriteByte();
            readResponseTime = Math.max(readResponseTime, volumeMemberPerfCounter.getReadResponseTime());
            writeResponseTime = Math.max(writeResponseTime, volumeMemberPerfCounter.getWriteResponseTime());
            totalResponseTime = Math.max(totalResponseTime, volumeMemberPerfCounter.getTotalResponseTime());
            writePending += volumeMemberPerfCounter.getWritePending();

            totalReadHitIO += volumeMemberPerfCounter.getReadCacheHitRatio() * volumeMemberPerfCounter.getNumRead();
            totalWriteHitIO += volumeMemberPerfCounter.getWriteCacheHitRatio() * volumeMemberPerfCounter.getNumWrite();
        }

        final long readCachHit = readIO > 0 ? totalReadHitIO / readIO : 0;
        final long writeCacheHit = writeIO > 0 ? totalWriteHitIO / writeIO : 0;
        final long totalCacheHit = writeIO + readIO > 0 ? (totalWriteHitIO + totalReadHitIO) / (writeIO + readIO) : 0;

        return new NetAppVolumePerfCounters(interval, readIO, writeIO, numReadByte, numWriteByte, readResponseTime, writeResponseTime, totalResponseTime, readCachHit, writeCacheHit, totalCacheHit, writePending);
    }

    // We should stop using DiskSpecification sor calculating response time, we should get it from the device

//    @Deprecated
//    public VolumeMeasurements convertToVolumeMeasurements(final Collection<NetAppDiskSpecification> disksSpecs) {
//
//        final long numIO = numRead + numWrite;
//        final long byteIO = numReadByte + numWriteByte;
//
//        final float readAccessTime = NetAppMetricCalculator.figureOutAccessTime(numRead, numReadByte, disksSpecs);
//        final float writeAccessTime = NetAppMetricCalculator.figureOutAccessTime(numWrite, numWriteByte, disksSpecs);
//        final float totalAccessTime = NetAppMetricCalculator.figureOutAccessTime(numIO, byteIO, disksSpecs);

// 17589 : We renamed Response Time to Access Time,
// then added a new metric Response Time.
// All existing array performance data sources need to report Access Time,
// not Response Time; Hence sending only AccessTime to server
// default value -1 for response time
// For NetApp Data Sources, we need to report Response Time too
// For Non NetApp, only report Access Time
//        final float readResponseTime = MetricCalculator.figureOutResponseTime(numRead, timestamp);
//        final float writeResponseTime = MetricCalculator.figureOutResponseTime(numWrite, timestamp);
//        final float totalResponseTime = MetricCalculator.figureOutResponseTime(numIO, timestamp);
//
//        final float readThroughput = NetAppMetricCalculator.getThoughputInMBPerSec(numReadByte, interval);
//        final float writeThroughput = NetAppMetricCalculator.getThoughputInMBPerSec(numWriteByte, interval);
//
//        final float readIOPS = NetAppMetricCalculator.getIOPS(numRead, interval);
//        final float writeIOPS = NetAppMetricCalculator.getIOPS(numWrite, interval);
//
//        readResponseTime = Math.max(readResponseTime, (long) readAccessTime);
//        writeResponseTime = Math.max(writeResponseTime, (long) writeAccessTime);
//        totalResponseTime = Math.max(totalResponseTime, (long) totalAccessTime);
//        final boolean accessed = numWrite + numRead > 0;
//        return new VolumeMeasurements((int) (readResponseTime < 0 ? -1 : readResponseTime),
//                (int) (writeResponseTime < 0 ? -1 : writeResponseTime),
//                (int) (totalResponseTime < 0 ? -1 : totalResponseTime),
//                (int) readThroughput,
//                (int) writeThroughput,
//                (int) readIOPS,
//                (int) writeIOPS,
//                readCacheHitRatio < 0 ? -1 : readCacheHitRatio,
//                writeCacheHitRatio < 0 ? -1 : writeCacheHitRatio,
//                totalCacheHitRatio < 0 ? -1 : totalCacheHitRatio,
//                -1,
//                (int) (writePending < 0 ? -1 : writePending),
//                accessed);
//    }


    public long getInterval() {
        return interval;
    }

    public void setInterval(final long interval) {
        this.interval = interval;
    }

    public long getNumRead() {
        return numRead;
    }

    public void setNumRead(final long numRead) {
        this.numRead = numRead;
    }

    public long getNumWrite() {
        return numWrite;
    }

    public void setNumWrite(final long numWrite) {
        this.numWrite = numWrite;
    }

    public long getNumReadByte() {
        return numReadByte;
    }

    public void setNumReadByte(final long numReadByte) {
        this.numReadByte = numReadByte;
    }

    public long getNumWriteByte() {
        return numWriteByte;
    }

    public void setNumWriteByte(final long numWriteByte) {
        this.numWriteByte = numWriteByte;
    }


    public long getReadResponseTime() {
        return readResponseTime;
    }

    public void setReadResponseTime(final long readResponseTime) {
        this.readResponseTime = readResponseTime;
    }

    public long getWriteResponseTime() {
        return writeResponseTime;
    }

    public void setWriteResponseTime(final long writeResponseTime) {
        this.writeResponseTime = writeResponseTime;
    }

    public long getTotalResponseTime() {
        return totalResponseTime;
    }

    public void setTotalResponseTime(final long totalResponseTime) {
        this.totalResponseTime = totalResponseTime;
    }

    public double getReadCacheHitRatio() {
        return readCacheHitRatio;
    }

    public void setReadCacheHitRatio(final double readCacheHitRatio) {
        this.readCacheHitRatio = readCacheHitRatio;
    }

    public double getWriteCacheHitRatio() {
        return writeCacheHitRatio;
    }

    public void setWriteCacheHitRatio(final double writeCacheHitRatio) {
        this.writeCacheHitRatio = writeCacheHitRatio;
    }

    public double getTotalCacheHitRatio() {
        return totalCacheHitRatio;
    }

    public void setTotalCacheHitRatio(final double totalCacheHitRatio) {
        this.totalCacheHitRatio = totalCacheHitRatio;
    }

    public long getWritePending() {
        return writePending;
    }

    public void setWritePending(final long writePending) {
        this.writePending = writePending;
    }

    /**
     * MetaVolumePerfInformation respnsible for aggregating volume members counters and diskspec
     * <p/>
     * calling getMetaVolumePerfCounters  returns MetaVolume Counters.
     */
    public static class MetaVolumePerfInformation {

        final private String metaVolumeName;
        final private List<NetAppVolumePerfCounters> membersCounters = new ArrayList<NetAppVolumePerfCounters>();
        final private List<NetAppDiskSpecification> membersDiskSpec = new ArrayList<NetAppDiskSpecification>();

        public MetaVolumePerfInformation(final String metaVolumeName) {
            if (metaVolumeName == null)
                throw new IllegalArgumentException("Meta Volume name can not be null");

            this.metaVolumeName = metaVolumeName;
        }

        public List<NetAppVolumePerfCounters> getMembersCounters() {
            return membersCounters;
        }

        public List<NetAppDiskSpecification> getMembersDiskSpec() {
            return membersDiskSpec;
        }

        public String getMetaVolumeName() {
            return metaVolumeName;
        }

        /**
         * @return VolumePerfCounters after aggregating all volume member counters
         */
        public NetAppVolumePerfCounters getMetaVolumePerfCounters() {
            return aggregateVolumesPrefCounters(membersCounters);
        }

    }

}
