package com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.config;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.BaseDataSource;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceSampleListener;
import com.onaro.sanscreen.acquisition.framework.datasource.config.DataSourceConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.metadata.PackageMetaData;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.NetAppClockSynchronizer;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.NetAppFoundationPackage;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.config.NetAppFoundationConfiguration;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.dependant.NetAppFoundationDependantPackage;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.dependant.config.NetAppFoundationDependantConfiguration;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppStoragePerformanceListener;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppStoragePerformanceUpdatesListener;
import com.onaro.util.DeviceFilter;
import org.apache.log4j.Logger;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Properties;

/**
 * Represents the configuration of a data source (types, properties, etc...)
 */
public abstract class NetAppStoragePerformanceConfiguration<PP extends NetAppFoundationDependantPackage> extends NetAppFoundationDependantConfiguration<PP> {
    private static final Logger logger = Logger.getLogger(NetAppStoragePerformanceConfiguration.class);

    private final boolean enabled;
    private final DeviceFilter filter;
    private final DeviceFilterType filterType;


    public static enum DeviceFilterType {
        INCLUDE("include"),
        EXCLUDE("exclude");

        String name;
        ArrayList nameList = new ArrayList();

        private DeviceFilterType(String name) {
            this.name = name;
            this.nameList.add(name);
        }

        public static DeviceFilterType typeFromName(String name) throws FrameworkException {
            if (name.equals("include")) {
                return INCLUDE;
            }
            else if (name.equals("exclude")) {
                return EXCLUDE;
            }
            else {
                throw new FrameworkException("unsupported Device Filter Type: " + name);
            }
        }
    }

    public NetAppStoragePerformanceConfiguration(final DataSourceConfiguration dataSourceConfiguration, final PackageMetaData metaData, final String dataSourceName, final long dataSourceID, final Properties props) throws FrameworkException {
        //Note package "name" here is hard-coded since "Feature" basis doesn't support OCIE 
        super(dataSourceConfiguration, metaData, dataSourceName, dataSourceID, "Performance", props, false);

        if (getNetAppFoundationConfiguration() == null)
            throw new FrameworkException("Foundation configuration is missing (it has to be defined before performance in the deploy descriptor file)");

        enabled = Boolean.parseBoolean(props.getProperty("enabled", Boolean.TRUE.toString()));
        props.remove("enabled");

        filter = new DeviceFilter(getProperty("storageperformance.filter", ""));
        filterType = DeviceFilterType.typeFromName(getProperty("storageperformance.filtertype", "exclude"));
    }

    public NetAppFoundationConfiguration getNetAppFoundationConfiguration() {
        return getDataSourceConfiguration().getPackageConfiguration(NetAppFoundationPackage.PACKAGE);
    }


    public boolean isEnabled() {
        return enabled;  //TODO - License check was removed for netapp performance....revisit before release.
    }


    public DeviceFilter getFilter() {
        return filter;
    }

    public DeviceFilterType getFilterType() {
        return filterType;
    }

    public boolean isCrawlSupported() {
        return false;
    }

    /**
     * The storage performance package saves both the foundation and the performance properties!
     */
    public Properties getPropertiesForRecording() {
        final Properties props = new Properties();
        props.putAll(getNetAppFoundationConfiguration().getPropertiesForRecording());
        props.putAll(super.getPropertiesForRecording());
        return props;
    }

    /**
     * Instantiate a new configuration object for this package based on the meta data described in this class.
     *
     * <p/>The configuration object was already created in the right class loader (see {@link PackageMetaData#instantiateConfiguration})
     * so we can simply instantiate a new Foundation package without having to worry about it.
     */
    public PP instantiatePackage(final BaseDataSource dataSource) throws FrameworkException {
        // Create a foundation listener for this package...
        final NetAppStoragePerformanceListener listener = new NetAppStoragePerformanceListener(this);

        return instantiatePackage(dataSource, listener, listener, listener);
    }

    /**
     * Package factory method used to instantiate a package with non default listener classes.
     */
    public PP instantiatePackage(final BaseDataSource dataSource, final DataSourceSampleListener dataSourceSampleListener, final NetAppStoragePerformanceUpdatesListener storagePerformanceListener, final NetAppClockSynchronizer clockSynchronizer) throws FrameworkException {
        try {
            final Class packageImplementationClass = getMetaData().getPackageImplementationClass();

            final Class[] constructorSignature = {BaseDataSource.class, this.getClass(), DataSourceSampleListener.class, NetAppStoragePerformanceUpdatesListener.class, NetAppClockSynchronizer.class};

            final Constructor constructor = packageImplementationClass.getConstructor(constructorSignature);

            // Construct the data source instance
            final Object[] constructorValues = {dataSource, this, dataSourceSampleListener, storagePerformanceListener, clockSynchronizer};

            //noinspection unchecked
            return (PP) constructor.newInstance(constructorValues);
        }
        catch (Throwable e) {
            throw new FrameworkException("Failed to instantiate storage performance package for data source " + getDataSourceName(), e);
        }
    }

}
