package com.onaro.sanscreen.acquisition.netappstorageperformance.standalone;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import com.onaro.sanscreen.acquisition.netappfoundation.standalone.NetAppFoundationStandalone;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppStoragePerformancePackage;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.config.NetAppStoragePerformanceConfiguration;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.util.Properties;

/**
 * Standalone class to run foundation polls
 */
public class NetAppStoragePerformanceStandalone extends NetAppFoundationStandalone {
    private final static Logger logger = Logger.getLogger(NetAppFoundationStandalone.class);
    protected final NetAppStoragePerformanceStandaloneListener storagePerformanceListener;
    protected final NetAppStoragePerformancePackage storagePerformancePackage;


    public NetAppStoragePerformanceStandalone(final File deployDescriptor, final Properties properties) throws IOException, FrameworkException {
        super(deployDescriptor, properties);

        // Create the storage performance package
        final NetAppStoragePerformanceConfiguration<NetAppStoragePerformancePackage> performanceConfiguration = configuration.getPackageConfiguration(NetAppStoragePerformancePackage.PACKAGE);
        storagePerformanceListener = new NetAppStoragePerformanceStandaloneListener();

        storagePerformancePackage =
                performanceConfiguration.instantiatePackage(dataSource, storagePerformanceListener, storagePerformanceListener, storagePerformanceListener);
        dataSource.registerPackage(storagePerformancePackage);
    }


    /**
     * Creates an instance of the target datasource, then run methods: start, run and stop
     *
     * @throws FrameworkException On any framework problem
     * @throws DataSourceErrorException On poll failure
     */
    public void run() throws FrameworkException, DataSourceErrorException {
        super.run();
        final NetAppStoragePerformanceConfiguration perfConfiguration = configuration.getPackageConfiguration(NetAppStoragePerformancePackage.PACKAGE);
        final Properties properties = perfConfiguration.getProperties();
        if (!perfConfiguration.isEnabled()) {
            return;
        }
        final long waitInBetweenMS = perfConfiguration.getPollTimeMillis();
        final int perfSamples = Integer.parseInt(properties.getProperty("recording.samples", "3"));

        // Now execute performance several times
        for (int j = 1 ; j <= perfSamples; j++) {
            logger.info("Starting performance sample " + j + " of " + perfSamples);
            dataSource.run(NetAppStoragePerformancePackage.PACKAGE);
            if (j < perfSamples) {
                try {
                    logger.info("Waiting " + waitInBetweenMS + "ms for next poll to start...");
                    Thread.sleep(waitInBetweenMS);
                }
                catch (InterruptedException e) {
                    // Fall through...
                }
            }
        }
    }

    /**
     * Main class, run it as stand-alone application, usually for recording purposes.
     *
     * @param args Arguments
     */
    public static void main(final String[] args) {
        setup(NetAppStoragePerformanceStandalone.class, args);
    }

    /**
     * This is what main does - but if also has a parameter for which class should be instantiated when running the
     * standalone.
     *
     * This is used to allow reusing of this code - while still supporting different implementations.
     *
     * @param clazz Standalone class to instantiate
     * @param args Arguments
     */
    protected static void setup(final Class clazz, final String[] args) {
        if (args.length < 3) {
            System.out.println("Usage: standalone -netappfoundation.ip=<ip> -netappfoundation.user=<user> -netappfoundation.password<password> -loglevel=<DEBUG|INFO|WARN|ERROR>");
            return;
        }

        try {
            setLogLevel(Level.ERROR);

            ServiceManager.getInstance().startForStandalone();
            ServiceManager.getInstance().start();

            // final NetAppStoragePerformanceStandalone standalone = new NetAppStoragePerformanceStandalone();

            try {

                final File deployDescriptor = new File("netapp_ocie.xml");
                if (!deployDescriptor.exists())
                    throw new FileNotFoundException("Could not find netapp_ocie.xml (expected to be located at current directory");

                final Properties properties = createPropertiesFromCommandLineArgs(args);

                final String logLevelRaw = properties.getProperty("loglevel");
                final Level logLevel = Level.toLevel(logLevelRaw, Level.INFO);

                // Construct the standalone instance using reflection to allow other implementations
                final Class[] constructorSignature = {java.io.File.class, java.util.Properties.class};
                final Constructor constructor = clazz.getConstructor(constructorSignature);
                final Object[] constructorValues = {deployDescriptor, properties};
                final NetAppStoragePerformanceStandalone standalone = (NetAppStoragePerformanceStandalone)constructor.newInstance(constructorValues);

                standalone.start();

                setLogLevel(logLevel);
                standalone.run();

                setLogLevel(Level.ERROR);
                standalone.stop();
            }
            finally {
                ServiceManager.getInstance().stop();
            }
        }
        catch (DataSourceErrorException e) {
            logger.error("Data source failed: " + e.getMessage(), e);
        }
        catch (Exception e) {
            logger.error("Internal error: " + e.getMessage(), e);
        }
    }
}
