package com.onaro.sanscreen.acquisition.schema;

import com.onaro.sanscreen.acquisition.framework.datasource.*;

import java.io.*;
import java.util.*;

/**
 * Provides a root for device model data that can be marshaled to xml
 *
 */
@SuppressWarnings({"UnusedDeclaration", "unchecked"})
abstract public class BaseModelRoot extends XMLNode<BaseRootFactory> implements Root, MarshallingCache {

    protected final Map<Comparable, Object> cache = new TreeMap<Comparable, Object>();

    // keep ordered in a tree set to keep diffing xml easier
    private final Map<String, String> stringAttributes = new TreeMap<String, String>();
    private final List<BaseXMLNode<MarshallingCache>> children = new ArrayList<BaseXMLNode<MarshallingCache>>();
    private final Map<String, List<BaseXMLNode<MarshallingCache>>> childrenUnderNode = new TreeMap<String, List<BaseXMLNode<MarshallingCache>>>();

    protected BaseModelRoot() throws SchemaException {
        super(null, "root");
    }

    protected final boolean hasChildren() {
        return !children.isEmpty() || !childrenUnderNode.isEmpty();
    }

    /**
     * Set up root attributes and children to be marshaled
     *  use methods:
     *  {@link #addMarshalingAttribute addMarshalingAttribute})
     *  {@link #addMarshalingChild addMarshalingChild})
     *  {@link #addMarshalingChildrenUnderNode addMarshalingChildrenUnderNode})
     *
     *
     * @throws SchemaException If an XML integrity error occurs
     */
    abstract public void addAttributesAndChildren() throws SchemaException;

    /**
     *
     * @param writer Writer to dump XML to
     *
     * @throws SchemaException If an XML integrity error occurs
     * @throws java.io.IOException If an IO error occurs
     */
    public void marshall(final Writer writer) throws IOException, SchemaException{
        addAttributesAndChildren();
        marshall(this, writer);
    }
     
    protected final void marshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        super.marshallChildren(cache, writer, tabs);

        for (final Map.Entry<String, String> entry: stringAttributes.entrySet()){
            addAttribute(writer, entry.getKey(), entry.getValue());
        }

        for (final BaseXMLNode child: children){
            addChild(cache, writer, child, tabs);
        }

        for (final Map.Entry<String, List<BaseXMLNode<MarshallingCache>>> entry: childrenUnderNode.entrySet()){
            final String nodeName = entry.getKey();
            final Collection<BaseXMLNode<MarshallingCache>> childList =  entry.getValue();
            this.addAllChildrenUnderNode(cache, writer, nodeName, childList, tabs);
        }
    }


    /**
     * Adds an XML attribute to be marshaled as children of the root.
     *
     * @param attributeName XML attribute name.
     * @param attributeValue Attribute value. If null, the attribute will not be set.
     *
     * @throws SchemaException If an <code>attributeName</code> is null or a duplicate attribute
     */
    protected final void addMarshalingAttribute(final String attributeName, final String attributeValue) throws SchemaException {
        if (attributeName == null)
            throw new SchemaException("model root attribute name can not be null");

        if (stringAttributes.containsKey(attributeName))
            throw new SchemaException("model root attribute name can not exist more than once: " +attributeName);

        stringAttributes.put(attributeName, attributeValue);
    }


    /**
     * Adds a list of  XMLNodes to be marshaled as children of the root.
     *
     * @param childList List of child object to marshal as the object's XML root children elements
     *
     * @throws SchemaException If <code>childList</code> is null or an element in <code>childList</code> is null
     */
    protected final void addMarshalingChildren(final Collection<? extends XMLNode> childList) throws SchemaException {
        if (childList == null)
            throw new SchemaException("model root childList can not be null");

        for (final XMLNode child: childList){
            addMarshalingChild(child);
        }
    }

    /**
     * Adds an XMLNode to be marshaled as a child of the root.
     *
     * @param childNode Child object to marshal as the object's XML root children elements
     *
     * @throws SchemaException If <code>childList</code> is null or an element in <code>childList</code> is null
     */
    protected final void addMarshalingChild(final BaseXMLNode<MarshallingCache> childNode) throws SchemaException{
        if (childNode == null)
                throw new SchemaException("model root child can not be null");

        children.add(childNode);
    }


    /**
     * Adds a list of  XMLNodes to be marshaled as children as root child node <code>nodeName</code>.
     *
     * @param nodeName Name of node to put children under
     * @param childList List of child object to marshal as the object's XML child elements
     *
     * @throws SchemaException If <code>nodeName</code> is null or <code>childList</code> is null or an element in <code>childList</code> is null
     */
    protected final void addMarshalingChildrenUnderNode(final String nodeName, final Collection<BaseXMLNode<MarshallingCache>> childList) throws SchemaException {
         if (nodeName == null)
            throw new SchemaException("model root child node name can not be null");

        if (childList == null)
            throw new SchemaException("model root child node list can not be null");


        List<BaseXMLNode<MarshallingCache>> list = childrenUnderNode.get(nodeName);
        if (list == null){
            list = new ArrayList<BaseXMLNode<MarshallingCache>>();
            childrenUnderNode.put(nodeName, list);
        }

        for (final BaseXMLNode child: childList){
            if (child == null)
                throw new SchemaException("model root child node list element can not be null");

            list.add(child);
        }
    }

    public final Object put(final Comparable key, final Object value) {
        return cache.put(key, value);
    }

    public final Object get(final Comparable key) {
        return cache.get(key);
    }

    public <K extends Key> K getKey() {
        return null;
    }

    public Info getInfo() {
        // DO NOTHING!
        return null;
    }

    public OriginatorType getOriginatorType() {
        // DO NOTHING!
        return null;
    }

    public Originator getOriginator() {
        // DO NOTHING!
        return null;
    }

    public String getOriginatorKey() {
        // DO NOTHING!
        return null;
    }

    public void setOriginatorKey(final String originatorKey) {
        // DO NOTHING!
    }

    abstract public void checkIntegrity() throws IntegrityException;
}