package com.onaro.sanscreen.acquisition.schema;

import com.onaro.util.Capacity;

import java.io.IOException;
import java.io.Serializable;
import java.io.Writer;
import java.util.*;

public abstract class BaseXMLNode<Cache> implements Serializable {
    private String nodeName;
    protected static final String lineSeperator = System.getProperty("line.separator");

    protected BaseXMLNode(String name) {
        nodeName = name;
    }

    public final String getNodeName() {
        return nodeName;
    }

    /**
     * Marshal the object into the appropriate XML schema.
     *
     * @param cache Object passed during marshalling that can be used to pass information between marshalled objects
     * @param writer Writer to dump XML to
     *
     * @throws SchemaException If an XML integrity error occurs
     * @throws java.io.IOException If an IO error occurs
     */
    protected void marshall(final Cache cache, final Writer writer) throws SchemaException, IOException {
        marshall(cache, writer, 0);
    }

    /**
     * Marshal the object into the appropriate XML schema.
     * <p/>
     * Usually this method should not be overriden. Insead, override {@link #marshallAttributes}, {@link #marshallChildren} and {@link #hasChildren}
     * to handle the object marshalling.
     *
     * @param cache Object passed during marshalling that can be used to pass information between marshalled objects
     * @param writer Writer to dump XML to
     * @param tabs The indentation of this node (to make XML more readable)
     *
     * @throws SchemaException If an XML integrity error occurs
     * @throws java.io.IOException If an IO error occurs
     */
    protected void marshall(final Cache cache, final Writer writer, int tabs) throws IOException, SchemaException {
        addTabulation(writer, tabs);
        writer.append("<").append(getNodeName()).append(' ');

        marshallAttributes(cache, writer);

        if (!hasChildren()) {
            writer.append("/>").append(lineSeperator);
        }
        else {
            writer.append(">").append(lineSeperator);

            marshallChildren(cache, writer, tabs + 1);

            addTabulation(writer, tabs);
            writer.append("</").append(getNodeName()).append(">").append(lineSeperator);
        }
    }

    /**
     * Adds all the object's XML attribute during marshalling
     *
     * @param cache Object passed during marshalling that can be used to pass information between marshalled objects
     * @param writer Writer to dump XML to
     *
     * @throws SchemaException If an XML integrity error occurs
     * @throws java.io.IOException If an IO error occurs
     */
    @SuppressWarnings({"UnusedDeclaration"})
    protected void marshallAttributes(final Cache cache, final Writer writer) throws IOException, SchemaException {
        // Default implementation marshals with no attributes
    }

    /**
     * Indicates if this object has any XML child elements.
     * When true, {@link #marshallChildren} should also be overriden.
     *
     * @return true if node has children
     */
    protected boolean hasChildren() {
        return false;
    }

    /**
     * Adds the objects XML child elements during marshalling
     *
     * @param cache Object passed during marshalling that can be used to pass information between marshalled objects
     * @param writer Writer to dump XML to
     * @param tabs The indentation of this node (to make XML more readable)
     *
     * @throws SchemaException If an XML integrity error occurs
     * @throws java.io.IOException If an IO error occurs
     */
    @SuppressWarnings({"UnusedDeclaration"})
    protected void marshallChildren(final Cache cache, final Writer writer, int tabs) throws IOException, SchemaException {
        // Default implementation marshals with no children
    }

    /**
     * Adds an XML attribute to the object during marshalling.
     * This method takes care of null values (the attribute will not be added) and XML escaping.
     *
     * @param writer Writer to dump XML to
     * @param attributeName XML attribute name.
     * @param attributeValue Attribute value. If null, the attribute will not be set. The {@link String#valueOf(Object)} method
     * will be used to convert the object's value into a string.
     *
     * @throws java.io.IOException If an IO error occurs
     */
    protected final void addAttribute(final Writer writer, final String attributeName, final Object attributeValue) throws IOException {
        if (attributeValue == null)
            return;

        writer.append(attributeName).append("=\"").append(escape(String.valueOf(attributeValue))).append("\" ");
    }

    /**
     * Adds a capacity XML attribute to the object during marshalling.
     * This method takes care of null values (the attribute will not be added) and XML escaping. This also will translate
     * the capacity to the desired unit.
     *
     * @param writer Writer to dump XML to
     * @param attributeName XML attribute name.
     * @param capacity value. If null, the attribute will not be set.
     *
     * @throws java.io.IOException If an IO error occurs
     */
    protected final void addAttribute(final Writer writer, final String attributeName, final Capacity capacity, final Capacity.Unit unit) throws IOException {
        if (capacity == null)
            return;

        writer.append(attributeName).append("=\"").append(String.valueOf(capacity.value(unit))).append("\" ");
    }

    /**
     * Helper method to be used when marshling an object's XML child elements (during {@link #marshallChildren}).
     *
     * @param cache Object passed during marshalling that can be used to pass information between marshalled objects
     * @param writer Writer to dump XML to
     * @param childList List of child object to marshal as the object's XML child elements
     * @param tabs The indentation of this node (to make XML more readable)
     *
     * @throws SchemaException If an XML integrity error occurs
     * @throws java.io.IOException If an IO error occurs
     */
    protected final void addAllChildren(final Cache cache, final Writer writer, final Collection<? extends BaseXMLNode<Cache>> childList, int tabs) throws IOException, SchemaException {
        for (final BaseXMLNode<Cache> node : childList)
            node.marshall(cache, writer, tabs);
    }

    /**
     * Helper method to be used when marshling an object's XML child elements (during {@link #marshallChildren}) and these
     * elements go under a sub node (like Storage->Volumes in the odata schema).
     *
     * @param cache Object passed during marshalling that can be used to pass information between marshalled objects
     * @param writer Writer to dump XML to
     * @param nodeName Name of node to put children under
     * @param childList List of child object to marshal as the object's XML child elements
     * @param tabs The indentation of this node (to make XML more readable)
     *
     * @throws SchemaException If an XML integrity error occurs
     * @throws java.io.IOException If an IO error occurs
     */
    protected final void addAllChildrenUnderNode(final Cache cache, final Writer writer, final String nodeName, final Collection<? extends BaseXMLNode<Cache>> childList, int tabs) throws IOException, SchemaException {
        //Dont print container element if the child list is empty
        //this will save us the check for every container element
        if (childList == null || childList.size() == 0)
            return;

        addTabulation(writer, tabs);
        writer.append("<").append(nodeName).append(">").append(lineSeperator);

        for (final BaseXMLNode<Cache> node : childList)
            node.marshall(cache, writer, tabs + 1);

        addTabulation(writer, tabs);
        writer.append("</").append(nodeName).append(">").append(lineSeperator);
    }


    /**
     * Helper method to be used when marshling an object's XML child elements (during {@link #marshallChildren}) and these
     * elements go under a sub node (like Storage->Volumes in the odata schema).
     *
     * @param cache Object passed during marshalling that can be used to pass information between marshalled objects
     * @param writer Writer to dump XML to
     * @param nodeName Name of node to put children under
     * @param childList List of child object to marshal as the object's XML child elements
     * @param tabs The indentation of this node (to make XML more readable)
     * @param comparator used to override the sorting of nodes based on the keys
     *
     * @throws SchemaException If an XML integrity error occurs
     * @throws java.io.IOException If an IO error occurs
     */
    protected final void addAllChildrenUnderNode(final Cache cache, final Writer writer, final String nodeName,
                                                 final Collection<? extends BaseXMLNode<Cache>> childList,
                                                 final int tabs, final Comparator comparator) throws IOException, SchemaException {
        // First sort the collection using the comparator, then marshall using the resulting collection
        List<? extends BaseXMLNode<Cache>> sortedChildList = new ArrayList(childList);
        Collections.sort(sortedChildList, comparator);
        addAllChildrenUnderNode(cache, writer, nodeName, sortedChildList, tabs);
    }




    /**
     * Helper method to be used when marshling an object's XML child elements (during {@link #marshallChildren}).
     *
     * @param cache Object passed during marshalling that can be used to pass information between marshalled objects
     * @param writer Writer to dump XML to
     * @param childNode Child object to marshal as one of the object's XML child elements (when null, NOOP)
     * @param tabs The indentation of this node (to make XML more readable)
     *
     * @throws SchemaException If an XML integrity error occurs
     * @throws java.io.IOException If an IO error occurs
     */
    protected final void addChild(final Cache cache, final Writer writer, final BaseXMLNode<Cache> childNode, int tabs) throws SchemaException, IOException {
        if (childNode == null)
            return;

        childNode.marshall(cache, writer, tabs);
    }

    /**
     * Sets up the tabulation to present the XML is a better format
     *
     * @param writer Writer to dump XML to
     * @param tabs The indentation of this node (to make XML more readable)
     *
     * @throws java.io.IOException If an IO error occurs
     */
    protected void addTabulation(final Writer writer, int tabs) throws IOException {
        for (int i = 0; i < tabs; i++)
            writer.append('\t');
    }

    /**
     * Converts the specified string to an XML appropriate version by escaping reserved ascii characters
     *
     * @param unescapedString String to encode
     *
     * @return XML legal string representation of suspect string
     */
    @SuppressWarnings({"ConstantConditions"})
    protected final String escape(final String unescapedString) {
        StringBuilder escapedString = new StringBuilder(unescapedString.length());

        for (int i = 0; i < unescapedString.length(); i++) {
            char unescapeCharacter = unescapedString.charAt(i);

            switch (unescapeCharacter) {
                case '"':
                    escapedString.append("&quot;");
                    break;
                case '\'':
                    escapedString.append("&apos;");
                    break;
                case '&':
                    escapedString.append("&amp;");
                    break;
                case '<':
                    escapedString.append("&lt;");
                    break;
                case '>':
                    escapedString.append("&gt;");
                    break;
                default:
                    escapedString.append(unescapeCharacter);
            }
        }

        return escapedString.toString();
    }
}
