package com.onaro.sanscreen.acquisition.schema;

import java.util.*;
import java.io.Serializable;

/**
 * Utility class for representing map of nodes. This list implements the {@link java.util.Map} interface and behaves in terms of
 * performance like {@link java.util.TreeMap}. It performs parental checks to verify that map refers to elements from the same
 * parent.
 */
public class NodeMap<P, K extends Comparable, N extends XMLNode> implements Map<K, N>, Serializable{
    private final static long serialVersionUID = 1;
    private final P parent;
    private final Map<K, N> map = new TreeMap<K, N>();

   protected NodeMap(final P parent) {
        this.parent = parent;
    }

    public P getParent() {
        return parent;
    }

    public int size() {
        return map.size();
    }

    public boolean isEmpty() {
        return map.isEmpty();
    }

    public boolean containsKey(final Object key) {
        //noinspection SuspiciousMethodCalls
        return map.containsKey(key);
    }

    public boolean containsValue(final Object value) {
        //noinspection SuspiciousMethodCalls
        return map.containsValue(value);
    }

    public N get(final Object key) {
        //noinspection SuspiciousMethodCalls
        return map.get(key);
    }

    public N put(final K key, final N value) {
        if (value.getParent() != parent)
            throw new IllegalArgumentException("Node parent does not match map parent");

        return map.put(key, value);
    }

    public N remove(final Object key) {
        //noinspection SuspiciousMethodCalls
        return map.remove(key);
    }

    public void putAll(final Map<? extends K, ? extends N> t) {
        for (final N n : t.values()) {
            if (n.getParent() != parent)
                throw new IllegalArgumentException("Node parent does not match map parent");
        }

        map.putAll(t);
    }

    public void clear() {
        map.clear();
    }

    public Set<K> keySet() {
        return map.keySet();
    }

    public Collection<N> values() {
        return map.values();
    }

    public Set<Entry<K, N>> entrySet() {
        return map.entrySet();
    }

    @SuppressWarnings({"EqualsWhichDoesntCheckParameterClass"})
    public boolean equals(final Object o) {
        return map.equals(o);
    }

    public int hashCode() {
        return map.hashCode();
    }
}
