package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.schema.*;

import java.io.IOException;
import java.io.Writer;

abstract class BaseRootImpl extends XMLNode<BaseRootFactory> implements OData2Root {
    private final Originator originator;
    private final OriginatorType originatorType;

    private final Info info;

    private final WWNAliases wwnAliases;
    private final Unknowns unknowns;

    private boolean alreadyMarshalled = false;

    private final RootKey key = new RootKey();

    // keep a strong reference to the root factory
    // the parent is kept as a weak reference and can be garbage collected (first seen in bug 20629) 
    private final RootFactory rootFactory;

    protected BaseRootImpl(final RootFactory parent, final long originatorID, final String originatorType, final String dataSourceName) throws SchemaException {
        super(parent, "root");

        if (parent == null || originatorType == null || originatorType.length() == 0 || dataSourceName == null || dataSourceName.length() == 0)
            throw new SchemaException(getNodeName() + " Type requires non-null parent, originator, and originatorType");

        this.rootFactory = parent;
        this.originator = new Originator(originatorID);
        this.originatorType = new OriginatorType(originatorType);

        this.info = new Info(this, originatorID, dataSourceName);

        wwnAliases = new WWNAliases(this);
        unknowns = new Unknowns(this);
    }

    public BaseRootFactory getRootFactory() {
        return rootFactory;
    }

    public RootKey getKey() {
        return key;
    }

    public Info getInfo() {
        return info;
    }

    public Originator getOriginator() {
        return originator;
    }

    public OriginatorType getOriginatorType() {
        return originatorType;
    }

    public String getOriginatorKey() {
        return info.getOriginatorKey();
    }

    public void setOriginatorKey(String originatorKey) {
        this.info.setOriginatorKey(originatorKey);
    }

    /**
     * This method cannot be overriden since it defined the order of the element under root.
     * Use {@link #doMarshallChildren} to add additional elements
     */
    protected final void marshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        super.marshallChildren(cache, writer, tabs);

        // TODO: Change schema to add data source meta data!!!
        addChild(cache, writer, originator, tabs);
        addChild(cache, writer, originatorType, tabs);
        addChild(cache, writer, info, tabs);

        doMarshallChildren(cache, writer, tabs);
    }

    /**
     * Adds all the children for this root.
     *
     * Default implementation will add the common objects (unknowns and wwn aliases)
     */
    protected void doMarshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        addAllChildren(cache, writer, unknowns, tabs);
        addAllChildren(cache, writer, wwnAliases.prepareForMarshalling(), tabs);
    }

    protected boolean hasChildren() {
        return true;
    }

    public void marshall(final Writer writer) throws IOException, SchemaException {
        if (alreadyMarshalled)
            throw new SchemaException("Unable to marshall more than once (originator " + getOriginator().getOriginatorID() + ")");

        alreadyMarshalled = true;
        marshall(getRootFactory(), writer);
    }

    public WWNAliases getWWNAliases() {
        return wwnAliases;
    }

    public Unknowns getUnknowns() {
        return unknowns;
    }

    public Unknown createUnknown(final String wwn) throws SchemaException {
        return new Unknown(this, wwn);
    }

    public class RootKey implements Key<RootKey> {
        public int compareTo(final RootKey o) {
            // Scale down the long result into an integer (we just need to maintain the negative, positive and zero attributes)
            long l = getOriginatorID() - o.getOriginatorID();

            if (l == 0)
                return 0;
            else if (l > 0)
                return 1;
            else
                return -1;
        }

        public long getOriginatorID() {
            return originator.getOriginatorID();
        }
    }
}



