package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;
import com.onaro.util.Capacity;

import java.io.IOException;
import java.io.Writer;

public class DataStore extends XMLNode<VirtualizationRoot> {
    private final String managementIp;
    private final String moid;

    private String name;
    private Capacity capacity;
    private Capacity usedCapacity;
    private Capacity provisionedCapacity;
    private final DataStoreKey key;

    DataStore(final VirtualizationRoot parent, final String moid, final String managementIp) throws SchemaException {
        super(parent, "HV_DataStore");

        if (moid == null || moid.length() == 0)
            throw new SchemaException("MOID is mandatory!");

        if (managementIp == null || managementIp.length() == 0)
            throw new SchemaException("virtualCenterIP is mandatory!");

        this.managementIp = managementIp;
        this.moid = moid;
        this.key = new DataStoreKey(moid, managementIp);
    }

    public DataStoreKey getKey() {
        return key;
    }

    public String getMOID() {
        return moid;
    }

    public String getName() {
        return name;
    }

    /**
     * Maximum (configured) capacity of this datastore.
     *
     * @return maximum capacity of DataStore as a capacity object.
     */
    public Capacity getCapacity() {
        return capacity;
    }

    public void setName(final String name) {
        this.name = name;
    }

    public void setCapacity(final Capacity capacity) {
        this.capacity = capacity != null ? Capacity.create(capacity) : null;
    }

    /**
     * The actual used capacity of this DataStore.  Must be less than the amount in provisioned capacity.  This is
     * optional.
     *
     * @return the actual used capacity in the Datastore.  NULL if not set.
     */
    public Capacity getUsedCapacity() {
        return usedCapacity;
    }

    public void setUsedCapacity(final Capacity usedCapacity) {
        this.usedCapacity = usedCapacity != null ? Capacity.create(usedCapacity) : null;
    }

    /**
     * Amount of storage set-aside for use by a datastore.
     * Files on the datastore can expand to this size, but not beyond it.
     *
     * This is optional.  NULL if not set.
     *
     * @return the provisioned capacity of the DataStore.  NULL if not set.
     */
    public Capacity getProvisionedCapacity() {
        return provisionedCapacity;
    }

    public void setProvisionedCapacity(final Capacity provisionedCapacity) {
        this.provisionedCapacity = provisionedCapacity != null ? Capacity.create(provisionedCapacity) : null;
    }

    public String getManagementIp() {
        return managementIp;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException, SchemaException {
        super.marshallAttributes(cache, writer);

        addAttribute(writer, "OID", getKey().toString());

        addAttribute(writer, "Name", name);

        addAttribute(writer, "CapacityMB", capacity != null ? capacity.value(Capacity.Unit.MB_1024) : -1);

        if (usedCapacity != null) {
            addAttribute(writer, "UsedCapacityMB", usedCapacity.value(Capacity.Unit.MB_1024));
        }
        if (provisionedCapacity != null) {
            addAttribute(writer, "ProvisionedCapacityMB", provisionedCapacity.value(Capacity.Unit.MB_1024));
        }

        addAttribute(writer, "VirtualCenterIp", managementIp);
        addAttribute(writer, "MOID", moid);

    }

    public static class DataStoreKey implements Key<DataStoreKey> {
        private String managementIp;
        private String moid;

        public DataStoreKey(final String moid, final String managementIp) {
            this.moid = moid;
            this.managementIp = managementIp;
        }

        public boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final DataStoreKey that = (DataStoreKey) o;

            if (moid != null ? !moid.equals(that.moid) : that.moid != null) return false;
            if (managementIp != null ? !managementIp.equals(that.managementIp) : that.managementIp != null)
                return false;

            return true;
        }

        public int hashCode() {
            int result;
            result = (managementIp != null ? managementIp.hashCode() : 0);
            result = 31 * result + (moid != null ? moid.hashCode() : 0);
            return result;
        }

        public int compareTo(final DataStoreKey o) {
            int i = moid.compareTo(o.getMOID());
            if (i != 0)
                return i;

            return managementIp.compareTo(o.getManagementIp());
        }

        private String getManagementIp() {
            return managementIp;
        }

        public String getMOID() {
            return moid;
        }

        public String toString() {
            return getMOID() + ":" + getManagementIp();
        }
    }
}
