package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;
import com.onaro.util.Pair;

import java.io.IOException;
import java.io.Serializable;
import java.io.Writer;
import java.util.Comparator;
import java.util.TreeSet;

public class DeviceGroup extends XMLNode<StorageRoot> {
    private final String name;
    private final String storageManagementID;
    private String type = null;
    private VolumeRefSet volumeRefs = new VolumeRefSet(new VolPrefComparator());

    private final DeviceGroupKey key;

    /**
     * Use {@link StorageRoot#createDeviceGroup} to create this object
     */
    DeviceGroup(final StorageRoot parent, final String name, final String storageManagementId) throws SchemaException {
        super(parent, "DeviceGroup");

        if (name == null || name.length() == 0)
            throw new SchemaException(getNodeName() + " DeviceGroup must have non-empty name");

        this.name = name;

        if (storageManagementId == null || storageManagementId.length() == 0)
            throw new SchemaException(getNodeName() + " DeviceGroup must have non-empty storageManagementId");

        this.storageManagementID = storageManagementId;

        key = new DeviceGroupKey(name, storageManagementID);
    }

    public final DeviceGroupKey getKey() {
        return key;
    }

    public final StorageRoot getRoot() {
        return getParent();
    }

    public final String getName() {
        return name;
    }

    public final String getType() {
        return type;
    }

    public final void setType(final String type) {
        this.type = type;
    }

    public final String getStorageManagementID() {
        return storageManagementID;
    }

    public final VolumeRefSet getVolumeRefs() {
        return volumeRefs;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException {
        addAttribute(writer, "Name", name);
        addAttribute(writer, "Type", type);
        addAttribute(writer, "StorageManagementId", storageManagementID);
    }

    protected boolean hasChildren() {
        return volumeRefs.size() > 0;
    }

    protected void marshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        for (final Pair<String, String> volumeRef : volumeRefs) {
            addTabulation(writer, tabs);
            writer.append("<").append("VolumeRef").append(' ');
            writer.append("Name=\"").append(volumeRef.y).append("\" ");
            writer.append("StorageIp=\"").append(volumeRef.x).append("\" ");
            writer.append("/>").append(lineSeperator);
        }
    }

    /**
     * Stores mappigs between storage IPs and device group names
     */
    public static class VolumeRefSet extends TreeSet<Pair<String, String>> {
        public VolumeRefSet() {
        }

        public VolumeRefSet(final Comparator<? super Pair<String, String>> c) {
            super(c);
        }
    }

    public final static class DeviceGroupKey implements Key<DeviceGroupKey> {
        private final String name;
        private final String storageManagementID;

        public DeviceGroupKey(final String name, final String storageManagementID) {
            this.name = name;
            this.storageManagementID = storageManagementID;
        }

        public final int compareTo(final DeviceGroupKey o) {
            int i;

            i = getName().compareTo(o.getName());
            if (i != 0)
                return i;

            return getStorageManagementID().compareTo(o.getStorageManagementID());
        }

        public final boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final DeviceGroupKey that = (DeviceGroupKey) o;

            if (name != null ? !name.equals(that.name) : that.name != null) return false;
            if (storageManagementID != null ? !storageManagementID.equals(that.storageManagementID) : that.storageManagementID != null)
                return false;

            return true;
        }

        public final int hashCode() {
            int result;
            result = (name != null ? name.hashCode() : 0);
            result = 29 * result + (storageManagementID != null ? storageManagementID.hashCode() : 0);
            return result;
        }

        public final String toString() {
            return name + ":" + storageManagementID;
        }

        public final String getName() {
            return name;
        }

        public final String getStorageManagementID() {
            return storageManagementID;
        }
    }

    static class VolPrefComparator implements Comparator<Pair<String, String>>, Serializable {
        public int compare(final Pair<String, String> o1, final Pair<String, String> o2) {
            return o1.toString().compareTo(o2.toString());
        }
    }
}
