package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;

import java.io.IOException;
import java.io.Writer;
import java.util.HashSet;
import java.util.Set;

public class Fabric extends XMLNode<FabricRoot> {
    private String name = null;

    private ActiveZoningCapabilities activeZoningCapabilities;
    private final InactiveZoningCapabilitiesCollection inactiveZoningCapabilities;

    private final Switches switches;

    private final FabricKey key;

    /**
     * Use {@link FabricRoot#createFabric} to create this object
     * @param parent - FabricRoot that contains this Fabric
     * @param wwn - fabric wwn
     * @throws com.onaro.sanscreen.acquisition.schema.SchemaException - if there is an error changing the model
     */
    Fabric(final FabricRoot parent, final String wwn) throws SchemaException {
        super(parent, "Fabric");

        if (wwn == null || wwn.length() == 0)
            throw new SchemaException(getNodeName() + " Type requires non-null oid and wwn");

        inactiveZoningCapabilities = new InactiveZoningCapabilitiesCollection(this);
        switches = new Switches(this);

        key = new FabricKey(wwn.toUpperCase());
    }

    @SuppressWarnings({"unchecked"})
    public FabricKey getKey() {
        return key;
    }

    public FabricRoot getRoot() {
        return getParent();
    }

    public Switches getSwitches() {
        return switches;
    }

    public Switch createSwitch(final String wwn) throws SchemaException {
       return new Switch(this, wwn, SwitchType.Type.FABRIC_PHYSICAL);
    }

    @SuppressWarnings("UnusedDeclaration")
    public Switch createNpvChassisSwitch(final String wwn) throws SchemaException {
        return new Switch(this, wwn, SwitchType.Type.NPV_CHASSIS);
    }

    public Switch createNpvPhysicalSwitch(final String wwn) throws SchemaException {
        return new Switch(this, wwn, SwitchType.Type.NPV_PHYSICAL);
    }

    public Switch createNpvHubSwitch(final String wwn) throws SchemaException {
        return new Switch(this, wwn, SwitchType.Type.NPV_HUB);
    }

    public String getName() {
        return name;
    }

    public void setName(final String name) {
        this.name = name;
    }

    public String getWWN() {
        return key.wwn;
    }

    public ActiveZoningCapabilities createActiveZoningCapabilities(final String name) throws SchemaException {
        return new ActiveZoningCapabilities(this, name);
    }

    public InactiveZoningCapabilities createInactiveZoningCapabilities(final String name) throws SchemaException {
        return new InactiveZoningCapabilities(this, name);
    }

    public ActiveZoningCapabilities getActiveZoningCapabilities() {
        return activeZoningCapabilities;
    }

    /**
     * Configures the fabric with zoning disabled
     *
     * @param zoningBehavior Specifies if access is permitted or denied to all ports
     *
     * @throws SchemaException if any other zoning configuration was set (active / inactive zoning capabilites)
     * @see ZoningCapabilities.ZoningBehavior
     */
    @SuppressWarnings("UnusedDeclaration")
    public void disableZoningForFabric(final ZoningCapabilities.ZoningBehavior zoningBehavior) throws SchemaException {
        if (activeZoningCapabilities != null)
            throw new SchemaException(getNodeName() + " Active zoning capabilities was already set for this fabric");

        if (inactiveZoningCapabilities.size() != 0)
            throw new SchemaException(getNodeName() + " Inactive zoning capabilities ware already set for this fabric");

        activeZoningCapabilities = new ActiveZoningCapabilities(this, "ZONING-DISABLED");
        activeZoningCapabilities.setDefaultZoningBehavior(zoningBehavior);
    }

    public void setActiveZoningCapabilities(final ActiveZoningCapabilities activeZoningCapabilities) throws SchemaException {
        this.activeZoningCapabilities = activeZoningCapabilities;
    }

    public InactiveZoningCapabilitiesCollection getInactiveZoningCapabilities() {
        return inactiveZoningCapabilities;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException {
        addAttribute(writer, "WWN", key.wwn);
        addAttribute(writer, "Name", name);
    }

    protected void marshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        addAllChildren(cache, writer, switches, tabs);

        if (activeZoningCapabilities != null)
            addChild(cache, writer, activeZoningCapabilities, tabs);

        addAllChildren(cache, writer, inactiveZoningCapabilities, tabs);
    }

    protected boolean hasChildren() {
        return !inactiveZoningCapabilities.isEmpty() ||
               (activeZoningCapabilities != null) ||
               !switches.isEmpty();
    }

    public final static class FabricKey implements Key<FabricKey> {
        private final String wwn;

        public FabricKey(final String wwn) {
            this.wwn = wwn;
        }

        public boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final FabricKey fabricKey = (FabricKey) o;

            if (wwn != null ? !wwn.equals(fabricKey.wwn) : fabricKey.wwn != null) return false;

            return true;
        }

        @Override
        public int hashCode() {
            return (wwn != null ? wwn.hashCode() : 0);
        }

        public int compareTo(final FabricKey o) {
            return getWWN().compareTo(o.getWWN());
        }

        public String getWWN() {
            return wwn;
        }
    }
}
