package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.IntegrityException;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;

import java.io.IOException;
import java.io.Writer;
import java.util.*;

class FabricRootImpl extends BaseRootImpl implements FabricRoot {
    private final ChassisSwitches chassisSwitches;
    private final Fabrics fabrics;
    private final Fabrics npvFabrics;
    private IVRPhysicalFabric ivrPhysicalFabric;

    FabricRootImpl(final RootFactory parent, final long originatorID, final String originatorType, final String dataSourceName) throws SchemaException {
        super(parent, originatorID, originatorType, dataSourceName);

        fabrics = new Fabrics(this);
        npvFabrics = new Fabrics(this);
        chassisSwitches = new ChassisSwitches(this);
    }

    public ChassisSwitch createChassisSwitch(final String wwn, final String ip) throws SchemaException {
        final ChassisSwitch chassisSwitch = new ChassisSwitch(this, wwn, SwitchType.Type.CHASSIS);
        chassisSwitch.setIP(ip);

        return chassisSwitch;
    }

    // Added so that we can discover NPV or access gateway devices and report as Chassis with logical switches.
    public ChassisSwitch createNpvChassisSwitch(final String wwn, final String ip) throws SchemaException {
        final ChassisSwitch chassisSwitch = new ChassisSwitch(this, wwn, SwitchType.Type.NPV_CHASSIS);
        chassisSwitch.setIP(ip);

        return chassisSwitch;
    }

    public ChassisSwitches getChassisSwitches() {
        return chassisSwitches;
    }

    public Fabrics getFabrics() {
        return fabrics;
    }

    /**
     * Separate container for NPV fabrics to avoid key collision. The key for a fabric is the wwn
     * and we need to report the NPV switches in a separate fabric when we have logical switches. 
     * @return  Set of NPV fabrics
     */
    public Fabrics getNPVFabrics() {
        return npvFabrics;
    }

    public Fabric createFabric(final String wwn) throws SchemaException {
        return new Fabric(this, wwn);
    }

    public VSAN createVSAN(final String wwn) throws SchemaException {
        return new VSAN(this, wwn);
    }

    public IVRPhysicalFabric createIVRPhysicalFabric(final Collection<String> ivrChassisWWNs) throws SchemaException {
        return new IVRPhysicalFabric(this, ivrChassisWWNs);
    }

    public IVRPhysicalFabric getIVRPhysicalFabric() {
        return ivrPhysicalFabric;
    }

    public void setIVRPhysicalFabric(final IVRPhysicalFabric ivrPhysicalFabric) {
        this.ivrPhysicalFabric = ivrPhysicalFabric;
    }

    public WWNAlias createSwitchZoneAlias(final String alias, final String wwn) throws SchemaException {
        return new WWNAlias(this, WWNAlias.Source.Switch, alias, wwn, WWNAlias.ObjectType.Unknown);
    }

    protected void doMarshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        addAllChildren(cache, writer, chassisSwitches, tabs);
        addAllChildren(cache, writer, fabrics, tabs);
        addAllChildren(cache, writer, npvFabrics, tabs);
        addChild(cache, writer, ivrPhysicalFabric, tabs);

        super.doMarshallChildren(cache, writer, tabs);
    }

    /**
     * Perform integrity checks on switch device data
     *
     * @throws IntegrityException
     */
    public void checkIntegrity() throws IntegrityException {
        // Get a list of all the chassis WWNs
        final Set<String> chassisWWNs = new HashSet<String>();
        final Map<PortType.AbstractPortTypeKey, String> connectedPortsMap = new HashMap<PortType.AbstractPortTypeKey, String>(2048);

        for (final ChassisSwitch chassisSwitch : getChassisSwitches()) {
            chassisWWNs.add(chassisSwitch.getWWN());
        }

        // Map of all known fabric and node ports, indexed by port WWN
        final Map<String, Port> allPorts = new HashMap<String, Port>(2048);

        for (final Fabric fabric : getFabrics()) {
            for (final Switch switchDevice : fabric.getSwitches()) {
                // First check if switch has ports...
                if (switchDevice.getPorts().size() == 0) {
                    throw new IntegrityException("Switch (" + switchDevice.getIP() + ") missing ports");
                }

                checkIntegrityOfConnectedPorts(switchDevice.getPorts(), connectedPortsMap);

                checkFabricSwitchPorts(switchDevice, allPorts);
            }

/*
            if (fabric.getActiveZoningCapabilities() == null) {
                // Usually, this is an indication that there's somthing wrong, since a fabric should have at least an
                // active zoning capabilities. But since we implemented the Multiple Originator ID project, data sources
                // are reporting each switch with it's own originator ID / root object (without any zoning) and the zoning
                // for the entire fabric are reported under a special originator ID.
                // Therefore, an exception is thrown only if this fabric has switches defined. If it does, this hints us
                // that this is not a multiple originator ID enabled data source.
                throw new IntegrityException("Fabric (" + fabric.getName() + ") missing active zoning capabilities");
            }
*/
            if (fabric instanceof VSAN) {
                final VSAN vsan = (VSAN) fabric;

                // When working with VSANs, each logical switch must point to an existing Chassis switch.
                for (final LogicalSwitch logicalSwitch : vsan.getLogicalSwitches()) {
                    if (!chassisWWNs.contains(logicalSwitch.getChassisWWN())) {
                        throw new IntegrityException("Logical switch " + logicalSwitch.getName() + " points to an invalid/non existing chassis switch!");
                    }

                    checkIntegrityOfConnectedPorts(logicalSwitch.getPorts(), connectedPortsMap);

                    checkFabricSwitchPorts(logicalSwitch, allPorts);
                }

                if (vsan.getLogicalSwitches().size() > 0 && (vsan.isVsanEnabled() == null || !vsan.isVsanEnabled())) {
                    throw new IntegrityException("Logical switches are defined in VSAN " + vsan.getName() + " but VSAN attribute is not enabled!");
                }

                // TODO: Check that there is at least one zoning capabilities object
                // TODO: Check that there is at least one active zoning capabilities object (if zoning is enabled)
            }
        }

        for (final Unknown unknown : getUnknowns()) {
            for (final UnknownPort unknownPort : unknown.getPorts()) {
                checkPort(unknownPort, allPorts);
            }
        }

        checkUnknownConnectivity(allPorts);
    }

    private void checkIntegrityOfConnectedPorts(final SwitchPorts switchPorts, final Map<PortType.AbstractPortTypeKey, String> connectedPortsMap) throws IntegrityException {
        for (final SwitchPort switchPort : switchPorts) {
            final ConnectedPort connectedPort = switchPort.getConnectedPort();
            if (connectedPort != null) {
                final PortType.AbstractPortTypeKey connectedPortKey = connectedPort.getKey();
                final String connectingPortWwn = connectedPortsMap.get(connectedPortKey);
                if (connectingPortWwn != null) {
                    throw new IntegrityException("The port " + connectedPortKey.toString() + " is already being connected by port " + connectingPortWwn);
                }
                connectedPortsMap.put(connectedPortKey, switchPort.getWWN());
            }
        }
    }

    private void checkFabricSwitchPorts(final FabricSwitchType fabricSwitch,
            final Map<String, Port> allPorts) throws IntegrityException {
        for (final SwitchPort switchPort : fabricSwitch.getPorts()) {
            checkPort(switchPort, allPorts);
        }
    }

    private void checkPort(final Port port, final Map<String, Port> allPorts)
            throws IntegrityException {
        final Port previousPort = allPorts.put(port.getWWN(), port);
        if (previousPort != null) {
            throw new IntegrityException("Duplicate ports " + port.getWWN()
                    + " found on nodes " + previousPort.getParent() + " and "
                    + port.getParent());
        }
    }

    /**
     * For any pair of connected ports, check that the <tt>UnknownConnectivity</tt>
     * flag is <tt>true</tt> on both ports if either port is generated.
     */
    private void checkUnknownConnectivity(final Map<String, Port> allPorts) throws IntegrityException {
        for (final Port port : allPorts.values()) {
            final ConnectedPort connectedPort = port.getConnectedPort();
            if ((connectedPort != null) && connectedPort.hasWWN()) {
                final String wwn;
                try {
                    wwn = connectedPort.getWWN();
                }
                catch (SchemaException e) {
                    throw new IntegrityException("Invalid connected port: " + connectedPort);
                }
                final Port otherPort = allPorts.get(wwn);
                if (otherPort != null) {
                    if ((port.isGenerated() || otherPort.isGenerated())
                            && (!port.isUnknownConnectivity() || !otherPort.isUnknownConnectivity())) {
                        throw new IntegrityException("Unknown connectivity flag must be set on both ports "
                                + port.getWWN() + " and " + otherPort.getWWN());
                    }
                }
            }
        }
    }
}
