package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;

import java.io.IOException;
import java.io.Writer;

public class FileShare extends XMLNode<Storage> {
    private final String name;
    private final String path;
    private final Qtree qtree;
    //this file share represent the whole volume
    private Boolean IsInternalVolume = Boolean.FALSE;
    private String status;
    private String securityType;
    private Boolean isOplocksEnabled;
    private String virtualStorage;
    private final ShareKey key;

    final Shares shares = new Shares(this);

    FileShare(final Storage parent, final String name, final String path, final Qtree qtree) throws SchemaException {
        super(parent, "FileShare");
        if (name == null || qtree == null || path == null)
            throw new SchemaException(getNodeName() + " Type requires non-null name path and qtree");

        this.name = name;
        this.qtree = qtree;
        this.path = path;
        key = new ShareKey(qtree.getKey().toString(), name, path);
    }

    public Qtree getQtree() {
        return qtree;
    }

    @SuppressWarnings({"unchecked"})
    public ShareKey getKey() {
        return key;
    }

    public String getName() {
        return name;
    }


    public String getPath() {
        return path;
    }

    public Boolean isInternalVolume() {
        return IsInternalVolume;
    }

    public void setIsInternalVolume(final Boolean isInternalVolume) {
        IsInternalVolume = isInternalVolume;
    }

    public Share createShare(final String shareName, final Share.ShareType type) throws SchemaException {
        return new Share(this, shareName, type);
    }

    public String getStatus() {
        return status;
    }

    public void setStatus(final String status) {
        this.status = status;
    }

    public String getSecurityType() {
        return securityType;
    }

    public void setSecurityType(final String securityType) {
        this.securityType = securityType;
    }

    /**
     * Get the name of the virtual storage on which this FileShare exists.  The meaning of this may differ by DS.  In NetApp,
     * for instance, it will hold the vFiler name to which the FileShare belongs.
     *
     * @return Virtual storage, if any, on which this FileShare exists.  May be null.
     */
    public String getVirtualStorage() {
        return virtualStorage;
    }

    /**
     * Set the name of the virtual storage on which this FileShare exists.  Optional component.
     *
     * @param virtualStorage the name virtual storage on which this FileShare exists.
     */
    public void setVirtualStorage(String virtualStorage) {
        this.virtualStorage = virtualStorage;
    }

    public Shares getShares() {
        return shares;
    }

    public Boolean getOplocksEnabled() {
        return isOplocksEnabled;
    }

    public void setOplocksEnabled(final Boolean oplocksEnabled) {
        isOplocksEnabled = oplocksEnabled;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException {
        //TODO:  Return and marshall virtualStorage when 6.0 released.

        addAttribute(writer, "Name", name);
        addAttribute(writer, "Path", path);
        addAttribute(writer, "QtreeId", getQtree().getKey());
        addAttribute(writer, "IsInternalVolume", isInternalVolume());
        addAttribute(writer, "IsShared", shares.size() > 0);
        addAttribute(writer, "Status", status);
        addAttribute(writer, "SecurityType", getSecurityType());
        addAttribute(writer, "IsOplocksEnabled", getOplocksEnabled());
    }

    protected void marshallChildren(final MarshallingCache marshallingCache, final Writer writer, final int tabs) throws IOException, SchemaException {
        addAllChildrenUnderNode(marshallingCache, writer, "Shares", shares, tabs);
    }

    protected boolean hasChildren() {
        return shares.size() > 0;
    }

    public static final class ShareKey implements Key<ShareKey> {
        private final String name;
        private final String internalVolumeName;
        private final String path;

        public ShareKey(final String internalVolumeName, final String name, final String path) {
            this.internalVolumeName = internalVolumeName;
            this.name = name;
            this.path = path;
        }

        @Override
        public int compareTo(final ShareKey o) {
            int i = getName().compareTo(o.getName());
            if (i != 0)
                return i;

            i = getPath().compareTo(o.getPath());
            if (i != 0)
                return i;

            return getInternalVolumeName().compareTo(o.getInternalVolumeName());
        }

        public String getPath() {
            return path;
        }

        public String getName() {
            return name;
        }

        public String getInternalVolumeName() {
            return internalVolumeName;
        }

        @Override
        public String toString() {
            return name + ":" + internalVolumeName + ":" + path;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            ShareKey shareKey = (ShareKey) o;

            if (internalVolumeName != null ? !internalVolumeName.equals(shareKey.internalVolumeName) : shareKey.internalVolumeName != null)
                return false;
            if (name != null ? !name.equals(shareKey.name) : shareKey.name != null) return false;
            if (path != null ? !path.equals(shareKey.path) : shareKey.path != null) return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = name != null ? name.hashCode() : 0;
            result = 31 * result + (internalVolumeName != null ? internalVolumeName.hashCode() : 0);
            result = 31 * result + (path != null ? path.hashCode() : 0);
            return result;
        }
    }
}
