package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;

import java.io.IOException;
import java.io.Writer;

public class Host extends XMLNode<VirtualizationRoot> {
    private final String oid;

    private String name;
    private String domainName;
    private String platformType;
    private Integer memoryInMB;
    private String vendorName;
    private String model;
    private Integer processors;
    private Long cpuSpeed;
    private Integer nicCount;
    private String ips;
    private Integer nicSpeed;


    private HostKey key;

    private final HostAdapters adapters = new HostAdapters(this);
    private final ISCSIInitiatorNodes initiators = new ISCSIInitiatorNodes(this);

    /**
     * Create new Host.   OID specified must be the same as the IPs field.
     *
     * @param parent parent virtualization root.
     * @param ipsAsOid OID which must be the IPs list (comma separated, ordered).
     * @throws SchemaException    for model building errors.
     */
    Host(final VirtualizationRoot parent, final String ipsAsOid) throws SchemaException {
        super(parent, "Host");

        if (ipsAsOid == null || ipsAsOid.length() == 0)
            throw new SchemaException("Object id is mandatory!");

        this.oid = ipsAsOid;
        this.key = new HostKey(ipsAsOid);
    }

    /**
     * OID for Host will always be same as IPs field.   This is a requirement of server.
     *
     * @return the OID of the Host (same as IPs).
     */
    public String getOID() {
        return oid;
    }

    public String getIPs() {
        return ips;
    }

    public void setIPs(final String ips) {
        this.ips = ips;
    }

    public HostKey getKey() {
        return key;
    }

    public HostAdapter createAdapter(final String www) {
        return new HostAdapter(this, www);
    }

    public ISCSIInitiatorNode createISCSINode(final String name) throws SchemaException {
        return new ISCSIInitiatorNode(this, name);
    }

    public ISCSIInitiatorNodes getInitiators() {
        return initiators;
    }

    public String getName() {
        return name;
    }

    public String getDomainName() {
        return domainName;
    }

    public String getPlatformType() {
        return platformType;
    }

    public long getMemoryInMB() {
        return memoryInMB;
    }

    public String getVendorName() {
        return vendorName;
    }

    public String getModel() {
        return model;
    }

    public int getProcessors() {
        return processors;
    }

    public long getCpuSpeed() {
        return cpuSpeed;
    }

    public int getNicCount() {
        return nicCount;
    }

    public int getNicSpeed() {
        return nicSpeed;
    }

    public HostAdapters getAdapters() {
        return adapters;
    }

    public void setName(final String name) {
        this.name = name;
    }

    public void setDomainName(final String domainName) {
        this.domainName = domainName;
    }

    public void setPlatformType(final String platformType) {
        this.platformType = platformType;
    }

    public void setMemoryInMB(final int memoryInMB) {
        this.memoryInMB = memoryInMB;
    }

    public void setVendorName(final String vendorName) {
        this.vendorName = vendorName;
    }

    public void setModel(final String model) {
        this.model = model;
    }

    public void setProcessors(final int processors) {
        this.processors = processors;
    }

    public void setCpuSpeed(final long cpuSpeed) {
        this.cpuSpeed = cpuSpeed;
    }

    public void setNicCount(final int nicCount) {
        this.nicCount = nicCount;
    }

    public void setNicSpeed(final int nicSpeed) {
        this.nicSpeed = nicSpeed;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException, SchemaException {
        super.marshallAttributes(cache, writer);

        // Check schema...
        addAttribute(writer, "OID", getKey());

        addAttribute(writer, "Name", name);
        addAttribute(writer, "IPs", ips);
        addAttribute(writer, "HostDomain", domainName);
        addAttribute(writer, "PlatformType", platformType);
        if (memoryInMB != null) {
            addAttribute(writer, "HostInstalledMemory", memoryInMB);
        }
        addAttribute(writer, "Manufacturer", vendorName);
        addAttribute(writer, "HostModel", model);
        if (processors != null) {
            addAttribute(writer, "HostCpuCount", processors);
        }
        if (cpuSpeed != null) {
            addAttribute(writer, "HostCpuSpeed", cpuSpeed);
        }
        if (nicCount != null) {
            addAttribute(writer, "HostNicCount", nicCount);
        }
        if (nicSpeed != null) {
            addAttribute(writer, "HostNicSpeed", nicSpeed);
        }

    }

    protected void marshallChildren(final MarshallingCache cache, final Writer writer, int tabs) throws IOException, SchemaException {
        addAllChildren(cache, writer, adapters, tabs);

        if (initiators.size() > 0)
            addAllChildrenUnderNode(cache, writer, "ISCSINodes", initiators, tabs);
    }

    protected boolean hasChildren() {
        return (!adapters.isEmpty() || !initiators.isEmpty());
    }

    public final static class HostKey implements Key<HostKey> {
        final String ips;

        public HostKey(String ip) {
            this.ips = ip;
        }

        public int compareTo(final HostKey o) {
            return ips.compareTo(o.getIPs());
        }

        private String getIPs() {
            return ips;
        }

        public String toString() {
            return getIPs();
        }

        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final HostKey hostKey = (HostKey) o;

            if (ips != null ? !ips.equals(hostKey.ips) : hostKey.ips != null) return false;

            return true;
        }

        public int hashCode() {
            return (ips != null ? ips.hashCode() : 0);
        }
    }
}
