package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;
import com.onaro.util.Capacity;

import java.io.IOException;
import java.io.Writer;
import java.text.NumberFormat;
import java.util.Locale;

/**
 * InternalVolume is an object that you can carve lun / shares
 * like netapp volumes.
 */
@SuppressWarnings({"unchecked"})
public class InternalVolume extends XMLNode<Storage> {
    final static NumberFormat usaLocaleNumberFormat = NumberFormat.getInstance(Locale.US);

    static {
        usaLocaleNumberFormat.setMaximumFractionDigits(4);
    }

    private final StoragePool storagePool;
    private final String type;
    private final InternalVolumeCapacity capacity;

    private SpaceGuarantee spaceGuarantee;
    private boolean thinProvisioned = false;
    private boolean thinProvisioningSupported = false;
    private boolean isDedupEnabled = false;
    private Float dedupeSaving;
    private Long lastSnapshotTime;
    private Long snapshotCount;
    private String status;
    private String cloneSourceName;
    private String virtualStorage;

    // Auto Tiering attributes

    /**
     * Only valid for NetApp
     * Whether the volume can participate in hybrid caching
     */
    private String flashPoolEligibility = null;

    private final Qtrees qtrees = new Qtrees(this);

    private final InternalVolumeKey key;
    private String protectionType = null;

    InternalVolume(final Storage parent, final String name, final StoragePool storagePool, final String type) throws SchemaException {
        super(parent, "InternalVolume");

        if (storagePool == null || name == null || type == null || name.length() == 0 || type.length() == 0)
            throw new SchemaException("Internal Volume must have non-null storage pool, name and type");

        // Match parents of storage pool and internal volume
        if (!storagePool.getParent().equals(parent))
            throw new SchemaException("Internal Volume and Storage Pool have different parent Storage.");

        this.storagePool = storagePool;
        this.type = type;
        this.capacity = new InternalVolumeCapacity();

        this.key = new InternalVolumeKey(parent.getKey(), name);
    }

    public final InternalVolumeKey getKey() {
        return key;
    }

    public final String getName() {
        return getKey().getName();
    }

    public final Storage getStorage() {
        return getParent();
    }

    public final String getRedundency() {
        return null;
    }

    public final SpaceGuarantee getSpaceGuarantee() {
        return spaceGuarantee;
    }

    public final void setSpaceGuarantee(final SpaceGuarantee spaceGuarantee) {
        this.spaceGuarantee = spaceGuarantee;
    }

    public final boolean isThinProvisioningSupported() {
        return thinProvisioningSupported;
    }

    public final void setThinProvisioningSupported(final boolean thinProvisioningSupported) {
        this.thinProvisioningSupported = thinProvisioningSupported;
    }

    public boolean isDedupEnabled() {
        return isDedupEnabled;
    }

    public void setDedupEnabled(final boolean dedupEnabled) {
        this.isDedupEnabled = dedupEnabled;
    }

    public float getDedupeSaving() {
        return dedupeSaving;
    }

    public void setDedupeSaving(final float dedupeSaving) {
        this.dedupeSaving = dedupeSaving;
    }

    public long getLastSnapshotTime() {
        return lastSnapshotTime;
    }

    public void setLastSnapshotTime(final long lastSnapshotTime) {
        this.lastSnapshotTime = lastSnapshotTime;
    }

    public InternalVolumeCapacity getCapacity() {
        return capacity;
    }

    public long getSnapshotCount() {
        return snapshotCount;
    }

    public void setSnapshotCount(final Long snapshotCount) {
        this.snapshotCount = snapshotCount;
    }

    public String getStatus() {
        return status;
    }

    public void setStatus(final String status) {
        this.status = status;
    }

    public String getCloneSourceName() {
        return cloneSourceName;
    }

    public void setCloneSourceName(final String cloneSourceName) {
        this.cloneSourceName = cloneSourceName;
    }

    public final String getType() {
        return type;
    }

    public boolean isThinProvisioned() {
        return thinProvisioned;
    }

    public void setThinProvisioned(final boolean thinProvisioned) {
        this.thinProvisioned = thinProvisioned;
    }

    /**
     * returns how the volume can use the SSD cache
     *
     * Expected values are: read, read-write, none, or NULL
     *
     * @return May return null.
     */
    public String getFlashPoolEligibility() {
        return flashPoolEligibility;
    }

    /**
     * @param flashPoolEligibility sets how the volume can use the SSD cache
     *                            Expected values are: read, read-write, none
     */
    public void setFlashPoolEligibility(final String flashPoolEligibility) {
        this.flashPoolEligibility = flashPoolEligibility;
    }

    public StoragePool getStoragePool() {
        return storagePool;
    }

    public String getVirtualStorage() {
        return virtualStorage;
    }

    public void setVirtualStorage(final String virtualStorage) {
        this.virtualStorage = virtualStorage;
    }

    public Qtrees getQtrees() {
        return qtrees;
    }

    public Qtree createQtree(final String name, final Qtree.Type type) throws SchemaException {
        return new Qtree(this, name, type);
    }

    @Override
    protected boolean hasChildren() {
        return true;
    }

    @Override
    protected void marshallChildren(MarshallingCache marshallingCache, Writer writer, int tabs) throws IOException, SchemaException {
        super.marshallChildren(marshallingCache, writer, tabs);

        if (qtrees.size() < 1)
            throw new SchemaException("InternalVolume should have at least one Qtree defined");

        addAllChildrenUnderNode(marshallingCache, writer, "Qtrees", qtrees, tabs);
    }

    protected void marshallAttributes(final MarshallingCache marshallingCache, final Writer writer) throws IOException, SchemaException {
        super.marshallAttributes(marshallingCache, writer);

        // mandatory
        addAttribute(writer, "Name", getName());
        addAttribute(writer, "StoragePoolId", storagePool.getStoragePoolId());
        addAttribute(writer, "Type", type);

        // optional
        addAttribute(writer, "SpaceGuarantee", spaceGuarantee);
        addAttribute(writer, "ThinProvisioned", thinProvisioned);
        addAttribute(writer, "ThinProvisioningSupported", thinProvisioningSupported);
        addAttribute(writer, "DedupeEnabled", isDedupEnabled);
        addAttribute(writer, "DedupeSavings", dedupeSaving);
        addAttribute(writer, "LastSnapshotTime", lastSnapshotTime);
        addAttribute(writer, "SnapshotCount", snapshotCount);
        addAttribute(writer, "Status", status);
        addAttribute(writer, "CloneSourceName", cloneSourceName);
        addAttribute(writer, "VirtualStorage", virtualStorage);
        addAttribute(writer, "ProtectionType", protectionType);
        addAttribute(writer, "FlashPoolEligibility", flashPoolEligibility);

        // Capacity
        addAttribute(writer, "SnapshotUsedCapacityMB", getCapacity().getSnapshotUsed(), Capacity.Unit.MB_1024);
        addAttribute(writer, "SnapshotAllocatedCapacityMB", getCapacity().getSnapshotAllocated(), Capacity.Unit.MB_1024);
        addAttribute(writer, "DataUsedCapacityMB", getCapacity().getDataUsed(), Capacity.Unit.MB_1024);
        addAttribute(writer, "DataAllocatedCapacityMB", getCapacity().getDataAllocated(), Capacity.Unit.MB_1024);
        addAttribute(writer, "TotalUsedCapacityMB", getCapacity().getTotalUsed(), Capacity.Unit.MB_1024);
        addAttribute(writer, "TotalUsedCapacityMB_2", getCapacity().getTotalUsedCapacityFromDevice(), Capacity.Unit.MB_1024);
        addAttribute(writer, "TotalAllocatedCapacityMB", getCapacity().getTotalAllocated(), Capacity.Unit.MB_1024);
        addAttribute(writer, "OtherUsedCapacityMB", getCapacity().getOtherUsedCapacity(), Capacity.Unit.MB_1024);
        addAttribute(writer, "OtherAllocatedCapacityMB", getCapacity().getOtherAllocatedCapacity(), Capacity.Unit.MB_1024);
        addAttribute(writer, "TotalCloneSavedCapacityMB", getCapacity().getTotalCloneSavedCapacityMB(), Capacity.Unit.MB_1024);
        addAttribute(writer, "RawToUsableRatio", usaLocaleNumberFormat.format(getStoragePool().getCapacity().getRawToUsableRatio()));
    }

    public String getProtectionType() {
        return protectionType;
    }

    public void setProtectionType(final String protectionType) {
        this.protectionType = protectionType;
    }

    public static enum SpaceGuarantee {
        /**
         * The storage guarantee space.
         */
        volume,
        /**
         * The storage partial guarantee space e.g. thin provisining
         */
        file,
        /**
         * no guarantee e.g. thin provisining
         */
        none,
        /**
         * for FlexCache only
         */
        partial
    }

    public final static class InternalVolumeKey implements Key<InternalVolumeKey> {
        private final Storage.StorageKey storageKey;
        private final String name;

        public InternalVolumeKey(final Storage.StorageKey storageKey, final String name) {
            this.storageKey = storageKey;
            this.name = name;
        }

        public final int compareTo(final InternalVolumeKey o) {
            final int i = getStorageKey().compareTo(o.getStorageKey());
            if (i != 0)
                return i;

            return getName().compareTo(o.getName());
        }

        public final boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final InternalVolumeKey volumeKey = (InternalVolumeKey) o;

            if (name != null ? !name.equals(volumeKey.name) : volumeKey.name != null) return false;
            //noinspection RedundantIfStatement
            if (storageKey != null ? !storageKey.equals(volumeKey.storageKey) : volumeKey.storageKey != null)
                return false;

            return true;
        }

        public final int hashCode() {
            int result;
            result = (storageKey != null ? storageKey.hashCode() : 0);
            result = 29 * result + (name != null ? name.hashCode() : 0);
            return result;
        }

        public final String toString() {
            return storageKey.toString() + ":" + getName();
        }

        public final String getName() {
            return name;
        }

        public final Storage.StorageKey getStorageKey() {
            return storageKey;
        }
    }
}