package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;
import com.onaro.util.Capacity;

import java.io.IOException;
import java.io.Writer;

/**
 * qtree is uniquely identified by internal volume and name
 */
@SuppressWarnings({"UnusedDeclaration"})
public class Qtree extends XMLNode<InternalVolume> {

    private final String name;
    private final Type type;

    private SecurityStyle securityStyle;
    private String status;
    private Boolean oplocks;

    // Represents quota "unlimited" limit values
    final static Long UNLIMITED_LONG = Long.MAX_VALUE;
    final static Capacity UNLIMITED_CAPACITY = Capacity.create(UNLIMITED_LONG, Capacity.Unit.BYTE);

    //Capacity values obtained from quota.
    private Capacity quotaHardCapacityLimit;
    private Capacity quotaSoftCapacityLimit;

    private Capacity quotaUsedCapacity;

    private QtreeKey key;

    Qtree(final InternalVolume parent, final String name, final Type type) throws SchemaException {
        super(parent, "Qtree");

        if (name == null || type == null)
            throw new SchemaException(getNodeName() + " Type requires non-null name and type");

        this.name = name;
        this.type = type;

        key = new QtreeKey(parent.getKey().getStorageKey(), parent.getName(), name);
    }

    @SuppressWarnings({"unchecked"})
    public final QtreeKey getKey() {
        return key;
    }

    public String getName() {
        return name;
    }

    public SecurityStyle getSecurityStyle() {
        return securityStyle;
    }

    public void setSecurityStyle(SecurityStyle securityStyle) {
        this.securityStyle = securityStyle;
    }

    public String getStatus() {
        return status;
    }

    public void setStatus(String status) {
        this.status = status;
    }

    public Boolean getOplocks() {
        return oplocks;
    }

    public void setOplocks(Boolean oplocks) {
        this.oplocks = oplocks;
    }

    /**
     * Get the Hard Capacity limit as defined on the Quota linked to this Qtree.  Should be in synch with the quota,
     * which will be of type 'tree'.   Null otherwise or if quota is not defined.
     *
     * @return hard capacity limit of linked quota object.
     */
    public Capacity getQuotaHardCapacityLimit() {
        return quotaHardCapacityLimit;
    }

    /**
     * Set the hard capacity limit from the linked quota object.   Package private - not to be used by clients directly.
     *
     * @param quotaHardCapacityLimit the hard capacity limit of the quota.
     */
    void setQuotaHardCapacityLimit(final Capacity quotaHardCapacityLimit) {
        this.quotaHardCapacityLimit = quotaHardCapacityLimit;
    }

    void setQuotaHardCapacityUnlimited() {
        this.quotaHardCapacityLimit = UNLIMITED_CAPACITY;
    }

    /**
     * Get the Soft Capacity limit as defined on the Quota linked to this Qtree.  Should be in synch with the quota,
     * which will be of type 'tree'.   Null otherwise or if quota is not defined.
     *
     * @return soft capacity limit of linked quota object.
     */
    public Capacity getQuotaSoftCapacityLimit() {
        return quotaSoftCapacityLimit;
    }

    /**
     * Set the soft capacity limit from the linked quota object.   Package private - not to be used by clients directly.
     *
     * @param quotaSoftCapacityLimit the soft capacity limit of the quota.
     */
    void setQuotaSoftCapacityLimit(final Capacity quotaSoftCapacityLimit) {
        this.quotaSoftCapacityLimit = quotaSoftCapacityLimit;
    }

    void setQuotaSoftCapacityUnlimited() {
        this.quotaSoftCapacityLimit = UNLIMITED_CAPACITY;
    }

    /**
     * Get the Used Capacity  as defined on the Quota linked to this Qtree.  Should be in synch with the quota,
     * which will be of type 'tree'.   Null otherwise or if quota is not defined.
     *
     * @return used capacity  of linked quota object.
     */
    public Capacity getQuotaUsedCapacity() {
        return quotaUsedCapacity;
    }

    /**
     * Set the used capacity from the linked quota object.   Package private - not to be used by clients directly.
     *
     * @param quotaUsedCapacity the used capacity limit of the quota.
     */
    void setQuotaUsedCapacity(final Capacity quotaUsedCapacity) {
        this.quotaUsedCapacity = quotaUsedCapacity;
    }

    public Type getType() {
        return type;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException {
        addAttribute(writer, "QtreeId", key.toString());

        // Changing behavior of DEFAULT qtree name for 20440, DEFAULT qtree has qtree name "-"

        //Server would like "name" of Qtree to be "" if it is the DEFAULT Qtree (i.e. it represents the internal volume).
        //This way shares in the UI appear to not have a Qtree.   To limit the inevitable mass confusion, we only
        //enforce this at the last minute when marshaling the ODATA.
        final String marshaledName =  (type == Type.DEFAULT) ? "-" : name;
        addAttribute(writer, "Name", marshaledName);
        addAttribute(writer, "SecurityStyle", securityStyle);
        addAttribute(writer, "Status", status);
        addAttribute(writer, "Oplocks", oplocks);
        addAttribute(writer, "Type", type);

        if (quotaHardCapacityLimit != null) {
            addAttribute(writer, "QuotaHardCapacityLimitMB", quotaHardCapacityLimit == UNLIMITED_CAPACITY ? "-1" : Capacity.valueForXML(quotaHardCapacityLimit, Capacity.Unit.MB_1024));
        }

        if (quotaSoftCapacityLimit != null) {
            addAttribute(writer, "QuotaSoftCapacityLimitMB", quotaSoftCapacityLimit == UNLIMITED_CAPACITY ? "-1" : Capacity.valueForXML(quotaSoftCapacityLimit, Capacity.Unit.MB_1024));
        }
        
        if (quotaUsedCapacity != null) {
            addAttribute(writer, "QuotaUsedCapacityMB", quotaUsedCapacity.value(Capacity.Unit.MB_1024));
        }

    }

    public StoragePool getStoragePool() {
        return getParent().getStoragePool();
    }

    public static enum SecurityStyle {
        unix,
        ntfs,
        mixed
    }

    public static enum Type {
        DEFAULT,
        EXPLICIT
    }

    public final static class QtreeKey implements Key<QtreeKey> {
        private final Storage.StorageKey storageKey;
        private final String internalVolumeName;
        private final String name;

        public QtreeKey(final Storage.StorageKey storageKey, final String internalVolumeName, final String name) throws SchemaException {
            this.storageKey = storageKey;
            this.internalVolumeName = internalVolumeName;
            this.name = name;
        }

        public String getName() {
            return name;
        }

        public Storage.StorageKey getStorageKey() {
            return storageKey;
        }

        public String getInternalVolumeName() {
            return internalVolumeName;
        }

        public String toString() {
            return storageKey.toString() + ":" + internalVolumeName + ":" + name;
        }

        @Override
        public boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            QtreeKey qtreeKey = (QtreeKey) o;

            if (internalVolumeName != null ? !internalVolumeName.equals(qtreeKey.internalVolumeName) : qtreeKey.internalVolumeName != null)
                return false;
            if (name != null ? !name.equals(qtreeKey.name) : qtreeKey.name != null) return false;
            if (storageKey != null ? !storageKey.equals(qtreeKey.storageKey) : qtreeKey.storageKey != null)
                return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = storageKey != null ? storageKey.hashCode() : 0;
            result = 31 * result + (internalVolumeName != null ? internalVolumeName.hashCode() : 0);
            result = 31 * result + (name != null ? name.hashCode() : 0);
            return result;
        }

        public int compareTo(final QtreeKey o) {
            int i;

            i = compareObjects(storageKey, o.getStorageKey());
            if (i != 0)
                return i;

            i = compareObjects(internalVolumeName, o.getInternalVolumeName());
            if (i != 0)
                return i;

            return compareObjects(name, o.getName());
        }
    }
}