package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;
import com.onaro.util.Capacity;

import java.io.IOException;
import java.io.Writer;

/**
 * quota is uniquely identified by storage, type, internalVolume, and qtree
 */
@SuppressWarnings({"UnusedDeclaration"})
public class Quota extends XMLNode<Storage> {

    private final QuotaType type;
    // The name of the internalVolume this quota is on
    private final InternalVolume internalVolume;
    // The qtree this quota refers to(only for tree quota)
    private final Qtree qtree;
    // The target for user or group quota
    private final String userOrGroupTarget;

    // Represents quota "unlimited" limit values
    final static Long UNLIMITED_LONG = Long.MAX_VALUE;
    public final static Capacity UNLIMITED_CAPACITY = Capacity.create(UNLIMITED_LONG, Capacity.Unit.BYTE);

    // Maximum number of files allowed for the quota target.
    private Long hardFileLimit = null;
    // Maximum number of files allowed for the quota target (Soft limit - gives warning to the user). 
    private Long softFileLimit = null;
    // Maximum amount of disk space, allowed for the quota target (Hard limit)
    private Capacity hardCapacityLimit = null;
    // Maximum amount of disk space, allowed for the quota target (Soft limit - gives warning to the user).
    private Capacity softCapacityLimit = null;
    // Disk space threshold, for the quota target.
    private Capacity threshold = null;
    // Number of files currently used
    private Long usedFiles = null;

    // Space in MB currently used
    private Capacity usedCapacity = null;

    private final QuotaKey key;

    Quota(final Storage parent, final QuotaType type, final InternalVolume internalVolume, final Qtree qtree, final String target) throws SchemaException {
        super(parent, "Quota");

        if (type == null || internalVolume == null)
            throw new SchemaException(getNodeName() + " Type requires non-null type, internalVolume name");

        this.type = type;
        this.internalVolume = internalVolume;
        this.qtree = qtree;

        // UserOrGroupTarget is null unless user or group type of quota have specific (i.e. not empty or "*") target
        if (type == QuotaType.tree)
            userOrGroupTarget = null;
        else if (target == null)
            userOrGroupTarget = null;
        else if (target.length() == 0)
            userOrGroupTarget = null;
        else
            userOrGroupTarget = target;

        key = new QuotaKey(parent.getKey(), type, internalVolume.getName(), qtree == null ? null : qtree.getName(), userOrGroupTarget);
    }

    @SuppressWarnings({"unchecked"})
    public final QuotaKey getKey() {
        return key;
    }

    public QuotaType getType() {
        return type;
    }

    public InternalVolume getInternalVolume() {
        return internalVolume;
    }

    public Qtree getQtree() {
        return qtree;
    }

    public String getUserOrGroupTarget() {
        return userOrGroupTarget;
    }

    public long getHardFileLimit() {
        return hardFileLimit;
    }

    public void setHardFileLimit(long hardFileLimit) {
        this.hardFileLimit = hardFileLimit;
    }

    public void setHardFileUnlimited() {
        this.hardFileLimit = UNLIMITED_LONG;
    }

    public long getSoftFileLimit() {
        return softFileLimit;
    }

    public void setSoftFileLimit(long softFileLimit) {
        this.softFileLimit = softFileLimit;
    }

    public void setSoftFileUnlimited() {
        this.softFileLimit = UNLIMITED_LONG;
    }

    public Capacity getHardCapacityLimit() {
        return hardCapacityLimit;
    }

    /**
     * Set hard capacity limit of quota.
     * <p/>
     * NOTE.  If this is a tree type quota and the qtree is defined, this ALSO sets the corresponding value on
     * the qtree object.   This is a restriction imposed by server requirements.
     *
     * @param hardCapacityLimit the hard capacity limit.
     */
    public void setHardCapacityLimit(final Capacity hardCapacityLimit) {
        this.hardCapacityLimit = hardCapacityLimit;
        if (type == QuotaType.tree && qtree != null) {
            qtree.setQuotaHardCapacityLimit(hardCapacityLimit);
        }
    }

    public void setHardCapacityUnlimited() {
        this.hardCapacityLimit = UNLIMITED_CAPACITY;
        if (type == QuotaType.tree && qtree != null) {
            qtree.setQuotaHardCapacityUnlimited();
        }
    }

    public Capacity getSoftCapacityLimit() {
        return softCapacityLimit;
    }

    /**
     * Set soft capacity limit of quota.
     * <p/>
     * NOTE.  If this is a tree type quota and the qtree is defined, this ALSO sets the corresponding value on
     * the qtree object.  This is a restriction imposed by server requirements.
     *
     * @param softCapacityLimit the soft capacity limit.
     */
    public void setSoftCapacityLimit(final Capacity softCapacityLimit) {
        this.softCapacityLimit = softCapacityLimit;
        if (type == QuotaType.tree && qtree != null) {
            qtree.setQuotaSoftCapacityLimit(softCapacityLimit);
        }
    }

    public void setSoftCapacityUnlimited() {
        this.softCapacityLimit = UNLIMITED_CAPACITY;
        if (type == QuotaType.tree && qtree != null) {
            qtree.setQuotaSoftCapacityUnlimited();
        }
    }

    public Capacity getThreshold() {
        return threshold;
    }

    public void setThreshold(final Capacity threshold) {
        this.threshold = threshold;
    }

    public void setThresholdUnlimited() {
        this.threshold = UNLIMITED_CAPACITY;
    }

    public long getUsedFiles() {
        return usedFiles;
    }

    public void setUsedFiles(final long usedFiles) {
        this.usedFiles = usedFiles;
    }

    public Capacity getUsedCapacity() {
        return usedCapacity;
    }

    /**
     * Set used capacity limit of quota.
     * <p/>
     * NOTE.  If this is a tree type quota and the qtree is defined, this ALSO sets the corresponding value on
     * the qtree object.    This is a restriction imposed by server requirements.
     *
     * @param usedCapacity the soft capacity limit.
     */
    public void setUsedCapacity(final Capacity usedCapacity) {
        this.usedCapacity = usedCapacity;
        if (type == QuotaType.tree && qtree != null) {
            qtree.setQuotaUsedCapacity(usedCapacity);
        }
    }

    @SuppressWarnings({"NumberEquality"})
    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException {

        addAttribute(writer, "QuotaId", getKey());
        addAttribute(writer, "Type", type);
        addAttribute(writer, "InternalVolumeName", internalVolume.getName());
        if (qtree != null)
            addAttribute(writer, "QtreeId", qtree.getKey());

        addAttribute(writer, "UserOrGroupTarget", userOrGroupTarget);
        addAttribute(writer, "HardFileLimit", hardFileLimit == UNLIMITED_LONG ? "-1" : hardFileLimit);
        addAttribute(writer, "SoftFileLimit", softFileLimit == UNLIMITED_LONG ? "-1" : softFileLimit);

        addAttribute(writer, "HardCapacityLimitMB", hardCapacityLimit == UNLIMITED_CAPACITY ? "-1" : Capacity.valueForXML(hardCapacityLimit, Capacity.Unit.MB_1024));
        addAttribute(writer, "SoftCapacityLimitMB", softCapacityLimit == UNLIMITED_CAPACITY ? "-1" : Capacity.valueForXML(softCapacityLimit, Capacity.Unit.MB_1024));

        addAttribute(writer, "ThresholdMB", threshold == UNLIMITED_CAPACITY ? "-1" : Capacity.valueForXML(threshold, Capacity.Unit.MB_1024));

        addAttribute(writer, "UsedFiles", usedFiles == UNLIMITED_LONG ? "-1" : usedFiles);

        addAttribute(writer, "UsedCapacityMB", usedCapacity == null ? "-1" : usedCapacity.value(Capacity.Unit.MB_1024));
    }

    public static enum QuotaType {
        user,
        group,
        tree
    }

    public final static class QuotaKey implements Key<QuotaKey> {
        private final QuotaType type;
        private final Storage.StorageKey storageKey;
        private final String internalVolumeName;
        private final String qtreeName;
        private final String userOrGroupTarget;

        public QuotaKey(final Storage.StorageKey storageKey, final QuotaType type, final String internalVolumeName, final String qtreeName, final String userOrGroupTarget) throws SchemaException {
            if (type == null || internalVolumeName == null)
                throw new SchemaException(" Quota Key requires non-null type and internalVolumeName");

            this.type = type;
            this.storageKey = storageKey;
            this.internalVolumeName = internalVolumeName;
            this.qtreeName = qtreeName;
            this.userOrGroupTarget = userOrGroupTarget;
        }

        public String toString() {
            return new StringBuilder().append(storageKey.toString()).append(":")
                    .append(type).append(":")
                    .append(internalVolumeName).append(":")
                    .append(qtreeName == null ? "" : qtreeName).append(":")
                    .append(userOrGroupTarget == null ? "" : userOrGroupTarget).toString();
        }

        public Storage.StorageKey getStorageKey() {
            return storageKey;
        }

        public String getInternalVolumeName() {
            return internalVolumeName;
        }

        public QuotaType getType() {
            return type;
        }

        public String getQtreeName() {
            return qtreeName;
        }

        public String getUserOrGroupTarget() {
            return userOrGroupTarget;
        }

        public int compareTo(final QuotaKey o) {
            int i;

            i = compareObjects(storageKey, o.getStorageKey());
            if (i != 0)
                return i;

            i = compareObjects(type, o.getType());
            if (i != 0)
                return i;

            i = compareObjects(internalVolumeName, o.getInternalVolumeName());
            if (i != 0)
                return i;

            i = compareObjects(userOrGroupTarget, o.getUserOrGroupTarget());
            if (i != 0)
                return i;

            return compareObjects(qtreeName, o.getQtreeName());
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            QuotaKey quotaKey = (QuotaKey) o;

            if (internalVolumeName != null ? !internalVolumeName.equals(quotaKey.internalVolumeName) : quotaKey.internalVolumeName != null)
                return false;
            if (qtreeName != null ? !qtreeName.equals(quotaKey.qtreeName) : quotaKey.qtreeName != null) return false;
            if (storageKey != null ? !storageKey.equals(quotaKey.storageKey) : quotaKey.storageKey != null)
                return false;
            if (type != quotaKey.type) return false;
            if (userOrGroupTarget != null ? !userOrGroupTarget.equals(quotaKey.userOrGroupTarget) : quotaKey.userOrGroupTarget != null)
                return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = type != null ? type.hashCode() : 0;
            result = 31 * result + (storageKey != null ? storageKey.hashCode() : 0);
            result = 31 * result + (internalVolumeName != null ? internalVolumeName.hashCode() : 0);
            result = 31 * result + (qtreeName != null ? qtreeName.hashCode() : 0);
            result = 31 * result + (userOrGroupTarget != null ? userOrGroupTarget.hashCode() : 0);
            return result;
        }
    }

}
