package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;

import java.io.IOException;
import java.io.Writer;

public class RemoteStorageSynchronization extends StorageSynchronization {
    private final String sourceStorageIP;
    private final String targetStorageIP;
    private final ProtocolEndPointRefs protocolEndPointRefs;

    private final RemoteStorageSynchronizationKey key;

    /**
     * Use {@link Storage#createRemoteStorageSynchronization} to create this object
     */
    RemoteStorageSynchronization(final Storage parent, final String sourceStorageIP, final String sourceVolume, final String targetStorageIP, final String targetVolume) throws SchemaException {
        super(parent, "StorageSynchronization", sourceVolume, targetVolume);

        if (sourceStorageIP == null || sourceStorageIP.length() == 0 || targetStorageIP == null || targetStorageIP.length() == 0)
            throw new SchemaException(getNodeName() + " Type requires non-null SourceStorageIP and TargetStorageIP");

        this.sourceStorageIP = sourceStorageIP;
        this.targetStorageIP = targetStorageIP;

        protocolEndPointRefs = new ProtocolEndPointRefs(this);

        key = new RemoteStorageSynchronizationKey(parent.getKey(), sourceStorageIP, sourceVolume, targetStorageIP, targetVolume);
    }

    public final RemoteStorageSynchronizationKey getKey() {
        return key;
    }

    public final ProtocolEndPointRefs getProtocolEndPointRefs() {
        return protocolEndPointRefs;
    }

    public final ProtocolEndPointRef createProtocolEndPointRef(final String id) throws SchemaException {
        return new ProtocolEndPointRef(this, id);
    }

    public final String getSourceStorageIP() {
        return sourceStorageIP;
    }

    public final String getTargetStorageIP() {
        return targetStorageIP;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException {
        super.marshallAttributes(cache, writer);
        addAttribute(writer, "SourceVolume", getSource());
        addAttribute(writer, "TargetVolume", getTarget());
        addAttribute(writer, "Mode", getMode()!= null ? getMode() : getModeEnum().name() );    // if mode is not explicitly set, take the enum
        addAttribute(writer, "ModeEnum", getModeEnum().name());
        addAttribute(writer, "State", getState()!= null ? getState() : getStateEnum().name());   // if state is not explicitly set, take the enum
        addAttribute(writer, "StateEnum", getStateEnum().name());
        addAttribute(writer, "SourceStorageIp", sourceStorageIP);
        addAttribute(writer, "TargetStorageIp", targetStorageIP);
        addAttribute(writer, "Technology", getTechnology());
    }

    protected void marshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        // When DR is off, ProtocolEndpoints are not reported so
        // these references will not point to elements in the odata.
        // Only add the protocol refs when dr is on.
        if (getParent().isDrPackageInstalled()) {
            addAllChildren(cache, writer, protocolEndPointRefs, tabs);
        }
    }

    protected boolean hasChildren() {
        return (!protocolEndPointRefs.isEmpty());
    }

    public static final class RemoteStorageSynchronizationKey implements Key<RemoteStorageSynchronizationKey> {
        private final Storage.StorageKey storageKey;
        private final String sourceStorageIP;
        private final String sourceVolume;
        private final String targetStorageIP;
        private final String targetVolume;

        public RemoteStorageSynchronizationKey(final Storage.StorageKey storageKey, final String sourceStorageIP, final String sourceVolume, final String targetStorageIP, final String targetVolume) {
            this.storageKey = storageKey;
            this.sourceStorageIP = sourceStorageIP;
            this.sourceVolume = sourceVolume;
            this.targetStorageIP = targetStorageIP;
            this.targetVolume = targetVolume;
        }

        public final int compareTo(final RemoteStorageSynchronizationKey o) {
            int i;

            // Must be the same volume
            i = storageKey.compareTo(o.getStorageKey());
            if (i != 0)
                return i;

            i = getSourceVolume().compareTo(o.getSourceVolume());
            if (i != 0)
                return i;

            i = getTargetVolume().compareTo(o.getTargetVolume());
            if (i != 0)
                return i;

            i = getSourceStorageIP().compareTo(o.getSourceStorageIP());
            if (i != 0)
                return i;

            return getTargetStorageIP().compareTo(o.getTargetStorageIP());
        }

        public final boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final RemoteStorageSynchronizationKey that = (RemoteStorageSynchronizationKey) o;

            if (sourceStorageIP != null ? !sourceStorageIP.equals(that.sourceStorageIP) : that.sourceStorageIP != null)
                return false;
            if (sourceVolume != null ? !sourceVolume.equals(that.sourceVolume) : that.sourceVolume != null)
                return false;
            if (storageKey != null ? !storageKey.equals(that.storageKey) : that.storageKey != null) return false;
            if (targetStorageIP != null ? !targetStorageIP.equals(that.targetStorageIP) : that.targetStorageIP != null)
                return false;
            if (targetVolume != null ? !targetVolume.equals(that.targetVolume) : that.targetVolume != null)
                return false;

            return true;
        }

        public final int hashCode() {
            int result;
            result = (storageKey != null ? storageKey.hashCode() : 0);
            result = 29 * result + (sourceStorageIP != null ? sourceStorageIP.hashCode() : 0);
            result = 29 * result + (sourceVolume != null ? sourceVolume.hashCode() : 0);
            result = 29 * result + (targetStorageIP != null ? targetStorageIP.hashCode() : 0);
            result = 29 * result + (targetVolume != null ? targetVolume.hashCode() : 0);
            return result;
        }

        public final String toString() {
            return storageKey.toString() + ":" + sourceStorageIP + ":" + sourceVolume + ":" + targetStorageIP + ":" + targetVolume;
        }

        public final Storage.StorageKey getStorageKey() {
            return storageKey;
        }

        public final String getSourceVolume() {
            return sourceVolume;
        }

        public final String getTargetVolume() {
            return targetVolume;
        }

        public final String getSourceStorageIP() {
            return sourceStorageIP;
        }

        public final String getTargetStorageIP() {
            return targetStorageIP;
        }
    }
}
