package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;
import com.onaro.util.Capacity;

import java.io.IOException;
import java.io.Writer;
import java.util.*;


@SuppressWarnings({"unchecked"})
public class Storage extends XMLNode<StorageRoot> {
    private final String ip;
    private String displayIP = null;
    private String name = null;
    private String manufacturer = null;
    private String model = null;
    private String family = null;
    private String serialNumber = null;
    private String microcodeVersion = null;
    private Boolean supportActiveActive;

    private String manageURL = null;
    private boolean virtual = false;

    private Long memoryMB;
    private Integer cpuCount;

    private final StorageControllers controllers;
    private final Volumes volumes;
    private final ProtocolEndPoints protocolEndPoints;
    private final NetworkPipes networkPipes;
    private final StorageSynchronizations<LocalStorageSynchronization> localSynchronizations;
    private final StorageSynchronizations<RemoteStorageSynchronization> remoteSynchronizations;

    private final StorageSynchronizations<LocalInternalVolumeStorageSynchronization> localInternalVolumeSynchronizations;
    private final StorageSynchronizations<RemoteInternalVolumeStorageSynchronization> remoteInternalVolumeSynchronizations;

    private final StorageSynchronizations<LocalQtreeStorageSynchronization> localQtreeSynchronizations;
    private final StorageSynchronizations<RemoteQtreeStorageSynchronization> remoteQtreeSynchronizations;

    private final Disks disks;
    private final StoragePools storagePools;
    private final AutoTierPolicies autoTierPolicies;
    private final BackEndLuns backEndLuns;

    private final ISCSINetworkPortalGroups iscsiNetworkPortalGroups;
    private final ISCSITargetNodes iscsiTargetNodes;
    private final DeviceCapabilities capabilities;
    private final InternalVolumes internalVolumes;
    private final ISCSIAuthenticationInfo iscsiAuthInfo;
    private final StorageCapacity capacity;
    private final FileShares fileShares;
    private final Quotas quotas;
    private final StorageEfficiencies storageEfficiencies;

    /**
     * Flag that states if any mapping was defined for this storage. This is important for integrity checks of the storage.
     * A storage with no mapping can indicate a problem with the sample (or an unconfigured storage).
     */
    private boolean mappingDefined = false;
    private boolean maskingDefined = false;

    private Boolean cluster = null;

    private final StorageKey key;

    /**
     * Use {@link StorageRoot#createStorage} to create this object
     */
    Storage(final StorageRoot parent, final String ip) throws SchemaException {
        super(parent, "Storage");

        if (ip == null || ip.length() == 0)
            throw new SchemaException(getNodeName() + " Type requires non-null oid and ip");

        this.ip = ip;

        // Make the default value of the display IP (the IP showed in the GUI) to be the IP (the stroage key). In data sources
        // where the IP cannot act as the key of the storage, we will change this display IP to reflect the actual IP address of
        // the storage (when applicable).
        this.displayIP = ip;

        controllers = new StorageControllers(this);
        volumes = new Volumes(this);
        protocolEndPoints = new ProtocolEndPoints(this);
        networkPipes = new NetworkPipes(this);
        localSynchronizations = new StorageSynchronizations<LocalStorageSynchronization>(this);
        remoteSynchronizations = new StorageSynchronizations<RemoteStorageSynchronization>(this);
        localInternalVolumeSynchronizations = new StorageSynchronizations<LocalInternalVolumeStorageSynchronization>(this);
        remoteInternalVolumeSynchronizations = new StorageSynchronizations<RemoteInternalVolumeStorageSynchronization>(this);
        localQtreeSynchronizations = new StorageSynchronizations<LocalQtreeStorageSynchronization>(this);
        remoteQtreeSynchronizations = new StorageSynchronizations<RemoteQtreeStorageSynchronization>(this);
        disks = new Disks(this);
        storagePools = new StoragePools(this);
        autoTierPolicies = new AutoTierPolicies(this);
        backEndLuns = new BackEndLuns(this);
        iscsiNetworkPortalGroups = new ISCSINetworkPortalGroups(this);
        iscsiTargetNodes = new ISCSITargetNodes(this);
        capabilities = new DeviceCapabilities(this);
        iscsiAuthInfo = new ISCSIAuthenticationInfo(this);

        key = new StorageKey(ip);
        capacity = new StorageCapacity();
        internalVolumes = new InternalVolumes(this);
        fileShares = new FileShares(this);
        quotas = new Quotas(this);
        storageEfficiencies = new StorageEfficiencies(this);
    }

    public StorageKey getKey() {
        return key;
    }

    public StorageRoot getRoot() {
        return getParent();
    }

    public boolean isDrPackageInstalled() {
        return getParent().isDrPackageInstalled();
    }

    public InternalVolumes getInternalVolumes() {
        return internalVolumes;
    }

    public Controller createController(final String wwn) throws SchemaException {
        return new Controller(this, wwn);
    }

    public LocalStorageSynchronization createLocalStorageSynchronization(final String sourceVolume, final String targetVolume) throws SchemaException {
        return new LocalStorageSynchronization(this, sourceVolume, targetVolume);
    }

    public RemoteStorageSynchronization createRemoteStorageSynchronization(final String sourceStorageIP, final String sourceVolume, final String targetStorageIP, final String targetVolume) throws SchemaException {
        return new RemoteStorageSynchronization(this, sourceStorageIP, sourceVolume, targetStorageIP, targetVolume);
    }

    public LocalInternalVolumeStorageSynchronization createLocalInternalVolumeStorageSynchronization(final String sourceInternalVolume, final String targetInternalVolume) throws SchemaException {
        return new LocalInternalVolumeStorageSynchronization(this, sourceInternalVolume, targetInternalVolume);
    }

    public RemoteInternalVolumeStorageSynchronization createRemoteInternalVolumeStorageSynchronization(final String sourceStorageIP, final String sourceInternalVolume, final String targetStorageIP, final String targetInternalVolume) throws SchemaException {
        return new RemoteInternalVolumeStorageSynchronization(this, sourceStorageIP, sourceInternalVolume, targetStorageIP, targetInternalVolume);
    }

    public AutoTierPolicy createAutoTierPolicy(final String id) throws SchemaException {
        return new AutoTierPolicy(this, id);
    }

    /**
     * Create a local (to this storage) qtree synchronization.
     *
     * @param sourceQtreeId souce qtree ID (as appears in ODATA).
     * @param targetQtreeId target qtree ID (as appears in ODATA).
     *
     * @return the synchronization element.
     *
     * @throws SchemaException for problems building schema.
     */
    public LocalQtreeStorageSynchronization createLocalQtreeStorageSynchronization(final String sourceQtreeId, final String targetQtreeId) throws SchemaException {
        return new LocalQtreeStorageSynchronization(this, sourceQtreeId, targetQtreeId);
    }

    /**
     * Create a remote (to this storage) qtree synchronization.
     *
     * @param sourceStorageIP the "IP" of the source storage.  As reported in odata.
     * @param sourceQtreeId   souce qtree ID (as appears in ODATA).
     * @param targetStorageIP the "IP" of the remote storage.  As reported in odata.
     * @param targetQtreeId   target qtree ID (as appears in ODATA).
     *
     * @return the synchronization element.
     *
     * @throws SchemaException for problems building schema.
     */
    public RemoteQtreeStorageSynchronization createRemoteQtreeStorageSynchronization(final String sourceStorageIP, final String sourceQtreeId, final String targetStorageIP, final String targetQtreeId) throws SchemaException {
        return new RemoteQtreeStorageSynchronization(this, sourceStorageIP, sourceQtreeId, targetStorageIP, targetQtreeId);
    }

    public ProtocolEndPoint createProtocolEndPoint(final String id, final String name, final String storageIP) throws SchemaException {
        return new ProtocolEndPoint(this, id, name, storageIP);
    }

    /**
     * Factory method for internal volumes.
     *
     * @param volumeName  Name of the internal volume.
     * @param storagePool StoragePool object on which internal volume sits.
     * @param type        Internal volume type: e.g. "FlexClone", "FlexVol", "traditional", "fileSystem".
     * @return Newly created internal volume
     * @throws SchemaException On any error
     */
    public InternalVolume createInternalVolume(final String volumeName, final StoragePool storagePool, final String type) throws SchemaException {
        return new InternalVolume(this, volumeName, storagePool, type);
    }

    public FileShare createFileShare(final String fileShareName, final String path, final Qtree qtree) throws SchemaException {
        return new FileShare(this, fileShareName, path, qtree);
    }

    public Quota createQuota(final Quota.QuotaType type, final InternalVolume internalVolume, final Qtree qtree, final String userOrGroupTarget) throws SchemaException {
        return new Quota(this, type, internalVolume, qtree, userOrGroupTarget);
    }

    public NetworkPipe createNetworkPipe(final String sourceID, final String targetID) throws SchemaException {
        return new NetworkPipe(this, sourceID, targetID);
    }

    /**
     * Create a volume that is based on an internal volume
     *
     * @param name  Name of volume
     * @param qtree Parent qtree (cannot be null!)
     * @return The new volume
     * @throws SchemaException On any error
     */
    public Volume createVolume(final String name, final Qtree qtree) throws SchemaException {
        if (qtree == null) {
            throw new SchemaException("qtree cannot be null!");
        }

        return new Volume(this, name, null, qtree, isDrPackageInstalled());
    }

    /**
     * Create a volume that is based on a storage pool
     *
     * @param name        Name of volume
     * @param storagePool Parent storage pool (cannot be null!)
     * @return The new volume
     * @throws SchemaException On any error
     */
    public Volume createVolume(final String name, final StoragePool storagePool) throws SchemaException {
        if (storagePool == null) {
            throw new SchemaException("StoragePool cannot be null!");
        }

        return new Volume(this, name, storagePool, null, isDrPackageInstalled());
    }

    /**
     * @param name
     * @return
     * @throws SchemaException
     *
     * @throws SchemaException
     * @deprecated Method provided only to allow Symmetrix CLI to set the storage pool
     *             after the volume is created. Should be removed when Symmetrix CLI is rewritten.
     *             <p/>
     *             Use {@link Storage#createVolume(String, Qtree)} or
     *             {@link Storage#createVolume(String, StoragePool)} )} for Volumes,
     *             {@link Storage#createMetaVolume(String)} for MetaVolumes,
     *             {@link MetaVolume#createVolumeMember(String, int, StoragePool)}
     *             or {@link MetaVolume#createVolumeMember(String, int, StoragePool)}
     *             for VolumeMembers.
     */
    public Volume createVolume(final String name) throws SchemaException {
        return new Volume(this, name, isDrPackageInstalled());
    }
    public MetaVolume createMetaVolume(final String name) throws SchemaException {
        return new MetaVolume(this, name, null, null, isDrPackageInstalled());
    }

    public Disk createDisk(final String diskID, final String name) throws SchemaException {
        return new Disk(this, diskID, name);
    }

    public StoragePool createRAID(final String name) throws SchemaException {
        final StoragePool raid = new StoragePool(this, name, StoragePool.StoragePoolType.RAID_GROUP);
        raid.setThinProvisioningSupported(false);
        raid.setIncludeInDwhCapacity(false);
        raid.setDedupeEnabled(false);

        return raid;
    }

    /**
     * Create a non-thin-provisioned backend storage pool to support Array Virtualization.
     */
    public StoragePool createBackendStoragePool(final String name) throws SchemaException {
        final StoragePool raid = new StoragePool(this, name, StoragePool.StoragePoolType.BACKEND_GROUP);
        raid.setThinProvisioningSupported(false);
        raid.setIncludeInDwhCapacity(false);
        raid.setDedupeEnabled(false);


        return raid;
    }

    /**
     * Create a NetApp aggregate.
     */
    public StoragePool createAggregate(final String name) throws SchemaException {
        final StoragePool aggregate = new StoragePool(this, name, StoragePool.StoragePoolType.AGGREGATE);
        aggregate.setThinProvisioningSupported(true);
        aggregate.setIncludeInDwhCapacity(true);

        return aggregate;
    }

    /**
     * Create a NetApp Flash Pool(an aggregate that uses SSD cache)
     */
    public StoragePool createFlashPool(final String name) throws SchemaException {
        final StoragePool aggregate = new StoragePool(this, name, StoragePool.StoragePoolType.FLASH_POOL);
        aggregate.setThinProvisioningSupported(true);
        aggregate.setIncludeInDwhCapacity(true);
        aggregate.setUsesFlashPools(true);
        return aggregate;
    }


    public StoragePool createThinProvisionedStoragePool(final String name) throws SchemaException {
        final StoragePool tpPool = new StoragePool(this, name, StoragePool.StoragePoolType.THIN_PROVISIONING);
        tpPool.setThinProvisioningSupported(true);
        tpPool.setIncludeInDwhCapacity(true);

        return tpPool;
    }

    public StoragePool createReservedStoragePool(final String name) throws SchemaException {
        final StoragePool reservedPool = new StoragePool(this, name, StoragePool.StoragePoolType.RESERVED);
        reservedPool.setThinProvisioningSupported(false);
        reservedPool.setIncludeInDwhCapacity(true);

        return reservedPool;
    }


    public ISCSINetworkPortalGroup createISCSIPortalGroup(final String portalGroupName, final String portalGroupTag) throws SchemaException {
        return new ISCSINetworkPortalGroup(this, portalGroupName, portalGroupTag);
    }

    public ISCSITargetNode createISCSITargetNode(final String iscsiNodeName) throws SchemaException {
        return new ISCSITargetNode(this, iscsiNodeName);
    }

    public DeviceCapability createDeviceCapability(final SupportedTechnology supportedTechnology) throws SchemaException {
        return new DeviceCapability(this, supportedTechnology.toString());
    }

    public DeviceCapabilities getCapabilities() {
        return capabilities;
    }

    public StorageCapacity getCapacity() {
        return capacity;
    }

    public ISCSITargetNodes getISCSINodes() {
        return iscsiTargetNodes;
    }

    public FileShares getFileShares() {
        return fileShares;
    }

    public Quotas getQuotas() {
        return quotas;
    }

    public StorageControllers getControllers() {
        return controllers;
    }

    public NetworkPipes getNetworkPipes() {
        return networkPipes;
    }

    public ProtocolEndPoints getProtocolEndPoints() {
        return protocolEndPoints;
    }

    public StorageSynchronizations<LocalStorageSynchronization> getLocalStorageSynchronizations() {
        return localSynchronizations;
    }

    public StorageSynchronizations<RemoteStorageSynchronization> getRemoteStorageSynchronizations() {
        return remoteSynchronizations;
    }

    public StorageSynchronizations<LocalInternalVolumeStorageSynchronization> getLocalInternalVolumeStorageSynchronizations() {
        return localInternalVolumeSynchronizations;
    }

    public StorageSynchronizations<RemoteInternalVolumeStorageSynchronization> getRemoteInternalVolumeStorageSynchronizations() {
        return remoteInternalVolumeSynchronizations;
    }

    public StorageSynchronizations<LocalQtreeStorageSynchronization> getLocalQtreeStorageSynchronizations() {
        return localQtreeSynchronizations;
    }

    public StorageSynchronizations<RemoteQtreeStorageSynchronization> getRemoteQtreeStorageSynchronizations() {
        return remoteQtreeSynchronizations;
    }

    public ISCSIAuthenticationInfo getAuthInfo() {
        return iscsiAuthInfo;
    }

    public Volumes getVolumes() {
        return volumes;
    }

    public Disks getDisks() {
        return disks;
    }

    public StoragePools getStoragePools() {
        return storagePools;
    }

    public AutoTierPolicies getAutoTierPolicies() {
        return autoTierPolicies;
    }

    public String getSerialNumber() {
        return serialNumber;
    }

    public void setSerialNumber(final String serialNumber) {
        this.serialNumber = serialNumber;
    }

    public String getName() {
        return name;
    }

    public void setName(final String name) {
        this.name = name;
    }

    public String getModel() {
        return model;
    }

    public void setModel(final String model) {
        this.model = model;
    }

    public String getFamily() {
        return family;
    }

    public void setFamily(final String family) {
        this.family = family;
    }

    public String getMicrocodeVersion() {
        return microcodeVersion;
    }

    public void setMicrocodeVersion(final String microcodeVersion) {
        this.microcodeVersion = microcodeVersion;
    }

    public String getManufacturer() {
        return manufacturer;
    }

    public void setManufacturer(final String manufacturer) {
        this.manufacturer = manufacturer;
    }

    public String getIP() {
        return ip;
    }

    public String getDisplayIP() {
        return displayIP;
    }

    public void setDisplayIP(final String displayIP) {
        this.displayIP = displayIP;
    }

    public String getManageURL() {
        return manageURL;
    }

    public void setManageURL(final String manageURL) {
        this.manageURL = manageURL;
    }

    public Boolean isSupportActiveActive() {
        return supportActiveActive;
    }

    /**
     * Specified if the storage supports active-active configurations (default is true)
     *
     * @param supportActiveActive
     */
    public void setSupportActiveActive(final Boolean supportActiveActive) {
        this.supportActiveActive = supportActiveActive;
    }

    /**
     * Sets if any mapping was defined for this storage. This is important for integrity checks of the storage.
     * A storage with no mapping can indicate a problem with the sample (or an unconfigured storage).
     */
    void setMappingDefined(final boolean mappingDefined) {
        this.mappingDefined = mappingDefined;
    }

    public Long getMemoryMB() {
        return memoryMB;
    }

    public void setMemoryMB(final Long memoryMB) {
        this.memoryMB = memoryMB;
    }

    public Integer getCpuCount() {
        return cpuCount;
    }

    public void setCpuCount(final Integer cpuCount) {
        this.cpuCount = cpuCount;
    }

    /**
     * Checks if any mapping was defined for this storage. This is important for integrity checks of the storage.
     * A storage with no mapping can indicate a problem with the sample (or an unconfigured storage).
     */
    public boolean isMappingDefined() {
        return mappingDefined;
    }

    /**
     * Sets if any masking was defined for this storage. This is important for integrity checks of the storage.
     * A storage with no masking can indicate a problem with the sample (or an unconfigured storage).
     */
    void setMaskingDefined(final boolean maskingDefined) {
        this.maskingDefined = maskingDefined;
    }

    /**
     * Checks if any masking was defined for this storage. This is important for integrity checks of the storage.
     * A storage with no masking can indicate a problem with the sample (or an unconfigured storage).
     */
    public boolean isMaskingDefined() {
        return maskingDefined;
    }

    /**
     * A virtual storage is like the IBM SVC - a server that virtualize backend storage. As far as the back end storage
     * devices are concerned, this is a host, but the hosts consider it as a storage.
     *
     * @return true if this storage is virtual
     */
    public boolean isVirtual() {
        return virtual;
    }

    /**
     * For details, see {@link #isVirtual}
     */
    public void setVirtual(final boolean virtual) {
        this.virtual = virtual;
    }

    public Boolean getCluster() {
        return cluster;
    }

    public void setCluster(final Boolean cluster) {
        this.cluster = cluster;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException, SchemaException {
        addAttribute(writer, "IP", ip);
        addAttribute(writer, "DisplayIP", displayIP);
        addAttribute(writer, "Name", name);
        addAttribute(writer, "Manufacturer", manufacturer);
        addAttribute(writer, "Model", model);
        addAttribute(writer, "Family", family);
        addAttribute(writer, "SerialNumber", serialNumber);
        addAttribute(writer, "MicrocodeVersion", microcodeVersion);
        // Server wanted to have RawCapacity attribute outside capacity tag as well
        addAttribute(writer, "Virtual", virtual);
        addAttribute(writer, "ManageURL", manageURL);
        addAttribute(writer, "SupportActiveActive", supportActiveActive);

        addAttribute(writer, "CpuCount", cpuCount);
        addAttribute(writer, "MemoryMB", memoryMB);

        addAttribute(writer, "Cluster", cluster);

        // <!-- Raw capacity in MB. default is Capacity.MIN_VALUE (for backward compatibility)-->
        addAttribute(writer, "RawCapacityMB", getCapacity().getRawCapacity(), Capacity.Unit.MB_1024);
        // <!-- Raw capacity of spare disks. default is -1 -->
        addAttribute(writer, "SpareRawCapacityMB", getCapacity().getSpareRaw(), Capacity.Unit.MB_1024);
        //<!-- Raw capacity of failed disks. default is -1 -->
        addAttribute(writer, "FailedRawCapacityMB", getCapacity().getFailedRaw(), Capacity.Unit.MB_1024);

    }

    /**
     * We create a comparator used to sort InternalVolumes differently
     * than we sort using the InternalVolumeKey objects. We need to ensure
     * that IVs that are clones (have clone sources) appear after their
     * source IVs in the marshalling. To get a globally consistent ordering
     * we order by checking the following in order of precedence:
     * First, a clone comes after a non-clone
     * Then, a clone comes after its source internal volume
     * Finally, if nothing above applies, check the InternalVolumeKey ordering
     */
    static class InternalVolumeComparator implements Comparator<InternalVolume> {

        @Override
        public int compare(final InternalVolume vol1, final InternalVolume vol2) {

            // clones come after non-clones
            if (vol1.getCloneSourceName() != null && vol2.getCloneSourceName() == null) {
                return 1;   // vol1 will be considered larger
            }
            if (vol1.getCloneSourceName() == null && vol2.getCloneSourceName() != null) {
                return -1;   // vol1 will be considered smaller
            }

            // now order clones in case we have clones of clones
            if (vol1.getCloneSourceName() != null && vol1.getCloneSourceName().equals(vol2.getName())) {
                return 1;   // vol1 will be considered larger than its source
            }
            if (vol2.getCloneSourceName() != null && vol2.getCloneSourceName().equals(vol1.getName())) {
                return -1;   // vol1 will be considered smaller than its clone
            }

            // otherwise, we just compare keys
            return vol1.getKey().compareTo(vol2.getKey());

        }
    }

    /**
     * The purpose of this method is to arrange the internal volumes in such a way that
     * all the clones come after their parent clones/internal volumes.
     * The comparator did not work so we had to write this logic to order them as
     * server expects a InternalVolume which is a clone should come after
     * its Partent Internal Volume.
     * <p/>
     * This method uses breadh first search algorithm to sort the internal volumes.
     * 1. Put all the parents in the Linked Hash Map first
     * 2. And put all the clones in one Linked List in the same interation as above
     * 3. Now Iterate through all the clones till the Clones List is empty
     * 4. For each clone, if the parent is already in teh map, then put the clone in the map
     * and remove it from the clones list.
     * <p/>
     * The worst case performance of this algorithm is O(n-square).
     * But assuming that we do not have more than 200 internal volumes for a single array,
     * it looks ok.
     *
     * @return
     */
    public Map<String, InternalVolume> sortInternalVolumes() {
        final Map<String, InternalVolume> sortedInternalVols = new LinkedHashMap<String, InternalVolume>();
        final List<InternalVolume> clones = new ArrayList<InternalVolume>();

        //add all the non-clones to the sorted volumes list first.
        //and prepare the list of clones seperately.
        for (final InternalVolume internalVol : internalVolumes) {
            if (internalVol.getCloneSourceName() == null) {
                sortedInternalVols.put(internalVol.getName(), internalVol);
            } else {
                clones.add(internalVol);
            }
        }


        while (clones.size() > 0) {
            for (int index = 0; index < clones.size(); index++) {

                final InternalVolume internalVol = clones.get(index);
                final InternalVolume parentVolume = sortedInternalVols.get(internalVol.getCloneSourceName());
                //Add the clone to the sorted internal volumes list only after the parent is in the list.
                if (parentVolume != null) {
                    sortedInternalVols.put(internalVol.getName(), internalVol);
                    clones.remove(index);
                    index--;
                }
            }
        }
        return sortedInternalVols;

    }

    protected void marshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        addAllChildrenUnderNode(cache, writer, "Capabilities", capabilities.values(), tabs);

        addAllChildren(cache, writer, controllers, tabs);

        addAllChildrenUnderNode(cache, writer, "Disks", disks.values(), tabs);
        addAllChildrenUnderNode(cache, writer, "BackendLuns", backEndLuns, tabs);
        addAllChildrenUnderNode(cache, writer, "AutoTierPolicies", autoTierPolicies.values(), tabs);
        addAllChildrenUnderNode(cache, writer, "StoragePools", storagePools.values(), tabs);
        
        final Map<String, InternalVolume> sortedInternalVolumes = sortInternalVolumes();
        addAllChildrenUnderNode(cache, writer, "InternalVolumes", sortedInternalVolumes.values(), tabs);

        addAllChildrenUnderNode(cache, writer, "Volumes", volumes.values(), tabs);

        // Local and Remote replicas (volume, internal volume qtree) should be marshaled without dr license
        // (this was changed to enable proper chargeback
        // for protection capacity in DWH)

        final List<StorageSynchronization> allSynchronizations = new LinkedList<StorageSynchronization>();
        if (isDrPackageInstalled()) {
            // The protocol end-points and Network pipes are only added to the XML in case the DR license is installed.
            addAllChildrenUnderNode(cache, writer, "ProtocolEndPoints", protocolEndPoints, tabs);
            addAllChildrenUnderNode(cache, writer, "NetworkPipes", networkPipes, tabs);
        }

        allSynchronizations.addAll(remoteSynchronizations);
        allSynchronizations.addAll(localSynchronizations);
        addAllChildrenUnderNode(cache, writer, "StorageSynchronizations", allSynchronizations, tabs);

        final List<InternalVolumeStorageSynchronization> allInternalVolumeSynchronizations = new LinkedList<InternalVolumeStorageSynchronization>();
        allInternalVolumeSynchronizations.addAll(remoteInternalVolumeSynchronizations);
        allInternalVolumeSynchronizations.addAll(localInternalVolumeSynchronizations);
        addAllChildrenUnderNode(cache, writer, "InternalVolumeStorageSynchronizations", allInternalVolumeSynchronizations, tabs);

        final List<QtreeStorageSynchronization> allQtreeSynchronizations = new LinkedList<QtreeStorageSynchronization>();
        allQtreeSynchronizations.addAll(remoteQtreeSynchronizations);
        allQtreeSynchronizations.addAll(localQtreeSynchronizations);
        addAllChildrenUnderNode(cache, writer, "QtreeStorageSynchronizations", allQtreeSynchronizations, tabs);

        addAllChildrenUnderNode(cache, writer, "ISCSINetworkPortalGroups", iscsiNetworkPortalGroups, tabs);
        addAllChildrenUnderNode(cache, writer, "ISCSINodes", iscsiTargetNodes, tabs);
        addAllChildrenUnderNode(cache, writer, "ISCSISecurityInfos", iscsiAuthInfo, tabs);
        addAllChildrenUnderNode(cache, writer, "FileShares", fileShares, tabs);
        addAllChildrenUnderNode(cache, writer, "Quotas", quotas, tabs);
        addAllChildrenUnderNode(cache, writer, "StorageEfficiencies", storageEfficiencies.values(), tabs);
    }

    public ISCSINetworkPortalGroup creatISCSITargetPortal(final String portalGroupName, final String portalGroupTag) throws SchemaException {
        return new ISCSINetworkPortalGroup(this, portalGroupName, portalGroupTag);
    }

    public BackEndLun createBackEndLun(final String name) throws SchemaException {
        return new BackEndLun(this, name);
    }

    public ISCSIAuthentication createISCSIAuthentication(final String initiatorNodeName, final String targetNodeName) throws SchemaException {
        return new ISCSIAuthentication(this, initiatorNodeName, targetNodeName);
    }

    public StorageEfficiency createStorageEfficiency(final String technology) throws SchemaException {
        return new StorageEfficiency(this, technology);
    }

    public StorageEfficiencies getStorageEfficiencies() {
        return storageEfficiencies;
    }

    public BackEndLuns getBackEndLuns() {
        return backEndLuns;
    }

    public ISCSINetworkPortalGroups getISCSIPortalGroups() {
        return iscsiNetworkPortalGroups;
    }

    protected boolean hasChildren() {
        return (!controllers.isEmpty() ||
                volumes.size() > 0 ||
                disks.size() > 0 ||
                protocolEndPoints.size() > 0 ||
                networkPipes.size() > 0 ||
                localSynchronizations.size() > 0 ||
                remoteSynchronizations.size() > 0 ||
                backEndLuns.size() > 0 ||
                iscsiNetworkPortalGroups.size() > 0 ||
                iscsiTargetNodes.size() > 0 ||
                capabilities.size() > 0 ||
                storagePools.size() > 0 ||
                internalVolumes.size() > 0);
    }

    public static class StorageKey implements Key<StorageKey> {
        private String ip;

        @SuppressWarnings({"RedundantIfStatement"})
        public boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final StorageKey that = (StorageKey) o;

            if (ip != null ? !ip.equals(that.ip) : that.ip != null) return false;

            return true;
        }

        public int hashCode() {
            return (ip != null ? ip.hashCode() : 0);
        }

        public StorageKey(final String ip) {
            this.ip = ip;
        }

        public int compareTo(final StorageKey o) {
            return getIP().compareTo(o.getIP());
        }

        public String getIP() {
            return ip;
        }

        public String toString() {
            return ip;
        }
    }

    /**
     * Known storage technologies. Each storage reports about which technologies it supports (today's storages can support
     * several technologies at the same time (ISCSI, FC & NAS for NetApp filers)
     */
    public static enum SupportedTechnology {
        FC,
        ISCSI {
            public String toString() {
                return "iSCSI";
            }
        },
        FCOE,
        FCIP,
        NFS,
        NAS,
        CIFS
    }
}
