package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;

import java.io.IOException;
import java.io.Writer;

/**
 * <p>Abstract parent class for storage based synchronizations.
 *
 * <p>Note the children of this class currently extend across several different XSD schema types.   Thus this class
 * does not marshal anything directly, but rather leaves this to the childern.   Further, the "schemaName" is
 * passed up from the sub-classes before being passed up to XML Node.
 *
 * <p>This class should only have the very generic attributes that extend across all these schemas (not necessarily
 * with the same formal XSD attribute names, but with the same data intent).
 *
 * <p>The only limitation here is that all derived schemas be under the Storage element.
 */
public abstract class StorageSynchronization extends XMLNode<Storage> {

    private final String source;
    private final String target;
    private String mode = null;
    private StorageSynchronizationMode modeEnum = StorageSynchronizationMode.UNKNOWN;
    private String state = null;
    private StorageSynchronizationState stateEnum = StorageSynchronizationState.UNKNOWN;
    private String technology = "";     // empty string is the default

    /**
     * Create a new storage synchronization.
     *
     * @param parent  parent storage object.
     * @param schemaName  the schema name to use based on child's implementation.
     * @param source  the source entity name/id
     * @param target the target entity name/id
     * @throws SchemaException for internal problems building model
     */
    protected StorageSynchronization(final Storage parent,
                                     final String schemaName,
                                     final String source,
                                     final String target) throws SchemaException {
        super(parent, schemaName);

        if (source == null || source.length() == 0 || target == null || target.length() == 0)
            throw new SchemaException(getNodeName() + " Type requires non-null Source and Target");

        this.source = source;
        this.target = target;
    }

    /**
     * Get a handle to the parent storage object.
     *
     * @return the storage object.
     */
    public final Storage getStorage() {
        return getParent();
    }

    /**
     * Get the synchronization mode for this synchronization.  See enum for possible values.    Optional - may be null.
     *
     * @return get the synchronization mode - may be null.
     */
    public final String getMode() {
        return mode;
    }

    /**
     * Set the synchronization mode for this synchronization.  See enum for possible values.    Optional.
     *
     * @param mode the synchronization mode/
     */
    public final void setMode(String mode) {
        this.mode = mode;
    }

    public StorageSynchronizationMode getModeEnum() {
        return modeEnum;
    }

    public void setModeEnum(final StorageSynchronizationMode modeEnum) {
        this.modeEnum = modeEnum;
    }

    public String getState() {
        return state;
    }

    public void setState(final String state) {
        this.state = state;
    }

    public StorageSynchronizationState getStateEnum() {
        return stateEnum;
    }

    public void setStateEnum(final StorageSynchronizationState stateEnum) {
        this.stateEnum = stateEnum;
    }

    public String getTechnology() {
        return technology;
    }

    public void setTechnology(final String technology) {
        this.technology = technology;
    }

    /**
     * Get the source of this synchronization.   This may have slightly different meanings based on the most
     * derived schema of the children.   For instance, may be a volume name, internal volume, qtree id, etc.
     *
     * @return the name/id of the source entity.
     */
    public final String getSource() {
        return source;
    }

    /**
     * Get the target of this synchronization.   This may have slightly different meanings based on the most
     * derived schema of the children.   For instance, may be a volume name, internal volume, qtree id, etc.
     *
     * @return the name/id of the target entity.
     */
    public final String getTarget() {
        return target;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException {
        //No-op for now.
        //Sub classes handle marshaling as they may name source, target, etc. by different names.
        //Each sub-class may have a slightly different schema to implement against.
    }

 
}
