package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;

import java.io.IOException;
import java.io.Writer;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Set;

public abstract class SwitchType<P> extends XMLNode<P> {
    private String ip = null;
    private String firmwareVersion = null;
    private String manufacturer = null;
    private String model = null;
    private String name = null;
    private Role switchRole = null;
    private String serialNumber = null;
    private String manageURL = null;

    private final SwitchPorts ports;

    private final SwitchTypeKey key;
    protected Type switchType = null;
    protected State switchState = null;
    protected Status switchStatus = null;
    protected String switchTimeZone = null;

    protected SwitchType(final P parent, final String elementName, final String wwn, final Type switchType) throws SchemaException {
        super(parent, elementName);

        if (wwn == null || wwn.length() == 0)
            throw new SchemaException(getNodeName() + " Type requires non-null oid and wwn");

        this.switchType = switchType;

        ports = new SwitchPorts(this);

        key = new SwitchTypeKey(wwn.toUpperCase());
    }

    @SuppressWarnings("unchecked")
    public final SwitchTypeKey getKey() {
        return key;
    }

    public final SwitchPort createPort(final String wwn) throws SchemaException {
        return new SwitchPort(this, wwn);
    }

    public final String getFirmwareVersion() {
        return firmwareVersion;
    }

    public final void setFirmwareVersion(final String firmwareversion) {
        this.firmwareVersion = firmwareversion;
    }

    public final String getIP() {
        return ip;
    }

    public final void setIP(final String ip) {
        this.ip = ip;
    }

    public final String getManufacturer() {
        return manufacturer;
    }

    @SuppressWarnings("UnusedDeclaration")
    public final String getManageURL() {
        return manageURL;
    }

    public final void setManageURL(final String manageURL) {
        this.manageURL = manageURL;
    }

    public final void setManufacturer(final String manufacturer) {
        this.manufacturer = manufacturer;
    }

    public final String getModel() {
        return model;
    }

    public final void setModel(final String model) {
        this.model = model;
    }

    public final String getName() {
        return name;
    }

    public final void setName(final String name) {
        this.name = name;
    }

    public final String getSerialNumber() {
        return serialNumber;
    }

    public final void setSerialNumber(final String serialNumber) {
        this.serialNumber = serialNumber;
    }

    public final Role getSwitchRole() {
        return switchRole;
    }

    public final void setSwitchRole(final Role switchrole) {
        this.switchRole = switchrole;
    }

    public final String getWWN() {
        return key.wwn;
    }

    public Type getSwitchType() {
        return switchType;
    }

    public final SwitchPorts getPorts() {
        return ports;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException {
        addAttribute(writer, "WWN", key.wwn);
        addAttribute(writer, "IP", ip);
        addAttribute(writer, "FirmwareVersion", firmwareVersion);
        addAttribute(writer, "Manufacturer", manufacturer);
        addAttribute(writer, "Model", model);
        addAttribute(writer, "Name", name);
        addAttribute(writer, "SwitchRole", switchRole);
        addAttribute(writer, "Type", switchType);
        addAttribute(writer, "SerialNumber", serialNumber);
        addAttribute(writer, "ManageURL", manageURL);

        addAttribute(writer, "SwitchState", switchState);
        addAttribute(writer, "SwitchStatus", switchStatus);
        addAttribute(writer, "SwitchTimeZone", switchTimeZone);

    }

    /**
     * HEADSUP: The marshling of the ports has to take into account that the port list can change during the iteration.
     * This can happen when marshalling an arbitrated loop (which currently adds additional ports to the switch).
     * The problem that we will encounter is a concurrent exception when trying to change (add) the list during iteration.
     * To cope with that, we iterate first on a local copy of the ports and then check if ports were added and iterate over
     * them too.
     *
     * @param writer - odata xml writer
     * @param tabs - for indenting properly in the odata xml
     *
     * @throws IOException
     * @throws SchemaException
     */
    protected void marshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        final SwitchPorts marshalledPorts = new SwitchPorts(this);

        while (marshalledPorts.size() < ports.size()) {
            // Create a list of ports that were not marshalled yet
            final LinkedHashSet<SwitchPort> yetToBeMarshalledPorts = new LinkedHashSet<SwitchPort>();
            yetToBeMarshalledPorts.addAll(ports);
            yetToBeMarshalledPorts.removeAll(marshalledPorts);

            for (final SwitchPort port : yetToBeMarshalledPorts)
                port.marshall(cache, writer, tabs);

            marshalledPorts.addAll(yetToBeMarshalledPorts);
        }
    }

    protected boolean hasChildren() {
        return (!ports.isEmpty());
    }

    public final State getSwitchState() {
        return switchState;
    }

    public final void setSwitchState(final State switchstate) {
        this.switchState = switchstate;
    }

    @SuppressWarnings("UnusedDeclaration")
    public final Status getSwitchStatus() {
        return switchStatus;
    }

    public final void setSwitchStatus(final Status switchstatus) {
        this.switchStatus = switchstatus;
    }

    @SuppressWarnings("UnusedDeclaration")
    public final String getSwitchTimeZone() {
        return switchTimeZone;
    }

    public final void setSwitchTimeZone(final String switchtimezone) {
        this.switchTimeZone = switchtimezone;
    }

    public static enum ZoningEnforcement {
        UNKNOWN(0),
        SOFTWARE(1),
        HARDWARE(2);

        private final int e;

        ZoningEnforcement(final int e) {
            this.e = e;
        }

        @SuppressWarnings("UnusedDeclaration")
        public int toEnum() {
            return e;
        }

        public String toString() {
            return Integer.toString(e);
        }
    }

    public static enum Role {
        /**
         * Switch is the principal switch in the fabric
         */
        PRINCIPAL(566),
        /**
         * Switch is a subordinate switch in the fabric
         */
        SUBORDINATE(567);

        private final int e;

        Role(final int e) {
            this.e = e;
        }

        @SuppressWarnings("UnusedDeclaration")
        public int toEnum() {
            return e;
        }

        public String toString() {
            return Integer.toString(e);
        }
    }

    public static enum Type {
        /**
         * Physical switch that is not associated with a fabric.
         * Typically it supports logical switches,
         * each associated with a different fabric.
         */
        CHASSIS,

        /**
         * Logical switch associated with a fabric. It must have
         * and reference a chassis switch.
         */
        FABRIC_LOGICAL,

        /**
         * Physical switch that is entirely associated with a single fabric.
         * It does not need to reference a chassis switch.
         */
        FABRIC_PHYSICAL,
        
        /**
         * Physical (Cisco) NPV or (Brocade) Access Gateway switch.
         * These support one or more logical NPV switches, so they
         * will not be associated with a fabric.
         * 
         */
        NPV_CHASSIS,

        /**
         * Logical (Cisco) NPV or (Brocade) Access Gateway switch.
         * These are associated with a single fabric and must have and
         * reference an NPV_CHASSIS switch.
         *
         */
        NPV_LOGICAL,

        /**
         * Physical (Cisco) NPV or (Brocade) Access Gateway switch that is
         * associated with a single fabric. It does not need to reference
         * an  NPV_CHASSIS switch. These appear as physical switches
         * associated with a single fabric.
         *
         * This type is associated with switches that are generated when we see
         * a fabric switch port that connects to a (Cisco) NPV
         * or (Brocade) Access Gateway.
         *
         */
        NPV_PHYSICAL,

        /**
         * Switch to represent NPIV connectivity to a fabric switch port.
         * It does not need to reference an  NPV_CHASSIS switch.
         * These appear as physical switches associated with a single fabric.
         *
         * This type is associated with switches that are generated when we see
         * a fabric switch port that connects to more than one node.
         *
         */
        NPV_HUB
    }

    public static enum State {
        UNKNOWN(560),
        ONLINE(561),
        OFFLINE(562),
        TESTING(563),
        COLDREBOOT(5),
        WARMREBOOT(6);

        private final int e;

        State(final int e) {
            this.e = e;
        }

        @SuppressWarnings("UnusedDeclaration")
        public int toEnum() {
            return e;
        }

        public String toString() {
            return Integer.toString(e);
        }
    }

    public static enum Status {
        UNKNOWN(1),
        NOTSUP(2),
        OK(3),
        WARNING(4),
        FAULTY(5),
        SEC_BUSY(6),
        SEC_ERROR(7);

        private final int e;

        Status(final int e) {
            this.e = e;
        }

        @SuppressWarnings("UnusedDeclaration")
        public int toEnum() {
            return e;
        }

        public String toString() {
            return Integer.toString(e);
        }
    }

    /**
     * The type of domain ID specified for this switch. If type is <code>static</code>, then any conflict on domain IDs
     * when trying merge fabrics will result with isolation. If type is <code>preferred</code> then the final domain ID
     * may change when merging fabrics.
     */
    public enum DomainIdType {
        STATIC(0),
        PREFERRED(1);

        private final int e;

        DomainIdType(final int e) {
            this.e = e;
        }

        @SuppressWarnings("UnusedDeclaration")
        public int toEnum() {
            return e;
        }

        public String toString() {
            return Integer.toString(e);
        }
    }

    public final static class SwitchTypeKey implements Key<SwitchTypeKey> {
        final String wwn;

        public SwitchTypeKey(final String wwn) {
            this.wwn = wwn;
        }

        public int compareTo(final SwitchTypeKey o) {
            return getWWN().compareTo(o.getWWN());
        }

        public boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final SwitchTypeKey that = (SwitchTypeKey) o;

            if (wwn != null ? !wwn.equals(that.wwn) : that.wwn != null) return false;

            return true;
        }

        public int hashCode() {
            return (wwn != null ? wwn.hashCode() : 0);
        }

        public String toString() {
            return wwn;
        }

        public String getWWN() {
            return wwn;
        }
    }
}