package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;
import com.onaro.util.Capacity;

import java.io.IOException;
import java.io.Writer;
import java.util.Collection;
import java.util.Set;
import java.util.TreeSet;

/**
 * A root level element representing a virtual machine and it's association with the virtual disks and the
 * containing datastore
 */
public class VirtualMachine extends XMLNode<VirtualizationRoot> {
    private final String moid;
    private final String managementIp;

    private String name;
    private String dnsName;
    private String os;
    private Integer processors;
    private Integer memoryInMB;
    private long powerStateChangeTime;
    private PowerState powerState;
    private String guestState;
    private DataStore.DataStoreKey dataStoreKey;
    private Host.HostKey host;
    private Capacity provisionedCapacity;
    private Capacity usedCapacity;

    private final Set<String> ips = new TreeSet<String>();
    private final VirtualMachineKey key;

    VirtualMachine(final VirtualizationRoot parent, final String moid, final String managementIp) throws SchemaException {
        super(parent, "HV_VirtualMachine");

        if (moid == null || moid.length() == 0)
            throw new SchemaException("moid is mandatory!");

        this.moid = moid;
        this.managementIp = managementIp;
        this.key = new VirtualMachineKey(managementIp, moid);
    }

    public String getName() {
        return name;
    }

    /**
     * A unique ID for this VM in the scope of the management IP.
     *
     * @return the unique ID of the VM in the scope of the management IP.
     */
    public String getMOID() {
        return moid;
    }

    /**
     * The IP of the management "station" of this VM.   For instance, this will be the virtual center IP in
     * VMWare and the HMC address for PowerVM, and so on.
     *
     * @return  the management station IP for this VM.
     */
    public String getManagementIp() {
        return managementIp;
    }

    public String getDnsName() {
        return dnsName;
    }

    public Set<String> getIPs() {
        return ips;
    }

    public String getOS() {
        return os;
    }

    public int getProcessors() {
        return processors;
    }

    public int getMemoryInMB() {
        return memoryInMB;
    }

    public DataStore.DataStoreKey getDataStoreKey() {
        return dataStoreKey;
    }

    public void setDataStoreKey(DataStore.DataStoreKey dataStoreKey) {
        this.dataStoreKey = dataStoreKey;
    }

    /**
     * Get the "power" state of this VM.   This may be different than "guest" state, but not necessarily.   Power
     * may represent a more fundamental operational state of the VM (where available).
     *
     * @return the power state as an enum.
     */
    public PowerState getPowerState() {
        return powerState;
    }

    public long getPowerStateChangeTime() {
        return powerStateChangeTime;
    }

    public Host.HostKey getHost() {
        return host;
    }

    public void setHost(Host.HostKey host) {
        this.host = host;
    }

    public void setName(final String name) {
        this.name = name;
    }

    public void setDnsName(final String dnsName) {
        this.dnsName = dnsName;
    }

    public void setOs(final String os) {
        this.os = os;
    }

    public void setProcessors(final Integer processors) {
        this.processors = processors;
    }

    public void setMemoryInMB(final Integer memoryInMb) {
        this.memoryInMB = memoryInMb;
    }

    public void setPowerState(final PowerState powerState) {
        this.powerState = powerState;
    }

    public void setPowerStateChangeTime(final long powerStateChangeTime) {
        this.powerStateChangeTime = powerStateChangeTime;
    }

//    /**
//     * Total configured capacity of this VM.  Defined as the sum of capacity over all Virtual disks allocated.
//     *
//     * @return the total configured capacity of the VM.
//     */
//    public Capacity getCapacity() {
//        return capacity;
//    }
//
//    public void setCapacity(final Capacity capacity) {
//        this.capacity = capacity != null ? Capacity.create(capacity) : null;
//    }

    /**
     * Get the provisioned capacity of this VM.   Files on the VM storage can expand to this size and no further.
     * This is an aggregate size available to the VM, across all DataStores and VDMKs.   This is an optional attribute.
     *
     * @return aggregated provisioned capacity of the VM.  NULL if not set.
     */
    public Capacity getProvisionedCapacity() {
        return provisionedCapacity;
    }

    public void setProvisionedCapacity(final Capacity provisionedCapacity) {
        this.provisionedCapacity =  provisionedCapacity != null ? Capacity.create(provisionedCapacity) : null;
    }

    /**
     * How much space this VM is actually using, aggregated across all DataStores and VDMKs.   This is an optional
     * attribute.   Must be less than or equal to provisioned capacity.       Note this includes ALL types of
     * space used by a VM, not just "virtual disk files" (includes snapshot overhead, swap, and other space).
     *
     * @return the actual aggregated used capacity of this VM.  NULL if not set.
     */
    public Capacity getUsedCapacity() {
        return usedCapacity;
    }

    public void setUsedCapacity(final Capacity usedCapacity) {
        this.usedCapacity = usedCapacity != null ? Capacity.create(usedCapacity) : null;
    }
    public String getMoid() {
        return moid;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException, SchemaException {
        super.marshallAttributes(cache, writer);

        addAttribute(writer, "OID", getKey().toString());

        addAttribute(writer, "Name", name);
        addAttribute(writer, "DnsName", dnsName);
        addAttribute(writer, "OS", os);
        addAttribute(writer, "Processors", processors);
        addAttribute(writer, "Memory", memoryInMB);
        addAttribute(writer, "HV_DataStore_OID", dataStoreKey);
        addAttribute(writer, "PowerState", powerState);
        addAttribute(writer, "PowerStateChangeTime", powerStateChangeTime);
        addAttribute(writer, "Host_OID", host);
        addAttribute(writer, "IPs", convertToCSV(ips));
        addAttribute(writer, "GuestState", guestState);

        if (usedCapacity != null) {
            addAttribute(writer, "UsedCapacityMB", usedCapacity.value(Capacity.Unit.MB_1024));
        }
        if (provisionedCapacity != null) {
            addAttribute(writer, "ProvisionedCapacityMB", provisionedCapacity.value(Capacity.Unit.MB_1024));
        }

        addAttribute(writer, "MOID", moid);
    }

    private String convertToCSV(final Collection<String> set) {
        if (set.size() == 0)
            return null;

        StringBuilder sb = new StringBuilder();
        for (final String str : set)
            sb.append(str).append(",");

        return sb.deleteCharAt(sb.length() - 1).toString();
    }

    /**
     * Get the "guest" state of this VM, which may be different than the power state (but not necessarily).   This is
     * a string value and may represent a high order, logical, state of the VM as oppposed to just the power state.
     *
     * @return  the guest state.
     */
    public String getGuestState() {
        return guestState;
    }

    public void setGuestState(String guestState) {
        this.guestState = guestState;
    }

    public VirtualMachineKey getKey() {
        return key;
    }

    public static enum PowerState {

        POWEREDOFF("poweredOff"),
        POWEREDON("poweredOn"),
        SUSPENDED("suspended");

        private String e;

        PowerState(String e) {
            this.e = e;
        }

        public String toString() {
            return e;
        }
    }

    public static class VirtualMachineKey implements Key<VirtualMachineKey> {
        private final String managementIp;
        private final String moid;

        public VirtualMachineKey(final String managementIp, final String moid) {
            this.managementIp = managementIp;
            this.moid = moid;
        }

        public final int compareTo(final VirtualMachineKey o) {
            int i = moid.compareTo(o.getMOID());
            if (i != 0)
                return i;

            if (managementIp != null) {
                if (o.getManagementIp() == null) {
                    return 1;
                }
                return managementIp.compareTo(o.getManagementIp());
            }
            else {
                if (o.getManagementIp() == null) {
                    return 0;
                }
                else {
                      return -1;
                }
            }
        }

        public final String getManagementIp() {
            return managementIp;
        }

        public final String getMOID() {
            return moid;
        }

        public final boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final VirtualMachineKey that = (VirtualMachineKey) o;

            if (moid != null ? !moid.equals(that.moid) : that.moid != null) return false;
            if (managementIp != null ? !managementIp.equals(that.managementIp) : that.managementIp != null)
                return false;

            return true;
        }

        public final int hashCode() {
            int result;
            result = (managementIp != null ? managementIp.hashCode() : 0);
            result = 29 * result + (moid != null ? moid.hashCode() : 0);
            return result;
        }

        public final String toString() {
            return getMOID() + (getManagementIp () != null ? (":" + getManagementIp()) : "");
        }
    }
}
