package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.util.Capacity;

import java.io.Serializable;

/**
 * Volume Capacity rules:
 * usedCapacityMB <= capacityMB
 * actualCapacityMB == capacityMB  or   actualCapacityMB = 0
 * rawToUsableRatio == rawCapacityMB / capacityMB
 */
public class VolumeCapacity implements Serializable {
    private Capacity capacity;
    private Capacity usedCapacity;
    private String redundancy;
    private boolean rawCapacityInExternalPool = false;
    private BaseVolume parent;

    protected VolumeCapacity(final BaseVolume parent) {
        this.parent = parent;
    }

    public Capacity getCapacity() {
        return capacity;
    }

    public void setCapacity(final Capacity capacity) {
        this.capacity = Capacity.create(capacity);
    }

    /**
     * Return the used capacity of the volume.
     *
     * <p/>Note that if this volume is indicated to have raw capacity in an external pool (like CLARiiON and Symmetrix
     * snapshots work), the used capacity here will be indicated as zero.
     *
     * @return Used capacity of volume (or zero if raw capacity is in external pool)
     *
     * @see #setRawCapacityInExternalPool()
     */
    public Capacity getUsedCapacity() {
        if (isRawCapacityInExternalPool())
            return Capacity.create();

        return usedCapacity;
    }

    public void setUsedCapacity(final Capacity usedCapacity) {
        this.usedCapacity = Capacity.create(usedCapacity);
    }

    public String getRedundancy() {
        return redundancy;
    }

    public void setRedundancy(final String redundancy) {
        this.redundancy = redundancy;
    }

    public final Capacity getRawCapacity() {
        // Virtual volumes always have raw capacity 0
        if (parent.isVirtual() || isRawCapacityInExternalPool())
            return Capacity.create();

        if (parent instanceof MetaVolume) {
            final Capacity rawCapacity = Capacity.create();

            // Meta volume raw capacity is the sum of its members
            final MetaVolume meta = (MetaVolume) parent;
            for (final VolumeMember member : meta.getMembers())
                rawCapacity.add(member.getCapacity().getRawCapacity());

            return rawCapacity;
        }

        final float ratio = parent.getUnderlyingStoragePool().getCapacity().getRawToUsableRatio();
        long rawCapacity = (long) (capacity.value(Capacity.Unit.BYTE) * ratio);
        return Capacity.create(rawCapacity, Capacity.Unit.BYTE);
    }

    /**
     * Special case for raw capacity. In some cases (CLARiiON Snapshots, Symmetrix snapshots, etc...) we found that the raw capacity
     * of that object is not taken from this storage pool at all. Instead, there is a dedicated storage pool (Snap virtual pool
     * in Symmetrix) that handles all the copy-on-write data for all snapshots in the array.
     *
     * In this case, the raw capacity of the volume, needs to be zero!
     */
    public final void setRawCapacityInExternalPool() {
        rawCapacityInExternalPool = true;
    }

    public final boolean isRawCapacityInExternalPool() {
        return rawCapacityInExternalPool;
    }
}
