package com.onaro.sanscreen.acquisition.util.cli;

import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorType;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileWriter;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSInputStream;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.*;

/**
 * The Test Simulator is a helper class to implement recording/simulation capabilities in a text-based data source.
 * This class provides utilities to store and read multiple test-based requests that compose an acquisition session
 * in single recording file.
 */
public class TextSimulator {
    private static final String NEWLINE = System.getProperty("line.separator");
    private static final int ONE_KB = 1024;
    private static final String requestHeader = "<<< REQUEST >>>";
    private static final String requestFooter = "<<< REQUESTEND >>>";
    private static final String responseHeader = "<<< RESPONSE >>>";
    private static final String responseFooter = "<<< RESPONSEEND >>>";
    private static final String timeHeader = "<<< TIME >>>";
    private static final String timeFooter = "<<< TIMEEND >>>";

    private final Map<String, ResponseWrapper> commands = Collections.synchronizedMap(new HashMap<String, ResponseWrapper>());

    private final RequestNormalizer requestNormalizer;
    private volatile DSFileWriter recordingFile = null;


    public TextSimulator() {
        this(null);
    }

    public TextSimulator(final RequestNormalizer requestNormalizer) {
        this.requestNormalizer = requestNormalizer;
    }

    public synchronized void setRecordingFile(final DSFileWriter recordingFile) {
        if (this.recordingFile != null) {
            throw new IllegalStateException("Recording file already initialized: " + this.recordingFile + ". Call close() first.");
        }
        this.recordingFile = recordingFile;
    }

    /**
     * Explicitly close the file opened.
     *
     * <p/>It is highly recommended to call this method when you are done recording.  This will prevent situations where
     * the file could be opened with permission privilege that may cause another entity to not be able to process
     * this file.
     * @throws com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException if an IOException occurs
     */
    public synchronized void close() throws DataSourceErrorException {
        if (recordingFile != null) {
            try {
                recordingFile.close();
            }
            catch (IOException e) {
                throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, null, "Failed to close recording file.");
            }
            finally {
                recordingFile = null;
            }
        }
    }

    /**
     * Record a request into the recording file
     *
     * @param command The request issued to the device
     * @param response The response returned from the device
     * @param time The time it took to process the request (millis)
     * @throws com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException if the request could not be recorded
     */
    public void recordRequest(final String command, final String response, final long time) throws DataSourceErrorException {
        try {
            final StringBuilder record = new StringBuilder(ONE_KB);

            record.append(requestHeader).append(NEWLINE);
            record.append(command).append(NEWLINE);
            record.append(requestFooter).append(NEWLINE);
            record.append(responseHeader).append(NEWLINE);
            record.append(response).append(NEWLINE);
            record.append(responseFooter).append(NEWLINE);
            record.append(timeHeader).append(NEWLINE);
            record.append(Long.toString(time)).append(NEWLINE);
            record.append(timeFooter).append(NEWLINE).append(NEWLINE);

            recordingFile.write(record.toString());
            recordingFile.flush();
        }
        catch (IOException e) {
            throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, null, "Failed to record request", e);
        }
    }

    /**
     * Initialize the Simulator from recording files. This will read all the given recording files and parse them
     * to extract the requests / responses encoded inside. The result will build a map of request-response to allow
     * simulating the exact behavior of the device.
     *
     * @param recordingStreams List of recording files containing the recorded session
     *
     * @throws IOException
     */
    public void initializeSimulation(final Collection<DSInputStream> recordingStreams) throws IOException {
        commands.clear();

        final StringBuilder requestBuff = new StringBuilder(255);
        final StringBuilder responseBuff = new StringBuilder(ONE_KB * 10);
        final StringBuilder timeBuff = new StringBuilder(40);

        for (final DSInputStream recordingStream : recordingStreams) {
            BufferedReader dataFileReader = null;
            try {
                // Open the data file. This file may contain many requestBuff & responseBuff pairs
                dataFileReader = new BufferedReader(new InputStreamReader(recordingStream, "utf-8"));

                // HEADSUP we have access to the stream now, lets normalize while we go along.
                while (true) {
                    String line = dataFileReader.readLine();
                    if (line == null)
                        break;

                    if (line.equals(requestHeader)) {
                        if (requestBuff.length() != 0 || responseBuff.length() != 0)
                            throw new IllegalStateException("Failed to parse recording: new request being parsed before finishing previous one.");

                        line = dataFileReader.readLine();

                        final StringBuilder requestBuffer = new StringBuilder(ONE_KB * 2);

                        while (line != null && !line.equals(requestFooter)) {
                            requestBuffer.append(normalizeCommandMapEntry(line));
                            line = dataFileReader.readLine();
                        }

                        if (line == null)
                            break;

                        requestBuff.append(requestBuffer.toString());
                    }

                    if (line.equals(responseHeader)) {
                        line = dataFileReader.readLine();
                        if (line == null)
                            break;

                        while (line != null && !line.equals(responseFooter)) {
                            responseBuff.append(line);
                            responseBuff.append(NEWLINE);
                            line = dataFileReader.readLine();
                        }

                        if (line == null)
                            break;
                    }

                    if (line.equals(timeHeader)) {
                        line = dataFileReader.readLine();
                        if (line == null)
                            break;

                        final StringBuilder timeString = new StringBuilder(100);

                        while (line != null && !line.endsWith(timeFooter)) {
                            timeString.append(line);
                            line = dataFileReader.readLine();
                        }

                        if (line == null)
                            break;

                        timeBuff.append(timeString.toString().trim());
                    }

                    final String rawRequest = requestBuff.toString();
                    final String normalizedRequest = (this.requestNormalizer == null)
                            ? rawRequest
                            : requestNormalizer.normalizeRequest(rawRequest);
                    final String response = responseBuff.toString();

                    if ((rawRequest.length() != 0) && (response.length() != 0)) {

                        final ResponseWrapper responseWrapper;
                        if(commands.containsKey(normalizedRequest)) {
                            responseWrapper = commands.get(normalizedRequest);
                        }
                        else {
                            responseWrapper = new ResponseWrapper();
                        }

                        responseWrapper.responses.add(response);

                        final String time = timeBuff.toString();
                        if (time.length() > 0)
                            responseWrapper.responseTimes.add(Long.parseLong(time));

                        commands.put(normalizedRequest, responseWrapper);

                        requestBuff.setLength(0);
                        responseBuff.setLength(0);
                        timeBuff.setLength(0);
                    }
                }
            }
            finally {
                if (dataFileReader != null)
                    dataFileReader.close();
            }
        }

        if (commands.size() == 0)
            throw new IOException("Did not find any recorded requests!");
    }

    /**
     * Simulates a device response based on the given request. {@link #initializeSimulation} must be called before using
     * this method!
     *
     * @param command Request to simulate response for
     *
     * @return The simulated response string
     *
     * @throws DataSourceErrorException if the request could not be found
     */
    public String simulateResponse(final String command) throws DataSourceErrorException {
        final String intermediateNormalizedCommand = normalizeCommandMapEntry(command);
        final String normalizedCommand = (requestNormalizer == null)
                ? intermediateNormalizedCommand
                : requestNormalizer.normalizeRequest(intermediateNormalizedCommand);

        final ResponseWrapper response = commands.get(normalizedCommand);
        if (response == null) {
            // Send the command not normalized to the caller.
            // It'll be easier to the caller identify the command. The command not found can be retrieved as normalized
            // command using <code>getNormalizedLastCommandNotFound</code> method.
            throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, "Failed to simulate request", "Not found request: [" + command + "]");
        }

        // To handle the situation where a command is called twice but the recording is older
        // and only has one of the calls, return the previous response
        if(!response.responseIterator.hasNext()) {
            return response.responses.get(response.responseIterator.previousIndex()); // Using index so as not turn move the iterator
        }

        final ResponseWrapper newResponse = new ResponseWrapper(response);
        final String responseString = newResponse.responseIterator.next();
        commands.put(normalizedCommand, newResponse);

        return responseString;
    }

    /**
     * Normalize the command to avoid any mismatch due to extra spaces or control chars that are not
     * importand.
     *
     * @param command No normalized command.
     *
     * @return command normalized
     */
    private String normalizeCommandMapEntry(final String command) {
        return command.toLowerCase().replaceAll(" ", "").replaceAll(NEWLINE, "").replace("\t", "").trim();
    }

    public static interface RequestNormalizer {
        public String normalizeRequest(final String rawRequest);
    }

    /**
     * A class that wraps the responses for a given command
     */
    private static class ResponseWrapper {
        public List<String> responses;
        public ListIterator<String> responseIterator;
        public List<Long> responseTimes;

        public ResponseWrapper() {
            responses = new ArrayList<String>();
            responseIterator = responses.listIterator();
            responseTimes = new ArrayList<Long>();
        }

        public ResponseWrapper(final ResponseWrapper src) {
            responses = new LinkedList<String>(src.responses);
            responseIterator = responses.listIterator(src.responseIterator.previousIndex() + 1);
            responseTimes = new LinkedList<Long>(src.responseTimes);
        }
    }

}
