/*
 * Copyright (c) 2009 NetApp
 * All rights reserved
 */

package com.onaro.sanscreen.server.sessions;

import com.netapp.oci.platform.common.ServerConstant;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.originator.OriginatorRemoveTask;
import com.netapp.oci.platform.originator.OriginatorUpdateTask;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.plugin.data.PlatformPlugin;
import com.netapp.oci.platform.plugin.remote.PlatformPluginRemote;
import com.netapp.oci.server.UpdateTaskException;
import com.netapp.oci.server.UpdateTaskSynchronizer;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceNew;
import com.onaro.sanscreen.server.originatorupdate.OriginatorCacheChanges;
import com.onaro.sanscreen.server.tables.acquisition.DataSourceTableNew;
import com.onaro.sanscreen.util.Constant;
import com.onaro.sanscreen.util.Schema;
import com.onaro.sanscreen.util.xml.DefaultHandler;
import com.onaro.sanscreen.util.xml.SaxParserWrapper;
import com.onaro.sanscreen.util.xml.XMLException;
import com.onaro.sanscreen.util.xml.XMLUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.jboss.ejb3.annotation.LocalBinding;
import org.jboss.ejb3.annotation.RemoteBinding;
import org.jboss.ejb3.annotation.RemoteBindings;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.annotation.security.RunAs;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.sql.DataSource;
import java.io.Reader;
import java.io.StringReader;
import java.sql.Connection;

@Stateless
@RemoteBindings({
        @RemoteBinding(jndiBinding = OriginatorUpdateSession.JNDI_NAME),
        @RemoteBinding(jndiBinding = ServerConstant.HTTPS_JNDI_BINDING_PREFIX + OriginatorUpdateSession.JNDI_NAME)})
@LocalBinding(jndiBinding = OriginatorUpdateSessionLocal.JNDI_NAME)
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "acquisition", "admin"})
@RunAs("admin")
public class OriginatorUpdateSessionEJB implements OriginatorUpdateSession, OriginatorUpdateSessionLocal {
    private static final Logger LOG = LoggerFactory.getLogger(OriginatorUpdateSessionEJB.class);

    @Resource(name = "jdbc/OnaroDataSource",
            mappedName = "java:/onaro_pool",
            type = DataSource.class,
            authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    @EJB
    private PlatformPluginRemote platformPluginRemote;

    private static class DSInfo {
        public String name;
        public Long id;

        public DSInfo(String name, Long id) {
            this.name = name;
            this.id = id;
        }
    }

    public boolean update(String xml, Schema schema) throws UpdateTaskException {
        OriginatorCacheChanges update = getUpdate(xml, schema);
        return update != null && update.isChanged();
    }

    /**
     * Parses data XML and stores data in memory cache and DB
     *
     * @param xml    xml data
     * @param schema schema type
     * @return changes versus previously stored data
     * @throws UpdateTaskException any exception caught within this method
     */
    public OriginatorCacheChanges getUpdate(String xml, Schema schema) throws UpdateTaskException {
        try {
            Long originId = XMLUtil.getLongElementVal(xml, "Originator");

            synchronized (OriginatorUpdateImpl.INSTANCE.getOriginatorSyncObject(originId)) {
                if (schema != Schema.NETAPP) {
                    LOG.info("License exception");
                    throw new UpdateTaskException("License exception");
                }

                String dataSourceName = XMLUtil.getElementAttributeVal(xml, "Info", "DataSourceName");
                return getUpdate(new StringReader(xml), schema, originId, dataSourceName);
            }
        } catch (XMLException e) {
            throw new UpdateTaskException("Failed to get update server with the xml", e);
        }
    }

    /**
     * Reads and processes XML data from a stream
     *
     * @param reader         reader for the data XML
     * @param schema         XML schema
     * @param originId       originator ID
     * @param dataSourceName data source name.  Can be <code>null</code>
     * @return changes versus previously stored data
     * @throws UpdateTaskException a generic exception in case anything goes wrong
     */
    public OriginatorCacheChanges getUpdate(Reader reader, Schema schema, Long originId, String dataSourceName) throws UpdateTaskException {
        long starttime = System.currentTimeMillis();

        OriginatorUpdateTask updateTask = null;
        Connection connection = null;
        try {

            synchronized (OriginatorUpdateImpl.INSTANCE.getOriginatorSyncObject(originId)) {
                if (schema != Schema.NETAPP) {
                    LOG.info("License exception");
                    throw new UpdateTaskException("License exception");
                }

                connection = ConnectionHelper.newConnection(dataSource);

                DSInfo datasource = findDatasourceInfo(connection, originId);
                PlatformPlugin platformPlugin = platformPluginRemote.getPlugin(schema);
                updateTask = new OriginatorUpdateTask(datasource.name, platformPlugin, originId, datasource.id, starttime, true);
                boolean isUpdate = updateTask.update(connection, reader, true, schema);

                boolean hasObjectsToPurge = updateTask.isPurgingRequired(connection, schema, originId);
                if (isUpdate || hasObjectsToPurge) {
                    updateTask.setHasObjectsToPurge(hasObjectsToPurge);
                    UpdateTaskSynchronizer.getInstance().execute(updateTask);
                    LOG.info("Notify change: originator = {}, updateTime = {}", originId, updateTask.getUpdateTime());
                    platformPlugin.changesCommitted(connection, datasource.id, datasource.name, updateTask.getOriginatorChanges());
                }

                OriginatorCacheChanges cacheChanges = new OriginatorCacheChanges();
                cacheChanges.setChangesCount(updateTask.getChangeCount());
                return cacheChanges;
            }
        } catch (Exception e) {
            LOG.error(String.format("Failed in processing xml for originId: %d, dataSourceName: %s", originId, dataSourceName), e);
            if (updateTask != null) {
                updateTask.rollback();
            }
            throw new UpdateTaskException("Failed to get update server with the xml, originator " + originId, e);
        } finally {
            ConnectionHelper.close(connection);
            LOG.info("Originator: {} cache time (ms): {}", originId, System.currentTimeMillis() - starttime);
        }
    }

    public OriginatorCacheChanges removeOriginUpdate(Long originId) throws UpdateTaskException {
        synchronized (OriginatorUpdateImpl.INSTANCE.getOriginatorSyncObject(originId)) {
            Connection connection = null;
            try {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("removeOriginUpdate() - (originId: {})", originId);
                }
                connection = ConnectionHelper.newConnection(dataSource);
                DSInfo datasource = findDatasourceInfo(connection, originId);
                if (LOG.isDebugEnabled()) {
                    LOG.debug("removeOriginUpdate()  - (datasourceName: {})", datasource.name);
                }

                //Since we don't have a schema context, try removal on all handler systems until
                //we get a successful update.
                for (Schema schema : Schema.values()) {
                    PlatformPlugin platformPlugin = platformPluginRemote.getPlugin(schema);
                    if (platformPlugin != null){
                        OriginatorRemoveTask removeTask = new OriginatorRemoveTask(connection, platformPlugin, originId, datasource.name);
                        if (removeTask.isDeleteRequired()) {
                            UpdateTaskSynchronizer.getInstance().execute(removeTask);
                            platformPlugin.notifyOriginatorRemove(connection, datasource.id, datasource.name, removeTask.getRootId(), removeTask.getUpdateTime());
                            OriginatorCacheChanges cacheChanges = new OriginatorCacheChanges();
                            cacheChanges.setChangesCount(removeTask.getTotalObjCount());
                            return cacheChanges;
                        }
                    }
                }
            } catch (Exception e) {
                throw new UpdateTaskException("Failed to remove originator for originId: " + originId, e);
            } finally {
                ConnectionHelper.close(connection);
            }
        }
        return new OriginatorCacheChanges();
    }

    public void validate(String xml, Schema schema) throws XMLException {
        SaxParserWrapper parser = new SaxParserWrapper(new DefaultHandler());
        parser.setSchemaFileUri(schema.getFilename());
        parser.parseXmlString(xml);
    }

    /**
     * Finds Datasource's name and ID based on a given Originator Id that this Datasource had reported about it.
     *
     * @param connection database connection.
     * @param originId   originator ID
     * @return the actual data source name, if found in the database.  Else, <code>Constant.UNKNOWN_ORIGINATOR_NAME</code> is returned.
     * @throws TableException
     */
    private DSInfo findDatasourceInfo(Connection connection, Long originId) throws TableException {
        DataSourceTableNew dataSourceTable = new DataSourceTableNew(connection);
        DataSourceNew dataSource = dataSourceTable.getByOriginatorId(originId.longValue());
        if (dataSource != null) {
            return new DSInfo(dataSource.getName(), dataSource.getId());
        }

        return new DSInfo(Constant.UNKNOWN_ORIGINATOR_NAME, Constant.LONG_NULL_ID);
    }

    public void resetAllCache() throws UpdateTaskException {
        try {
            for (PlatformPlugin plugin : platformPluginRemote.getPlugins()) {
                KeyCache.clear(plugin);
            }
        } catch (Exception e) {
            LOG.error("Failed to reset all cache", e);
            throw new UpdateTaskException("Failed to reset all cache", e);
        }
    }
}
