package com.onaro.util.sessions;

import com.onaro.sanscreen.util.LoginConfiguration;
import org.jboss.security.RunAsIdentity;
import org.jboss.security.SecurityContextAssociation;
import org.jboss.security.auth.callback.UsernamePasswordHandler;
import org.jboss.security.client.SecurityClient;
import org.jboss.security.client.SecurityClientFactory;
import org.jboss.security.identity.Identity;

import javax.security.auth.login.Configuration;
import java.rmi.RemoteException;
import java.security.Principal;

/**
 * This class facilitates remote login for a JBoss client. It simplifies
 * the process by removing the requirement of the auth.conf file.
 * <p/>
 * Note: Logging calls should be avoided as the client which uses this class
 * might not have the log4j.jar installed.
 */
public class DefaultClientLogin implements ClientLogin {
    private final String user;
    private final String password;

    public DefaultClientLogin(String user, String password) {
        this.user = user;
        this.password = password;
    }

    /**
     * Login to the server using RMI
     * <p/>
     * Note: Has to be called from within a synchronized method
     *
     * @throws RemoteException
     */
    public void login() throws RemoteException {
        if (hasSecurityContext()) {
            clean();
        }

        SecurityClient securityClient = null;
        try {
            Configuration.setConfiguration(new LoginConfiguration());

            securityClient = SecurityClientFactory.getSecurityClient();
            securityClient.setJAAS("onaro-login", new UsernamePasswordHandler(user, password));
            // login assumes only one authentication per JVM, if not the case, this code will trample
            // other concurrent authentications
            securityClient.setVmwideAssociation(true);
            securityClient.login();

        } catch (Exception e) {
            // logout if we failed processing login
            if (securityClient != null) {
                securityClient.logout();
            }
            throw new RemoteException("Failed to login to server", e);
        }
    }

    public boolean isLoggedIn() {
        return hasSecurityContext() && !needRelogin();
    }

    private boolean hasSecurityContext() {
        return SecurityContextAssociation.getSecurityContext() != null;
    }

    public String getUser() {
        return user;
    }


    public void clean() {
        try {
            SecurityClientFactory.getSecurityClient().logout();
        }
        catch (Exception ignored) {
        }
    }

    private boolean needRelogin() {
        if (SecurityContextAssociation.getSecurityContext() != null) {
            if (SecurityContextAssociation.getSecurityContext().getOutgoingRunAs() != null &&
                ((RunAsIdentity) SecurityContextAssociation.getSecurityContext().getOutgoingRunAs()).getPrincipalsSet() != null) {
                for (Principal principal : ((RunAsIdentity) SecurityContextAssociation.getSecurityContext().getOutgoingRunAs()).getPrincipalsSet()) {
                    return !"anonymous".equals(principal.getName()) && !user.equals(principal.getName());  // return after the first one
                }
            }
            if (SecurityContextAssociation.getSecurityContext().getSubjectInfo() != null &&
                SecurityContextAssociation.getSecurityContext().getSubjectInfo().getIdentities() != null) {
                for (Identity identity : SecurityContextAssociation.getSecurityContext().getSubjectInfo().getIdentities()) {
                    return (identity != null) && !"anonymous".equals(identity.getName()) && !user.equals(identity.getName());  // return after the first one                    
                }
            }
        }

        return false;
    }
}
