package com.onaro.util.sessions;

import com.netapp.oci.platform.common.ServerConstant;
import javax.ejb.CreateException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.rmi.RemoteException;
import java.rmi.UnmarshalException;
import java.util.HashMap;
import java.util.Map;

public class SessionCache {

    private static final Logger LOG = LoggerFactory.getLogger(SessionCache.class);

    private Map<String, Object> sessionCache = new HashMap<String, Object>();
    private Context initialContext = null;
    private boolean stopping = false;
    private boolean rmiOverHttp = Boolean.getBoolean("com.onaro.rmiOverHttp");

    private int timeoutMillis;

    private final ClientLogin clientLogin;

    private ClassLoader classLoader;

    /**
     * Initializes the session cache.
     *
     * @param classLoader   The class loader used to load the proxy object. This should be the loader of the class using the session.
     * @param clientLogin   the client logging in to JBoss
     * @param timeoutMillis Total tims to try to recover from communication problems before giving up. When 0, no retries will be attempted
     */
    public SessionCache(ClassLoader classLoader, ClientLogin clientLogin, final int timeoutMillis) {
        this.timeoutMillis = timeoutMillis;
        this.clientLogin = clientLogin;
        this.classLoader = classLoader;
    }

    /**
     * Initializes the session cache.
     *
     * @param classLoader   The class loader used to load the proxy object. This should be the loader of the class using the session.
     * @param clientLogin   the client logging in to JBoss
     * @param timeoutMillis Total tims to try to recover from communication problems before giving up. When 0, no retries will be attempted
     * @param logExceptions
     * @deprecated Use {@link #SessionCache(ClassLoader, ClientLogin, int).
     */
    @Deprecated
    public SessionCache(ClassLoader classLoader, ClientLogin clientLogin, final int timeoutMillis, boolean logExceptions) {
        this(classLoader, clientLogin, timeoutMillis);
    }

    public synchronized void stop() {
        stopping = true;

        // TODO: Interrupt all proxies
    }

    public synchronized boolean isStopping() {
        return stopping;
    }

    /**
     * Retrieves an EJB session.
     * This method will try to reuse previously created sessions and home interfaces (if possible).
     * The retrieved object is a Proxy that wrappes the Session object to detect network failures and try to recover
     * from then.
     *
     * @param jndiName JNDI name of EJB session to retrieve
     * @throws SessionCacheException
     */
    public synchronized <E> E getSession(final String jndiName) throws SessionCacheException {
        final Object session;

        try {
            // Perform an additional lookup to check the type of object retrurned.
            session = lookupSession(jndiName);
        }
        catch (RemoteException e) {
            throw new SessionCacheException("Failed to lookup session " + jndiName, e);
        }

        return (E) Proxy.newProxyInstance(classLoader, session.getClass().getInterfaces(), new ProxyInvocationHandler(jndiName));
    }

    public synchronized void cleanCache() {
        clean();
    }

    public synchronized String[] getCachedSessions() {
        final String[] descriptions = new String[sessionCache.size()];
        sessionCache.keySet().toArray(descriptions);

        return descriptions;
    }

    /**
     * Creates the initial context.
     * <p/>
     * Note: Has to be called from within a synchronized method
     *
     * @throws NamingException
     */
    private Context getInitialContext() throws NamingException {
        if (initialContext == null) {
            initialContext = new InitialContext();
        }

        return initialContext;
    }

    /**
     * Closes the initial context.
     * <p/>
     * Note: Has to be called from within a synchronized method
     */
    private void closeInitialContext() {
        if (initialContext != null) {
            try {
                initialContext.close();
            }
            catch (NamingException ignore) {}

            initialContext = null;
        }
    }

    /**
     * Clean session cache!
     * <p/>
     * Note: Has to be called from within a synchronized method
     */
    private void clean() {
        clientLogin.clean();
        closeInitialContext();
        sessionCache.clear();
    }

    /**
     * This prefix is used to change the name of the EJB that we're connecting to. This is important
     * since the naming server returns us a proxy that is already embedded with the protocol used to communicate.
     * What we do is advertive the EJB under two names. One for RMI and the other for RMI-OVER-HTTP (same name
     * with the HTTP prefix).
     *
     * @param jndiName
     */
    private String resolveJndiName(String jndiName) {
        return rmiOverHttp ? ServerConstant.HTTPS_JNDI_BINDING_PREFIX + jndiName : jndiName;
    }

    /**
     * Retrieves an EJB session reference (after logging in, looking up, etc...). This method will try to reuse the
     * session cache if possible to save network calls.
     */
    public synchronized Object lookupSession(final String jndiName) throws RemoteException {
        if (isStopping()) {
            throw new RemoteException("Communication Manager is currently stopping!");
        }

        if (!clientLogin.isLoggedIn()) {
            clientLogin.login();
        }

        // Lookup the session cache...
        String resolvedJndiName = resolveJndiName(jndiName);

        Object session = sessionCache.get(resolvedJndiName);
        if (session == null) {
            try {
                session = getInitialContext().lookup(resolvedJndiName);
                if (session == null) {
                    LOG.debug("Creating session '{}'", resolvedJndiName);

                    closeInitialContext();
                    session = getInitialContext().lookup(resolvedJndiName);
                    if (session == null) {
                        throw new NamingException("Context.lookup(\"" + resolvedJndiName + "\") returned null twice in a row");
                    }
                }

                sessionCache.put(resolvedJndiName, session);
            }
            catch (Exception e) {
                clean();
                throw new RemoteException("Failed to get the " + resolvedJndiName + " session", e);
            }
        }

        return session;
    }

    /**
     * Represent a wrapper or handler class to the underline EJB Session.
     * The main intention of this class is to create a generic way to encapsulate problem related to:
     * <p/>
     * <ul>
     * <li> Create an instance of the right type.
     * <li> Provides a mechanism of reconnection is case of a broken communication.
     * </ul>
     * <p/>
     * Once established a session the User can call EJB session's method that are
     * executed by {@link ProxyInvocationHandler#invoke(Object,java.lang.reflect.Method,Object[])}
     * which uses underling implementation represented by <code>obj</code>.
     * <code>obj</code> is an instance of a particular EJB Session.  In order to create the right EJB Session
     * we use the <code>jndiName</code>
     */
    private class ProxyInvocationHandler implements InvocationHandler {
        private String jndiName;

        public ProxyInvocationHandler(final String jndiName) {
            this.jndiName = jndiName;
        }

        @Override
        public Object invoke(Object proxy, Method m, Object[] args) throws Throwable {
            final int retries = Math.max(timeoutMillis / 1000, 1);
            final int sleepTime = Math.max(timeoutMillis / retries, 1000);

            int retriesLeft = retries;
            while (!isStopping()) {
                try {
                    return m.invoke(lookupSession(jndiName), args);
                }
                catch (InvocationTargetException e) {
                    final Throwable targetException = e.getTargetException();

                    if (timeoutMillis > 0 && !isRecoverableCommunicationError(targetException)) {
                        LOG.error("failed to invoke jndiName: " + jndiName, targetException);
                        throw targetException;
                    }

                    // We are handling only these kind of exceptions because this means that the target method that we're executing
                    // threw an exception. This means that EJB communication layer failed somehow when calling the server.
                    // Stop and Restart, this will clean the cache, create and initialize all the related instances.
                    retriesLeft--;

                    if (retriesLeft <= 0) {
                        StringBuilder sb = new StringBuilder();
                        sb.append("failed to invoke ").append(jndiName).append(" (no retries left)");
                        LOG.error(sb.toString(), targetException);
                        throw targetException;
                    }

                    // Provide only basic info in the log. The detailed exception, including stack trace will be written when
                    // no retries are left.
                    LOG.warn("failed to invoke {} - {} retries left: {}:{}",
                            new Object[] { jndiName, retriesLeft, targetException.getClass().getName(),
                            targetException.getMessage() });

                    // Clean state and retry...
                    cleanCache();

                    try {
                        Thread.sleep(sleepTime);
                    }
                    catch (InterruptedException ignore) {}
                }
            }

            // If we're here, the service is stopping...
            throw new InterruptedException("Session Cache was stopped");
        }

        /**
         * HACK:
         * On these exceptions, we will try to silently retry the connection before propagating the problem
         */
        private boolean isRecoverableCommunicationError(Throwable targetException) {
            // HACK: If the server sent us an exception and we could not marshal this exceptio, there is no point in
            // recovering
            try {
                if (targetException instanceof UnmarshalException) {
                    if (targetException.getCause() instanceof ClassNotFoundException) {
                        return false;
                    }
                }
            }
            catch (NoClassDefFoundError e) {
                return false;
            }

            // Exceptions thrown by the JBoss communication layer
            if (targetException instanceof RemoteException) {
                return true;
            }

            // Exceptions thrown during session.create
            if (targetException instanceof CreateException) {
                return true;
            }

            // HACK: Exception usually thrown when the server reboots and the client was in the middle of an RMI call
            if (targetException.getClass().getName().equals(IllegalArgumentException.class.getName())) {
                return true;
            }

            // OutOfMemory is also a recoverable error...
            if (targetException instanceof OutOfMemoryError) {
                return true;
            }

            return false;
        }
    }
}
