/*
 * counter_data_info.c
 *
 *  Created on: Feb 20, 2014
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/types.h>
#include "cmapl_wrapper_header.h"
#define CM_MAX_COUNTERS_PER_OBJECT 512

#ifdef DEBUG_ONLY // Run this routine only in Debug mode, as this routine is for testing JNA translation.
/**
 * Function takes ccma file, objects specific timestamp and the object name and prints all counter values comma separated into the file passed.
 * ccma_file_path  = "ccma file path location"
 * output_file_path = output file path location
 * time = timestamp value for an object where we gather counter info
 * current_obj_char = object for which we gather counter data (resource, workload_queue_dblade,..)
 */

bool write_counter_values_for_timestamp_to_file ( const char* ccma_file_path,char *output_file_path, uint64_t time, uint64_t time_period, char* current_obj_char ) {


	setlogmask (LOG_UPTO (LOG_NOTICE));
	openlog ("cmapl_wrapper_get_counter_data", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

	//Open the output file for reading
	FILE *fp = fopen( output_file_path, "w");
	if( fp == 0 ) {
		syslog (LOG_ERR, "\nCould not dump counter values to file. Error opening output file : %s", output_file_path);
		return CMAPL_WRAPPER_ERROR;
	}

	uint32_t instance_count = 0;
	cmapl_wrapper_counter_data_t * get_counter_data_list;
	uint32_t headers_count =0;
	counter_wrapper_headers * get_counter_headers_list;
	char * error = NULL;

	//test Counter data return
	uint32_t ret_value = get_counter_values_for_timestamp(ccma_file_path, time, time_period, current_obj_char, &get_counter_data_list, &instance_count, &get_counter_headers_list, &headers_count, &error );
	if(ret_value == CMAPL_WRAPPER_ERROR) {
		goto error_1;
	}

	uint16_t i, j, k;
	for (i=0; i < instance_count; i++) {
		for (j = 0; j < get_counter_data_list[i].num_of_counters_count; j++) {
			fprintf(fp, "\"");
			for (k = 0 ; k < get_counter_data_list[i].counter_values[j].num_values ; k++) {
				switch(get_counter_data_list[i].counter_values[j].value_type){
				case CMAPL_CNT_4BYTES:
					fprintf(fp, "%d", get_counter_data_list[i].counter_values[j].values[k].val_4B);
					break;
				case CMAPL_CNT_8BYTES:
					fprintf(fp, "%"PRIu64, get_counter_data_list[i].counter_values[j].values[k].val_8B);
					break;
				case CMAPL_CNT_STRING:
					fprintf(fp, "%s", get_counter_data_list[i].counter_values[j].values[k].val_str);
					break;
				}
				// print out multiple values if it's an array counter
				if (k != (get_counter_data_list[i].counter_values[j].num_values - 1)){
					fprintf(fp, ", ");
				}
			}
			fprintf(fp, "\"");
		}
		fprintf(fp, "\n");
	}

	free_cmapl_wrapper_counter_data (get_counter_data_list, instance_count, get_counter_headers_list, headers_count, error);
	fclose(fp);
	return CMAPL_WRAPPER_SUCCESS;
    closelog ();

error_1:
	if(get_counter_data_list != NULL) {
		free_cmapl_wrapper_counter_data (get_counter_data_list, instance_count, get_counter_headers_list, headers_count, error);
	}

	if(fp) fclose(fp);
    closelog ();
	return CMAPL_WRAPPER_ERROR;
}
#endif // Run the above routine in debug mode.

/**
 * Function takes ccma file, objects specific timestamp and the object name and returns an array of cmapl_counter_data
 * ccma_file_path  = "ccma file path location"
 * time = timestamp value for an object where we gather counter info
 * current_obj_char = object for which we gather counter information (resource, workload_queue_dblade,..)
 * wrapper_counter_data = output parameter which gives array of counter data values
 * counter_data_handle_count = output parameter which gives number of instances for an object at specified timestamp
 * wrapper_headers = output parameter which gives metadata(headers) info for counter (cpu_busy, ...)
 * header_count = output parameter which gives number of headers count
 */
bool get_counter_values_for_timestamp( const char* ccma_file_path, uint64_t time, uint64_t time_period, char* current_obj_char, cmapl_wrapper_counter_data_t **wrapper_counter_data, uint32_t * counter_data_handle_count,
											counter_wrapper_headers **wrapper_headers, uint32_t * header_count, char **error) {


	// On UNIX to enable syslog, add to /etc/syslog.conf or /etc/rsyslog.conf
	// local1.*         /var/log/mylog
	// restart syslogd
	// /etc/init.d/syslog restart (or) /etc/init.d/rsyslog
	setlogmask (LOG_UPTO (LOG_NOTICE));
	openlog ("cmapl_wrapper_get_counter_data", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

	bool ret_value = CMAPL_WRAPPER_SUCCESS;


	*wrapper_headers = NULL;
	*wrapper_counter_data = NULL;
	*error = NULL; // error string array

	uint32_t instance_count = 0;
	// Open Ccma file to read
	cmapl_file_handle_t* file;
	const char* meta_path;
	const char* data_path;
	cmapl_error_t err = cmapl_open_data_file(ccma_file_path, "r", &file, &meta_path, &data_path);
	if (err != CMAPL_SUCCESS){
		syslog (LOG_ERR, "\nError opening file: %s, error : %s\n", ccma_file_path, cmapl_error_labels[err]);
		file = NULL; // any error above cleans up data_path and meta_path, ensuire file is NULL
		goto error_1;
	}

	// Set the parameters to get the header handles
	uint64_t start_time = 0;
	err = cmapl_get_oldest_record_time(file, &start_time);
	if (err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "\nCould not get the oldest record time error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		goto error_1;
	}
	uint64_t end_time = 0;
	err = cmapl_get_newest_record_time(file, &end_time);
	if (err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "\nCould not get the newest record time error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		goto error_1;
	}

	// get the header handle
	cmapl_header_handle_t* header;
	err = cmapl_get_header(file, &header);
	if (err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "\nCould not get the header handle error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		goto error_1;
	}

	// get Object info
	cmapl_object_handle_t* current_object;
	err =  cmapl_find_object(header, current_obj_char, &current_object);
	if (err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "\nCould not get the object : [%s], error : %s for file : %s",current_obj_char, cmapl_error_labels[err], ccma_file_path);
		goto error_1;
	}

	// get Object Info
	cmapl_object_def_t obj_info;
	if( current_object != NULL ) {
		err = cmapl_get_object_info(current_object, &obj_info);
		if (err != CMAPL_SUCCESS) {
			syslog(LOG_ERR, "\nCould not get the object info for object : [%s], error : %s for file : %s",current_obj_char, cmapl_error_labels[err], ccma_file_path);
			goto error_1;
		}
	}

	uint32_t cnt_idx;
	const char* cnt_names[CM_MAX_COUNTERS_PER_OBJECT];
	// Get counter info, like cnt_name (e.g, instance_uuid, instance_name, )
	// and also get no, of counters. This is needed for query.
	err = get_counter_info(current_object, time_period, cnt_names, &cnt_idx);
	if(err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "Couldn't get the counter information, error: %s, object : %s for file : %s",cmapl_error_labels[err], current_object, ccma_file_path);
		goto error_1;
	}

	// Construct a query
	cmapl_query_def_t query_def;
	cmapl_set_latest_version(&query_def);
	query_def.object_name = obj_info.name;
	query_def.num_cnts = cnt_idx; // This needs to be optional, It is ONTAP burt and currently is workaround only.
	query_def.cnt_names = cnt_names; // This needs to be optional, It is ONTAP burt and currently is workaround only.
	query_def.start_time = start_time;
	query_def.end_time = end_time;
	query_def.period = time_period; //input param time_period
	query_def.num_key_cnts = 0;

	// create and get the query handle for the ccma file
	cmapl_query_handle_t* query;
	err = cmapl_query(file, &query_def, &query);
	if (err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "\nCould not run the query error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		goto error_1;
	}


	// iterate through all the timestamps in the query result that have counter data
	cmapl_timestamp_handle_t* timestamp;
	err = cmapl_first_timestamp(query, &timestamp);
	if (err != CMAPL_SUCCESS){
		syslog(LOG_ERR, "\nCould not get the first timestamp error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		goto error_1;
	}


	uint32_t track_key_cnt = 0; // used to track key counters to construct headers data

	// counter values which keep track of headers e.g, Instance_name, instance_uuid, ....
	uint32_t counter_headers_count = 0;

	while (timestamp != NULL){
		uint64_t timeutc = 0;
		err = cmapl_get_timestamp(timestamp, &timeutc);
		if (err != CMAPL_SUCCESS){
			syslog(LOG_ERR, "\nCould not get the timestamp error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
			goto error_1;
		}

		// Check for the required timestamp, the "time" passed as input parameter
		if(timeutc == time) {
			cmapl_instance_handle_t* instance;
			err = cmapl_first_instance(timestamp, &instance);
			if (err != CMAPL_SUCCESS){
				syslog(LOG_ERR, "\nCould not get the first instance error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
				goto error_1;
			}


			//counter which holds number of instances
			instance_count = 0;
			//Iterate through all the instances
			while (instance != NULL) {

				if(instance_count > 0) {
					// more memory needed, if there are more instances. We do this because we can never know the
					// lenght of number of instances before through cmapl calls.
					*wrapper_counter_data = (cmapl_wrapper_counter_data_t *) realloc (*wrapper_counter_data, (instance_count + 1) * sizeof(cmapl_wrapper_counter_data_t));
					if(*wrapper_counter_data == NULL){
						syslog(LOG_ERR, "\nCould not allocate memory for wrapper_counter_data after query in file : %s", ccma_file_path);
						goto error_1;
					}
				} else {
						// Allocate memory for the return counter types at instance level
						// wrapper_counter_data is of the following format
						// index(instance) : {coutner_data}
						// E.g.
						// 1 : {"CPU_idle""131944448082""17959651368""8"}
						// 2 : {"CPU_kahuna""372797805""17959651368""1"}
						// 3 : {"CPU_exempt""9135396610""17959651368""8"}
						*wrapper_counter_data = (cmapl_wrapper_counter_data_t *) malloc (sizeof(cmapl_wrapper_counter_data_t));
						if(*wrapper_counter_data == NULL){
							syslog(LOG_ERR, "\nCould not allocate memory for wrapper_counter_data after query in file : %s", ccma_file_path);
							goto error_1;
						}
				}

				// get instance data
				cmapl_instance_data_t instance_data;
				err = cmapl_get_instance_data(instance, &instance_data);
				if (err != CMAPL_SUCCESS){
					syslog(LOG_ERR, "\nCould not get instance data error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
					goto error_1;
				}

				// Set index to instance_count, this indicates the instance data it's parsing.
				(*wrapper_counter_data)[instance_count].index = instance_count;

				// iterate through all the counter data in this instance
				cmapl_counter_handle_t* ret_cnt;
				cmapl_counter_data_handle_t* cnt_data;
				err = cmapl_first_counter_data(instance, &ret_cnt, &cnt_data);
				if (err != CMAPL_SUCCESS){
					syslog(LOG_ERR, "\nCould not get first counter data error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
					goto error_1;
				}

				// Variable to hold number of counters per instance
				uint32_t counter_count = 0;

				//Iterate through counter data handle for a each instance handle
				while (cnt_data != NULL){
					// get counter info
					cmapl_counter_def_t cnt_info;
					err = cmapl_get_counter_info(ret_cnt, &cnt_info);
					if (err != CMAPL_SUCCESS){
						syslog(LOG_ERR, "\nCould not get counter info error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
						goto error_1;
					}


					// Add more counters headers to wrapper header value
					if (track_key_cnt == 0) {
						if(counter_headers_count == 0) {
							// Instantiate memory for wrapper headers
							*wrapper_headers = (counter_wrapper_headers *) malloc (sizeof(counter_wrapper_headers));
							if(*wrapper_counter_data == NULL){
								syslog(LOG_ERR, "\nCould not allocate memory for wrapper_header in file : %s", ccma_file_path);
								goto error_1;
							}
							(*wrapper_headers)[counter_headers_count].counter_name = strdup(cnt_info.name);
							(*wrapper_headers)[counter_headers_count].counter_id = cnt_info.cnt_id;
							counter_headers_count++;
							} else {
							// Re allocate memory if there are more counters.
							*wrapper_headers = (counter_wrapper_headers *) realloc (*wrapper_headers, (counter_headers_count + 1) *sizeof(counter_wrapper_headers));
							(*wrapper_headers)[counter_headers_count].counter_name = strdup(cnt_info.name);
							(*wrapper_headers)[counter_headers_count].counter_id = cnt_info.cnt_id;
							counter_headers_count++;
						}
					}

					// get counter data
					cmapl_counter_data_t cnt_data_value;
					err = cmapl_get_counter_data_value(cnt_data, &cnt_data_value);
					if (err != CMAPL_SUCCESS){
						syslog(LOG_ERR, "\nCould not get counter data value error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
						goto error_1;
					}

					// Allocate memory for the counter data type for a particular index
					if(counter_count == 0) {
						(*wrapper_counter_data)[instance_count].counter_values = (cmapl_counter_data_t *) malloc (sizeof(cmapl_counter_data_t));
					} else {// more counter data so, increase the memory allocation. We do this because we can never know the lenght beforehand through cmapl.
						(*wrapper_counter_data)[instance_count].counter_values = (cmapl_counter_data_t *) realloc ((*wrapper_counter_data)[instance_count].counter_values, (counter_count + 1) * sizeof(cmapl_counter_data_t));
					}

					// Check if allocation succeeded
					if((*wrapper_counter_data)[instance_count].counter_values == NULL){
						syslog(LOG_ERR, "\nCould not allocate memory for wrapper_counter_data's counter_values for file : %s", ccma_file_path);
						goto error_1;
					}

					uint32_t j;
					(*wrapper_counter_data)[instance_count].counter_values[counter_count].values = (cmapl_cnt_value_t * ) malloc (cnt_data_value.num_values * sizeof(cmapl_cnt_value_t));
					if((*wrapper_counter_data)[instance_count].counter_values[counter_count].values == NULL){
						syslog(LOG_ERR, "\nCould not allocate memory for wrapper_counter_data's counter_values values for file : %s", ccma_file_path);
						goto error_1;
					}

					(*wrapper_counter_data)[instance_count].counter_values[counter_count].value_type = cnt_data_value.value_type;
					(*wrapper_counter_data)[instance_count].counter_values[counter_count].num_values = cnt_data_value.num_values;
					for (j = 0; j < cnt_data_value.num_values; j++){
						switch(cnt_data_value.value_type){
							case CMAPL_CNT_4BYTES:
								(*wrapper_counter_data)[instance_count].counter_values[counter_count].values[j].val_4B =  cnt_data_value.values[j].val_4B;
								break;
							case CMAPL_CNT_8BYTES:
								(*wrapper_counter_data)[instance_count].counter_values[counter_count].values[j].val_8B =  cnt_data_value.values[j].val_8B;
								break;
							case CMAPL_CNT_STRING:
								(*wrapper_counter_data)[instance_count].counter_values[counter_count].values[j].val_str = strdup(cnt_data_value.values[j].val_str);
								break;
						}
					}

					// get next counter data in this instance
					err = cmapl_next_counter_data(instance, &ret_cnt, &cnt_data);
					if (err != CMAPL_SUCCESS){
						syslog(LOG_ERR, "\nCould not get next counter data error : %d for file : %s",err, ccma_file_path);
						goto error_1;
					}

					counter_count++;
				}

				err = cmapl_next_instance(timestamp, &instance);
				if (err != CMAPL_SUCCESS){
					syslog(LOG_ERR, "\nCould not get next instance error : %d for file : %s",err, ccma_file_path);
					goto error_1;
				}

				(*wrapper_counter_data)[instance_count].num_of_counters_count = counter_count;
				instance_count++;
				track_key_cnt++;
			}

			//Output param for the length of counter instances
			*counter_data_handle_count = instance_count;

			// break out as we found the timestamp we are looking for.
			break;
			// Iteration of instance ends
		}

		// Get the next timestamp in the query to see if it matches the timestamp(time) passed as input parameter.
		err = cmapl_next_timestamp(query, &timestamp);
		if (err == CMAPL_RECORD_EMPTY){
			err = CMAPL_SUCCESS;
		} else if (err != CMAPL_SUCCESS){
			syslog(LOG_ERR, "\nCould not get next timestamp error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
			goto error_1;
		}
	}

	*header_count = counter_headers_count;

	// close the archive file handle
	err = cmapl_close_data_file(file);
	if (err != CMAPL_SUCCESS){
		syslog(LOG_ERR, "Error closing file: %s, error : %s\n", ccma_file_path, cmapl_error_labels[err]);
		file = NULL;
		goto error_1;
	}

	//Close syslog
	closelog ();
	// Return 0 or 1 for success or failure, at this point wrapper_cnt_data is already populated with data
	return ret_value;

error_1:
    // Any error caught until this point should be propagated to the JNA interface
    if (err != CMAPL_SUCCESS) {
    	*error = (char *)malloc(strlen(cmapl_error_labels[err]) + 1);
    	strcpy(*error, cmapl_error_labels[err]);
    }

	// All information about the ccma data file still in memory will be de-allocated
	if(file != NULL) {
		err = cmapl_close_data_file(file);
		if (err != CMAPL_SUCCESS){
			syslog(LOG_ERR, "\nCould not close ccma data file handler, error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		}
		file = NULL;
	}

	// De-Allocation wrapper_counter_Data explicitly by application by calling free_cmapl_wrapper_counter_data

	closelog ();
	return CMAPL_WRAPPER_ERROR;
}

/**
 * De-allocate memory allocated for the counter data information. It needs the length calculated from the function get_counter_values_for_timestamp
 */
void free_cmapl_wrapper_counter_data (cmapl_wrapper_counter_data_t * counter_data_output, uint32_t instance_count, counter_wrapper_headers * wrapper_headers, uint32_t header_count, char * error) {
	uint32_t i = 0, j =0, k =0;
	if(counter_data_output != NULL) {
		for( i =0; counter_data_output[i].counter_values != NULL  && i < instance_count; i++) {
			for (j = 0; counter_data_output[i].counter_values[j].values != NULL && j < counter_data_output[i].num_of_counters_count; j++) {
				for (k = 0 ; k < counter_data_output[i].counter_values[j].num_values ; k++) {
					switch(counter_data_output[i].counter_values[j].value_type){
					case CMAPL_CNT_STRING:
						if(counter_data_output[i].counter_values[j].values[k].val_str != NULL){
							free((char *)counter_data_output[i].counter_values[j].values[k].val_str);
							counter_data_output[i].counter_values[j].values[k].val_str = NULL;
						}
						break;
					}
				}
				free(counter_data_output[i].counter_values[j].values);
				counter_data_output[i].counter_values[j].values = NULL;
			}
			free(counter_data_output[i].counter_values);
			counter_data_output[i].counter_values = NULL;
		}
		free(counter_data_output);
		counter_data_output = NULL;
	}

	// Also free counter headers
	if(wrapper_headers != NULL){
		free_counter_headers(wrapper_headers, header_count);
	}

	// Also de-allocate error message if there were any
	if(error != NULL) {
		free(error);
		error = NULL;
	}
}


/**
 *  De-allocate memory for the  list counter headers(metadata) data.
 *  It needs the length from the function call from get_counter_headers.
 */
void free_counter_headers(counter_wrapper_headers * counter_values_headers_array, uint32_t header_count) {
	uint32_t i;
	for(i = 0; i < (header_count) ; i++) {
		free (counter_values_headers_array[i].counter_name);
		counter_values_headers_array[i].counter_name = NULL;
	}
	free(counter_values_headers_array);
	counter_values_headers_array = NULL;
}
