/*
 * counter_periods.c
 *
 *  Created on: Dec 16, 2014
 *      Author: root
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <cmapl/cmapl_ext.h>
#include "cmapl_wrapper_header.h"

/**
 * Takes ccma archiver file, object and returns the List of timestamps values for that object and its length in num_timestamps
 * ccma_file_path = path to ccma file
 * current_obj = object for which we find timestamps (resource, workload_queue_dblade, ...)
 * time_period_list = out parameter which gives number of periods for an object specified.
 */
bool get_periods_list(const char* ccma_file_path, char* current_obj, uint64_t** time_period_list, uint32_t* num_periods, char **error) {
	// On UNIX to enable syslog, add to /etc/syslog.conf or /etc/rsyslog.conf
	// local1.*         /var/log/mylog
	// restart syslogd
	// /etc/init.d/syslog restart (or) /etc/init.d/rsyslog
	setlogmask (LOG_UPTO (LOG_NOTICE));
	openlog ("cmapl_wrapper_get_timestamps", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

	//Open ccma file to read
	cmapl_file_handle_t* file;
	const char* meta_path;
	const char* data_path;

	*time_period_list = NULL;
	*error = NULL; // error message to be propagated to JNA interface
	uint64_t* periods;

	// Get the cmapl file_handle
	cmapl_error_t err = cmapl_open_data_file(ccma_file_path, "r", &file, &meta_path, &data_path);
	if(err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "Error opening file: %s, error : %s\n", ccma_file_path, cmapl_error_labels[err]);
		file = NULL;
		goto error;
	}

	// get start and end time parameters to look into the ccma file which is required by the query
	uint64_t start_time = 0;
	err = cmapl_get_oldest_record_time(file, &start_time);
	if(err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "Couldn't get the oldest record time error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		goto error;
	}

	uint64_t end_time = 0;
	err = cmapl_get_newest_record_time(file, &end_time);
	if(err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "Couldn't get the newest record time error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		goto error;
	}

	// Get the header handle
	cmapl_header_handle_t* header;
	err = cmapl_get_header(file, &header);
	if(err != CMAPL_SUCCESS){
		syslog(LOG_ERR, "Couldn't get the header handle error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		goto error;
	}

	//get object handle for the input parameter : current_obj
	cmapl_object_handle_t* current_object;
	err = cmapl_find_object(header, current_obj, &current_object);
	if(err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "Couldn't get the object error : %s, object : %s for file : %s",cmapl_error_labels[err], current_obj, ccma_file_path);
		goto error;
	}

	// Get Object info.
	cmapl_object_def_t obj_info;
	if(current_object != NULL) {
		err = cmapl_get_object_info(current_object, &obj_info);
		if(err != CMAPL_SUCCESS) {
			syslog(LOG_ERR, "Couldn't get the object info error : %s, object : %s for file : %s",cmapl_error_labels[err], current_obj, ccma_file_path);
			goto error;
		}
	}

	//Iterate through all periods of an object
	uint32_t no_periods = 0;
	err = cmapl_get_object_periods(current_object, &periods, &no_periods);
	if(err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "Couldn't get the object periods error : %s, object : %s for file : %s",cmapl_error_labels[err], current_obj, ccma_file_path);
		goto error;
	}

	uint32_t i;
	*time_period_list = (uint64_t*)malloc(no_periods * sizeof(uint64_t));
	if(*time_period_list == NULL) {
		syslog(LOG_ERR, "Couldn't allocate memory for timestamp_list for file : %s", ccma_file_path);
		goto error;
	}


	for (i=0; i < no_periods; i++){
		// get the counter name of each counter at this period for this object
		(*time_period_list)[i] = periods[i];
	}

	*num_periods = no_periods;

	// free periods
	if(periods != NULL) {
		free(periods);
	}

	err = cmapl_close_data_file(file);
	if (err != CMAPL_SUCCESS){
		syslog(LOG_ERR, "Error closing file: %s, error : %s\n", ccma_file_path, cmapl_error_labels[err]);
		file = NULL;
		goto error;
	}
	closelog();
	return CMAPL_WRAPPER_SUCCESS;

//Cleanup memory and return error
error:
	// Any error caught until this point should be propagated to the JNA interface
    if (err != CMAPL_SUCCESS) { // For  other error scenarios, err will still be CMAPL_SUCCESS
    	*error = (char *)malloc(strlen(cmapl_error_labels[err]) + 1);
    	strcpy(*error, cmapl_error_labels[err]);
    }

	// All information about the ccma data file still in memory will be de-allocated
	if(file != NULL) {
		err = cmapl_close_data_file(file);
		if (err != CMAPL_SUCCESS){
			syslog(LOG_ERR, "\nCould not close ccma data file handler, error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		}
	    file = NULL;
	}

	// free periods
	if(periods != NULL) {
		free(periods);
	}

	//De-allocate timestamp_list is done explicitly by application calling free_cmapl_wrapper_timestamps_data routine.
	closelog();
	return CMAPL_WRAPPER_ERROR;
}



/**
 * De-allocate memory for the list of periods
 */
void free_cmapl_wrapper_periods_data(uint64_t * time_period_list, char * error) {
	if(time_period_list != NULL){
		free(time_period_list);
		time_period_list = NULL;
	}

	// Also de-allocate error message if there were any
	if(error != NULL) {
		free(error);
		error = NULL;
	}
}
