/*
 * timestamps_list.c
 *
 *  Created on: Feb 20, 2014
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <cmapl/cmapl_ext.h>
#include "cmapl_wrapper_header.h"
#define CM_MAX_COUNTERS_PER_OBJECT 512

/**
 * Takes ccma archiver file, object and returns the List of timestamps values for that object and its length in num_timestamps
 * ccma_file_path = path to ccma file
 * current_obj = object for which we find timestamps (resource, workload_queue_dblade, ...)
 * num_tiomestamps = out parameter which gives number of timestmaps returned by the query.
 */
bool get_timestamp_list(const char* ccma_file_path, uint64_t time_period, char* current_obj, uint64_t** timestamp_list, uint32_t* num_timestamps, char** error){

	// On UNIX to enable syslog, add to /etc/syslog.conf or /etc/rsyslog.conf
	// local1.*         /var/log/mylog
	// restart syslogd
	// /etc/init.d/syslog restart (or) /etc/init.d/rsyslog
	setlogmask (LOG_UPTO (LOG_NOTICE));
	openlog ("cmapl_wrapper_get_timestamps", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

	//Open ccma file to read
	cmapl_file_handle_t* file;
	const char* meta_path;
	const char* data_path;

	*timestamp_list = NULL;
	*error = NULL; // error message to be propagated to JNA interface

	// Get the cmapl file_handle
	cmapl_error_t err = cmapl_open_data_file(ccma_file_path, "r", &file, &meta_path, &data_path);
	if(err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "Error opening file: %s, error : %s\n", ccma_file_path, cmapl_error_labels[err]);
		file = NULL;
		goto error;
	}

	// get start and end time parameters to look into the ccma file which is required by the query
	uint64_t start_time = 0;
	err = cmapl_get_oldest_record_time(file, &start_time);
	if(err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "Couldn't get the oldest record time error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		goto error;
	}

	uint64_t end_time = 0;
	err = cmapl_get_newest_record_time(file, &end_time);
	if(err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "Couldn't get the newest record time error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		goto error;
	}

	// Get the header handle
	cmapl_header_handle_t* header;
	err = cmapl_get_header(file, &header);
	if(err != CMAPL_SUCCESS){
		syslog(LOG_ERR, "Couldn't get the header handle error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		goto error;
	}

	//get object handle for the input parameter : current_obj
	cmapl_object_handle_t* current_object;
	err = cmapl_find_object(header, current_obj, &current_object);
	if(err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "Couldn't get the object error : %s, object : %s for file : %s",cmapl_error_labels[err], current_obj, ccma_file_path);
		goto error;
	}

	// Get Object info.
	cmapl_object_def_t obj_info;
	if(current_object != NULL) {
		err = cmapl_get_object_info(current_object, &obj_info);
		if(err != CMAPL_SUCCESS) {
			syslog(LOG_ERR, "Couldn't get the object info error : %s, object : %s for file : %s",cmapl_error_labels[err], current_obj, ccma_file_path);
			goto error;
		}
	}

	uint32_t j=0;
	*timestamp_list = (uint64_t*)malloc(sizeof(uint64_t));
	if(*timestamp_list == NULL) {
		syslog(LOG_ERR, "Couldn't allocate memory for timestamp_list for file : %s", ccma_file_path);
		goto error;
	}

	uint32_t cnt_idx;
	const char* cnt_names[CM_MAX_COUNTERS_PER_OBJECT];
	// Get counter info needed for cmapl_query.
	// Info includes cnt_name (instance_uuid, instance_name, archiver_uuid_lo...) and no. of counters
	err = get_counter_info(current_object, time_period, cnt_names, &cnt_idx);
	if(err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "Couldn't get the counter information, error: %s, object : %s for file : %s",cmapl_error_labels[err], current_object, ccma_file_path);
		goto error;
	}

	// Define query parameters
	cmapl_query_def_t query_def;
	cmapl_set_latest_version(&query_def);
	query_def.object_name = obj_info.name;
	query_def.num_cnts = cnt_idx; // This needs to be optional, It is ONTAP burt and currently is workaround only.
	query_def.cnt_names = cnt_names; // This needs to be optional, It is ONTAP burt and currently is workaround only.
	query_def.start_time = start_time;
	query_def.end_time = end_time;
	query_def.period = time_period; // input param time_period
	query_def.num_key_cnts = 0;

	// Get query_handle for the ccma file
	cmapl_query_handle_t* query;
	err = cmapl_query(file, &query_def, &query);
	if(err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "Couldn't run the query error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		goto error;
	}

	//get timestamp handle and iterate through it until  we find all timestamps for an object.
	cmapl_timestamp_handle_t* timestamp;
	err = cmapl_first_timestamp(query, &timestamp);
	if(err != CMAPL_SUCCESS){
		syslog(LOG_ERR, "Couldn't get the timestamp error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		goto error;
	}

	while(timestamp != NULL){
		uint64_t time = 0;
		err = cmapl_get_timestamp(timestamp, &time);
		if(err != CMAPL_SUCCESS){
			syslog(LOG_ERR, "Couldn't get timestamp error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
			goto error;
		}
		*timestamp_list = realloc(*timestamp_list,(j+1)*sizeof(uint64_t));
		if(*timestamp_list == NULL) {
			syslog(LOG_ERR, "Couldn't re-allocate memory for timestamp_list for file : %s", ccma_file_path);
			goto error;
		}

		(*timestamp_list)[j] = time;
		err = cmapl_next_timestamp(query, &timestamp);
		if(err == CMAPL_RECORD_EMPTY) {
			err = CMAPL_SUCCESS;
		} else if (err != CMAPL_SUCCESS){
			syslog(LOG_ERR, "Couldn't get the next timestamp error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
			goto error;
		}
		j++;
	}

	*num_timestamps = j;
	err = cmapl_close_data_file(file);
	if (err != CMAPL_SUCCESS){
		syslog(LOG_ERR, "Error closing file: %s, error : %s\n", ccma_file_path, cmapl_error_labels[err]);
		file = NULL;
		goto error;
	}
	closelog();
	return CMAPL_WRAPPER_SUCCESS;

//Cleanup memory and return error
error:
	// Any error caught until this point should be propagated to the JNA interface
    if (err != CMAPL_SUCCESS) { // For  other error scenarios, err will still be CMAPL_SUCCESS
    	*error = (char *)malloc(strlen(cmapl_error_labels[err]) + 1);
    	strcpy(*error, cmapl_error_labels[err]);
    }

	// All information about the ccma data file still in memory will be de-allocated
	if(file != NULL) {
		err = cmapl_close_data_file(file);
		if (err != CMAPL_SUCCESS){
			syslog(LOG_ERR, "\nCould not close ccma data file handler, error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		}
	    file = NULL;
	}

	//De-allocate timestamp_list is done explicitly by application calling free_cmapl_wrapper_timestamps_data routine.

	closelog();
	return CMAPL_WRAPPER_ERROR;
}


#ifdef DEBUG_ONLY // Run this routine only in Debug mode, as this routine is for testing JNA translation.
/**
 * Print the timestamps values into a file
 * ccma_file_path = location of ccma file
 * output_file_path = location of the output file where the timestamps are written
 * current_obj = object for which timestamps are needed. (reource, workload_queue_dblade,....)
 */
bool write_timestamps_to_file(const char* ccma_file_path, char* output_file_path, uint64_t time_period, char* current_obj){
	uint32_t num_timestamps=0;
	uint64_t* timestamps;
	uint32_t i=0;

	setlogmask (LOG_UPTO (LOG_NOTICE));
	openlog ("cmapl_wrapper_get_counter_data", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

	char * error = NULL;

	bool return_val = get_timestamp_list(ccma_file_path, time_period, current_obj, &timestamps, &num_timestamps, &error);
	if(!return_val) {
		syslog (LOG_ERR, "get_timestamp_list routine failed to retrieve timestamps for file : %s", ccma_file_path);
		return CMAPL_WRAPPER_ERROR;
	}
	//open the file for writing.
	FILE *fp = fopen(output_file_path,"w");
	if( !fp ){
		syslog (LOG_ERR, "Could not open output file to dump the timestamp values for file : %s", ccma_file_path);
		free_cmapl_wrapper_timestamps_data(timestamps, error);
		closelog();
		return CMAPL_WRAPPER_ERROR;
	}
	for(i=0; i<num_timestamps; i++) {
		fprintf(fp, "Timestamp: %"PRIu64"\n",*(timestamps+i));
	}
	free_cmapl_wrapper_timestamps_data(timestamps, error);
	fclose(fp);
	closelog ();
	return CMAPL_WRAPPER_SUCCESS;
}
#endif // Run the above routine in Debug mode only.

/**
 * De-allocate memory for the list of timestamps
 */
void free_cmapl_wrapper_timestamps_data(uint64_t * timestamps, char * error) {
	if(timestamps != NULL){
		free(timestamps);
		timestamps = NULL;
	}

	// Also de-allocate error message if there were any
	if(error != NULL) {
		free(error);
		error = NULL;
	}
}
